diff -pruN 0.11+ds1-5/build/build.go 0.12+ds1-1/build/build.go
--- 0.11+ds1-5/build/build.go	2018-01-18 02:03:11.000000000 +0000
+++ 0.12+ds1-1/build/build.go	1970-01-01 00:00:00.000000000 +0000
@@ -1,11 +0,0 @@
-// Package build contains build information.
-//
-// Build information should be set during compilation by passing
-// -ldflags "-X github.com/elves/elvish/build.Var=value" to "go build" or
-// "go get".
-package build
-
-var (
-	Version = "unknown"
-	Builder = "unknown"
-)
diff -pruN 0.11+ds1-5/buildall.sh 0.12+ds1-1/buildall.sh
--- 0.11+ds1-5/buildall.sh	1970-01-01 00:00:00.000000000 +0000
+++ 0.12+ds1-1/buildall.sh	2018-07-11 02:32:25.000000000 +0000
@@ -0,0 +1,53 @@
+#!/bin/sh -e
+
+IFS=
+
+VERSION=`git describe --tags --always --dirty=-dirty`
+
+: ${BUILDER:=`id -un`@`hostname`}
+: ${VERSION_SUFFIX:=$VERSION}
+: ${BIN_DIR:=./bin}
+
+export GOOS GOARCH
+
+buildone() {
+    GOOS=$1
+    GOARCH=$2
+    STEM=elvish-$GOOS-$GOARCH-$VERSION_SUFFIX
+    if test $GOOS = windows; then
+        BIN=$STEM.exe
+        ARCHIVE=$STEM.zip
+    else
+        BIN=$STEM
+        ARCHIVE=$STEM.tar.gz
+    fi
+
+    echo "Going to build $BIN"
+    go build -o $BIN_DIR/$BIN -ldflags \
+        "-X github.com/elves/elvish/buildinfo.Version=$VERSION \
+         -X github.com/elves/elvish/buildinfo.GoRoot=`go env GOROOT` \
+         -X github.com/elves/elvish/buildinfo.GoPath=`go env GOPATH`"
+
+    (
+    cd $BIN_DIR
+    if test $GOOS = windows; then
+        zip $ARCHIVE $BIN
+    else
+        tar cfz $ARCHIVE $BIN
+    fi
+    )
+
+    echo "Built $BIN and archived $ARCHIVE"
+}
+
+build() {
+    local os=$1 arch=
+    shift
+    for arch in $@; do
+        buildone $os $arch
+    done
+}
+
+build darwin  amd64
+build windows amd64 386
+build linux   amd64 386 arm64
diff -pruN 0.11+ds1-5/build-and-upload.elv 0.12+ds1-1/build-and-upload.elv
--- 0.11+ds1-5/build-and-upload.elv	2018-01-18 02:03:11.000000000 +0000
+++ 0.12+ds1-1/build-and-upload.elv	1970-01-01 00:00:00.000000000 +0000
@@ -1,34 +0,0 @@
-if (eq $E:UPLOAD_TOKEN '') {
-    echo 'UPLOAD_TOKEN must be set'
-    exit 2
-}
-
-version = (git describe --tags --always)
-fname-suffix = ''
-if (not-eq $E:TRAVIS_TAG '') {
-    fname-suffix = -$E:TRAVIS_TAG
-}
-
-fn build [os arch]{
-    bin = elvish-$os'-'$arch$fname-suffix
-    archive = $bin.tar.gz
-    echo 'Going to build '$bin
-    E:GOOS=$os E:GOARCH=$arch go build -ldflags "-X main.Version="$version -o $bin
-    if (eq $os windows) {
-        archive = $bin.zip
-        cp $bin $bin.exe
-        zip $archive $bin.exe
-    } else {
-        tar cfz $archive $bin
-    }
-    curl https://ul.elvish.io/ -F name=$archive -F token=$E:UPLOAD_TOKEN -F file=@$archive
-    echo 'Built '$bin' and uploaded '$archive
-}
-
-build darwin amd64
-for arch [386 amd64] {
-    build windows $arch
-}
-for arch [386 amd64 arm64] {
-    build linux $arch
-}
diff -pruN 0.11+ds1-5/buildinfo/buildinfo.go 0.12+ds1-1/buildinfo/buildinfo.go
--- 0.11+ds1-5/buildinfo/buildinfo.go	1970-01-01 00:00:00.000000000 +0000
+++ 0.12+ds1-1/buildinfo/buildinfo.go	2018-07-11 02:32:25.000000000 +0000
@@ -0,0 +1,12 @@
+// Package buildinfo contains build information.
+//
+// Build information should be set during compilation by passing
+// -ldflags "-X github.com/elves/elvish/buildinfo.Var=value" to "go build" or
+// "go get".
+package buildinfo
+
+var (
+	Version = "unknown"
+	GoPath  = "unknown"
+	GoRoot  = "unknown"
+)
diff -pruN 0.11+ds1-5/daemon/api.go 0.12+ds1-1/daemon/api.go
--- 0.11+ds1-5/daemon/api.go	2018-01-18 02:03:11.000000000 +0000
+++ 0.12+ds1-1/daemon/api.go	2018-07-11 02:32:25.000000000 +0000
@@ -9,7 +9,7 @@ const (
 	ServiceName = "Daemon"
 
 	// Version is the API version. It should be bumped any time the API changes.
-	Version = -97
+	Version = -94
 )
 
 // Basic requests.
@@ -42,6 +42,13 @@ type AddCmdResponse struct {
 	Seq int
 }
 
+type DelCmdRequest struct {
+	Seq int
+}
+
+type DelCmdResponse struct {
+}
+
 type CmdRequest struct {
 	Seq int
 }
@@ -88,6 +95,12 @@ type AddDirRequest struct {
 
 type AddDirResponse struct{}
 
+type DelDirRequest struct {
+	Dir string
+}
+
+type DelDirResponse struct{}
+
 type DirsRequest struct {
 	Blacklist map[string]struct{}
 }
diff -pruN 0.11+ds1-5/daemon/client.go 0.12+ds1-1/daemon/client.go
--- 0.11+ds1-5/daemon/client.go	2018-01-18 02:03:11.000000000 +0000
+++ 0.12+ds1-1/daemon/client.go	2018-07-11 02:32:25.000000000 +0000
@@ -123,6 +123,13 @@ func (c *Client) AddCmd(text string) (in
 	return res.Seq, err
 }
 
+func (c *Client) DelCmd(seq int) error {
+	req := &DelCmdRequest{seq}
+	res := &DelCmdResponse{}
+	err := c.call("DelCmd", req, res)
+	return err
+}
+
 func (c *Client) Cmd(seq int) (string, error) {
 	req := &CmdRequest{seq}
 	res := &CmdResponse{}
@@ -158,6 +165,13 @@ func (c *Client) AddDir(dir string, incF
 	return err
 }
 
+func (c *Client) DelDir(dir string) error {
+	req := &DelDirRequest{dir}
+	res := &DelDirResponse{}
+	err := c.call("DelDir", req, res)
+	return err
+}
+
 func (c *Client) Dirs(blacklist map[string]struct{}) ([]storedefs.Dir, error) {
 	req := &DirsRequest{blacklist}
 	res := &DirsResponse{}
diff -pruN 0.11+ds1-5/daemon/daemon_test.go 0.12+ds1-1/daemon/daemon_test.go
--- 0.11+ds1-5/daemon/daemon_test.go	2018-01-18 02:03:11.000000000 +0000
+++ 0.12+ds1-1/daemon/daemon_test.go	2018-07-11 02:32:25.000000000 +0000
@@ -16,16 +16,17 @@ func TestDaemon(t *testing.T) {
 		}()
 
 		client := NewClient("sock")
-		for i := 0; i < 10; i++ {
+		for i := 0; i < 100; i++ {
 			client.ResetConn()
 			_, err := client.Version()
 			if err == nil {
 				break
-			} else if i == 9 {
-				t.Fatal("Failed to connect after 100ms")
+			} else if i == 99 {
+				t.Fatal("Failed to connect after 1s")
 			}
 			time.Sleep(10 * time.Millisecond)
 		}
+
 		_, err := client.AddCmd("test cmd")
 		if err != nil {
 			t.Errorf("client.AddCmd -> error %v", err)
diff -pruN 0.11+ds1-5/daemon/service.go 0.12+ds1-1/daemon/service.go
--- 0.11+ds1-5/daemon/service.go	2018-01-18 02:03:11.000000000 +0000
+++ 0.12+ds1-1/daemon/service.go	2018-07-11 02:32:25.000000000 +0000
@@ -132,6 +132,14 @@ func (s *Service) AddCmd(req *AddCmdRequ
 	return err
 }
 
+func (s *Service) DelCmd(req *DelCmdRequest, res *DelCmdResponse) error {
+	if s.err != nil {
+		return s.err
+	}
+	err := s.store.DelCmd(req.Seq)
+	return err
+}
+
 func (s *Service) Cmd(req *CmdRequest, res *CmdResponse) error {
 	if s.err != nil {
 		return s.err
@@ -175,6 +183,13 @@ func (s *Service) AddDir(req *AddDirRequ
 	return s.store.AddDir(req.Dir, req.IncFactor)
 }
 
+func (s *Service) DelDir(req *DelDirRequest, res *DelDirResponse) error {
+	if s.err != nil {
+		return s.err
+	}
+	return s.store.DelDir(req.Dir)
+}
+
 func (s *Service) Dirs(req *DirsRequest, res *DirsResponse) error {
 	if s.err != nil {
 		return s.err
diff -pruN 0.11+ds1-5/daemon/sock_windows.go 0.12+ds1-1/daemon/sock_windows.go
--- 0.11+ds1-5/daemon/sock_windows.go	2018-01-18 02:03:11.000000000 +0000
+++ 0.12+ds1-1/daemon/sock_windows.go	2018-07-11 02:32:25.000000000 +0000
@@ -1,15 +1,12 @@
 package daemon
 
 import (
-	"errors"
 	"fmt"
 	"io/ioutil"
 	"net"
 	"os"
 )
 
-var errSockExists = errors.New("socket file already exists")
-
 func listen(path string) (net.Listener, error) {
 	file, err := os.OpenFile(path, os.O_CREATE|os.O_WRONLY, 0006)
 	if err != nil {
diff -pruN 0.11+ds1-5/debian/changelog 0.12+ds1-1/debian/changelog
--- 0.11+ds1-5/debian/changelog	2018-03-02 09:29:22.000000000 +0000
+++ 0.12+ds1-1/debian/changelog	2018-07-11 02:37:51.000000000 +0000
@@ -1,3 +1,25 @@
+elvish (0.12+ds1-1) unstable; urgency=medium
+
+  * New upstream version 0.12+ds1
+  * d/rules: make DATE reproducible
+  * d/rules: s/markdown_github/gfm in pandoc arg
+
+ -- Shengjing Zhu <i@zhsj.me>  Wed, 11 Jul 2018 10:37:51 +0800
+
+elvish (0.12~rc1+ds1-1) experimental; urgency=medium
+
+  * New upstream version 0.12~rc1+ds1
+  * d/control: update Standards-Version to 4.1.4 (no changes)
+  * d/patches: drop merged patches
+  * d/control: bump golang-github-xiaq-persistent-dev to 20180301
+  * d/rules: update ldflags for buildinfo
+  * d/rules: add website to DH_GOLANG_EXCLUDES
+  * d/rules: disable editor e2e test
+  * d/elvish.1: update manpage
+  * Use pandoc to generate reference manpage.
+
+ -- Shengjing Zhu <i@zhsj.me>  Wed, 27 Jun 2018 17:10:04 +0800
+
 elvish (0.11+ds1-5) unstable; urgency=medium
 
   * d/control: stop using gccgo-go on ppc64el/ppc64.
diff -pruN 0.11+ds1-5/debian/control 0.12+ds1-1/debian/control
--- 0.11+ds1-5/debian/control	2018-03-02 09:29:22.000000000 +0000
+++ 0.12+ds1-1/debian/control	2018-07-11 02:37:51.000000000 +0000
@@ -8,9 +8,10 @@ Build-Depends: debhelper (>= 11),
                golang-any,
                golang-github-boltdb-bolt-dev,
                golang-github-kr-pty-dev,
-               golang-github-xiaq-persistent-dev (>=0.0~git20180105.221501f),
+               golang-github-xiaq-persistent-dev (>=0.0~git20180301.cd415c6),
                golang-golang-x-sys-dev,
-Standards-Version: 4.1.3
+               pandoc,
+Standards-Version: 4.1.4
 Homepage: https://github.com/elves/elvish
 Vcs-Browser: https://salsa.debian.org/go-team/packages/elvish
 Vcs-Git: https://salsa.debian.org/go-team/packages/elvish.git
diff -pruN 0.11+ds1-5/debian/elvish.1 0.12+ds1-1/debian/elvish.1
--- 0.11+ds1-5/debian/elvish.1	2018-03-02 09:29:22.000000000 +0000
+++ 0.12+ds1-1/debian/elvish.1	2018-07-11 02:37:51.000000000 +0000
@@ -1,4 +1,4 @@
-.TH elvish 1 "Jan 16, 2018" "elvish 0.11"
+.TH elvish 1 "June 27, 2018" "elvish 0.12"
 
 .SH NAME
 elvish - Friendly and expressive shell
@@ -47,9 +47,6 @@ the port of the web backend (default 317
 -sock string
 path to the daemon socket
 .TP
--version
-show version and quit
-.TP
 -web
 run backend of web interface
 
diff -pruN 0.11+ds1-5/debian/elvish.manpages 0.12+ds1-1/debian/elvish.manpages
--- 0.11+ds1-5/debian/elvish.manpages	2018-03-02 09:29:22.000000000 +0000
+++ 0.12+ds1-1/debian/elvish.manpages	2018-07-11 02:37:51.000000000 +0000
@@ -1 +1,8 @@
+debian/elvish-builtin.7
+debian/elvish-bundled.7
+debian/elvish-edit.7
+debian/elvish-epm.7
+debian/elvish-language.7
+debian/elvish-philosophy.7
+debian/elvish-re.7
 debian/elvish.1
diff -pruN 0.11+ds1-5/debian/patches/0001-sys-Work-around-nil-references-in-Select.patch 0.12+ds1-1/debian/patches/0001-sys-Work-around-nil-references-in-Select.patch
--- 0.11+ds1-5/debian/patches/0001-sys-Work-around-nil-references-in-Select.patch	2018-03-02 09:29:22.000000000 +0000
+++ 0.12+ds1-1/debian/patches/0001-sys-Work-around-nil-references-in-Select.patch	1970-01-01 00:00:00.000000000 +0000
@@ -1,169 +0,0 @@
-From: Qi Xiao <xiaqqaix@gmail.com>
-Date: Thu, 1 Mar 2018 13:53:12 -0500
-Subject: sys: Work around nil references in Select.
-
-Origin: backport, https://github.com/elves/elvish/commit/9865e3c0c8ecea688be73230b335a9ade2b2d483
-Bug: https://github.com/elves/elvish/issues/612
-Bug-Debian: #891822
----
- sys/fdset_freebsd.go    | 11 ++++++-----
- sys/fdset_notfreebsd.go | 11 ++++++-----
- sys/select_bsd.go       |  6 +++---
- sys/select_linux.go     |  9 ++++++---
- sys/select_test.go      | 23 ++++++++++++-----------
- sys/waitforread_unix.go |  2 +-
- 6 files changed, 34 insertions(+), 28 deletions(-)
-
-diff --git a/sys/fdset_freebsd.go b/sys/fdset_freebsd.go
-index dc9b705..6a7f879 100644
---- a/sys/fdset_freebsd.go
-+++ b/sys/fdset_freebsd.go
-@@ -8,15 +8,16 @@ package sys
- 
- import (
- 	"reflect"
--	"syscall"
-+
-+	"golang.org/x/sys/unix"
- )
- 
--var nFdBits = (uint)(reflect.TypeOf(syscall.FdSet{}.X__fds_bits[0]).Size() * 8)
-+var nFdBits = (uint)(reflect.TypeOf(unix.FdSet{}.X__fds_bits[0]).Size() * 8)
- 
--type FdSet syscall.FdSet
-+type FdSet unix.FdSet
- 
--func (fs *FdSet) s() *syscall.FdSet {
--	return (*syscall.FdSet)(fs)
-+func (fs *FdSet) s() *unix.FdSet {
-+	return (*unix.FdSet)(fs)
- }
- 
- func NewFdSet(fds ...int) *FdSet {
-diff --git a/sys/fdset_notfreebsd.go b/sys/fdset_notfreebsd.go
-index a226eed..7cb6ee7 100644
---- a/sys/fdset_notfreebsd.go
-+++ b/sys/fdset_notfreebsd.go
-@@ -3,16 +3,17 @@
- package sys
- 
- import (
--	"syscall"
- 	"unsafe"
-+
-+	"golang.org/x/sys/unix"
- )
- 
--var nFdBits = uint(8 * unsafe.Sizeof(syscall.FdSet{}.Bits[0]))
-+var nFdBits = uint(8 * unsafe.Sizeof(unix.FdSet{}.Bits[0]))
- 
--type FdSet syscall.FdSet
-+type FdSet unix.FdSet
- 
--func (fs *FdSet) s() *syscall.FdSet {
--	return (*syscall.FdSet)(fs)
-+func (fs *FdSet) s() *unix.FdSet {
-+	return (*unix.FdSet)(fs)
- }
- 
- func NewFdSet(fds ...int) *FdSet {
-diff --git a/sys/select_bsd.go b/sys/select_bsd.go
-index 0f5a509..bbd54cc 100644
---- a/sys/select_bsd.go
-+++ b/sys/select_bsd.go
-@@ -2,8 +2,8 @@
- 
- package sys
- 
--import "syscall"
-+import "golang.org/x/sys/unix"
- 
--func Select(nfd int, r *FdSet, w *FdSet, e *FdSet, timeout *syscall.Timeval) (err error) {
--	return syscall.Select(nfd, r.s(), w.s(), e.s(), timeout)
-+func Select(nfd int, r *FdSet, w *FdSet, e *FdSet) error {
-+	return unix.Select(nfd, r.s(), w.s(), e.s(), nil)
- }
-diff --git a/sys/select_linux.go b/sys/select_linux.go
-index 093c1aa..a30b03c 100644
---- a/sys/select_linux.go
-+++ b/sys/select_linux.go
-@@ -2,9 +2,12 @@
- 
- package sys
- 
--import "syscall"
-+import "golang.org/x/sys/unix"
- 
--func Select(nfd int, r *FdSet, w *FdSet, e *FdSet, timeout *syscall.Timeval) error {
--	_, err := syscall.Select(nfd, r.s(), w.s(), e.s(), timeout)
-+func Select(nfd int, r *FdSet, w *FdSet, e *FdSet) error {
-+	// On ARM64, MIPS64 and MIPS64LE, unix.Select is emulated in userland and
-+	// will dereference timeout. In that case, we use Pselect to work around the
-+	// problem. Bug: https://github.com/golang/go/issues/24189
-+	_, err := unix.Pselect(nfd, r.s(), w.s(), e.s(), nil, nil)
- 	return err
- }
-diff --git a/sys/select_test.go b/sys/select_test.go
-index 530bfab..4086ac3 100644
---- a/sys/select_test.go
-+++ b/sys/select_test.go
-@@ -3,8 +3,9 @@
- package sys
- 
- import (
--	"syscall"
- 	"testing"
-+
-+	"golang.org/x/sys/unix"
- )
- 
- func TestFdSet(t *testing.T) {
-@@ -30,8 +31,8 @@ func TestFdSet(t *testing.T) {
- 
- func TestSelect(t *testing.T) {
- 	var p1, p2 [2]int
--	mustNil(syscall.Pipe(p1[:]))
--	mustNil(syscall.Pipe(p2[:]))
-+	mustNil(unix.Pipe(p1[:]))
-+	mustNil(unix.Pipe(p2[:]))
- 	fs := NewFdSet(p1[0], p2[0])
- 	var maxfd int
- 	if p1[0] > p2[0] {
-@@ -40,16 +41,16 @@ func TestSelect(t *testing.T) {
- 		maxfd = p2[0] + 1
- 	}
- 	go func() {
--		syscall.Write(p1[1], []byte("to p1"))
--		syscall.Write(p2[1], []byte("to p2"))
--		syscall.Close(p1[1])
--		syscall.Close(p2[1])
-+		unix.Write(p1[1], []byte("to p1"))
-+		unix.Write(p2[1], []byte("to p2"))
-+		unix.Close(p1[1])
-+		unix.Close(p2[1])
- 	}()
--	e := Select(maxfd+1, fs, nil, nil, nil)
-+	e := Select(maxfd+1, fs, nil, nil)
- 	if e != nil {
--		t.Errorf("Select(%v, %v, nil, nil, nil) => %v, want <nil>",
-+		t.Errorf("Select(%v, %v, nil, nil) => %v, want <nil>",
- 			maxfd+1, fs, e)
- 	}
--	syscall.Close(p1[0])
--	syscall.Close(p2[0])
-+	unix.Close(p1[0])
-+	unix.Close(p2[0])
- }
-diff --git a/sys/waitforread_unix.go b/sys/waitforread_unix.go
-index b741663..51cb807 100644
---- a/sys/waitforread_unix.go
-+++ b/sys/waitforread_unix.go
-@@ -20,7 +20,7 @@ func WaitForRead(files ...*os.File) (ready []bool, err error) {
- 		}
- 		fdset.Set(fd)
- 	}
--	err = Select(maxfd+1, fdset, nil, nil, nil)
-+	err = Select(maxfd+1, fdset, nil, nil)
- 	ready = make([]bool, len(files))
- 	for i, file := range files {
- 		ready[i] = fdset.IsSet(int(file.Fd()))
diff -pruN 0.11+ds1-5/debian/patches/series 0.12+ds1-1/debian/patches/series
--- 0.11+ds1-5/debian/patches/series	2018-03-02 09:29:22.000000000 +0000
+++ 0.12+ds1-1/debian/patches/series	1970-01-01 00:00:00.000000000 +0000
@@ -1 +0,0 @@
-0001-sys-Work-around-nil-references-in-Select.patch
diff -pruN 0.11+ds1-5/debian/rules 0.12+ds1-1/debian/rules
--- 0.11+ds1-5/debian/rules	2018-03-02 09:29:22.000000000 +0000
+++ 0.12+ds1-1/debian/rules	2018-07-11 02:37:51.000000000 +0000
@@ -2,16 +2,29 @@
 
 include /usr/share/dpkg/default.mk
 
-BUILDER    := $(shell sed -ne 's,^Maintainer: .[^<]*<\([^>]*\)>,\1,p' debian/control)
-BUILDPKG   := github.com/elves/elvish/build
+BUILDPKG   := github.com/elves/elvish/buildinfo
 BUILDFLAGS := -ldflags\
-  "-X $(BUILDPKG).Version=$(DEB_VERSION)\
-   -X $(BUILDPKG).Builder=$(BUILDER)"
+  "-X $(BUILDPKG).Version=$(DEB_VERSION)"
+
+export DH_GOLANG_EXCLUDES := website
+
+DATE := $(shell date -u --date=@$(SOURCE_DATE_EPOCH) "+%b %d, %Y")
+VER := $(shell echo $(DEB_VERSION_UPSTREAM) | sed -e 's/+ds[0-9]*//')
+PANDOC_ARG := -s -f gfm -t man -V section:7 -V header:"Miscellaneous Information Manual" -V footer:"Elvish $(VER)" -M date:"$(DATE)"
 
 %:
 	dh $@ --buildsystem=golang --with=golang
 
+override_dh_auto_configure:
+	# e2e_test.go needs a working terminal, it fails in my sbuild env.
+	# https://github.com/elves/elvish/issues/708
+	-rm -f edit/edcore/e2e_test.go
+	dh_auto_configure
+
 override_dh_auto_build:
+	for file in builtin bundled edit epm language philosophy re; do \
+	  pandoc $(PANDOC_ARG) -M title:"elvish-$$file" -o debian/elvish-$$file.7 website/src/ref/$$file.md; \
+	done
 	dh_auto_build -- $(BUILDFLAGS)
 
 override_dh_auto_install:
diff -pruN 0.11+ds1-5/Dockerfile 0.12+ds1-1/Dockerfile
--- 0.11+ds1-5/Dockerfile	2018-01-18 02:03:11.000000000 +0000
+++ 0.12+ds1-1/Dockerfile	2018-07-11 02:32:25.000000000 +0000
@@ -1 +1,12 @@
-FROM golang:onbuild
+FROM golang:alpine
+
+RUN apk update && \
+    apk add --virtual build-deps make git
+
+COPY . /go/src/github.com/elves/elvish
+RUN make -C /go/src/github.com/elves/elvish get
+
+RUN apk del --purge build-deps
+
+WORKDIR /root
+CMD /go/bin/elvish
diff -pruN 0.11+ds1-5/.dockerignore 0.12+ds1-1/.dockerignore
--- 0.11+ds1-5/.dockerignore	1970-01-01 00:00:00.000000000 +0000
+++ 0.12+ds1-1/.dockerignore	2018-07-11 02:32:25.000000000 +0000
@@ -0,0 +1,2 @@
+/cover
+/website
diff -pruN 0.11+ds1-5/edit/abbr.go 0.12+ds1-1/edit/abbr.go
--- 0.11+ds1-5/edit/abbr.go	2018-01-18 02:03:11.000000000 +0000
+++ 0.12+ds1-1/edit/abbr.go	1970-01-01 00:00:00.000000000 +0000
@@ -1,31 +0,0 @@
-package edit
-
-import (
-	"github.com/elves/elvish/eval/types"
-	"github.com/elves/elvish/eval/vartypes"
-	"github.com/xiaq/persistent/hashmap"
-)
-
-var _ = RegisterVariable("abbr", func() vartypes.Variable {
-	return vartypes.NewValidatedPtr(
-		types.NewMap(hashmap.Empty), vartypes.ShouldBeMap)
-})
-
-func (ed *Editor) abbr() types.Map {
-	return ed.variables["abbr"].Get().(types.Map)
-}
-
-func (ed *Editor) abbrIterate(cb func(abbr, full string) bool) {
-	m := ed.abbr()
-	m.IteratePair(func(abbrValue, fullValue types.Value) bool {
-		abbr, ok := abbrValue.(types.String)
-		if !ok {
-			return true
-		}
-		full, ok := fullValue.(types.String)
-		if !ok {
-			return true
-		}
-		return cb(string(abbr), string(full))
-	})
-}
diff -pruN 0.11+ds1-5/edit/action.go 0.12+ds1-1/edit/action.go
--- 0.11+ds1-5/edit/action.go	2018-01-18 02:03:11.000000000 +0000
+++ 0.12+ds1-1/edit/action.go	1970-01-01 00:00:00.000000000 +0000
@@ -1,27 +0,0 @@
-package edit
-
-// action is used in the nextAction field of editorState to schedule a special
-// action after a keybinding has been executed.
-type action int
-
-const (
-	noAction action = iota
-	// reprocessKey makes the editor to reprocess the keybinding.
-	reprocessKey
-	// commitLine makes the editor return with the current line.
-	commitLine
-	// commitEOF makes the editor return with an EOF.
-	commitEOF
-)
-
-func (ed *Editor) setAction(action action) {
-	if ed.nextAction == noAction {
-		ed.nextAction = action
-	}
-}
-
-func (ed *Editor) popAction() action {
-	action := ed.nextAction
-	ed.nextAction = noAction
-	return action
-}
diff -pruN 0.11+ds1-5/edit/api.go 0.12+ds1-1/edit/api.go
--- 0.11+ds1-5/edit/api.go	2018-01-18 02:03:11.000000000 +0000
+++ 0.12+ds1-1/edit/api.go	1970-01-01 00:00:00.000000000 +0000
@@ -1,211 +0,0 @@
-package edit
-
-import (
-	"errors"
-	"strconv"
-	"unicode/utf8"
-	"unsafe"
-
-	"github.com/elves/elvish/edit/history"
-	"github.com/elves/elvish/edit/ui"
-	"github.com/elves/elvish/eval"
-	"github.com/elves/elvish/eval/types"
-	"github.com/elves/elvish/eval/vartypes"
-	"github.com/xiaq/persistent/hash"
-)
-
-// This file implements types and functions for interactions with the
-// Elvishscript runtime.
-
-var (
-	errNotNav             = errors.New("not in navigation mode")
-	errLineMustBeString   = errors.New("line must be string")
-	errDotMustBeString    = errors.New("dot must be string")
-	errDotMustBeInt       = errors.New("dot must be integer")
-	errDotOutOfRange      = errors.New("dot out of range")
-	errDotInsideCodepoint = errors.New("dot cannot be inside a codepoint")
-	errEditorInvalid      = errors.New("internal error: editor not set up")
-	errEditorInactive     = errors.New("editor inactive")
-)
-
-// BuiltinFn represents an editor builtin.
-type BuiltinFn struct {
-	name string
-	impl func(ed *Editor)
-}
-
-var _ eval.Fn = &BuiltinFn{}
-
-// Kind returns "fn".
-func (*BuiltinFn) Kind() string {
-	return "fn"
-}
-
-// Equal compares based on identity.
-func (bf *BuiltinFn) Equal(a interface{}) bool {
-	return bf == a
-}
-
-func (bf *BuiltinFn) Hash() uint32 {
-	return hash.Pointer(unsafe.Pointer(bf))
-}
-
-// Repr returns the representation of a builtin function as a variable name.
-func (bf *BuiltinFn) Repr(int) string {
-	return "$" + bf.name
-}
-
-// Call calls a builtin function.
-func (bf *BuiltinFn) Call(ec *eval.Frame, args []types.Value, opts map[string]types.Value) {
-	eval.TakeNoOpt(opts)
-	eval.TakeNoArg(args)
-	ed, ok := ec.Editor.(*Editor)
-	if !ok {
-		throw(errEditorInvalid)
-	}
-	if !ed.active {
-		throw(errEditorInactive)
-	}
-	bf.impl(ed)
-}
-
-// installModules installs edit: and edit:* modules.
-func installModules(builtin eval.Ns, ed *Editor) {
-	// Construct the edit: module, starting with builtins.
-	ns := makeNsFromBuiltins(builtinMaps[""])
-
-	// TODO(xiaq): Everything here should be registered to some registry instead
-	// of centralized here.
-
-	// Editor configurations.
-	for name, variable := range ed.variables {
-		ns[name] = variable
-	}
-
-	// Internal states.
-	ns["history"] = vartypes.NewRo(history.List{&ed.historyMutex, ed.daemon})
-	ns["current-command"] = vartypes.NewCallback(
-		func(v types.Value) error {
-			if !ed.active {
-				return errEditorInactive
-			}
-			if s, ok := v.(types.String); ok {
-				ed.buffer = string(s)
-				ed.dot = len(ed.buffer)
-			} else {
-				return errLineMustBeString
-			}
-			return nil
-		},
-		func() types.Value { return types.String(ed.buffer) },
-	)
-	ns["-dot"] = vartypes.NewCallback(
-		func(v types.Value) error {
-			s, ok := v.(types.String)
-			if !ok {
-				return errDotMustBeString
-			}
-			i, err := strconv.Atoi(string(s))
-			if err != nil {
-				if err.(*strconv.NumError).Err == strconv.ErrRange {
-					return errDotOutOfRange
-				} else {
-					return errDotMustBeInt
-				}
-			}
-			if i < 0 || i > len(ed.buffer) {
-				return errDotOutOfRange
-			}
-			if i < len(ed.buffer) {
-				r, _ := utf8.DecodeRuneInString(ed.buffer[i:])
-				if r == utf8.RuneError {
-					return errDotInsideCodepoint
-				}
-			}
-			ed.dot = i
-			return nil
-		},
-		func() types.Value { return types.String(strconv.Itoa(ed.dot)) },
-	)
-	ns["selected-file"] = vartypes.NewRoCallback(
-		func() types.Value {
-			if !ed.active {
-				throw(errEditorInactive)
-			}
-			nav, ok := ed.mode.(*navigation)
-			if !ok {
-				throw(errNotNav)
-			}
-			return types.String(nav.current.selectedName())
-		},
-	)
-
-	// Completers.
-	for _, bac := range argCompletersData {
-		ns[bac.name+eval.FnSuffix] = vartypes.NewRo(bac)
-	}
-
-	// Matchers.
-	eval.AddBuiltinFns(ns, matchers...)
-
-	// Functions.
-	eval.AddBuiltinFns(ns,
-		&eval.BuiltinFn{"edit:binding-table", makeBindingTable},
-		&eval.BuiltinFn{"edit:command-history", CommandHistory},
-		&eval.BuiltinFn{"edit:complete-getopt", complGetopt},
-		&eval.BuiltinFn{"edit:complex-candidate", outputComplexCandidate},
-		&eval.BuiltinFn{"edit:insert-at-dot", InsertAtDot},
-		&eval.BuiltinFn{"edit:replace-input", ReplaceInput},
-		&eval.BuiltinFn{"edit:styled", styled},
-		&eval.BuiltinFn{"edit:key", ui.KeyBuiltin},
-		&eval.BuiltinFn{"edit:wordify", Wordify},
-		&eval.BuiltinFn{"edit:-dump-buf", _dumpBuf},
-		&eval.BuiltinFn{"edit:-narrow-read", NarrowRead},
-	)
-
-	builtin["edit"+eval.NsSuffix] = vartypes.NewValidatedPtr(ns, eval.ShouldBeNs)
-	submods := make(map[string]eval.Ns)
-	// Install other modules.
-	for module, builtins := range builtinMaps {
-		if module != "" {
-			submods[module] = makeNsFromBuiltins(builtins)
-		}
-	}
-
-	// Add $edit:{mode}:binding variables.
-	for mode, bindingVar := range ed.bindings {
-		submod, ok := submods[mode]
-		if !ok {
-			submod = make(eval.Ns)
-			submods[mode] = submod
-		}
-		submod["binding"] = bindingVar
-	}
-
-	for name, ns := range submods {
-		builtin["edit:"+name+eval.NsSuffix] = vartypes.NewValidatedPtr(ns, eval.ShouldBeNs)
-	}
-}
-
-// CallFn calls an Fn, displaying its outputs and possible errors as editor
-// notifications. It is the preferred way to call a Fn while the editor is
-// active.
-func (ed *Editor) CallFn(fn eval.Fn, args ...types.Value) {
-	if b, ok := fn.(*BuiltinFn); ok {
-		// Builtin function: quick path.
-		b.impl(ed)
-		return
-	}
-
-	ports := []*eval.Port{
-		eval.DevNullClosedChan, ed.notifyPort, ed.notifyPort,
-	}
-	// XXX There is no source to pass to NewTopEvalCtx.
-	ec := eval.NewTopFrame(ed.evaler, eval.NewInternalSource("[editor]"), ports)
-	ex := ec.PCall(fn, args, eval.NoOpts)
-	if ex != nil {
-		ed.Notify("function error: %s", ex.Error())
-	}
-
-	ed.refresh(true, true)
-}
diff -pruN 0.11+ds1-5/edit/api_test.go 0.12+ds1-1/edit/api_test.go
--- 0.11+ds1-5/edit/api_test.go	2018-01-18 02:03:11.000000000 +0000
+++ 0.12+ds1-1/edit/api_test.go	1970-01-01 00:00:00.000000000 +0000
@@ -1,56 +0,0 @@
-package edit
-
-import (
-	"testing"
-
-	"github.com/elves/elvish/eval"
-	"github.com/elves/elvish/eval/types"
-	"github.com/elves/elvish/util"
-)
-
-func TestBuiltinFn(t *testing.T) {
-	called := false
-	builtinFn := &BuiltinFn{"foobar", func(*Editor) {
-		if called {
-			t.Errorf("builtin impl called multiple times, called not reset")
-		}
-		called = true
-	}}
-
-	if kind := builtinFn.Kind(); kind != "fn" {
-		t.Errorf("Kind of BuiltinFn should be fn, is %q", kind)
-	}
-	if repr := builtinFn.Repr(10); repr != "$foobar" {
-		t.Errorf("Repr of BuiltinFn should be $foobar, is %q", repr)
-	}
-
-	ec := &eval.Frame{Evaler: &eval.Evaler{}}
-
-	if !util.Throws(func() { builtinFn.Call(ec, nil, nil) }, errEditorInvalid) {
-		t.Errorf("BuiltinFn should error when Editor is nil, didn't")
-	}
-
-	ec.Editor = &Editor{active: false}
-	if !util.Throws(func() { builtinFn.Call(ec, nil, nil) }, errEditorInactive) {
-		t.Errorf("BuiltinFn should error when Editor is inactive, didn't")
-	}
-
-	ec.Editor = &Editor{active: true}
-
-	if !util.Throws(func() {
-		builtinFn.Call(ec, []types.Value{types.String("2")}, nil)
-	}, eval.ErrNoArgAccepted) {
-		t.Errorf("BuiltinFn should error when argument was supplied, didn't")
-	}
-
-	if !util.Throws(func() {
-		builtinFn.Call(ec, nil, map[string]types.Value{"a": types.String("b")})
-	}, eval.ErrNoOptAccepted) {
-		t.Errorf("BuiltinFn should error when option was supplied, didn't")
-	}
-
-	builtinFn.Call(ec, nil, nil)
-	if !called {
-		t.Errorf("BuiltinFn should call its implementation, didn't")
-	}
-}
diff -pruN 0.11+ds1-5/edit/api_utils.go 0.12+ds1-1/edit/api_utils.go
--- 0.11+ds1-5/edit/api_utils.go	2018-01-18 02:03:11.000000000 +0000
+++ 0.12+ds1-1/edit/api_utils.go	1970-01-01 00:00:00.000000000 +0000
@@ -1,23 +0,0 @@
-package edit
-
-// Trivial utilities for the elvishscript API.
-
-import (
-	"fmt"
-
-	"github.com/elves/elvish/util"
-)
-
-func throw(e error) {
-	util.Throw(e)
-}
-
-func maybeThrow(e error) {
-	if e != nil {
-		util.Throw(e)
-	}
-}
-
-func throwf(format string, args ...interface{}) {
-	util.Throw(fmt.Errorf(format, args...))
-}
diff -pruN 0.11+ds1-5/edit/arg_completers.go 0.12+ds1-1/edit/arg_completers.go
--- 0.11+ds1-5/edit/arg_completers.go	2018-01-18 02:03:11.000000000 +0000
+++ 0.12+ds1-1/edit/arg_completers.go	1970-01-01 00:00:00.000000000 +0000
@@ -1,244 +0,0 @@
-package edit
-
-import (
-	"bufio"
-	"errors"
-	"io"
-	"os"
-	"strings"
-	"unsafe"
-
-	"github.com/elves/elvish/eval"
-	"github.com/elves/elvish/eval/types"
-	"github.com/elves/elvish/eval/vartypes"
-	"github.com/xiaq/persistent/hash"
-	"github.com/xiaq/persistent/hashmap"
-)
-
-// For an overview of completion, see the comment in completers.go.
-//
-// When completing arguments of commands (as opposed to variable names, map
-// indicies, etc.), the list of candidates often depends on the command in
-// question; e.g. "ls <Tab>" and "apt <Tab>" should yield different results
-// because the commands "ls" and "apt" accept different arguments. To reflec
-// this, Elvish has a map of "argument completers", with the key being the name
-// of the command, and the value being the argument completer itself, accessible
-// to script as $edit:arg-completer. The one entry with an empty string as the
-// key is the fallback completer, and is used when an argument completer for the
-// current command has not been defined.
-//
-// When completing an argument, Elvish first finds out the name of the command
-// (e.g. "ls" or "apt") can tries to evaluate its arguments. It then calls the
-// suitable completer with the name of the command and the arguments. The
-// arguments are in evaluated forms: e.g. if an argument is 'foo' (with quotes),
-// the argument is its value foo, not a literal 'foo'. The last argument is what
-// needs to be completed; if the user is starting a new argument, e.g. by typing
-// "ls a " (without quotes), the last argument passed to the argument completer
-// will be an empty string.
-//
-// The argument completer should then return a list of what can replace the last
-// argument. The results are of type rawCandidate, which basically means that
-// argument completers do not need to worry about quoting of candidates; the raw
-// candidates will be "cooked" into actual candidates that appear in the
-// interface, which includes quoting.
-//
-// There are a few finer points in this process:
-//
-// 1. If some of the arguments cannot be evaluated statically (for instance,
-//    consider this: echo (uname)), it will be an empty string. There needs
-//    probably be a better way to distinguish empty arguments and unknown
-//    arguments, but normally there is not much argument completers can do for
-//    unknown arguments.
-//
-// 2. The argument completer normally **should not** perform filtering. For
-//    instance, if the user has typed "ls x", the argument completer for "ls"
-//    should return **all** files, not just those whose names start with x. This
-//    is to make it possible for user to specify a different matching algorithm
-//    than the default prefix matching.
-//
-//    However, argument completers **should** look at the argument to decide
-//    which **type** of candidates to generate. For instance, if the user has
-//    typed "ls --x", the argument completer should generate all long options
-//    for "ls", but not only those starting with "x".
-
-var (
-	// ErrCompleterMustBeFn is thrown if the user has put a non-function entry
-	// in $edit:completer, and that entry needs to be used for completion.
-	// TODO(xiaq): Detect the type violation when the user modifies
-	// $edit:completer.
-	ErrCompleterMustBeFn = errors.New("completer must be fn")
-	// ErrCompleterArgMustBeString is thrown when a builtin argument completer
-	// is called with non-string arguments.
-	ErrCompleterArgMustBeString = errors.New("arguments to arg completers must be string")
-	// ErrTooFewArguments is thrown when a builtin argument completer is called
-	// with too few arguments.
-	ErrTooFewArguments = errors.New("too few arguments")
-)
-
-var (
-	argCompletersData = map[string]*builtinArgCompleter{
-		"":     {"complete-filename", complFilename},
-		"sudo": {"complete-sudo", complSudo},
-	}
-)
-
-var _ = RegisterVariable("arg-completer", argCompleterVariable)
-
-func argCompleterVariable() vartypes.Variable {
-	m := hashmap.Empty
-	for k, v := range argCompletersData {
-		m = m.Assoc(types.String(k), v)
-	}
-	return vartypes.NewValidatedPtr(types.NewMap(m), vartypes.ShouldBeMap)
-}
-
-func (ed *Editor) argCompleter() types.Map {
-	return ed.variables["arg-completer"].Get().(types.Map)
-}
-
-// completeArg calls the correct argument completers according to the command
-// name. It is used by complArg and can also be useful when further dispatching
-// based on command name is needed -- e.g. in the argument completer for "sudo".
-func completeArg(words []string, ev *eval.Evaler, rawCands chan<- rawCandidate) error {
-	logger.Printf("completing argument: %q", words)
-	// XXX(xiaq): not the best way to get argCompleter.
-	m := ev.Editor.(*Editor).argCompleter()
-	var v types.Value
-	if m.HasKey(types.String(words[0])) {
-		v = m.IndexOne(types.String(words[0]))
-	} else {
-		v = m.IndexOne(types.String(""))
-	}
-	fn, ok := v.(eval.Fn)
-	if !ok {
-		return ErrCompleterMustBeFn
-	}
-	return callArgCompleter(fn, ev, words, rawCands)
-}
-
-type builtinArgCompleter struct {
-	name string
-	impl func([]string, *eval.Evaler, chan<- rawCandidate) error
-}
-
-var _ eval.Fn = &builtinArgCompleter{}
-
-func (bac *builtinArgCompleter) Kind() string {
-	return "fn"
-}
-
-// Equal compares by identity.
-func (bac *builtinArgCompleter) Equal(a interface{}) bool {
-	return bac == a
-}
-
-func (bac *builtinArgCompleter) Hash() uint32 {
-	return hash.Pointer(unsafe.Pointer(bac))
-}
-
-func (bac *builtinArgCompleter) Repr(int) string {
-	return "$edit:" + bac.name + eval.FnSuffix
-}
-
-func (bac *builtinArgCompleter) Call(ec *eval.Frame, args []types.Value, opts map[string]types.Value) {
-	eval.TakeNoOpt(opts)
-	words := make([]string, len(args))
-	for i, arg := range args {
-		s, ok := arg.(types.String)
-		if !ok {
-			throw(ErrCompleterArgMustBeString)
-		}
-		words[i] = string(s)
-	}
-
-	rawCands := make(chan rawCandidate)
-	var err error
-	go func() {
-		defer close(rawCands)
-		err = bac.impl(words, ec.Evaler, rawCands)
-	}()
-
-	output := ec.OutputChan()
-	for rc := range rawCands {
-		output <- rc
-	}
-	maybeThrow(err)
-}
-
-func complFilename(words []string, ev *eval.Evaler, rawCands chan<- rawCandidate) error {
-	if len(words) < 1 {
-		return ErrTooFewArguments
-	}
-	return complFilenameInner(words[len(words)-1], false, rawCands)
-}
-
-func complSudo(words []string, ev *eval.Evaler, rawCands chan<- rawCandidate) error {
-	if len(words) < 2 {
-		return ErrTooFewArguments
-	}
-	if len(words) == 2 {
-		return complFormHeadInner(words[1], ev, rawCands)
-	}
-	return completeArg(words[1:], ev, rawCands)
-}
-
-// callArgCompleter calls a Fn, assuming that it is an arg completer. It calls
-// the Fn with specified arguments and closed input, and converts its output to
-// candidate objects.
-func callArgCompleter(fn eval.Fn,
-	ev *eval.Evaler, words []string, rawCands chan<- rawCandidate) error {
-
-	// Quick path for builtin arg completers.
-	if builtin, ok := fn.(*builtinArgCompleter); ok {
-		return builtin.impl(words, ev, rawCands)
-	}
-
-	args := make([]types.Value, len(words))
-	for i, word := range words {
-		args[i] = types.String(word)
-	}
-
-	ports := []*eval.Port{
-		eval.DevNullClosedChan,
-		{}, // Will be replaced when capturing output
-		{File: os.Stderr},
-	}
-
-	valuesCb := func(ch <-chan types.Value) {
-		for v := range ch {
-			switch v := v.(type) {
-			case rawCandidate:
-				rawCands <- v
-			case types.String:
-				rawCands <- plainCandidate(v)
-			default:
-				logger.Printf("completer must output string or candidate")
-			}
-		}
-	}
-
-	bytesCb := func(r *os.File) {
-		buffered := bufio.NewReader(r)
-		for {
-			line, err := buffered.ReadString('\n')
-			if line != "" {
-				rawCands <- plainCandidate(strings.TrimSuffix(line, "\n"))
-			}
-			if err != nil {
-				if err != io.EOF {
-					logger.Println("error on reading:", err)
-				}
-				break
-			}
-		}
-	}
-
-	// XXX There is no source to pass to NewTopEvalCtx.
-	ec := eval.NewTopFrame(ev, eval.NewInternalSource("[editor completer]"), ports)
-	err := ec.PCaptureOutputInner(fn, args, eval.NoOpts, valuesCb, bytesCb)
-	if err != nil {
-		err = errors.New("completer error: " + err.Error())
-	}
-
-	return err
-}
diff -pruN 0.11+ds1-5/edit/binding_table.go 0.12+ds1-1/edit/binding_table.go
--- 0.11+ds1-5/edit/binding_table.go	2018-01-18 02:03:11.000000000 +0000
+++ 0.12+ds1-1/edit/binding_table.go	1970-01-01 00:00:00.000000000 +0000
@@ -1,91 +0,0 @@
-package edit
-
-import (
-	"errors"
-	"sort"
-
-	"github.com/elves/elvish/edit/ui"
-	"github.com/elves/elvish/eval"
-	"github.com/elves/elvish/eval/types"
-	"github.com/elves/elvish/eval/vartypes"
-	"github.com/elves/elvish/parse"
-)
-
-var errValueShouldBeFn = errors.New("value should be function")
-
-func getBinding(bindingVar vartypes.Variable, k ui.Key) eval.Fn {
-	binding := bindingVar.Get().(BindingTable)
-	switch {
-	case binding.HasKey(k):
-		return binding.IndexOne(k).(eval.Fn)
-	case binding.HasKey(ui.Default):
-		return binding.IndexOne(ui.Default).(eval.Fn)
-	default:
-		return nil
-	}
-}
-
-// BindingTable is a special Map that converts its key to ui.Key and ensures
-// that its values satisfy eval.CallableValue.
-type BindingTable struct {
-	types.Map
-}
-
-// Repr returns the representation of the binding table as if it were an
-// ordinary map keyed by strings.
-func (bt BindingTable) Repr(indent int) string {
-	var builder types.MapReprBuilder
-	builder.Indent = indent
-
-	var keys ui.Keys
-	bt.Map.IterateKey(func(k types.Value) bool {
-		keys = append(keys, k.(ui.Key))
-		return true
-	})
-	sort.Sort(keys)
-
-	for _, k := range keys {
-		v := bt.Map.IndexOne(k)
-		builder.WritePair(parse.Quote(k.String()), indent+2, v.Repr(indent+2))
-	}
-
-	return builder.String()
-}
-
-// IndexOne converts the index to ui.Key and uses the IndexOne of the inner Map.
-func (bt BindingTable) IndexOne(idx types.Value) types.Value {
-	return bt.Map.IndexOne(ui.ToKey(idx))
-}
-
-func (bt BindingTable) get(k ui.Key) eval.Fn {
-	return bt.Map.IndexOne(k).(eval.Fn)
-}
-
-// Assoc converts the index to ui.Key, ensures that the value is CallableValue,
-// uses the Assoc of the inner Map and converts the result to a BindingTable.
-func (bt BindingTable) Assoc(k, v types.Value) types.Value {
-	key := ui.ToKey(k)
-	f, ok := v.(eval.Fn)
-	if !ok {
-		throw(errValueShouldBeFn)
-	}
-	return BindingTable{bt.Map.Assoc(key, f).(types.Map)}
-}
-
-func makeBindingTable(f *eval.Frame, args []types.Value, opts map[string]types.Value) {
-	var raw types.Map
-	eval.ScanArgs(args, &raw)
-	eval.TakeNoOpt(opts)
-
-	converted := types.EmptyMap
-	raw.IteratePair(func(k, v types.Value) bool {
-		f, ok := v.(eval.Fn)
-		if !ok {
-			throw(errValueShouldBeFn)
-		}
-		converted = converted.Assoc(ui.ToKey(k), f).(types.Map)
-		return true
-	})
-
-	f.OutputChan() <- BindingTable{converted}
-}
diff -pruN 0.11+ds1-5/edit/candidate.go 0.12+ds1-1/edit/candidate.go
--- 0.11+ds1-5/edit/candidate.go	2018-01-18 02:03:11.000000000 +0000
+++ 0.12+ds1-1/edit/candidate.go	1970-01-01 00:00:00.000000000 +0000
@@ -1,163 +0,0 @@
-package edit
-
-import (
-	"fmt"
-	"os"
-	"sort"
-
-	"github.com/elves/elvish/edit/ui"
-	"github.com/elves/elvish/eval"
-	"github.com/elves/elvish/eval/types"
-	"github.com/elves/elvish/parse"
-	"github.com/xiaq/persistent/hash"
-)
-
-type candidate struct {
-	code string    // This is what will be substituted on the command line.
-	menu ui.Styled // This is what is displayed in the completion menu.
-}
-
-// rawCandidate is what can be converted to a candidate.
-type rawCandidate interface {
-	types.Value
-	text() string
-	cook(q parse.PrimaryType) *candidate
-}
-
-type rawCandidates []rawCandidate
-
-func (cs rawCandidates) Len() int           { return len(cs) }
-func (cs rawCandidates) Swap(i, j int)      { cs[i], cs[j] = cs[j], cs[i] }
-func (cs rawCandidates) Less(i, j int) bool { return cs[i].text() < cs[j].text() }
-
-// plainCandidate is a minimal implementation of rawCandidate.
-type plainCandidate types.String
-
-func (plainCandidate) Kind() string               { return "string" }
-func (p plainCandidate) Equal(a interface{}) bool { return p == a }
-func (p plainCandidate) Hash() uint32             { return hash.String(string(p)) }
-func (p plainCandidate) Repr(l int) string        { return types.String(p).Repr(l) }
-func (p plainCandidate) text() string             { return string(p) }
-
-func (p plainCandidate) cook(q parse.PrimaryType) *candidate {
-	s := string(p)
-	quoted, _ := parse.QuoteAs(s, q)
-	return &candidate{code: quoted, menu: ui.Unstyled(s)}
-}
-
-// noQuoteCandidate is a rawCandidate that does not quote when cooked.
-type noQuoteCandidate string
-
-func (noQuoteCandidate) Kind() string                { return "string" }
-func (nq noQuoteCandidate) Equal(a interface{}) bool { return nq == a }
-func (nq noQuoteCandidate) Hash() uint32             { return hash.String(string(nq)) }
-func (nq noQuoteCandidate) Repr(l int) string        { return types.String(nq).Repr(l) }
-func (nq noQuoteCandidate) text() string             { return string(nq) }
-
-func (nq noQuoteCandidate) cook(parse.PrimaryType) *candidate {
-	s := string(nq)
-	return &candidate{code: s, menu: ui.Unstyled(s)}
-}
-
-// complexCandidate is an implementation of rawCandidate that offers
-// customization options.
-type complexCandidate struct {
-	stem          string    // Used in the code and the menu.
-	codeSuffix    string    // Appended to the code.
-	displaySuffix string    // Appended to the display.
-	style         ui.Styles // Used in the menu.
-}
-
-func (c *complexCandidate) Kind() string { return "map" }
-
-func (c *complexCandidate) Equal(a interface{}) bool {
-	rhs, ok := a.(*complexCandidate)
-	return ok && c.stem == rhs.stem && c.codeSuffix == rhs.codeSuffix && c.displaySuffix == rhs.displaySuffix && c.style.Eq(rhs.style)
-}
-
-func (c *complexCandidate) Hash() uint32 {
-	h := hash.DJBInit
-	h = hash.DJBCombine(h, hash.String(c.stem))
-	h = hash.DJBCombine(h, hash.String(c.codeSuffix))
-	h = hash.DJBCombine(h, hash.String(c.displaySuffix))
-	h = hash.DJBCombine(h, c.style.Hash())
-	return h
-}
-
-func (c *complexCandidate) Repr(indent int) string {
-	// TODO(xiaq): Pretty-print when indent >= 0
-	return fmt.Sprintf("(edit:complex-candidate %s &code-suffix=%s &display-suffix=%s style=%s)",
-		parse.Quote(c.stem), parse.Quote(c.codeSuffix),
-		parse.Quote(c.displaySuffix), parse.Quote(c.style.String()))
-}
-
-func (c *complexCandidate) text() string { return c.stem }
-
-func (c *complexCandidate) cook(q parse.PrimaryType) *candidate {
-	quoted, _ := parse.QuoteAs(c.stem, q)
-	return &candidate{
-		code: quoted + c.codeSuffix,
-		menu: ui.Styled{c.stem + c.displaySuffix, c.style},
-	}
-}
-
-// outputComplexCandidate composes a complexCandidate.
-func outputComplexCandidate(ec *eval.Frame,
-	args []types.Value, opts map[string]types.Value) {
-
-	var style string
-	c := &complexCandidate{}
-
-	eval.ScanArgs(args, &c.stem)
-	eval.ScanOpts(opts,
-		eval.OptToScan{"code-suffix", &c.codeSuffix, types.String("")},
-		eval.OptToScan{"display-suffix", &c.displaySuffix, types.String("")},
-		eval.OptToScan{"style", &style, types.String("")},
-	)
-	if style != "" {
-		c.style = ui.StylesFromString(style)
-	}
-
-	ec.OutputChan() <- c
-}
-
-func filterRawCandidates(ev *eval.Evaler, matcher eval.Fn,
-	seed string, chanRawCandidate <-chan rawCandidate) ([]rawCandidate, error) {
-
-	matcherInput := make(chan types.Value)
-	stopCollector := make(chan struct{})
-	var collected []rawCandidate
-	go func() {
-		defer close(matcherInput)
-		for rc := range chanRawCandidate {
-			collected = append(collected, rc)
-			select {
-			case matcherInput <- types.String(rc.text()):
-			case <-stopCollector:
-				return
-			}
-		}
-	}()
-	defer close(stopCollector)
-
-	ports := []*eval.Port{
-		{Chan: matcherInput, File: eval.DevNull}, {File: os.Stdout}, {File: os.Stderr}}
-	ec := eval.NewTopFrame(ev, eval.NewInternalSource("[editor matcher]"), ports)
-
-	args := []types.Value{types.String(seed)}
-	values, err := ec.PCaptureOutput(matcher, args, eval.NoOpts)
-	if err != nil {
-		return nil, err
-	} else if len(values) != len(collected) {
-		return nil, errIncorrectNumOfResults
-	}
-
-	var filtered []rawCandidate
-	for i, value := range values {
-		if types.ToBool(value) {
-			filtered = append(filtered, collected[i])
-		}
-	}
-	sort.Sort(rawCandidates(filtered))
-	return filtered, nil
-}
diff -pruN 0.11+ds1-5/edit/complete_arg.go 0.12+ds1-1/edit/complete_arg.go
--- 0.11+ds1-5/edit/complete_arg.go	2018-01-18 02:03:11.000000000 +0000
+++ 0.12+ds1-1/edit/complete_arg.go	1970-01-01 00:00:00.000000000 +0000
@@ -1,127 +0,0 @@
-package edit
-
-import (
-	"fmt"
-	"io/ioutil"
-	"os"
-	"path/filepath"
-	"strings"
-
-	"github.com/elves/elvish/edit/lscolors"
-	"github.com/elves/elvish/edit/ui"
-	"github.com/elves/elvish/eval"
-	"github.com/elves/elvish/parse"
-)
-
-type argComplContext struct {
-	complContextCommon
-	words []string
-}
-
-func (*argComplContext) name() string { return "argument" }
-
-func findArgComplContext(n parse.Node, ev pureEvaler) complContext {
-	if sep, ok := n.(*parse.Sep); ok {
-		if form, ok := sep.Parent().(*parse.Form); ok && form.Head != nil {
-			return &argComplContext{
-				complContextCommon{
-					"", quotingForEmptySeed, n.End(), n.End()},
-				evalFormPure(form, "", n.End(), ev),
-			}
-		}
-	}
-	if primary, ok := n.(*parse.Primary); ok {
-		if compound, seed := primaryInSimpleCompound(primary, ev); compound != nil {
-			if form, ok := compound.Parent().(*parse.Form); ok {
-				if form.Head != nil && form.Head != compound {
-					return &argComplContext{
-						complContextCommon{
-							seed, primary.Type, compound.Begin(), compound.End()},
-						evalFormPure(form, seed, compound.Begin(), ev),
-					}
-				}
-			}
-		}
-	}
-	return nil
-}
-
-func evalFormPure(form *parse.Form, seed string, seedBegin int, ev pureEvaler) []string {
-	// Find out head of the form and preceding arguments.
-	// If form.Head is not a simple compound, head will be "", just what we want.
-	head, _ := ev.PurelyEvalPartialCompound(form.Head, nil)
-	words := []string{head}
-	for _, compound := range form.Args {
-		if compound.Begin() >= seedBegin {
-			break
-		}
-		if arg, err := ev.PurelyEvalCompound(compound); err == nil {
-			// XXX Arguments that are not simple compounds are simply ignored.
-			words = append(words, arg)
-		}
-	}
-
-	words = append(words, seed)
-	return words
-}
-
-// To complete an argument, delegate the actual completion work to a suitable
-// complContext.
-func (ctx *argComplContext) generate(ev *eval.Evaler, ch chan<- rawCandidate) error {
-	return completeArg(ctx.words, ev, ch)
-}
-
-// TODO: getStyle does redundant stats.
-func complFilenameInner(head string, executableOnly bool, rawCands chan<- rawCandidate) error {
-	dir, fileprefix := filepath.Split(head)
-	dirToRead := dir
-	if dirToRead == "" {
-		dirToRead = "."
-	}
-
-	infos, err := ioutil.ReadDir(dirToRead)
-	if err != nil {
-		return fmt.Errorf("cannot list directory %s: %v", dirToRead, err)
-	}
-
-	lsColor := lscolors.GetColorist()
-	// Make candidates out of elements that match the file component.
-	for _, info := range infos {
-		name := info.Name()
-		// Show dot files iff file part of pattern starts with dot, and vice
-		// versa.
-		if dotfile(fileprefix) != dotfile(name) {
-			continue
-		}
-		// Only accept searchable directories and executable files if
-		// executableOnly is true.
-		if executableOnly && !(info.IsDir() || (info.Mode()&0111) != 0) {
-			continue
-		}
-
-		// Full filename for source and getStyle.
-		full := dir + name
-
-		suffix := " "
-		if info.IsDir() {
-			suffix = string(filepath.Separator)
-		} else if info.Mode()&os.ModeSymlink != 0 {
-			stat, err := os.Stat(full)
-			if err == nil && stat.IsDir() {
-				// Symlink to directory.
-				suffix = string(filepath.Separator)
-			}
-		}
-
-		rawCands <- &complexCandidate{
-			stem: full, codeSuffix: suffix,
-			style: ui.StylesFromString(lsColor.GetStyle(full)),
-		}
-	}
-
-	return nil
-}
-
-func dotfile(fname string) bool {
-	return strings.HasPrefix(fname, ".")
-}
diff -pruN 0.11+ds1-5/edit/complete_arg_test.go 0.12+ds1-1/edit/complete_arg_test.go
--- 0.11+ds1-5/edit/complete_arg_test.go	2018-01-18 02:03:11.000000000 +0000
+++ 0.12+ds1-1/edit/complete_arg_test.go	1970-01-01 00:00:00.000000000 +0000
@@ -1,18 +0,0 @@
-package edit
-
-import (
-	"testing"
-
-	"github.com/elves/elvish/parse"
-)
-
-func TestFindArgComplContext(t *testing.T) {
-	testComplContextFinder(t, "findArgComplContext", findArgComplContext, []complContextFinderTest{
-		{"a ", &argComplContext{
-			complContextCommon{"", quotingForEmptySeed, 2, 2}, []string{"a", ""}}},
-		{"a b", &argComplContext{
-			complContextCommon{"b", parse.Bareword, 2, 3}, []string{"a", "b"}}},
-		// No space after command; won't complete arg
-		{"a", nil},
-	})
-}
diff -pruN 0.11+ds1-5/edit/complete_arg_unix_test.go 0.12+ds1-1/edit/complete_arg_unix_test.go
--- 0.11+ds1-5/edit/complete_arg_unix_test.go	2018-01-18 02:03:11.000000000 +0000
+++ 0.12+ds1-1/edit/complete_arg_unix_test.go	1970-01-01 00:00:00.000000000 +0000
@@ -1,103 +0,0 @@
-// +build !windows,!plan9
-
-// TODO: Enable on Windows.
-package edit
-
-import (
-	"os"
-	"reflect"
-	"sort"
-	"testing"
-
-	"github.com/elves/elvish/edit/ui"
-	"github.com/elves/elvish/util"
-)
-
-var (
-	fileStyle = ui.StylesFromString("1")
-	exeStyle  = ui.StylesFromString("2")
-	dirStyle  = ui.StylesFromString("4")
-)
-
-var complFilenameInnerTests = []struct {
-	head           string
-	executableOnly bool
-	wantCandidates rawCandidates
-}{
-	// Match all non-hidden files and dirs, in alphabetical order.
-	// Files have suffix " " and directories "/". Styles are set according to
-	// the LS_COLORS variable, which are set in the beginning of the test.
-	{"haha", false, rawCandidates{
-		&complexCandidate{stem: "Documents", codeSuffix: "/", style: dirStyle},
-		&complexCandidate{stem: "bar", codeSuffix: " ", style: fileStyle},
-		&complexCandidate{stem: "elvish", codeSuffix: " ", style: exeStyle},
-		&complexCandidate{stem: "foo", codeSuffix: " ", style: fileStyle},
-	}},
-	// Only match executables and directories.
-	{"haha", true, rawCandidates{
-		&complexCandidate{stem: "Documents", codeSuffix: "/", style: dirStyle},
-		&complexCandidate{stem: "elvish", codeSuffix: " ", style: exeStyle},
-	}},
-	// Match hidden files and directories.
-	{".haha", false, rawCandidates{
-		&complexCandidate{stem: ".elvish", codeSuffix: "/", style: dirStyle},
-		&complexCandidate{stem: ".vimrc", codeSuffix: " ", style: fileStyle},
-	}},
-}
-
-func TestComplFilenameInner(t *testing.T) {
-	os.Setenv("LS_COLORS", "rs=1:ex=2:di=4")
-	util.InTempDir(func(string) {
-		create("foo", 0600)
-		create(".vimrc", 0600)
-		create("bar", 0600)
-
-		create("elvish", 0700)
-
-		mkdir("Documents", 0700)
-		mkdir(".elvish", 0700)
-
-		for _, test := range complFilenameInnerTests {
-			var (
-				err   error
-				cands rawCandidates
-				gets  = make(chan rawCandidate)
-			)
-			go func() {
-				defer close(gets)
-				err = complFilenameInner(test.head, test.executableOnly, gets)
-			}()
-			for v := range gets {
-				cands = append(cands, v)
-			}
-			if err != nil {
-				t.Errorf("complFilenameInner(%v, %v) returns error %v, want nil",
-					test.head, test.executableOnly, err)
-			}
-			sort.Sort(cands)
-			if !reflect.DeepEqual(cands, test.wantCandidates) {
-				t.Errorf("complFilenameInner(%v, %v) returns %v, want %v",
-					test.head, test.executableOnly, cands, test.wantCandidates)
-				t.Log("returned candidates are:")
-				for _, cand := range cands {
-					t.Logf("%#v", cand)
-				}
-			}
-		}
-	})
-}
-
-func mkdir(dirname string, perm os.FileMode) {
-	err := os.Mkdir(dirname, perm)
-	if err != nil {
-		panic(err)
-	}
-}
-
-func create(fname string, perm os.FileMode) {
-	f, err := os.OpenFile(fname, os.O_CREATE, perm)
-	if err != nil {
-		panic(err)
-	}
-	f.Close()
-}
diff -pruN 0.11+ds1-5/edit/complete_command.go 0.12+ds1-1/edit/complete_command.go
--- 0.11+ds1-5/edit/complete_command.go	2018-01-18 02:03:11.000000000 +0000
+++ 0.12+ds1-1/edit/complete_command.go	1970-01-01 00:00:00.000000000 +0000
@@ -1,104 +0,0 @@
-package edit
-
-import (
-	"strings"
-
-	"github.com/elves/elvish/edit/ui"
-
-	"github.com/elves/elvish/eval"
-	"github.com/elves/elvish/parse"
-	"github.com/elves/elvish/util"
-)
-
-type commandComplContext struct {
-	complContextCommon
-}
-
-const quotingForEmptySeed = parse.Bareword
-
-func findCommandComplContext(n parse.Node, ev pureEvaler) complContext {
-	// Determine if we are starting a new command. There are 3 cases:
-	// 1. The whole chunk is empty (nothing entered at all): the leaf is a
-	//    Chunk.
-	// 2. Just after a newline or semicolon: the leaf is a Sep and its parent is
-	//    a Chunk.
-	// 3. Just after a pipe: the leaf is a Sep and its parent is a Pipeline.
-	if parse.IsChunk(n) {
-		return &commandComplContext{
-			complContextCommon{"", parse.Bareword, n.End(), n.End()}}
-	}
-	if parse.IsSep(n) {
-		parent := n.Parent()
-		switch {
-		case parse.IsChunk(parent), parse.IsPipeline(parent):
-			return &commandComplContext{
-				complContextCommon{"", quotingForEmptySeed, n.End(), n.End()}}
-		case parse.IsPrimary(parent):
-			ptype := parent.(*parse.Primary).Type
-			if ptype == parse.OutputCapture || ptype == parse.ExceptionCapture {
-				return &commandComplContext{
-					complContextCommon{"", quotingForEmptySeed, n.End(), n.End()}}
-			}
-		}
-	}
-
-	if primary, ok := n.(*parse.Primary); ok {
-		if compound, seed := primaryInSimpleCompound(primary, ev); compound != nil {
-			if form, ok := compound.Parent().(*parse.Form); ok {
-				if form.Head == compound {
-					return &commandComplContext{
-						complContextCommon{seed, primary.Type, compound.Begin(), compound.End()}}
-				}
-			}
-		}
-	}
-	return nil
-}
-
-func (*commandComplContext) name() string { return "command" }
-
-func (ctx *commandComplContext) generate(ev *eval.Evaler, ch chan<- rawCandidate) error {
-	return complFormHeadInner(ctx.seed, ev, ch)
-}
-
-func complFormHeadInner(head string, ev *eval.Evaler, rawCands chan<- rawCandidate) error {
-	if util.DontSearch(head) {
-		return complFilenameInner(head, true, rawCands)
-	}
-
-	got := func(s string) {
-		rawCands <- plainCandidate(s)
-	}
-	for special := range eval.IsBuiltinSpecial {
-		got(special)
-	}
-	explode, ns, _ := eval.ParseVariable(head)
-	if !explode {
-		ev.EachVariableInTop(ns, func(varname string) {
-			if strings.HasSuffix(varname, eval.FnSuffix) {
-				got(eval.MakeVariableName(false, ns, varname[:len(varname)-len(eval.FnSuffix)]))
-			} else {
-				name := eval.MakeVariableName(false, ns, varname)
-				rawCands <- &complexCandidate{name, " = ", " = ", ui.Styles{}}
-			}
-		})
-	}
-	eval.EachExternal(func(command string) {
-		got(command)
-		if strings.HasPrefix(head, "e:") {
-			got("e:" + command)
-		}
-	})
-	// TODO Support non-module namespaces.
-	for name := range ev.Global {
-		if head != name && strings.HasSuffix(name, eval.NsSuffix) {
-			got(name)
-		}
-	}
-	for name := range ev.Builtin {
-		if head != name && strings.HasSuffix(name, eval.NsSuffix) {
-			got(name)
-		}
-	}
-	return nil
-}
diff -pruN 0.11+ds1-5/edit/complete_command_test.go 0.12+ds1-1/edit/complete_command_test.go
--- 0.11+ds1-5/edit/complete_command_test.go	2018-01-18 02:03:11.000000000 +0000
+++ 0.12+ds1-1/edit/complete_command_test.go	1970-01-01 00:00:00.000000000 +0000
@@ -1,59 +0,0 @@
-package edit
-
-import (
-	"testing"
-
-	"github.com/elves/elvish/parse"
-)
-
-func TestFindCommandComplContext(t *testing.T) {
-	testComplContextFinder(t, "findCommandComplContext", findCommandComplContext, []complContextFinderTest{
-		// Very beginning, empty seed
-		{"", &commandComplContext{
-			complContextCommon{"", quotingForEmptySeed, 0, 0}}},
-		// Very beginning, nonempty seed
-		{"a", &commandComplContext{
-			complContextCommon{"a", parse.Bareword, 0, 1}}},
-		// Very beginning, nonempty seed with quoting
-		{`"a"`, &commandComplContext{
-			complContextCommon{"a", parse.DoubleQuoted, 0, 3}}},
-		// Very beginning, nonempty seed with partial quoting
-		{`"a`, &commandComplContext{
-			complContextCommon{"a", parse.DoubleQuoted, 0, 2}}},
-
-		// Beginning of output capture, empty seed
-		{"a (", &commandComplContext{
-			complContextCommon{"", quotingForEmptySeed, 3, 3}}},
-		// Beginning of output capture, nonempty seed
-		{"a (b", &commandComplContext{
-			complContextCommon{"b", parse.Bareword, 3, 4}}},
-
-		// Beginning of exception capture, empty seed
-		{"a ?(", &commandComplContext{
-			complContextCommon{"", quotingForEmptySeed, 4, 4}}},
-		// Beginning of exception capture, nonempty seed
-		{"a ?(b", &commandComplContext{
-			complContextCommon{"b", parse.Bareword, 4, 5}}},
-
-		// Beginning of lambda, empty seed
-		{"a { ", &commandComplContext{
-			complContextCommon{"", quotingForEmptySeed, 4, 4}}},
-		// Beginning of lambda, nonempty seed
-		{"a { b", &commandComplContext{
-			complContextCommon{"b", parse.Bareword, 4, 5}}},
-
-		// After another command and pipe, empty seed
-		{"a|", &commandComplContext{
-			complContextCommon{"", quotingForEmptySeed, 2, 2}}},
-		// After another command and pipe, nonempty seed
-		{"a|b", &commandComplContext{
-			complContextCommon{"b", parse.Bareword, 2, 3}}},
-
-		// After another pipeline, empty seed
-		{"a\n", &commandComplContext{
-			complContextCommon{"", quotingForEmptySeed, 2, 2}}},
-		// After another pipeline, nonempty seed
-		{"a\nb", &commandComplContext{
-			complContextCommon{"b", parse.Bareword, 2, 3}}},
-	})
-}
diff -pruN 0.11+ds1-5/edit/complete_index.go 0.12+ds1-1/edit/complete_index.go
--- 0.11+ds1-5/edit/complete_index.go	2018-01-18 02:03:11.000000000 +0000
+++ 0.12+ds1-1/edit/complete_index.go	1970-01-01 00:00:00.000000000 +0000
@@ -1,99 +0,0 @@
-package edit
-
-import (
-	"errors"
-
-	"github.com/elves/elvish/eval"
-	"github.com/elves/elvish/eval/types"
-	"github.com/elves/elvish/parse"
-)
-
-var errCannotIterateKey = errors.New("indexee does not support iterating keys")
-
-type indexComplContext struct {
-	complContextCommon
-	indexee types.Value
-}
-
-func (*indexComplContext) name() string { return "index" }
-
-// Find context information for complIndex.
-//
-// Right now we only support cases where there is only one level of indexing,
-// e.g. $a[<Tab> is supported but $a[x][<Tab> is not.
-func findIndexComplContext(n parse.Node, ev pureEvaler) complContext {
-	if parse.IsSep(n) {
-		if parse.IsIndexing(n.Parent()) {
-			// We are just after an opening bracket.
-			indexing := parse.GetIndexing(n.Parent())
-			if len(indexing.Indicies) == 1 {
-				if indexee := ev.PurelyEvalPrimary(indexing.Head); indexee != nil {
-					return &indexComplContext{
-						complContextCommon{
-							"", quotingForEmptySeed, n.End(), n.End()},
-						indexee,
-					}
-				}
-			}
-		}
-		if parse.IsArray(n.Parent()) {
-			array := n.Parent()
-			if parse.IsIndexing(array.Parent()) {
-				// We are after an existing index and spaces.
-				indexing := parse.GetIndexing(array.Parent())
-				if len(indexing.Indicies) == 1 {
-					if indexee := ev.PurelyEvalPrimary(indexing.Head); indexee != nil {
-						return &indexComplContext{
-							complContextCommon{
-								"", quotingForEmptySeed, n.End(), n.End()},
-							indexee,
-						}
-					}
-				}
-			}
-		}
-	}
-
-	if parse.IsPrimary(n) {
-		primary := parse.GetPrimary(n)
-		compound, seed := primaryInSimpleCompound(primary, ev)
-		if compound != nil {
-			if parse.IsArray(compound.Parent()) {
-				array := compound.Parent()
-				if parse.IsIndexing(array.Parent()) {
-					// We are just after an incomplete index.
-					indexing := parse.GetIndexing(array.Parent())
-					if len(indexing.Indicies) == 1 {
-						if indexee := ev.PurelyEvalPrimary(indexing.Head); indexee != nil {
-							return &indexComplContext{
-								complContextCommon{
-									seed, primary.Type, compound.Begin(), compound.End()},
-								indexee,
-							}
-						}
-					}
-				}
-			}
-		}
-	}
-
-	return nil
-}
-
-func (ctx *indexComplContext) generate(ev *eval.Evaler, ch chan<- rawCandidate) error {
-	m, ok := ctx.indexee.(types.IterateKeyer)
-	if !ok {
-		return errCannotIterateKey
-	}
-	complIndexInner(m, ch)
-	return nil
-}
-
-func complIndexInner(m types.IterateKeyer, ch chan<- rawCandidate) {
-	m.IterateKey(func(v types.Value) bool {
-		if keyv, ok := v.(types.String); ok {
-			ch <- plainCandidate(keyv)
-		}
-		return true
-	})
-}
diff -pruN 0.11+ds1-5/edit/complete_index_test.go 0.12+ds1-1/edit/complete_index_test.go
--- 0.11+ds1-5/edit/complete_index_test.go	2018-01-18 02:03:11.000000000 +0000
+++ 0.12+ds1-1/edit/complete_index_test.go	1970-01-01 00:00:00.000000000 +0000
@@ -1,54 +0,0 @@
-package edit
-
-import (
-	"reflect"
-	"sort"
-	"testing"
-
-	"github.com/elves/elvish/eval/types"
-	"github.com/elves/elvish/parse"
-)
-
-var testIndexee = types.String("a")
-
-func TestFindIndexComplContext(t *testing.T) {
-	testComplContextFinder(t, "findIndexComplContext", findIndexComplContext, []complContextFinderTest{
-		{"a[", &indexComplContext{
-			complContextCommon{"", quotingForEmptySeed, 2, 2}, testIndexee}},
-		{"a[x", &indexComplContext{
-			complContextCommon{"x", parse.Bareword, 2, 3}, testIndexee}},
-		{"a[x ", &indexComplContext{
-			complContextCommon{"", quotingForEmptySeed, 4, 4}, testIndexee}},
-		// Not supported when indexee cannot be evaluated statically
-		{"(x)[", nil},
-		// Multi-layer indexing not supported yet
-		{"a[x][", nil},
-	})
-}
-
-func TestComplIndexInner(t *testing.T) {
-	m := types.MakeMap(map[types.Value]types.Value{
-		types.String("foo"):   types.String("bar"),
-		types.String("lorem"): types.String("ipsum"),
-	})
-	var (
-		candidates     rawCandidates
-		wantCandidates = rawCandidates{
-			plainCandidate("foo"), plainCandidate("lorem"),
-		}
-	)
-
-	gets := make(chan rawCandidate)
-	go func() {
-		defer close(gets)
-		complIndexInner(m, gets)
-	}()
-	for v := range gets {
-		candidates = append(candidates, v)
-	}
-	sort.Sort(candidates)
-	if !reflect.DeepEqual(candidates, wantCandidates) {
-		t.Errorf("complIndexInner(%v) = %v, want %v",
-			m, candidates, wantCandidates)
-	}
-}
diff -pruN 0.11+ds1-5/edit/complete_redir.go 0.12+ds1-1/edit/complete_redir.go
--- 0.11+ds1-5/edit/complete_redir.go	2018-01-18 02:03:11.000000000 +0000
+++ 0.12+ds1-1/edit/complete_redir.go	1970-01-01 00:00:00.000000000 +0000
@@ -1,34 +0,0 @@
-package edit
-
-import (
-	"github.com/elves/elvish/eval"
-	"github.com/elves/elvish/parse"
-)
-
-type redirComplContext struct {
-	complContextCommon
-}
-
-func (*redirComplContext) name() string { return "redir" }
-
-func findRedirComplContext(n parse.Node, ev pureEvaler) complContext {
-	if parse.IsSep(n) {
-		if parse.IsRedir(n.Parent()) {
-			return &redirComplContext{complContextCommon{
-				"", quotingForEmptySeed, n.End(), n.End()}}
-		}
-	}
-	if primary, ok := n.(*parse.Primary); ok {
-		if compound, seed := primaryInSimpleCompound(primary, ev); compound != nil {
-			if parse.IsRedir(compound.Parent()) {
-				return &redirComplContext{complContextCommon{
-					seed, primary.Type, compound.Begin(), compound.End()}}
-			}
-		}
-	}
-	return nil
-}
-
-func (ctx *redirComplContext) generate(ev *eval.Evaler, ch chan<- rawCandidate) error {
-	return complFilenameInner(ctx.seed, false, ch)
-}
diff -pruN 0.11+ds1-5/edit/complete_redir_test.go 0.12+ds1-1/edit/complete_redir_test.go
--- 0.11+ds1-5/edit/complete_redir_test.go	2018-01-18 02:03:11.000000000 +0000
+++ 0.12+ds1-1/edit/complete_redir_test.go	1970-01-01 00:00:00.000000000 +0000
@@ -1,12 +0,0 @@
-package edit
-
-import "testing"
-
-func TestFindRedirComplContext(t *testing.T) {
-	testComplContextFinder(t, "findRedirComplContext", findRedirComplContext, []complContextFinderTest{
-		{"a >", &redirComplContext{
-			complContextCommon{"", quotingForEmptySeed, 3, 3}}},
-		{"a >b", &redirComplContext{
-			complContextCommon{"b", quotingForEmptySeed, 3, 4}}},
-	})
-}
diff -pruN 0.11+ds1-5/edit/complete_variable.go 0.12+ds1-1/edit/complete_variable.go
--- 0.11+ds1-5/edit/complete_variable.go	2018-01-18 02:03:11.000000000 +0000
+++ 0.12+ds1-1/edit/complete_variable.go	1970-01-01 00:00:00.000000000 +0000
@@ -1,63 +0,0 @@
-package edit
-
-import (
-	"strings"
-
-	"github.com/elves/elvish/eval"
-	"github.com/elves/elvish/parse"
-)
-
-type variableComplContext struct {
-	complContextCommon
-	ns, nsPart string
-}
-
-func (*variableComplContext) name() string { return "variable" }
-
-func findVariableComplContext(n parse.Node, _ pureEvaler) complContext {
-	primary := parse.GetPrimary(n)
-	if primary != nil && primary.Type == parse.Variable {
-		explode, qname := eval.ParseVariableSplice(primary.Value)
-		nsPart, nameSeed := eval.ParseVariableQName(qname)
-		// Move past "$", "@" and "<ns>:".
-		begin := primary.Begin() + 1 + len(explode) + len(nsPart)
-		ns := nsPart
-		if len(ns) > 0 {
-			ns = ns[:len(ns)-1]
-		}
-		return &variableComplContext{
-			complContextCommon{nameSeed, parse.Bareword, begin, primary.End()},
-			ns, nsPart,
-		}
-	}
-	return nil
-}
-
-type evalerScopes interface {
-	EachVariableInTop(string, func(string))
-	EachNsInTop(func(string))
-}
-
-func (ctx *variableComplContext) generate(ev *eval.Evaler, ch chan<- rawCandidate) error {
-	complVariable(ctx.ns, ctx.nsPart, ev, ch)
-	return nil
-}
-
-func complVariable(ctxNs, ctxNsPart string, ev evalerScopes, ch chan<- rawCandidate) {
-	ev.EachVariableInTop(ctxNs, func(varname string) {
-		ch <- noQuoteCandidate(varname)
-	})
-
-	ev.EachNsInTop(func(ns string) {
-		nsPart := ns + ":"
-		// This is to match namespaces that are "nested" under the current
-		// namespace.
-		if hasProperPrefix(nsPart, ctxNsPart) {
-			ch <- noQuoteCandidate(nsPart[len(ctxNsPart):])
-		}
-	})
-}
-
-func hasProperPrefix(s, p string) bool {
-	return len(s) > len(p) && strings.HasPrefix(s, p)
-}
diff -pruN 0.11+ds1-5/edit/complete_variable_test.go 0.12+ds1-1/edit/complete_variable_test.go
--- 0.11+ds1-5/edit/complete_variable_test.go	2018-01-18 02:03:11.000000000 +0000
+++ 0.12+ds1-1/edit/complete_variable_test.go	1970-01-01 00:00:00.000000000 +0000
@@ -1,90 +0,0 @@
-package edit
-
-import (
-	"reflect"
-	"sort"
-	"testing"
-
-	"github.com/elves/elvish/parse"
-)
-
-func TestFindVariableComplContext(t *testing.T) {
-	testComplContextFinder(t, "findVariableComplContext", findVariableComplContext, []complContextFinderTest{
-		{"$", &variableComplContext{
-			complContextCommon{"", parse.Bareword, 1, 1}, "", ""}},
-		{"$a", &variableComplContext{
-			complContextCommon{"a", parse.Bareword, 1, 2}, "", ""}},
-		{"$a:", &variableComplContext{
-			complContextCommon{"", parse.Bareword, 3, 3}, "a", "a:"}},
-		{"$a:b", &variableComplContext{
-			complContextCommon{"b", parse.Bareword, 3, 4}, "a", "a:"}},
-		// Wrong contexts
-		{"", nil},
-		{"echo", nil},
-	})
-}
-
-type testEvalerScopes struct{}
-
-var testScopes = map[string]map[string]int{
-	"":        {"veni": 0, "vidi": 0, "vici": 0},
-	"foo":     {"lorem": 0, "ipsum": 0},
-	"foo:bar": {"lorem": 0, "dolor": 0},
-}
-
-func (testEvalerScopes) EachNsInTop(f func(string)) {
-	for ns := range testScopes {
-		if ns != "" {
-			f(ns)
-		}
-	}
-}
-
-func (testEvalerScopes) EachVariableInTop(ns string, f func(string)) {
-	for name := range testScopes[ns] {
-		f(name)
-	}
-}
-
-var complVariableTests = []struct {
-	ns     string
-	nsPart string
-	want   []rawCandidate
-}{
-	// No namespace: complete variables and namespaces
-	{"", "", []rawCandidate{
-		noQuoteCandidate("foo:"), noQuoteCandidate("foo:bar:"),
-		noQuoteCandidate("veni"), noQuoteCandidate("vici"), noQuoteCandidate("vidi"),
-	}},
-	// Nonempty namespace: complete variables in namespace and subnamespaces
-	// (but not variables in subnamespaces)
-	{"foo", "foo:", []rawCandidate{
-		noQuoteCandidate("bar:"),
-		noQuoteCandidate("ipsum"), noQuoteCandidate("lorem"),
-	}},
-	// Bad namespace
-	{"bad", "bad:", nil},
-}
-
-func TestComplVariable(t *testing.T) {
-	for _, test := range complVariableTests {
-		got := collectComplVariable(test.ns, test.nsPart, testEvalerScopes{})
-		if !reflect.DeepEqual(got, test.want) {
-			t.Errorf("complVariable(%q, %q, ...) => %v, want %v", test.ns, test.nsPart, got, test.want)
-		}
-	}
-}
-
-func collectComplVariable(ns, nsPart string, ev evalerScopes) []rawCandidate {
-	ch := make(chan rawCandidate)
-	go func() {
-		complVariable(ns, nsPart, ev, ch)
-		close(ch)
-	}()
-	var results []rawCandidate
-	for result := range ch {
-		results = append(results, result)
-	}
-	sort.Sort(rawCandidates(results))
-	return results
-}
diff -pruN 0.11+ds1-5/edit/completion/arg_completers.go 0.12+ds1-1/edit/completion/arg_completers.go
--- 0.11+ds1-5/edit/completion/arg_completers.go	1970-01-01 00:00:00.000000000 +0000
+++ 0.12+ds1-1/edit/completion/arg_completers.go	2018-07-11 02:32:25.000000000 +0000
@@ -0,0 +1,238 @@
+package completion
+
+import (
+	"bufio"
+	"errors"
+	"io"
+	"os"
+	"strings"
+	"unsafe"
+
+	"github.com/elves/elvish/eval"
+	"github.com/xiaq/persistent/hash"
+	"github.com/xiaq/persistent/hashmap"
+)
+
+// For an overview of completion, see the comment in completers.go.
+//
+// When completing arguments of commands (as opposed to variable names, map
+// indicies, etc.), the list of candidates often depends on the command in
+// question; e.g. "ls <Tab>" and "apt <Tab>" should yield different results
+// because the commands "ls" and "apt" accept different arguments. To reflec
+// this, Elvish has a map of "argument completers", with the key being the name
+// of the command, and the value being the argument completer itself, accessible
+// to script as $edit:arg-completer. The one entry with an empty string as the
+// key is the fallback completer, and is used when an argument completer for the
+// current command has not been defined.
+//
+// When completing an argument, Elvish first finds out the name of the command
+// (e.g. "ls" or "apt") can tries to evaluate its arguments. It then calls the
+// suitable completer with the name of the command and the arguments. The
+// arguments are in evaluated forms: e.g. if an argument is 'foo' (with quotes),
+// the argument is its value foo, not a literal 'foo'. The last argument is what
+// needs to be completed; if the user is starting a new argument, e.g. by typing
+// "ls a " (without quotes), the last argument passed to the argument completer
+// will be an empty string.
+//
+// The argument completer should then return a list of what can replace the last
+// argument. The results are of type rawCandidate, which basically means that
+// argument completers do not need to worry about quoting of candidates; the raw
+// candidates will be "cooked" into actual candidates that appear in the
+// interface, which includes quoting.
+//
+// There are a few finer points in this process:
+//
+// 1. If some of the arguments cannot be evaluated statically (for instance,
+//    consider this: echo (uname)), it will be an empty string. There needs
+//    probably be a better way to distinguish empty arguments and unknown
+//    arguments, but normally there is not much argument completers can do for
+//    unknown arguments.
+//
+// 2. The argument completer normally **should not** perform filtering. For
+//    instance, if the user has typed "ls x", the argument completer for "ls"
+//    should return **all** files, not just those whose names start with x. This
+//    is to make it possible for user to specify a different matching algorithm
+//    than the default prefix matching.
+//
+//    However, argument completers **should** look at the argument to decide
+//    which **type** of candidates to generate. For instance, if the user has
+//    typed "ls --x", the argument completer should generate all long options
+//    for "ls", but not only those starting with "x".
+
+var (
+	// errCompleterMustBeFn is thrown if the user has put a non-function entry
+	// in $edit:completer, and that entry needs to be used for completion.
+	// TODO(xiaq): Detect the type violation when the user modifies
+	// $edit:completer.
+	errCompleterMustBeFn = errors.New("completer must be fn")
+	// errNoMatchingCompleter is thrown if there is no completer matching the
+	// current command.
+	errNoMatchingCompleter = errors.New("no matching completer")
+	// errCompleterArgMustBeString is thrown when a builtin argument completer
+	// is called with non-string arguments.
+	errCompleterArgMustBeString = errors.New("arguments to arg completers must be string")
+	// errTooFewArguments is thrown when a builtin argument completer is called
+	// with too few arguments.
+	errTooFewArguments = errors.New("too few arguments")
+)
+
+var (
+	argCompletersData = map[string]*argCompleterEntry{
+		"":     {"complete-filename", complFilename},
+		"sudo": {"complete-sudo", complSudo},
+	}
+)
+
+type argCompleterEntry struct {
+	name string
+	impl func([]string, *eval.Evaler, hashmap.Map, chan<- rawCandidate) error
+}
+
+// completeArg calls the correct argument completers according to the command
+// name. It is used by complArg and can also be useful when further dispatching
+// based on command name is needed -- e.g. in the argument completer for "sudo".
+func completeArg(words []string, ev *eval.Evaler, ac hashmap.Map, rawCands chan<- rawCandidate) error {
+	logger.Printf("completing argument: %q", words)
+	var v interface{}
+	index := words[0]
+	v, ok := ac.Index(index)
+	if !ok {
+		v, ok = ac.Index("")
+		if !ok {
+			return errNoMatchingCompleter
+		}
+	}
+	fn, ok := v.(eval.Callable)
+	if !ok {
+		return errCompleterMustBeFn
+	}
+	return callArgCompleter(fn, ev, words, rawCands)
+}
+
+type builtinArgCompleter struct {
+	name string
+	impl func([]string, *eval.Evaler, hashmap.Map, chan<- rawCandidate) error
+
+	argCompleter hashmap.Map
+}
+
+var _ eval.Callable = &builtinArgCompleter{}
+
+func (bac *builtinArgCompleter) Kind() string {
+	return "fn"
+}
+
+// Equal compares by identity.
+func (bac *builtinArgCompleter) Equal(a interface{}) bool {
+	return bac == a
+}
+
+func (bac *builtinArgCompleter) Hash() uint32 {
+	return hash.Pointer(unsafe.Pointer(bac))
+}
+
+func (bac *builtinArgCompleter) Repr(int) string {
+	return "$edit:" + bac.name + eval.FnSuffix
+}
+
+func (bac *builtinArgCompleter) Call(ec *eval.Frame, args []interface{}, opts map[string]interface{}) error {
+	eval.TakeNoOpt(opts)
+	words := make([]string, len(args))
+	for i, arg := range args {
+		s, ok := arg.(string)
+		if !ok {
+			throw(errCompleterArgMustBeString)
+		}
+		words[i] = s
+	}
+
+	rawCands := make(chan rawCandidate)
+	var err error
+	go func() {
+		defer close(rawCands)
+		err = bac.impl(words, ec.Evaler, bac.argCompleter, rawCands)
+	}()
+
+	output := ec.OutputChan()
+	for rc := range rawCands {
+		output <- rc
+	}
+	return err
+}
+
+func complFilename(words []string, ev *eval.Evaler, ac hashmap.Map, rawCands chan<- rawCandidate) error {
+	if len(words) < 1 {
+		return errTooFewArguments
+	}
+	return complFilenameInner(words[len(words)-1], false, rawCands)
+}
+
+func complSudo(words []string, ev *eval.Evaler, ac hashmap.Map, rawCands chan<- rawCandidate) error {
+	if len(words) < 2 {
+		return errTooFewArguments
+	}
+	if len(words) == 2 {
+		return complFormHeadInner(words[1], ev, rawCands)
+	}
+	return completeArg(words[1:], ev, ac, rawCands)
+}
+
+// callArgCompleter calls a Fn, assuming that it is an arg completer. It calls
+// the Fn with specified arguments and closed input, and converts its output to
+// candidate objects.
+func callArgCompleter(fn eval.Callable, ev *eval.Evaler, words []string, rawCands chan<- rawCandidate) error {
+
+	// Quick path for builtin arg completers.
+	if builtin, ok := fn.(*builtinArgCompleter); ok {
+		return builtin.impl(words, ev, builtin.argCompleter, rawCands)
+	}
+
+	args := make([]interface{}, len(words))
+	for i, word := range words {
+		args[i] = word
+	}
+
+	ports := []*eval.Port{
+		eval.DevNullClosedChan,
+		{}, // Will be replaced when capturing output
+		{File: os.Stderr},
+	}
+
+	valuesCb := func(ch <-chan interface{}) {
+		for v := range ch {
+			switch v := v.(type) {
+			case rawCandidate:
+				rawCands <- v
+			case string:
+				rawCands <- plainCandidate(v)
+			default:
+				logger.Printf("completer must output string or candidate")
+			}
+		}
+	}
+
+	bytesCb := func(r *os.File) {
+		buffered := bufio.NewReader(r)
+		for {
+			line, err := buffered.ReadString('\n')
+			if line != "" {
+				rawCands <- plainCandidate(strings.TrimSuffix(line, "\n"))
+			}
+			if err != nil {
+				if err != io.EOF {
+					logger.Println("error on reading:", err)
+				}
+				break
+			}
+		}
+	}
+
+	// XXX There is no source to pass to NewTopEvalCtx.
+	ec := eval.NewTopFrame(ev, eval.NewInternalSource("[editor completer]"), ports)
+	err := ec.CallWithOutputCallback(fn, args, eval.NoOpts, valuesCb, bytesCb)
+	if err != nil {
+		err = errors.New("completer error: " + err.Error())
+	}
+
+	return err
+}
diff -pruN 0.11+ds1-5/edit/completion/candidate.go 0.12+ds1-1/edit/completion/candidate.go
--- 0.11+ds1-5/edit/completion/candidate.go	1970-01-01 00:00:00.000000000 +0000
+++ 0.12+ds1-1/edit/completion/candidate.go	2018-07-11 02:32:25.000000000 +0000
@@ -0,0 +1,176 @@
+package completion
+
+import (
+	"fmt"
+	"os"
+	"sort"
+
+	"github.com/elves/elvish/edit/ui"
+	"github.com/elves/elvish/eval"
+	"github.com/elves/elvish/eval/vals"
+	"github.com/elves/elvish/parse"
+	"github.com/xiaq/persistent/hash"
+)
+
+type candidate struct {
+	code string    // This is what will be substituted on the command line.
+	menu ui.Styled // This is what is displayed in the completion menu.
+}
+
+// rawCandidate is what can be converted to a candidate.
+type rawCandidate interface {
+	text() string
+	cook(q parse.PrimaryType) *candidate
+}
+
+type rawCandidates []rawCandidate
+
+func (cs rawCandidates) Len() int           { return len(cs) }
+func (cs rawCandidates) Swap(i, j int)      { cs[i], cs[j] = cs[j], cs[i] }
+func (cs rawCandidates) Less(i, j int) bool { return cs[i].text() < cs[j].text() }
+
+// plainCandidate is a minimal implementation of rawCandidate.
+type plainCandidate string
+
+func (plainCandidate) Kind() string               { return "string" }
+func (p plainCandidate) Equal(a interface{}) bool { return p == a }
+func (p plainCandidate) Hash() uint32             { return hash.String(string(p)) }
+func (p plainCandidate) Repr(l int) string        { return vals.Repr(string(p), l) }
+func (p plainCandidate) text() string             { return string(p) }
+
+func (p plainCandidate) cook(q parse.PrimaryType) *candidate {
+	s := string(p)
+	quoted, _ := parse.QuoteAs(s, q)
+	return &candidate{code: quoted, menu: ui.Unstyled(s)}
+}
+
+// noQuoteCandidate is a rawCandidate that does not quote when cooked.
+type noQuoteCandidate string
+
+func (noQuoteCandidate) Kind() string                { return "string" }
+func (nq noQuoteCandidate) Equal(a interface{}) bool { return nq == a }
+func (nq noQuoteCandidate) Hash() uint32             { return hash.String(string(nq)) }
+func (nq noQuoteCandidate) Repr(l int) string        { return vals.Repr(string(nq), l) }
+func (nq noQuoteCandidate) text() string             { return string(nq) }
+
+func (nq noQuoteCandidate) cook(parse.PrimaryType) *candidate {
+	s := string(nq)
+	return &candidate{code: s, menu: ui.Unstyled(s)}
+}
+
+// complexCandidate is an implementation of rawCandidate that offers
+// customization options.
+type complexCandidate struct {
+	stem          string    // Used in the code and the menu.
+	codeSuffix    string    // Appended to the code.
+	displaySuffix string    // Appended to the display.
+	style         ui.Styles // Used in the menu.
+}
+
+func (c *complexCandidate) Index(k interface{}) (interface{}, bool) {
+	switch k {
+	case "stem":
+		return c.stem, true
+	case "code-suffix":
+		return c.codeSuffix, true
+	case "display-suffix":
+		return c.displaySuffix, true
+	case "style":
+		return c.style.String(), true
+	}
+	return nil, false
+}
+
+func (c *complexCandidate) IterateKeys(f func(interface{}) bool) {
+	vals.Feed(f, "stem", "code-suffix", "display-suffix", "style")
+}
+
+func (c *complexCandidate) Kind() string { return "map" }
+
+func (c *complexCandidate) Equal(a interface{}) bool {
+	rhs, ok := a.(*complexCandidate)
+	return ok && c.stem == rhs.stem && c.codeSuffix == rhs.codeSuffix && c.displaySuffix == rhs.displaySuffix && c.style.Eq(rhs.style)
+}
+
+func (c *complexCandidate) Hash() uint32 {
+	h := hash.DJBInit
+	h = hash.DJBCombine(h, hash.String(c.stem))
+	h = hash.DJBCombine(h, hash.String(c.codeSuffix))
+	h = hash.DJBCombine(h, hash.String(c.displaySuffix))
+	h = hash.DJBCombine(h, c.style.Hash())
+	return h
+}
+
+func (c *complexCandidate) Repr(indent int) string {
+	// TODO(xiaq): Pretty-print when indent >= 0
+	return fmt.Sprintf("(edit:complex-candidate %s &code-suffix=%s &display-suffix=%s &style=%s)",
+		parse.Quote(c.stem), parse.Quote(c.codeSuffix),
+		parse.Quote(c.displaySuffix), parse.Quote(c.style.String()))
+}
+
+func (c *complexCandidate) text() string { return c.stem }
+
+func (c *complexCandidate) cook(q parse.PrimaryType) *candidate {
+	quoted, _ := parse.QuoteAs(c.stem, q)
+	return &candidate{
+		code: quoted + c.codeSuffix,
+		menu: ui.Styled{c.stem + c.displaySuffix, c.style},
+	}
+}
+
+// makeComplexCandidate composes a complexCandidate.
+func makeComplexCandidate(rawOpts eval.RawOptions, stem string) *complexCandidate {
+	opts := struct {
+		CodeSuffix    string
+		DisplaySuffix string
+		Style         string
+	}{}
+	rawOpts.Scan(&opts)
+
+	return &complexCandidate{
+		stem:          stem,
+		codeSuffix:    opts.CodeSuffix,
+		displaySuffix: opts.DisplaySuffix,
+		style:         ui.StylesFromString(opts.Style),
+	}
+}
+
+func filterRawCandidates(ev *eval.Evaler, matcher eval.Callable, seed string, chanRawCandidate <-chan rawCandidate) ([]rawCandidate, error) {
+
+	matcherInput := make(chan interface{})
+	stopCollector := make(chan struct{})
+	var collected []rawCandidate
+	go func() {
+		defer close(matcherInput)
+		for rc := range chanRawCandidate {
+			collected = append(collected, rc)
+			select {
+			case matcherInput <- rc.text():
+			case <-stopCollector:
+				return
+			}
+		}
+	}()
+	defer close(stopCollector)
+
+	ports := []*eval.Port{
+		{Chan: matcherInput, File: eval.DevNull}, {File: os.Stdout}, {File: os.Stderr}}
+	ec := eval.NewTopFrame(ev, eval.NewInternalSource("[editor matcher]"), ports)
+
+	args := []interface{}{seed}
+	values, err := ec.CaptureOutput(matcher, args, eval.NoOpts)
+	if err != nil {
+		return nil, err
+	} else if len(values) != len(collected) {
+		return nil, errIncorrectNumOfResults
+	}
+
+	var filtered []rawCandidate
+	for i, value := range values {
+		if vals.Bool(value) {
+			filtered = append(filtered, collected[i])
+		}
+	}
+	sort.Sort(rawCandidates(filtered))
+	return filtered, nil
+}
diff -pruN 0.11+ds1-5/edit/completion/candidate_test.go 0.12+ds1-1/edit/completion/candidate_test.go
--- 0.11+ds1-5/edit/completion/candidate_test.go	1970-01-01 00:00:00.000000000 +0000
+++ 0.12+ds1-1/edit/completion/candidate_test.go	2018-07-11 02:32:25.000000000 +0000
@@ -0,0 +1,57 @@
+package completion
+
+import (
+	"testing"
+
+	"github.com/elves/elvish/eval"
+	"github.com/elves/elvish/eval/vals"
+)
+
+func TestRawFilterCandidates(t *testing.T) {
+	passAll := eval.NewBuiltinFn("test:passAll",
+		func(fm *eval.Frame, opts eval.RawOptions, pattern string, inputs eval.Inputs) {
+			out := fm.OutputChan()
+			inputs(func(v interface{}) {
+				out <- vals.Bool(true)
+			})
+		})
+	blockAll := eval.NewBuiltinFn("test:blockAll",
+		func(fm *eval.Frame, opts eval.RawOptions, pattern string, inputs eval.Inputs) {
+			out := fm.OutputChan()
+			inputs(func(v interface{}) {
+				out <- vals.Bool(false)
+			})
+		})
+
+	tests := []filterRawCandidatesTest{
+		{
+			name:    "passAll",
+			matcher: passAll,
+			src:     []string{"1", "2", "3"},
+			want:    []string{"1", "2", "3"},
+		},
+		{
+			name:    "blockAll",
+			matcher: blockAll,
+			src:     []string{"1", "2", "3"},
+			want:    []string{},
+		},
+	}
+
+	testRawFilterCandidates(t, tests)
+}
+
+func TestComplexCandidate(t *testing.T) {
+	setup := func(ev *eval.Evaler) {
+		ev.Builtin.AddNs("edit", eval.NewNs().AddBuiltinFn(
+			"edit:", "complex-candidate", makeComplexCandidate))
+	}
+	That := eval.That
+	eval.TestWithSetup(t, setup,
+		That("put  (edit:complex-candidate lorem &code-suffix=ipsum &display-suffix=dolor &style=red)[stem]").Puts("lorem"),
+		That("put  (edit:complex-candidate lorem &code-suffix=ipsum &display-suffix=dolor &style=red)[code-suffix]").Puts("ipsum"),
+		That("put  (edit:complex-candidate lorem &code-suffix=ipsum &display-suffix=dolor &style=red)[display-suffix]").Puts("dolor"),
+		That("put  (edit:complex-candidate lorem &code-suffix=ipsum &display-suffix=dolor &style=red)[style]").Puts("31"),
+		That("keys (edit:complex-candidate lorem &code-suffix=ipsum &display-suffix=dolor &style=red)").Puts("stem", "code-suffix", "display-suffix", "style"),
+	)
+}
diff -pruN 0.11+ds1-5/edit/completion/complete_arg.go 0.12+ds1-1/edit/completion/complete_arg.go
--- 0.11+ds1-5/edit/completion/complete_arg.go	1970-01-01 00:00:00.000000000 +0000
+++ 0.12+ds1-1/edit/completion/complete_arg.go	2018-07-11 02:32:25.000000000 +0000
@@ -0,0 +1,126 @@
+package completion
+
+import (
+	"fmt"
+	"io/ioutil"
+	"os"
+	"path/filepath"
+	"strings"
+
+	"github.com/elves/elvish/edit/lscolors"
+	"github.com/elves/elvish/edit/ui"
+	"github.com/elves/elvish/parse"
+)
+
+type argComplContext struct {
+	complContextCommon
+	words []string
+}
+
+func (*argComplContext) name() string { return "argument" }
+
+func findArgComplContext(n parse.Node, ev pureEvaler) complContext {
+	if sep, ok := n.(*parse.Sep); ok {
+		if form, ok := sep.Parent().(*parse.Form); ok && form.Head != nil {
+			return &argComplContext{
+				complContextCommon{
+					"", quotingForEmptySeed, n.End(), n.End()},
+				evalFormPure(form, "", n.End(), ev),
+			}
+		}
+	}
+	if primary, ok := n.(*parse.Primary); ok {
+		if compound, seed := primaryInSimpleCompound(primary, ev); compound != nil {
+			if form, ok := compound.Parent().(*parse.Form); ok {
+				if form.Head != nil && form.Head != compound {
+					return &argComplContext{
+						complContextCommon{
+							seed, primary.Type, compound.Begin(), compound.End()},
+						evalFormPure(form, seed, compound.Begin(), ev),
+					}
+				}
+			}
+		}
+	}
+	return nil
+}
+
+func evalFormPure(form *parse.Form, seed string, seedBegin int, ev pureEvaler) []string {
+	// Find out head of the form and preceding arguments.
+	// If form.Head is not a simple compound, head will be "", just what we want.
+	head, _ := ev.PurelyEvalPartialCompound(form.Head, nil)
+	words := []string{head}
+	for _, compound := range form.Args {
+		if compound.Begin() >= seedBegin {
+			break
+		}
+		if arg, err := ev.PurelyEvalCompound(compound); err == nil {
+			// XXX Arguments that are not simple compounds are simply ignored.
+			words = append(words, arg)
+		}
+	}
+
+	words = append(words, seed)
+	return words
+}
+
+// To complete an argument, delegate the actual completion work to a suitable
+// complContext.
+func (ctx *argComplContext) generate(env *complEnv, ch chan<- rawCandidate) error {
+	return completeArg(ctx.words, env.evaler, env.argCompleter, ch)
+}
+
+// TODO: getStyle does redundant stats.
+func complFilenameInner(head string, executableOnly bool, rawCands chan<- rawCandidate) error {
+	dir, fileprefix := filepath.Split(head)
+	dirToRead := dir
+	if dirToRead == "" {
+		dirToRead = "."
+	}
+
+	infos, err := ioutil.ReadDir(dirToRead)
+	if err != nil {
+		return fmt.Errorf("cannot list directory %s: %v", dirToRead, err)
+	}
+
+	lsColor := lscolors.GetColorist()
+	// Make candidates out of elements that match the file component.
+	for _, info := range infos {
+		name := info.Name()
+		// Show dot files iff file part of pattern starts with dot, and vice
+		// versa.
+		if dotfile(fileprefix) != dotfile(name) {
+			continue
+		}
+		// Only accept searchable directories and executable files if
+		// executableOnly is true.
+		if executableOnly && !(info.IsDir() || (info.Mode()&0111) != 0) {
+			continue
+		}
+
+		// Full filename for source and getStyle.
+		full := dir + name
+
+		suffix := " "
+		if info.IsDir() {
+			suffix = string(filepath.Separator)
+		} else if info.Mode()&os.ModeSymlink != 0 {
+			stat, err := os.Stat(full)
+			if err == nil && stat.IsDir() {
+				// Symlink to directory.
+				suffix = string(filepath.Separator)
+			}
+		}
+
+		rawCands <- &complexCandidate{
+			stem: full, codeSuffix: suffix,
+			style: ui.StylesFromString(lsColor.GetStyle(full)),
+		}
+	}
+
+	return nil
+}
+
+func dotfile(fname string) bool {
+	return strings.HasPrefix(fname, ".")
+}
diff -pruN 0.11+ds1-5/edit/completion/complete_arg_test.go 0.12+ds1-1/edit/completion/complete_arg_test.go
--- 0.11+ds1-5/edit/completion/complete_arg_test.go	1970-01-01 00:00:00.000000000 +0000
+++ 0.12+ds1-1/edit/completion/complete_arg_test.go	2018-07-11 02:32:25.000000000 +0000
@@ -0,0 +1,18 @@
+package completion
+
+import (
+	"testing"
+
+	"github.com/elves/elvish/parse"
+)
+
+func TestFindArgComplContext(t *testing.T) {
+	testComplContextFinder(t, "findArgComplContext", findArgComplContext, []complContextFinderTest{
+		{"a ", &argComplContext{
+			complContextCommon{"", quotingForEmptySeed, 2, 2}, []string{"a", ""}}},
+		{"a b", &argComplContext{
+			complContextCommon{"b", parse.Bareword, 2, 3}, []string{"a", "b"}}},
+		// No space after command; won't complete arg
+		{"a", nil},
+	})
+}
diff -pruN 0.11+ds1-5/edit/completion/complete_arg_unix_test.go 0.12+ds1-1/edit/completion/complete_arg_unix_test.go
--- 0.11+ds1-5/edit/completion/complete_arg_unix_test.go	1970-01-01 00:00:00.000000000 +0000
+++ 0.12+ds1-1/edit/completion/complete_arg_unix_test.go	2018-07-11 02:32:25.000000000 +0000
@@ -0,0 +1,103 @@
+// +build !windows,!plan9
+
+// TODO: Enable on Windows.
+package completion
+
+import (
+	"os"
+	"reflect"
+	"sort"
+	"testing"
+
+	"github.com/elves/elvish/edit/ui"
+	"github.com/elves/elvish/util"
+)
+
+var (
+	fileStyle = ui.StylesFromString("1")
+	exeStyle  = ui.StylesFromString("2")
+	dirStyle  = ui.StylesFromString("4")
+)
+
+var complFilenameInnerTests = []struct {
+	head           string
+	executableOnly bool
+	wantCandidates rawCandidates
+}{
+	// Match all non-hidden files and dirs, in alphabetical order.
+	// Files have suffix " " and directories "/". Styles are set according to
+	// the LS_COLORS variable, which are set in the beginning of the test.
+	{"haha", false, rawCandidates{
+		&complexCandidate{stem: "Documents", codeSuffix: "/", style: dirStyle},
+		&complexCandidate{stem: "bar", codeSuffix: " ", style: fileStyle},
+		&complexCandidate{stem: "elvish", codeSuffix: " ", style: exeStyle},
+		&complexCandidate{stem: "foo", codeSuffix: " ", style: fileStyle},
+	}},
+	// Only match executables and directories.
+	{"haha", true, rawCandidates{
+		&complexCandidate{stem: "Documents", codeSuffix: "/", style: dirStyle},
+		&complexCandidate{stem: "elvish", codeSuffix: " ", style: exeStyle},
+	}},
+	// Match hidden files and directories.
+	{".haha", false, rawCandidates{
+		&complexCandidate{stem: ".elvish", codeSuffix: "/", style: dirStyle},
+		&complexCandidate{stem: ".vimrc", codeSuffix: " ", style: fileStyle},
+	}},
+}
+
+func TestComplFilenameInner(t *testing.T) {
+	os.Setenv("LS_COLORS", "rs=1:ex=2:di=4")
+	util.InTempDir(func(string) {
+		create("foo", 0600)
+		create(".vimrc", 0600)
+		create("bar", 0600)
+
+		create("elvish", 0700)
+
+		mkdir("Documents", 0700)
+		mkdir(".elvish", 0700)
+
+		for _, test := range complFilenameInnerTests {
+			var (
+				err   error
+				cands rawCandidates
+				gets  = make(chan rawCandidate)
+			)
+			go func() {
+				defer close(gets)
+				err = complFilenameInner(test.head, test.executableOnly, gets)
+			}()
+			for v := range gets {
+				cands = append(cands, v)
+			}
+			if err != nil {
+				t.Errorf("complFilenameInner(%v, %v) returns error %v, want nil",
+					test.head, test.executableOnly, err)
+			}
+			sort.Sort(cands)
+			if !reflect.DeepEqual(cands, test.wantCandidates) {
+				t.Errorf("complFilenameInner(%v, %v) returns %v, want %v",
+					test.head, test.executableOnly, cands, test.wantCandidates)
+				t.Log("returned candidates are:")
+				for _, cand := range cands {
+					t.Logf("%#v", cand)
+				}
+			}
+		}
+	})
+}
+
+func mkdir(dirname string, perm os.FileMode) {
+	err := os.Mkdir(dirname, perm)
+	if err != nil {
+		panic(err)
+	}
+}
+
+func create(fname string, perm os.FileMode) {
+	f, err := os.OpenFile(fname, os.O_CREATE, perm)
+	if err != nil {
+		panic(err)
+	}
+	f.Close()
+}
diff -pruN 0.11+ds1-5/edit/completion/complete_command.go 0.12+ds1-1/edit/completion/complete_command.go
--- 0.11+ds1-5/edit/completion/complete_command.go	1970-01-01 00:00:00.000000000 +0000
+++ 0.12+ds1-1/edit/completion/complete_command.go	2018-07-11 02:32:25.000000000 +0000
@@ -0,0 +1,97 @@
+package completion
+
+import (
+	"strings"
+
+	"github.com/elves/elvish/edit/ui"
+
+	"github.com/elves/elvish/eval"
+	"github.com/elves/elvish/parse"
+	"github.com/elves/elvish/util"
+)
+
+type commandComplContext struct {
+	complContextCommon
+}
+
+const quotingForEmptySeed = parse.Bareword
+
+func findCommandComplContext(n parse.Node, ev pureEvaler) complContext {
+	// Determine if we are starting a new command. There are 3 cases:
+	// 1. The whole chunk is empty (nothing entered at all): the leaf is a
+	//    Chunk.
+	// 2. Just after a newline or semicolon: the leaf is a Sep and its parent is
+	//    a Chunk.
+	// 3. Just after a pipe: the leaf is a Sep and its parent is a Pipeline.
+	if parse.IsChunk(n) {
+		return &commandComplContext{
+			complContextCommon{"", parse.Bareword, n.End(), n.End()}}
+	}
+	if parse.IsSep(n) {
+		parent := n.Parent()
+		switch {
+		case parse.IsChunk(parent), parse.IsPipeline(parent):
+			return &commandComplContext{
+				complContextCommon{"", quotingForEmptySeed, n.End(), n.End()}}
+		case parse.IsPrimary(parent):
+			ptype := parent.(*parse.Primary).Type
+			if ptype == parse.OutputCapture || ptype == parse.ExceptionCapture {
+				return &commandComplContext{
+					complContextCommon{"", quotingForEmptySeed, n.End(), n.End()}}
+			}
+		}
+	}
+
+	if primary, ok := n.(*parse.Primary); ok {
+		if compound, seed := primaryInSimpleCompound(primary, ev); compound != nil {
+			if form, ok := compound.Parent().(*parse.Form); ok {
+				if form.Head == compound {
+					return &commandComplContext{
+						complContextCommon{seed, primary.Type, compound.Begin(), compound.End()}}
+				}
+			}
+		}
+	}
+	return nil
+}
+
+func (*commandComplContext) name() string { return "command" }
+
+func (ctx *commandComplContext) generate(env *complEnv, ch chan<- rawCandidate) error {
+	return complFormHeadInner(ctx.seed, env.evaler, ch)
+}
+
+func complFormHeadInner(head string, ev *eval.Evaler, rawCands chan<- rawCandidate) error {
+	if util.DontSearch(head) {
+		return complFilenameInner(head, true, rawCands)
+	}
+
+	got := func(s string) {
+		rawCands <- plainCandidate(s)
+	}
+	for special := range eval.IsBuiltinSpecial {
+		got(special)
+	}
+	explode, ns, _ := eval.ParseIncompleteVariableRef(head)
+	if !explode {
+		logger.Printf("completing commands in ns %q", ns)
+		ev.EachVariableInTop(ns, func(varname string) {
+			switch {
+			case strings.HasSuffix(varname, eval.FnSuffix):
+				got(eval.MakeVariableRef(false, ns, varname[:len(varname)-len(eval.FnSuffix)]))
+			case strings.HasSuffix(varname, eval.NsSuffix):
+				got(eval.MakeVariableRef(false, ns, varname))
+			default:
+				name := eval.MakeVariableRef(false, ns, varname)
+				rawCands <- &complexCandidate{name, " = ", " = ", ui.Styles{}}
+			}
+		})
+	}
+	eval.EachExternal(func(command string) {
+		got(command)
+		if strings.HasPrefix(head, "e:") {
+			got("e:" + command)
+		}
+	})
+	return nil
+}
diff -pruN 0.11+ds1-5/edit/completion/complete_command_test.go 0.12+ds1-1/edit/completion/complete_command_test.go
--- 0.11+ds1-5/edit/completion/complete_command_test.go	1970-01-01 00:00:00.000000000 +0000
+++ 0.12+ds1-1/edit/completion/complete_command_test.go	2018-07-11 02:32:25.000000000 +0000
@@ -0,0 +1,59 @@
+package completion
+
+import (
+	"testing"
+
+	"github.com/elves/elvish/parse"
+)
+
+func TestFindCommandComplContext(t *testing.T) {
+	testComplContextFinder(t, "findCommandComplContext", findCommandComplContext, []complContextFinderTest{
+		// Very beginning, empty seed
+		{"", &commandComplContext{
+			complContextCommon{"", quotingForEmptySeed, 0, 0}}},
+		// Very beginning, nonempty seed
+		{"a", &commandComplContext{
+			complContextCommon{"a", parse.Bareword, 0, 1}}},
+		// Very beginning, nonempty seed with quoting
+		{`"a"`, &commandComplContext{
+			complContextCommon{"a", parse.DoubleQuoted, 0, 3}}},
+		// Very beginning, nonempty seed with partial quoting
+		{`"a`, &commandComplContext{
+			complContextCommon{"a", parse.DoubleQuoted, 0, 2}}},
+
+		// Beginning of output capture, empty seed
+		{"a (", &commandComplContext{
+			complContextCommon{"", quotingForEmptySeed, 3, 3}}},
+		// Beginning of output capture, nonempty seed
+		{"a (b", &commandComplContext{
+			complContextCommon{"b", parse.Bareword, 3, 4}}},
+
+		// Beginning of exception capture, empty seed
+		{"a ?(", &commandComplContext{
+			complContextCommon{"", quotingForEmptySeed, 4, 4}}},
+		// Beginning of exception capture, nonempty seed
+		{"a ?(b", &commandComplContext{
+			complContextCommon{"b", parse.Bareword, 4, 5}}},
+
+		// Beginning of lambda, empty seed
+		{"a { ", &commandComplContext{
+			complContextCommon{"", quotingForEmptySeed, 4, 4}}},
+		// Beginning of lambda, nonempty seed
+		{"a { b", &commandComplContext{
+			complContextCommon{"b", parse.Bareword, 4, 5}}},
+
+		// After another command and pipe, empty seed
+		{"a|", &commandComplContext{
+			complContextCommon{"", quotingForEmptySeed, 2, 2}}},
+		// After another command and pipe, nonempty seed
+		{"a|b", &commandComplContext{
+			complContextCommon{"b", parse.Bareword, 2, 3}}},
+
+		// After another pipeline, empty seed
+		{"a\n", &commandComplContext{
+			complContextCommon{"", quotingForEmptySeed, 2, 2}}},
+		// After another pipeline, nonempty seed
+		{"a\nb", &commandComplContext{
+			complContextCommon{"b", parse.Bareword, 2, 3}}},
+	})
+}
diff -pruN 0.11+ds1-5/edit/completion/complete_index.go 0.12+ds1-1/edit/completion/complete_index.go
--- 0.11+ds1-5/edit/completion/complete_index.go	1970-01-01 00:00:00.000000000 +0000
+++ 0.12+ds1-1/edit/completion/complete_index.go	2018-07-11 02:32:25.000000000 +0000
@@ -0,0 +1,85 @@
+package completion
+
+import (
+	"github.com/elves/elvish/eval/vals"
+	"github.com/elves/elvish/parse"
+)
+
+type indexComplContext struct {
+	complContextCommon
+	indexee interface{}
+}
+
+func (*indexComplContext) name() string { return "index" }
+
+// Find context information for complIndex.
+//
+// Right now we only support cases where there is only one level of indexing,
+// e.g. $a[<Tab> is supported but $a[x][<Tab> is not.
+func findIndexComplContext(n parse.Node, ev pureEvaler) complContext {
+	if parse.IsSep(n) {
+		if parse.IsIndexing(n.Parent()) {
+			// We are just after an opening bracket.
+			indexing := parse.GetIndexing(n.Parent())
+			if len(indexing.Indicies) == 1 {
+				if indexee := ev.PurelyEvalPrimary(indexing.Head); indexee != nil {
+					return &indexComplContext{
+						complContextCommon{
+							"", quotingForEmptySeed, n.End(), n.End()},
+						indexee,
+					}
+				}
+			}
+		}
+		if parse.IsArray(n.Parent()) {
+			array := n.Parent()
+			if parse.IsIndexing(array.Parent()) {
+				// We are after an existing index and spaces.
+				indexing := parse.GetIndexing(array.Parent())
+				if len(indexing.Indicies) == 1 {
+					if indexee := ev.PurelyEvalPrimary(indexing.Head); indexee != nil {
+						return &indexComplContext{
+							complContextCommon{
+								"", quotingForEmptySeed, n.End(), n.End()},
+							indexee,
+						}
+					}
+				}
+			}
+		}
+	}
+
+	if parse.IsPrimary(n) {
+		primary := parse.GetPrimary(n)
+		compound, seed := primaryInSimpleCompound(primary, ev)
+		if compound != nil {
+			if parse.IsArray(compound.Parent()) {
+				array := compound.Parent()
+				if parse.IsIndexing(array.Parent()) {
+					// We are just after an incomplete index.
+					indexing := parse.GetIndexing(array.Parent())
+					if len(indexing.Indicies) == 1 {
+						if indexee := ev.PurelyEvalPrimary(indexing.Head); indexee != nil {
+							return &indexComplContext{
+								complContextCommon{
+									seed, primary.Type, compound.Begin(), compound.End()},
+								indexee,
+							}
+						}
+					}
+				}
+			}
+		}
+	}
+
+	return nil
+}
+
+func (ctx *indexComplContext) generate(env *complEnv, ch chan<- rawCandidate) error {
+	return vals.IterateKeys(ctx.indexee, func(k interface{}) bool {
+		if kstring, ok := k.(string); ok {
+			ch <- plainCandidate(kstring)
+		}
+		return true
+	})
+}
diff -pruN 0.11+ds1-5/edit/completion/complete_index_test.go 0.12+ds1-1/edit/completion/complete_index_test.go
--- 0.11+ds1-5/edit/completion/complete_index_test.go	1970-01-01 00:00:00.000000000 +0000
+++ 0.12+ds1-1/edit/completion/complete_index_test.go	2018-07-11 02:32:25.000000000 +0000
@@ -0,0 +1,24 @@
+package completion
+
+import (
+	"testing"
+
+	"github.com/elves/elvish/parse"
+)
+
+var testIndexee = "a"
+
+func TestFindIndexComplContext(t *testing.T) {
+	testComplContextFinder(t, "findIndexComplContext", findIndexComplContext, []complContextFinderTest{
+		{"a[", &indexComplContext{
+			complContextCommon{"", quotingForEmptySeed, 2, 2}, testIndexee}},
+		{"a[x", &indexComplContext{
+			complContextCommon{"x", parse.Bareword, 2, 3}, testIndexee}},
+		{"a[x ", &indexComplContext{
+			complContextCommon{"", quotingForEmptySeed, 4, 4}, testIndexee}},
+		// Not supported when indexee cannot be evaluated statically
+		{"(x)[", nil},
+		// Multi-layer indexing not supported yet
+		{"a[x][", nil},
+	})
+}
diff -pruN 0.11+ds1-5/edit/completion/complete_redir.go 0.12+ds1-1/edit/completion/complete_redir.go
--- 0.11+ds1-5/edit/completion/complete_redir.go	1970-01-01 00:00:00.000000000 +0000
+++ 0.12+ds1-1/edit/completion/complete_redir.go	2018-07-11 02:32:25.000000000 +0000
@@ -0,0 +1,33 @@
+package completion
+
+import (
+	"github.com/elves/elvish/parse"
+)
+
+type redirComplContext struct {
+	complContextCommon
+}
+
+func (*redirComplContext) name() string { return "redir" }
+
+func findRedirComplContext(n parse.Node, ev pureEvaler) complContext {
+	if parse.IsSep(n) {
+		if parse.IsRedir(n.Parent()) {
+			return &redirComplContext{complContextCommon{
+				"", quotingForEmptySeed, n.End(), n.End()}}
+		}
+	}
+	if primary, ok := n.(*parse.Primary); ok {
+		if compound, seed := primaryInSimpleCompound(primary, ev); compound != nil {
+			if parse.IsRedir(compound.Parent()) {
+				return &redirComplContext{complContextCommon{
+					seed, primary.Type, compound.Begin(), compound.End()}}
+			}
+		}
+	}
+	return nil
+}
+
+func (ctx *redirComplContext) generate(env *complEnv, ch chan<- rawCandidate) error {
+	return complFilenameInner(ctx.seed, false, ch)
+}
diff -pruN 0.11+ds1-5/edit/completion/complete_redir_test.go 0.12+ds1-1/edit/completion/complete_redir_test.go
--- 0.11+ds1-5/edit/completion/complete_redir_test.go	1970-01-01 00:00:00.000000000 +0000
+++ 0.12+ds1-1/edit/completion/complete_redir_test.go	2018-07-11 02:32:25.000000000 +0000
@@ -0,0 +1,12 @@
+package completion
+
+import "testing"
+
+func TestFindRedirComplContext(t *testing.T) {
+	testComplContextFinder(t, "findRedirComplContext", findRedirComplContext, []complContextFinderTest{
+		{"a >", &redirComplContext{
+			complContextCommon{"", quotingForEmptySeed, 3, 3}}},
+		{"a >b", &redirComplContext{
+			complContextCommon{"b", quotingForEmptySeed, 3, 4}}},
+	})
+}
diff -pruN 0.11+ds1-5/edit/completion/complete_variable.go 0.12+ds1-1/edit/completion/complete_variable.go
--- 0.11+ds1-5/edit/completion/complete_variable.go	1970-01-01 00:00:00.000000000 +0000
+++ 0.12+ds1-1/edit/completion/complete_variable.go	2018-07-11 02:32:25.000000000 +0000
@@ -0,0 +1,62 @@
+package completion
+
+import (
+	"strings"
+
+	"github.com/elves/elvish/eval"
+	"github.com/elves/elvish/parse"
+)
+
+type variableComplContext struct {
+	complContextCommon
+	ns, nsPart string
+}
+
+func (*variableComplContext) name() string { return "variable" }
+
+func findVariableComplContext(n parse.Node, _ pureEvaler) complContext {
+	primary := parse.GetPrimary(n)
+	if primary != nil && primary.Type == parse.Variable {
+		explode, nsPart, nameSeed := eval.SplitIncompleteVariableRef(primary.Value)
+		// Move past "$", "@" and "<ns>:".
+		begin := primary.Begin() + 1 + len(explode) + len(nsPart)
+		ns := nsPart
+		if len(ns) > 0 {
+			ns = ns[:len(ns)-1]
+		}
+		return &variableComplContext{
+			complContextCommon{nameSeed, parse.Bareword, begin, primary.End()},
+			ns, nsPart,
+		}
+	}
+	return nil
+}
+
+type evalerScopes interface {
+	EachVariableInTop(string, func(string))
+	EachNsInTop(func(string))
+}
+
+func (ctx *variableComplContext) generate(env *complEnv, ch chan<- rawCandidate) error {
+	complVariable(ctx.ns, ctx.nsPart, env.evaler, ch)
+	return nil
+}
+
+func complVariable(ctxNs, ctxNsPart string, ev evalerScopes, ch chan<- rawCandidate) {
+	ev.EachVariableInTop(ctxNs, func(varname string) {
+		ch <- noQuoteCandidate(varname)
+	})
+
+	ev.EachNsInTop(func(ns string) {
+		nsPart := ns + ":"
+		// This is to match namespaces that are "nested" under the current
+		// namespace.
+		if hasProperPrefix(nsPart, ctxNsPart) {
+			ch <- noQuoteCandidate(nsPart[len(ctxNsPart):])
+		}
+	})
+}
+
+func hasProperPrefix(s, p string) bool {
+	return len(s) > len(p) && strings.HasPrefix(s, p)
+}
diff -pruN 0.11+ds1-5/edit/completion/complete_variable_test.go 0.12+ds1-1/edit/completion/complete_variable_test.go
--- 0.11+ds1-5/edit/completion/complete_variable_test.go	1970-01-01 00:00:00.000000000 +0000
+++ 0.12+ds1-1/edit/completion/complete_variable_test.go	2018-07-11 02:32:25.000000000 +0000
@@ -0,0 +1,90 @@
+package completion
+
+import (
+	"reflect"
+	"sort"
+	"testing"
+
+	"github.com/elves/elvish/parse"
+)
+
+func TestFindVariableComplContext(t *testing.T) {
+	testComplContextFinder(t, "findVariableComplContext", findVariableComplContext, []complContextFinderTest{
+		{"$", &variableComplContext{
+			complContextCommon{"", parse.Bareword, 1, 1}, "", ""}},
+		{"$a", &variableComplContext{
+			complContextCommon{"a", parse.Bareword, 1, 2}, "", ""}},
+		{"$a:", &variableComplContext{
+			complContextCommon{"", parse.Bareword, 3, 3}, "a", "a:"}},
+		{"$a:b", &variableComplContext{
+			complContextCommon{"b", parse.Bareword, 3, 4}, "a", "a:"}},
+		// Wrong contexts
+		{"", nil},
+		{"echo", nil},
+	})
+}
+
+type testEvalerScopes struct{}
+
+var testScopes = map[string]map[string]int{
+	"":        {"veni": 0, "vidi": 0, "vici": 0},
+	"foo":     {"lorem": 0, "ipsum": 0},
+	"foo:bar": {"lorem": 0, "dolor": 0},
+}
+
+func (testEvalerScopes) EachNsInTop(f func(string)) {
+	for ns := range testScopes {
+		if ns != "" {
+			f(ns)
+		}
+	}
+}
+
+func (testEvalerScopes) EachVariableInTop(ns string, f func(string)) {
+	for name := range testScopes[ns] {
+		f(name)
+	}
+}
+
+var complVariableTests = []struct {
+	ns     string
+	nsPart string
+	want   []rawCandidate
+}{
+	// No namespace: complete variables and namespaces
+	{"", "", []rawCandidate{
+		noQuoteCandidate("foo:"), noQuoteCandidate("foo:bar:"),
+		noQuoteCandidate("veni"), noQuoteCandidate("vici"), noQuoteCandidate("vidi"),
+	}},
+	// Nonempty namespace: complete variables in namespace and subnamespaces
+	// (but not variables in subnamespaces)
+	{"foo", "foo:", []rawCandidate{
+		noQuoteCandidate("bar:"),
+		noQuoteCandidate("ipsum"), noQuoteCandidate("lorem"),
+	}},
+	// Bad namespace
+	{"bad", "bad:", nil},
+}
+
+func TestComplVariable(t *testing.T) {
+	for _, test := range complVariableTests {
+		got := collectComplVariable(test.ns, test.nsPart, testEvalerScopes{})
+		if !reflect.DeepEqual(got, test.want) {
+			t.Errorf("complVariable(%q, %q, ...) => %v, want %v", test.ns, test.nsPart, got, test.want)
+		}
+	}
+}
+
+func collectComplVariable(ns, nsPart string, ev evalerScopes) []rawCandidate {
+	ch := make(chan rawCandidate)
+	go func() {
+		complVariable(ns, nsPart, ev, ch)
+		close(ch)
+	}()
+	var results []rawCandidate
+	for result := range ch {
+		results = append(results, result)
+	}
+	sort.Sort(rawCandidates(results))
+	return results
+}
diff -pruN 0.11+ds1-5/edit/completion/completion.go 0.12+ds1-1/edit/completion/completion.go
--- 0.11+ds1-5/edit/completion/completion.go	1970-01-01 00:00:00.000000000 +0000
+++ 0.12+ds1-1/edit/completion/completion.go	2018-07-11 02:32:25.000000000 +0000
@@ -0,0 +1,127 @@
+package completion
+
+// Completion in Elvish is organized around the concept of "completers",
+// functions that take the current AST Node (the Node that the cursor is at,
+// always a leaf in the AST) and an eval.Evaler and returns a specification for
+// the completion (a complSpec) -- a list of completion candidates, and which
+// part of the source code they can **replace**. When completion is requested,
+// the editor calls each completer; it is up to the completer to decide whether
+// they apply to the current context. As soon as one completer returns results,
+// the remaining completers are not tried.
+//
+// As an example instance, if the user writes the following and presses Tab:
+//
+// echo $p
+//
+// assuming that only the builtin variables $paths, $pid and $pwd are viable
+// candidates, one of the completers -- the variable completer -- will return a
+// complSpec that means "any of paths, pid and pwd can replace the 'p' in the
+// source code".
+//
+// Note that the "replace" part in the semantics of complSpec is important: in
+// the default setting of prefix matching, it might be easier to define
+// complSpec in such a way that completers say "any of aths, id and wd can be
+// appended to the 'p' in the source code". However, this is not flexible enough
+// for alternative matching mechanism like substring matching or subsequence
+// matching, where the "seed" of completion (here, p) may not be a prefix of the
+// candidates.
+//
+// There is one completer that deserves more attention than others, the
+// completer for arguments. Unlike other completers, it delegates most of its
+// work to argument completers. See the comment in arg_completers.go for
+// details.
+
+import (
+	"github.com/elves/elvish/eval"
+	"github.com/elves/elvish/parse"
+	"github.com/elves/elvish/util"
+	"github.com/xiaq/persistent/hashmap"
+)
+
+var logger = util.GetLogger("[edit/completion] ")
+
+type complContext interface {
+	name() string
+	common() *complContextCommon
+	generate(*complEnv, chan<- rawCandidate) error
+}
+
+type complContextCommon struct {
+	seed       string
+	quoting    parse.PrimaryType
+	begin, end int
+}
+
+func (c *complContextCommon) common() *complContextCommon { return c }
+
+// complEnv contains environment information that may affect candidate
+// generation.
+type complEnv struct {
+	evaler       *eval.Evaler
+	matcher      hashmap.Map
+	argCompleter hashmap.Map
+}
+
+// complSpec is the result of a completion, meaning that any of the candidates
+// can replace the text in the interval [begin, end).
+type complSpec struct {
+	begin      int
+	end        int
+	candidates []*candidate
+}
+
+// A complContextFinder takes the current Node (always a leaf in the AST) and an
+// Evaler, and returns a complContext. If the complContext does not apply to the
+// type of the current Node, it should return nil.
+type complContextFinder func(parse.Node, pureEvaler) complContext
+
+type pureEvaler interface {
+	PurelyEvalCompound(*parse.Compound) (string, error)
+	PurelyEvalPartialCompound(cn *parse.Compound, upto *parse.Indexing) (string, error)
+	PurelyEvalPrimary(*parse.Primary) interface{}
+}
+
+var complContextFinders = []complContextFinder{
+	findVariableComplContext,
+	findCommandComplContext,
+	findIndexComplContext,
+	findRedirComplContext,
+	findArgComplContext,
+}
+
+// complete takes a Node and Evaler and tries all complContexts. It returns the
+// name of the complContext, and the result and error it gave. If no complContext is
+// available, it returns an empty complContext name.
+func complete(n parse.Node, env *complEnv) (string, *complSpec, error) {
+	for _, finder := range complContextFinders {
+		ctx := finder(n, env.evaler)
+		if ctx == nil {
+			continue
+		}
+		name := ctx.name()
+		ctxCommon := ctx.common()
+
+		matcher, ok := lookupMatcher(env.matcher, name)
+		if !ok {
+			return name, nil, errMatcherMustBeFn
+		}
+
+		chanRawCandidate := make(chan rawCandidate)
+		chanErrGenerate := make(chan error)
+		go func() {
+			err := ctx.generate(env, chanRawCandidate)
+			close(chanRawCandidate)
+			chanErrGenerate <- err
+		}()
+
+		rawCandidates, errFilter := filterRawCandidates(env.evaler, matcher, ctxCommon.seed, chanRawCandidate)
+		candidates := make([]*candidate, len(rawCandidates))
+		for i, raw := range rawCandidates {
+			candidates[i] = raw.cook(ctxCommon.quoting)
+		}
+		spec := &complSpec{ctxCommon.begin, ctxCommon.end, candidates}
+		return name, spec, util.Errors(<-chanErrGenerate, errFilter)
+
+	}
+	return "", nil, nil
+}
diff -pruN 0.11+ds1-5/edit/completion/completion_mode.go 0.12+ds1-1/edit/completion/completion_mode.go
--- 0.11+ds1-5/edit/completion/completion_mode.go	1970-01-01 00:00:00.000000000 +0000
+++ 0.12+ds1-1/edit/completion/completion_mode.go	2018-07-11 02:32:25.000000000 +0000
@@ -0,0 +1,407 @@
+package completion
+
+import (
+	"fmt"
+	"strings"
+	"unicode/utf8"
+
+	"github.com/elves/elvish/edit/eddefs"
+	"github.com/elves/elvish/edit/ui"
+	"github.com/elves/elvish/eval"
+	"github.com/elves/elvish/eval/vals"
+	"github.com/elves/elvish/eval/vars"
+	"github.com/elves/elvish/parse/parseutil"
+	"github.com/elves/elvish/util"
+	"github.com/xiaq/persistent/hashmap"
+)
+
+// Completion mode.
+
+// Interface.
+
+type completion struct {
+	binding      eddefs.BindingMap
+	matcher      hashmap.Map
+	argCompleter hashmap.Map
+	completionState
+}
+
+type completionState struct {
+	complSpec
+	completer string
+
+	filtering       bool
+	filter          string
+	filtered        []*candidate
+	selected        int
+	firstShown      int
+	lastShownInFull int
+	height          int
+}
+
+func Init(ed eddefs.Editor, ns eval.Ns) {
+	c := &completion{
+		binding:      eddefs.EmptyBindingMap,
+		matcher:      vals.MakeMapFromKV("", matchPrefix),
+		argCompleter: makeArgCompleter(),
+	}
+
+	ns.AddNs("completion",
+		eval.Ns{
+			"binding":       vars.FromPtr(&c.binding),
+			"matcher":       vars.FromPtr(&c.matcher),
+			"arg-completer": vars.FromPtr(&c.argCompleter),
+		}.AddBuiltinFns("edit:completion:", map[string]interface{}{
+			"start":          func() { c.start(ed, false) },
+			"smart-start":    func() { c.start(ed, true) },
+			"up":             func() { c.prev(false) },
+			"up-cycle":       func() { c.prev(true) },
+			"down":           func() { c.next(false) },
+			"down-cycle":     func() { c.next(true) },
+			"left":           c.left,
+			"right":          c.right,
+			"accept":         func() { c.accept(ed) },
+			"trigger-filter": c.triggerFilter,
+			"default":        func() { c.complDefault(ed) },
+		}))
+
+	// Exposing arg completers.
+	for _, v := range argCompletersData {
+		ns[v.name+eval.FnSuffix] = vars.NewRo(
+			&builtinArgCompleter{v.name, v.impl, c.argCompleter})
+	}
+
+	// Matchers.
+	ns.AddFn("match-prefix", matchPrefix)
+	ns.AddFn("match-substr", matchSubstr)
+	ns.AddFn("match-subseq", matchSubseq)
+
+	// Other functions.
+	ns.AddBuiltinFns("edit:", map[string]interface{}{
+		"complete-getopt":   complGetopt,
+		"complex-candidate": makeComplexCandidate,
+	})
+}
+
+func makeArgCompleter() hashmap.Map {
+	m := vals.EmptyMap
+	for k, v := range argCompletersData {
+		m = m.Assoc(k, &builtinArgCompleter{v.name, v.impl, m})
+	}
+	return m
+}
+
+func (c *completion) Teardown() {
+	c.completionState = completionState{}
+}
+
+func (c *completion) Binding(k ui.Key) eval.Callable {
+	return c.binding.GetOrDefault(k)
+}
+
+func (c *completion) Replacement() (int, int, string) {
+	return c.begin, c.end, c.selectedCandidate().code
+}
+
+func (*completion) RedrawModeLine() {}
+
+func (c *completion) needScrollbar() bool {
+	return c.firstShown > 0 || c.lastShownInFull < len(c.filtered)-1
+}
+
+func (c *completion) ModeLine() ui.Renderer {
+	ml := ui.NewModeLineRenderer(
+		fmt.Sprintf(" COMPLETING %s ", c.completer), c.filter)
+	if !c.needScrollbar() {
+		return ml
+	}
+	return ui.NewModeLineWithScrollBarRenderer(ml,
+		len(c.filtered), c.firstShown, c.lastShownInFull+1)
+}
+
+func (c *completion) CursorOnModeLine() bool {
+	return c.filtering
+}
+
+func (c *completion) left() {
+	if x := c.selected - c.height; x >= 0 {
+		c.selected = x
+	}
+}
+
+func (c *completion) right() {
+	if x := c.selected + c.height; x < len(c.filtered) {
+		c.selected = x
+	}
+}
+
+// acceptCompletion accepts currently selected completion candidate.
+func (c *completion) accept(ed eddefs.Editor) {
+	if 0 <= c.selected && c.selected < len(c.filtered) {
+		ed.SetBuffer(c.apply(ed.Buffer()))
+	}
+	ed.SetModeInsert()
+}
+
+func (c *completion) complDefault(ed eddefs.Editor) {
+	k := ed.LastKey()
+	if c.filtering && likeChar(k) {
+		c.changeFilter(c.filter + string(k.Rune))
+	} else if c.filtering && k == (ui.Key{ui.Backspace, 0}) {
+		_, size := utf8.DecodeLastRuneInString(c.filter)
+		if size > 0 {
+			c.changeFilter(c.filter[:len(c.filter)-size])
+		}
+	} else {
+		c.accept(ed)
+		ed.SetAction(eddefs.ReprocessKey)
+	}
+}
+
+func (c *completion) triggerFilter() {
+	if c.filtering {
+		c.filtering = false
+		c.changeFilter("")
+	} else {
+		c.filtering = true
+	}
+}
+
+func (c *completion) selectedCandidate() *candidate {
+	if c.selected == -1 {
+		return &candidate{}
+	}
+	return c.filtered[c.selected]
+}
+
+// apply returns the line and dot after applying a candidate.
+func (c *completion) apply(line string, dot int) (string, int) {
+	text := c.selectedCandidate().code
+	return line[:c.begin] + text + line[c.end:], c.begin + len(text)
+}
+
+func (c *completion) prev(cycle bool) {
+	c.selected--
+	if c.selected == -1 {
+		if cycle {
+			c.selected = len(c.filtered) - 1
+		} else {
+			c.selected++
+		}
+	}
+}
+
+func (c *completion) next(cycle bool) {
+	c.selected++
+	if c.selected == len(c.filtered) {
+		if cycle {
+			c.selected = 0
+		} else {
+			c.selected--
+		}
+	}
+}
+
+func (c *completion) start(ed eddefs.Editor, acceptSingleton bool) {
+	_, dot := ed.Buffer()
+	chunk := ed.ParsedBuffer()
+	node := parseutil.FindLeafNode(chunk, dot)
+	if node == nil {
+		return
+	}
+
+	completer, complSpec, err := complete(
+		node, &complEnv{ed.Evaler(), c.matcher, c.argCompleter})
+
+	if err != nil {
+		ed.AddTip("%v", err)
+		// We don't show the full stack trace. To make debugging still possible,
+		// we log it.
+		if pprinter, ok := err.(util.Pprinter); ok {
+			logger.Println("matcher error:")
+			logger.Println(pprinter.Pprint(""))
+		}
+	} else if completer == "" {
+		ed.AddTip("unsupported completion :(")
+		logger.Println("path to current leaf, leaf first")
+		for n := node; n != nil; n = n.Parent() {
+			logger.Printf("%T (%d-%d)", n, n.Begin(), n.End())
+		}
+	} else if len(complSpec.candidates) == 0 {
+		ed.AddTip("no candidate for %s", completer)
+	} else {
+		if acceptSingleton && len(complSpec.candidates) == 1 {
+			// Just accept this single candidate.
+			repl := complSpec.candidates[0].code
+			buffer, _ := ed.Buffer()
+			ed.SetBuffer(
+				buffer[:complSpec.begin]+repl+buffer[complSpec.end:],
+				complSpec.begin+len(repl))
+			return
+		}
+		c.completionState = completionState{
+			completer: completer,
+			complSpec: *complSpec,
+			filtered:  complSpec.candidates,
+		}
+		ed.SetMode(c)
+	}
+}
+
+// commonPrefix returns the longest common prefix of two strings.
+func commonPrefix(s, t string) string {
+	for i, r := range s {
+		if i >= len(t) {
+			return s[:i]
+		}
+		r2, _ := utf8.DecodeRuneInString(t[i:])
+		if r2 != r {
+			return s[:i]
+		}
+	}
+	return s
+}
+
+const (
+	completionColMarginLeft  = 1
+	completionColMarginRight = 1
+	completionColMarginTotal = completionColMarginLeft + completionColMarginRight
+)
+
+// maxWidth finds the maximum wcwidth of display texts of candidates [lo, hi).
+// hi may be larger than the number of candidates, in which case it is truncated
+// to the number of candidates.
+func (c *completion) maxWidth(lo, hi int) int {
+	if hi > len(c.filtered) {
+		hi = len(c.filtered)
+	}
+	width := 0
+	for i := lo; i < hi; i++ {
+		w := util.Wcswidth(c.filtered[i].menu.Text)
+		if width < w {
+			width = w
+		}
+	}
+	return width
+}
+
+func (c *completion) ListRender(width, maxHeight int) *ui.Buffer {
+	bb := ui.NewBufferBuilder(width)
+	cands := c.filtered
+	if len(cands) == 0 {
+		bb.WriteString(util.TrimWcwidth("(no result)", width), "")
+		return bb.Buffer()
+	}
+	if maxHeight <= 1 || width <= 2 {
+		bb.WriteString(util.TrimWcwidth("(terminal too small)", width), "")
+		return bb.Buffer()
+	}
+
+	// Reserve the the rightmost row as margins.
+	width--
+
+	// Determine comp.height and comp.firstShown.
+	// First determine whether all candidates can be fit in the screen,
+	// assuming that they are all of maximum width. If that is the case, we use
+	// the computed height as the height for the listing, and the first
+	// candidate to show is 0. Otherwise, we use min(height, len(cands)) as the
+	// height and find the first candidate to show.
+	perLine := max(1, width/(c.maxWidth(0, len(cands))+completionColMarginTotal))
+	heightBound := util.CeilDiv(len(cands), perLine)
+	first := 0
+	height := 0
+	if heightBound < maxHeight {
+		height = heightBound
+	} else {
+		height = min(maxHeight, len(cands))
+		// Determine the first column to show. We start with the column in which the
+		// selected one is found, moving to the left until either the width is
+		// exhausted, or the old value of firstShown has been hit.
+		first = c.selected / height * height
+		w := c.maxWidth(first, first+height) + completionColMarginTotal
+		for ; first > c.firstShown; first -= height {
+			dw := c.maxWidth(first-height, first) + completionColMarginTotal
+			if w+dw > width {
+				break
+			}
+			w += dw
+		}
+	}
+	c.height = height
+	c.firstShown = first
+
+	var i, j int
+	remainedWidth := width
+	trimmed := false
+	// Show the results in columns, until width is exceeded.
+	for i = first; i < len(cands); i += height {
+		// Determine the width of the column (without the margin)
+		colWidth := c.maxWidth(i, min(i+height, len(cands)))
+		totalColWidth := colWidth + completionColMarginTotal
+		if totalColWidth > remainedWidth {
+			totalColWidth = remainedWidth
+			colWidth = totalColWidth - completionColMarginTotal
+			trimmed = true
+		}
+
+		col := ui.NewBufferBuilder(totalColWidth)
+		for j = i; j < i+height; j++ {
+			if j > i {
+				col.Newline()
+			}
+			if j >= len(cands) {
+				// Write padding to make the listing a rectangle.
+				col.WriteSpaces(totalColWidth, styleForCompletion.String())
+			} else {
+				col.WriteSpaces(completionColMarginLeft, styleForCompletion.String())
+				s := ui.JoinStyles(styleForCompletion, cands[j].menu.Styles)
+				if j == c.selected {
+					s = append(s, styleForSelectedCompletion.String())
+				}
+				col.WriteString(util.ForceWcwidth(cands[j].menu.Text, colWidth), s.String())
+				col.WriteSpaces(completionColMarginRight, styleForCompletion.String())
+				if !trimmed {
+					c.lastShownInFull = j
+				}
+			}
+		}
+
+		bb.ExtendRight(col.Buffer(), 0)
+		remainedWidth -= totalColWidth
+		if remainedWidth <= completionColMarginTotal {
+			break
+		}
+	}
+	// When the listing is incomplete, always use up the entire width.
+	if remainedWidth > 0 && c.needScrollbar() {
+		col := ui.NewBufferBuilder(remainedWidth)
+		for i := 0; i < height; i++ {
+			if i > 0 {
+				col.Newline()
+			}
+			col.WriteSpaces(remainedWidth, styleForCompletion.String())
+		}
+		bb.ExtendRight(col.Buffer(), 0)
+		remainedWidth = 0
+	}
+	return bb.Buffer()
+}
+
+func (c *completion) changeFilter(f string) {
+	c.filter = f
+	if f == "" {
+		c.filtered = c.candidates
+		return
+	}
+	c.filtered = nil
+	for _, cand := range c.candidates {
+		if strings.Contains(cand.menu.Text, f) {
+			c.filtered = append(c.filtered, cand)
+		}
+	}
+	if len(c.filtered) > 0 {
+		c.selected = 0
+	} else {
+		c.selected = -1
+	}
+}
diff -pruN 0.11+ds1-5/edit/completion/completion_test.go 0.12+ds1-1/edit/completion/completion_test.go
--- 0.11+ds1-5/edit/completion/completion_test.go	1970-01-01 00:00:00.000000000 +0000
+++ 0.12+ds1-1/edit/completion/completion_test.go	2018-07-11 02:32:25.000000000 +0000
@@ -0,0 +1,65 @@
+package completion
+
+import (
+	"reflect"
+	"testing"
+
+	"github.com/elves/elvish/eval"
+	"github.com/elves/elvish/parse"
+	"github.com/elves/elvish/parse/parseutil"
+)
+
+type complContextFinderTest struct {
+	src  string
+	want complContext
+}
+
+func testComplContextFinder(t *testing.T, name string, finder complContextFinder, tests []complContextFinderTest) {
+	ev := eval.NewEvaler()
+	defer ev.Close()
+	for _, test := range tests {
+		n, err := parse.Parse("[test]", test.src)
+		// Ignore error as long is n is non-nil
+		if n == nil {
+			panic(err)
+		}
+		leaf := parseutil.FindLeafNode(n, len(test.src))
+		got := finder(leaf, ev)
+		if !reflect.DeepEqual(got, test.want) {
+			t.Errorf("For %q, %s(leaf) => %v, want %v", test.src, name, got, test.want)
+		}
+	}
+}
+
+type filterRawCandidatesTest struct {
+	name    string
+	matcher eval.Callable
+	seed    string
+	src     []string
+	want    []string
+}
+
+func testRawFilterCandidates(t *testing.T, tests []filterRawCandidatesTest) {
+	ev := eval.NewEvaler()
+	defer ev.Close()
+	for _, test := range tests {
+		input := make(chan rawCandidate, len(test.src))
+		for _, s := range test.src {
+			input <- plainCandidate(s)
+		}
+		close(input)
+
+		result, err := filterRawCandidates(ev, test.matcher, test.seed, input)
+		if err != nil {
+			t.Errorf("For %s, got unexpected error %v", test.name, err)
+		}
+
+		got := make([]string, len(result))
+		for i, r := range result {
+			got[i] = r.text()
+		}
+		if !reflect.DeepEqual(got, test.want) {
+			t.Errorf("For %s, got unexpected result => %v, want %v", test.name, got, test.want)
+		}
+	}
+}
diff -pruN 0.11+ds1-5/edit/completion/compl_getopt.go 0.12+ds1-1/edit/completion/compl_getopt.go
--- 0.11+ds1-5/edit/completion/compl_getopt.go	1970-01-01 00:00:00.000000000 +0000
+++ 0.12+ds1-1/edit/completion/compl_getopt.go	2018-07-11 02:32:25.000000000 +0000
@@ -0,0 +1,195 @@
+package completion
+
+import (
+	"strings"
+	"unicode/utf8"
+
+	"github.com/elves/elvish/eval"
+	"github.com/elves/elvish/eval/vals"
+	"github.com/elves/elvish/getopt"
+	"github.com/elves/elvish/parse"
+	"github.com/xiaq/persistent/hashmap"
+)
+
+func complGetopt(fm *eval.Frame, elemsv, optsv, argsv interface{}) {
+	var (
+		elems    []string
+		opts     []*getopt.Option
+		args     []eval.Callable
+		variadic bool
+	)
+	desc := make(map[*getopt.Option]string)
+	argdesc := make(map[*getopt.Option]string)
+	// Convert arguments.
+	err := vals.Iterate(elemsv, func(v interface{}) bool {
+		elem, ok := v.(string)
+		if !ok {
+			throwf("arg should be string, got %s", vals.Kind(v))
+		}
+		elems = append(elems, elem)
+		return true
+	})
+	maybeThrow(err)
+	err = vals.Iterate(optsv, func(v interface{}) bool {
+		m, ok := v.(hashmap.Map)
+		if !ok {
+			throwf("opt should be map, got %s", vals.Kind(v))
+		}
+
+		opt := &getopt.Option{}
+
+		getStringField := func(k string) (string, bool) {
+			v, ok := m.Index(k)
+			if !ok {
+				return "", false
+			}
+			if vs, ok := v.(string); ok {
+				return vs, true
+			}
+			throwf("%s should be string, got %s", k, vals.Kind(v))
+			panic("unreachable")
+		}
+		if s, ok := getStringField("short"); ok {
+			r, size := utf8.DecodeRuneInString(s)
+			if r == utf8.RuneError || size != len(s) {
+				throwf("short option should be exactly one rune, got %v", parse.Quote(s))
+			}
+			opt.Short = r
+		}
+		if s, ok := getStringField("long"); ok {
+			opt.Long = s
+		}
+		if opt.Short == 0 && opt.Long == "" {
+			throwf("opt should have at least one of short and long forms")
+		}
+
+		getBoolField := func(k string) (bool, bool) {
+			v, ok := m.Index(k)
+			if !ok {
+				return false, false
+			}
+			if vb, ok := v.(bool); ok {
+				return vb, true
+			}
+			throwf("%s should be bool, got %s", k, vals.Kind(v))
+			panic("unreachable")
+		}
+		argRequired, _ := getBoolField("arg-required")
+		argOptional, _ := getBoolField("arg-optional")
+		switch {
+		case argRequired && argOptional:
+			throwf("opt cannot have both arg-required and arg-optional")
+		case argRequired:
+			opt.HasArg = getopt.RequiredArgument
+		case argOptional:
+			opt.HasArg = getopt.OptionalArgument
+		}
+
+		if s, ok := getStringField("desc"); ok {
+			desc[opt] = s
+		}
+		if s, ok := getStringField("arg-desc"); ok {
+			argdesc[opt] = s
+		}
+		opts = append(opts, opt)
+		return true
+	})
+	maybeThrow(err)
+	err = vals.Iterate(argsv, func(v interface{}) bool {
+		sv, ok := v.(string)
+		if ok {
+			if sv == "..." {
+				variadic = true
+				return true
+			}
+			throwf("string except for ... not allowed as argument handler, got %s", parse.Quote(sv))
+		}
+		arg, ok := v.(eval.Callable)
+		if !ok {
+			throwf("argument handler should be fn, got %s", vals.Kind(v))
+		}
+		args = append(args, arg)
+		return true
+	})
+	maybeThrow(err)
+
+	// TODO Configurable config
+	g := getopt.Getopt{opts, getopt.GNUGetoptLong}
+	_, parsedArgs, ctx := g.Parse(elems)
+	out := fm.OutputChan()
+
+	putShortOpt := func(opt *getopt.Option) {
+		c := &complexCandidate{stem: "-" + string(opt.Short)}
+		if d, ok := desc[opt]; ok {
+			if e, ok := argdesc[opt]; ok {
+				c.displaySuffix = " " + e + " (" + d + ")"
+			} else {
+				c.displaySuffix = " (" + d + ")"
+			}
+		}
+		out <- c
+	}
+	putLongOpt := func(opt *getopt.Option) {
+		c := &complexCandidate{stem: "--" + opt.Long}
+		if d, ok := desc[opt]; ok {
+			if e, ok := argdesc[opt]; ok {
+				c.displaySuffix = " " + e + " (" + d + ")"
+			} else {
+				c.displaySuffix = " (" + d + ")"
+			}
+		}
+		out <- c
+	}
+
+	switch ctx.Type {
+	case getopt.NewOptionOrArgument, getopt.Argument:
+		// Find argument completer
+		var argCompl eval.Callable
+		if len(parsedArgs) < len(args) {
+			argCompl = args[len(parsedArgs)]
+		} else if variadic {
+			argCompl = args[len(args)-1]
+		}
+		if argCompl != nil {
+			rawCands := make(chan rawCandidate)
+			defer close(rawCands)
+			go func() {
+				for rc := range rawCands {
+					out <- rc
+				}
+			}()
+			err := callArgCompleter(argCompl, fm.Evaler, []string{ctx.Text}, rawCands)
+			maybeThrow(err)
+		}
+		// TODO Notify that there is no suitable argument completer
+	case getopt.NewOption:
+		for _, opt := range opts {
+			if opt.Short != 0 {
+				putShortOpt(opt)
+			}
+			if opt.Long != "" {
+				putLongOpt(opt)
+			}
+		}
+	case getopt.NewLongOption:
+		for _, opt := range opts {
+			if opt.Long != "" {
+				putLongOpt(opt)
+			}
+		}
+	case getopt.LongOption:
+		for _, opt := range opts {
+			if strings.HasPrefix(opt.Long, ctx.Text) {
+				putLongOpt(opt)
+			}
+		}
+	case getopt.ChainShortOption:
+		for _, opt := range opts {
+			if opt.Short != 0 {
+				// XXX loses chained options
+				putShortOpt(opt)
+			}
+		}
+	case getopt.OptionArgument:
+	}
+}
diff -pruN 0.11+ds1-5/edit/completion/matcher.go 0.12+ds1-1/edit/completion/matcher.go
--- 0.11+ds1-5/edit/completion/matcher.go	1970-01-01 00:00:00.000000000 +0000
+++ 0.12+ds1-1/edit/completion/matcher.go	2018-07-11 02:32:25.000000000 +0000
@@ -0,0 +1,80 @@
+package completion
+
+import (
+	"errors"
+	"strings"
+
+	"github.com/elves/elvish/eval"
+	"github.com/elves/elvish/eval/vals"
+	"github.com/elves/elvish/util"
+	"github.com/xiaq/persistent/hashmap"
+)
+
+var (
+	errIncorrectNumOfResults    = errors.New("matcher must return a bool for each candidate")
+	errMatcherMustBeFn          = errors.New("matcher must be a function")
+	errMatcherInputMustBeString = errors.New("matcher input must be string")
+)
+
+var (
+	matchPrefix = eval.NewBuiltinFn(
+		"edit:match-prefix", wrapMatcher(strings.HasPrefix))
+	matchSubstr = eval.NewBuiltinFn(
+		"edit:match-substr", wrapMatcher(strings.Contains))
+	matchSubseq = eval.NewBuiltinFn(
+		"edit:match-subseq", wrapMatcher(util.HasSubseq))
+)
+
+func lookupMatcher(m hashmap.Map, name string) (eval.Callable, bool) {
+	key := name
+	if !hashmap.HasKey(m, key) {
+		// Use fallback matcher
+		if !hashmap.HasKey(m, "") {
+			return nil, false
+		}
+		key = ""
+	}
+	value, _ := m.Index(key)
+	matcher, ok := value.(eval.Callable)
+	return matcher, ok
+}
+
+func wrapMatcher(matcher func(s, p string) bool) interface{} {
+	return func(fm *eval.Frame,
+		opts eval.RawOptions, pattern string, inputs eval.Inputs) {
+
+		var options struct {
+			IgnoreCase bool
+			SmartCase  bool
+		}
+		opts.Scan(&options)
+		switch {
+		case options.IgnoreCase && options.SmartCase:
+			throwf("-ignore-case and -smart-case cannot be used together")
+		case options.IgnoreCase:
+			innerMatcher := matcher
+			matcher = func(s, p string) bool {
+				return innerMatcher(strings.ToLower(s), strings.ToLower(p))
+			}
+		case options.SmartCase:
+			innerMatcher := matcher
+			matcher = func(s, p string) bool {
+				if p == strings.ToLower(p) {
+					// Ignore case is pattern is all lower case.
+					return innerMatcher(strings.ToLower(s), p)
+				} else {
+					return innerMatcher(s, p)
+				}
+			}
+		}
+
+		out := fm.OutputChan()
+		inputs(func(v interface{}) {
+			s, ok := v.(string)
+			if !ok {
+				throw(errMatcherInputMustBeString)
+			}
+			out <- vals.Bool(matcher(s, pattern))
+		})
+	}
+}
diff -pruN 0.11+ds1-5/edit/completion/matcher_test.go 0.12+ds1-1/edit/completion/matcher_test.go
--- 0.11+ds1-5/edit/completion/matcher_test.go	1970-01-01 00:00:00.000000000 +0000
+++ 0.12+ds1-1/edit/completion/matcher_test.go	2018-07-11 02:32:25.000000000 +0000
@@ -0,0 +1,33 @@
+package completion
+
+import (
+	"testing"
+)
+
+func TestBuiltinMatcher(t *testing.T) {
+	tests := []filterRawCandidatesTest{
+		{
+			name:    "matchPrefix",
+			matcher: matchPrefix,
+			seed:    "x",
+			src:     []string{"x1", "x2", "3"},
+			want:    []string{"x1", "x2"},
+		},
+		{
+			name:    "matchSubstr",
+			matcher: matchSubstr,
+			seed:    "x",
+			src:     []string{"1x", "2x", "3"},
+			want:    []string{"1x", "2x"},
+		},
+		{
+			name:    "matchSubseq",
+			matcher: matchSubseq,
+			seed:    "xyz",
+			src:     []string{"1xyz1", "2abc2", "123"},
+			want:    []string{"1xyz1"},
+		},
+	}
+
+	testRawFilterCandidates(t, tests)
+}
diff -pruN 0.11+ds1-5/edit/completion/nodeutil.go 0.12+ds1-1/edit/completion/nodeutil.go
--- 0.11+ds1-5/edit/completion/nodeutil.go	1970-01-01 00:00:00.000000000 +0000
+++ 0.12+ds1-1/edit/completion/nodeutil.go	2018-07-11 02:32:25.000000000 +0000
@@ -0,0 +1,23 @@
+package completion
+
+import (
+	"github.com/elves/elvish/parse"
+)
+
+// Utilities for insepcting the AST. Used for completers and stylists.
+
+func primaryInSimpleCompound(pn *parse.Primary, ev pureEvaler) (*parse.Compound, string) {
+	indexing := parse.GetIndexing(pn.Parent())
+	if indexing == nil {
+		return nil, ""
+	}
+	compound := parse.GetCompound(indexing.Parent())
+	if compound == nil {
+		return nil, ""
+	}
+	head, err := ev.PurelyEvalPartialCompound(compound, indexing)
+	if err != nil {
+		return nil, ""
+	}
+	return compound, head
+}
diff -pruN 0.11+ds1-5/edit/completion/utils.go 0.12+ds1-1/edit/completion/utils.go
--- 0.11+ds1-5/edit/completion/utils.go	1970-01-01 00:00:00.000000000 +0000
+++ 0.12+ds1-1/edit/completion/utils.go	2018-07-11 02:32:25.000000000 +0000
@@ -0,0 +1,51 @@
+package completion
+
+import (
+	"fmt"
+	"unicode"
+
+	"github.com/elves/elvish/edit/ui"
+	"github.com/elves/elvish/util"
+)
+
+func throw(e error) {
+	util.Throw(e)
+}
+
+func maybeThrow(e error) {
+	if e != nil {
+		util.Throw(e)
+	}
+}
+
+func throwf(format string, args ...interface{}) {
+	util.Throw(fmt.Errorf(format, args...))
+}
+
+func min(a, b int) int {
+	if a <= b {
+		return a
+	}
+	return b
+}
+
+func max(a, b int) int {
+	if a >= b {
+		return a
+	}
+	return b
+}
+
+// Styles for UI.
+var (
+	// Use default style for completion listing
+	styleForCompletion = ui.Styles{}
+	// Use inverse style for selected completion entry
+	styleForSelectedCompletion = ui.Styles{"inverse"}
+)
+
+// likeChar returns if a key looks like a character meant to be input (as
+// opposed to a function key).
+func likeChar(k ui.Key) bool {
+	return k.Mod == 0 && k.Rune > 0 && unicode.IsGraphic(k.Rune)
+}
diff -pruN 0.11+ds1-5/edit/completion.go 0.12+ds1-1/edit/completion.go
--- 0.11+ds1-5/edit/completion.go	2018-01-18 02:03:11.000000000 +0000
+++ 0.12+ds1-1/edit/completion.go	1970-01-01 00:00:00.000000000 +0000
@@ -1,118 +0,0 @@
-package edit
-
-// Completion in Elvish is organized around the concept of "completers",
-// functions that take the current AST Node (the Node that the cursor is at,
-// always a leaf in the AST) and an eval.Evaler and returns a specification for
-// the completion (a complSpec) -- a list of completion candidates, and which
-// part of the source code they can **replace**. When completion is requested,
-// the editor calls each completer; it is up to the completer to decide whether
-// they apply to the current context. As soon as one completer returns results,
-// the remaining completers are not tried.
-//
-// As an example instance, if the user writes the following and presses Tab:
-//
-// echo $p
-//
-// assuming that only the builtin variables $paths, $pid and $pwd are viable
-// candidates, one of the completers -- the variable completer -- will return a
-// complSpec that means "any of paths, pid and pwd can replace the 'p' in the
-// source code".
-//
-// Note that the "replace" part in the semantics of complSpec is important: in
-// the default setting of prefix matching, it might be easier to define
-// complSpec in such a way that completers say "any of aths, id and wd can be
-// appended to the 'p' in the source code". However, this is not flexible enough
-// for alternative matching mechanism like substring matching or subsequence
-// matching, where the "seed" of completion (here, p) may not be a prefix of the
-// candidates.
-//
-// There is one completer that deserves more attention than others, the
-// completer for arguments. Unlike other completers, it delegates most of its
-// work to argument completers. See the comment in arg_completers.go for
-// details.
-
-import (
-	"github.com/elves/elvish/eval"
-	"github.com/elves/elvish/eval/types"
-	"github.com/elves/elvish/parse"
-	"github.com/elves/elvish/util"
-)
-
-type complContext interface {
-	name() string
-	common() *complContextCommon
-	generate(*eval.Evaler, chan<- rawCandidate) error
-}
-
-type complContextCommon struct {
-	seed       string
-	quoting    parse.PrimaryType
-	begin, end int
-}
-
-func (c *complContextCommon) common() *complContextCommon { return c }
-
-// complSpec is the result of a completion, meaning that any of the candidates
-// can replace the text in the interval [begin, end).
-type complSpec struct {
-	begin      int
-	end        int
-	candidates []*candidate
-}
-
-// A complContextFinder takes the current Node (always a leaf in the AST) and an
-// Evaler, and returns a complContext. If the complContext does not apply to the
-// type of the current Node, it should return nil.
-type complContextFinder func(parse.Node, pureEvaler) complContext
-
-type pureEvaler interface {
-	PurelyEvalCompound(*parse.Compound) (string, error)
-	PurelyEvalPartialCompound(cn *parse.Compound, upto *parse.Indexing) (string, error)
-	PurelyEvalPrimary(*parse.Primary) types.Value
-}
-
-var complContextFinders = []complContextFinder{
-	findVariableComplContext,
-	findCommandComplContext,
-	findIndexComplContext,
-	findRedirComplContext,
-	findArgComplContext,
-}
-
-// complete takes a Node and Evaler and tries all complContexts. It returns the
-// name of the complContext, and the result and error it gave. If no complContext is
-// available, it returns an empty complContext name.
-func complete(n parse.Node, ev *eval.Evaler) (string, *complSpec, error) {
-	ed := ev.Editor.(*Editor)
-	for _, finder := range complContextFinders {
-		ctx := finder(n, ev)
-		if ctx == nil {
-			continue
-		}
-		name := ctx.name()
-		ctxCommon := ctx.common()
-
-		matcher, ok := ed.lookupMatcher(name)
-		if !ok {
-			return name, nil, errMatcherMustBeFn
-		}
-
-		chanRawCandidate := make(chan rawCandidate)
-		chanErrGenerate := make(chan error)
-		go func() {
-			err := ctx.generate(ev, chanRawCandidate)
-			close(chanRawCandidate)
-			chanErrGenerate <- err
-		}()
-
-		rawCandidates, errFilter := filterRawCandidates(ev, matcher, ctxCommon.seed, chanRawCandidate)
-		candidates := make([]*candidate, len(rawCandidates))
-		for i, raw := range rawCandidates {
-			candidates[i] = raw.cook(ctxCommon.quoting)
-		}
-		spec := &complSpec{ctxCommon.begin, ctxCommon.end, candidates}
-		return name, spec, util.Errors(<-chanErrGenerate, errFilter)
-
-	}
-	return "", nil, nil
-}
diff -pruN 0.11+ds1-5/edit/completion_mode.go 0.12+ds1-1/edit/completion_mode.go
--- 0.11+ds1-5/edit/completion_mode.go	2018-01-18 02:03:11.000000000 +0000
+++ 0.12+ds1-1/edit/completion_mode.go	1970-01-01 00:00:00.000000000 +0000
@@ -1,383 +0,0 @@
-package edit
-
-import (
-	"fmt"
-	"strings"
-	"unicode/utf8"
-
-	"github.com/elves/elvish/edit/ui"
-	"github.com/elves/elvish/eval"
-	"github.com/elves/elvish/eval/vartypes"
-	"github.com/elves/elvish/util"
-)
-
-// Completion mode.
-
-// Interface.
-
-var _ = registerBuiltins(modeCompletion, map[string]func(*Editor){
-	"smart-start":    complSmartStart,
-	"start":          complStart,
-	"up":             complUp,
-	"up-cycle":       complUpCycle,
-	"down":           complDown,
-	"down-cycle":     complDownCycle,
-	"left":           complLeft,
-	"right":          complRight,
-	"accept":         complAccept,
-	"trigger-filter": complTriggerFilter,
-	"default":        complDefault,
-})
-
-type completion struct {
-	complSpec
-	completer string
-
-	filtering       bool
-	filter          string
-	filtered        []*candidate
-	selected        int
-	firstShown      int
-	lastShownInFull int
-	height          int
-}
-
-func (*completion) Binding(m map[string]vartypes.Variable, k ui.Key) eval.Fn {
-	return getBinding(m[modeCompletion], k)
-}
-
-func (c *completion) needScrollbar() bool {
-	return c.firstShown > 0 || c.lastShownInFull < len(c.filtered)-1
-}
-
-func (c *completion) ModeLine() ui.Renderer {
-	ml := modeLineRenderer{fmt.Sprintf(" COMPLETING %s ", c.completer), c.filter}
-	if !c.needScrollbar() {
-		return ml
-	}
-	return modeLineWithScrollBarRenderer{ml,
-		len(c.filtered), c.firstShown, c.lastShownInFull + 1}
-}
-
-func (c *completion) CursorOnModeLine() bool {
-	return c.filtering
-}
-
-func complStart(ed *Editor) {
-	startCompletionInner(ed, false)
-}
-
-func complSmartStart(ed *Editor) {
-	startCompletionInner(ed, true)
-}
-
-func complUp(ed *Editor) {
-	ed.completion.prev(false)
-}
-
-func complDown(ed *Editor) {
-	ed.completion.next(false)
-}
-
-func complLeft(ed *Editor) {
-	if c := ed.completion.selected - ed.completion.height; c >= 0 {
-		ed.completion.selected = c
-	}
-}
-
-func complRight(ed *Editor) {
-	if c := ed.completion.selected + ed.completion.height; c < len(ed.completion.filtered) {
-		ed.completion.selected = c
-	}
-}
-
-func complUpCycle(ed *Editor) {
-	ed.completion.prev(true)
-}
-
-func complDownCycle(ed *Editor) {
-	ed.completion.next(true)
-}
-
-// acceptCompletion accepts currently selected completion candidate.
-func complAccept(ed *Editor) {
-	c := ed.completion
-	if 0 <= c.selected && c.selected < len(c.filtered) {
-		ed.buffer, ed.dot = c.apply(ed.buffer, ed.dot)
-	}
-	ed.mode = &ed.insert
-}
-
-func complDefault(ed *Editor) {
-	k := ed.lastKey
-	c := &ed.completion
-	if c.filtering && likeChar(k) {
-		c.changeFilter(c.filter + string(k.Rune))
-	} else if c.filtering && k == (ui.Key{ui.Backspace, 0}) {
-		_, size := utf8.DecodeLastRuneInString(c.filter)
-		if size > 0 {
-			c.changeFilter(c.filter[:len(c.filter)-size])
-		}
-	} else {
-		complAccept(ed)
-		ed.setAction(reprocessKey)
-	}
-}
-
-func complTriggerFilter(ed *Editor) {
-	c := &ed.completion
-	if c.filtering {
-		c.filtering = false
-		c.changeFilter("")
-	} else {
-		c.filtering = true
-	}
-}
-
-func (c *completion) selectedCandidate() *candidate {
-	if c.selected == -1 {
-		return &candidate{}
-	}
-	return c.filtered[c.selected]
-}
-
-// apply returns the line and dot after applying a candidate.
-func (c *completion) apply(line string, dot int) (string, int) {
-	text := c.selectedCandidate().code
-	return line[:c.begin] + text + line[c.end:], c.begin + len(text)
-}
-
-func (c *completion) prev(cycle bool) {
-	c.selected--
-	if c.selected == -1 {
-		if cycle {
-			c.selected = len(c.filtered) - 1
-		} else {
-			c.selected++
-		}
-	}
-}
-
-func (c *completion) next(cycle bool) {
-	c.selected++
-	if c.selected == len(c.filtered) {
-		if cycle {
-			c.selected = 0
-		} else {
-			c.selected--
-		}
-	}
-}
-
-func startCompletionInner(ed *Editor, acceptPrefix bool) {
-	node := findLeafNode(ed.chunk, ed.dot)
-	if node == nil {
-		return
-	}
-
-	completer, complSpec, err := complete(node, ed.evaler)
-
-	if err != nil {
-		ed.addTip("%v", err)
-		// We don't show the full stack trace. To make debugging still possible,
-		// we log it.
-		if pprinter, ok := err.(util.Pprinter); ok {
-			logger.Println("matcher error:")
-			logger.Println(pprinter.Pprint(""))
-		}
-	} else if completer == "" {
-		ed.addTip("unsupported completion :(")
-		logger.Println("path to current leaf, leaf first")
-		for n := node; n != nil; n = n.Parent() {
-			logger.Printf("%T (%d-%d)", n, n.Begin(), n.End())
-		}
-	} else if len(complSpec.candidates) == 0 {
-		ed.addTip("no candidate for %s", completer)
-	} else {
-		if acceptPrefix {
-			// If there is a non-empty longest common prefix, insert it and
-			// don't start completion mode.
-			//
-			// As a special case, when there is exactly one candidate, it is
-			// immeidately accepted.
-			prefix := complSpec.candidates[0].code
-			for _, cand := range complSpec.candidates[1:] {
-				prefix = commonPrefix(prefix, cand.code)
-				if prefix == "" {
-					break
-				}
-			}
-
-			if prefix != "" && len(prefix) > complSpec.end-complSpec.begin {
-				ed.buffer = ed.buffer[:complSpec.begin] + prefix + ed.buffer[complSpec.end:]
-				ed.dot = complSpec.begin + len(prefix)
-
-				return
-			}
-		}
-		ed.completion = completion{
-			completer: completer,
-			complSpec: *complSpec,
-			filtered:  complSpec.candidates,
-		}
-		ed.mode = &ed.completion
-	}
-}
-
-// commonPrefix returns the longest common prefix of two strings.
-func commonPrefix(s, t string) string {
-	for i, r := range s {
-		if i >= len(t) {
-			return s[:i]
-		}
-		r2, _ := utf8.DecodeRuneInString(t[i:])
-		if r2 != r {
-			return s[:i]
-		}
-	}
-	return s
-}
-
-const (
-	completionColMarginLeft  = 1
-	completionColMarginRight = 1
-	completionColMarginTotal = completionColMarginLeft + completionColMarginRight
-)
-
-// maxWidth finds the maximum wcwidth of display texts of candidates [lo, hi).
-// hi may be larger than the number of candidates, in which case it is truncated
-// to the number of candidates.
-func (c *completion) maxWidth(lo, hi int) int {
-	if hi > len(c.filtered) {
-		hi = len(c.filtered)
-	}
-	width := 0
-	for i := lo; i < hi; i++ {
-		w := util.Wcswidth(c.filtered[i].menu.Text)
-		if width < w {
-			width = w
-		}
-	}
-	return width
-}
-
-func (c *completion) ListRender(width, maxHeight int) *ui.Buffer {
-	b := ui.NewBuffer(width)
-	cands := c.filtered
-	if len(cands) == 0 {
-		b.WriteString(util.TrimWcwidth("(no result)", width), "")
-		return b
-	}
-	if maxHeight <= 1 || width <= 2 {
-		b.WriteString(util.TrimWcwidth("(terminal too small)", width), "")
-		return b
-	}
-
-	// Reserve the the rightmost row as margins.
-	width--
-
-	// Determine comp.height and comp.firstShown.
-	// First determine whether all candidates can be fit in the screen,
-	// assuming that they are all of maximum width. If that is the case, we use
-	// the computed height as the height for the listing, and the first
-	// candidate to show is 0. Otherwise, we use min(height, len(cands)) as the
-	// height and find the first candidate to show.
-	perLine := max(1, width/(c.maxWidth(0, len(cands))+completionColMarginTotal))
-	heightBound := util.CeilDiv(len(cands), perLine)
-	first := 0
-	height := 0
-	if heightBound < maxHeight {
-		height = heightBound
-	} else {
-		height = min(maxHeight, len(cands))
-		// Determine the first column to show. We start with the column in which the
-		// selected one is found, moving to the left until either the width is
-		// exhausted, or the old value of firstShown has been hit.
-		first = c.selected / height * height
-		w := c.maxWidth(first, first+height) + completionColMarginTotal
-		for ; first > c.firstShown; first -= height {
-			dw := c.maxWidth(first-height, first) + completionColMarginTotal
-			if w+dw > width {
-				break
-			}
-			w += dw
-		}
-	}
-	c.height = height
-	c.firstShown = first
-
-	var i, j int
-	remainedWidth := width
-	trimmed := false
-	// Show the results in columns, until width is exceeded.
-	for i = first; i < len(cands); i += height {
-		// Determine the width of the column (without the margin)
-		colWidth := c.maxWidth(i, min(i+height, len(cands)))
-		totalColWidth := colWidth + completionColMarginTotal
-		if totalColWidth > remainedWidth {
-			totalColWidth = remainedWidth
-			colWidth = totalColWidth - completionColMarginTotal
-			trimmed = true
-		}
-
-		col := ui.NewBuffer(totalColWidth)
-		for j = i; j < i+height; j++ {
-			if j > i {
-				col.Newline()
-			}
-			if j >= len(cands) {
-				// Write padding to make the listing a rectangle.
-				col.WriteSpaces(totalColWidth, styleForCompletion.String())
-			} else {
-				col.WriteSpaces(completionColMarginLeft, styleForCompletion.String())
-				s := ui.JoinStyles(styleForCompletion, cands[j].menu.Styles)
-				if j == c.selected {
-					s = append(s, styleForSelectedCompletion.String())
-				}
-				col.WriteString(util.ForceWcwidth(cands[j].menu.Text, colWidth), s.String())
-				col.WriteSpaces(completionColMarginRight, styleForCompletion.String())
-				if !trimmed {
-					c.lastShownInFull = j
-				}
-			}
-		}
-
-		b.ExtendRight(col, 0)
-		remainedWidth -= totalColWidth
-		if remainedWidth <= completionColMarginTotal {
-			break
-		}
-	}
-	// When the listing is incomplete, always use up the entire width.
-	if remainedWidth > 0 && c.needScrollbar() {
-		col := ui.NewBuffer(remainedWidth)
-		for i := 0; i < height; i++ {
-			if i > 0 {
-				col.Newline()
-			}
-			col.WriteSpaces(remainedWidth, styleForCompletion.String())
-		}
-		b.ExtendRight(col, 0)
-		remainedWidth = 0
-	}
-	return b
-}
-
-func (c *completion) changeFilter(f string) {
-	c.filter = f
-	if f == "" {
-		c.filtered = c.candidates
-		return
-	}
-	c.filtered = nil
-	for _, cand := range c.candidates {
-		if strings.Contains(cand.menu.Text, f) {
-			c.filtered = append(c.filtered, cand)
-		}
-	}
-	if len(c.filtered) > 0 {
-		c.selected = 0
-	} else {
-		c.selected = -1
-	}
-}
diff -pruN 0.11+ds1-5/edit/completion_test.go 0.12+ds1-1/edit/completion_test.go
--- 0.11+ds1-5/edit/completion_test.go	2018-01-18 02:03:11.000000000 +0000
+++ 0.12+ds1-1/edit/completion_test.go	1970-01-01 00:00:00.000000000 +0000
@@ -1,31 +0,0 @@
-package edit
-
-import (
-	"reflect"
-	"testing"
-
-	"github.com/elves/elvish/eval"
-	"github.com/elves/elvish/parse"
-)
-
-type complContextFinderTest struct {
-	src  string
-	want complContext
-}
-
-func testComplContextFinder(t *testing.T, name string, finder complContextFinder, tests []complContextFinderTest) {
-	ev := eval.NewEvaler()
-	defer ev.Close()
-	for _, test := range tests {
-		n, err := parse.Parse("[test]", test.src)
-		// Ignore error as long is n is non-nil
-		if n == nil {
-			panic(err)
-		}
-		leaf := findLeafNode(n, len(test.src))
-		got := finder(leaf, ev)
-		if !reflect.DeepEqual(got, test.want) {
-			t.Errorf("For %q, %s(leaf) => %v, want %v", test.src, name, got, test.want)
-		}
-	}
-}
diff -pruN 0.11+ds1-5/edit/compl_getopt.go 0.12+ds1-1/edit/compl_getopt.go
--- 0.11+ds1-5/edit/compl_getopt.go	2018-01-18 02:03:11.000000000 +0000
+++ 0.12+ds1-1/edit/compl_getopt.go	1970-01-01 00:00:00.000000000 +0000
@@ -1,160 +0,0 @@
-package edit
-
-import (
-	"strings"
-	"unicode/utf8"
-
-	"github.com/elves/elvish/eval"
-	"github.com/elves/elvish/eval/types"
-	"github.com/elves/elvish/getopt"
-	"github.com/elves/elvish/parse"
-)
-
-func complGetopt(ec *eval.Frame, a []types.Value, o map[string]types.Value) {
-	var elemsv, optsv, argsv types.IteratorValue
-	eval.ScanArgs(a, &elemsv, &optsv, &argsv)
-	eval.TakeNoOpt(o)
-
-	var (
-		elems    []string
-		opts     []*getopt.Option
-		args     []eval.Fn
-		variadic bool
-	)
-	desc := make(map[*getopt.Option]string)
-	// Convert arguments.
-	elemsv.Iterate(func(v types.Value) bool {
-		elem, ok := v.(types.String)
-		if !ok {
-			throwf("arg should be string, got %s", v.Kind())
-		}
-		elems = append(elems, string(elem))
-		return true
-	})
-	optsv.Iterate(func(v types.Value) bool {
-		m, ok := v.(types.MapLike)
-		if !ok {
-			throwf("opt should be map-like, got %s", v.Kind())
-		}
-		get := func(ks string) (string, bool) {
-			kv := types.String(ks)
-			if !m.HasKey(kv) {
-				return "", false
-			}
-			vv := m.IndexOne(kv)
-			if vs, ok := vv.(types.String); ok {
-				return string(vs), true
-			} else {
-				throwf("%s should be string, got %s", ks, vs.Kind())
-				panic("unreachable")
-			}
-		}
-
-		opt := &getopt.Option{}
-		if s, ok := get("short"); ok {
-			r, size := utf8.DecodeRuneInString(s)
-			if r == utf8.RuneError || size != len(s) {
-				throwf("short option should be exactly one rune, got %v", parse.Quote(s))
-			}
-			opt.Short = r
-		}
-		if s, ok := get("long"); ok {
-			opt.Long = s
-		}
-		if opt.Short == 0 && opt.Long == "" {
-			throwf("opt should have at least one of short and long forms")
-		}
-		if s, ok := get("desc"); ok {
-			desc[opt] = s
-		}
-		opts = append(opts, opt)
-		return true
-	})
-	argsv.Iterate(func(v types.Value) bool {
-		sv, ok := v.(types.String)
-		if ok {
-			if string(sv) == "..." {
-				variadic = true
-				return true
-			}
-			throwf("string except for ... not allowed as argument handler, got %s", parse.Quote(string(sv)))
-		}
-		arg, ok := v.(eval.Fn)
-		if !ok {
-			throwf("argument handler should be fn, got %s", v.Kind())
-		}
-		args = append(args, arg)
-		return true
-	})
-	// TODO Configurable config
-	g := getopt.Getopt{opts, getopt.GNUGetoptLong}
-	_, parsedArgs, ctx := g.Parse(elems)
-	out := ec.OutputChan()
-
-	putShortOpt := func(opt *getopt.Option) {
-		c := &complexCandidate{stem: "-" + string(opt.Short)}
-		if d, ok := desc[opt]; ok {
-			c.displaySuffix = " (" + d + ")"
-		}
-		out <- c
-	}
-	putLongOpt := func(opt *getopt.Option) {
-		c := &complexCandidate{stem: "--" + string(opt.Long)}
-		if d, ok := desc[opt]; ok {
-			c.displaySuffix = " (" + d + ")"
-		}
-		out <- c
-	}
-
-	switch ctx.Type {
-	case getopt.NewOptionOrArgument, getopt.Argument:
-		// Find argument completer
-		var argCompl eval.Fn
-		if len(parsedArgs) < len(args) {
-			argCompl = args[len(parsedArgs)]
-		} else if variadic {
-			argCompl = args[len(args)-1]
-		}
-		if argCompl != nil {
-			rawCands := make(chan rawCandidate)
-			defer close(rawCands)
-			go func() {
-				for rc := range rawCands {
-					out <- rc
-				}
-			}()
-			err := callArgCompleter(argCompl, ec.Evaler, []string{ctx.Text}, rawCands)
-			maybeThrow(err)
-		}
-		// TODO Notify that there is no suitable argument completer
-	case getopt.NewOption:
-		for _, opt := range opts {
-			if opt.Short != 0 {
-				putShortOpt(opt)
-			}
-			if opt.Long != "" {
-				putLongOpt(opt)
-			}
-		}
-	case getopt.NewLongOption:
-		for _, opt := range opts {
-			if opt.Long != "" {
-				putLongOpt(opt)
-			}
-		}
-	case getopt.LongOption:
-		for _, opt := range opts {
-			if strings.HasPrefix(opt.Long, ctx.Text) {
-				putLongOpt(opt)
-			}
-		}
-	case getopt.ChainShortOption:
-		for _, opt := range opts {
-			if opt.Short != 0 {
-				// XXX loses chained options
-				putShortOpt(opt)
-			}
-		}
-	case getopt.OptionArgument:
-	}
-}
diff -pruN 0.11+ds1-5/edit/dump_buf.go 0.12+ds1-1/edit/dump_buf.go
--- 0.11+ds1-5/edit/dump_buf.go	2018-01-18 02:03:11.000000000 +0000
+++ 0.12+ds1-1/edit/dump_buf.go	1970-01-01 00:00:00.000000000 +0000
@@ -1,39 +0,0 @@
-package edit
-
-import (
-	"fmt"
-	"html"
-	"strings"
-
-	"github.com/elves/elvish/eval"
-	"github.com/elves/elvish/eval/types"
-)
-
-func _dumpBuf(ec *eval.Frame, args []types.Value, opts map[string]types.Value) {
-	out := ec.OutputFile()
-	buf := ec.Editor.(*Editor).writer.CurrentBuffer()
-	for _, line := range buf.Lines {
-		style := ""
-		openedSpan := false
-		for _, c := range line {
-			if c.Style != style {
-				if openedSpan {
-					fmt.Fprint(out, "</span>")
-				}
-				var classes []string
-				for _, c := range strings.Split(c.Style, ";") {
-					classes = append(classes, "sgr-"+c)
-				}
-				fmt.Fprintf(out,
-					`<span class="%s">`, strings.Join(classes, " "))
-				style = c.Style
-				openedSpan = true
-			}
-			fmt.Fprintf(out, "%s", html.EscapeString(c.Text))
-		}
-		if openedSpan {
-			fmt.Fprint(out, "</span>")
-		}
-		fmt.Fprint(out, "\n")
-	}
-}
diff -pruN 0.11+ds1-5/edit/edcore/abbr.go 0.12+ds1-1/edit/edcore/abbr.go
--- 0.11+ds1-5/edit/edcore/abbr.go	1970-01-01 00:00:00.000000000 +0000
+++ 0.12+ds1-1/edit/edcore/abbr.go	2018-07-11 02:32:25.000000000 +0000
@@ -0,0 +1,32 @@
+package edcore
+
+import (
+	"github.com/elves/elvish/eval"
+	"github.com/elves/elvish/eval/vals"
+	"github.com/elves/elvish/eval/vars"
+	"github.com/xiaq/persistent/hashmap"
+)
+
+func init() {
+	atEditorInit(func(ed *editor, ns eval.Ns) {
+		ed.abbr = vals.EmptyMap
+		ns["abbr"] = vars.FromPtr(&ed.abbr)
+	})
+}
+
+func abbrIterate(abbr hashmap.Map, cb func(abbr, full string) bool) {
+	for it := abbr.Iterator(); it.HasElem(); it.Next() {
+		abbrValue, fullValue := it.Elem()
+		abbr, ok := abbrValue.(string)
+		if !ok {
+			continue
+		}
+		full, ok := fullValue.(string)
+		if !ok {
+			continue
+		}
+		if !cb(abbr, full) {
+			break
+		}
+	}
+}
diff -pruN 0.11+ds1-5/edit/edcore/action.go 0.12+ds1-1/edit/edcore/action.go
--- 0.11+ds1-5/edit/edcore/action.go	1970-01-01 00:00:00.000000000 +0000
+++ 0.12+ds1-1/edit/edcore/action.go	2018-07-11 02:32:25.000000000 +0000
@@ -0,0 +1,15 @@
+package edcore
+
+import "github.com/elves/elvish/edit/eddefs"
+
+func (ed *editor) SetAction(action eddefs.Action) {
+	if ed.nextAction == noAction {
+		ed.nextAction = action
+	}
+}
+
+func (ed *editor) popAction() eddefs.Action {
+	action := ed.nextAction
+	ed.nextAction = noAction
+	return action
+}
diff -pruN 0.11+ds1-5/edit/edcore/api.go 0.12+ds1-1/edit/edcore/api.go
--- 0.11+ds1-5/edit/edcore/api.go	1970-01-01 00:00:00.000000000 +0000
+++ 0.12+ds1-1/edit/edcore/api.go	2018-07-11 02:32:25.000000000 +0000
@@ -0,0 +1,122 @@
+package edcore
+
+import (
+	"errors"
+	"strconv"
+	"unicode/utf8"
+
+	"github.com/elves/elvish/edit/eddefs"
+	"github.com/elves/elvish/edit/ui"
+	"github.com/elves/elvish/eval"
+	"github.com/elves/elvish/eval/vars"
+)
+
+// This file implements types and functions for interactions with the
+// Elvishscript runtime.
+
+var (
+	errNotNav             = errors.New("not in navigation mode")
+	errLineMustBeString   = errors.New("line must be string")
+	errDotMustBeString    = errors.New("dot must be string")
+	errDotMustBeInt       = errors.New("dot must be integer")
+	errDotOutOfRange      = errors.New("dot out of range")
+	errDotInsideCodepoint = errors.New("dot cannot be inside a codepoint")
+	errEditorInactive     = errors.New("editor inactive")
+)
+
+// makeNs makes the edit: namespace.
+func makeNs(ed *editor) eval.Ns {
+	ns := eval.NewNs()
+
+	// TODO(xiaq): Everything here should be registered to some registry instead
+	// of centralized here.
+
+	// Internal states.
+	ns["current-command"] = vars.FromSetGet(
+		func(v interface{}) error {
+			if !ed.active {
+				return errEditorInactive
+			}
+			if s, ok := v.(string); ok {
+				ed.buffer = s
+				ed.dot = len(ed.buffer)
+			} else {
+				return errLineMustBeString
+			}
+			return nil
+		},
+		func() interface{} { return ed.buffer },
+	)
+	ns["-dot"] = vars.FromSetGet(
+		func(v interface{}) error {
+			s, ok := v.(string)
+			if !ok {
+				return errDotMustBeString
+			}
+			i, err := strconv.Atoi(s)
+			if err != nil {
+				if err.(*strconv.NumError).Err == strconv.ErrRange {
+					return errDotOutOfRange
+				} else {
+					return errDotMustBeInt
+				}
+			}
+			if i < 0 || i > len(ed.buffer) {
+				return errDotOutOfRange
+			}
+			if i < len(ed.buffer) {
+				r, _ := utf8.DecodeRuneInString(ed.buffer[i:])
+				if r == utf8.RuneError {
+					return errDotInsideCodepoint
+				}
+			}
+			ed.dot = i
+			return nil
+		},
+		func() interface{} { return strconv.Itoa(ed.dot) },
+	)
+	ns["selected-file"] = vars.FromGet(
+		func() interface{} {
+			if !ed.active {
+				throw(errEditorInactive)
+			}
+			nav, ok := ed.mode.(*navigation)
+			if !ok {
+				throw(errNotNav)
+			}
+			return nav.current.selectedName()
+		},
+	)
+
+	// Functions.
+	fns := map[string]interface{}{
+		"binding-table": eddefs.MakeBindingMap,
+		"insert-at-dot": ed.InsertAtDot,
+		"replace-input": ed.replaceInput,
+		"styled":        styled,
+		"key":           ui.ToKey,
+		"wordify":       wordifyBuiltin,
+		"-dump-buf":     ed.dumpBuf,
+	}
+	ns.AddBuiltinFns("edit:", fns)
+
+	return ns
+}
+
+// CallFn calls an Fn, displaying its outputs and possible errors as editor
+// notifications. It is the preferred way to call a Fn while the editor is
+// active.
+func (ed *editor) CallFn(fn eval.Callable, args ...interface{}) {
+	ports := []*eval.Port{
+		eval.DevNullClosedChan, ed.notifyPort, ed.notifyPort,
+	}
+	// XXX There is no source to pass to NewTopEvalCtx.
+	ec := eval.NewTopFrame(ed.evaler, eval.NewInternalSource("[editor]"), ports)
+	ex := ec.Call(fn, args, eval.NoOpts)
+	if ex != nil {
+		ed.Notify("function error: %s", ex.Error())
+	}
+
+	// XXX Concurrency-dangerous!
+	ed.refresh(true, true)
+}
diff -pruN 0.11+ds1-5/edit/edcore/api_utils.go 0.12+ds1-1/edit/edcore/api_utils.go
--- 0.11+ds1-5/edit/edcore/api_utils.go	1970-01-01 00:00:00.000000000 +0000
+++ 0.12+ds1-1/edit/edcore/api_utils.go	2018-07-11 02:32:25.000000000 +0000
@@ -0,0 +1,23 @@
+package edcore
+
+// Trivial utilities for the elvishscript API.
+
+import (
+	"fmt"
+
+	"github.com/elves/elvish/util"
+)
+
+func throw(e error) {
+	util.Throw(e)
+}
+
+func maybeThrow(e error) {
+	if e != nil {
+		util.Throw(e)
+	}
+}
+
+func throwf(format string, args ...interface{}) {
+	util.Throw(fmt.Errorf(format, args...))
+}
diff -pruN 0.11+ds1-5/edit/edcore/dump_buf.go 0.12+ds1-1/edit/edcore/dump_buf.go
--- 0.11+ds1-5/edit/edcore/dump_buf.go	1970-01-01 00:00:00.000000000 +0000
+++ 0.12+ds1-1/edit/edcore/dump_buf.go	2018-07-11 02:32:25.000000000 +0000
@@ -0,0 +1,38 @@
+package edcore
+
+import (
+	"fmt"
+	"html"
+	"strings"
+
+	"github.com/elves/elvish/eval"
+)
+
+func (ed *editor) dumpBuf(fm *eval.Frame) {
+	out := fm.OutputFile()
+	buf := ed.writer.CurrentBuffer()
+	for _, line := range buf.Lines {
+		style := ""
+		openedSpan := false
+		for _, c := range line {
+			if c.Style != style {
+				if openedSpan {
+					fmt.Fprint(out, "</span>")
+				}
+				var classes []string
+				for _, c := range strings.Split(c.Style, ";") {
+					classes = append(classes, "sgr-"+c)
+				}
+				fmt.Fprintf(out,
+					`<span class="%s">`, strings.Join(classes, " "))
+				style = c.Style
+				openedSpan = true
+			}
+			fmt.Fprintf(out, "%s", html.EscapeString(c.Text))
+		}
+		if openedSpan {
+			fmt.Fprint(out, "</span>")
+		}
+		fmt.Fprint(out, "\n")
+	}
+}
diff -pruN 0.11+ds1-5/edit/edcore/e2e_test.go 0.12+ds1-1/edit/edcore/e2e_test.go
--- 0.11+ds1-5/edit/edcore/e2e_test.go	1970-01-01 00:00:00.000000000 +0000
+++ 0.12+ds1-1/edit/edcore/e2e_test.go	2018-07-11 02:32:25.000000000 +0000
@@ -0,0 +1,160 @@
+// +build !windows,!plan9
+
+// End-to-end tests for the editor. Only enabled on UNIX where pseudo-terminals
+// are supported.
+
+package edcore
+
+import (
+	"io"
+	"os"
+	"syscall"
+	"testing"
+	"time"
+
+	"github.com/elves/elvish/eval"
+	"github.com/elves/elvish/sys"
+	"github.com/kr/pty"
+)
+
+// readLineTest contains the data for a test case of ReadLine.
+type readLineTest struct {
+	// The input to write to the master side.
+	input []byte
+	// Places where SIGINT should be sent to the editor, as indicies into the
+	// input string. For example, if sigints is {1}, it means that a SIGINT
+	// should be sent right after the first byte is sent.
+	sigints []int
+	// Expected line to be returned from ReadLine.
+	wantLine string
+}
+
+func newTest(input string, wantLine string) *readLineTest {
+	return &readLineTest{input: []byte(input), wantLine: wantLine}
+}
+
+func (t *readLineTest) sigint(x ...int) *readLineTest {
+	t.sigints = x
+	return t
+}
+
+var readLineTests = []*readLineTest{
+	newTest("\n", ""),
+	newTest("test\n", "test"),
+	// \x7f is DEL and erases the previous character
+	newTest("abc\x7fd\n", "abd"),
+	// \x17 is ^U and erases the line before the cursor
+	newTest("abc\x17d\n", "d"),
+	// SIGINT resets the editor and erases the line. Disabled for now.
+	// newTest("000123\n", "123").sigint(3),
+}
+
+var readLineTimeout = 5 * time.Second
+
+func TestReadLine(t *testing.T) {
+	ev := eval.NewEvaler()
+	defer ev.Close()
+
+	for _, test := range readLineTests {
+		// Editor output is only used in failure messages.
+		var outputs []byte
+		sigs := make(chan os.Signal, 10)
+		defer close(sigs)
+		master, lineChan, errChan, err := run(ev, sigs, &outputs)
+		defer master.Close()
+		if err != nil {
+			t.Skip(err)
+		}
+
+		write(master, sigs, test.input, test.sigints)
+
+		select {
+		case line := <-lineChan:
+			if line != test.wantLine {
+				t.Errorf("ReadLine() => %q, want %q (input %q)", line, test.wantLine, test.input)
+			}
+		case err := <-errChan:
+			t.Errorf("ReadLine() => error %v (input %q)", err, test.input)
+		case <-time.After(readLineTimeout):
+			t.Errorf("ReadLine() timed out (input %q)", test.input)
+			t.Log("Stack trace: \n" + sys.DumpStack())
+			t.Logf("Terminal output: %q", outputs)
+			t.FailNow()
+		}
+	}
+}
+
+// run sets up a testing environment for an Editor, and calls its ReadLine
+// method in a goroutine. It returns the master end of the pty the Editor is
+// connected to, and two channels onto which the result of ReadLine will be
+// delivered. The caller is responsible for closing the master file.
+func run(ev *eval.Evaler, sigs <-chan os.Signal, ptrOutputs *[]byte) (*os.File,
+	<-chan string, <-chan error, error) {
+
+	master, tty, err := pty.Open()
+	if err != nil {
+		return nil, nil, nil, err
+	}
+	// Continually consume tty outputs so that the editor is not blocked on
+	// writing.
+	go drain(master, ptrOutputs)
+
+	lineChan := make(chan string)
+	errChan := make(chan error)
+
+	go func() {
+		ed := NewEditor(tty, tty, nil, ev)
+		line, err := ed.ReadLine()
+		if err != nil {
+			errChan <- err
+		} else {
+			lineChan <- line
+		}
+		ed.Close()
+		tty.Close()
+		close(lineChan)
+		close(errChan)
+	}()
+
+	return master, lineChan, errChan, nil
+}
+
+// drain drains the given reader. If a non-nil []byte pointer is passed, it also
+// makes the outputs available. It returns when r.Read returns an error.
+func drain(r io.Reader, ptrOutputs *[]byte) {
+	var buf [256]byte
+	for {
+		nr, err := r.Read(buf[:])
+		if err != nil {
+			return
+		}
+		if ptrOutputs != nil {
+			*ptrOutputs = append(*ptrOutputs, buf[:nr]...)
+		}
+	}
+}
+
+// write interprets the input and sigints arguments, and write inputs and
+// signals to the writer and signal channel.
+func write(w *os.File, sigs chan<- os.Signal, input []byte, sigints []int) {
+	if len(sigints) == 0 {
+		mustWrite(w, input)
+		return
+	}
+	for i, idx := range sigints {
+		lastidx := 0
+		if i > 0 {
+			lastidx = sigints[i-1]
+		}
+		mustWrite(w, input[lastidx:idx])
+		sigs <- syscall.SIGINT
+	}
+	mustWrite(w, input[sigints[len(sigints)-1]:])
+}
+
+func mustWrite(w io.Writer, p []byte) {
+	_, err := w.Write(p)
+	if err != nil {
+		panic(err)
+	}
+}
diff -pruN 0.11+ds1-5/edit/edcore/eddefs_aliases.go 0.12+ds1-1/edit/edcore/eddefs_aliases.go
--- 0.11+ds1-5/edit/edcore/eddefs_aliases.go	1970-01-01 00:00:00.000000000 +0000
+++ 0.12+ds1-1/edit/edcore/eddefs_aliases.go	2018-07-11 02:32:25.000000000 +0000
@@ -0,0 +1,14 @@
+package edcore
+
+import "github.com/elves/elvish/edit/eddefs"
+
+// Aliases for variables and functions from the eddefs package.
+// NOTE: We cannot use type aliases before we drop support for Go 1.8.
+var (
+	noAction     = eddefs.NoAction
+	reprocessKey = eddefs.ReprocessKey
+	commitLine   = eddefs.CommitLine
+	commitEOF    = eddefs.CommitEOF
+
+	emptyBindingMap = eddefs.EmptyBindingMap
+)
diff -pruN 0.11+ds1-5/edit/edcore/edit.go 0.12+ds1-1/edit/edcore/edit.go
--- 0.11+ds1-5/edit/edcore/edit.go	1970-01-01 00:00:00.000000000 +0000
+++ 0.12+ds1-1/edit/edcore/edit.go	2018-07-11 02:32:25.000000000 +0000
@@ -0,0 +1,532 @@
+// Package edcore implements the core of the Elvish command editor.
+package edcore
+
+import (
+	"bufio"
+	"bytes"
+	"fmt"
+	"io"
+	"os"
+	"sync"
+	"syscall"
+	"time"
+
+	"github.com/elves/elvish/daemon"
+	"github.com/elves/elvish/edit/eddefs"
+	"github.com/elves/elvish/edit/highlight"
+	"github.com/elves/elvish/edit/tty"
+	"github.com/elves/elvish/edit/ui"
+	"github.com/elves/elvish/eval"
+	"github.com/elves/elvish/eval/vals"
+	"github.com/elves/elvish/parse"
+	"github.com/elves/elvish/sys"
+	"github.com/elves/elvish/util"
+	"github.com/xiaq/persistent/hashmap"
+)
+
+var logger = util.GetLogger("[edit] ")
+
+// editor implements the line editor.
+type editor struct {
+	in     *os.File
+	out    *os.File
+	writer tty.Writer
+	reader tty.Reader
+	sigs   <-chan os.Signal
+	daemon *daemon.Client
+	evaler *eval.Evaler
+
+	active      bool
+	activeMutex sync.Mutex
+
+	// notifyPort is a write-only port that turns data written to it into editor
+	// notifications.
+	notifyPort *eval.Port
+	// notifyRead is the read end of notifyPort.File.
+	notifyRead *os.File
+
+	// Configurations. Each of the following fields have an initializer defined
+	// using atEditorInit.
+	editorHooks
+	abbr      hashmap.Map
+	maxHeight float64
+
+	prompt, rprompt   eddefs.Prompt
+	RpromptPersistent bool
+
+	// Modes.
+	insert     *insert
+	command    *command
+	navigation *navigation
+	listing    *listingMode
+
+	restoreTerminal func() error
+
+	editorState
+}
+
+type editorState struct {
+	// States used during ReadLine. Reset at the beginning of ReadLine.
+	notificationMutex sync.Mutex
+
+	notifications []string
+	tips          []string
+
+	buffer string
+	dot    int
+
+	chunk           *parse.Chunk
+	styling         *highlight.Styling
+	parseErrorAtEnd bool
+
+	promptContent, rpromptContent []*ui.Styled
+
+	mode eddefs.Mode
+
+	// A cache of external commands, used in stylist.
+	isExternal map[string]bool
+
+	// Used for builtins.
+	lastKey    ui.Key
+	nextAction eddefs.Action
+}
+
+// NewEditor creates an Editor. When the instance is no longer used, its Close
+// method should be called.
+func NewEditor(in *os.File, out *os.File, sigs <-chan os.Signal, ev *eval.Evaler) eddefs.Editor {
+	daemon := ev.DaemonClient
+
+	ed := &editor{
+		in:     in,
+		out:    out,
+		writer: tty.NewWriter(out),
+		reader: tty.NewReader(in),
+		sigs:   sigs,
+		daemon: daemon,
+		evaler: ev,
+	}
+
+	notifyChan := make(chan interface{})
+	notifyRead, notifyWrite, err := os.Pipe()
+	if err != nil {
+		panic(err)
+	}
+	ed.notifyPort = &eval.Port{File: notifyWrite, Chan: notifyChan}
+	ed.notifyRead = notifyRead
+	// Forward reads from notifyRead to notification.
+	go func() {
+		reader := bufio.NewReader(notifyRead)
+		for {
+			line, err := reader.ReadString('\n')
+			if err != nil {
+				break
+			}
+			ed.Notify("[bytes out] %s", line[:len(line)-1])
+		}
+		if err != io.EOF {
+			logger.Println("notifyRead error:", err)
+		}
+	}()
+	// Forward reads from notifyChan to notification.
+	go func() {
+		for v := range notifyChan {
+			ed.Notify("[value out] %s", vals.Repr(v, vals.NoPretty))
+		}
+	}()
+
+	ev.Editor = ed
+
+	ns := makeNs(ed)
+	for _, f := range editorInitFuncs {
+		f(ed, ns)
+	}
+	ev.Builtin.AddNs("edit", ns)
+
+	err = ev.EvalSource(eval.NewScriptSource("[editor]", "[editor]", "use binding; binding:install"))
+	if err != nil {
+		fmt.Fprintln(out, "Failed to load default binding:", err)
+	}
+
+	return ed
+}
+
+func (ed *editor) Close() {
+	ed.reader.Close()
+	close(ed.notifyPort.Chan)
+	ed.notifyPort.File.Close()
+	ed.notifyRead.Close()
+	ed.prompt.Close()
+	ed.rprompt.Close()
+}
+
+func (ed *editor) Evaler() *eval.Evaler {
+	return ed.evaler
+}
+
+func (ed *editor) Daemon() *daemon.Client {
+	return ed.daemon
+}
+
+func (ed *editor) Buffer() (string, int) {
+	return ed.buffer, ed.dot
+}
+
+func (ed *editor) SetBuffer(buffer string, dot int) {
+	ed.buffer, ed.dot = buffer, dot
+}
+
+func (ed *editor) ParsedBuffer() *parse.Chunk {
+	return ed.chunk
+}
+
+func (ed *editor) SetMode(m eddefs.Mode) {
+	if ed.mode != nil {
+		ed.mode.Teardown()
+	}
+	ed.mode = m
+}
+
+func (ed *editor) SetModeInsert() {
+	ed.SetMode(ed.insert)
+}
+
+func (ed *editor) SetModeListing(b eddefs.BindingMap, p eddefs.ListingProvider) {
+	ed.listing.setup(b, p)
+	ed.SetMode(ed.listing)
+}
+
+func (ed *editor) RefreshListing() {
+	if l, ok := ed.mode.(*listingMode); ok {
+		l.refresh()
+	}
+}
+
+func (ed *editor) flash() {
+	// TODO implement fish-like flash effect
+}
+
+// AddTip adds a message to the tip area.
+func (ed *editor) AddTip(format string, args ...interface{}) {
+	ed.tips = append(ed.tips, fmt.Sprintf(format, args...))
+}
+
+// Notify writes out a message in a way that does not interrupt the editor
+// display. When the editor is not active, it simply writes the message to the
+// terminal. When the editor is active, it appends the message to the
+// notification queue, which will be written out during the update cycle. It can
+// be safely used concurrently.
+func (ed *editor) Notify(format string, args ...interface{}) {
+	msg := fmt.Sprintf(format, args...)
+	ed.activeMutex.Lock()
+	defer ed.activeMutex.Unlock()
+	// If the editor is not active, simply write out the message.
+	if !ed.active {
+		ed.out.WriteString(msg + "\n")
+		return
+	}
+	ed.notificationMutex.Lock()
+	defer ed.notificationMutex.Unlock()
+	ed.notifications = append(ed.notifications, msg)
+}
+
+func (ed *editor) LastKey() ui.Key {
+	return ed.lastKey
+}
+
+func (ed *editor) refresh(fullRefresh bool, addErrorsToTips bool) error {
+	src := ed.buffer
+	// Parse the current line
+	n, err := parse.Parse("[interactive]", src)
+	ed.chunk = n
+
+	ed.parseErrorAtEnd = err != nil && atEnd(err, len(src))
+	// If all parse errors are at the end, it is likely caused by incomplete
+	// input. In that case, do not complain about parse errors.
+	// TODO(xiaq): Find a more reliable way to determine incomplete input.
+	// Ideally the parser should report it.
+	if err != nil && addErrorsToTips && !ed.parseErrorAtEnd {
+		ed.AddTip("%s", err)
+	}
+
+	ed.styling = &highlight.Styling{}
+	doHighlight(n, ed)
+
+	_, err = ed.evaler.Compile(n, eval.NewInteractiveSource(src))
+	if err != nil && !atEnd(err, len(src)) {
+		if addErrorsToTips {
+			ed.AddTip("%s", err)
+		}
+		// Highlight errors in the input buffer.
+		ctx := err.(*eval.CompilationError).Context
+		ed.styling.Add(ctx.Begin, ctx.End, styleForCompilerError.String())
+	}
+
+	// Render onto a buffer.
+	height, width := sys.GetWinsize(ed.out)
+	height = min(height, maxHeightToInt(ed.maxHeight))
+	er := &editorRenderer{&ed.editorState, height, nil}
+	buf := ui.Render(er, width)
+	return ed.writer.CommitBuffer(er.bufNoti, buf, fullRefresh)
+}
+
+func atEnd(e error, n int) bool {
+	switch e := e.(type) {
+	case *eval.CompilationError:
+		return e.Context.Begin == n
+	case *parse.Error:
+		for _, entry := range e.Entries {
+			if entry.Context.Begin != n {
+				return false
+			}
+		}
+		return true
+	default:
+		logger.Printf("atEnd called with error type %T", e)
+		return false
+	}
+}
+
+// InsertAtDot inserts text at the dot and moves the dot after it.
+func (ed *editor) InsertAtDot(text string) {
+	ed.buffer = ed.buffer[:ed.dot] + text + ed.buffer[ed.dot:]
+	ed.dot += len(text)
+}
+
+func (ed *editor) SetPrompt(prompt eddefs.Prompt) {
+	ed.prompt = prompt
+}
+
+func (ed *editor) SetRPrompt(rprompt eddefs.Prompt) {
+	ed.rprompt = rprompt
+}
+
+// startReadLine prepares the terminal for the editor.
+func (ed *editor) startReadLine() error {
+	ed.activeMutex.Lock()
+	defer ed.activeMutex.Unlock()
+	ed.active = true
+
+	restoreTerminal, err := tty.Setup(ed.in, ed.out)
+	if err != nil {
+		if restoreTerminal != nil {
+			restoreTerminal()
+		}
+		return err
+	}
+	// Save the restore callback when Readline is called for the first time, and
+	// reuse that callback. This is guard against commands that misbehave and
+	// put the terminal in a bad state when it quits.
+	if ed.restoreTerminal == nil {
+		ed.restoreTerminal = restoreTerminal
+	}
+
+	return nil
+}
+
+// finishReadLine puts the terminal in a state suitable for other programs to
+// use.
+func (ed *editor) finishReadLine() error {
+	// After-readline hooks should be called before most teardown happens as
+	// they can cause the editor to refresh.
+	for _, f := range ed.afterReadline {
+		f(ed.buffer)
+	}
+
+	ed.activeMutex.Lock()
+	defer ed.activeMutex.Unlock()
+	ed.active = false
+
+	// Refresh the terminal for the last time in a clean-ish state.
+	ed.SetModeInsert()
+	ed.tips = nil
+	ed.dot = len(ed.buffer)
+	if !ed.RpromptPersistent {
+		ed.rpromptContent = nil
+	}
+	errRefresh := ed.refresh(false, false)
+	ed.mode.Teardown()
+	ed.out.WriteString("\n")
+	ed.writer.ResetCurrentBuffer()
+
+	ed.reader.Stop()
+
+	// Restore termios.
+	errRestore := ed.restoreTerminal()
+
+	// Reset all of editorState.
+	ed.editorState = editorState{}
+
+	return util.Errors(errRefresh, errRestore)
+}
+
+// ReadLine reads a line interactively.
+func (ed *editor) ReadLine() (string, error) {
+	err := ed.startReadLine()
+	if err != nil {
+		return "", err
+	}
+	defer func() {
+		err := ed.finishReadLine()
+		if err != nil {
+			fmt.Fprintln(ed.out, "error:", err)
+		}
+	}()
+
+	ed.SetModeInsert()
+
+	// Find external commands asynchronously, so that slow I/O won't block the
+	// editor.
+	isExternalCh := make(chan map[string]bool, 1)
+	go getIsExternal(isExternalCh)
+
+	ed.reader.Start()
+
+	fullRefresh := false
+
+	for _, f := range ed.beforeReadline {
+		f()
+	}
+
+	ed.promptContent = ed.prompt.Last()
+	ed.rpromptContent = ed.rprompt.Last()
+	fresh := true
+MainLoop:
+	for {
+		ed.prompt.Update(fresh)
+		ed.rprompt.Update(fresh)
+		fresh = false
+
+	refresh:
+		err := ed.refresh(fullRefresh, true)
+		fullRefresh = false
+		if err != nil {
+			return "", err
+		}
+
+		ed.tips = nil
+
+		select {
+		case ed.promptContent = <-ed.prompt.Chan():
+			logger.Println("prompt fetched late")
+			goto refresh
+		case ed.rpromptContent = <-ed.rprompt.Chan():
+			logger.Println("rprompt fetched late")
+			goto refresh
+		case m := <-isExternalCh:
+			ed.isExternal = m
+			goto refresh
+		case sig := <-ed.sigs:
+			// TODO(xiaq): Maybe support customizable handling of signals
+			switch sig {
+			case syscall.SIGHUP:
+				return "", io.EOF
+			case syscall.SIGINT:
+				// Start over
+				ed.mode.Teardown()
+				ed.editorState = editorState{
+					isExternal: ed.isExternal,
+				}
+				ed.SetModeInsert()
+				fresh = true
+				continue MainLoop
+			case sys.SIGWINCH:
+				fullRefresh = true
+				continue MainLoop
+			default:
+				ed.AddTip("ignored signal %s", sig)
+			}
+		case event := <-ed.reader.EventChan():
+			switch event := event.(type) {
+			case tty.NonfatalErrorEvent:
+				ed.Notify("error when reading terminal: %v", event.Err)
+			case tty.FatalErrorEvent:
+				ed.Notify("fatal error when reading terminal: %v", event.Err)
+				return "", event.Err
+			case tty.MouseEvent:
+				ed.AddTip("mouse: %+v", event)
+			case tty.CursorPosition:
+				// Ignore CPR
+			case tty.PasteSetting:
+				if !event {
+					continue
+				}
+				var buf bytes.Buffer
+				timer := time.NewTimer(tty.DefaultSeqTimeout)
+			paste:
+				for {
+					// XXX Should also select on other chans. However those chans
+					// will be unified (again) into one later so we don't do
+					// busywork here.
+					select {
+					case event := <-ed.reader.EventChan():
+						switch event := event.(type) {
+						case tty.KeyEvent:
+							k := ui.Key(event)
+							if k.Mod != 0 {
+								ed.Notify("function key within paste, aborting")
+								break paste
+							}
+							buf.WriteRune(k.Rune)
+							timer.Reset(tty.DefaultSeqTimeout)
+						case tty.PasteSetting:
+							if !event {
+								break paste
+							}
+						default: // Ignore other things.
+						}
+					case <-timer.C:
+						ed.Notify("bracketed paste timeout")
+						break paste
+					}
+				}
+				topaste := buf.String()
+				if ed.insert.quotePaste {
+					topaste = parse.Quote(topaste)
+				}
+				ed.InsertAtDot(topaste)
+			case tty.RawRune:
+				insertRaw(ed, rune(event))
+			case tty.KeyEvent:
+				k := ui.Key(event)
+			lookupKey:
+				fn := ed.mode.Binding(k)
+				if fn == nil {
+					ed.AddTip("Unbound and no default binding: %s", k)
+					continue MainLoop
+				}
+
+				ed.insert.insertedLiteral = false
+				ed.lastKey = k
+				ed.CallFn(fn)
+				if ed.insert.insertedLiteral {
+					ed.insert.literalInserts++
+				} else {
+					ed.insert.literalInserts = 0
+				}
+
+				switch ed.popAction() {
+				case reprocessKey:
+					err := ed.refresh(false, true)
+					if err != nil {
+						return "", err
+					}
+					goto lookupKey
+				case commitLine:
+					return ed.buffer, nil
+				case commitEOF:
+					return "", io.EOF
+				}
+			}
+		}
+	}
+}
+
+// getIsExternal finds a set of all external commands and puts it on the result
+// channel.
+func getIsExternal(result chan<- map[string]bool) {
+	isExternal := make(map[string]bool)
+	eval.EachExternal(func(name string) {
+		isExternal[name] = true
+	})
+	result <- isExternal
+}
diff -pruN 0.11+ds1-5/edit/edcore/highlight.go 0.12+ds1-1/edit/edcore/highlight.go
--- 0.11+ds1-5/edit/edcore/highlight.go	1970-01-01 00:00:00.000000000 +0000
+++ 0.12+ds1-1/edit/edcore/highlight.go	2018-07-11 02:32:25.000000000 +0000
@@ -0,0 +1,56 @@
+package edcore
+
+import (
+	"os"
+
+	"github.com/elves/elvish/edit/highlight"
+	"github.com/elves/elvish/eval"
+	"github.com/elves/elvish/parse"
+	"github.com/elves/elvish/util"
+)
+
+func doHighlight(n parse.Node, ed *editor) {
+	s := &highlight.Emitter{
+		func(s string) bool { return goodFormHead(s, ed) },
+		ed.styling.Add,
+	}
+	s.EmitAll(n)
+}
+
+func goodFormHead(head string, ed *editor) bool {
+	if eval.IsBuiltinSpecial[head] {
+		return true
+	} else if util.DontSearch(head) {
+		// XXX don't stat twice
+		return util.IsExecutable(head) || isDir(head)
+	} else {
+		ev := ed.evaler
+		explode, ns, name := eval.ParseVariableRef(head)
+		if !explode {
+			switch ns {
+			case "":
+				if ev.Builtin[name+eval.FnSuffix] != nil || ev.Global[name+eval.FnSuffix] != nil {
+					return true
+				}
+			case "e":
+				if ed.isExternal[name] {
+					return true
+				}
+			default:
+				mod := ev.Global[ns+eval.NsSuffix]
+				if mod == nil {
+					mod = ev.Builtin[ns+eval.NsSuffix]
+				}
+				if mod != nil && mod.Get().(eval.Ns)[name+eval.FnSuffix] != nil {
+					return true
+				}
+			}
+		}
+		return ed.isExternal[head]
+	}
+}
+
+func isDir(fname string) bool {
+	stat, err := os.Stat(fname)
+	return err == nil && stat.IsDir()
+}
diff -pruN 0.11+ds1-5/edit/edcore/hooks.go 0.12+ds1-1/edit/edcore/hooks.go
--- 0.11+ds1-5/edit/edcore/hooks.go	1970-01-01 00:00:00.000000000 +0000
+++ 0.12+ds1-1/edit/edcore/hooks.go	2018-07-11 02:32:25.000000000 +0000
@@ -0,0 +1,52 @@
+package edcore
+
+import (
+	"github.com/elves/elvish/eval"
+	"github.com/elves/elvish/eval/vals"
+	"github.com/elves/elvish/eval/vars"
+	"github.com/xiaq/persistent/vector"
+)
+
+// The $edit:{before,after}-readline lists that contain hooks. We might have more
+// hooks in future.
+
+// editorHooks contain hooks for the editor. They are just slices of functions;
+// each of them is initialized with a function that calls all Elvish functions
+// contained in the eponymous variable under edit:.
+type editorHooks struct {
+	beforeReadline []func()
+	afterReadline  []func(string)
+}
+
+func init() {
+	atEditorInit(func(ed *editor, ns eval.Ns) {
+		beforeReadline := vals.EmptyList
+		ns["before-readline"] = vars.FromPtr(&beforeReadline)
+		ed.AddBeforeReadline(func() { callHooks(ed, beforeReadline) })
+
+		afterReadline := vals.EmptyList
+		ns["after-readline"] = vars.FromPtr(&afterReadline)
+		ed.AddAfterReadline(func(s string) { callHooks(ed, afterReadline, s) })
+	})
+}
+
+// AddBeforeReadline adds a function to the before-readline hook.
+func (h *editorHooks) AddBeforeReadline(f func()) {
+	h.beforeReadline = append(h.beforeReadline, f)
+}
+
+// AddAfterReadline adds a function to the after-readline hook.
+func (h *editorHooks) AddAfterReadline(f func(string)) {
+	h.afterReadline = append(h.afterReadline, f)
+}
+
+func callHooks(ed *editor, li vector.Vector, args ...interface{}) {
+	for it := li.Iterator(); it.HasElem(); it.Next() {
+		fn, ok := it.Elem().(eval.Callable)
+		if !ok {
+			// TODO More detailed error message.
+			ed.Notify("hook not a function")
+		}
+		ed.CallFn(fn, args...)
+	}
+}
diff -pruN 0.11+ds1-5/edit/edcore/insert.go 0.12+ds1-1/edit/edcore/insert.go
--- 0.11+ds1-5/edit/edcore/insert.go	1970-01-01 00:00:00.000000000 +0000
+++ 0.12+ds1-1/edit/edcore/insert.go	2018-07-11 02:32:25.000000000 +0000
@@ -0,0 +1,378 @@
+package edcore
+
+import (
+	"strings"
+	"unicode"
+	"unicode/utf8"
+
+	"github.com/elves/elvish/edit/eddefs"
+	"github.com/elves/elvish/edit/ui"
+	"github.com/elves/elvish/eval"
+	"github.com/elves/elvish/eval/vars"
+	"github.com/elves/elvish/parse/parseutil"
+	"github.com/elves/elvish/util"
+)
+
+// Builtins related to insert and command mode.
+
+func init() {
+	atEditorInit(initCoreFns)
+	atEditorInit(initInsert)
+	atEditorInit(initCommand)
+}
+
+func initCoreFns(ed *editor, ns eval.Ns) {
+	ns.AddBuiltinFns("edit:", map[string]interface{}{
+		"kill-line-left":       ed.killLineLeft,
+		"kill-line-right":      ed.killLineRight,
+		"kill-word-left":       ed.killWordLeft,
+		"kill-small-word-left": ed.killSmallWordLeft,
+		"kill-rune-left":       ed.killRuneLeft,
+		"kill-rune-right":      ed.killRuneRight,
+
+		"move-dot-left":       ed.moveDotLeft,
+		"move-dot-right":      ed.moveDotRight,
+		"move-dot-left-word":  ed.moveDotLeftWord,
+		"move-dot-right-word": ed.moveDotRightWord,
+		"move-dot-sol":        ed.moveDotSOL,
+		"move-dot-eol":        ed.moveDotEOL,
+		"move-dot-up":         ed.moveDotUp,
+		"move-dot-down":       ed.moveDotDown,
+
+		"insert-last-word": ed.insertLastWord,
+		"insert-key":       ed.insertKey,
+
+		"return-line": ed.returnLine,
+		"smart-enter": ed.smartEnter,
+		"return-eof":  ed.returnEOF,
+
+		"toggle-quote-paste": ed.toggleQuotePaste,
+		"insert-raw":         ed.startInsertRaw,
+
+		"end-of-history": ed.endOfHistory,
+		"redraw":         ed.redraw,
+	})
+}
+
+type insert struct {
+	binding eddefs.BindingMap
+	insertState
+}
+
+type insertState struct {
+	quotePaste bool
+	// The number of consecutive key inserts. Used for abbreviation expansion.
+	literalInserts int
+	// Indicates whether a key was inserted (via insert-default). A hack for
+	// maintaining the inserts field.
+	insertedLiteral bool
+}
+
+func initInsert(ed *editor, ns eval.Ns) {
+	insert := &insert{binding: emptyBindingMap}
+	ed.insert = insert
+
+	insertNs := eval.Ns{
+		"binding": vars.FromPtr(&insert.binding),
+	}
+	insertNs.AddBuiltinFns("edit:insert:", map[string]interface{}{
+		"start":   ed.SetModeInsert,
+		"default": ed.insertDefault,
+	})
+	ns.AddNs("insert", insertNs)
+}
+
+func (ins *insert) Teardown() {
+	ins.insertState = insertState{}
+}
+
+// Insert mode is the default mode and has an empty mode, unless quotePaste is
+// true.
+func (ins *insert) ModeLine() ui.Renderer {
+	if ins.quotePaste {
+		return ui.NewModeLineRenderer(" INSERT (quote paste) ", "")
+	}
+	return nil
+}
+
+func (ins *insert) Binding(k ui.Key) eval.Callable {
+	return ins.binding.GetOrDefault(k)
+}
+
+type command struct {
+	binding eddefs.BindingMap
+}
+
+func initCommand(ed *editor, ns eval.Ns) {
+	command := &command{binding: emptyBindingMap}
+	ed.command = command
+
+	commandNs := eval.Ns{
+		"binding": vars.FromPtr(&command.binding),
+	}
+	commandNs.AddBuiltinFns("edit:command:", map[string]interface{}{
+		"start":   ed.commandStart,
+		"default": ed.commandDefault,
+	})
+	ns.AddNs("command", commandNs)
+}
+
+func (*command) Teardown() {}
+
+func (*command) ModeLine() ui.Renderer {
+	return ui.NewModeLineRenderer(" COMMAND ", "")
+}
+
+func (cmd *command) Binding(k ui.Key) eval.Callable {
+	return cmd.binding.GetOrDefault(k)
+}
+
+func (ed *editor) commandStart() {
+	ed.SetMode(ed.command)
+}
+
+func (ed *editor) killLineLeft() {
+	sol := util.FindLastSOL(ed.buffer[:ed.dot])
+	ed.buffer = ed.buffer[:sol] + ed.buffer[ed.dot:]
+	ed.dot = sol
+}
+
+func (ed *editor) killLineRight() {
+	eol := util.FindFirstEOL(ed.buffer[ed.dot:]) + ed.dot
+	ed.buffer = ed.buffer[:ed.dot] + ed.buffer[eol:]
+}
+
+// NOTE(xiaq): A word is a run of non-space runes. When killing a word,
+// trimming spaces are removed as well. Examples:
+// "abc  xyz" -> "abc  ", "abc xyz " -> "abc  ".
+
+func (ed *editor) killWordLeft() {
+	if ed.dot == 0 {
+		return
+	}
+	space := strings.LastIndexFunc(
+		strings.TrimRightFunc(ed.buffer[:ed.dot], unicode.IsSpace),
+		unicode.IsSpace) + 1
+	ed.buffer = ed.buffer[:space] + ed.buffer[ed.dot:]
+	ed.dot = space
+}
+
+// NOTE(xiaq): A small word is either a run of alphanumeric (Unicode category L
+// or N) runes or a run of non-alphanumeric runes. This is consistent with vi's
+// definition of word, except that "_" is not considered alphanumeric. When
+// killing a small word, trimming spaces are removed as well. Examples:
+// "abc/~" -> "abc", "~/abc" -> "~/", "abc* " -> "abc"
+
+func (ed *editor) killSmallWordLeft() {
+	left := strings.TrimRightFunc(ed.buffer[:ed.dot], unicode.IsSpace)
+	// The case of left == "" is handled as well.
+	r, _ := utf8.DecodeLastRuneInString(left)
+	if isAlnum(r) {
+		left = strings.TrimRightFunc(left, isAlnum)
+	} else {
+		left = strings.TrimRightFunc(
+			left, func(r rune) bool { return !isAlnum(r) })
+	}
+	ed.buffer = left + ed.buffer[ed.dot:]
+	ed.dot = len(left)
+}
+
+func isAlnum(r rune) bool {
+	return unicode.IsLetter(r) || unicode.IsNumber(r)
+}
+
+func (ed *editor) killRuneLeft() {
+	if ed.dot > 0 {
+		_, w := utf8.DecodeLastRuneInString(ed.buffer[:ed.dot])
+		ed.buffer = ed.buffer[:ed.dot-w] + ed.buffer[ed.dot:]
+		ed.dot -= w
+	} else {
+		ed.flash()
+	}
+}
+
+func (ed *editor) killRuneRight() {
+	if ed.dot < len(ed.buffer) {
+		_, w := utf8.DecodeRuneInString(ed.buffer[ed.dot:])
+		ed.buffer = ed.buffer[:ed.dot] + ed.buffer[ed.dot+w:]
+	} else {
+		ed.flash()
+	}
+}
+
+func (ed *editor) moveDotLeft() {
+	_, w := utf8.DecodeLastRuneInString(ed.buffer[:ed.dot])
+	ed.dot -= w
+}
+
+func (ed *editor) moveDotRight() {
+	_, w := utf8.DecodeRuneInString(ed.buffer[ed.dot:])
+	ed.dot += w
+}
+
+func (ed *editor) moveDotLeftWord() {
+	if ed.dot == 0 {
+		return
+	}
+	space := strings.LastIndexFunc(
+		strings.TrimRightFunc(ed.buffer[:ed.dot], unicode.IsSpace),
+		unicode.IsSpace) + 1
+	ed.dot = space
+}
+
+func (ed *editor) moveDotRightWord() {
+	// Move to first space
+	p := strings.IndexFunc(ed.buffer[ed.dot:], unicode.IsSpace)
+	if p == -1 {
+		ed.dot = len(ed.buffer)
+		return
+	}
+	ed.dot += p
+	// Move to first nonspace
+	p = strings.IndexFunc(ed.buffer[ed.dot:], notSpace)
+	if p == -1 {
+		ed.dot = len(ed.buffer)
+		return
+	}
+	ed.dot += p
+}
+
+func notSpace(r rune) bool {
+	return !unicode.IsSpace(r)
+}
+
+func (ed *editor) moveDotSOL() {
+	sol := util.FindLastSOL(ed.buffer[:ed.dot])
+	ed.dot = sol
+}
+
+func (ed *editor) moveDotEOL() {
+	eol := util.FindFirstEOL(ed.buffer[ed.dot:]) + ed.dot
+	ed.dot = eol
+}
+
+func (ed *editor) moveDotUp() {
+	sol := util.FindLastSOL(ed.buffer[:ed.dot])
+	if sol == 0 {
+		ed.flash()
+		return
+	}
+	prevEOL := sol - 1
+	prevSOL := util.FindLastSOL(ed.buffer[:prevEOL])
+	width := util.Wcswidth(ed.buffer[sol:ed.dot])
+	ed.dot = prevSOL + len(util.TrimWcwidth(ed.buffer[prevSOL:prevEOL], width))
+}
+
+func (ed *editor) moveDotDown() {
+	eol := util.FindFirstEOL(ed.buffer[ed.dot:]) + ed.dot
+	if eol == len(ed.buffer) {
+		ed.flash()
+		return
+	}
+	nextSOL := eol + 1
+	nextEOL := util.FindFirstEOL(ed.buffer[nextSOL:]) + nextSOL
+	sol := util.FindLastSOL(ed.buffer[:ed.dot])
+	width := util.Wcswidth(ed.buffer[sol:ed.dot])
+	ed.dot = nextSOL + len(util.TrimWcwidth(ed.buffer[nextSOL:nextEOL], width))
+}
+
+func (ed *editor) insertLastWord() {
+	if ed.daemon == nil {
+		ed.AddTip("daemon offline")
+		return
+	}
+	_, cmd, err := ed.daemon.PrevCmd(-1, "")
+	if err == nil {
+		ed.InsertAtDot(lastWord(cmd))
+	} else {
+		ed.AddTip("db error: %s", err.Error())
+	}
+}
+
+func lastWord(s string) string {
+	words := parseutil.Wordify(s)
+	if len(words) == 0 {
+		return ""
+	}
+	return words[len(words)-1]
+}
+
+func (ed *editor) insertKey() {
+	k := ed.lastKey
+	ed.InsertAtDot(string(k.Rune))
+}
+
+func (ed *editor) returnLine() {
+	ed.SetAction(commitLine)
+}
+
+func (ed *editor) smartEnter() {
+	if ed.parseErrorAtEnd {
+		// There is a parsing error at the end. ui.Insert a newline and copy
+		// indents from previous line.
+		indent := findLastIndent(ed.buffer[:ed.dot])
+		ed.InsertAtDot("\n" + indent)
+	} else {
+		ed.returnLine()
+	}
+}
+
+func findLastIndent(s string) string {
+	line := s[util.FindLastSOL(s):]
+	trimmed := strings.TrimLeft(line, " \t")
+	return line[:len(line)-len(trimmed)]
+}
+
+func (ed *editor) returnEOF() {
+	if len(ed.buffer) == 0 {
+		ed.SetAction(commitEOF)
+	}
+}
+
+func (ed *editor) toggleQuotePaste() {
+	ed.insert.quotePaste = !ed.insert.quotePaste
+}
+
+func (ed *editor) endOfHistory() {
+	ed.Notify("End of history")
+}
+
+func (ed *editor) redraw() {
+	ed.refresh(true, true)
+}
+
+func (ed *editor) insertDefault() {
+	k := ed.lastKey
+	if likeChar(k) {
+		ed.insertKey()
+		// Match abbreviations.
+		expanded := false
+		literals := ed.buffer[ed.dot-ed.insert.literalInserts-1 : ed.dot]
+		abbrIterate(ed.abbr, func(abbr, full string) bool {
+			if strings.HasSuffix(literals, abbr) {
+				ed.buffer = ed.buffer[:ed.dot-len(abbr)] + full + ed.buffer[ed.dot:]
+				ed.dot += len(full) - len(abbr)
+				expanded = true
+				return false
+			}
+			return true
+		})
+		// No match.
+		if !expanded {
+			ed.insert.insertedLiteral = true
+		}
+	} else {
+		ed.Notify("Unbound: %s", k)
+	}
+}
+
+// likeChar returns if a key looks like a character meant to be input (as
+// opposed to a function key).
+func likeChar(k ui.Key) bool {
+	return k.Mod == 0 && k.Rune > 0 && unicode.IsGraphic(k.Rune)
+}
+
+func (ed *editor) commandDefault() {
+	k := ed.lastKey
+	ed.Notify("Unbound: %s", k)
+}
diff -pruN 0.11+ds1-5/edit/edcore/listing.go 0.12+ds1-1/edit/edcore/listing.go
--- 0.11+ds1-5/edit/edcore/listing.go	1970-01-01 00:00:00.000000000 +0000
+++ 0.12+ds1-1/edit/edcore/listing.go	2018-07-11 02:32:25.000000000 +0000
@@ -0,0 +1,340 @@
+package edcore
+
+import (
+	"container/list"
+	"fmt"
+	"strings"
+	"unicode/utf8"
+
+	"github.com/elves/elvish/edit/eddefs"
+	"github.com/elves/elvish/edit/ui"
+	"github.com/elves/elvish/eval"
+	"github.com/elves/elvish/eval/vars"
+	"github.com/elves/elvish/util"
+)
+
+// listingMode implements a mode that supports listing, selecting and filtering
+// entries.
+type listingMode struct {
+	commonBinding eddefs.BindingMap
+	listingState
+}
+
+type listingState struct {
+	binding     eddefs.BindingMap
+	provider    eddefs.ListingProvider
+	selected    int
+	filter      string
+	pagesize    int
+	headerWidth int
+}
+
+func init() { atEditorInit(initListing) }
+
+func initListing(ed *editor, ns eval.Ns) {
+	l := &listingMode{commonBinding: emptyBindingMap}
+	ed.listing = l
+
+	subns := eval.Ns{
+		"binding": vars.FromPtr(&l.commonBinding),
+	}
+	subns.AddBuiltinFns("edit:listing:", map[string]interface{}{
+		"up":         func() { l.up(false) },
+		"up-cycle":   func() { l.up(true) },
+		"page-up":    func() { l.pageUp() },
+		"down":       func() { l.down(false) },
+		"down-cycle": func() { l.down(true) },
+		"page-down":  func() { l.pageDown() },
+		"backspace":  func() { l.backspace() },
+		"accept":     func() { l.accept(ed) },
+		"accept-close": func() {
+			l.accept(ed)
+			ed.SetModeInsert()
+		},
+		"default": func() { l.defaultBinding(ed) },
+	})
+	ns.AddNs("listing", subns)
+}
+
+type placeholderer interface {
+	Placeholder() string
+}
+
+func (l *listingMode) Teardown() {
+	l.listingState = listingState{}
+	if p, ok := l.provider.(teardowner); ok {
+		p.Teardown()
+	}
+}
+
+type teardowner interface {
+	Teardown()
+}
+
+func (l *listingMode) Binding(k ui.Key) eval.Callable {
+	specificBindings := l.binding
+	listingBindings := l.commonBinding
+	// mode-specific binding -> listing binding ->
+	// mode-specific default -> listing default
+	switch {
+	case specificBindings.HasKey(k):
+		return specificBindings.GetKey(k)
+	case listingBindings.HasKey(k):
+		return listingBindings.GetKey(k)
+	case specificBindings.HasKey(ui.Default):
+		return specificBindings.GetKey(ui.Default)
+	case listingBindings.HasKey(ui.Default):
+		return listingBindings.GetKey(ui.Default)
+	default:
+		return nil
+	}
+}
+
+func newListing(b eddefs.BindingMap, p eddefs.ListingProvider) *listingState {
+	l := &listingState{}
+	l.setup(b, p)
+	return l
+}
+
+func (l *listingState) setup(b eddefs.BindingMap, p eddefs.ListingProvider) {
+	*l = listingState{b, p, 0, "", 0, 0}
+	l.refresh()
+	for i := 0; i < p.Len(); i++ {
+		header, _ := p.Show(i)
+		width := util.Wcswidth(header)
+		if l.headerWidth < width {
+			l.headerWidth = width
+		}
+	}
+}
+
+func (l *listingState) ModeLine() ui.Renderer {
+	return ui.NewModeLineRenderer(l.provider.ModeTitle(l.selected), l.filter)
+}
+
+func (l *listingState) CursorOnModeLine() bool {
+	if c, ok := l.provider.(cursorOnModeLiner); ok {
+		return c.CursorOnModeLine()
+	}
+	return false
+}
+
+func (l *listingState) List(maxHeight int) ui.Renderer {
+	n := l.provider.Len()
+	if n == 0 {
+		var ph string
+		if pher, ok := l.provider.(placeholderer); ok {
+			ph = pher.Placeholder()
+		} else {
+			ph = "(no result)"
+		}
+		return placeholderRenderer(ph)
+	}
+
+	// Collect the entries to show. We start from the selected entry and extend
+	// in both directions alternatingly. The entries are split into lines and
+	// then collected in a list.
+	low := l.selected
+	if low == -1 {
+		low = 0
+	}
+	high := low
+	height := 0
+	var listOfLines list.List
+	getEntry := func(i int) []ui.Styled {
+		header, content := l.provider.Show(i)
+		lines := strings.Split(content.Text, "\n")
+		styles := content.Styles
+		if i == l.selected {
+			styles = append(styles, styleForSelected...)
+		}
+		styleds := make([]ui.Styled, len(lines))
+		for i, line := range lines {
+			if l.headerWidth > 0 {
+				if i == 0 {
+					line = fmt.Sprintf("%*s %s", l.headerWidth, header, line)
+				} else {
+					line = fmt.Sprintf("%*s %s", l.headerWidth, "", line)
+				}
+			}
+			styleds[i] = ui.Styled{line, styles}
+		}
+		return styleds
+	}
+	// We start by extending high, so that the first entry to include is
+	// l.selected.
+	extendLow := false
+	lastShownIncomplete := false
+	for height < maxHeight && !(low == 0 && high == n) {
+		var i int
+		if (extendLow && low > 0) || high == n {
+			low--
+
+			entry := getEntry(low)
+			// Prepend at most the last (height - maxHeight) lines.
+			for i = len(entry) - 1; i >= 0 && height < maxHeight; i-- {
+				listOfLines.PushFront(entry[i])
+				height++
+			}
+			if i >= 0 {
+				lastShownIncomplete = true
+			}
+		} else {
+			entry := getEntry(high)
+			// Append at most the first (height - maxHeight) lines.
+			for i = 0; i < len(entry) && height < maxHeight; i++ {
+				listOfLines.PushBack(entry[i])
+				height++
+			}
+			if i < len(entry) {
+				lastShownIncomplete = true
+			}
+
+			high++
+		}
+		extendLow = !extendLow
+	}
+
+	l.pagesize = high - low
+
+	// Convert the List to a slice.
+	lines := make([]ui.Styled, 0, listOfLines.Len())
+	for p := listOfLines.Front(); p != nil; p = p.Next() {
+		lines = append(lines, p.Value.(ui.Styled))
+	}
+
+	ls := listingRenderer{lines}
+	if low > 0 || high < n || lastShownIncomplete {
+		// Need scrollbar
+		return listingWithScrollBarRenderer{ls, n, low, high, height}
+	}
+	return ls
+}
+
+func renderScrollbar(n, low, high, height int) *ui.Buffer {
+	slow, shigh := findScrollInterval(n, low, high, height)
+	// Logger.Printf("low = %d, high = %d, n = %d, slow = %d, shigh = %d", low, high, n, slow, shigh)
+	bb := ui.NewBufferBuilder(1)
+	for i := 0; i < height; i++ {
+		if i > 0 {
+			bb.Newline()
+		}
+		if slow <= i && i < shigh {
+			bb.Write(' ', styleForScrollBarThumb.String())
+		} else {
+			bb.Write('│', styleForScrollBarArea.String())
+		}
+	}
+	return bb.Buffer()
+}
+
+func findScrollInterval(n, low, high, height int) (int, int) {
+	f := func(i int) int {
+		return int(float64(i)/float64(n)*float64(height) + 0.5)
+	}
+	scrollLow, scrollHigh := f(low), f(high)
+	if scrollLow == scrollHigh {
+		if scrollHigh == high {
+			scrollLow--
+		} else {
+			scrollHigh++
+		}
+	}
+	return scrollLow, scrollHigh
+}
+
+func (l *listingState) changeFilter(newfilter string) {
+	l.filter = newfilter
+	l.refresh()
+}
+
+func (l *listingState) refresh() {
+	l.selected = l.provider.Filter(l.filter)
+}
+
+func (l *listingState) backspace() bool {
+	_, size := utf8.DecodeLastRuneInString(l.filter)
+	if size > 0 {
+		l.changeFilter(l.filter[:len(l.filter)-size])
+		return true
+	}
+	return false
+}
+
+func (l *listingState) up(cycle bool) {
+	n := l.provider.Len()
+	if n == 0 {
+		return
+	}
+	l.selected--
+	if l.selected == -1 {
+		if cycle {
+			l.selected += n
+		} else {
+			l.selected++
+		}
+	}
+}
+
+func (l *listingState) pageUp() {
+	n := l.provider.Len()
+	if n == 0 {
+		return
+	}
+	l.selected -= l.pagesize
+	if l.selected < 0 {
+		l.selected = 0
+	}
+}
+
+func (l *listingState) down(cycle bool) {
+	n := l.provider.Len()
+	if n == 0 {
+		return
+	}
+	l.selected++
+	if l.selected == n {
+		if cycle {
+			l.selected -= n
+		} else {
+			l.selected--
+		}
+	}
+}
+
+func (l *listingState) pageDown() {
+	n := l.provider.Len()
+	if n == 0 {
+		return
+	}
+	l.selected += l.pagesize
+	if l.selected >= n {
+		l.selected = n - 1
+	}
+}
+
+func (l *listingState) accept(ed *editor) {
+	if l.selected >= 0 {
+		l.provider.Accept(l.selected, ed)
+	}
+}
+
+func (l *listingState) defaultBinding(ed *editor) {
+	k := ed.LastKey()
+	if likeChar(k) {
+		// Append to filter
+		l.changeFilter(l.filter + string(k.Rune))
+		if aa, ok := l.provider.(autoAccepter); ok {
+			if aa.AutoAccept() {
+				l.accept(ed)
+			}
+		}
+	} else {
+		ed.SetModeInsert()
+		ed.SetAction(reprocessKey)
+	}
+}
+
+type autoAccepter interface {
+	AutoAccept() bool
+}
diff -pruN 0.11+ds1-5/edit/edcore/listing_test.go 0.12+ds1-1/edit/edcore/listing_test.go
--- 0.11+ds1-5/edit/edcore/listing_test.go	1970-01-01 00:00:00.000000000 +0000
+++ 0.12+ds1-1/edit/edcore/listing_test.go	2018-07-11 02:32:25.000000000 +0000
@@ -0,0 +1,74 @@
+package edcore
+
+import (
+	"fmt"
+	"reflect"
+	"strconv"
+	"testing"
+
+	"github.com/elves/elvish/edit/eddefs"
+	"github.com/elves/elvish/edit/ui"
+)
+
+type provider struct {
+	elems    []string
+	accepted int
+}
+
+func (p provider) Len() int                       { return len(p.elems) }
+func (p provider) Filter(string) int              { return 0 }
+func (p provider) Accept(i int, ed eddefs.Editor) { p.accepted = i }
+func (p provider) ModeTitle(i int) string         { return fmt.Sprintf("test %d", i) }
+
+func (p provider) Show(i int) (string, ui.Styled) {
+	return strconv.Itoa(i), ui.Unstyled(p.elems[i])
+}
+
+var (
+	p  = provider{[]string{"foo", "bar", "foobar", "lorem", "ipsum"}, -1}
+	ls = newListing(emptyBindingMap, p)
+)
+
+func TestListing(t *testing.T) {
+	wantedModeLine := ui.NewModeLineRenderer("test 0", "")
+	if modeLine := ls.ModeLine(); modeLine != wantedModeLine {
+		t.Errorf("ls.ModeLine() = %v, want %v", modeLine, wantedModeLine)
+	}
+
+	// Selecting the first element and rendering with height=2. We expect to see
+	// the first 2 elements, with the first being shown as selected.
+	testListingList(t, 0, 2, listingWithScrollBarRenderer{
+		listingRenderer: listingRenderer{[]ui.Styled{
+			{"0 foo", styleForSelected},
+			{"1 bar", ui.Styles{}},
+		}},
+		n: 5, low: 0, high: 2, height: 2,
+	})
+	// Selecting the last element and rendering with height=2. We expect to see
+	// the last 2 elements, with the last being shown as selected.
+	testListingList(t, 4, 2, listingWithScrollBarRenderer{
+		listingRenderer: listingRenderer{[]ui.Styled{
+			{"3 lorem", ui.Styles{}},
+			{"4 ipsum", styleForSelected},
+		}},
+		n: 5, low: 3, high: 5, height: 2,
+	})
+	// Selecting the middle element and rendering with height=3. We expect to
+	// see the middle element and two elements around it, with the middle being
+	// shown as selected.
+	testListingList(t, 2, 3, listingWithScrollBarRenderer{
+		listingRenderer: listingRenderer{[]ui.Styled{
+			{"1 bar", ui.Styles{}},
+			{"2 foobar", styleForSelected},
+			{"3 lorem", ui.Styles{}},
+		}},
+		n: 5, low: 1, high: 4, height: 3,
+	})
+}
+
+func testListingList(t *testing.T, i, h int, want ui.Renderer) {
+	ls.selected = i
+	if r := ls.List(h); !reflect.DeepEqual(r, want) {
+		t.Errorf("selecting %d, ls.List(%d) = %v, want %v", i, h, r, want)
+	}
+}
diff -pruN 0.11+ds1-5/edit/edcore/max_height.go 0.12+ds1-1/edit/edcore/max_height.go
--- 0.11+ds1-5/edit/edcore/max_height.go	1970-01-01 00:00:00.000000000 +0000
+++ 0.12+ds1-1/edit/edcore/max_height.go	2018-07-11 02:32:25.000000000 +0000
@@ -0,0 +1,23 @@
+package edcore
+
+import (
+	"math"
+
+	"github.com/elves/elvish/eval"
+	"github.com/elves/elvish/eval/vars"
+	"github.com/elves/elvish/util"
+)
+
+func init() {
+	atEditorInit(func(ed *editor, ns eval.Ns) {
+		ed.maxHeight = math.Inf(1)
+		ns["max-height"] = vars.FromPtr(&ed.maxHeight)
+	})
+}
+
+func maxHeightToInt(h float64) int {
+	if math.IsInf(h, 1) {
+		return util.MaxInt
+	}
+	return int(h)
+}
diff -pruN 0.11+ds1-5/edit/edcore/minmax.go 0.12+ds1-1/edit/edcore/minmax.go
--- 0.11+ds1-5/edit/edcore/minmax.go	1970-01-01 00:00:00.000000000 +0000
+++ 0.12+ds1-1/edit/edcore/minmax.go	2018-07-11 02:32:25.000000000 +0000
@@ -0,0 +1,15 @@
+package edcore
+
+func min(a, b int) int {
+	if a <= b {
+		return a
+	}
+	return b
+}
+
+func max(a, b int) int {
+	if a >= b {
+		return a
+	}
+	return b
+}
diff -pruN 0.11+ds1-5/edit/edcore/mode.go 0.12+ds1-1/edit/edcore/mode.go
--- 0.11+ds1-5/edit/edcore/mode.go	1970-01-01 00:00:00.000000000 +0000
+++ 0.12+ds1-1/edit/edcore/mode.go	2018-07-11 02:32:25.000000000 +0000
@@ -0,0 +1,38 @@
+package edcore
+
+import "github.com/elves/elvish/edit/ui"
+
+// Additional interfaces mode implementations may satisfy.
+
+// cursorOnModeLiner is an optional interface that modes can implement. If a
+// mode does and the method returns true, the cursor is placed on the modeline
+// when that mode is active.
+type cursorOnModeLiner interface {
+	CursorOnModeLine() bool
+}
+
+type replacementer interface {
+	// Replacement returns the part of the buffer that is replaced.
+	Replacement() (begin, end int, text string)
+}
+
+type redrawModeLiner interface {
+	// RedrawModelLine indicates that the modeline should be redrawn after
+	// listing. This is only used in completion mode now.
+	RedrawModeLine()
+}
+
+// lister is an optional interface that modes can implement. If a mode
+// implements this interface, the result of this method will be shown in the
+// listing area.
+type lister interface {
+	List(maxHeight int) ui.Renderer
+}
+
+// listRenderer is similar to lister, but the mode handles the rendering itself.
+// NOTE(xiaq): This interface is being deprecated in favor of Lister.
+type listRenderer interface {
+	// ListRender renders the listing under the given constraint of width and
+	// maximum height. It returns a rendered buffer.
+	ListRender(width, maxHeight int) *ui.Buffer
+}
diff -pruN 0.11+ds1-5/edit/edcore/modules.go 0.12+ds1-1/edit/edcore/modules.go
--- 0.11+ds1-5/edit/edcore/modules.go	1970-01-01 00:00:00.000000000 +0000
+++ 0.12+ds1-1/edit/edcore/modules.go	2018-07-11 02:32:25.000000000 +0000
@@ -0,0 +1,20 @@
+package edcore
+
+import (
+	"github.com/elves/elvish/edit/completion"
+	"github.com/elves/elvish/edit/history"
+	"github.com/elves/elvish/edit/lastcmd"
+	"github.com/elves/elvish/edit/location"
+	"github.com/elves/elvish/edit/prompt"
+	"github.com/elves/elvish/eval"
+)
+
+func init() {
+	atEditorInit(func(ed *editor, ns eval.Ns) {
+		location.Init(ed, ns)
+		lastcmd.Init(ed, ns)
+		history.Init(ed, ns)
+		completion.Init(ed, ns)
+		prompt.Init(ed, ns)
+	})
+}
diff -pruN 0.11+ds1-5/edit/edcore/narrow.go 0.12+ds1-1/edit/edcore/narrow.go
--- 0.11+ds1-5/edit/edcore/narrow.go	1970-01-01 00:00:00.000000000 +0000
+++ 0.12+ds1-1/edit/edcore/narrow.go	2018-07-11 02:32:25.000000000 +0000
@@ -0,0 +1,456 @@
+package edcore
+
+import (
+	"container/list"
+	"strings"
+	"unicode/utf8"
+
+	"github.com/elves/elvish/edit/eddefs"
+	"github.com/elves/elvish/edit/ui"
+	"github.com/elves/elvish/eval"
+	"github.com/elves/elvish/eval/vals"
+	"github.com/elves/elvish/eval/vars"
+	"github.com/elves/elvish/parse/parseutil"
+	"github.com/xiaq/persistent/hashmap"
+)
+
+// narrow implements a listing mode that supports the notion of selecting an
+// entry and filtering entries.
+type narrow struct {
+	binding eddefs.BindingMap
+	narrowState
+}
+
+func init() { atEditorInit(initNarrow) }
+
+func initNarrow(ed *editor, ns eval.Ns) {
+	n := &narrow{binding: emptyBindingMap}
+	subns := eval.Ns{
+		"binding": vars.FromPtr(&n.binding),
+	}
+	subns.AddBuiltinFns("edit:narrow:", map[string]interface{}{
+		"up":         func() { n.up(false) },
+		"up-cycle":   func() { n.up(true) },
+		"page-up":    func() { n.pageUp() },
+		"down":       func() { n.down(false) },
+		"down-cycle": func() { n.down(true) },
+		"page-down":  func() { n.pageDown() },
+		"backspace":  func() { n.backspace() },
+		"accept":     func() { n.accept(ed) },
+		"accept-close": func() {
+			n.accept(ed)
+			ed.SetModeInsert()
+		},
+		"toggle-ignore-duplication": func() {
+			n.opts.IgnoreDuplication = !n.opts.IgnoreDuplication
+			n.refresh()
+		},
+		"toggle-ignore-case": func() {
+			n.opts.IgnoreCase = !n.opts.IgnoreCase
+			n.refresh()
+		},
+		"default": func() { n.defaultBinding(ed) },
+	})
+	ns.AddNs("narrow", subns)
+	ns.AddBuiltinFn("edit:", "-narrow-read", n.NarrowRead)
+}
+
+type narrowState struct {
+	name        string
+	selected    int
+	filter      string
+	pagesize    int
+	headerWidth int
+
+	placehold string
+	source    func() []narrowItem
+	action    func(*editor, narrowItem)
+	match     func(string, string) bool
+	filtered  []narrowItem
+	opts      narrowOptions
+}
+
+func (l *narrow) Teardown() {
+	l.narrowState = narrowState{}
+}
+
+func (l *narrow) Binding(k ui.Key) eval.Callable {
+	if l.opts.bindingMap != nil {
+		if f, ok := l.opts.bindingMap[k]; ok {
+			return f
+		}
+	}
+	return l.binding.GetOrDefault(k)
+}
+
+func (l *narrowState) ModeLine() ui.Renderer {
+	ml := l.opts.Modeline
+	var opt []string
+	if l.opts.AutoCommit {
+		opt = append(opt, "A")
+	}
+	if l.opts.IgnoreCase {
+		opt = append(opt, "C")
+	}
+	if l.opts.IgnoreDuplication {
+		opt = append(opt, "D")
+	}
+	if len(opt) != 0 {
+		ml += "[" + strings.Join(opt, " ") + "]"
+	}
+	return ui.NewModeLineRenderer(ml, l.filter)
+}
+
+func (l *narrowState) CursorOnModeLine() bool {
+	return true
+}
+
+func (l *narrowState) List(maxHeight int) ui.Renderer {
+	if l.opts.MaxLines > 0 && l.opts.MaxLines < maxHeight {
+		maxHeight = l.opts.MaxLines
+	}
+
+	if l.filtered == nil {
+		l.refresh()
+	}
+	n := len(l.filtered)
+	if n == 0 {
+		return placeholderRenderer(l.placehold)
+	}
+
+	// Collect the entries to show. We start from the selected entry and extend
+	// in both directions alternatingly. The entries are split into lines and
+	// then collected in a list.
+	low := l.selected
+	if low == -1 {
+		low = 0
+	}
+	high := low
+	height := 0
+	var listOfLines list.List
+	getEntry := func(i int) []ui.Styled {
+		display := l.filtered[i].Display()
+		lines := strings.Split(display.Text, "\n")
+		styles := display.Styles
+		if i == l.selected {
+			styles = append(styles, styleForSelected...)
+		}
+		styleds := make([]ui.Styled, len(lines))
+		for i, line := range lines {
+			styleds[i] = ui.Styled{line, styles}
+		}
+		return styleds
+	}
+	// We start by extending high, so that the first entry to include is
+	// l.selected.
+	extendLow := false
+	lastShownIncomplete := false
+	for height < maxHeight && !(low == 0 && high == n) {
+		var i int
+		if (extendLow && low > 0) || high == n {
+			low--
+
+			entry := getEntry(low)
+			// Prepend at most the last (height - maxHeight) lines.
+			for i = len(entry) - 1; i >= 0 && height < maxHeight; i-- {
+				listOfLines.PushFront(entry[i])
+				height++
+			}
+			if i >= 0 {
+				lastShownIncomplete = true
+			}
+		} else {
+			entry := getEntry(high)
+			// Append at most the first (height - maxHeight) lines.
+			for i = 0; i < len(entry) && height < maxHeight; i++ {
+				listOfLines.PushBack(entry[i])
+				height++
+			}
+			if i < len(entry) {
+				lastShownIncomplete = true
+			}
+
+			high++
+		}
+		extendLow = !extendLow
+	}
+
+	l.pagesize = high - low
+
+	// Convert the List to a slice.
+	lines := make([]ui.Styled, 0, listOfLines.Len())
+	for p := listOfLines.Front(); p != nil; p = p.Next() {
+		lines = append(lines, p.Value.(ui.Styled))
+	}
+
+	ls := listingRenderer{lines}
+	if low > 0 || high < n || lastShownIncomplete {
+		// Need scrollbar
+		return listingWithScrollBarRenderer{ls, n, low, high, height}
+	}
+	return ls
+}
+
+func (l *narrowState) refresh() {
+	var candidates []narrowItem
+	if l.source != nil {
+		candidates = l.source()
+	}
+	l.filtered = make([]narrowItem, 0, len(candidates))
+
+	filter := l.filter
+	if l.opts.IgnoreCase {
+		filter = strings.ToLower(filter)
+	}
+
+	set := make(map[string]struct{})
+
+	for _, item := range candidates {
+		text := item.FilterText()
+		s := text
+		if l.opts.IgnoreCase {
+			s = strings.ToLower(s)
+		}
+		if !l.match(s, filter) {
+			continue
+		}
+		if l.opts.IgnoreDuplication {
+			if _, ok := set[text]; ok {
+				continue
+			}
+			set[text] = struct{}{}
+		}
+		l.filtered = append(l.filtered, item)
+	}
+
+	if l.opts.KeepBottom {
+		l.selected = len(l.filtered) - 1
+	} else {
+		l.selected = 0
+	}
+}
+
+func (l *narrowState) changeFilter(newfilter string) {
+	l.filter = newfilter
+	l.refresh()
+}
+
+func (l *narrowState) backspace() bool {
+	_, size := utf8.DecodeLastRuneInString(l.filter)
+	if size > 0 {
+		l.changeFilter(l.filter[:len(l.filter)-size])
+		return true
+	}
+	return false
+}
+
+func (l *narrowState) up(cycle bool) {
+	n := len(l.filtered)
+	if n == 0 {
+		return
+	}
+	l.selected--
+	if l.selected == -1 {
+		if cycle {
+			l.selected += n
+		} else {
+			l.selected++
+		}
+	}
+}
+
+func (l *narrowState) pageUp() {
+	n := len(l.filtered)
+	if n == 0 {
+		return
+	}
+	l.selected -= l.pagesize
+	if l.selected < 0 {
+		l.selected = 0
+	}
+}
+
+func (l *narrowState) down(cycle bool) {
+	n := len(l.filtered)
+	if n == 0 {
+		return
+	}
+	l.selected++
+	if l.selected == n {
+		if cycle {
+			l.selected -= n
+		} else {
+			l.selected--
+		}
+	}
+}
+
+func (l *narrowState) pageDown() {
+	n := len(l.filtered)
+	if n == 0 {
+		return
+	}
+	l.selected += l.pagesize
+	if l.selected >= n {
+		l.selected = n - 1
+	}
+}
+
+func (l *narrowState) accept(ed *editor) {
+	if l.selected >= 0 {
+		l.action(ed, l.filtered[l.selected])
+	}
+}
+
+func (l *narrowState) handleFilterKey(ed *editor) bool {
+	k := ed.lastKey
+	if likeChar(k) {
+		l.changeFilter(l.filter + string(k.Rune))
+		if len(l.filtered) == 1 && l.opts.AutoCommit {
+			l.accept(ed)
+			ed.SetModeInsert()
+		}
+		return true
+	}
+	return false
+}
+
+func (l *narrowState) defaultBinding(ed *editor) {
+	if !l.handleFilterKey(ed) {
+		ed.SetModeInsert()
+		ed.SetAction(reprocessKey)
+	}
+}
+
+type narrowItem interface {
+	Display() ui.Styled
+	Content() string
+	FilterText() string
+}
+
+type narrowOptions struct {
+	AutoCommit        bool
+	Bindings          hashmap.Map
+	IgnoreDuplication bool
+	IgnoreCase        bool
+	KeepBottom        bool
+	MaxLines          int
+	Modeline          string
+
+	bindingMap map[ui.Key]eval.Callable
+}
+
+type narrowItemString struct {
+	String string
+}
+
+func (s *narrowItemString) Content() string {
+	return s.String
+}
+
+func (s *narrowItemString) Display() ui.Styled {
+	return ui.Unstyled(s.String)
+}
+
+func (s *narrowItemString) FilterText() string {
+	return s.Content()
+}
+
+type narrowItemComplex struct {
+	hashmap.Map
+}
+
+func (c *narrowItemComplex) Content() string {
+	if v, ok := c.Map.Index("content"); ok {
+		if s, ok := v.(string); ok {
+			return s
+		}
+	}
+	return ""
+}
+
+// TODO: add style
+func (c *narrowItemComplex) Display() ui.Styled {
+	if v, ok := c.Map.Index("display"); ok {
+		if s, ok := v.(string); ok {
+			return ui.Unstyled(s)
+		}
+	}
+	return ui.Unstyled("")
+}
+
+func (c *narrowItemComplex) FilterText() string {
+	if v, ok := c.Map.Index("filter-text"); ok {
+		if s, ok := v.(string); ok {
+			return s
+		}
+	}
+	return c.Content()
+}
+
+func (n *narrow) NarrowRead(fm *eval.Frame, opts eval.RawOptions, source, action eval.Callable) {
+	l := &narrowState{
+		opts: narrowOptions{
+			Bindings: vals.EmptyMap,
+		},
+	}
+
+	opts.Scan(&l.opts)
+
+	for it := l.opts.Bindings.Iterator(); it.HasElem(); it.Next() {
+		k, v := it.Elem()
+		key := ui.ToKey(k)
+		val, ok := v.(eval.Callable)
+		if !ok {
+			throwf("should be fn")
+		}
+		if l.opts.bindingMap == nil {
+			l.opts.bindingMap = make(map[ui.Key]eval.Callable)
+		}
+		l.opts.bindingMap[key] = val
+	}
+
+	l.source = narrowGetSource(fm, source)
+	l.action = func(ed *editor, item narrowItem) {
+		ed.CallFn(action, item)
+	}
+	// TODO: user customize varible
+	l.match = strings.Contains
+
+	l.changeFilter("")
+
+	n.narrowState = *l
+	fm.Editor.(*editor).SetMode(n)
+}
+
+func narrowGetSource(ec *eval.Frame, source eval.Callable) func() []narrowItem {
+	return func() []narrowItem {
+		ed := ec.Editor.(*editor)
+		vs, err := ec.CaptureOutput(source, eval.NoArgs, eval.NoOpts)
+		if err != nil {
+			ed.Notify(err.Error())
+			return nil
+		}
+		var lis []narrowItem
+		for _, v := range vs {
+			switch raw := v.(type) {
+			case string:
+				lis = append(lis, &narrowItemString{raw})
+			case hashmap.Map:
+				lis = append(lis, &narrowItemComplex{raw})
+			}
+		}
+		return lis
+	}
+}
+
+func (ed *editor) replaceInput(text string) {
+	ed.buffer = text
+}
+
+func wordifyBuiltin(fm *eval.Frame, text string) {
+	out := fm.OutputChan()
+	for _, s := range parseutil.Wordify(text) {
+		out <- s
+	}
+}
diff -pruN 0.11+ds1-5/edit/edcore/navigation.go 0.12+ds1-1/edit/edcore/navigation.go
--- 0.11+ds1-5/edit/edcore/navigation.go	1970-01-01 00:00:00.000000000 +0000
+++ 0.12+ds1-1/edit/edcore/navigation.go	2018-07-11 02:32:25.000000000 +0000
@@ -0,0 +1,437 @@
+package edcore
+
+import (
+	"errors"
+	"os"
+	"path"
+	"sort"
+	"strings"
+	"unicode/utf8"
+
+	"github.com/elves/elvish/edit/eddefs"
+	"github.com/elves/elvish/edit/lscolors"
+	"github.com/elves/elvish/edit/ui"
+	"github.com/elves/elvish/eval"
+	"github.com/elves/elvish/eval/vars"
+	"github.com/elves/elvish/parse"
+	"github.com/elves/elvish/util"
+)
+
+// Navigation subsystem.
+
+// Interface.
+
+type navigation struct {
+	binding eddefs.BindingMap
+	chdir   func(string) error
+	navigationState
+}
+
+type navigationState struct {
+	current    *navColumn
+	parent     *navColumn
+	preview    navPreview
+	showHidden bool
+	filtering  bool
+	filter     string
+}
+
+func init() { atEditorInit(initNavigation) }
+
+func initNavigation(ed *editor, ns eval.Ns) {
+	n := &navigation{
+		binding: emptyBindingMap,
+		chdir:   ed.Evaler().Chdir,
+	}
+	ed.navigation = n
+
+	subns := eval.Ns{
+		"binding": vars.FromPtr(&n.binding),
+	}
+	subns.AddBuiltinFns("edit:navigation:", map[string]interface{}{
+		"start":                    func() { n.start(ed) },
+		"up":                       n.prev,
+		"down":                     n.next,
+		"page-up":                  n.pageUp,
+		"page-down":                n.pageDown,
+		"left":                     n.ascend,
+		"right":                    n.descend,
+		"file-preview-up":          n.filePreviewUp,
+		"file-preview-down":        n.filePreviewDown,
+		"trigger-shown-hidden":     n.triggerShowHidden,
+		"trigger-filter":           n.triggerFilter,
+		"insert-selected":          func() { n.insertSelected(ed) },
+		"insert-selected-and-quit": func() { n.insertSelectedAndQuit(ed) },
+		"default":                  func() { n.defaultFn(ed) },
+	})
+	ns.AddNs("navigation", subns)
+}
+
+type navPreview interface {
+	FullWidth(int) int
+	List(int) ui.Renderer
+}
+
+func (n *navigation) Teardown() {
+	n.navigationState = navigationState{}
+}
+
+func (n *navigation) Binding(k ui.Key) eval.Callable {
+	return n.binding.GetOrDefault(k)
+}
+
+func (n *navigation) ModeLine() ui.Renderer {
+	title := " NAVIGATING "
+	if n.showHidden {
+		title += "(show hidden) "
+	}
+	return ui.NewModeLineRenderer(title, n.filter)
+}
+
+func (n *navigation) CursorOnModeLine() bool {
+	return n.filtering
+}
+
+func (n *navigation) start(ed *editor) {
+	n.refresh()
+	ed.SetMode(n)
+}
+
+func (n *navigation) pageUp() {
+	n.current.pageUp()
+	n.refresh()
+}
+
+func (n *navigation) pageDown() {
+	n.current.pageDown()
+	n.refresh()
+}
+
+func (n *navigation) filePreviewUp() {
+	fp, ok := n.preview.(*navFilePreview)
+	if ok {
+		if fp.beginLine > 0 {
+			fp.beginLine--
+		}
+	}
+}
+
+func (n *navigation) filePreviewDown() {
+	fp, ok := n.preview.(*navFilePreview)
+	if ok {
+		if fp.beginLine < len(fp.lines)-1 {
+			fp.beginLine++
+		}
+	}
+}
+
+func (n *navigation) triggerShowHidden() {
+	n.showHidden = !n.showHidden
+	n.refresh()
+}
+
+func (n *navigation) triggerFilter() {
+	n.filtering = !n.filtering
+}
+
+func (n *navigation) insertSelected(ed *editor) {
+	ed.InsertAtDot(parse.Quote(n.current.selectedName()) + " ")
+}
+
+func (n *navigation) insertSelectedAndQuit(ed *editor) {
+	ed.InsertAtDot(parse.Quote(n.current.selectedName()) + " ")
+	ed.SetModeInsert()
+}
+
+func (n *navigation) defaultFn(ed *editor) {
+	// Use key binding for insert mode without exiting nigation mode.
+	k := ed.lastKey
+	if n.filtering && likeChar(k) {
+		n.filter += k.String()
+		n.refreshCurrent()
+		n.refreshDirPreview()
+	} else if n.filtering && k == (ui.Key{ui.Backspace, 0}) {
+		_, size := utf8.DecodeLastRuneInString(n.filter)
+		if size > 0 {
+			n.filter = n.filter[:len(n.filter)-size]
+			n.refreshCurrent()
+			n.refreshDirPreview()
+		}
+	} else {
+		fn := ed.insert.binding.GetOrDefault(k)
+		if fn == nil {
+			ed.Notify("key %s unbound and no default binding", k)
+		} else {
+			ed.CallFn(fn)
+		}
+	}
+}
+
+// Implementation.
+// TODO(xiaq): Remember which file was selected in each directory.
+
+var errorEmptyCwd = errors.New("current directory is empty")
+
+func (n *navigation) maintainSelected(name string) {
+	n.current.selected = 0
+	for i, s := range n.current.candidates {
+		if s.Text > name {
+			break
+		}
+		n.current.selected = i
+	}
+}
+
+func (n *navigation) refreshCurrent() {
+	selectedName := n.current.selectedName()
+	all, err := n.loaddir(".")
+	if err != nil {
+		n.current = newErrNavColumn(err)
+		return
+	}
+	// Try to select the old selected file.
+	// XXX(xiaq): This would break when we support alternative ordering.
+	n.current = newNavColumn(all, func(i int) bool {
+		return i == 0 || all[i].Text <= selectedName
+	})
+	n.current.changeFilter(n.filter)
+	n.maintainSelected(selectedName)
+}
+
+func (n *navigation) refreshParent() {
+	wd, err := os.Getwd()
+	if err != nil {
+		n.parent = newErrNavColumn(err)
+		return
+	}
+	if wd == "/" {
+		n.parent = newNavColumn(nil, nil)
+	} else {
+		all, err := n.loaddir("..")
+		if err != nil {
+			n.parent = newErrNavColumn(err)
+			return
+		}
+		cwd, err := os.Stat(".")
+		if err != nil {
+			n.parent = newErrNavColumn(err)
+			return
+		}
+		n.parent = newNavColumn(all, func(i int) bool {
+			d, _ := os.Lstat("../" + all[i].Text)
+			return os.SameFile(d, cwd)
+		})
+	}
+}
+
+func (n *navigation) refreshDirPreview() {
+	if n.current.selected != -1 {
+		name := n.current.selectedName()
+		fi, err := os.Stat(name)
+		if err != nil {
+			n.preview = newErrNavColumn(err)
+			return
+		}
+		if fi.Mode().IsDir() {
+			all, err := n.loaddir(name)
+			if err != nil {
+				n.preview = newErrNavColumn(err)
+				return
+			}
+			n.preview = newNavColumn(all, func(int) bool { return false })
+		} else {
+			n.preview = makeNavFilePreview(name)
+		}
+	} else {
+		n.preview = nil
+	}
+}
+
+// refresh rereads files in current and parent directories and maintains the
+// selected file if possible.
+func (n *navigation) refresh() {
+	n.refreshCurrent()
+	n.refreshParent()
+	n.refreshDirPreview()
+}
+
+// ascend changes current directory to the parent.
+// TODO(xiaq): navigation.{ascend descend} bypasses the cd builtin. This can be
+// problematic if cd acquires more functionality (e.g. trigger a hook).
+func (n *navigation) ascend() error {
+	wd, err := os.Getwd()
+	if err != nil {
+		return err
+	}
+	if wd == "/" {
+		return nil
+	}
+
+	name := n.parent.selectedName()
+	err = os.Chdir("..")
+	if err != nil {
+		return err
+	}
+	n.filter = ""
+	n.refresh()
+	n.maintainSelected(name)
+	// XXX Refresh dir preview again. We should perhaps not have used refresh
+	// above.
+	n.refreshDirPreview()
+	return nil
+}
+
+// descend changes current directory to the selected file, if it is a
+// directory.
+func (n *navigation) descend() error {
+	if n.current.selected == -1 {
+		return errorEmptyCwd
+	}
+	name := n.current.selectedName()
+	err := n.chdir(name)
+	if err != nil {
+		return err
+	}
+	n.filter = ""
+	n.current.selected = -1
+	n.refresh()
+	n.refreshDirPreview()
+	return nil
+}
+
+// prev selects the previous file.
+func (n *navigation) prev() {
+	if n.current.selected > 0 {
+		n.current.selected--
+	}
+	n.refresh()
+}
+
+// next selects the next file.
+func (n *navigation) next() {
+	if n.current.selected != -1 && n.current.selected < len(n.current.candidates)-1 {
+		n.current.selected++
+	}
+	n.refresh()
+}
+
+func (n *navigation) loaddir(dir string) ([]ui.Styled, error) {
+	f, err := os.Open(dir)
+	if err != nil {
+		return nil, err
+	}
+	names, err := f.Readdirnames(-1)
+	if err != nil {
+		return nil, err
+	}
+	sort.Strings(names)
+
+	var all []ui.Styled
+	lsColor := lscolors.GetColorist()
+	for _, name := range names {
+		if n.showHidden || name[0] != '.' {
+			all = append(all, ui.Styled{name,
+				ui.StylesFromString(lsColor.GetStyle(path.Join(dir, name)))})
+		}
+	}
+
+	return all, nil
+}
+
+func (n *navigation) List(maxHeight int) ui.Renderer {
+	return makeNavRenderer(
+		maxHeight,
+		n.parent.FullWidth(maxHeight),
+		n.current.FullWidth(maxHeight),
+		n.preview.FullWidth(maxHeight),
+		n.parent.List(maxHeight),
+		n.current.List(maxHeight),
+		n.preview.List(maxHeight),
+	)
+}
+
+// navColumn is a column in the navigation layout.
+type navColumn struct {
+	listingMode
+	all        []ui.Styled
+	candidates []ui.Styled
+	// selected int
+	err error
+}
+
+func newNavColumn(all []ui.Styled, sel func(int) bool) *navColumn {
+	nc := &navColumn{all: all, candidates: all}
+	nc.provider = nc
+	nc.selected = -1
+	for i := range all {
+		if sel(i) {
+			nc.selected = i
+		}
+	}
+	return nc
+}
+
+func newErrNavColumn(err error) *navColumn {
+	nc := &navColumn{err: err}
+	nc.provider = nc
+	return nc
+}
+
+func (nc *navColumn) Placeholder() string {
+	if nc.err != nil {
+		return nc.err.Error()
+	}
+	return ""
+}
+
+func (nc *navColumn) Len() int {
+	return len(nc.candidates)
+}
+
+func (nc *navColumn) Show(i int) (string, ui.Styled) {
+	cand := nc.candidates[i]
+	return "", ui.Styled{" " + cand.Text + " ", cand.Styles}
+}
+
+func (nc *navColumn) Filter(filter string) int {
+	nc.candidates = nc.candidates[:0]
+	for _, s := range nc.all {
+		if strings.Contains(s.Text, filter) {
+			nc.candidates = append(nc.candidates, s)
+		}
+	}
+	return 0
+}
+
+func (nc *navColumn) FullWidth(h int) int {
+	if nc == nil {
+		return 0
+	}
+	if nc.err != nil {
+		return util.Wcswidth(nc.err.Error())
+	}
+	maxw := 0
+	for _, s := range nc.candidates {
+		maxw = max(maxw, util.Wcswidth(s.Text)+2)
+	}
+	if len(nc.candidates) > h {
+		maxw++
+	}
+	return maxw
+}
+
+func (nc *navColumn) Accept(i int, ed eddefs.Editor) {
+	// TODO
+}
+
+func (nc *navColumn) ModeTitle(i int) string {
+	// Not used
+	return ""
+}
+
+func (nc *navColumn) selectedName() string {
+	if nc == nil || nc.selected == -1 || nc.selected >= len(nc.candidates) {
+		return ""
+	}
+	return nc.candidates[nc.selected].Text
+}
diff -pruN 0.11+ds1-5/edit/edcore/navigation_preview.go 0.12+ds1-1/edit/edcore/navigation_preview.go
--- 0.11+ds1-5/edit/edcore/navigation_preview.go	1970-01-01 00:00:00.000000000 +0000
+++ 0.12+ds1-1/edit/edcore/navigation_preview.go	2018-07-11 02:32:25.000000000 +0000
@@ -0,0 +1,91 @@
+package edcore
+
+import (
+	"errors"
+	"io"
+	"os"
+	"strings"
+	"unicode/utf8"
+
+	"github.com/elves/elvish/edit/ui"
+	"github.com/elves/elvish/util"
+)
+
+// previewBytes is the maximum number of bytes to preview a file.
+const previewBytes = 64 * 1024
+
+// Errors displayed in the preview area of navigation mode.
+var (
+	errNotRegular   = errors.New("no preview for non-regular file")
+	errNotValidUTF8 = errors.New("no preview for non-utf8 file")
+)
+
+type navFilePreview struct {
+	lines     []ui.Styled
+	fullWidth int
+	beginLine int
+}
+
+func newNavFilePreview(lines []string) *navFilePreview {
+	width := 0
+	convertedLines := make([]ui.Styled, len(lines))
+	for i, line := range lines {
+		// BUG: Handle tabstops correctly
+		convertedLine := strings.Replace(line, "\t", "    ", -1)
+		convertedLines[i] = ui.Unstyled(convertedLine)
+		width = max(width, util.Wcswidth(convertedLine))
+	}
+	return &navFilePreview{convertedLines, width, 0}
+}
+
+func (fp *navFilePreview) FullWidth(h int) int {
+	width := fp.fullWidth
+	if h < len(fp.lines) {
+		return width + 1
+	}
+	return width
+}
+
+func (fp *navFilePreview) List(h int) ui.Renderer {
+	if len(fp.lines) <= h {
+		logger.Printf("Height %d fit all lines", h)
+		return listingRenderer{fp.lines}
+	}
+	shown := fp.lines[fp.beginLine:]
+	if len(shown) > h {
+		shown = shown[:h]
+	}
+	logger.Printf("Showing lines %d to %d", fp.beginLine, fp.beginLine+len(shown))
+	return listingWithScrollBarRenderer{
+		listingRenderer{shown}, len(fp.lines),
+		fp.beginLine, fp.beginLine + len(shown), h}
+}
+
+func makeNavFilePreview(fname string) navPreview {
+	file, err := os.Open(fname)
+	if err != nil {
+		return newErrNavColumn(err)
+	}
+
+	info, err := file.Stat()
+	if err != nil {
+		return newErrNavColumn(err)
+	}
+	if (info.Mode() & (os.ModeDevice | os.ModeNamedPipe | os.ModeSocket | os.ModeCharDevice)) != 0 {
+		return newErrNavColumn(errNotRegular)
+	}
+
+	// BUG: when the file is bigger than the buffer, the scrollbar is wrong.
+	var buf [previewBytes]byte
+	nr, err := file.Read(buf[:])
+	if err != nil && err != io.EOF {
+		return newErrNavColumn(err)
+	}
+
+	content := buf[:nr]
+	if !utf8.Valid(content) {
+		return newErrNavColumn(errNotValidUTF8)
+	}
+
+	return newNavFilePreview(strings.Split(string(content), "\n"))
+}
diff -pruN 0.11+ds1-5/edit/edcore/navigation_width.go 0.12+ds1-1/edit/edcore/navigation_width.go
--- 0.11+ds1-5/edit/edcore/navigation_width.go	1970-01-01 00:00:00.000000000 +0000
+++ 0.12+ds1-1/edit/edcore/navigation_width.go	2018-07-11 02:32:25.000000000 +0000
@@ -0,0 +1,30 @@
+package edcore
+
+// getNavWidths calculates the widths for the three (parent, current and
+// preview) columns in the navigation mode. It takes the available width, full
+// width required to display the current and preview columns, and returns
+// suitable widths for the columns.
+//
+// The parent column always gets 1/6 of the total width. The current and preview
+// columns initially get 1/2 of the remaining width each, but if one of them
+// does not have enough widths and another has some spare width, the amount of
+// spare width or the needed width (whichever is smaller) is donated from the
+// latter to the former.
+func getNavWidths(total, currentFull, previewFull int) (int, int, int) {
+	parent := total / 6
+
+	remain := total - parent
+	current := remain / 2
+	preview := remain - current
+
+	if current < currentFull && preview > previewFull {
+		donate := min(currentFull-current, preview-previewFull)
+		current += donate
+		preview -= donate
+	} else if preview < previewFull && current > currentFull {
+		donate := min(previewFull-preview, current-currentFull)
+		preview += donate
+		current -= donate
+	}
+	return parent, current, preview
+}
diff -pruN 0.11+ds1-5/edit/edcore/navigation_widths_test.go 0.12+ds1-1/edit/edcore/navigation_widths_test.go
--- 0.11+ds1-5/edit/edcore/navigation_widths_test.go	1970-01-01 00:00:00.000000000 +0000
+++ 0.12+ds1-1/edit/edcore/navigation_widths_test.go	2018-07-11 02:32:25.000000000 +0000
@@ -0,0 +1,25 @@
+package edcore
+
+import (
+	"testing"
+
+	"github.com/elves/elvish/tt"
+)
+
+var (
+	getNavWidthsTests = tt.Table{
+		// Enough room for both current and preview: parent gets 1/6, current
+		// and preview gets 1/2 of remain
+		tt.Args(120, 10, 10).Rets(20, 50, 50),
+		// Not enough room for either of current and preview: same as above
+		tt.Args(120, 100, 100).Rets(20, 50, 50),
+		// Enough room for current but not preview; current donates to preview
+		tt.Args(120, 10, 100).Rets(20, 10, 90),
+		// Enough room for preview but not current; preview donates to current
+		tt.Args(120, 100, 10).Rets(20, 90, 10),
+	}
+)
+
+func TestGetNavWidths(t *testing.T) {
+	tt.Test(t, tt.Fn("getNavWidths", getNavWidths), getNavWidthsTests)
+}
diff -pruN 0.11+ds1-5/edit/edcore/raw_insert.go 0.12+ds1-1/edit/edcore/raw_insert.go
--- 0.11+ds1-5/edit/edcore/raw_insert.go	1970-01-01 00:00:00.000000000 +0000
+++ 0.12+ds1-1/edit/edcore/raw_insert.go	2018-07-11 02:32:25.000000000 +0000
@@ -0,0 +1,35 @@
+package edcore
+
+import (
+	"github.com/elves/elvish/edit/ui"
+	"github.com/elves/elvish/eval"
+)
+
+// Raw insert mode is a special mode, in that it does not use the normal key
+// binding. Rather, insertRaw is called directly from the main loop in
+// Editor.ReadLine.
+
+type rawInsert struct {
+}
+
+func (ed *editor) startInsertRaw() {
+	ed.reader.SetRaw(true)
+	ed.mode = rawInsert{}
+}
+
+func insertRaw(ed *editor, r rune) {
+	ed.InsertAtDot(string(r))
+	ed.reader.SetRaw(false)
+	ed.SetModeInsert()
+}
+
+func (rawInsert) Teardown() {}
+
+func (rawInsert) Binding(ui.Key) eval.Callable {
+	// The raw insert mode does not handle keys.
+	return nil
+}
+
+func (ri rawInsert) ModeLine() ui.Renderer {
+	return ui.NewModeLineRenderer(" RAW ", "")
+}
diff -pruN 0.11+ds1-5/edit/edcore/registry.go 0.12+ds1-1/edit/edcore/registry.go
--- 0.11+ds1-5/edit/edcore/registry.go	1970-01-01 00:00:00.000000000 +0000
+++ 0.12+ds1-1/edit/edcore/registry.go	2018-07-11 02:32:25.000000000 +0000
@@ -0,0 +1,13 @@
+package edcore
+
+import (
+	"github.com/elves/elvish/eval"
+)
+
+// This file contains utilities that facilitates modularization of the editor.
+
+var editorInitFuncs []func(*editor, eval.Ns)
+
+func atEditorInit(f func(*editor, eval.Ns)) {
+	editorInitFuncs = append(editorInitFuncs, f)
+}
diff -pruN 0.11+ds1-5/edit/edcore/render.go 0.12+ds1-1/edit/edcore/render.go
--- 0.11+ds1-5/edit/edcore/render.go	1970-01-01 00:00:00.000000000 +0000
+++ 0.12+ds1-1/edit/edcore/render.go	2018-07-11 02:32:25.000000000 +0000
@@ -0,0 +1,275 @@
+package edcore
+
+import (
+	"strings"
+	"unicode/utf8"
+
+	"github.com/elves/elvish/edit/highlight"
+	"github.com/elves/elvish/edit/ui"
+	"github.com/elves/elvish/util"
+)
+
+type placeholderRenderer string
+
+func (lp placeholderRenderer) Render(bb *ui.BufferBuilder) {
+	bb.WriteString(util.TrimWcwidth(string(lp), bb.Width), "")
+}
+
+type listingRenderer struct {
+	lines []ui.Styled
+}
+
+func (ls listingRenderer) Render(bb *ui.BufferBuilder) {
+	for i, line := range ls.lines {
+		if i > 0 {
+			bb.Newline()
+		}
+		bb.WriteString(util.ForceWcwidth(line.Text, bb.Width), line.Styles.String())
+	}
+}
+
+type listingWithScrollBarRenderer struct {
+	listingRenderer
+	n, low, high, height int
+}
+
+func (ls listingWithScrollBarRenderer) Render(bb *ui.BufferBuilder) {
+	b1 := ui.Render(ls.listingRenderer, bb.Width-1)
+	bb.ExtendRight(b1, 0)
+
+	scrollbar := renderScrollbar(ls.n, ls.low, ls.high, ls.height)
+	bb.ExtendRight(scrollbar, bb.Width-1)
+}
+
+type navRenderer struct {
+	maxHeight                      int
+	fwParent, fwCurrent, fwPreview int
+	parent, current, preview       ui.Renderer
+}
+
+func makeNavRenderer(h int, w1, w2, w3 int, r1, r2, r3 ui.Renderer) ui.Renderer {
+	return &navRenderer{h, w1, w2, w3, r1, r2, r3}
+}
+
+const navColMargin = 1
+
+func (nr *navRenderer) Render(bb *ui.BufferBuilder) {
+	wParent, wCurrent, wPreview := getNavWidths(bb.Width-navColMargin*2,
+		nr.fwCurrent, nr.fwPreview)
+
+	bParent := ui.Render(nr.parent, wParent)
+	bb.ExtendRight(bParent, 0)
+
+	bCurrent := ui.Render(nr.current, wCurrent)
+	bb.ExtendRight(bCurrent, wParent+navColMargin)
+
+	if wPreview > 0 {
+		bPreview := ui.Render(nr.preview, wPreview)
+		bb.ExtendRight(bPreview, wParent+wCurrent+2*navColMargin)
+	}
+}
+
+// linesRenderer renders lines with a uniform style.
+type linesRenderer struct {
+	lines []string
+	style string
+}
+
+func (nr linesRenderer) Render(bb *ui.BufferBuilder) {
+	bb.WriteString(strings.Join(nr.lines, "\n"), "")
+}
+
+// cmdlineRenderer renders the command line, including the prompt, the user's
+// input and the rprompt.
+type cmdlineRenderer struct {
+	prompt  []*ui.Styled
+	line    string
+	styling *highlight.Styling
+	dot     int
+	rprompt []*ui.Styled
+
+	hasRepl   bool
+	replBegin int
+	replEnd   int
+	replText  string
+}
+
+func newCmdlineRenderer(p []*ui.Styled, l string, s *highlight.Styling, d int, rp []*ui.Styled) *cmdlineRenderer {
+	return &cmdlineRenderer{prompt: p, line: l, styling: s, dot: d, rprompt: rp}
+}
+
+func (clr *cmdlineRenderer) setRepl(bb, e int, t string) {
+	clr.hasRepl = true
+	clr.replBegin, clr.replEnd, clr.replText = bb, e, t
+}
+
+func (clr *cmdlineRenderer) Render(bb *ui.BufferBuilder) {
+	bb.EagerWrap = true
+
+	bb.WriteStyleds(clr.prompt)
+
+	// If the prompt takes less than half of a line, set the indent.
+	if len(bb.Lines) == 1 && bb.Col*2 < bb.Width {
+		bb.Indent = bb.Col
+	}
+
+	// i keeps track of number of bytes written.
+	i := 0
+
+	applier := clr.styling.Apply()
+
+	// nowAt is called at every rune boundary.
+	nowAt := func(i int) {
+		applier.At(i)
+		// Replacement should be written before setting bb.Dot. This way, if the
+		// replacement starts right at the dot, the cursor is correctly placed
+		// after the replacement.
+		if clr.hasRepl && i == clr.replBegin {
+			bb.WriteString(clr.replText, styleForReplacement.String())
+		}
+		if i == clr.dot {
+			bb.Dot = bb.Cursor()
+		}
+	}
+	nowAt(0)
+
+	for _, r := range clr.line {
+		if clr.hasRepl && clr.replBegin <= i && i < clr.replEnd {
+			// Do nothing. This part is replaced by the replacement.
+		} else {
+			bb.Write(r, applier.Get())
+		}
+		i += utf8.RuneLen(r)
+
+		nowAt(i)
+	}
+
+	// Write rprompt
+	if len(clr.rprompt) > 0 {
+		padding := bb.Width - bb.Col
+		for _, s := range clr.rprompt {
+			padding -= util.Wcswidth(s.Text)
+		}
+		if padding >= 1 {
+			bb.EagerWrap = false
+			bb.WriteSpaces(padding, "")
+			bb.WriteStyleds(clr.rprompt)
+		}
+	}
+}
+
+var logEditorRender = false
+
+// editorRenderer renders the entire editor.
+type editorRenderer struct {
+	*editorState
+	height  int
+	bufNoti *ui.Buffer
+}
+
+func (er *editorRenderer) Render(bb *ui.BufferBuilder) {
+	height, width, es := er.height, bb.Width, er.editorState
+
+	var bufNoti, bufLine, bufMode, bufTips, bufListing *ui.Buffer
+	// butNoti
+	if len(es.notifications) > 0 {
+		bufNoti = ui.Render(linesRenderer{es.notifications, ""}, width)
+		es.notifications = nil
+	}
+
+	// bufLine
+	clr := newCmdlineRenderer(es.promptContent, es.buffer, es.styling, es.dot, es.rpromptContent)
+	if repl, ok := es.mode.(replacementer); ok {
+		clr.setRepl(repl.Replacement())
+	}
+	bufLine = ui.Render(clr, width)
+
+	// bufMode
+	bufMode = ui.Render(es.mode.ModeLine(), width)
+
+	// bufTips
+	// TODO tips is assumed to contain no newlines.
+	if len(es.tips) > 0 {
+		bufTips = ui.Render(linesRenderer{es.tips, styleForTip.String()}, width)
+	}
+
+	hListing := 0
+	// Trim lines and determine the maximum height for bufListing
+	// TODO come up with a UI to tell the user that something is not shown.
+	switch {
+	case height >= ui.BuffersHeight(bufNoti, bufLine, bufMode, bufTips):
+		hListing = height - ui.BuffersHeight(bufLine, bufMode, bufTips)
+	case height >= ui.BuffersHeight(bufNoti, bufLine, bufTips):
+		bufMode = nil
+	case height >= ui.BuffersHeight(bufNoti, bufLine):
+		bufMode = nil
+		if bufTips != nil {
+			bufTips.TrimToLines(0, height-ui.BuffersHeight(bufNoti, bufLine))
+		}
+	case height >= ui.BuffersHeight(bufLine):
+		bufTips, bufMode = nil, nil
+		if bufNoti != nil {
+			n := len(bufNoti.Lines)
+			bufNoti.TrimToLines(n-(height-ui.BuffersHeight(bufLine)), n)
+		}
+	case height >= 1:
+		bufNoti, bufTips, bufMode = nil, nil, nil
+		// Determine a window of bufLine that has $height lines around the line
+		// where the dot is currently on.
+		low := bufLine.Dot.Line - height/2
+		high := low + height
+		if low < 0 {
+			low = 0
+			high = low + height
+		} else if high > len(bufLine.Lines) {
+			high = len(bufLine.Lines)
+			low = high - height
+		}
+		bufLine.TrimToLines(low, high)
+	default:
+		// Broken terminal. Still try to render one line of bufLine.
+		bufNoti, bufTips, bufMode = nil, nil, nil
+		dotLine := bufLine.Dot.Line
+		bufLine.TrimToLines(dotLine, dotLine+1)
+	}
+
+	// bufListing.
+	if hListing > 0 {
+		switch mode := es.mode.(type) {
+		case listRenderer:
+			bufListing = mode.ListRender(width, hListing)
+		case lister:
+			bufListing = ui.Render(mode.List(hListing), width)
+		}
+		// XXX When in completion mode, we re-render the mode line, since the
+		// scrollbar in the mode line depends on completion.lastShown which is
+		// only known after the listing has been rendered. Since rendering the
+		// scrollbar never adds additional lines to bufMode, we may do this
+		// without recalculating the layout.
+		if _, ok := es.mode.(redrawModeLiner); ok {
+			bufMode = ui.Render(es.mode.ModeLine(), width)
+		}
+	}
+
+	if logEditorRender {
+		logger.Printf("bufNoti %d, bufLine %d, bufMode %d, bufTips %d, "+
+			"hListing %d, bufListing %d",
+			ui.BuffersHeight(bufNoti), ui.BuffersHeight(bufLine),
+			ui.BuffersHeight(bufMode), ui.BuffersHeight(bufTips),
+			hListing, ui.BuffersHeight(bufListing))
+	}
+
+	// XXX
+	bb.Lines = nil
+	// Combine buffers (reusing bufLine)
+	bb.Extend(bufLine, true)
+	cursorOnModeLine := false
+	if coml, ok := es.mode.(cursorOnModeLiner); ok {
+		cursorOnModeLine = coml.CursorOnModeLine()
+	}
+	bb.Extend(bufMode, cursorOnModeLine)
+	bb.Extend(bufTips, false)
+	bb.Extend(bufListing, false)
+
+	er.bufNoti = bufNoti
+}
diff -pruN 0.11+ds1-5/edit/edcore/rprompt_persistent.go 0.12+ds1-1/edit/edcore/rprompt_persistent.go
--- 0.11+ds1-5/edit/edcore/rprompt_persistent.go	1970-01-01 00:00:00.000000000 +0000
+++ 0.12+ds1-1/edit/edcore/rprompt_persistent.go	2018-07-11 02:32:25.000000000 +0000
@@ -0,0 +1,13 @@
+package edcore
+
+import (
+	"github.com/elves/elvish/eval"
+	"github.com/elves/elvish/eval/vars"
+)
+
+func init() {
+	atEditorInit(func(ed *editor, ns eval.Ns) {
+		ed.RpromptPersistent = false
+		ns["rprompt-persistent"] = vars.FromPtr(&ed.RpromptPersistent)
+	})
+}
diff -pruN 0.11+ds1-5/edit/edcore/styled.go 0.12+ds1-1/edit/edcore/styled.go
--- 0.11+ds1-5/edit/edcore/styled.go	1970-01-01 00:00:00.000000000 +0000
+++ 0.12+ds1-1/edit/edcore/styled.go	2018-07-11 02:32:25.000000000 +0000
@@ -0,0 +1,30 @@
+package edcore
+
+import (
+	"errors"
+
+	"github.com/elves/elvish/edit/ui"
+	"github.com/xiaq/persistent/vector"
+)
+
+var errStyledStyles = errors.New("styles must either be a string or list of strings")
+
+// A constructor for *ui.Styled, for use in Elvish script.
+func styled(text string, styles interface{}) (*ui.Styled, error) {
+	switch styles := styles.(type) {
+	case string:
+		return &ui.Styled{text, ui.StylesFromString(styles)}, nil
+	case vector.Vector:
+		converted := make([]string, 0, styles.Len())
+		for it := styles.Iterator(); it.HasElem(); it.Next() {
+			elem, ok := it.Elem().(string)
+			if !ok {
+				return nil, errStyledStyles
+			}
+			converted = append(converted, elem)
+		}
+		return &ui.Styled{text, ui.Styles(converted)}, nil
+	default:
+		return nil, errStyledStyles
+	}
+}
diff -pruN 0.11+ds1-5/edit/edcore/style.go 0.12+ds1-1/edit/edcore/style.go
--- 0.11+ds1-5/edit/edcore/style.go	1970-01-01 00:00:00.000000000 +0000
+++ 0.12+ds1-1/edit/edcore/style.go	2018-07-11 02:32:25.000000000 +0000
@@ -0,0 +1,16 @@
+package edcore
+
+import (
+	"github.com/elves/elvish/edit/ui"
+)
+
+// Styles for UI.
+var (
+	styleForReplacement    = ui.Styles{"underlined"}
+	styleForTip            = ui.Styles{}
+	styleForSelected       = ui.Styles{"inverse"}
+	styleForScrollBarArea  = ui.Styles{"magenta"}
+	styleForScrollBarThumb = ui.Styles{"magenta", "inverse"}
+
+	styleForCompilerError = ui.Styles{"white", "bg-red"}
+)
diff -pruN 0.11+ds1-5/edit/eddefs/action.go 0.12+ds1-1/edit/eddefs/action.go
--- 0.11+ds1-5/edit/eddefs/action.go	1970-01-01 00:00:00.000000000 +0000
+++ 0.12+ds1-1/edit/eddefs/action.go	2018-07-11 02:32:25.000000000 +0000
@@ -0,0 +1,15 @@
+package eddefs
+
+// Action is used in the SetAction method of the Editor to schedule a special
+// Action after a keybinding has been executed.
+type Action int
+
+const (
+	NoAction Action = iota
+	// ReprocessKey makes the editor to reprocess the keybinding.
+	ReprocessKey
+	// CommitLine makes the editor return with the current line.
+	CommitLine
+	// CommitEOF makes the editor return with an EOF.
+	CommitEOF
+)
diff -pruN 0.11+ds1-5/edit/eddefs/binding_map.go 0.12+ds1-1/edit/eddefs/binding_map.go
--- 0.11+ds1-5/edit/eddefs/binding_map.go	1970-01-01 00:00:00.000000000 +0000
+++ 0.12+ds1-1/edit/eddefs/binding_map.go	2018-07-11 02:32:25.000000000 +0000
@@ -0,0 +1,103 @@
+package eddefs
+
+import (
+	"errors"
+	"sort"
+
+	"github.com/elves/elvish/edit/ui"
+	"github.com/elves/elvish/eval"
+	"github.com/elves/elvish/eval/vals"
+	"github.com/elves/elvish/parse"
+	"github.com/xiaq/persistent/hashmap"
+)
+
+var errValueShouldBeFn = errors.New("value should be function")
+
+// BindingMap is a special Map that converts its key to ui.Key and ensures
+// that its values satisfy eval.CallableValue.
+type BindingMap struct {
+	hashmap.Map
+}
+
+var EmptyBindingMap = BindingMap{vals.EmptyMap}
+
+// Repr returns the representation of the binding table as if it were an
+// ordinary map keyed by strings.
+func (bt BindingMap) Repr(indent int) string {
+	var builder vals.MapReprBuilder
+	builder.Indent = indent
+
+	var keys ui.Keys
+	for it := bt.Map.Iterator(); it.HasElem(); it.Next() {
+		k, _ := it.Elem()
+		keys = append(keys, k.(ui.Key))
+	}
+	sort.Sort(keys)
+
+	for _, k := range keys {
+		v, _ := bt.Map.Index(k)
+		builder.WritePair(parse.Quote(k.String()), indent+2, vals.Repr(v, indent+2))
+	}
+
+	return builder.String()
+}
+
+// Index converts the index to ui.Key and uses the Index of the inner Map.
+func (bt BindingMap) Index(index interface{}) (interface{}, error) {
+	return vals.Index(bt.Map, ui.ToKey(index))
+}
+
+func (bt BindingMap) HasKey(k interface{}) bool {
+	_, ok := bt.Map.Index(k)
+	return ok
+}
+
+func (bt BindingMap) GetKey(k ui.Key) eval.Callable {
+	v, ok := bt.Map.Index(k)
+	if !ok {
+		panic("get called when key not present")
+	}
+	return v.(eval.Callable)
+}
+
+func (bt BindingMap) GetOrDefault(k ui.Key) eval.Callable {
+	switch {
+	case bt.HasKey(k):
+		return bt.GetKey(k)
+	case bt.HasKey(ui.Default):
+		return bt.GetKey(ui.Default)
+	}
+	return nil
+}
+
+// Assoc converts the index to ui.Key, ensures that the value is CallableValue,
+// uses the Assoc of the inner Map and converts the result to a BindingTable.
+func (bt BindingMap) Assoc(k, v interface{}) (interface{}, error) {
+	key := ui.ToKey(k)
+	f, ok := v.(eval.Callable)
+	if !ok {
+		return nil, errValueShouldBeFn
+	}
+	map2 := bt.Map.Assoc(key, f)
+	return BindingMap{map2}, nil
+}
+
+// Dissoc converts the key to ui.Key and calls the Dissoc method of the inner
+// map.
+func (bt BindingMap) Dissoc(k interface{}) interface{} {
+	return BindingMap{bt.Map.Dissoc(ui.ToKey(k))}
+}
+
+func MakeBindingMap(raw hashmap.Map) (BindingMap, error) {
+	converted := vals.EmptyMap
+	for it := raw.Iterator(); it.HasElem(); it.Next() {
+		k, v := it.Elem()
+		f, ok := v.(eval.Callable)
+		if !ok {
+			return EmptyBindingMap, errValueShouldBeFn
+		}
+		converted = converted.Assoc(ui.ToKey(k), f)
+	}
+
+	return BindingMap{converted}, nil
+}
diff -pruN 0.11+ds1-5/edit/eddefs/eddefs.go 0.12+ds1-1/edit/eddefs/eddefs.go
--- 0.11+ds1-5/edit/eddefs/eddefs.go	1970-01-01 00:00:00.000000000 +0000
+++ 0.12+ds1-1/edit/eddefs/eddefs.go	2018-07-11 02:32:25.000000000 +0000
@@ -0,0 +1,3 @@
+// Package eddefs contains types used in the Editor. It is separate package so
+// that editor plugins do not need to depend on the edit package.
+package eddefs
diff -pruN 0.11+ds1-5/edit/eddefs/eddefs_test.go 0.12+ds1-1/edit/eddefs/eddefs_test.go
--- 0.11+ds1-5/edit/eddefs/eddefs_test.go	1970-01-01 00:00:00.000000000 +0000
+++ 0.12+ds1-1/edit/eddefs/eddefs_test.go	2018-07-11 02:32:25.000000000 +0000
@@ -0,0 +1,7 @@
+package eddefs
+
+import "testing"
+
+func Test(t *testing.T) {
+	// TODO: Add tests
+}
diff -pruN 0.11+ds1-5/edit/eddefs/editor.go 0.12+ds1-1/edit/eddefs/editor.go
--- 0.11+ds1-5/edit/eddefs/editor.go	1970-01-01 00:00:00.000000000 +0000
+++ 0.12+ds1-1/edit/eddefs/editor.go	2018-07-11 02:32:25.000000000 +0000
@@ -0,0 +1,66 @@
+package eddefs
+
+import (
+	"github.com/elves/elvish/daemon"
+	"github.com/elves/elvish/edit/ui"
+	"github.com/elves/elvish/eval"
+	"github.com/elves/elvish/parse"
+)
+
+// Editor is the interface for the Elvish line editor.
+type Editor interface {
+	// ReadLine reads a line interactively.
+	ReadLine() (string, error)
+	// Close releases resources used by the editor.
+	Close()
+
+	// Evaler returns the Evaler associated with the Editor.
+	Evaler() *eval.Evaler
+	// Daemon returns the daemon client associated with the Editor.
+	Daemon() *daemon.Client
+
+	// Buffer returns the current content and dot position of the buffer.
+	Buffer() (string, int)
+	// SetBuffer sets the current content and dot position of the buffer.
+	SetBuffer(buffer string, dot int)
+	// ParsedBuffer returns the node from parsing the buffer.
+	ParsedBuffer() *parse.Chunk
+	// InsertAtDot inserts text at the dot and moves the dot after it.
+	InsertAtDot(text string)
+
+	// SetPrompt sets the prompt of the editor.
+	SetPrompt(prompt Prompt)
+	// SetPrompt sets the rprompt of the editor.
+	SetRPrompt(rprompt Prompt)
+
+	// SetMode sets the current mode of the Editor.
+	SetMode(m Mode)
+	// SetModeInsert sets the current mode of the Editor to insert mode.
+	SetModeInsert()
+	// SetModeListing sets the current mode of the Editor to listing mode with
+	// the supplied binding and provider.
+	SetModeListing(b BindingMap, p ListingProvider)
+	// RefreshListing refreshes the listing mode, recalculating the listing
+	// items. It is useful when the underlying listing provider has been
+	// changed. If the editor is not in listing mode, it does nothing.
+	RefreshListing()
+
+	// AddTip adds a message to the tip area.
+	AddTip(format string, args ...interface{})
+	// Notify writes out a message in a way that does not interrupt the editor
+	// display. When the editor is not active, it simply writes the message to
+	// the terminal. When the editor is active, it appends the message to the
+	// notification queue, which will be written out during the update cycle. It
+	// can be safely used concurrently.
+	Notify(format string, args ...interface{})
+
+	// LastKey returns the last key received from the user. It is useful mainly
+	// in keybindings.
+	LastKey() ui.Key
+
+	// SetAction sets the action to execute after the key binding has finished.
+	SetAction(a Action)
+
+	// AddAfterReadline adds a hook function that runs after readline ends.
+	AddAfterReadline(func(string))
+}
diff -pruN 0.11+ds1-5/edit/eddefs/listing.go 0.12+ds1-1/edit/eddefs/listing.go
--- 0.11+ds1-5/edit/eddefs/listing.go	1970-01-01 00:00:00.000000000 +0000
+++ 0.12+ds1-1/edit/eddefs/listing.go	2018-07-11 02:32:25.000000000 +0000
@@ -0,0 +1,13 @@
+package eddefs
+
+import (
+	"github.com/elves/elvish/edit/ui"
+)
+
+type ListingProvider interface {
+	Len() int
+	Show(i int) (string, ui.Styled)
+	Filter(filter string) int
+	Accept(i int, ed Editor)
+	ModeTitle(int) string
+}
diff -pruN 0.11+ds1-5/edit/eddefs/listing_testutils.go 0.12+ds1-1/edit/eddefs/listing_testutils.go
--- 0.11+ds1-5/edit/eddefs/listing_testutils.go	1970-01-01 00:00:00.000000000 +0000
+++ 0.12+ds1-1/edit/eddefs/listing_testutils.go	2018-07-11 02:32:25.000000000 +0000
@@ -0,0 +1,38 @@
+package eddefs
+
+import (
+	"reflect"
+	"testing"
+
+	"github.com/elves/elvish/edit/ui"
+)
+
+type ListingShown struct {
+	Header  string
+	Content ui.Styled
+}
+
+type ListingProviderFilterTest struct {
+	Filter     string
+	WantShowns []ListingShown
+}
+
+func TestListingProviderFilter(t *testing.T, name string, ls ListingProvider, testcases []ListingProviderFilterTest) {
+	for _, testcase := range testcases {
+		ls.Filter(testcase.Filter)
+
+		l := ls.Len()
+		if l != len(testcase.WantShowns) {
+			t.Errorf("%s.Len() -> %d, want %d (filter was %q)",
+				name, l, len(testcase.WantShowns), testcase.Filter)
+		} else {
+			for i, want := range testcase.WantShowns {
+				header, content := ls.Show(i)
+				if header != want.Header || !reflect.DeepEqual(content, want.Content) {
+					t.Errorf("%s.Show(%d) => (%v, %v), want (%v, %v) (filter was %q)",
+						name, i, header, content, want.Header, want.Content, testcase.Filter)
+				}
+			}
+		}
+	}
+}
diff -pruN 0.11+ds1-5/edit/eddefs/mode.go 0.12+ds1-1/edit/eddefs/mode.go
--- 0.11+ds1-5/edit/eddefs/mode.go	1970-01-01 00:00:00.000000000 +0000
+++ 0.12+ds1-1/edit/eddefs/mode.go	2018-07-11 02:32:25.000000000 +0000
@@ -0,0 +1,13 @@
+package eddefs
+
+import (
+	"github.com/elves/elvish/edit/ui"
+	"github.com/elves/elvish/eval"
+)
+
+// Mode is an editor mode.
+type Mode interface {
+	ModeLine() ui.Renderer
+	Binding(ui.Key) eval.Callable
+	Teardown()
+}
diff -pruN 0.11+ds1-5/edit/eddefs/prompt.go 0.12+ds1-1/edit/eddefs/prompt.go
--- 0.11+ds1-5/edit/eddefs/prompt.go	1970-01-01 00:00:00.000000000 +0000
+++ 0.12+ds1-1/edit/eddefs/prompt.go	2018-07-11 02:32:25.000000000 +0000
@@ -0,0 +1,17 @@
+package eddefs
+
+import "github.com/elves/elvish/edit/ui"
+
+// Prompt is the interface for a general prompt.
+type Prompt interface {
+	// Chan returns a prompt on which the content of the prompt is made
+	// available.
+	Chan() <-chan []*ui.Styled
+	// Update signifies that the prompt should be updated.
+	Update(force bool)
+	// Last returns the last content written to Chan. If Chan was never written,
+	// it should return some content representing an unknown prompt.
+	Last() []*ui.Styled
+	// Close releases resources associated with the prompt.
+	Close() error
+}
diff -pruN 0.11+ds1-5/edit/edit.go 0.12+ds1-1/edit/edit.go
--- 0.11+ds1-5/edit/edit.go	2018-01-18 02:03:11.000000000 +0000
+++ 0.12+ds1-1/edit/edit.go	2018-07-11 02:32:25.000000000 +0000
@@ -1,491 +1,16 @@
-// Package edit implements a command line editor.
+// Package edit implements the Elvish command editor.
 package edit
 
 import (
-	"bufio"
-	"bytes"
-	"fmt"
-	"io"
 	"os"
-	"sync"
-	"syscall"
-	"time"
 
-	"github.com/elves/elvish/daemon"
-	"github.com/elves/elvish/edit/highlight"
-	"github.com/elves/elvish/edit/history"
-	"github.com/elves/elvish/edit/prompt"
-	"github.com/elves/elvish/edit/tty"
-	"github.com/elves/elvish/edit/ui"
+	"github.com/elves/elvish/edit/edcore"
+	"github.com/elves/elvish/edit/eddefs"
 	"github.com/elves/elvish/eval"
-	"github.com/elves/elvish/eval/types"
-	"github.com/elves/elvish/eval/vartypes"
-	"github.com/elves/elvish/parse"
-	"github.com/elves/elvish/sys"
-	"github.com/elves/elvish/util"
 )
 
-var logger = util.GetLogger("[edit] ")
-
-// Editor keeps the status of the line editor.
-type Editor struct {
-	in     *os.File
-	out    *os.File
-	writer tty.Writer
-	reader tty.Reader
-	sigs   chan os.Signal
-	daemon *daemon.Client
-	evaler *eval.Evaler
-
-	variables map[string]vartypes.Variable
-	bindings  map[string]vartypes.Variable
-
-	active      bool
-	activeMutex sync.Mutex
-
-	historyFuser *history.Fuser
-	historyMutex sync.RWMutex
-
-	// notifyPort is a write-only port that turns data written to it into editor
-	// notifications.
-	notifyPort *eval.Port
-	// notifyRead is the read end of notifyPort.File.
-	notifyRead *os.File
-
-	editorState
-}
-
-type editorState struct {
-	// States used during ReadLine. Reset at the beginning of ReadLine.
-	restoreTerminal func() error
-
-	notificationMutex sync.Mutex
-
-	notifications []string
-	tips          []string
-
-	buffer string
-	dot    int
-
-	chunk           *parse.Chunk
-	styling         *highlight.Styling
-	parseErrorAtEnd bool
-
-	promptContent  []*ui.Styled
-	rpromptContent []*ui.Styled
-
-	mode Mode
-
-	insert     insert
-	command    command
-	completion completion
-	navigation navigation
-
-	// A cache of external commands, used in stylist.
-	isExternal map[string]bool
-
-	// Used for builtins.
-	lastKey    ui.Key
-	nextAction action
-}
-
 // NewEditor creates an Editor. When the instance is no longer used, its Close
 // method should be called.
-func NewEditor(in *os.File, out *os.File, sigs chan os.Signal, ev *eval.Evaler) *Editor {
-	daemon := ev.DaemonClient
-
-	ed := &Editor{
-		in:     in,
-		out:    out,
-		writer: tty.NewWriter(out),
-		reader: tty.NewReader(in),
-		sigs:   sigs,
-		daemon: daemon,
-		evaler: ev,
-
-		bindings:  makeBindings(),
-		variables: makeVariables(),
-	}
-
-	notifyChan := make(chan types.Value)
-	notifyRead, notifyWrite, err := os.Pipe()
-	if err != nil {
-		panic(err)
-	}
-	ed.notifyPort = &eval.Port{File: notifyWrite, Chan: notifyChan}
-	ed.notifyRead = notifyRead
-	// Forward reads from notifyRead to notification.
-	go func() {
-		reader := bufio.NewReader(notifyRead)
-		for {
-			line, err := reader.ReadString('\n')
-			if err != nil {
-				break
-			}
-			ed.Notify("[bytes out] %s", line[:len(line)-1])
-		}
-		if err != io.EOF {
-			logger.Println("notifyRead error:", err)
-		}
-	}()
-	// Forward reads from notifyChan to notification.
-	go func() {
-		for v := range notifyChan {
-			ed.Notify("[value out] %s", v.Repr(types.NoPretty))
-		}
-	}()
-
-	if daemon != nil {
-		f, err := history.NewFuser(daemon)
-		if err != nil {
-			fmt.Fprintln(os.Stderr, "Failed to initialize command history. Disabled.")
-		} else {
-			ed.historyFuser = f
-		}
-	}
-	ev.Editor = ed
-
-	installModules(ev.Builtin, ed)
-	err = ev.SourceText(eval.NewScriptSource("[editor]", "[editor]", "use binding; binding:install"))
-	if err != nil {
-		fmt.Fprintln(out, "Failed to load default binding:", err)
-	}
-
-	return ed
-}
-
-// Close releases resources used by the editor.
-func (ed *Editor) Close() {
-	ed.reader.Close()
-	close(ed.notifyPort.Chan)
-	ed.notifyPort.File.Close()
-	ed.notifyRead.Close()
-}
-
-// Active returns the activeness of the Editor.
-func (ed *Editor) Active() bool {
-	return ed.active
-}
-
-// ActiveMutex returns a mutex that must be used when changing the activeness of
-// the Editor.
-func (ed *Editor) ActiveMutex() *sync.Mutex {
-	return &ed.activeMutex
-}
-
-func (ed *Editor) Evaler() *eval.Evaler {
-	return ed.evaler
-}
-
-func (ed *Editor) Variable(name string) vartypes.Variable {
-	return ed.variables[name]
-}
-
-func (ed *Editor) flash() {
-	// TODO implement fish-like flash effect
-}
-
-func (ed *Editor) addTip(format string, args ...interface{}) {
-	ed.tips = append(ed.tips, fmt.Sprintf(format, args...))
-}
-
-// Notify adds one notification entry. It is concurrency-safe.
-func (ed *Editor) Notify(format string, args ...interface{}) {
-	ed.notificationMutex.Lock()
-	defer ed.notificationMutex.Unlock()
-	ed.notifications = append(ed.notifications, fmt.Sprintf(format, args...))
-}
-
-func (ed *Editor) refresh(fullRefresh bool, addErrorsToTips bool) error {
-	src := ed.buffer
-	// Parse the current line
-	n, err := parse.Parse("[interactive]", src)
-	ed.chunk = n
-
-	ed.parseErrorAtEnd = err != nil && atEnd(err, len(src))
-	// If all parse errors are at the end, it is likely caused by incomplete
-	// input. In that case, do not complain about parse errors.
-	// TODO(xiaq): Find a more reliable way to determine incomplete input.
-	// Ideally the parser should report it.
-	if err != nil && addErrorsToTips && !ed.parseErrorAtEnd {
-		ed.addTip("%s", err)
-	}
-
-	ed.styling = &highlight.Styling{}
-	doHighlight(n, ed)
-
-	_, err = ed.evaler.Compile(n, eval.NewInteractiveSource(src))
-	if err != nil && !atEnd(err, len(src)) {
-		if addErrorsToTips {
-			ed.addTip("%s", err)
-		}
-		// Highlight errors in the input buffer.
-		ctx := err.(*eval.CompilationError).Context
-		ed.styling.Add(ctx.Begin, ctx.End, styleForCompilerError.String())
-	}
-
-	// Render onto a buffer.
-	height, width := sys.GetWinsize(ed.out)
-	height = min(height, ed.maxHeight())
-	er := &editorRenderer{&ed.editorState, height, nil}
-	buf := ui.Render(er, width)
-	return ed.writer.CommitBuffer(er.bufNoti, buf, fullRefresh)
-}
-
-func atEnd(e error, n int) bool {
-	switch e := e.(type) {
-	case *eval.CompilationError:
-		return e.Context.Begin == n
-	case *parse.Error:
-		for _, entry := range e.Entries {
-			if entry.Context.Begin != n {
-				return false
-			}
-		}
-		return true
-	default:
-		logger.Printf("atEnd called with error type %T", e)
-		return false
-	}
-}
-
-// insertAtDot inserts text at the dot and moves the dot after it.
-func (ed *Editor) insertAtDot(text string) {
-	ed.buffer = ed.buffer[:ed.dot] + text + ed.buffer[ed.dot:]
-	ed.dot += len(text)
-}
-
-// startReadLine prepares the terminal for the editor.
-func (ed *Editor) startReadLine() error {
-	ed.activeMutex.Lock()
-	defer ed.activeMutex.Unlock()
-	ed.active = true
-
-	restoreTerminal, err := tty.Setup(ed.in, ed.out)
-	if err != nil {
-		if restoreTerminal != nil {
-			restoreTerminal()
-		}
-		return err
-	}
-	ed.restoreTerminal = restoreTerminal
-
-	return nil
-}
-
-// finishReadLine puts the terminal in a state suitable for other programs to
-// use.
-func (ed *Editor) finishReadLine() error {
-	ed.activeMutex.Lock()
-	defer ed.activeMutex.Unlock()
-	ed.active = false
-
-	// Refresh the terminal for the last time in a clean-ish state.
-	ed.mode = &ed.insert
-	ed.tips = nil
-	ed.dot = len(ed.buffer)
-	if !prompt.RpromptPersistent(ed) {
-		ed.rpromptContent = nil
-	}
-	errRefresh := ed.refresh(false, false)
-	ed.out.WriteString("\n")
-	ed.writer.ResetCurrentBuffer()
-
-	ed.reader.Stop()
-
-	// Restore termios.
-	errRestore := ed.restoreTerminal()
-
-	// Save the line before resetting all of editorState.
-	line := ed.buffer
-	ed.editorState = editorState{}
-
-	callHooks(ed.evaler, ed.afterReadLine(), types.String(line))
-
-	return util.Errors(errRefresh, errRestore)
-}
-
-// ReadLine reads a line interactively.
-func (ed *Editor) ReadLine() (string, error) {
-	err := ed.startReadLine()
-	if err != nil {
-		return "", err
-	}
-	defer func() {
-		err := ed.finishReadLine()
-		if err != nil {
-			fmt.Fprintln(ed.out, "error:", err)
-		}
-	}()
-
-	ed.mode = &ed.insert
-
-	// Find external commands asynchronously, so that slow I/O won't block the
-	// editor.
-	isExternalCh := make(chan map[string]bool, 1)
-	go getIsExternal(ed.evaler, isExternalCh)
-
-	ed.reader.Start()
-
-	fullRefresh := false
-
-	callHooks(ed.evaler, ed.beforeReadLine())
-
-	promptUpdater := prompt.NewUpdater(prompt.Prompt)
-	rpromptUpdater := prompt.NewUpdater(prompt.Rprompt)
-
-MainLoop:
-	for {
-		promptCh := promptUpdater.Update(ed)
-		rpromptCh := rpromptUpdater.Update(ed)
-		promptTimeout := prompt.MakeMaxWaitChan(ed)
-		rpromptTimeout := prompt.MakeMaxWaitChan(ed)
-
-		select {
-		case ed.promptContent = <-promptCh:
-			logger.Println("prompt fetched")
-		case <-promptTimeout:
-			logger.Println("stale prompt")
-			ed.promptContent = promptUpdater.Staled
-		}
-		select {
-		case ed.rpromptContent = <-rpromptCh:
-			logger.Println("rprompt fetched")
-		case <-rpromptTimeout:
-			logger.Println("stale rprompt")
-			ed.rpromptContent = rpromptUpdater.Staled
-		}
-
-	refresh:
-		err := ed.refresh(fullRefresh, true)
-		fullRefresh = false
-		if err != nil {
-			return "", err
-		}
-
-		ed.tips = nil
-
-		select {
-		case ed.promptContent = <-promptCh:
-			logger.Println("prompt fetched late")
-			goto refresh
-		case ed.rpromptContent = <-rpromptCh:
-			logger.Println("rprompt fetched late")
-			goto refresh
-		case m := <-isExternalCh:
-			ed.isExternal = m
-		case sig := <-ed.sigs:
-			// TODO(xiaq): Maybe support customizable handling of signals
-			switch sig {
-			case syscall.SIGHUP:
-				return "", io.EOF
-			case syscall.SIGINT:
-				// Start over
-				ed.editorState = editorState{
-					restoreTerminal: ed.restoreTerminal,
-					isExternal:      ed.isExternal,
-				}
-				ed.mode = &ed.insert
-				continue MainLoop
-			case sys.SIGWINCH:
-				fullRefresh = true
-				continue MainLoop
-			default:
-				ed.addTip("ignored signal %s", sig)
-			}
-		case event := <-ed.reader.EventChan():
-			switch event := event.(type) {
-			case tty.NonfatalErrorEvent:
-				ed.Notify("error when reading terminal: %v", event.Err)
-			case tty.FatalErrorEvent:
-				ed.Notify("fatal error when reading terminal: %v", event.Err)
-				return "", event.Err
-			case tty.MouseEvent:
-				ed.addTip("mouse: %+v", event)
-			case tty.CursorPosition:
-				// Ignore CPR
-			case tty.PasteSetting:
-				if !event {
-					continue
-				}
-				var buf bytes.Buffer
-				timer := time.NewTimer(tty.DefaultSeqTimeout)
-			paste:
-				for {
-					// XXX Should also select on other chans. However those chans
-					// will be unified (again) into one later so we don't do
-					// busywork here.
-					select {
-					case event := <-ed.reader.EventChan():
-						switch event := event.(type) {
-						case tty.KeyEvent:
-							k := ui.Key(event)
-							if k.Mod != 0 {
-								ed.Notify("function key within paste, aborting")
-								break paste
-							}
-							buf.WriteRune(k.Rune)
-							timer.Reset(tty.DefaultSeqTimeout)
-						case tty.PasteSetting:
-							if !event {
-								break paste
-							}
-						default: // Ignore other things.
-						}
-					case <-timer.C:
-						ed.Notify("bracketed paste timeout")
-						break paste
-					}
-				}
-				topaste := buf.String()
-				if ed.insert.quotePaste {
-					topaste = parse.Quote(topaste)
-				}
-				ed.insertAtDot(topaste)
-			case tty.RawRune:
-				insertRaw(ed, rune(event))
-			case tty.KeyEvent:
-				k := ui.Key(event)
-			lookupKey:
-				fn := ed.mode.Binding(ed.bindings, k)
-				if fn == nil {
-					ed.addTip("Unbound and no default binding: %s", k)
-					continue MainLoop
-				}
-
-				ed.insert.insertedLiteral = false
-				ed.lastKey = k
-				ed.CallFn(fn)
-				if ed.insert.insertedLiteral {
-					ed.insert.literalInserts++
-				} else {
-					ed.insert.literalInserts = 0
-				}
-
-				switch ed.popAction() {
-				case reprocessKey:
-					err := ed.refresh(false, true)
-					if err != nil {
-						return "", err
-					}
-					goto lookupKey
-				case commitLine:
-					ed.appendHistory(ed.buffer)
-					return ed.buffer, nil
-				case commitEOF:
-					return "", io.EOF
-				}
-			}
-		}
-	}
-}
-
-// getIsExternal finds a set of all external commands and puts it on the result
-// channel.
-func getIsExternal(ev *eval.Evaler, result chan<- map[string]bool) {
-	isExternal := make(map[string]bool)
-	eval.EachExternal(func(name string) {
-		isExternal[name] = true
-	})
-	result <- isExternal
+func NewEditor(in *os.File, out *os.File, sigs <-chan os.Signal, ev *eval.Evaler) eddefs.Editor {
+	return edcore.NewEditor(in, out, sigs, ev)
 }
diff -pruN 0.11+ds1-5/edit/editor_unix_test.go 0.12+ds1-1/edit/editor_unix_test.go
--- 0.11+ds1-5/edit/editor_unix_test.go	2018-01-18 02:03:11.000000000 +0000
+++ 0.12+ds1-1/edit/editor_unix_test.go	1970-01-01 00:00:00.000000000 +0000
@@ -1,86 +0,0 @@
-// +build !windows,!plan9
-
-package edit
-
-import (
-	"testing"
-	"time"
-
-	"github.com/elves/elvish/eval"
-	"github.com/elves/elvish/sys"
-	"github.com/kr/pty"
-)
-
-var readLineTests = []struct {
-	input string
-	want  string
-}{
-	{"\n", ""},
-	{"test\n", "test"},
-	{"abc\x7fd\n", "abd"},
-	{"abc\x17d\n", "d"},
-}
-
-var readLineTimeout = 5 * time.Second
-
-func TestReadLine(t *testing.T) {
-	master, tty, err := pty.Open()
-	if err != nil {
-		panic(err)
-	}
-	defer master.Close()
-	defer tty.Close()
-
-	// Continually consume tty outputs so that the editor is not blocked on
-	// writing.
-	var outputs []byte
-	go func() {
-		var buf [256]byte
-		for {
-			nr, err := master.Read(buf[:])
-			if err != nil {
-				break
-			}
-			outputs = append(outputs, buf[:nr]...)
-		}
-	}()
-
-	ev := eval.NewEvaler()
-	// XXX: Needed for "use" to work.
-	ev.SetLibDir("/non/exist/ent")
-	defer ev.Close()
-
-	for _, test := range readLineTests {
-		lineChan := make(chan string)
-		errChan := make(chan error)
-		go func() {
-			ed := NewEditor(tty, tty, nil, ev)
-			defer ed.Close()
-			line, err := ed.ReadLine()
-			if err != nil {
-				errChan <- err
-			} else {
-				lineChan <- line
-			}
-		}()
-
-		_, err := master.WriteString(test.input)
-		if err != nil {
-			panic(err)
-		}
-
-		select {
-		case line := <-lineChan:
-			if line != test.want {
-				t.Errorf("ReadLine() => %q, want %q (input %q)", line, test.want, test.input)
-			}
-		case err := <-errChan:
-			t.Errorf("ReadLine() => error %v (input %q)", err, test.input)
-		case <-time.After(readLineTimeout):
-			t.Errorf("ReadLine() timed out (input %q)", test.input)
-			t.Log("Stack trace: \n" + sys.DumpStack())
-			t.Logf("Terminal output: %q", outputs)
-			t.FailNow()
-		}
-	}
-}
diff -pruN 0.11+ds1-5/edit/highlight.go 0.12+ds1-1/edit/highlight.go
--- 0.11+ds1-5/edit/highlight.go	2018-01-18 02:03:11.000000000 +0000
+++ 0.12+ds1-1/edit/highlight.go	1970-01-01 00:00:00.000000000 +0000
@@ -1,56 +0,0 @@
-package edit
-
-import (
-	"os"
-
-	"github.com/elves/elvish/edit/highlight"
-	"github.com/elves/elvish/eval"
-	"github.com/elves/elvish/parse"
-	"github.com/elves/elvish/util"
-)
-
-func doHighlight(n parse.Node, ed *Editor) {
-	s := &highlight.Emitter{
-		func(s string) bool { return goodFormHead(s, ed) },
-		ed.styling.Add,
-	}
-	s.EmitAll(n)
-}
-
-func goodFormHead(head string, ed *Editor) bool {
-	if eval.IsBuiltinSpecial[head] {
-		return true
-	} else if util.DontSearch(head) {
-		// XXX don't stat twice
-		return util.IsExecutable(head) || isDir(head)
-	} else {
-		ev := ed.evaler
-		explode, ns, name := eval.ParseVariable(head)
-		if !explode {
-			switch ns {
-			case "":
-				if ev.Builtin[name+eval.FnSuffix] != nil || ev.Global[name+eval.FnSuffix] != nil {
-					return true
-				}
-			case "e":
-				if ed.isExternal[name] {
-					return true
-				}
-			default:
-				mod := ev.Global[ns+eval.NsSuffix]
-				if mod == nil {
-					mod = ev.Builtin[ns+eval.NsSuffix]
-				}
-				if mod != nil && mod.Get().(eval.Ns)[name+eval.FnSuffix] != nil {
-					return true
-				}
-			}
-		}
-		return ed.isExternal[head]
-	}
-}
-
-func isDir(fname string) bool {
-	stat, err := os.Stat(fname)
-	return err == nil && stat.IsDir()
-}
diff -pruN 0.11+ds1-5/edit/histlist.go 0.12+ds1-1/edit/histlist.go
--- 0.11+ds1-5/edit/histlist.go	2018-01-18 02:03:11.000000000 +0000
+++ 0.12+ds1-1/edit/histlist.go	1970-01-01 00:00:00.000000000 +0000
@@ -1,142 +0,0 @@
-package edit
-
-import (
-	"errors"
-	"fmt"
-	"strconv"
-	"strings"
-
-	"github.com/elves/elvish/edit/ui"
-)
-
-// Command history listing mode.
-
-var _ = registerBuiltins(modeHistoryListing, map[string]func(*Editor){
-	"start":                   histlistStart,
-	"toggle-dedup":            histlistToggleDedup,
-	"toggle-case-sensitivity": histlistToggleCaseSensitivity,
-})
-
-// ErrStoreOffline is thrown when an operation requires the storage backend, but
-// it is offline.
-var ErrStoreOffline = errors.New("store offline")
-
-type histlist struct {
-	all             []string
-	dedup           bool
-	caseInsensitive bool
-	last            map[string]int
-	shown           []string
-	index           []int
-	indexWidth      int
-}
-
-func newHistlist(cmds []string) *listing {
-	last := make(map[string]int)
-	for i, entry := range cmds {
-		last[entry] = i
-	}
-	hl := &histlist{
-		// This has to be here for the initialization to work :(
-		all:        cmds,
-		dedup:      true,
-		last:       last,
-		indexWidth: len(strconv.Itoa(len(cmds) - 1)),
-	}
-	l := newListing(modeHistoryListing, hl)
-	return &l
-}
-
-func (hl *histlist) ModeTitle(i int) string {
-	s := " HISTORY "
-	if hl.dedup {
-		s += "(dedup on) "
-	}
-	if hl.caseInsensitive {
-		s += "(case-insensitive) "
-	}
-	return s
-}
-
-func (*histlist) CursorOnModeLine() bool {
-	return true
-}
-
-func (hl *histlist) Len() int {
-	return len(hl.shown)
-}
-
-func (hl *histlist) Show(i int) (string, ui.Styled) {
-	return fmt.Sprintf("%d", hl.index[i]), ui.Unstyled(hl.shown[i])
-}
-
-func (hl *histlist) Filter(filter string) int {
-	hl.shown = nil
-	hl.index = nil
-	dedup := hl.dedup
-	if hl.caseInsensitive {
-		filter = strings.ToLower(filter)
-	}
-	for i, entry := range hl.all {
-		fentry := entry
-		if hl.caseInsensitive {
-			fentry = strings.ToLower(entry)
-		}
-		if (!dedup || hl.last[entry] == i) && strings.Contains(fentry, filter) {
-			hl.index = append(hl.index, i)
-			hl.shown = append(hl.shown, entry)
-		}
-	}
-	// TODO: Maintain old selection
-	return len(hl.shown) - 1
-}
-
-// Editor interface.
-
-func (hl *histlist) Accept(i int, ed *Editor) {
-	line := hl.shown[i]
-	if len(ed.buffer) > 0 {
-		line = "\n" + line
-	}
-	ed.insertAtDot(line)
-}
-
-func histlistStart(ed *Editor) {
-	cmds, err := getCmds(ed)
-	if err != nil {
-		ed.Notify("%v", err)
-		return
-	}
-
-	ed.mode = newHistlist(cmds)
-}
-
-func getCmds(ed *Editor) ([]string, error) {
-	if ed.daemon == nil {
-		return nil, ErrStoreOffline
-	}
-	return ed.historyFuser.AllCmds()
-}
-
-func histlistToggleDedup(ed *Editor) {
-	if l, hl, ok := getHistlist(ed); ok {
-		hl.dedup = !hl.dedup
-		l.refresh()
-	}
-}
-
-func histlistToggleCaseSensitivity(ed *Editor) {
-	if l, hl, ok := getHistlist(ed); ok {
-		hl.caseInsensitive = !hl.caseInsensitive
-		l.refresh()
-	}
-}
-
-func getHistlist(ed *Editor) (*listing, *histlist, bool) {
-	if l, ok := ed.mode.(*listing); ok {
-		if hl, ok := l.provider.(*histlist); ok {
-			return l, hl, true
-		}
-	}
-	return nil, nil, false
-}
diff -pruN 0.11+ds1-5/edit/histlist_test.go 0.12+ds1-1/edit/histlist_test.go
--- 0.11+ds1-5/edit/histlist_test.go	2018-01-18 02:03:11.000000000 +0000
+++ 0.12+ds1-1/edit/histlist_test.go	1970-01-01 00:00:00.000000000 +0000
@@ -1,37 +0,0 @@
-package edit
-
-import (
-	"testing"
-
-	"github.com/elves/elvish/edit/ui"
-)
-
-var (
-	theHistList = newHistlist([]string{"ls", "echo lalala", "ls"})
-
-	histlistDedupFilterTests = []listingFilterTestCases{
-		{"", []shown{
-			{"1", ui.Unstyled("echo lalala")},
-			{"2", ui.Unstyled("ls")}}},
-		{"l", []shown{
-			{"1", ui.Unstyled("echo lalala")},
-			{"2", ui.Unstyled("ls")}}},
-	}
-
-	histlistNoDedupFilterTests = []listingFilterTestCases{
-		{"", []shown{
-			{"0", ui.Unstyled("ls")},
-			{"1", ui.Unstyled("echo lalala")},
-			{"2", ui.Unstyled("ls")}}},
-		{"l", []shown{
-			{"0", ui.Unstyled("ls")},
-			{"1", ui.Unstyled("echo lalala")},
-			{"2", ui.Unstyled("ls")}}},
-	}
-)
-
-func TestHistlist(t *testing.T) {
-	testListingFilter(t, "theHistList", theHistList, histlistDedupFilterTests)
-	theHistList.provider.(*histlist).dedup = false
-	testListingFilter(t, "theHistList", theHistList, histlistNoDedupFilterTests)
-}
diff -pruN 0.11+ds1-5/edit/history/fuser.go 0.12+ds1-1/edit/history/fuser.go
--- 0.11+ds1-5/edit/history/fuser.go	2018-01-18 02:03:11.000000000 +0000
+++ 0.12+ds1-1/edit/history/fuser.go	2018-07-11 02:32:25.000000000 +0000
@@ -29,6 +29,20 @@ func NewFuser(store Store) (*Fuser, erro
 	}, nil
 }
 
+func (f *Fuser) FastForward() error {
+	f.Lock()
+	defer f.Unlock()
+
+	upper, err := f.store.NextCmdSeq()
+	if err != nil {
+		return err
+	}
+	f.storeUpper = upper
+	f.cmds = nil
+	f.seqs = nil
+	return nil
+}
+
 func (f *Fuser) AddCmd(cmd string) error {
 	f.Lock()
 	defer f.Unlock()
diff -pruN 0.11+ds1-5/edit/history/histlist.go 0.12+ds1-1/edit/history/histlist.go
--- 0.11+ds1-5/edit/history/histlist.go	1970-01-01 00:00:00.000000000 +0000
+++ 0.12+ds1-1/edit/history/histlist.go	2018-07-11 02:32:25.000000000 +0000
@@ -0,0 +1,129 @@
+package history
+
+import (
+	"errors"
+	"fmt"
+	"strconv"
+	"strings"
+
+	"github.com/elves/elvish/edit/eddefs"
+	"github.com/elves/elvish/edit/ui"
+)
+
+// Command history listing mode.
+
+// errStoreOffline is thrown when an operation requires the storage backend, but
+// it is offline.
+var errStoreOffline = errors.New("store offline")
+
+type histlist struct {
+	all             []string
+	dedup           bool
+	caseInsensitive bool
+	last            map[string]int
+	shown           []string
+	index           []int
+	indexWidth      int
+}
+
+func newHistlist(cmds []string) *histlist {
+	last := make(map[string]int)
+	for i, entry := range cmds {
+		last[entry] = i
+	}
+	return &histlist{
+		// This has to be here for the initialization to work :(
+		all:        cmds,
+		dedup:      true,
+		last:       last,
+		indexWidth: len(strconv.Itoa(len(cmds) - 1)),
+	}
+}
+
+func (hl *histlist) Teardown() {
+	*hl = histlist{}
+}
+
+func (hl *histlist) ModeTitle(i int) string {
+	s := " HISTORY "
+	if hl.dedup {
+		s += "(dedup on) "
+	}
+	if hl.caseInsensitive {
+		s += "(case-insensitive) "
+	}
+	return s
+}
+
+func (*histlist) CursorOnModeLine() bool {
+	return true
+}
+
+func (hl *histlist) Len() int {
+	return len(hl.shown)
+}
+
+func (hl *histlist) Show(i int) (string, ui.Styled) {
+	return fmt.Sprintf("%d", hl.index[i]+1), ui.Unstyled(hl.shown[i])
+}
+
+func (hl *histlist) Filter(filter string) int {
+	hl.shown = nil
+	hl.index = nil
+	dedup := hl.dedup
+	if hl.caseInsensitive {
+		filter = strings.ToLower(filter)
+	}
+	for i, entry := range hl.all {
+		fentry := entry
+		if hl.caseInsensitive {
+			fentry = strings.ToLower(entry)
+		}
+		if (!dedup || hl.last[entry] == i) && strings.Contains(fentry, filter) {
+			hl.index = append(hl.index, i)
+			hl.shown = append(hl.shown, entry)
+		}
+	}
+	// TODO: Maintain old selection
+	return len(hl.shown) - 1
+}
+
+// Editor interface.
+
+func (hl *histlist) Accept(i int, ed eddefs.Editor) {
+	line := hl.shown[i]
+	buffer, _ := ed.Buffer()
+	if len(buffer) > 0 {
+		line = "\n" + line
+	}
+	ed.InsertAtDot(line)
+}
+
+func (hl *histlist) start(ed eddefs.Editor, fuser *Fuser, binding eddefs.BindingMap) {
+	cmds, err := getCmds(fuser)
+	if err != nil {
+		ed.Notify("%v", err)
+		return
+	}
+
+	*hl = *newHistlist(cmds)
+
+	ed.SetModeListing(binding, hl)
+}
+
+func getCmds(fuser *Fuser) ([]string, error) {
+	if fuser == nil {
+		return nil, errStoreOffline
+	}
+	return fuser.AllCmds()
+}
+
+func (hl *histlist) toggleDedup(ed eddefs.Editor) {
+	hl.dedup = !hl.dedup
+	ed.RefreshListing()
+}
+
+func (hl *histlist) toggleCaseSensitivity(ed eddefs.Editor) {
+	hl.caseInsensitive = !hl.caseInsensitive
+	ed.RefreshListing()
+}
diff -pruN 0.11+ds1-5/edit/history/histlist_test.go 0.12+ds1-1/edit/history/histlist_test.go
--- 0.11+ds1-5/edit/history/histlist_test.go	1970-01-01 00:00:00.000000000 +0000
+++ 0.12+ds1-1/edit/history/histlist_test.go	2018-07-11 02:32:25.000000000 +0000
@@ -0,0 +1,38 @@
+package history
+
+import (
+	"testing"
+
+	"github.com/elves/elvish/edit/eddefs"
+	"github.com/elves/elvish/edit/ui"
+)
+
+var (
+	theHistList = newHistlist([]string{"ls", "echo lalala", "ls"})
+
+	histlistDedupFilterTests = []eddefs.ListingProviderFilterTest{
+		{"", []eddefs.ListingShown{
+			{"2", ui.Unstyled("echo lalala")},
+			{"3", ui.Unstyled("ls")}}},
+		{"l", []eddefs.ListingShown{
+			{"2", ui.Unstyled("echo lalala")},
+			{"3", ui.Unstyled("ls")}}},
+	}
+
+	histlistNoDedupFilterTests = []eddefs.ListingProviderFilterTest{
+		{"", []eddefs.ListingShown{
+			{"1", ui.Unstyled("ls")},
+			{"2", ui.Unstyled("echo lalala")},
+			{"3", ui.Unstyled("ls")}}},
+		{"l", []eddefs.ListingShown{
+			{"1", ui.Unstyled("ls")},
+			{"2", ui.Unstyled("echo lalala")},
+			{"3", ui.Unstyled("ls")}}},
+	}
+)
+
+func TestHistlist(t *testing.T) {
+	eddefs.TestListingProviderFilter(t, "theHistList", theHistList, histlistDedupFilterTests)
+	theHistList.dedup = false
+	eddefs.TestListingProviderFilter(t, "theHistList", theHistList, histlistNoDedupFilterTests)
+}
diff -pruN 0.11+ds1-5/edit/history/history.go 0.12+ds1-1/edit/history/history.go
--- 0.11+ds1-5/edit/history/history.go	2018-01-18 02:03:11.000000000 +0000
+++ 0.12+ds1-1/edit/history/history.go	2018-07-11 02:32:25.000000000 +0000
@@ -1,2 +1,204 @@
-// Package history provides utilities for the command history.
 package history
+
+import (
+	"fmt"
+	"os"
+	"strconv"
+	"strings"
+	"sync"
+
+	"github.com/elves/elvish/edit/eddefs"
+	"github.com/elves/elvish/edit/ui"
+	"github.com/elves/elvish/eval"
+	"github.com/elves/elvish/eval/vals"
+	"github.com/elves/elvish/eval/vars"
+	"github.com/elves/elvish/util"
+)
+
+var logger = util.GetLogger("[edit/history] ")
+
+// Command history mode.
+
+type hist struct {
+	ed      eddefs.Editor
+	mutex   sync.RWMutex
+	fuser   *Fuser
+	binding eddefs.BindingMap
+
+	// Non-persistent state.
+	walker    *Walker
+	bufferLen int
+}
+
+func Init(ed eddefs.Editor, ns eval.Ns) {
+	hist := &hist{ed: ed, binding: eddefs.EmptyBindingMap}
+	if ed.Daemon() != nil {
+		fuser, err := NewFuser(ed.Daemon())
+		if err != nil {
+			fmt.Fprintln(os.Stderr, "Failed to initialize command history; disabled.")
+		} else {
+			hist.fuser = fuser
+			ed.AddAfterReadline(hist.appendHistory)
+		}
+	}
+
+	hl := &histlist{}
+	histlistBinding := eddefs.EmptyBindingMap
+
+	historyNs := eval.Ns{
+		"binding": vars.FromPtr(&hist.binding),
+		"list":    vars.NewRo(List{&hist.mutex, ed.Daemon()}),
+	}
+	historyNs.AddBuiltinFns("edit:history:", map[string]interface{}{
+		"start":        hist.start,
+		"up":           hist.up,
+		"down":         hist.down,
+		"down-or-quit": hist.downOrQuit,
+		"default":      hist.defaultFn,
+
+		"fast-forward": hist.fuser.FastForward,
+	})
+
+	histlistNs := eval.Ns{
+		"binding": vars.FromPtr(&histlistBinding),
+	}
+	histlistNs.AddBuiltinFns("edit:histlist:", map[string]interface{}{
+		"start": func() {
+			hl.start(ed, hist.fuser, histlistBinding)
+		},
+		"toggle-dedup":            func() { hl.toggleDedup(ed) },
+		"toggle-case-sensitivity": func() { hl.toggleCaseSensitivity(ed) },
+	})
+
+	ns.AddNs("history", historyNs)
+	ns.AddNs("histlist", histlistNs)
+	// TODO(xiaq): Rename and put in edit:history
+	ns.AddBuiltinFn("edit:", "command-history", hist.commandHistory)
+}
+
+func (h *hist) Teardown() {
+	h.walker = nil
+	h.bufferLen = 0
+}
+
+func (h *hist) Binding(k ui.Key) eval.Callable {
+	return h.binding.GetOrDefault(k)
+}
+
+func (h *hist) ModeLine() ui.Renderer {
+	return ui.NewModeLineRenderer(
+		fmt.Sprintf(" HISTORY #%d ", h.walker.CurrentSeq()), "")
+}
+
+func (h *hist) Replacement() (int, int, string) {
+	begin := len(h.walker.Prefix())
+	return begin, h.bufferLen, h.walker.CurrentCmd()[begin:]
+}
+
+func (hist *hist) start() {
+	ed := hist.ed
+	if hist.fuser == nil {
+		ed.Notify("history offline")
+		return
+	}
+
+	buffer, dot := ed.Buffer()
+	prefix := buffer[:dot]
+	walker := hist.fuser.Walker(prefix)
+	_, _, err := walker.Prev()
+
+	if err == nil {
+		hist.walker = walker
+		hist.bufferLen = len(buffer)
+		ed.SetMode(hist)
+	} else {
+		ed.AddTip("no matching history item")
+	}
+}
+
+func (hist *hist) up() {
+	_, _, err := hist.walker.Prev()
+	if err != nil {
+		hist.ed.Notify("%s", err)
+	}
+}
+
+func (hist *hist) down() {
+	_, _, err := hist.walker.Next()
+	if err != nil {
+		hist.ed.Notify("%s", err)
+	}
+}
+
+func (hist *hist) downOrQuit() {
+	_, _, err := hist.walker.Next()
+	if err != nil {
+		hist.ed.SetModeInsert()
+	}
+}
+
+func (hist *hist) defaultFn() {
+	newBuffer := hist.walker.CurrentCmd()
+	hist.ed.SetBuffer(newBuffer, len(newBuffer))
+	hist.ed.SetModeInsert()
+	hist.ed.SetAction(eddefs.ReprocessKey)
+}
+
+func (hist *hist) appendHistory(line string) {
+	// Do not add empty commands or commands with leading spaces to
+	// TODO: Make this customizable.
+	if line == "" || strings.HasPrefix(line, " ") {
+		return
+	}
+
+	if hist.fuser != nil {
+		hist.mutex.Lock()
+		go func() {
+			err := hist.fuser.AddCmd(line)
+			hist.mutex.Unlock()
+			if err != nil {
+				logger.Printf("Failed to AddCmd %q: %v", line, err)
+			}
+		}()
+	}
+}
+
+func (hist *hist) commandHistory(fm *eval.Frame, args ...int) {
+	var limit, start, end int
+
+	out := fm.OutputChan()
+	cmds, err := hist.fuser.AllCmds()
+	if err != nil {
+		return
+	}
+
+	if len(args) > 0 {
+		limit = args[0]
+	}
+
+	total := len(cmds)
+	switch {
+	case limit > 0:
+		start = 0
+		end = limit
+		if limit > total {
+			end = total
+		}
+	case limit < 0:
+		start = limit + total
+		if start < 0 {
+			start = 0
+		}
+		end = total
+	default:
+		start = 0
+		end = total
+	}
+
+	for i := start; i < end; i++ {
+		out <- vals.MakeMapFromKV(
+			"id", strconv.Itoa(i),
+			"cmd", cmds[i],
+		)
+	}
+}
diff -pruN 0.11+ds1-5/edit/history/list.go 0.12+ds1-1/edit/history/list.go
--- 0.11+ds1-5/edit/history/list.go	2018-01-18 02:03:11.000000000 +0000
+++ 0.12+ds1-1/edit/history/list.go	2018-07-11 02:32:25.000000000 +0000
@@ -4,7 +4,7 @@ import (
 	"sync"
 
 	"github.com/elves/elvish/daemon"
-	"github.com/elves/elvish/eval/types"
+	"github.com/elves/elvish/eval/vals"
 	"github.com/elves/elvish/util"
 )
 
@@ -15,11 +15,6 @@ type List struct {
 	Daemon *daemon.Client
 }
 
-var (
-	_ types.Value    = List{}
-	_ types.ListLike = List{}
-)
-
 func (hv List) Kind() string {
 	return "list"
 }
@@ -49,7 +44,7 @@ func (hv List) Len() int {
 	return nextseq - 1
 }
 
-func (hv List) Iterate(f func(types.Value) bool) {
+func (hv List) Iterate(f func(interface{}) bool) {
 	hv.RLock()
 	defer hv.RUnlock()
 
@@ -58,29 +53,33 @@ func (hv List) Iterate(f func(types.Valu
 	maybeThrow(err)
 
 	for _, cmd := range cmds {
-		if !f(types.String(cmd)) {
+		if !f(string(cmd)) {
 			break
 		}
 	}
 }
 
-func (hv List) IndexOne(idx types.Value) types.Value {
+func (hv List) Index(rawIndex interface{}) (interface{}, error) {
 	hv.RLock()
 	defer hv.RUnlock()
 
-	slice, i, j := types.ParseAndFixListIndex(types.ToString(idx), hv.Len())
-	if slice {
-		cmds, err := hv.Daemon.Cmds(i+1, j+1)
-		maybeThrow(err)
-		vs := make([]types.Value, len(cmds))
-		for i := range cmds {
-			vs[i] = types.String(cmds[i])
+	index, err := vals.ConvertListIndex(rawIndex, hv.Len())
+	if err != nil {
+		return nil, err
+	}
+	if index.Slice {
+		cmds, err := hv.Daemon.Cmds(index.Lower+1, index.Upper+1)
+		if err != nil {
+			return nil, err
 		}
-		return types.MakeList(vs...)
+		l := vals.EmptyList
+		for _, cmd := range cmds {
+			l = l.Cons(cmd)
+		}
+		return l, nil
 	}
-	s, err := hv.Daemon.Cmd(i + 1)
-	maybeThrow(err)
-	return types.String(s)
+	s, err := hv.Daemon.Cmd(index.Lower + 1)
+	return string(s), err
 }
 
 func maybeThrow(e error) {
diff -pruN 0.11+ds1-5/edit/history.go 0.12+ds1-1/edit/history.go
--- 0.11+ds1-5/edit/history.go	2018-01-18 02:03:11.000000000 +0000
+++ 0.12+ds1-1/edit/history.go	1970-01-01 00:00:00.000000000 +0000
@@ -1,120 +0,0 @@
-package edit
-
-import (
-	"errors"
-	"fmt"
-	"strings"
-
-	"github.com/elves/elvish/edit/history"
-	"github.com/elves/elvish/edit/ui"
-	"github.com/elves/elvish/eval"
-	"github.com/elves/elvish/eval/vartypes"
-)
-
-// Command history mode.
-
-var _ = registerBuiltins("history", map[string]func(*Editor){
-	"start":              historyStart,
-	"up":                 wrapHistoryBuiltin(historyUp),
-	"down":               wrapHistoryBuiltin(historyDown),
-	"down-or-quit":       wrapHistoryBuiltin(historyDownOrQuit),
-	"switch-to-histlist": wrapHistoryBuiltin(historySwitchToHistlist),
-	"default":            wrapHistoryBuiltin(historyDefault),
-})
-
-type hist struct {
-	*history.Walker
-}
-
-func (*hist) Binding(m map[string]vartypes.Variable, k ui.Key) eval.Fn {
-	return getBinding(m[modeHistory], k)
-}
-
-func (h *hist) ModeLine() ui.Renderer {
-	return modeLineRenderer{fmt.Sprintf(" HISTORY #%d ", h.CurrentSeq()), ""}
-}
-
-func historyStart(ed *Editor) {
-	if ed.historyFuser == nil {
-		ed.Notify("history offline")
-		return
-	}
-	prefix := ed.buffer[:ed.dot]
-	walker := ed.historyFuser.Walker(prefix)
-	hist := hist{walker}
-	_, _, err := hist.Prev()
-	if err == nil {
-		ed.mode = &hist
-	} else {
-		ed.addTip("no matching history item")
-	}
-}
-
-var errNotHistory = errors.New("not in history mode")
-
-func wrapHistoryBuiltin(f func(*Editor, *hist)) func(*Editor) {
-	return func(ed *Editor) {
-		hist, ok := ed.mode.(*hist)
-		if !ok {
-			throw(errNotHistory)
-		}
-		f(ed, hist)
-	}
-}
-
-func historyUp(ed *Editor, hist *hist) {
-	_, _, err := hist.Prev()
-	if err != nil {
-		ed.Notify("%s", err)
-	}
-}
-
-func historyDown(ed *Editor, hist *hist) {
-	_, _, err := hist.Next()
-	if err != nil {
-		ed.Notify("%s", err)
-	}
-}
-
-func historyDownOrQuit(ed *Editor, hist *hist) {
-	_, _, err := hist.Next()
-	if err != nil {
-		ed.mode = &ed.insert
-	}
-}
-
-func historySwitchToHistlist(ed *Editor, hist *hist) {
-	histlistStart(ed)
-	if l, _, ok := getHistlist(ed); ok {
-		ed.buffer = ""
-		ed.dot = 0
-		l.changeFilter(hist.Prefix())
-	}
-}
-
-func historyDefault(ed *Editor, hist *hist) {
-	ed.buffer = hist.CurrentCmd()
-	ed.dot = len(ed.buffer)
-	ed.mode = &ed.insert
-	ed.setAction(reprocessKey)
-}
-
-func (ed *Editor) appendHistory(line string) {
-	// TODO: should have a user variable to control the behavior
-	// Do not add command leading by space into history. This is
-	// useful for confidential operations.
-	if strings.HasPrefix(line, " ") {
-		return
-	}
-
-	if ed.daemon != nil && ed.historyFuser != nil {
-		ed.historyMutex.Lock()
-		go func() {
-			err := ed.historyFuser.AddCmd(line)
-			ed.historyMutex.Unlock()
-			if err != nil {
-				logger.Printf("Failed to AddCmd %q: %v", line, err)
-			}
-		}()
-	}
-}
diff -pruN 0.11+ds1-5/edit/hooks.go 0.12+ds1-1/edit/hooks.go
--- 0.11+ds1-5/edit/hooks.go	2018-01-18 02:03:11.000000000 +0000
+++ 0.12+ds1-1/edit/hooks.go	1970-01-01 00:00:00.000000000 +0000
@@ -1,52 +0,0 @@
-package edit
-
-import (
-	"fmt"
-	"os"
-
-	"github.com/elves/elvish/eval"
-	"github.com/elves/elvish/eval/types"
-	"github.com/elves/elvish/eval/vartypes"
-)
-
-// The $le:{before,after}-readline lists that contain hooks. We might have more
-// hooks in future.
-
-var _ = RegisterVariable("before-readline", makeListVariable)
-
-func (ed *Editor) beforeReadLine() types.List {
-	return ed.variables["before-readline"].Get().(types.List)
-}
-
-var _ = RegisterVariable("after-readline", makeListVariable)
-
-func (ed *Editor) afterReadLine() types.List {
-	return ed.variables["after-readline"].Get().(types.List)
-}
-
-func makeListVariable() vartypes.Variable {
-	return vartypes.NewValidatedPtr(types.EmptyList, vartypes.ShouldBeList)
-}
-
-func callHooks(ev *eval.Evaler, li types.List, args ...types.Value) {
-	if li.Len() == 0 {
-		return
-	}
-
-	li.Iterate(func(v types.Value) bool {
-		opfunc := func(ec *eval.Frame) {
-			fn, ok := v.(eval.Fn)
-			if !ok {
-				fmt.Fprintf(os.Stderr, "not a function: %s\n", v.Repr(types.NoPretty))
-				return
-			}
-			err := ec.PCall(fn, args, eval.NoOpts)
-			if err != nil {
-				// TODO Print stack trace.
-				fmt.Fprintf(os.Stderr, "function error: %s\n", err.Error())
-			}
-		}
-		ev.Eval(eval.Op{opfunc, -1, -1}, eval.NewInternalSource("[hooks]"))
-		return true
-	})
-}
diff -pruN 0.11+ds1-5/edit/insert.go 0.12+ds1-1/edit/insert.go
--- 0.11+ds1-5/edit/insert.go	2018-01-18 02:03:11.000000000 +0000
+++ 0.12+ds1-1/edit/insert.go	1970-01-01 00:00:00.000000000 +0000
@@ -1,340 +0,0 @@
-package edit
-
-import (
-	"strings"
-	"unicode"
-	"unicode/utf8"
-
-	"github.com/elves/elvish/edit/ui"
-	"github.com/elves/elvish/eval"
-	"github.com/elves/elvish/eval/vartypes"
-	"github.com/elves/elvish/util"
-)
-
-// Builtins related to insert and command mode.
-
-var (
-	_ = registerBuiltins("", map[string]func(*Editor){
-		"kill-line-left":       killLineLeft,
-		"kill-line-right":      killLineRight,
-		"kill-word-left":       killWordLeft,
-		"kill-small-word-left": killSmallWordLeft,
-		"kill-rune-left":       killRuneLeft,
-		"kill-rune-right":      killRuneRight,
-
-		"move-dot-left":       moveDotLeft,
-		"move-dot-right":      moveDotRight,
-		"move-dot-left-word":  moveDotLeftWord,
-		"move-dot-right-word": moveDotRightWord,
-		"move-dot-sol":        moveDotSOL,
-		"move-dot-eol":        moveDotEOL,
-		"move-dot-up":         moveDotUp,
-		"move-dot-down":       moveDotDown,
-
-		"insert-last-word": insertLastWord,
-		"insert-key":       insertKey,
-
-		"return-line": returnLine,
-		"smart-enter": smartEnter,
-		"return-eof":  returnEOF,
-
-		"toggle-quote-paste": toggleQuotePaste,
-		"insert-raw":         startInsertRaw,
-
-		"end-of-history": endOfHistory,
-		"redraw":         redraw,
-	})
-	_ = registerBuiltins("insert", map[string]func(*Editor){
-		"start":   insertStart,
-		"default": insertDefault,
-	})
-	_ = registerBuiltins("command", map[string]func(*Editor){
-		"start":   commandStart,
-		"default": commandDefault,
-	})
-)
-
-type insert struct {
-	quotePaste bool
-	// The number of consecutive key inserts. Used for abbreviation expansion.
-	literalInserts int
-	// Indicates whether a key was inserted (via insert-default). A hack for
-	// maintaining the inserts field.
-	insertedLiteral bool
-}
-
-// ui.Insert mode is the default mode and has an empty mode.
-func (ins *insert) ModeLine() ui.Renderer {
-	if ins.quotePaste {
-		return modeLineRenderer{" INSERT (quote paste) ", ""}
-	}
-	return nil
-}
-
-func (*insert) Binding(m map[string]vartypes.Variable, k ui.Key) eval.Fn {
-	return getBinding(m[modeInsert], k)
-}
-
-type command struct{}
-
-func (*command) ModeLine() ui.Renderer {
-	return modeLineRenderer{" COMMAND ", ""}
-}
-
-func (*command) Binding(m map[string]vartypes.Variable, k ui.Key) eval.Fn {
-	return getBinding(m[modeCommand], k)
-}
-
-func insertStart(ed *Editor) {
-	ed.mode = &ed.insert
-}
-
-func commandStart(ed *Editor) {
-	ed.mode = &ed.command
-}
-
-func killLineLeft(ed *Editor) {
-	sol := util.FindLastSOL(ed.buffer[:ed.dot])
-	ed.buffer = ed.buffer[:sol] + ed.buffer[ed.dot:]
-	ed.dot = sol
-}
-
-func killLineRight(ed *Editor) {
-	eol := util.FindFirstEOL(ed.buffer[ed.dot:]) + ed.dot
-	ed.buffer = ed.buffer[:ed.dot] + ed.buffer[eol:]
-}
-
-// NOTE(xiaq): A word is a run of non-space runes. When killing a word,
-// trimming spaces are removed as well. Examples:
-// "abc  xyz" -> "abc  ", "abc xyz " -> "abc  ".
-
-func killWordLeft(ed *Editor) {
-	if ed.dot == 0 {
-		return
-	}
-	space := strings.LastIndexFunc(
-		strings.TrimRightFunc(ed.buffer[:ed.dot], unicode.IsSpace),
-		unicode.IsSpace) + 1
-	ed.buffer = ed.buffer[:space] + ed.buffer[ed.dot:]
-	ed.dot = space
-}
-
-// NOTE(xiaq): A small word is either a run of alphanumeric (Unicode category L
-// or N) runes or a run of non-alphanumeric runes. This is consistent with vi's
-// definition of word, except that "_" is not considered alphanumeric. When
-// killing a small word, trimming spaces are removed as well. Examples:
-// "abc/~" -> "abc", "~/abc" -> "~/", "abc* " -> "abc"
-
-func killSmallWordLeft(ed *Editor) {
-	left := strings.TrimRightFunc(ed.buffer[:ed.dot], unicode.IsSpace)
-	// The case of left == "" is handled as well.
-	r, _ := utf8.DecodeLastRuneInString(left)
-	if isAlnum(r) {
-		left = strings.TrimRightFunc(left, isAlnum)
-	} else {
-		left = strings.TrimRightFunc(
-			left, func(r rune) bool { return !isAlnum(r) })
-	}
-	ed.buffer = left + ed.buffer[ed.dot:]
-	ed.dot = len(left)
-}
-
-func isAlnum(r rune) bool {
-	return unicode.IsLetter(r) || unicode.IsNumber(r)
-}
-
-func killRuneLeft(ed *Editor) {
-	if ed.dot > 0 {
-		_, w := utf8.DecodeLastRuneInString(ed.buffer[:ed.dot])
-		ed.buffer = ed.buffer[:ed.dot-w] + ed.buffer[ed.dot:]
-		ed.dot -= w
-	} else {
-		ed.flash()
-	}
-}
-
-func killRuneRight(ed *Editor) {
-	if ed.dot < len(ed.buffer) {
-		_, w := utf8.DecodeRuneInString(ed.buffer[ed.dot:])
-		ed.buffer = ed.buffer[:ed.dot] + ed.buffer[ed.dot+w:]
-	} else {
-		ed.flash()
-	}
-}
-
-func moveDotLeft(ed *Editor) {
-	_, w := utf8.DecodeLastRuneInString(ed.buffer[:ed.dot])
-	ed.dot -= w
-}
-
-func moveDotRight(ed *Editor) {
-	_, w := utf8.DecodeRuneInString(ed.buffer[ed.dot:])
-	ed.dot += w
-}
-
-func moveDotLeftWord(ed *Editor) {
-	if ed.dot == 0 {
-		return
-	}
-	space := strings.LastIndexFunc(
-		strings.TrimRightFunc(ed.buffer[:ed.dot], unicode.IsSpace),
-		unicode.IsSpace) + 1
-	ed.dot = space
-}
-
-func moveDotRightWord(ed *Editor) {
-	// Move to first space
-	p := strings.IndexFunc(ed.buffer[ed.dot:], unicode.IsSpace)
-	if p == -1 {
-		ed.dot = len(ed.buffer)
-		return
-	}
-	ed.dot += p
-	// Move to first nonspace
-	p = strings.IndexFunc(ed.buffer[ed.dot:], notSpace)
-	if p == -1 {
-		ed.dot = len(ed.buffer)
-		return
-	}
-	ed.dot += p
-}
-
-func notSpace(r rune) bool {
-	return !unicode.IsSpace(r)
-}
-
-func moveDotSOL(ed *Editor) {
-	sol := util.FindLastSOL(ed.buffer[:ed.dot])
-	ed.dot = sol
-}
-
-func moveDotEOL(ed *Editor) {
-	eol := util.FindFirstEOL(ed.buffer[ed.dot:]) + ed.dot
-	ed.dot = eol
-}
-
-func moveDotUp(ed *Editor) {
-	sol := util.FindLastSOL(ed.buffer[:ed.dot])
-	if sol == 0 {
-		ed.flash()
-		return
-	}
-	prevEOL := sol - 1
-	prevSOL := util.FindLastSOL(ed.buffer[:prevEOL])
-	width := util.Wcswidth(ed.buffer[sol:ed.dot])
-	ed.dot = prevSOL + len(util.TrimWcwidth(ed.buffer[prevSOL:prevEOL], width))
-}
-
-func moveDotDown(ed *Editor) {
-	eol := util.FindFirstEOL(ed.buffer[ed.dot:]) + ed.dot
-	if eol == len(ed.buffer) {
-		ed.flash()
-		return
-	}
-	nextSOL := eol + 1
-	nextEOL := util.FindFirstEOL(ed.buffer[nextSOL:]) + nextSOL
-	sol := util.FindLastSOL(ed.buffer[:ed.dot])
-	width := util.Wcswidth(ed.buffer[sol:ed.dot])
-	ed.dot = nextSOL + len(util.TrimWcwidth(ed.buffer[nextSOL:nextEOL], width))
-}
-
-func insertLastWord(ed *Editor) {
-	if ed.daemon == nil {
-		ed.addTip("daemon offline")
-		return
-	}
-	_, cmd, err := ed.daemon.PrevCmd(-1, "")
-	if err == nil {
-		ed.insertAtDot(lastWord(cmd))
-	} else {
-		ed.addTip("db error: %s", err.Error())
-	}
-}
-
-func lastWord(s string) string {
-	words := wordify(s)
-	if len(words) == 0 {
-		return ""
-	}
-	return words[len(words)-1]
-}
-
-func insertKey(ed *Editor) {
-	k := ed.lastKey
-	ed.insertAtDot(string(k.Rune))
-}
-
-func returnLine(ed *Editor) {
-	ed.setAction(commitLine)
-}
-
-func smartEnter(ed *Editor) {
-	if ed.parseErrorAtEnd {
-		// There is a parsing error at the end. ui.Insert a newline and copy
-		// indents from previous line.
-		indent := findLastIndent(ed.buffer[:ed.dot])
-		ed.insertAtDot("\n" + indent)
-	} else {
-		returnLine(ed)
-	}
-}
-
-func findLastIndent(s string) string {
-	line := s[util.FindLastSOL(s):]
-	trimmed := strings.TrimLeft(line, " \t")
-	return line[:len(line)-len(trimmed)]
-}
-
-func returnEOF(ed *Editor) {
-	if len(ed.buffer) == 0 {
-		ed.setAction(commitEOF)
-	}
-}
-
-func toggleQuotePaste(ed *Editor) {
-	ed.insert.quotePaste = !ed.insert.quotePaste
-}
-
-func endOfHistory(ed *Editor) {
-	ed.Notify("End of history")
-}
-
-func redraw(ed *Editor) {
-	ed.refresh(true, true)
-}
-
-func insertDefault(ed *Editor) {
-	k := ed.lastKey
-	if likeChar(k) {
-		insertKey(ed)
-		// Match abbreviations.
-		expanded := false
-		literals := ed.buffer[ed.dot-ed.insert.literalInserts-1 : ed.dot]
-		ed.abbrIterate(func(abbr, full string) bool {
-			if strings.HasSuffix(literals, abbr) {
-				ed.buffer = ed.buffer[:ed.dot-len(abbr)] + full + ed.buffer[ed.dot:]
-				ed.dot += len(full) - len(abbr)
-				expanded = true
-				return false
-			}
-			return true
-		})
-		// No match.
-		if !expanded {
-			ed.insert.insertedLiteral = true
-		}
-	} else {
-		ed.Notify("Unbound: %s", k)
-	}
-}
-
-// likeChar returns if a key looks like a character meant to be input (as
-// opposed to a function key).
-func likeChar(k ui.Key) bool {
-	return k.Mod == 0 && k.Rune > 0 && unicode.IsGraphic(k.Rune)
-}
-
-func commandDefault(ed *Editor) {
-	k := ed.lastKey
-	ed.Notify("Unbound: %s", k)
-}
diff -pruN 0.11+ds1-5/edit/lastcmd/lastcmd.go 0.12+ds1-1/edit/lastcmd/lastcmd.go
--- 0.11+ds1-5/edit/lastcmd/lastcmd.go	1970-01-01 00:00:00.000000000 +0000
+++ 0.12+ds1-1/edit/lastcmd/lastcmd.go	2018-07-11 02:32:25.000000000 +0000
@@ -0,0 +1,117 @@
+package lastcmd
+
+import (
+	"fmt"
+	"strconv"
+	"strings"
+
+	"github.com/elves/elvish/edit/eddefs"
+	"github.com/elves/elvish/edit/ui"
+	"github.com/elves/elvish/eval"
+	"github.com/elves/elvish/eval/vars"
+	"github.com/elves/elvish/parse/parseutil"
+)
+
+type state struct {
+	line     string
+	words    []string
+	filtered []entry
+	minus    bool
+}
+
+type entry struct {
+	i int
+	s string
+}
+
+// Init initializes the lastcmd module for an Editor.
+func Init(ed eddefs.Editor, ns eval.Ns) {
+	lc := &state{}
+	binding := eddefs.EmptyBindingMap
+
+	subns := eval.Ns{
+		"binding": vars.FromPtr(&binding),
+	}
+	subns.AddBuiltinFns("edit:lastcmd:", map[string]interface{}{
+		"start":       func() { lc.start(ed, binding) },
+		"accept-line": func() { lc.acceptLine(ed) },
+	})
+	ns.AddNs("lastcmd", subns)
+}
+
+func newState(line string) *state {
+	return &state{line, parseutil.Wordify(line), nil, false}
+}
+
+func (*state) AutoAccept() bool {
+	return true
+}
+
+func (*state) ModeTitle(int) string {
+	return " LASTCMD "
+}
+
+func (s *state) Len() int {
+	return len(s.filtered)
+}
+
+func (s *state) Show(i int) (string, ui.Styled) {
+	entry := s.filtered[i]
+	var head string
+	if entry.i == -1 {
+		head = "M-1"
+	} else if s.minus {
+		head = fmt.Sprintf("%d", entry.i-len(s.words))
+	} else {
+		head = fmt.Sprintf("%d", entry.i)
+	}
+	return head, ui.Unstyled(entry.s)
+}
+
+func (s *state) Filter(filter string) int {
+	s.filtered = nil
+	s.minus = len(filter) > 0 && filter[0] == '-'
+	if filter == "" || filter == "-" {
+		s.filtered = append(s.filtered, entry{-1, s.line})
+	} else if _, err := strconv.Atoi(filter); err != nil {
+		return -1
+	}
+	// Quite inefficient way to filter by prefix of stringified index.
+	n := len(s.words)
+	for i, word := range s.words {
+		if filter == "" ||
+			(!s.minus && strings.HasPrefix(strconv.Itoa(i), filter)) ||
+			(s.minus && strings.HasPrefix(strconv.Itoa(i-n), filter)) {
+			s.filtered = append(s.filtered, entry{i, word})
+		}
+	}
+	if len(s.filtered) == 0 {
+		return -1
+	}
+	return 0
+}
+
+func (s *state) Accept(i int, ed eddefs.Editor) {
+	ed.InsertAtDot(s.filtered[i].s)
+	ed.SetModeInsert()
+}
+
+func (s *state) start(ed eddefs.Editor, binding eddefs.BindingMap) {
+	daemon := ed.Daemon()
+	if daemon == nil {
+		ed.Notify("store offline, cannot start lastcmd mode")
+		return
+	}
+	_, cmd, err := daemon.PrevCmd(-1, "")
+	if err != nil {
+		ed.Notify("db error: %s", err.Error())
+		return
+	}
+	*s = *newState(cmd)
+	ed.SetModeListing(binding, s)
+}
+
+func (s *state) acceptLine(ed eddefs.Editor) {
+	ed.InsertAtDot(s.line)
+	ed.SetModeInsert()
+}
diff -pruN 0.11+ds1-5/edit/lastcmd/lastcmd_test.go 0.12+ds1-1/edit/lastcmd/lastcmd_test.go
--- 0.11+ds1-5/edit/lastcmd/lastcmd_test.go	1970-01-01 00:00:00.000000000 +0000
+++ 0.12+ds1-1/edit/lastcmd/lastcmd_test.go	2018-07-11 02:32:25.000000000 +0000
@@ -0,0 +1,32 @@
+package lastcmd
+
+import (
+	"testing"
+
+	"github.com/elves/elvish/edit/eddefs"
+	"github.com/elves/elvish/edit/ui"
+)
+
+var (
+	theLine    = "qw search 'foo bar ~y'"
+	theLastCmd = newState(theLine)
+
+	tests = []eddefs.ListingProviderFilterTest{
+		{"", []eddefs.ListingShown{
+			{"M-1", ui.Unstyled(theLine)},
+			{"0", ui.Unstyled("qw")},
+			{"1", ui.Unstyled("search")},
+			{"2", ui.Unstyled("'foo bar ~y'")}}},
+		{"1", []eddefs.ListingShown{{"1", ui.Unstyled("search")}}},
+		{"-", []eddefs.ListingShown{
+			{"M-1", ui.Unstyled(theLine)},
+			{"-3", ui.Unstyled("qw")},
+			{"-2", ui.Unstyled("search")},
+			{"-1", ui.Unstyled("'foo bar ~y'")}}},
+		{"-1", []eddefs.ListingShown{{"-1", ui.Unstyled("'foo bar ~y'")}}},
+	}
+)
+
+func TestLastCmd(t *testing.T) {
+	eddefs.TestListingProviderFilter(t, "theLastCmd", theLastCmd, tests)
+}
diff -pruN 0.11+ds1-5/edit/lastcmd.go 0.12+ds1-1/edit/lastcmd.go
--- 0.11+ds1-5/edit/lastcmd.go	2018-01-18 02:03:11.000000000 +0000
+++ 0.12+ds1-1/edit/lastcmd.go	1970-01-01 00:00:00.000000000 +0000
@@ -1,124 +0,0 @@
-package edit
-
-import (
-	"fmt"
-	"strconv"
-	"strings"
-
-	"github.com/elves/elvish/edit/ui"
-)
-
-// LastCmd mode.
-
-var _ = registerBuiltins(modeLastCmd, map[string]func(*Editor){
-	"start":       lastcmdStart,
-	"alt-default": lastcmdAltDefault,
-})
-
-type lastcmdEntry struct {
-	i int
-	s string
-}
-
-type lastcmd struct {
-	line     string
-	words    []string
-	filtered []lastcmdEntry
-	minus    bool
-}
-
-func newLastCmd(line string) *listing {
-	b := &lastcmd{line, wordify(line), nil, false}
-	l := newListing(modeLastCmd, b)
-	return &l
-}
-
-func (b *lastcmd) ModeTitle(int) string {
-	return " LASTCMD "
-}
-
-func (b *lastcmd) Len() int {
-	return len(b.filtered)
-}
-
-func (b *lastcmd) Show(i int) (string, ui.Styled) {
-	entry := b.filtered[i]
-	var head string
-	if entry.i == -1 {
-		head = "M-1"
-	} else if b.minus {
-		head = fmt.Sprintf("%d", entry.i-len(b.words))
-	} else {
-		head = fmt.Sprintf("%d", entry.i)
-	}
-	return head, ui.Unstyled(entry.s)
-}
-
-func (b *lastcmd) Filter(filter string) int {
-	b.filtered = nil
-	b.minus = len(filter) > 0 && filter[0] == '-'
-	if filter == "" || filter == "-" {
-		b.filtered = append(b.filtered, lastcmdEntry{-1, b.line})
-	} else if _, err := strconv.Atoi(filter); err != nil {
-		return -1
-	}
-	// Quite inefficient way to filter by prefix of stringified index.
-	n := len(b.words)
-	for i, word := range b.words {
-		if filter == "" ||
-			(!b.minus && strings.HasPrefix(strconv.Itoa(i), filter)) ||
-			(b.minus && strings.HasPrefix(strconv.Itoa(i-n), filter)) {
-			b.filtered = append(b.filtered, lastcmdEntry{i, word})
-		}
-	}
-	if len(b.filtered) == 0 {
-		return -1
-	}
-	return 0
-}
-
-// Editor interface.
-
-func (b *lastcmd) Accept(i int, ed *Editor) {
-	ed.insertAtDot(b.filtered[i].s)
-	insertStart(ed)
-}
-
-func lastcmdStart(ed *Editor) {
-	logger.Println("lastcmd-alt-start")
-	_, cmd, err := ed.daemon.PrevCmd(-1, "")
-	if err != nil {
-		ed.Notify("db error: %s", err.Error())
-		return
-	}
-	ed.mode = newLastCmd(cmd)
-}
-
-func lastcmdAltDefault(ed *Editor) {
-	l, lc := getLastcmd(ed)
-	if l == nil {
-		return
-	}
-	logger.Println("lastcmd-alt-default")
-	if ed.lastKey == (ui.Key{'1', ui.Alt}) {
-		lc.Accept(0, ed)
-		logger.Println("accepting")
-	} else if l.handleFilterKey(ed.lastKey) {
-		if lc.Len() == 1 {
-			lc.Accept(l.selected, ed)
-			logger.Println("accepting")
-		}
-	} else {
-		insertStart(ed)
-		ed.setAction(reprocessKey)
-	}
-}
-
-func getLastcmd(ed *Editor) (*listing, *lastcmd) {
-	if l, ok := ed.mode.(*listing); ok {
-		if lc, ok := l.provider.(*lastcmd); ok {
-			return l, lc
-		}
-	}
-	return nil, nil
-}
diff -pruN 0.11+ds1-5/edit/lastcmd_test.go 0.12+ds1-1/edit/lastcmd_test.go
--- 0.11+ds1-5/edit/lastcmd_test.go	2018-01-18 02:03:11.000000000 +0000
+++ 0.12+ds1-1/edit/lastcmd_test.go	1970-01-01 00:00:00.000000000 +0000
@@ -1,31 +0,0 @@
-package edit
-
-import (
-	"testing"
-
-	"github.com/elves/elvish/edit/ui"
-)
-
-var (
-	theLine    = "qw search 'foo bar ~y'"
-	theLastCmd = newLastCmd(theLine)
-
-	lastcmdFilterTests = []listingFilterTestCases{
-		{"", []shown{
-			{"M-1", ui.Unstyled(theLine)},
-			{"0", ui.Unstyled("qw")},
-			{"1", ui.Unstyled("search")},
-			{"2", ui.Unstyled("'foo bar ~y'")}}},
-		{"1", []shown{{"1", ui.Unstyled("search")}}},
-		{"-", []shown{
-			{"M-1", ui.Unstyled(theLine)},
-			{"-3", ui.Unstyled("qw")},
-			{"-2", ui.Unstyled("search")},
-			{"-1", ui.Unstyled("'foo bar ~y'")}}},
-		{"-1", []shown{{"-1", ui.Unstyled("'foo bar ~y'")}}},
-	}
-)
-
-func TestLastCmd(t *testing.T) {
-	testListingFilter(t, "theLastCmd", theLastCmd, lastcmdFilterTests)
-}
diff -pruN 0.11+ds1-5/edit/listing_common_test.go 0.12+ds1-1/edit/listing_common_test.go
--- 0.11+ds1-5/edit/listing_common_test.go	2018-01-18 02:03:11.000000000 +0000
+++ 0.12+ds1-1/edit/listing_common_test.go	1970-01-01 00:00:00.000000000 +0000
@@ -1,39 +0,0 @@
-package edit
-
-import (
-	"reflect"
-	"testing"
-
-	"github.com/elves/elvish/edit/ui"
-)
-
-type shown struct {
-	header  string
-	content ui.Styled
-}
-
-type listingFilterTestCases struct {
-	filter     string
-	wantShowns []shown
-}
-
-func testListingFilter(t *testing.T, name string, l *listing, testcases []listingFilterTestCases) {
-	ls := l.provider
-	for _, testcase := range testcases {
-		ls.Filter(testcase.filter)
-
-		l := ls.Len()
-		if l != len(testcase.wantShowns) {
-			t.Errorf("%s.Len() -> %d, want %d (filter was %q)",
-				name, l, len(testcase.wantShowns), testcase.filter)
-		} else {
-			for i, want := range testcase.wantShowns {
-				header, content := ls.Show(i)
-				if header != want.header || !reflect.DeepEqual(content, want.content) {
-					t.Errorf("%s.Show(%d) => (%v, %v), want (%v, %v) (filter was %q)",
-						name, i, header, content, want.header, want.content, testcase.filter)
-				}
-			}
-		}
-	}
-}
diff -pruN 0.11+ds1-5/edit/listing.go 0.12+ds1-1/edit/listing.go
--- 0.11+ds1-5/edit/listing.go	2018-01-18 02:03:11.000000000 +0000
+++ 0.12+ds1-1/edit/listing.go	1970-01-01 00:00:00.000000000 +0000
@@ -1,336 +0,0 @@
-package edit
-
-import (
-	"container/list"
-	"errors"
-	"fmt"
-	"strings"
-	"unicode/utf8"
-
-	"github.com/elves/elvish/edit/ui"
-	"github.com/elves/elvish/eval"
-	"github.com/elves/elvish/eval/vartypes"
-	"github.com/elves/elvish/util"
-)
-
-var _ = registerBuiltins(modeListing, map[string]func(*Editor){
-	"up":         func(ed *Editor) { getListing(ed).up(false) },
-	"up-cycle":   func(ed *Editor) { getListing(ed).up(true) },
-	"page-up":    func(ed *Editor) { getListing(ed).pageUp() },
-	"down":       func(ed *Editor) { getListing(ed).down(false) },
-	"down-cycle": func(ed *Editor) { getListing(ed).down(true) },
-	"page-down":  func(ed *Editor) { getListing(ed).pageDown() },
-	"backspace":  func(ed *Editor) { getListing(ed).backspace() },
-	"accept":     func(ed *Editor) { getListing(ed).accept(ed) },
-	"accept-close": func(ed *Editor) {
-		getListing(ed).accept(ed)
-		insertStart(ed)
-	},
-	"default": func(ed *Editor) { getListing(ed).defaultBinding(ed) },
-})
-
-// listing implements a listing mode that supports the notion of selecting an
-// entry and filtering entries.
-type listing struct {
-	name        string
-	provider    listingProvider
-	selected    int
-	filter      string
-	pagesize    int
-	headerWidth int
-}
-
-type listingProvider interface {
-	Len() int
-	Show(i int) (string, ui.Styled)
-	Filter(filter string) int
-	Accept(i int, ed *Editor)
-	ModeTitle(int) string
-}
-
-type placeholderer interface {
-	Placeholder() string
-}
-
-func newListing(t string, p listingProvider) listing {
-	l := listing{t, p, 0, "", 0, 0}
-	l.refresh()
-	for i := 0; i < p.Len(); i++ {
-		header, _ := p.Show(i)
-		width := util.Wcswidth(header)
-		if l.headerWidth < width {
-			l.headerWidth = width
-		}
-	}
-	return l
-}
-
-func (l *listing) Binding(m map[string]vartypes.Variable, k ui.Key) eval.Fn {
-	if m[l.name] == nil {
-		return getBinding(m[modeListing], k)
-	}
-	specificBindings := m[l.name].Get().(BindingTable)
-	listingBindings := m[modeListing].Get().(BindingTable)
-	// mode-specific binding -> listing binding ->
-	// mode-specific default -> listing default
-	switch {
-	case specificBindings.HasKey(k):
-		return specificBindings.get(k)
-	case listingBindings.HasKey(k):
-		return listingBindings.get(k)
-	case specificBindings.HasKey(ui.Default):
-		return specificBindings.get(ui.Default)
-	case listingBindings.HasKey(ui.Default):
-		return listingBindings.get(ui.Default)
-	default:
-		return nil
-	}
-}
-
-func (l *listing) ModeLine() ui.Renderer {
-	return modeLineRenderer{l.provider.ModeTitle(l.selected), l.filter}
-}
-
-func (l *listing) CursorOnModeLine() bool {
-	if c, ok := l.provider.(CursorOnModeLiner); ok {
-		return c.CursorOnModeLine()
-	}
-	return false
-}
-
-func (l *listing) List(maxHeight int) ui.Renderer {
-	n := l.provider.Len()
-	if n == 0 {
-		var ph string
-		if pher, ok := l.provider.(placeholderer); ok {
-			ph = pher.Placeholder()
-		} else {
-			ph = "(no result)"
-		}
-		return placeholderRenderer(ph)
-	}
-
-	// Collect the entries to show. We start from the selected entry and extend
-	// in both directions alternatingly. The entries are split into lines and
-	// then collected in a list.
-	low := l.selected
-	if low == -1 {
-		low = 0
-	}
-	high := low
-	height := 0
-	var listOfLines list.List
-	getEntry := func(i int) []ui.Styled {
-		header, content := l.provider.Show(i)
-		lines := strings.Split(content.Text, "\n")
-		styles := content.Styles
-		if i == l.selected {
-			styles = append(styles, styleForSelected...)
-		}
-		styleds := make([]ui.Styled, len(lines))
-		for i, line := range lines {
-			if l.headerWidth > 0 {
-				if i == 0 {
-					line = fmt.Sprintf("%*s %s", l.headerWidth, header, line)
-				} else {
-					line = fmt.Sprintf("%*s %s", l.headerWidth, "", line)
-				}
-			}
-			styleds[i] = ui.Styled{line, styles}
-		}
-		return styleds
-	}
-	// We start by extending high, so that the first entry to include is
-	// l.selected.
-	extendLow := false
-	lastShownIncomplete := false
-	for height < maxHeight && !(low == 0 && high == n) {
-		var i int
-		if (extendLow && low > 0) || high == n {
-			low--
-
-			entry := getEntry(low)
-			// Prepend at most the last (height - maxHeight) lines.
-			for i = len(entry) - 1; i >= 0 && height < maxHeight; i-- {
-				listOfLines.PushFront(entry[i])
-				height++
-			}
-			if i >= 0 {
-				lastShownIncomplete = true
-			}
-		} else {
-			entry := getEntry(high)
-			// Append at most the first (height - maxHeight) lines.
-			for i = 0; i < len(entry) && height < maxHeight; i++ {
-				listOfLines.PushBack(entry[i])
-				height++
-			}
-			if i < len(entry) {
-				lastShownIncomplete = true
-			}
-
-			high++
-		}
-		extendLow = !extendLow
-	}
-
-	l.pagesize = high - low
-
-	// Convert the List to a slice.
-	lines := make([]ui.Styled, 0, listOfLines.Len())
-	for p := listOfLines.Front(); p != nil; p = p.Next() {
-		lines = append(lines, p.Value.(ui.Styled))
-	}
-
-	ls := listingRenderer{lines}
-	if low > 0 || high < n || lastShownIncomplete {
-		// Need scrollbar
-		return listingWithScrollBarRenderer{ls, n, low, high, height}
-	}
-	return ls
-}
-
-func writeHorizontalScrollbar(b *ui.Buffer, n, low, high, width int) {
-	slow, shigh := findScrollInterval(n, low, high, width)
-	for i := 0; i < width; i++ {
-		if slow <= i && i < shigh {
-			b.Write(' ', styleForScrollBarThumb.String())
-		} else {
-			b.Write('━', styleForScrollBarArea.String())
-		}
-	}
-}
-
-func renderScrollbar(n, low, high, height int) *ui.Buffer {
-	slow, shigh := findScrollInterval(n, low, high, height)
-	// Logger.Printf("low = %d, high = %d, n = %d, slow = %d, shigh = %d", low, high, n, slow, shigh)
-	b := ui.NewBuffer(1)
-	for i := 0; i < height; i++ {
-		if i > 0 {
-			b.Newline()
-		}
-		if slow <= i && i < shigh {
-			b.Write(' ', styleForScrollBarThumb.String())
-		} else {
-			b.Write('│', styleForScrollBarArea.String())
-		}
-	}
-	return b
-}
-
-func findScrollInterval(n, low, high, height int) (int, int) {
-	f := func(i int) int {
-		return int(float64(i)/float64(n)*float64(height) + 0.5)
-	}
-	scrollLow, scrollHigh := f(low), f(high)
-	if scrollLow == scrollHigh {
-		if scrollHigh == high {
-			scrollLow--
-		} else {
-			scrollHigh++
-		}
-	}
-	return scrollLow, scrollHigh
-}
-
-func (l *listing) changeFilter(newfilter string) {
-	l.filter = newfilter
-	l.refresh()
-}
-
-func (l *listing) refresh() {
-	l.selected = l.provider.Filter(l.filter)
-}
-
-func (l *listing) backspace() bool {
-	_, size := utf8.DecodeLastRuneInString(l.filter)
-	if size > 0 {
-		l.changeFilter(l.filter[:len(l.filter)-size])
-		return true
-	}
-	return false
-}
-
-func (l *listing) up(cycle bool) {
-	n := l.provider.Len()
-	if n == 0 {
-		return
-	}
-	l.selected--
-	if l.selected == -1 {
-		if cycle {
-			l.selected += n
-		} else {
-			l.selected++
-		}
-	}
-}
-
-func (l *listing) pageUp() {
-	n := l.provider.Len()
-	if n == 0 {
-		return
-	}
-	l.selected -= l.pagesize
-	if l.selected < 0 {
-		l.selected = 0
-	}
-}
-
-func (l *listing) down(cycle bool) {
-	n := l.provider.Len()
-	if n == 0 {
-		return
-	}
-	l.selected++
-	if l.selected == n {
-		if cycle {
-			l.selected -= n
-		} else {
-			l.selected--
-		}
-	}
-}
-
-func (l *listing) pageDown() {
-	n := l.provider.Len()
-	if n == 0 {
-		return
-	}
-	l.selected += l.pagesize
-	if l.selected >= n {
-		l.selected = n - 1
-	}
-}
-
-func (l *listing) accept(ed *Editor) {
-	if l.selected >= 0 {
-		l.provider.Accept(l.selected, ed)
-	}
-}
-
-func (l *listing) handleFilterKey(k ui.Key) bool {
-	if likeChar(k) {
-		l.changeFilter(l.filter + string(k.Rune))
-		return true
-	}
-	return false
-}
-
-func (l *listing) defaultBinding(ed *Editor) {
-	if !l.handleFilterKey(ed.lastKey) {
-		insertStart(ed)
-		ed.setAction(reprocessKey)
-	}
-}
-
-var errNotListing = errors.New("not in a listing mode")
-
-func getListing(ed *Editor) *listing {
-	if l, ok := ed.mode.(*listing); ok {
-		return l
-	} else {
-		throw(errNotListing)
-		panic("unreachable")
-	}
-}
diff -pruN 0.11+ds1-5/edit/listing_test.go 0.12+ds1-1/edit/listing_test.go
--- 0.11+ds1-5/edit/listing_test.go	2018-01-18 02:03:11.000000000 +0000
+++ 0.12+ds1-1/edit/listing_test.go	1970-01-01 00:00:00.000000000 +0000
@@ -1,74 +0,0 @@
-package edit
-
-import (
-	"fmt"
-	"reflect"
-	"strconv"
-	"testing"
-
-	"github.com/elves/elvish/edit/ui"
-)
-
-type provider struct {
-	elems    []string
-	accepted int
-}
-
-func (p provider) Len() int                 { return len(p.elems) }
-func (p provider) Filter(string) int        { return 0 }
-func (p provider) Accept(i int, ed *Editor) { p.accepted = i }
-func (p provider) ModeTitle(i int) string   { return fmt.Sprintf("test %d", i) }
-
-func (p provider) Show(i int) (string, ui.Styled) {
-	return strconv.Itoa(i), ui.Unstyled(p.elems[i])
-}
-
-var (
-	mode = "test233"
-	p    = provider{[]string{"foo", "bar", "foobar", "lorem", "ipsum"}, -1}
-	ls   = newListing(mode, p)
-)
-
-func TestListing(t *testing.T) {
-	wantedModeLine := modeLineRenderer{"test 0", ""}
-	if modeLine := ls.ModeLine(); modeLine != wantedModeLine {
-		t.Errorf("ls.ModeLine() = %v, want %v", modeLine, wantedModeLine)
-	}
-
-	// Selecting the first element and rendering with height=2. We expect to see
-	// the first 2 elements, with the first being shown as selected.
-	testListingList(t, 0, 2, listingWithScrollBarRenderer{
-		listingRenderer: listingRenderer{[]ui.Styled{
-			{"0 foo", styleForSelected},
-			{"1 bar", ui.Styles{}},
-		}},
-		n: 5, low: 0, high: 2, height: 2,
-	})
-	// Selecting the last element and rendering with height=2. We expect to see
-	// the last 2 elements, with the last being shown as selected.
-	testListingList(t, 4, 2, listingWithScrollBarRenderer{
-		listingRenderer: listingRenderer{[]ui.Styled{
-			{"3 lorem", ui.Styles{}},
-			{"4 ipsum", styleForSelected},
-		}},
-		n: 5, low: 3, high: 5, height: 2,
-	})
-	// Selecting the middle element and rendering with height=3. We expect to
-	// see the middle element and two elements around it, with the middle being
-	// shown as selected.
-	testListingList(t, 2, 3, listingWithScrollBarRenderer{
-		listingRenderer: listingRenderer{[]ui.Styled{
-			{"1 bar", ui.Styles{}},
-			{"2 foobar", styleForSelected},
-			{"3 lorem", ui.Styles{}},
-		}},
-		n: 5, low: 1, high: 4, height: 3,
-	})
-}
-
-func testListingList(t *testing.T, i, h int, want ui.Renderer) {
-	ls.selected = i
-	if r := ls.List(h); !reflect.DeepEqual(r, want) {
-		t.Errorf("selecting %d, ls.List(%d) = %v, want %v", i, h, r, want)
-	}
-}
diff -pruN 0.11+ds1-5/edit/location/location.go 0.12+ds1-1/edit/location/location.go
--- 0.11+ds1-5/edit/location/location.go	1970-01-01 00:00:00.000000000 +0000
+++ 0.12+ds1-1/edit/location/location.go	2018-07-11 02:32:25.000000000 +0000
@@ -0,0 +1,354 @@
+// Package location implements the location mode for the editor.
+package location
+
+import (
+	"bytes"
+	"fmt"
+	"math"
+	"os"
+	"path/filepath"
+	"regexp"
+	"strings"
+
+	"github.com/elves/elvish/edit/eddefs"
+	"github.com/elves/elvish/edit/ui"
+	"github.com/elves/elvish/eval"
+	"github.com/elves/elvish/eval/vals"
+	"github.com/elves/elvish/eval/vars"
+	"github.com/elves/elvish/parse"
+	"github.com/elves/elvish/store/storedefs"
+	"github.com/elves/elvish/util"
+	"github.com/xiaq/persistent/hashmap"
+	"github.com/xiaq/persistent/vector"
+)
+
+var logger = util.GetLogger("[edit/location] ")
+
+// pinnedScore is a special value of Score in storedefs.Dir to represent that the
+// directory is pinned.
+var pinnedScore = math.Inf(1)
+
+type mode struct {
+	editor     eddefs.Editor
+	binding    eddefs.BindingMap
+	hidden     vector.Vector
+	pinned     vector.Vector
+	workspaces hashmap.Map
+	matcher    eval.Callable
+}
+
+var matchDirPatternBuiltin = eval.NewBuiltinFn("edit:location:match-dir-pattern", matchDirPattern)
+
+// Init initializes the location mode for an Editor.
+func Init(ed eddefs.Editor, ns eval.Ns) {
+	m := &mode{ed, eddefs.EmptyBindingMap,
+		vals.EmptyList, vals.EmptyList, vals.EmptyMap, matchDirPatternBuiltin}
+
+	ns.AddNs("location",
+		eval.Ns{
+			"binding":    vars.FromPtr(&m.binding),
+			"hidden":     vars.FromPtr(&m.hidden),
+			"pinned":     vars.FromPtr(&m.pinned),
+			"matcher":    vars.FromPtr(&m.matcher),
+			"workspaces": vars.FromPtr(&m.workspaces),
+		}.AddBuiltinFn("edit:location:", "start", m.start).
+			AddFn("match-dir-pattern", matchDirPatternBuiltin))
+
+	ed.Evaler().AddAfterChdir(func(string) {
+		store := ed.Daemon()
+		if store == nil {
+			return
+		}
+		pwd, err := os.Getwd()
+		if err != nil {
+			logger.Println("Failed to get pwd in after-chdir hook:", err)
+		}
+		go addDir(store, pwd, m.workspaces)
+	})
+}
+
+func addDir(store storedefs.Store, pwd string, workspaces hashmap.Map) {
+	err := store.AddDir(pwd, 1)
+	if err != nil {
+		logger.Println("add dir in after-chdir hook:", err)
+		return
+	}
+	ws := matchWorkspace(pwd, workspaces)
+	if ws == nil {
+		return
+	}
+	err = store.AddDir(ws.workspacify(pwd), 1)
+	if err != nil {
+		logger.Println("add workspacified dir in after-chdir hook:", err)
+	}
+}
+
+type wsInfo struct {
+	name string
+	root string
+}
+
+func matchWorkspace(dir string, workspaces hashmap.Map) *wsInfo {
+	for it := workspaces.Iterator(); it.HasElem(); it.Next() {
+		k, v := it.Elem()
+		name, ok := k.(string)
+		if !ok {
+			// TODO: Surface to user
+			logger.Println("$workspaces key not string", k)
+			continue
+		}
+		if strings.HasPrefix(name, "/") {
+			// TODO: Surface to user
+			logger.Println("$workspaces key starts with /", k)
+			continue
+		}
+		pattern, ok := v.(string)
+		if !ok {
+			// TODO: Surface to user
+			logger.Println("$workspaces value not string", v)
+			continue
+		}
+		if !strings.HasPrefix(pattern, "^") {
+			pattern = "^" + pattern
+		}
+		re, err := regexp.Compile(pattern)
+		if err != nil {
+			// TODO: Surface to user
+			logger.Println("$workspaces pattern invalid", pattern)
+			continue
+		}
+		if ws := re.FindString(dir); ws != "" {
+			return &wsInfo{name, ws}
+		}
+	}
+	return nil
+}
+
+func (w *wsInfo) workspacify(dir string) string {
+	return w.name + dir[len(w.root):]
+}
+
+func (w *wsInfo) unworkspacify(dir string) string {
+	return w.root + dir[len(w.name):]
+}
+
+func (m *mode) start() {
+	ed := m.editor
+
+	daemon := ed.Daemon()
+	if daemon == nil {
+		ed.Notify("store offline, cannot start location mode")
+		return
+	}
+
+	// Pinned directories are also blacklisted to prevent them from showing up
+	// twice.
+	black := convertListsToSet(m.hidden, m.pinned)
+	pwd, err := os.Getwd()
+	if err == nil {
+		black[pwd] = struct{}{}
+	}
+	stored, err := daemon.Dirs(black)
+	if err != nil {
+		ed.Notify("store error: %v", err)
+		return
+	}
+
+	// TODO: Move workspace filtering to the daemon.
+	ws := matchWorkspace(pwd, m.workspaces)
+	wsName := ""
+	if ws != nil {
+		wsName = ws.name
+	}
+	wsNameSlash := wsName + string(filepath.Separator)
+
+	var filtered []storedefs.Dir
+	for _, dir := range stored {
+		if filepath.IsAbs(dir.Path) || (wsName != "" && (dir.Path == wsName || strings.HasPrefix(dir.Path, wsNameSlash))) {
+			filtered = append(filtered, dir)
+		}
+	}
+
+	// Prepend pinned dirs.
+	pinnedDirs := convertListToDirs(m.pinned)
+	dirs := make([]storedefs.Dir, len(pinnedDirs)+len(filtered))
+	copy(dirs, pinnedDirs)
+	copy(dirs[len(pinnedDirs):], filtered)
+
+	// Drop the error. When there is an error, home is "", which is used to
+	// signify "no home known" in location.
+	home, _ := util.GetHome("")
+	ed.SetModeListing(m.binding,
+		newProvider(dirs, home, ws, ed.Evaler(), m.matcher))
+}
+
+// convertListToDirs converts a list of strings to []storedefs.Dir. It uses the
+// special score of pinnedScore to signify that the directory is pinned.
+func convertListToDirs(li vector.Vector) []storedefs.Dir {
+	pinned := make([]storedefs.Dir, 0, li.Len())
+	// XXX(xiaq): silently drops non-string items.
+	for it := li.Iterator(); it.HasElem(); it.Next() {
+		if s, ok := it.Elem().(string); ok {
+			pinned = append(pinned, storedefs.Dir{s, pinnedScore})
+		}
+	}
+	return pinned
+}
+
+func convertListsToSet(lis ...vector.Vector) map[string]struct{} {
+	set := make(map[string]struct{})
+	// XXX(xiaq): silently drops non-string items.
+	for _, li := range lis {
+		for it := li.Iterator(); it.HasElem(); it.Next() {
+			if s, ok := it.Elem().(string); ok {
+				set[s] = struct{}{}
+			}
+		}
+	}
+	return set
+}
+
+type provider struct {
+	all      []storedefs.Dir
+	filtered []storedefs.Dir
+	home     string // The home directory; leave empty if unknown.
+	ws       *wsInfo
+	ev       *eval.Evaler
+	matcher  eval.Callable
+}
+
+func newProvider(dirs []storedefs.Dir, home string, ws *wsInfo, ev *eval.Evaler, matcher eval.Callable) *provider {
+	return &provider{dirs, nil, home, ws, ev, matcher}
+}
+
+func (*provider) ModeTitle(i int) string {
+	return " LOCATION "
+}
+
+func (*provider) CursorOnModeLine() bool {
+	return true
+}
+
+func (p *provider) Len() int {
+	return len(p.filtered)
+}
+
+func (p *provider) Show(i int) (string, ui.Styled) {
+	var header string
+	score := p.filtered[i].Score
+	if score == pinnedScore {
+		header = "*"
+	} else {
+		header = fmt.Sprintf("%.0f", score)
+	}
+	return header, ui.Unstyled(showPath(p.filtered[i].Path, p.home))
+}
+
+func showPath(path, home string) string {
+	if home != "" && path == home {
+		return "~"
+	} else if home != "" && strings.HasPrefix(path, home+"/") {
+		return "~/" + parse.Quote(path[len(home)+1:])
+	} else {
+		return parse.Quote(path)
+	}
+}
+
+func (p *provider) Filter(filter string) int {
+	p.filtered = nil
+
+	// TODO: this is just a replica of `filterRawCandidates`.
+	matcherInput := make(chan interface{}, len(p.all))
+	stopCollector := make(chan struct{})
+	go func() {
+		defer close(matcherInput)
+		for _, item := range p.all {
+			select {
+			case matcherInput <- showPath(item.Path, p.home):
+				logger.Printf("put %s\n", item.Path)
+			case <-stopCollector:
+				return
+			}
+		}
+	}()
+	defer close(stopCollector)
+
+	ports := []*eval.Port{
+		{Chan: matcherInput, File: eval.DevNull}, {File: os.Stdout}, {File: os.Stderr}}
+	ec := eval.NewTopFrame(p.ev, eval.NewInternalSource("[editor matcher]"), ports)
+	args := []interface{}{filter}
+
+	values, err := ec.CaptureOutput(p.matcher, args, eval.NoOpts)
+	if err != nil {
+		logger.Printf("failed to match %s: %v", filter, err)
+		return -1
+	} else if got, expect := len(values), len(p.all); got != expect {
+		logger.Printf("wrong match count: got %d, want %d", got, expect)
+		return -1
+	}
+
+	for i, value := range values {
+		if vals.Bool(value) {
+			p.filtered = append(p.filtered, p.all[i])
+		}
+	}
+
+	if len(p.filtered) == 0 {
+		return -1
+	}
+	return 0
+}
+
+var emptyRegexp = regexp.MustCompile("")
+
+func makeLocationFilterPattern(s string, ignoreCase bool) *regexp.Regexp {
+	var b bytes.Buffer
+	if ignoreCase {
+		b.WriteString("(?i)")
+	}
+	b.WriteString(".*")
+	segs := strings.Split(s, "/")
+	for i, seg := range segs {
+		if i > 0 {
+			b.WriteString(".*/.*")
+		}
+		b.WriteString(regexp.QuoteMeta(seg))
+	}
+	b.WriteString(".*")
+	p, err := regexp.Compile(b.String())
+	if err != nil {
+		logger.Printf("failed to compile regexp %q: %v", b.String(), err)
+		return emptyRegexp
+	}
+	return p
+}
+
+func (p *provider) Accept(i int, ed eddefs.Editor) {
+	path := p.filtered[i].Path
+	if !filepath.IsAbs(path) {
+		path = p.ws.unworkspacify(path)
+	}
+	err := ed.Evaler().Chdir(path)
+	if err != nil {
+		ed.Notify("%v", err)
+	}
+	ed.SetModeInsert()
+}
+
+func matchDirPattern(fm *eval.Frame, opts eval.RawOptions, pattern string, inputs eval.Inputs) {
+	var options struct {
+		IgnoreCase bool
+	}
+	opts.Scan(&options)
+
+	p := makeLocationFilterPattern(pattern, options.IgnoreCase)
+	out := fm.OutputChan()
+	inputs(func(v interface{}) {
+		s, ok := v.(string)
+		if !ok {
+			logger.Printf("input item must be string, but got %#v", v)
+			return
+		}
+		out <- vals.Bool(p.MatchString(s))
+	})
+}
diff -pruN 0.11+ds1-5/edit/location/location_test.go 0.12+ds1-1/edit/location/location_test.go
--- 0.11+ds1-5/edit/location/location_test.go	1970-01-01 00:00:00.000000000 +0000
+++ 0.12+ds1-1/edit/location/location_test.go	2018-07-11 02:32:25.000000000 +0000
@@ -0,0 +1,91 @@
+package location
+
+import (
+	"strings"
+	"testing"
+
+	"github.com/elves/elvish/edit/eddefs"
+	"github.com/elves/elvish/edit/ui"
+	"github.com/elves/elvish/eval"
+	"github.com/elves/elvish/eval/vals"
+	"github.com/elves/elvish/store/storedefs"
+)
+
+var (
+	theLocation = newProvider([]storedefs.Dir{
+		{"/pinned", pinnedScore},
+		{"/src/github.com/elves/elvish", 300},
+		{"/src/home/xyz", 233},
+		{"/home/dir", 100},
+		{"/foo/\nbar", 77},
+		{"/usr/elves/elvish", 6},
+	}, "/home", nil, eval.NewEvaler(), matchDirPatternBuiltin)
+
+	locationFilterTests = []eddefs.ListingProviderFilterTest{
+		{"", []eddefs.ListingShown{
+			{"*", ui.Unstyled("/pinned")},
+			{"300", ui.Unstyled("/src/github.com/elves/elvish")},
+			{"233", ui.Unstyled("/src/home/xyz")},
+			{"100", ui.Unstyled("~/dir")},       // home is abbreviated
+			{"77", ui.Unstyled(`"/foo/\nbar"`)}, // special char is quoted
+			{"6", ui.Unstyled("/usr/elves/elvish")}}},
+		{"/s", []eddefs.ListingShown{
+			{"300", ui.Unstyled("/src/github.com/elves/elvish")},
+			{"233", ui.Unstyled("/src/home/xyz")},
+			{"6", ui.Unstyled("/usr/elves/elvish")}}},
+		{"/e/e", []eddefs.ListingShown{
+			{"300", ui.Unstyled("/src/github.com/elves/elvish")},
+			{"6", ui.Unstyled("/usr/elves/elvish")}}},
+		{"x", []eddefs.ListingShown{{"233", ui.Unstyled("/src/home/xyz")}}},
+		// Matchers operate on the displayed text, not the actual path.
+		// 1. Home directory is abbreviated to ~, and is matched by ~, but not by
+		//    the actual path.
+		{"~", []eddefs.ListingShown{{"100", ui.Unstyled("~/dir")}}},
+		{"home", []eddefs.ListingShown{{"233", ui.Unstyled("/src/home/xyz")}}},
+		// 2. Special characters are quoted, and are matched by the quoted form,
+		//    not by the actual form.
+		{"\n", []eddefs.ListingShown{}},
+		{"\\n", []eddefs.ListingShown{{"77", ui.Unstyled(`"/foo/\nbar"`)}}},
+	}
+
+	locationWithPrefixMatcher = newProvider([]storedefs.Dir{
+		{"/pinned", pinnedScore},
+		{"/src/github.com/elves/elvish", 300},
+		{"/src/home/xyz", 233},
+		{"/home/dir", 100},
+		{"/foo/\nbar", 77},
+		{"/usr/elves/elvish", 6},
+	},
+		"/home",
+		nil,
+		eval.NewEvaler(),
+		eval.NewBuiltinFn("edit:location:test:match-prefix", matchPrefix))
+
+	locationWithPrefixMatcherTests = []eddefs.ListingProviderFilterTest{
+		{"", []eddefs.ListingShown{
+			{"*", ui.Unstyled("/pinned")},
+			{"300", ui.Unstyled("/src/github.com/elves/elvish")},
+			{"233", ui.Unstyled("/src/home/xyz")},
+			{"100", ui.Unstyled("~/dir")},       // home is abbreviated
+			{"77", ui.Unstyled(`"/foo/\nbar"`)}, // special char is quoted
+			{"6", ui.Unstyled("/usr/elves/elvish")}}},
+		{"/src", []eddefs.ListingShown{
+			{"300", ui.Unstyled("/src/github.com/elves/elvish")},
+			{"233", ui.Unstyled("/src/home/xyz")}}},
+		{"home", []eddefs.ListingShown{}},
+	}
+)
+
+func matchPrefix(fm *eval.Frame, opts eval.RawOptions, pattern string, inputs eval.Inputs) {
+	out := fm.OutputChan()
+	inputs(func(v interface{}) {
+		out <- vals.Bool(strings.HasPrefix(v.(string), pattern))
+	})
+}
+
+func TestProvider(t *testing.T) {
+	eddefs.TestListingProviderFilter(
+		t, "theLocation", theLocation, locationFilterTests)
+	eddefs.TestListingProviderFilter(
+		t, "locationWithPrefixMatcher", locationWithPrefixMatcher, locationWithPrefixMatcherTests)
+}
diff -pruN 0.11+ds1-5/edit/location/location_unix_test.go 0.12+ds1-1/edit/location/location_unix_test.go
--- 0.11+ds1-5/edit/location/location_unix_test.go	1970-01-01 00:00:00.000000000 +0000
+++ 0.12+ds1-1/edit/location/location_unix_test.go	2018-07-11 02:32:25.000000000 +0000
@@ -0,0 +1,34 @@
+// +build !windows,!plan9
+
+package location
+
+import (
+	"testing"
+
+	"github.com/elves/elvish/eval/vals"
+	"github.com/elves/elvish/tt"
+)
+
+var workspaces = vals.MakeMapFromKV(
+	// Pattern is always anchored at beginning; this won't match anything
+	"bad", "bad",
+	// This is a normal pattern.
+	"linux", "/src/linux/[^/]+",
+	// Pattern may match a trailing /, in which case it only matches subdirs
+	"bsd", "/src/bsd/[^/]+/",
+)
+
+func TestMatchWorkspace(t *testing.T) {
+	var nilWsInfo *wsInfo
+
+	tt.Test(t, tt.Fn("matchWorkspace", matchWorkspace), tt.Table{
+		tt.Args("/bad", workspaces).Rets(nilWsInfo),
+		tt.Args("/src/linux/ws1", workspaces).Rets(
+			&wsInfo{"linux", "/src/linux/ws1"}),
+		tt.Args("/src/linux/ws1/dir", workspaces).Rets(
+			&wsInfo{"linux", "/src/linux/ws1"}),
+		tt.Args("/src/bsd/ws1", workspaces).Rets(nilWsInfo),
+		tt.Args("/src/bsd/ws1/dir", workspaces).Rets(
+			&wsInfo{"bsd", "/src/bsd/ws1/"}),
+	})
+}
diff -pruN 0.11+ds1-5/edit/location.go 0.12+ds1-1/edit/location.go
--- 0.11+ds1-5/edit/location.go	2018-01-18 02:03:11.000000000 +0000
+++ 0.12+ds1-1/edit/location.go	1970-01-01 00:00:00.000000000 +0000
@@ -1,195 +0,0 @@
-package edit
-
-import (
-	"bytes"
-	"fmt"
-	"math"
-	"os"
-	"regexp"
-	"strings"
-
-	"github.com/elves/elvish/edit/ui"
-	"github.com/elves/elvish/eval"
-	"github.com/elves/elvish/eval/types"
-	"github.com/elves/elvish/eval/vartypes"
-	"github.com/elves/elvish/parse"
-	"github.com/elves/elvish/store/storedefs"
-	"github.com/elves/elvish/util"
-)
-
-// Location mode.
-
-var _ = registerBuiltins(modeLocation, map[string]func(*Editor){
-	"start": locStart,
-})
-
-// PinnedScore is a special value of Score in storedefs.Dir to represent that the
-// directory is pinned.
-var PinnedScore = math.Inf(1)
-
-type location struct {
-	home     string // The home directory; leave empty if unknown.
-	all      []storedefs.Dir
-	filtered []storedefs.Dir
-}
-
-func newLocation(dirs []storedefs.Dir, home string) *listing {
-	l := newListing(modeLocation, &location{all: dirs, home: home})
-	return &l
-}
-
-func (loc *location) ModeTitle(i int) string {
-	return " LOCATION "
-}
-
-func (*location) CursorOnModeLine() bool {
-	return true
-}
-
-func (loc *location) Len() int {
-	return len(loc.filtered)
-}
-
-func (loc *location) Show(i int) (string, ui.Styled) {
-	var header string
-	score := loc.filtered[i].Score
-	if score == PinnedScore {
-		header = "*"
-	} else {
-		header = fmt.Sprintf("%.0f", score)
-	}
-	return header, ui.Unstyled(showPath(loc.filtered[i].Path, loc.home))
-}
-
-func (loc *location) Filter(filter string) int {
-	loc.filtered = nil
-	pattern := makeLocationFilterPattern(filter)
-	for _, item := range loc.all {
-		if pattern.MatchString(showPath(item.Path, loc.home)) {
-			loc.filtered = append(loc.filtered, item)
-		}
-	}
-
-	if len(loc.filtered) == 0 {
-		return -1
-	}
-	return 0
-}
-
-func showPath(path, home string) string {
-	if home != "" && path == home {
-		return "~"
-	} else if home != "" && strings.HasPrefix(path, home+"/") {
-		return "~/" + parse.Quote(path[len(home)+1:])
-	} else {
-		return parse.Quote(path)
-	}
-}
-
-var emptyRegexp = regexp.MustCompile("")
-
-func makeLocationFilterPattern(s string) *regexp.Regexp {
-	var b bytes.Buffer
-	b.WriteString(".*")
-	segs := strings.Split(s, "/")
-	for i, seg := range segs {
-		if i > 0 {
-			b.WriteString(".*/.*")
-		}
-		b.WriteString(regexp.QuoteMeta(seg))
-	}
-	b.WriteString(".*")
-	p, err := regexp.Compile(b.String())
-	if err != nil {
-		logger.Printf("failed to compile regexp %q: %v", b.String(), err)
-		return emptyRegexp
-	}
-	return p
-}
-
-// Editor interface.
-
-func (loc *location) Accept(i int, ed *Editor) {
-	err := eval.Chdir(loc.filtered[i].Path, ed.daemon)
-	if err != nil {
-		ed.Notify("%v", err)
-	}
-	ed.mode = &ed.insert
-}
-
-func locStart(ed *Editor) {
-	if ed.daemon == nil {
-		ed.Notify("%v", ErrStoreOffline)
-		return
-	}
-
-	// Pinned directories are also blacklisted to prevent them from showing up
-	// twice.
-	black := convertListsToSet(ed.locHidden(), ed.locPinned())
-	pwd, err := os.Getwd()
-	if err == nil {
-		black[pwd] = struct{}{}
-	}
-	stored, err := ed.daemon.Dirs(black)
-	if err != nil {
-		ed.Notify("store error: %v", err)
-		return
-	}
-
-	// Concatenate pinned and stored dirs, pinned first.
-	pinned := convertListToDirs(ed.locPinned())
-	dirs := make([]storedefs.Dir, len(pinned)+len(stored))
-	copy(dirs, pinned)
-	copy(dirs[len(pinned):], stored)
-
-	// Drop the error. When there is an error, home is "", which is used to
-	// signify "no home known" in location.
-	home, _ := util.GetHome("")
-	ed.mode = newLocation(dirs, home)
-}
-
-// convertListToDirs converts a list of strings to []storedefs.Dir. It uses the
-// special score of PinnedScore to signify that the directory is pinned.
-func convertListToDirs(li types.List) []storedefs.Dir {
-	pinned := make([]storedefs.Dir, 0, li.Len())
-	// XXX(xiaq): silently drops non-string items.
-	li.Iterate(func(v types.Value) bool {
-		if s, ok := v.(types.String); ok {
-			pinned = append(pinned, storedefs.Dir{string(s), PinnedScore})
-		}
-		return true
-	})
-	return pinned
-}
-
-func convertListsToSet(lis ...types.List) map[string]struct{} {
-	set := make(map[string]struct{})
-	// XXX(xiaq): silently drops non-string items.
-	for _, li := range lis {
-		li.Iterate(func(v types.Value) bool {
-			if s, ok := v.(types.String); ok {
-				set[string(s)] = struct{}{}
-			}
-			return true
-		})
-	}
-	return set
-}
-
-// Variables.
-
-var _ = RegisterVariable("loc-hidden", func() vartypes.Variable {
-	return vartypes.NewValidatedPtr(types.EmptyList, vartypes.ShouldBeList)
-})
-
-func (ed *Editor) locHidden() types.List {
-	return ed.variables["loc-hidden"].Get().(types.List)
-}
-
-var _ = RegisterVariable("loc-pinned", func() vartypes.Variable {
-	return vartypes.NewValidatedPtr(types.EmptyList, vartypes.ShouldBeList)
-})
-
-func (ed *Editor) locPinned() types.List {
-	return ed.variables["loc-pinned"].Get().(types.List)
-}
diff -pruN 0.11+ds1-5/edit/location_test.go 0.12+ds1-1/edit/location_test.go
--- 0.11+ds1-5/edit/location_test.go	2018-01-18 02:03:11.000000000 +0000
+++ 0.12+ds1-1/edit/location_test.go	1970-01-01 00:00:00.000000000 +0000
@@ -1,50 +0,0 @@
-package edit
-
-import (
-	"testing"
-
-	"github.com/elves/elvish/edit/ui"
-	"github.com/elves/elvish/store/storedefs"
-)
-
-var (
-	theLocation = newLocation([]storedefs.Dir{
-		{"/pinned", PinnedScore},
-		{"/src/github.com/elves/elvish", 300},
-		{"/src/home/xyz", 233},
-		{"/home/dir", 100},
-		{"/foo/\nbar", 77},
-		{"/usr/elves/elvish", 6},
-	}, "/home")
-
-	locationFilterTests = []listingFilterTestCases{
-		{"", []shown{
-			{"*", ui.Unstyled("/pinned")},
-			{"300", ui.Unstyled("/src/github.com/elves/elvish")},
-			{"233", ui.Unstyled("/src/home/xyz")},
-			{"100", ui.Unstyled("~/dir")},       // home is abbreviated
-			{"77", ui.Unstyled(`"/foo/\nbar"`)}, // special char is quoted
-			{"6", ui.Unstyled("/usr/elves/elvish")}}},
-		{"/s", []shown{
-			{"300", ui.Unstyled("/src/github.com/elves/elvish")},
-			{"233", ui.Unstyled("/src/home/xyz")},
-			{"6", ui.Unstyled("/usr/elves/elvish")}}},
-		{"/e/e", []shown{
-			{"300", ui.Unstyled("/src/github.com/elves/elvish")},
-			{"6", ui.Unstyled("/usr/elves/elvish")}}},
-		{"x", []shown{{"233", ui.Unstyled("/src/home/xyz")}}},
-		// Matchers operate on the displayed text, not the actual path.
-		// 1. Home directory is abbreviated to ~, and is matched by ~, but not by
-		//    the actual path.
-		{"~", []shown{{"100", ui.Unstyled("~/dir")}}},
-		{"home", []shown{{"233", ui.Unstyled("/src/home/xyz")}}},
-		// 2. Special characters are quoted, and are matched by the quoted form,
-		//    not by the actual form.
-		{"\n", []shown{}},
-		{"\\n", []shown{{"77", ui.Unstyled(`"/foo/\nbar"`)}}},
-	}
-)
-
-func TestLocation(t *testing.T) {
-	testListingFilter(t, "theLocation", theLocation, locationFilterTests)
-}
diff -pruN 0.11+ds1-5/edit/matcher.go 0.12+ds1-1/edit/matcher.go
--- 0.11+ds1-5/edit/matcher.go	2018-01-18 02:03:11.000000000 +0000
+++ 0.12+ds1-1/edit/matcher.go	1970-01-01 00:00:00.000000000 +0000
@@ -1,91 +0,0 @@
-package edit
-
-import (
-	"errors"
-	"strings"
-
-	"github.com/elves/elvish/eval"
-	"github.com/elves/elvish/eval/types"
-	"github.com/elves/elvish/eval/vartypes"
-	"github.com/elves/elvish/util"
-	"github.com/xiaq/persistent/hashmap"
-)
-
-var (
-	errIncorrectNumOfResults    = errors.New("matcher must return a bool for each candidate")
-	errMatcherMustBeFn          = errors.New("matcher must be a function")
-	errMatcherInputMustBeString = errors.New("matcher input must be string")
-)
-
-var (
-	matchPrefix = &eval.BuiltinFn{
-		"edit:match-prefix", wrapMatcher(strings.HasPrefix)}
-	matchSubstr = &eval.BuiltinFn{
-		"edit:match-substr", wrapMatcher(strings.Contains)}
-	matchSubseq = &eval.BuiltinFn{
-		"edit:match-subseq", wrapMatcher(util.HasSubseq)}
-	matchers = []*eval.BuiltinFn{
-		matchPrefix,
-		matchSubstr,
-		matchSubseq,
-	}
-
-	_ = RegisterVariable("-matcher", func() vartypes.Variable {
-		m := hashmap.Empty.Assoc(
-			// Fallback matcher uses empty string as key
-			types.String(""), matchPrefix)
-		return vartypes.NewValidatedPtr(types.NewMap(m), vartypes.ShouldBeMap)
-	})
-)
-
-func (ed *Editor) lookupMatcher(name string) (eval.Fn, bool) {
-	m := ed.variables["-matcher"].Get().(types.Map)
-	if !m.HasKey(types.String(name)) {
-		// Use fallback matcher
-		name = ""
-	}
-	matcher, ok := m.IndexOne(types.String(name)).(eval.Fn)
-	return matcher, ok
-}
-
-func wrapMatcher(matcher func(s, p string) bool) eval.BuiltinFnImpl {
-	return func(ec *eval.Frame,
-		args []types.Value, opts map[string]types.Value) {
-
-		var pattern types.String
-		iterate := eval.ScanArgsOptionalInput(ec, args, &pattern)
-		var options struct {
-			IgnoreCase bool
-			SmartCase  bool
-		}
-		eval.ScanOptsToStruct(opts, &options)
-		switch {
-		case options.IgnoreCase && options.SmartCase:
-			throwf("-ignore-case and -smart-case cannot be used together")
-		case options.IgnoreCase:
-			innerMatcher := matcher
-			matcher = func(s, p string) bool {
-				return innerMatcher(strings.ToLower(s), strings.ToLower(p))
-			}
-		case options.SmartCase:
-			innerMatcher := matcher
-			matcher = func(s, p string) bool {
-				if p == strings.ToLower(p) {
-					// Ignore case is pattern is all lower case.
-					return innerMatcher(strings.ToLower(s), p)
-				} else {
-					return innerMatcher(s, p)
-				}
-			}
-		}
-
-		out := ec.OutputChan()
-		iterate(func(v types.Value) {
-			s, ok := v.(types.String)
-			if !ok {
-				throw(errMatcherInputMustBeString)
-			}
-			out <- types.Bool(matcher(string(s), string(pattern)))
-		})
-	}
-}
diff -pruN 0.11+ds1-5/edit/max_height.go 0.12+ds1-1/edit/max_height.go
--- 0.11+ds1-5/edit/max_height.go	2018-01-18 02:03:11.000000000 +0000
+++ 0.12+ds1-1/edit/max_height.go	1970-01-01 00:00:00.000000000 +0000
@@ -1,22 +0,0 @@
-package edit
-
-import (
-	"math"
-	"strconv"
-
-	"github.com/elves/elvish/eval/types"
-	"github.com/elves/elvish/eval/vartypes"
-	"github.com/elves/elvish/util"
-)
-
-var _ = RegisterVariable("max-height", func() vartypes.Variable {
-	return vartypes.NewValidatedPtr(types.String("+Inf"), vartypes.ShouldBeNumber)
-})
-
-func (ed *Editor) maxHeight() int {
-	f, _ := strconv.ParseFloat(string(ed.variables["max-height"].Get().(types.String)), 64)
-	if math.IsInf(f, 1) {
-		return util.MaxInt
-	}
-	return int(f)
-}
diff -pruN 0.11+ds1-5/edit/minmax.go 0.12+ds1-1/edit/minmax.go
--- 0.11+ds1-5/edit/minmax.go	2018-01-18 02:03:11.000000000 +0000
+++ 0.12+ds1-1/edit/minmax.go	1970-01-01 00:00:00.000000000 +0000
@@ -1,15 +0,0 @@
-package edit
-
-func min(a, b int) int {
-	if a <= b {
-		return a
-	}
-	return b
-}
-
-func max(a, b int) int {
-	if a >= b {
-		return a
-	}
-	return b
-}
diff -pruN 0.11+ds1-5/edit/mode.go 0.12+ds1-1/edit/mode.go
--- 0.11+ds1-5/edit/mode.go	2018-01-18 02:03:11.000000000 +0000
+++ 0.12+ds1-1/edit/mode.go	1970-01-01 00:00:00.000000000 +0000
@@ -1,48 +0,0 @@
-package edit
-
-import (
-	"github.com/elves/elvish/edit/ui"
-	"github.com/elves/elvish/eval"
-	"github.com/elves/elvish/eval/vartypes"
-)
-
-// Names of modes, used for subnamespaces of edit:.
-const (
-	modeInsert         = "insert"
-	modeRawInsert      = "raw-insert"
-	modeCommand        = "command"
-	modeCompletion     = "completion"
-	modeNavigation     = "navigation"
-	modeHistory        = "history"
-	modeHistoryListing = "histlist"
-	modeLastCmd        = "lastcmd"
-	modeLocation       = "location"
-	modeListing        = "listing" // A "super mode" for histlist, lastcmd, loc
-	modeNarrow         = "narrow"  // a listing mode fork to be extended by scripts
-)
-
-// Mode is an editor mode.
-type Mode interface {
-	ModeLine() ui.Renderer
-	Binding(map[string]vartypes.Variable, ui.Key) eval.Fn
-}
-
-// CursorOnModeLiner is an optional interface that modes can implement. If a
-// mode does and the method returns true, the cursor is placed on the modeline
-// when that mode is active.
-type CursorOnModeLiner interface {
-	CursorOnModeLine() bool
-}
-
-// Lister is a mode with a listing.
-type Lister interface {
-	List(maxHeight int) ui.Renderer
-}
-
-// ListRenderer is a mode with a listing that handles the rendering itself.
-// NOTE(xiaq): This interface is being deprecated in favor of Lister.
-type ListRenderer interface {
-	// ListRender renders the listing under the given constraint of width and
-	// maximum height. It returns a rendered buffer.
-	ListRender(width, maxHeight int) *ui.Buffer
-}
diff -pruN 0.11+ds1-5/edit/narrow.go 0.12+ds1-1/edit/narrow.go
--- 0.11+ds1-5/edit/narrow.go	2018-01-18 02:03:11.000000000 +0000
+++ 0.12+ds1-1/edit/narrow.go	1970-01-01 00:00:00.000000000 +0000
@@ -1,524 +0,0 @@
-package edit
-
-import (
-	"container/list"
-	"errors"
-	"strconv"
-	"strings"
-	"unicode/utf8"
-
-	"github.com/elves/elvish/edit/ui"
-	"github.com/elves/elvish/eval"
-	"github.com/elves/elvish/eval/types"
-	"github.com/elves/elvish/eval/vartypes"
-	"github.com/xiaq/persistent/hashmap"
-)
-
-var _ = registerBuiltins(modeNarrow, map[string]func(*Editor){
-	"up":         func(ed *Editor) { getNarrow(ed).up(false) },
-	"up-cycle":   func(ed *Editor) { getNarrow(ed).up(true) },
-	"page-up":    func(ed *Editor) { getNarrow(ed).pageUp() },
-	"down":       func(ed *Editor) { getNarrow(ed).down(false) },
-	"down-cycle": func(ed *Editor) { getNarrow(ed).down(true) },
-	"page-down":  func(ed *Editor) { getNarrow(ed).pageDown() },
-	"backspace":  func(ed *Editor) { getNarrow(ed).backspace() },
-	"accept":     func(ed *Editor) { getNarrow(ed).accept(ed) },
-	"accept-close": func(ed *Editor) {
-		getNarrow(ed).accept(ed)
-		insertStart(ed)
-	},
-	"toggle-ignore-duplication": func(ed *Editor) {
-		l := getNarrow(ed)
-		l.opts.IgnoreDuplication = !l.opts.IgnoreDuplication
-		l.refresh()
-	},
-	"toggle-ignore-case": func(ed *Editor) {
-		l := getNarrow(ed)
-		l.opts.IgnoreCase = !l.opts.IgnoreCase
-		l.refresh()
-	},
-	"default": func(ed *Editor) { getNarrow(ed).defaultBinding(ed) },
-})
-
-// narrow implements a listing mode that supports the notion of selecting an
-// entry and filtering entries.
-type narrow struct {
-	name        string
-	selected    int
-	filter      string
-	pagesize    int
-	headerWidth int
-
-	placehold string
-	source    func() []narrowItem
-	action    func(*Editor, narrowItem)
-	match     func(string, string) bool
-	filtered  []narrowItem
-	opts      narrowOptions
-}
-
-func (l *narrow) Binding(m map[string]vartypes.Variable, k ui.Key) eval.Fn {
-	if l.opts.bindingMap != nil {
-		if f, ok := l.opts.bindingMap[k]; ok {
-			return f
-		}
-	}
-
-	return getBinding(m[modeNarrow], k)
-}
-
-func (l *narrow) ModeLine() ui.Renderer {
-	ml := l.opts.Modeline
-	var opt []string
-	if l.opts.AutoCommit {
-		opt = append(opt, "A")
-	}
-	if l.opts.IgnoreCase {
-		opt = append(opt, "C")
-	}
-	if l.opts.IgnoreDuplication {
-		opt = append(opt, "D")
-	}
-	if len(opt) != 0 {
-		ml += "[" + strings.Join(opt, " ") + "]"
-	}
-	return modeLineRenderer{ml, l.filter}
-}
-
-func (l *narrow) CursorOnModeLine() bool {
-	return true
-}
-
-func (l *narrow) List(maxHeight int) ui.Renderer {
-	if l.opts.MaxLines > 0 && l.opts.MaxLines < maxHeight {
-		maxHeight = l.opts.MaxLines
-	}
-
-	if l.filtered == nil {
-		l.refresh()
-	}
-	n := len(l.filtered)
-	if n == 0 {
-		return placeholderRenderer(l.placehold)
-	}
-
-	// Collect the entries to show. We start from the selected entry and extend
-	// in both directions alternatingly. The entries are split into lines and
-	// then collected in a list.
-	low := l.selected
-	if low == -1 {
-		low = 0
-	}
-	high := low
-	height := 0
-	var listOfLines list.List
-	getEntry := func(i int) []ui.Styled {
-		display := l.filtered[i].Display()
-		lines := strings.Split(display.Text, "\n")
-		styles := display.Styles
-		if i == l.selected {
-			styles = append(styles, styleForSelected...)
-		}
-		styleds := make([]ui.Styled, len(lines))
-		for i, line := range lines {
-			styleds[i] = ui.Styled{line, styles}
-		}
-		return styleds
-	}
-	// We start by extending high, so that the first entry to include is
-	// l.selected.
-	extendLow := false
-	lastShownIncomplete := false
-	for height < maxHeight && !(low == 0 && high == n) {
-		var i int
-		if (extendLow && low > 0) || high == n {
-			low--
-
-			entry := getEntry(low)
-			// Prepend at most the last (height - maxHeight) lines.
-			for i = len(entry) - 1; i >= 0 && height < maxHeight; i-- {
-				listOfLines.PushFront(entry[i])
-				height++
-			}
-			if i >= 0 {
-				lastShownIncomplete = true
-			}
-		} else {
-			entry := getEntry(high)
-			// Append at most the first (height - maxHeight) lines.
-			for i = 0; i < len(entry) && height < maxHeight; i++ {
-				listOfLines.PushBack(entry[i])
-				height++
-			}
-			if i < len(entry) {
-				lastShownIncomplete = true
-			}
-
-			high++
-		}
-		extendLow = !extendLow
-	}
-
-	l.pagesize = high - low
-
-	// Convert the List to a slice.
-	lines := make([]ui.Styled, 0, listOfLines.Len())
-	for p := listOfLines.Front(); p != nil; p = p.Next() {
-		lines = append(lines, p.Value.(ui.Styled))
-	}
-
-	ls := listingRenderer{lines}
-	if low > 0 || high < n || lastShownIncomplete {
-		// Need scrollbar
-		return listingWithScrollBarRenderer{ls, n, low, high, height}
-	}
-	return ls
-}
-
-func (l *narrow) refresh() {
-	var candidates []narrowItem
-	if l.source != nil {
-		candidates = l.source()
-	}
-	l.filtered = make([]narrowItem, 0, len(candidates))
-
-	filter := l.filter
-	if l.opts.IgnoreCase {
-		filter = strings.ToLower(filter)
-	}
-
-	set := make(map[string]struct{})
-
-	for _, item := range candidates {
-		text := item.FilterText()
-		s := text
-		if l.opts.IgnoreCase {
-			s = strings.ToLower(s)
-		}
-		if !l.match(s, filter) {
-			continue
-		}
-		if l.opts.IgnoreDuplication {
-			if _, ok := set[text]; ok {
-				continue
-			}
-			set[text] = struct{}{}
-		}
-		l.filtered = append(l.filtered, item)
-	}
-
-	if l.opts.KeepBottom {
-		l.selected = len(l.filtered) - 1
-	} else {
-		l.selected = 0
-	}
-}
-
-func (l *narrow) changeFilter(newfilter string) {
-	l.filter = newfilter
-	l.refresh()
-}
-
-func (l *narrow) backspace() bool {
-	_, size := utf8.DecodeLastRuneInString(l.filter)
-	if size > 0 {
-		l.changeFilter(l.filter[:len(l.filter)-size])
-		return true
-	}
-	return false
-}
-
-func (l *narrow) up(cycle bool) {
-	n := len(l.filtered)
-	if n == 0 {
-		return
-	}
-	l.selected--
-	if l.selected == -1 {
-		if cycle {
-			l.selected += n
-		} else {
-			l.selected++
-		}
-	}
-}
-
-func (l *narrow) pageUp() {
-	n := len(l.filtered)
-	if n == 0 {
-		return
-	}
-	l.selected -= l.pagesize
-	if l.selected < 0 {
-		l.selected = 0
-	}
-}
-
-func (l *narrow) down(cycle bool) {
-	n := len(l.filtered)
-	if n == 0 {
-		return
-	}
-	l.selected++
-	if l.selected == n {
-		if cycle {
-			l.selected -= n
-		} else {
-			l.selected--
-		}
-	}
-}
-
-func (l *narrow) pageDown() {
-	n := len(l.filtered)
-	if n == 0 {
-		return
-	}
-	l.selected += l.pagesize
-	if l.selected >= n {
-		l.selected = n - 1
-	}
-}
-
-func (l *narrow) accept(ed *Editor) {
-	if l.selected >= 0 {
-		l.action(ed, l.filtered[l.selected])
-	}
-}
-
-func (l *narrow) handleFilterKey(ed *Editor) bool {
-	k := ed.lastKey
-	if likeChar(k) {
-		l.changeFilter(l.filter + string(k.Rune))
-		if len(l.filtered) == 1 && l.opts.AutoCommit {
-			l.accept(ed)
-			insertStart(ed)
-		}
-		return true
-	}
-	return false
-}
-
-func (l *narrow) defaultBinding(ed *Editor) {
-	if !l.handleFilterKey(ed) {
-		insertStart(ed)
-		ed.setAction(reprocessKey)
-	}
-}
-
-var errNotNarrow = errors.New("not in a narrow mode")
-
-func getNarrow(ed *Editor) *narrow {
-	if l, ok := ed.mode.(*narrow); ok {
-		return l
-	} else {
-		throw(errNotNarrow)
-		panic("unreachable")
-	}
-}
-
-type narrowItem interface {
-	types.Value
-	Display() ui.Styled
-	Content() string
-	FilterText() string
-}
-
-type narrowOptions struct {
-	AutoCommit        bool
-	Bindings          types.Map
-	IgnoreDuplication bool
-	IgnoreCase        bool
-	KeepBottom        bool
-	MaxLines          int
-	Modeline          string
-
-	bindingMap map[ui.Key]eval.Fn
-}
-
-type narrowItemString struct {
-	types.String
-}
-
-func (s *narrowItemString) Content() string {
-	return string(s.String)
-}
-
-func (s *narrowItemString) Display() ui.Styled {
-	return ui.Unstyled(string(s.String))
-}
-
-func (s *narrowItemString) FilterText() string {
-	return s.Content()
-}
-
-type narrowItemComplex struct {
-	types.Map
-}
-
-func (c *narrowItemComplex) Content() string {
-	key := types.String("content")
-	if !c.Map.HasKey(key) {
-		return ""
-	}
-	if s, ok := c.Map.IndexOne(key).(types.String); !ok {
-		return ""
-	} else {
-		return s.String()
-	}
-}
-
-// TODO: add style
-func (c *narrowItemComplex) Display() ui.Styled {
-	key := types.String("display")
-	if !c.Map.HasKey(key) {
-		return ui.Unstyled("")
-	}
-	if s, ok := c.Map.IndexOne(key).(types.String); !ok {
-		return ui.Unstyled("")
-	} else {
-		return ui.Unstyled(s.String())
-	}
-}
-
-func (c *narrowItemComplex) FilterText() string {
-	key := types.String("filter-text")
-	if c.Map.HasKey(key) {
-		return c.Map.IndexOne(key).(types.String).String()
-	}
-	return c.Content()
-}
-
-func NarrowRead(ec *eval.Frame, args []types.Value, opts map[string]types.Value) {
-	var source, action eval.Fn
-	l := &narrow{
-		opts: narrowOptions{
-			Bindings: types.NewMap(hashmap.Empty),
-		},
-	}
-
-	eval.ScanArgs(args, &source, &action)
-	eval.ScanOptsToStruct(opts, &l.opts)
-
-	l.opts.Bindings.IterateKey(func(k types.Value) bool {
-		key := ui.ToKey(k)
-		f := l.opts.Bindings.IndexOne(k)
-		maybeThrow(eval.ShouldBeFn(f))
-		if l.opts.bindingMap == nil {
-			l.opts.bindingMap = make(map[ui.Key]eval.Fn)
-		}
-		l.opts.bindingMap[key] = f.(eval.Fn)
-		return true
-	})
-
-	l.source = narrowGetSource(ec, source)
-	l.action = func(ed *Editor, item narrowItem) {
-		ed.CallFn(action, item)
-	}
-	// TODO: user customize varible
-	l.match = strings.Contains
-
-	l.changeFilter("")
-	ed := ec.Editor.(*Editor)
-	ed.mode = l
-}
-
-func narrowGetSource(ec *eval.Frame, source eval.Fn) func() []narrowItem {
-	return func() []narrowItem {
-		ed := ec.Editor.(*Editor)
-		vs, err := ec.PCaptureOutput(source, eval.NoArgs, eval.NoOpts)
-		if err != nil {
-			ed.Notify(err.Error())
-			return nil
-		}
-		var lis []narrowItem
-		for _, v := range vs {
-			switch raw := v.(type) {
-			case types.String:
-				lis = append(lis, &narrowItemString{raw})
-			case types.Map:
-				lis = append(lis, &narrowItemComplex{raw})
-			}
-		}
-		return lis
-	}
-}
-
-func CommandHistory(ec *eval.Frame, args []types.Value, opts map[string]types.Value) {
-	var (
-		rest              []int
-		limit, start, end int
-	)
-
-	eval.ScanArgsVariadic(args, &rest)
-	eval.TakeNoOpt(opts)
-
-	out := ec.OutputChan()
-	ed := ec.Editor.(*Editor)
-	cmds, err := ed.historyFuser.AllCmds()
-	if err != nil {
-		return
-	}
-
-	if len(rest) > 0 {
-		limit = rest[0]
-	}
-
-	total := len(cmds)
-	switch {
-	case limit > 0:
-		start = 0
-		end = limit
-		if limit > total {
-			end = total
-		}
-	case limit < 0:
-		start = limit + total
-		if start < 0 {
-			start = 0
-		}
-		end = total
-	default:
-		start = 0
-		end = total
-	}
-
-	for i := start; i < end; i++ {
-		out <- types.MakeMap(map[types.Value]types.Value{
-			types.String("id"):  types.String(strconv.Itoa(i)),
-			types.String("cmd"): types.String(cmds[i]),
-		})
-	}
-}
-
-func InsertAtDot(ec *eval.Frame, args []types.Value, opts map[string]types.Value) {
-	var text types.String
-
-	eval.ScanArgs(args, &text)
-	eval.TakeNoOpt(opts)
-
-	ed := ec.Editor.(*Editor)
-	ed.insertAtDot(text.String())
-}
-
-func ReplaceInput(ec *eval.Frame, args []types.Value, opts map[string]types.Value) {
-	var text types.String
-
-	eval.ScanArgs(args, &text)
-	eval.TakeNoOpt(opts)
-
-	ed := ec.Editor.(*Editor)
-	ed.buffer = text.String()
-}
-
-func Wordify(ec *eval.Frame, args []types.Value, opts map[string]types.Value) {
-	var text types.String
-
-	eval.ScanArgs(args, &text)
-	eval.TakeNoOpt(opts)
-
-	out := ec.OutputChan()
-	for _, s := range wordify(text.String()) {
-		out <- types.String(s)
-	}
-}
diff -pruN 0.11+ds1-5/edit/navigation.go 0.12+ds1-1/edit/navigation.go
--- 0.11+ds1-5/edit/navigation.go	2018-01-18 02:03:11.000000000 +0000
+++ 0.12+ds1-1/edit/navigation.go	1970-01-01 00:00:00.000000000 +0000
@@ -1,432 +0,0 @@
-package edit
-
-import (
-	"errors"
-	"os"
-	"path"
-	"sort"
-	"strings"
-	"unicode/utf8"
-
-	"github.com/elves/elvish/edit/lscolors"
-	"github.com/elves/elvish/edit/ui"
-	"github.com/elves/elvish/eval"
-	"github.com/elves/elvish/eval/vartypes"
-	"github.com/elves/elvish/parse"
-	"github.com/elves/elvish/util"
-)
-
-// Navigation subsystem.
-
-// Interface.
-
-var _ = registerBuiltins(modeNavigation, map[string]func(*Editor){
-	"start":                    navStart,
-	"up":                       navUp,
-	"down":                     navDown,
-	"page-up":                  navPageUp,
-	"page-down":                navPageDown,
-	"left":                     navLeft,
-	"right":                    navRight,
-	"file-preview-up":          navFilePreviewUp,
-	"file-preview-down":        navFilePreviewDown,
-	"trigger-shown-hidden":     navTriggerShowHidden,
-	"trigger-filter":           navTriggerFilter,
-	"insert-selected":          navInsertSelected,
-	"insert-selected-and-quit": navInsertSelectedAndQuit,
-	"default":                  navDefault,
-})
-
-type navigation struct {
-	current    *navColumn
-	parent     *navColumn
-	preview    navPreview
-	showHidden bool
-	filtering  bool
-	filter     string
-	chdir      func(string) error
-}
-
-type navPreview interface {
-	FullWidth(int) int
-	List(int) ui.Renderer
-}
-
-func (*navigation) Binding(m map[string]vartypes.Variable, k ui.Key) eval.Fn {
-	return getBinding(m[modeNavigation], k)
-}
-
-func (n *navigation) ModeLine() ui.Renderer {
-	title := " NAVIGATING "
-	if n.showHidden {
-		title += "(show hidden) "
-	}
-	return modeLineRenderer{title, n.filter}
-}
-
-func (n *navigation) CursorOnModeLine() bool {
-	return n.filtering
-}
-
-func navStart(ed *Editor) {
-	initNavigation(&ed.navigation, ed)
-	ed.mode = &ed.navigation
-}
-
-func navUp(ed *Editor) {
-	ed.navigation.prev()
-}
-
-func navDown(ed *Editor) {
-	ed.navigation.next()
-}
-
-func navPageUp(ed *Editor) {
-	ed.navigation.current.pageUp()
-	ed.navigation.refresh()
-}
-
-func navPageDown(ed *Editor) {
-	ed.navigation.current.pageDown()
-	ed.navigation.refresh()
-}
-
-func navLeft(ed *Editor) {
-	ed.navigation.ascend()
-}
-
-func navRight(ed *Editor) {
-	ed.navigation.descend()
-}
-
-func navFilePreviewUp(ed *Editor) {
-	fp, ok := ed.navigation.preview.(*navFilePreview)
-	if ok {
-		if fp.beginLine > 0 {
-			fp.beginLine--
-		}
-	}
-}
-
-func navFilePreviewDown(ed *Editor) {
-	fp, ok := ed.navigation.preview.(*navFilePreview)
-	if ok {
-		if fp.beginLine < len(fp.lines)-1 {
-			fp.beginLine++
-		}
-	}
-}
-
-func navTriggerShowHidden(ed *Editor) {
-	ed.navigation.showHidden = !ed.navigation.showHidden
-	ed.navigation.refresh()
-}
-
-func navTriggerFilter(ed *Editor) {
-	ed.navigation.filtering = !ed.navigation.filtering
-}
-
-func navInsertSelected(ed *Editor) {
-	ed.insertAtDot(parse.Quote(ed.navigation.current.selectedName()) + " ")
-}
-
-func navInsertSelectedAndQuit(ed *Editor) {
-	ed.insertAtDot(parse.Quote(ed.navigation.current.selectedName()) + " ")
-	ed.mode = &ed.insert
-}
-
-func navDefault(ed *Editor) {
-	// Use key binding for insert mode without exiting nigation mode.
-	k := ed.lastKey
-	n := &ed.navigation
-	if n.filtering && likeChar(k) {
-		n.filter += k.String()
-		n.refreshCurrent()
-		n.refreshDirPreview()
-	} else if n.filtering && k == (ui.Key{ui.Backspace, 0}) {
-		_, size := utf8.DecodeLastRuneInString(n.filter)
-		if size > 0 {
-			n.filter = n.filter[:len(n.filter)-size]
-			n.refreshCurrent()
-			n.refreshDirPreview()
-		}
-	} else {
-		ed.CallFn(getBinding(ed.bindings[modeInsert], k))
-	}
-}
-
-// Implementation.
-// TODO(xiaq): Remember which file was selected in each directory.
-
-var errorEmptyCwd = errors.New("current directory is empty")
-
-func initNavigation(n *navigation, ed *Editor) {
-	*n = navigation{chdir: func(dir string) error {
-		return eval.Chdir(dir, ed.daemon)
-	}}
-	n.refresh()
-}
-
-func (n *navigation) maintainSelected(name string) {
-	n.current.selected = 0
-	for i, s := range n.current.candidates {
-		if s.Text > name {
-			break
-		}
-		n.current.selected = i
-	}
-}
-
-func (n *navigation) refreshCurrent() {
-	selectedName := n.current.selectedName()
-	all, err := n.loaddir(".")
-	if err != nil {
-		n.current = newErrNavColumn(err)
-		return
-	}
-	// Try to select the old selected file.
-	// XXX(xiaq): This would break when we support alternative ordering.
-	n.current = newNavColumn(all, func(i int) bool {
-		return i == 0 || all[i].Text <= selectedName
-	})
-	n.current.changeFilter(n.filter)
-	n.maintainSelected(selectedName)
-}
-
-func (n *navigation) refreshParent() {
-	wd, err := os.Getwd()
-	if err != nil {
-		n.parent = newErrNavColumn(err)
-		return
-	}
-	if wd == "/" {
-		n.parent = newNavColumn(nil, nil)
-	} else {
-		all, err := n.loaddir("..")
-		if err != nil {
-			n.parent = newErrNavColumn(err)
-			return
-		}
-		cwd, err := os.Stat(".")
-		if err != nil {
-			n.parent = newErrNavColumn(err)
-			return
-		}
-		n.parent = newNavColumn(all, func(i int) bool {
-			d, _ := os.Lstat("../" + all[i].Text)
-			return os.SameFile(d, cwd)
-		})
-	}
-}
-
-func (n *navigation) refreshDirPreview() {
-	if n.current.selected != -1 {
-		name := n.current.selectedName()
-		fi, err := os.Stat(name)
-		if err != nil {
-			n.preview = newErrNavColumn(err)
-			return
-		}
-		if fi.Mode().IsDir() {
-			all, err := n.loaddir(name)
-			if err != nil {
-				n.preview = newErrNavColumn(err)
-				return
-			}
-			n.preview = newNavColumn(all, func(int) bool { return false })
-		} else {
-			n.preview = makeNavFilePreview(name)
-		}
-	} else {
-		n.preview = nil
-	}
-}
-
-// refresh rereads files in current and parent directories and maintains the
-// selected file if possible.
-func (n *navigation) refresh() {
-	n.refreshCurrent()
-	n.refreshParent()
-	n.refreshDirPreview()
-}
-
-// ascend changes current directory to the parent.
-// TODO(xiaq): navigation.{ascend descend} bypasses the cd builtin. This can be
-// problematic if cd acquires more functionality (e.g. trigger a hook).
-func (n *navigation) ascend() error {
-	wd, err := os.Getwd()
-	if err != nil {
-		return err
-	}
-	if wd == "/" {
-		return nil
-	}
-
-	name := n.parent.selectedName()
-	err = os.Chdir("..")
-	if err != nil {
-		return err
-	}
-	n.filter = ""
-	n.refresh()
-	n.maintainSelected(name)
-	// XXX Refresh dir preview again. We should perhaps not have used refresh
-	// above.
-	n.refreshDirPreview()
-	return nil
-}
-
-// descend changes current directory to the selected file, if it is a
-// directory.
-func (n *navigation) descend() error {
-	if n.current.selected == -1 {
-		return errorEmptyCwd
-	}
-	name := n.current.selectedName()
-	err := n.chdir(name)
-	if err != nil {
-		return err
-	}
-	n.filter = ""
-	n.current.selected = -1
-	n.refresh()
-	n.refreshDirPreview()
-	return nil
-}
-
-// prev selects the previous file.
-func (n *navigation) prev() {
-	if n.current.selected > 0 {
-		n.current.selected--
-	}
-	n.refresh()
-}
-
-// next selects the next file.
-func (n *navigation) next() {
-	if n.current.selected != -1 && n.current.selected < len(n.current.candidates)-1 {
-		n.current.selected++
-	}
-	n.refresh()
-}
-
-func (n *navigation) loaddir(dir string) ([]ui.Styled, error) {
-	f, err := os.Open(dir)
-	if err != nil {
-		return nil, err
-	}
-	names, err := f.Readdirnames(-1)
-	if err != nil {
-		return nil, err
-	}
-	sort.Strings(names)
-
-	var all []ui.Styled
-	lsColor := lscolors.GetColorist()
-	for _, name := range names {
-		if n.showHidden || name[0] != '.' {
-			all = append(all, ui.Styled{name,
-				ui.StylesFromString(lsColor.GetStyle(path.Join(dir, name)))})
-		}
-	}
-
-	return all, nil
-}
-
-func (n *navigation) List(maxHeight int) ui.Renderer {
-	return makeNavRenderer(
-		maxHeight,
-		n.parent.FullWidth(maxHeight),
-		n.current.FullWidth(maxHeight),
-		n.preview.FullWidth(maxHeight),
-		n.parent.List(maxHeight),
-		n.current.List(maxHeight),
-		n.preview.List(maxHeight),
-	)
-}
-
-// navColumn is a column in the navigation layout.
-type navColumn struct {
-	listing
-	all        []ui.Styled
-	candidates []ui.Styled
-	// selected int
-	err error
-}
-
-func newNavColumn(all []ui.Styled, sel func(int) bool) *navColumn {
-	nc := &navColumn{all: all, candidates: all}
-	nc.provider = nc
-	nc.selected = -1
-	for i := range all {
-		if sel(i) {
-			nc.selected = i
-		}
-	}
-	return nc
-}
-
-func newErrNavColumn(err error) *navColumn {
-	nc := &navColumn{err: err}
-	nc.provider = nc
-	return nc
-}
-
-func (nc *navColumn) Placeholder() string {
-	if nc.err != nil {
-		return nc.err.Error()
-	}
-	return ""
-}
-
-func (nc *navColumn) Len() int {
-	return len(nc.candidates)
-}
-
-func (nc *navColumn) Show(i int) (string, ui.Styled) {
-	cand := nc.candidates[i]
-	return "", ui.Styled{" " + cand.Text + " ", cand.Styles}
-}
-
-func (nc *navColumn) Filter(filter string) int {
-	nc.candidates = nc.candidates[:0]
-	for _, s := range nc.all {
-		if strings.Contains(s.Text, filter) {
-			nc.candidates = append(nc.candidates, s)
-		}
-	}
-	return 0
-}
-
-func (nc *navColumn) FullWidth(h int) int {
-	if nc == nil {
-		return 0
-	}
-	if nc.err != nil {
-		return util.Wcswidth(nc.err.Error())
-	}
-	maxw := 0
-	for _, s := range nc.candidates {
-		maxw = max(maxw, util.Wcswidth(s.Text)+2)
-	}
-	if len(nc.candidates) > h {
-		maxw++
-	}
-	return maxw
-}
-
-func (nc *navColumn) Accept(i int, ed *Editor) {
-	// TODO
-}
-
-func (nc *navColumn) ModeTitle(i int) string {
-	// Not used
-	return ""
-}
-
-func (nc *navColumn) selectedName() string {
-	if nc == nil || nc.selected == -1 || nc.selected >= len(nc.candidates) {
-		return ""
-	}
-	return nc.candidates[nc.selected].Text
-}
diff -pruN 0.11+ds1-5/edit/navigation_preview.go 0.12+ds1-1/edit/navigation_preview.go
--- 0.11+ds1-5/edit/navigation_preview.go	2018-01-18 02:03:11.000000000 +0000
+++ 0.12+ds1-1/edit/navigation_preview.go	1970-01-01 00:00:00.000000000 +0000
@@ -1,91 +0,0 @@
-package edit
-
-import (
-	"errors"
-	"io"
-	"os"
-	"strings"
-	"unicode/utf8"
-
-	"github.com/elves/elvish/edit/ui"
-	"github.com/elves/elvish/util"
-)
-
-// PreviewBytes is the maximum number of bytes to preview a file.
-const PreviewBytes = 64 * 1024
-
-// Errors displayed in the preview area of navigation mode.
-var (
-	ErrNotRegular   = errors.New("no preview for non-regular file")
-	ErrNotValidUTF8 = errors.New("no preview for non-utf8 file")
-)
-
-type navFilePreview struct {
-	lines     []ui.Styled
-	fullWidth int
-	beginLine int
-}
-
-func newNavFilePreview(lines []string) *navFilePreview {
-	width := 0
-	convertedLines := make([]ui.Styled, len(lines))
-	for i, line := range lines {
-		// BUG: Handle tabstops correctly
-		convertedLine := strings.Replace(line, "\t", "    ", -1)
-		convertedLines[i] = ui.Unstyled(convertedLine)
-		width = max(width, util.Wcswidth(convertedLine))
-	}
-	return &navFilePreview{convertedLines, width, 0}
-}
-
-func (fp *navFilePreview) FullWidth(h int) int {
-	width := fp.fullWidth
-	if h < len(fp.lines) {
-		return width + 1
-	}
-	return width
-}
-
-func (fp *navFilePreview) List(h int) ui.Renderer {
-	if len(fp.lines) <= h {
-		logger.Printf("Height %d fit all lines", h)
-		return listingRenderer{fp.lines}
-	}
-	shown := fp.lines[fp.beginLine:]
-	if len(shown) > h {
-		shown = shown[:h]
-	}
-	logger.Printf("Showing lines %d to %d", fp.beginLine, fp.beginLine+len(shown))
-	return listingWithScrollBarRenderer{
-		listingRenderer{shown}, len(fp.lines),
-		fp.beginLine, fp.beginLine + len(shown), h}
-}
-
-func makeNavFilePreview(fname string) navPreview {
-	file, err := os.Open(fname)
-	if err != nil {
-		return newErrNavColumn(err)
-	}
-
-	info, err := file.Stat()
-	if err != nil {
-		return newErrNavColumn(err)
-	}
-	if (info.Mode() & (os.ModeDevice | os.ModeNamedPipe | os.ModeSocket | os.ModeCharDevice)) != 0 {
-		return newErrNavColumn(ErrNotRegular)
-	}
-
-	// BUG: when the file is bigger than the buffer, the scrollbar is wrong.
-	var buf [PreviewBytes]byte
-	nr, err := file.Read(buf[:])
-	if err != nil && err != io.EOF {
-		return newErrNavColumn(err)
-	}
-
-	content := string(buf[:nr])
-	if !utf8.ValidString(content) {
-		return newErrNavColumn(ErrNotValidUTF8)
-	}
-
-	return newNavFilePreview(strings.Split(content, "\n"))
-}
diff -pruN 0.11+ds1-5/edit/navigation_width.go 0.12+ds1-1/edit/navigation_width.go
--- 0.11+ds1-5/edit/navigation_width.go	2018-01-18 02:03:11.000000000 +0000
+++ 0.12+ds1-1/edit/navigation_width.go	1970-01-01 00:00:00.000000000 +0000
@@ -1,30 +0,0 @@
-package edit
-
-// getNavWidths calculates the widths for the three (parent, current and
-// preview) columns in the navigation mode. It takes the available width, full
-// width required to display the current and preview columns, and returns
-// suitable widths for the columns.
-//
-// The parent column always gets 1/6 of the total width. The current and preview
-// columns initially get 1/2 of the remaining width each, but if one of them
-// does not have enough widths and another has some spare width, the amount of
-// spare width or the needed width (whichever is smaller) is donated from the
-// latter to the former.
-func getNavWidths(total, currentFull, previewFull int) (int, int, int) {
-	parent := total / 6
-
-	remain := total - parent
-	current := remain / 2
-	preview := remain - current
-
-	if current < currentFull && preview > previewFull {
-		donate := min(currentFull-current, preview-previewFull)
-		current += donate
-		preview -= donate
-	} else if preview < previewFull && current > currentFull {
-		donate := min(previewFull-preview, current-currentFull)
-		preview += donate
-		current -= donate
-	}
-	return parent, current, preview
-}
diff -pruN 0.11+ds1-5/edit/navigation_widths_test.go 0.12+ds1-1/edit/navigation_widths_test.go
--- 0.11+ds1-5/edit/navigation_widths_test.go	2018-01-18 02:03:11.000000000 +0000
+++ 0.12+ds1-1/edit/navigation_widths_test.go	1970-01-01 00:00:00.000000000 +0000
@@ -1,25 +0,0 @@
-package edit
-
-import (
-	"testing"
-
-	"github.com/elves/elvish/tt"
-)
-
-var (
-	getNavWidthsTests = tt.Table{
-		// Enough room for both current and preview: parent gets 1/6, current
-		// and preview gets 1/2 of remain
-		tt.Args(120, 10, 10).Rets(20, 50, 50),
-		// Not enough room for either of current and preview: same as above
-		tt.Args(120, 100, 100).Rets(20, 50, 50),
-		// Enough room for current but not preview; current donates to preview
-		tt.Args(120, 10, 100).Rets(20, 10, 90),
-		// Enough room for preview but not current; preview donates to current
-		tt.Args(120, 100, 10).Rets(20, 90, 10),
-	}
-)
-
-func TestGetNavWidths(t *testing.T) {
-	tt.Test(t, tt.Fn("getNavWidths", getNavWidths), getNavWidthsTests)
-}
diff -pruN 0.11+ds1-5/edit/nodeutil.go 0.12+ds1-1/edit/nodeutil.go
--- 0.11+ds1-5/edit/nodeutil.go	2018-01-18 02:03:11.000000000 +0000
+++ 0.12+ds1-1/edit/nodeutil.go	1970-01-01 00:00:00.000000000 +0000
@@ -1,60 +0,0 @@
-package edit
-
-import (
-	"strings"
-
-	"github.com/elves/elvish/parse"
-)
-
-// Utilities for insepcting the AST. Used for completers and stylists.
-
-func primaryInSimpleCompound(pn *parse.Primary, ev pureEvaler) (*parse.Compound, string) {
-	indexing := parse.GetIndexing(pn.Parent())
-	if indexing == nil {
-		return nil, ""
-	}
-	compound := parse.GetCompound(indexing.Parent())
-	if compound == nil {
-		return nil, ""
-	}
-	head, err := ev.PurelyEvalPartialCompound(compound, indexing)
-	if err != nil {
-		return nil, ""
-	}
-	return compound, head
-}
-
-// leafNodeAtDot finds the leaf node at a specific position. It returns nil if
-// position is out of bound.
-func findLeafNode(n parse.Node, p int) parse.Node {
-descend:
-	for len(n.Children()) > 0 {
-		for _, ch := range n.Children() {
-			if ch.Begin() <= p && p <= ch.End() {
-				n = ch
-				continue descend
-			}
-		}
-		return nil
-	}
-	return n
-}
-
-func wordify(src string) []string {
-	n, _ := parse.Parse("[wordify]", src)
-	return wordifyInner(n, nil)
-}
-
-func wordifyInner(n parse.Node, words []string) []string {
-	if len(n.Children()) == 0 || parse.IsCompound(n) {
-		text := n.SourceText()
-		if strings.TrimFunc(text, parse.IsSpaceOrNewline) != "" {
-			return append(words, text)
-		}
-		return words
-	}
-	for _, ch := range n.Children() {
-		words = wordifyInner(ch, words)
-	}
-	return words
-}
diff -pruN 0.11+ds1-5/edit/prompt/defaults.go 0.12+ds1-1/edit/prompt/defaults.go
--- 0.11+ds1-5/edit/prompt/defaults.go	1970-01-01 00:00:00.000000000 +0000
+++ 0.12+ds1-1/edit/prompt/defaults.go	2018-07-11 02:32:25.000000000 +0000
@@ -0,0 +1,57 @@
+package prompt
+
+import (
+	"os"
+	"os/user"
+
+	"github.com/elves/elvish/edit/ui"
+	"github.com/elves/elvish/eval"
+	"github.com/elves/elvish/util"
+)
+
+var defaultPrompt, defaultRPrompt, defaultStaleTransform eval.Callable
+
+func init() {
+	user, err := user.Current()
+	isRoot := err == nil && user.Uid == "0"
+
+	prompt := func(fm *eval.Frame) {
+		out := fm.OutputChan()
+		out <- string(util.Getwd())
+		if isRoot {
+			out <- &ui.Styled{"# ", ui.Styles{"red"}}
+		} else {
+			out <- &ui.Styled{"> ", ui.Styles{}}
+		}
+	}
+	defaultPrompt = eval.NewBuiltinFn("default prompt", prompt)
+}
+
+func init() {
+	username := "???"
+	user, err := user.Current()
+	if err == nil {
+		username = user.Username
+	}
+	hostname, err := os.Hostname()
+	if err != nil {
+		hostname = "???"
+	}
+	rpromptStr := username + "@" + hostname
+	rprompt := func(fm *eval.Frame) {
+		out := fm.OutputChan()
+		out <- &ui.Styled{rpromptStr, ui.Styles{"inverse"}}
+	}
+	defaultRPrompt = eval.NewBuiltinFn("default rprompt", rprompt)
+}
+
+func init() {
+	staleTransform := func(fm *eval.Frame) {
+		out := fm.OutputChan()
+		fm.IterateInputs(func(i interface{}) {
+			s := i.(*ui.Styled)
+			out <- &ui.Styled{s.Text, ui.Styles{"inverse"}}
+		})
+	}
+	defaultStaleTransform = eval.NewBuiltinFn("default stale transform", staleTransform)
+}
diff -pruN 0.11+ds1-5/edit/prompt/prompt.go 0.12+ds1-1/edit/prompt/prompt.go
--- 0.11+ds1-5/edit/prompt/prompt.go	2018-01-18 02:03:11.000000000 +0000
+++ 0.12+ds1-1/edit/prompt/prompt.go	2018-07-11 02:32:25.000000000 +0000
@@ -1,134 +1,209 @@
-// Package prompt implements prompt-related functionalities of the editor.
+// Package prompt implements the prompt subsystem of the editor.
 package prompt
 
 import (
 	"io/ioutil"
 	"math"
 	"os"
-	"os/user"
-	"strconv"
 	"sync"
 	"time"
 
+	"github.com/elves/elvish/edit/eddefs"
 	"github.com/elves/elvish/edit/ui"
 	"github.com/elves/elvish/eval"
-	"github.com/elves/elvish/eval/types"
-	"github.com/elves/elvish/eval/vartypes"
+	"github.com/elves/elvish/eval/vals"
+	"github.com/elves/elvish/eval/vars"
+	"github.com/elves/elvish/styled"
 	"github.com/elves/elvish/util"
 )
 
 var logger = util.GetLogger("[edit/prompt] ")
 
-// Editor is the interface used by the prompt to access the editor.
-type Editor interface {
-	Evaler() *eval.Evaler
-	Variable(string) vartypes.Variable
-	Notify(string, ...interface{})
+// Init initializes the prompt subsystem of the editor.
+func Init(ed eddefs.Editor, ns eval.Ns) {
+	prompt := makePrompt(ed, defaultPrompt)
+	rprompt := makePrompt(ed, defaultRPrompt)
+	ed.SetPrompt(prompt)
+	ed.SetRPrompt(rprompt)
+	installAPI(ns, prompt, "prompt")
+	installAPI(ns, rprompt, "rprompt")
+}
+
+func installAPI(ns eval.Ns, p *prompt, basename string) {
+	ns.Add(basename, vars.FromPtr(&p.fn))
+	ns.Add(basename+"-stale-threshold", vars.FromPtr(&p.staleThreshold))
+	ns.Add(basename+"-stale-transform", vars.FromPtr(&p.staleTransform))
+	ns.Add("-"+basename+"-eagerness", vars.FromPtr(&p.eagerness))
+}
+
+type prompt struct {
+	ed eddefs.Editor
+	// The main callback.
+	fn eval.Callable
+	// Callback used to transform stale prompts.
+	staleTransform eval.Callable
+	// Threshold in seconds for a prompt to be considered as stale.
+	staleThreshold float64
+	// How eager the prompt should be updated. When >= 5, updated when directory
+	// is changed. When >= 10, always update. Default is 5.
+	eagerness int
+
+	// Working directory when prompt was last updated.
+	lastWd string
+	// Channel for update requests.
+	updateReq chan struct{}
+	// Channel on which prompt contents are sent.
+	ch chan []*ui.Styled
+	// Last prompt content
+	last      []*ui.Styled
+	lastMutex *sync.RWMutex
+}
+
+var unknownContent = []*ui.Styled{&ui.Styled{"???> ", ui.Styles{}}}
+
+func makePrompt(ed eddefs.Editor, fn eval.Callable) *prompt {
+	p := &prompt{
+		ed, fn, defaultStaleTransform, 0.2, 5,
+		"", make(chan struct{}, 1), make(chan []*ui.Styled, 1),
+		unknownContent, new(sync.RWMutex)}
+	go p.loop()
+	return p
 }
 
-// maxSeconds is the maximum number of seconds time.Duration can represent.
-const maxSeconds = float64(math.MaxInt64 / time.Second)
-
-// PromptVariable returns a variable for $edit:prompt.
-func PromptVariable() vartypes.Variable {
-	user, err := user.Current()
-	isRoot := err == nil && user.Uid == "0"
-
-	prompt := func(ec *eval.Frame,
-		args []types.Value, opts map[string]types.Value) {
-
-		out := ec.OutputChan()
-		out <- types.String(util.Getwd())
-		if isRoot {
-			out <- &ui.Styled{"# ", ui.Styles{"red"}}
-		} else {
-			out <- &ui.Styled{"> ", ui.Styles{}}
+func (p *prompt) loop() {
+	content := unknownContent
+	ch := make(chan []*ui.Styled)
+	for range p.updateReq {
+		go func() {
+			ch <- callPrompt(p.ed, p.fn)
+		}()
+
+		select {
+		case <-makeMaxWaitChan(p.staleThreshold):
+			// The prompt callback did not finish within the threshold. Send the
+			// previous content, marked as stale.
+			p.send(callTransformer(p.ed, p.staleTransform, content))
+			content = <-ch
+
+			select {
+			case <-p.updateReq:
+				// If another update is already requested by the time we finish,
+				// keep marking the prompt as stale. This reduces flickering.
+				p.send(callTransformer(p.ed, p.staleTransform, content))
+				p.queueUpdate()
+			default:
+				p.send(content)
+			}
+		case content = <-ch:
+			p.send(content)
 		}
 	}
-	return vartypes.NewValidatedPtr(
-		&eval.BuiltinFn{"default prompt", prompt}, eval.ShouldBeFn)
 }
 
-// Prompt extracts $edit:prompt.
-func Prompt(ed Editor) eval.Callable {
-	return ed.Variable("prompt").Get().(eval.Callable)
+func (p *prompt) Chan() <-chan []*ui.Styled {
+	return p.ch
 }
 
-// RpromptVariable returns a variable for $edit:rprompt.
-func RpromptVariable() vartypes.Variable {
-	username := "???"
-	user, err := user.Current()
-	if err == nil {
-		username = user.Username
-	}
-	hostname, err := os.Hostname()
-	if err != nil {
-		hostname = "???"
-	}
-	rpromptStr := username + "@" + hostname
-	rprompt := func(ec *eval.Frame,
-		args []types.Value, opts map[string]types.Value) {
-
-		out := ec.OutputChan()
-		out <- &ui.Styled{rpromptStr, ui.Styles{"inverse"}}
+func (p *prompt) Update(force bool) {
+	if force || p.shouldUpdate() {
+		p.queueUpdate()
 	}
-
-	return vartypes.NewValidatedPtr(
-		&eval.BuiltinFn{"default rprompt", rprompt}, eval.ShouldBeFn)
 }
 
-// Rprompt extracts $edit:rprompt.
-func Rprompt(ed Editor) eval.Callable {
-	return ed.Variable("rprompt").Get().(eval.Callable)
+func (p *prompt) Last() []*ui.Styled {
+	p.lastMutex.RLock()
+	defer p.lastMutex.RUnlock()
+	return p.last
 }
 
-// Implementation for $rprompt-persistent.
-
-// TODO Keep the underlying boolean and float64 values somewhere within the
-// Editor to avoid casts.
-
-// RpromptPersistentVariable returns a variable for $edit:rprompt-persistent.
-func RpromptPersistentVariable() vartypes.Variable {
-	b := false
-	return vartypes.NewBool(&b)
+func (p *prompt) Close() error {
+	// TODO: Close p.updateReq. However, doing this can cause
+	// write-to-closed-channel panics.
+	return nil
 }
 
-// RpromptPersistent extracts $edit:rprompt-persistent.
-func RpromptPersistent(ed Editor) bool {
-	return bool(ed.Variable("rprompt-persistent").Get().(types.Bool).Bool())
+func (p *prompt) queueUpdate() {
+	select {
+	case p.updateReq <- struct{}{}:
+	default:
+	}
 }
 
-// MaxWaitVariable returns a variable for $edit:-prompts-max-wait.
-func MaxWaitVariable() vartypes.Variable {
-	f := math.Inf(1)
-	return vartypes.NewNumber(&f)
+func (p *prompt) send(content []*ui.Styled) {
+	p.lastMutex.Lock()
+	p.last = content
+	p.lastMutex.Unlock()
+	p.ch <- content
 }
 
-// MaxWait extracts $edit:-prompts-max-wait.
-func MaxWait(ed Editor) float64 {
-	f, _ := strconv.ParseFloat(string(ed.Variable("-prompts-max-wait").Get().(types.String)), 64)
-	return f
+func (p *prompt) shouldUpdate() bool {
+	if p.eagerness >= 10 {
+		return true
+	}
+	if p.eagerness >= 5 {
+		wd, err := os.Getwd()
+		if err != nil {
+			wd = "error"
+		}
+		oldWd := p.lastWd
+		p.lastWd = wd
+		return wd != oldWd
+	}
+	return false
 }
 
-// MakeMaxWait makes a channel that sends the current time after
-// $edit:-prompts-max-wait seconds if the time fits in a time.Duration value, or
-// nil otherwise.
-func MakeMaxWaitChan(ed Editor) <-chan time.Time {
-	f := MaxWait(ed)
+// maxSeconds is the maximum number of seconds time.Duration can represent.
+const maxSeconds = float64(math.MaxInt64 / time.Second)
+
+// makeMaxWaitChan makes a channel that sends the current time after f seconds.
+// If f does not fits in a time.Duration value, it returns nil, which is a
+// channel that never sends any value.
+func makeMaxWaitChan(f float64) <-chan time.Time {
 	if f > maxSeconds {
 		return nil
 	}
 	return time.After(time.Duration(f * float64(time.Second)))
 }
 
-// callPrompt calls a Fn, assuming that it is a prompt. It calls the Fn with no
-// arguments and closed input, and converts its outputs to styled objects.
-func callPrompt(ed Editor, fn eval.Callable) []*ui.Styled {
+// callPrompt calls a function with no arguments and closed input, and converts
+// its outputs to styled objects. Used to call prompt callbacks.
+func callPrompt(ed eddefs.Editor, fn eval.Callable) []*ui.Styled {
 	ports := []*eval.Port{
 		eval.DevNullClosedChan,
 		{}, // Will be replaced when capturing output
 		{File: os.Stderr},
 	}
+
+	return callAndGetStyled(ed, fn, ports)
+}
+
+// callTransformer calls a function with no arguments and the given inputs, and
+// converts its outputs to styled objects. Used to call stale transformers.
+func callTransformer(ed eddefs.Editor, fn eval.Callable, currentPrompt []*ui.Styled) []*ui.Styled {
+	input := make(chan interface{})
+	stopInputWriter := make(chan struct{})
+
+	ports := []*eval.Port{
+		{Chan: input, File: eval.DevNull},
+		{}, // Will be replaced when capturing output
+		{File: os.Stderr},
+	}
+	go func() {
+		defer close(input)
+		for _, char := range currentPrompt {
+			select {
+			case input <- char:
+			case <-stopInputWriter:
+				return
+			}
+		}
+	}()
+	defer close(stopInputWriter)
+
+	return callAndGetStyled(ed, fn, ports)
+}
+
+func callAndGetStyled(ed eddefs.Editor, fn eval.Callable, ports []*eval.Port) []*ui.Styled {
 	var (
 		styleds      []*ui.Styled
 		styledsMutex sync.Mutex
@@ -140,12 +215,16 @@ func callPrompt(ed Editor, fn eval.Calla
 	}
 	// Value output may be of type ui.Styled or any other type, in which case
 	// they are converted to ui.Styled.
-	valuesCb := func(ch <-chan types.Value) {
+	valuesCb := func(ch <-chan interface{}) {
 		for v := range ch {
 			if s, ok := v.(*ui.Styled); ok {
 				add(s)
+			} else if s, ok := v.(*styled.Text); ok {
+				for _, seg := range *s {
+					add(seg.ToLegacyType())
+				}
 			} else {
-				add(&ui.Styled{types.ToString(v), ui.Styles{}})
+				add(&ui.Styled{vals.ToString(v), ui.Styles{}})
 			}
 		}
 	}
@@ -162,7 +241,7 @@ func callPrompt(ed Editor, fn eval.Calla
 
 	// XXX There is no source to pass to NewTopEvalCtx.
 	ec := eval.NewTopFrame(ed.Evaler(), eval.NewInternalSource("[prompt]"), ports)
-	err := ec.PCaptureOutputInner(fn, nil, eval.NoOpts, valuesCb, bytesCb)
+	err := ec.CallWithOutputCallback(fn, nil, eval.NoOpts, valuesCb, bytesCb)
 
 	if err != nil {
 		ed.Notify("prompt function error: %v", err)
@@ -171,30 +250,3 @@ func callPrompt(ed Editor, fn eval.Calla
 
 	return styleds
 }
-
-// Updater manages the update of a prompt.
-type Updater struct {
-	promptFn func(Editor) eval.Callable
-	Staled   []*ui.Styled
-}
-
-var staledPrompt = &ui.Styled{"?", ui.Styles{"inverse"}}
-
-// NewUpdater creates a new Updater.
-func NewUpdater(promptFn func(Editor) eval.Callable) *Updater {
-	return &Updater{promptFn, []*ui.Styled{staledPrompt}}
-}
-
-// Update updates the prompt, returning a channel onto which the result will be
-// written.
-func (pu *Updater) Update(ed Editor) <-chan []*ui.Styled {
-	ch := make(chan []*ui.Styled)
-	go func() {
-		result := callPrompt(ed, pu.promptFn(ed))
-		pu.Staled = make([]*ui.Styled, len(result)+1)
-		pu.Staled[0] = staledPrompt
-		copy(pu.Staled[1:], result)
-		ch <- result
-	}()
-	return ch
-}
diff -pruN 0.11+ds1-5/edit/prompt_vars.go 0.12+ds1-1/edit/prompt_vars.go
--- 0.11+ds1-5/edit/prompt_vars.go	2018-01-18 02:03:11.000000000 +0000
+++ 0.12+ds1-1/edit/prompt_vars.go	1970-01-01 00:00:00.000000000 +0000
@@ -1,10 +0,0 @@
-package edit
-
-import "github.com/elves/elvish/edit/prompt"
-
-var (
-	_ = RegisterVariable("prompt", prompt.PromptVariable)
-	_ = RegisterVariable("rprompt", prompt.RpromptVariable)
-	_ = RegisterVariable("rprompt-persistent", prompt.RpromptPersistentVariable)
-	_ = RegisterVariable("-prompts-max-wait", prompt.MaxWaitVariable)
-)
diff -pruN 0.11+ds1-5/edit/raw_insert.go 0.12+ds1-1/edit/raw_insert.go
--- 0.11+ds1-5/edit/raw_insert.go	2018-01-18 02:03:11.000000000 +0000
+++ 0.12+ds1-1/edit/raw_insert.go	1970-01-01 00:00:00.000000000 +0000
@@ -1,34 +0,0 @@
-package edit
-
-import (
-	"github.com/elves/elvish/edit/ui"
-	"github.com/elves/elvish/eval"
-	"github.com/elves/elvish/eval/vartypes"
-)
-
-// Raw insert mode is a special mode, in that it does not use the normal key
-// binding. Rather, insertRaw is called directly from the main loop in
-// Editor.ReadLine.
-
-type rawInsert struct {
-}
-
-func startInsertRaw(ed *Editor) {
-	ed.reader.SetRaw(true)
-	ed.mode = rawInsert{}
-}
-
-func insertRaw(ed *Editor, r rune) {
-	ed.insertAtDot(string(r))
-	ed.reader.SetRaw(false)
-	ed.mode = &ed.insert
-}
-
-func (rawInsert) Binding(map[string]vartypes.Variable, ui.Key) eval.Fn {
-	// The raw insert mode does not handle keys.
-	return nil
-}
-
-func (ri rawInsert) ModeLine() ui.Renderer {
-	return modeLineRenderer{" RAW ", ""}
-}
diff -pruN 0.11+ds1-5/edit/registry.go 0.12+ds1-1/edit/registry.go
--- 0.11+ds1-5/edit/registry.go	2018-01-18 02:03:11.000000000 +0000
+++ 0.12+ds1-1/edit/registry.go	1970-01-01 00:00:00.000000000 +0000
@@ -1,82 +0,0 @@
-package edit
-
-import (
-	"errors"
-
-	"github.com/elves/elvish/eval"
-	"github.com/elves/elvish/eval/types"
-	"github.com/elves/elvish/eval/vartypes"
-)
-
-// This file contains several "registries", data structure that are written
-// during program initialization and later used when initializing the Editor.
-//
-// The purpose of these registries is to decentralize the definition for
-// builtins, default bindings, and variables (e.g. $edit:prompt). For instance,
-// the definition for $edit:prompt can live in prompt.go instead of api.go.
-
-var variableRegistry = map[string]func() vartypes.Variable{}
-
-// RegisterVariable registers a variable: its name and a func to derive a
-// Variable instance. It is later to be used during Editor initialization to
-// populate Editor.variables as well as the edit: namespace.
-func RegisterVariable(name string, maker func() vartypes.Variable) struct{} {
-	variableRegistry[name] = maker
-	return struct{}{}
-}
-
-func makeVariables() map[string]vartypes.Variable {
-	m := make(map[string]vartypes.Variable, len(variableRegistry))
-	for name, maker := range variableRegistry {
-		m[name] = maker()
-	}
-	return m
-}
-
-var builtinMaps = map[string]map[string]*BuiltinFn{}
-
-// registerBuiltins registers builtins under a subnamespace of edit:, to be used
-// during the initialization of the Editor. It should be called for global
-// variable initializations to make sure every subnamespace is registered before
-// makeBindings is ever called.
-func registerBuiltins(module string, impls map[string]func(*Editor)) struct{} {
-	if _, ok := builtinMaps[module]; !ok {
-		builtinMaps[module] = make(map[string]*BuiltinFn)
-	}
-	for name, impl := range impls {
-		ns := "edit"
-		if module != "" {
-			ns += ":" + module
-		}
-		fullName := ns + ":" + name + eval.FnSuffix
-		builtinMaps[module][name] = &BuiltinFn{fullName, impl}
-	}
-	return struct{}{}
-}
-
-func makeNsFromBuiltins(builtins map[string]*BuiltinFn) eval.Ns {
-	ns := make(eval.Ns)
-	for name, builtin := range builtins {
-		ns[name+eval.FnSuffix] = vartypes.NewPtr(builtin)
-	}
-	return ns
-}
-
-func makeBindings() map[string]vartypes.Variable {
-	bindings := map[string]vartypes.Variable{}
-	// XXX This abuses the builtin registry to get a list of mode names
-	for mode := range builtinMaps {
-		bindings[mode] = vartypes.NewValidatedPtr(
-			BindingTable{types.EmptyMap}, shouldBeBindingTable)
-	}
-	return bindings
-}
-
-var errShouldBeBindingTable = errors.New("should be binding table")
-
-func shouldBeBindingTable(v types.Value) error {
-	if _, ok := v.(BindingTable); !ok {
-		return errShouldBeBindingTable
-	}
-	return nil
-}
diff -pruN 0.11+ds1-5/edit/render.go 0.12+ds1-1/edit/render.go
--- 0.11+ds1-5/edit/render.go	2018-01-18 02:03:11.000000000 +0000
+++ 0.12+ds1-1/edit/render.go	1970-01-01 00:00:00.000000000 +0000
@@ -1,311 +0,0 @@
-package edit
-
-import (
-	"strings"
-	"unicode/utf8"
-
-	"github.com/elves/elvish/edit/highlight"
-	"github.com/elves/elvish/edit/ui"
-	"github.com/elves/elvish/util"
-)
-
-type modeLineRenderer struct {
-	title  string
-	filter string
-}
-
-func (ml modeLineRenderer) Render(b *ui.Buffer) {
-	b.WriteString(ml.title, styleForMode.String())
-	b.WriteSpaces(1, "")
-	b.WriteString(ml.filter, styleForFilter.String())
-	b.Dot = b.Cursor()
-}
-
-type modeLineWithScrollBarRenderer struct {
-	modeLineRenderer
-	n, low, high int
-}
-
-func (ml modeLineWithScrollBarRenderer) Render(b *ui.Buffer) {
-	ml.modeLineRenderer.Render(b)
-
-	scrollbarWidth := b.Width - ui.CellsWidth(b.Lines[len(b.Lines)-1]) - 2
-	if scrollbarWidth >= 3 {
-		b.WriteSpaces(1, "")
-		writeHorizontalScrollbar(b, ml.n, ml.low, ml.high, scrollbarWidth)
-	}
-}
-
-type placeholderRenderer string
-
-func (lp placeholderRenderer) Render(b *ui.Buffer) {
-	b.WriteString(util.TrimWcwidth(string(lp), b.Width), "")
-}
-
-type listingRenderer struct {
-	lines []ui.Styled
-}
-
-func (ls listingRenderer) Render(b *ui.Buffer) {
-	for i, line := range ls.lines {
-		if i > 0 {
-			b.Newline()
-		}
-		b.WriteString(util.ForceWcwidth(line.Text, b.Width), line.Styles.String())
-	}
-}
-
-type listingWithScrollBarRenderer struct {
-	listingRenderer
-	n, low, high, height int
-}
-
-func (ls listingWithScrollBarRenderer) Render(b *ui.Buffer) {
-	b1 := ui.Render(ls.listingRenderer, b.Width-1)
-	b.ExtendRight(b1, 0)
-
-	scrollbar := renderScrollbar(ls.n, ls.low, ls.high, ls.height)
-	b.ExtendRight(scrollbar, b.Width-1)
-}
-
-type navRenderer struct {
-	maxHeight                      int
-	fwParent, fwCurrent, fwPreview int
-	parent, current, preview       ui.Renderer
-}
-
-func makeNavRenderer(h int, w1, w2, w3 int, r1, r2, r3 ui.Renderer) ui.Renderer {
-	return &navRenderer{h, w1, w2, w3, r1, r2, r3}
-}
-
-const navColMargin = 1
-
-func (nr *navRenderer) Render(b *ui.Buffer) {
-	wParent, wCurrent, wPreview := getNavWidths(b.Width-navColMargin*2,
-		nr.fwCurrent, nr.fwPreview)
-
-	bParent := ui.Render(nr.parent, wParent)
-	b.ExtendRight(bParent, 0)
-
-	bCurrent := ui.Render(nr.current, wCurrent)
-	b.ExtendRight(bCurrent, wParent+navColMargin)
-
-	if wPreview > 0 {
-		bPreview := ui.Render(nr.preview, wPreview)
-		b.ExtendRight(bPreview, wParent+wCurrent+2*navColMargin)
-	}
-}
-
-// linesRenderer renders lines with a uniform style.
-type linesRenderer struct {
-	lines []string
-	style string
-}
-
-func (nr linesRenderer) Render(b *ui.Buffer) {
-	b.WriteString(strings.Join(nr.lines, "\n"), "")
-}
-
-// cmdlineRenderer renders the command line, including the prompt, the user's
-// input and the rprompt.
-type cmdlineRenderer struct {
-	prompt  []*ui.Styled
-	line    string
-	styling *highlight.Styling
-	dot     int
-	rprompt []*ui.Styled
-
-	hasComp   bool
-	compBegin int
-	compEnd   int
-	compText  string
-
-	hasHist   bool
-	histBegin int
-	histText  string
-}
-
-func newCmdlineRenderer(p []*ui.Styled, l string, s *highlight.Styling, d int, rp []*ui.Styled) *cmdlineRenderer {
-	return &cmdlineRenderer{prompt: p, line: l, styling: s, dot: d, rprompt: rp}
-}
-
-func (clr *cmdlineRenderer) setComp(b, e int, t string) {
-	clr.hasComp = true
-	clr.compBegin, clr.compEnd, clr.compText = b, e, t
-}
-
-func (clr *cmdlineRenderer) setHist(b int, t string) {
-	clr.hasHist = true
-	clr.histBegin, clr.histText = b, t
-}
-
-func (clr *cmdlineRenderer) Render(b *ui.Buffer) {
-	b.EagerWrap = true
-
-	b.WriteStyleds(clr.prompt)
-
-	// If the prompt takes less than half of a line, set the indent.
-	if len(b.Lines) == 1 && b.Col*2 < b.Width {
-		b.Indent = b.Col
-	}
-
-	// i keeps track of number of bytes written.
-	i := 0
-
-	applier := clr.styling.Apply()
-
-	// nowAt is called at every rune boundary.
-	nowAt := func(i int) {
-		applier.At(i)
-		if clr.hasComp && i == clr.compBegin {
-			b.WriteString(clr.compText, styleForCompleted.String())
-		}
-		if i == clr.dot {
-			b.Dot = b.Cursor()
-		}
-	}
-	nowAt(0)
-
-	for _, r := range clr.line {
-		if clr.hasComp && clr.compBegin <= i && i < clr.compEnd {
-			// Do nothing. This part is replaced by the completion candidate.
-		} else {
-			b.Write(r, applier.Get())
-		}
-		i += utf8.RuneLen(r)
-
-		nowAt(i)
-		if clr.hasHist && i == clr.histBegin {
-			break
-		}
-	}
-
-	if clr.hasHist {
-		// Put the rest of current history and position the cursor at the
-		// end of the line.
-		b.WriteString(clr.histText, styleForCompletedHistory.String())
-		b.Dot = b.Cursor()
-	}
-
-	// Write rprompt
-	if len(clr.rprompt) > 0 {
-		padding := b.Width - b.Col
-		for _, s := range clr.rprompt {
-			padding -= util.Wcswidth(s.Text)
-		}
-		if padding >= 1 {
-			b.EagerWrap = false
-			b.WriteSpaces(padding, "")
-			b.WriteStyleds(clr.rprompt)
-		}
-	}
-}
-
-var logEditorRender = false
-
-// editorRenderer renders the entire editor.
-type editorRenderer struct {
-	*editorState
-	height  int
-	bufNoti *ui.Buffer
-}
-
-func (er *editorRenderer) Render(buf *ui.Buffer) {
-	height, width, es := er.height, buf.Width, er.editorState
-
-	var bufNoti, bufLine, bufMode, bufTips, bufListing *ui.Buffer
-	// butNoti
-	if len(es.notifications) > 0 {
-		bufNoti = ui.Render(linesRenderer{es.notifications, ""}, width)
-		es.notifications = nil
-	}
-
-	// bufLine
-	clr := newCmdlineRenderer(es.promptContent, es.buffer, es.styling, es.dot, es.rpromptContent)
-	// TODO(xiaq): Instead of doing a type switch, expose an API for modes to
-	// modify the text (and mark their part as modified).
-	switch mode := es.mode.(type) {
-	case *completion:
-		c := es.completion
-		clr.setComp(c.begin, c.end, c.selectedCandidate().code)
-	case *hist:
-		begin := len(mode.Prefix())
-		clr.setHist(begin, mode.CurrentCmd()[begin:])
-	}
-	bufLine = ui.Render(clr, width)
-
-	// bufMode
-	bufMode = ui.Render(es.mode.ModeLine(), width)
-
-	// bufTips
-	// TODO tips is assumed to contain no newlines.
-	if len(es.tips) > 0 {
-		bufTips = ui.Render(linesRenderer{es.tips, styleForTip.String()}, width)
-	}
-
-	hListing := 0
-	// Trim lines and determine the maximum height for bufListing
-	// TODO come up with a UI to tell the user that something is not shown.
-	switch {
-	case height >= ui.BuffersHeight(bufNoti, bufLine, bufMode, bufTips):
-		hListing = height - ui.BuffersHeight(bufLine, bufMode, bufTips)
-	case height >= ui.BuffersHeight(bufNoti, bufLine, bufTips):
-		bufMode = nil
-	case height >= ui.BuffersHeight(bufNoti, bufLine):
-		bufMode = nil
-		if bufTips != nil {
-			bufTips.TrimToLines(0, height-ui.BuffersHeight(bufNoti, bufLine))
-		}
-	case height >= ui.BuffersHeight(bufLine):
-		bufTips, bufMode = nil, nil
-		if bufNoti != nil {
-			n := len(bufNoti.Lines)
-			bufNoti.TrimToLines(n-(height-ui.BuffersHeight(bufLine)), n)
-		}
-	case height >= 1:
-		bufNoti, bufTips, bufMode = nil, nil, nil
-		dotLine := bufLine.Dot.Line
-		bufLine.TrimToLines(dotLine+1-height, dotLine+1)
-	default:
-		// Broken terminal. Still try to render one line of bufLine.
-		bufNoti, bufTips, bufMode = nil, nil, nil
-		dotLine := bufLine.Dot.Line
-		bufLine.TrimToLines(dotLine, dotLine+1)
-	}
-
-	// bufListing.
-	if hListing > 0 {
-		if lister, ok := es.mode.(ListRenderer); ok {
-			bufListing = lister.ListRender(width, hListing)
-		} else if lister, ok := es.mode.(Lister); ok {
-			bufListing = ui.Render(lister.List(hListing), width)
-		}
-		// XXX When in completion mode, we re-render the mode line, since the
-		// scrollbar in the mode line depends on completion.lastShown which is
-		// only known after the listing has been rendered. Since rendering the
-		// scrollbar never adds additional lines to bufMode, we may do this
-		// without recalculating the layout.
-		if _, ok := es.mode.(*completion); ok {
-			bufMode = ui.Render(es.mode.ModeLine(), width)
-		}
-	}
-
-	if logEditorRender {
-		logger.Printf("bufLine %d, bufMode %d, bufTips %d, bufListing %d",
-			ui.BuffersHeight(bufLine), ui.BuffersHeight(bufMode), ui.BuffersHeight(bufTips), ui.BuffersHeight(bufListing))
-	}
-
-	// XXX
-	buf.Lines = nil
-	// Combine buffers (reusing bufLine)
-	buf.Extend(bufLine, true)
-	cursorOnModeLine := false
-	if coml, ok := es.mode.(CursorOnModeLiner); ok {
-		cursorOnModeLine = coml.CursorOnModeLine()
-	}
-	buf.Extend(bufMode, cursorOnModeLine)
-	buf.Extend(bufTips, false)
-	buf.Extend(bufListing, false)
-
-	er.bufNoti = bufNoti
-}
diff -pruN 0.11+ds1-5/edit/styled.go 0.12+ds1-1/edit/styled.go
--- 0.11+ds1-5/edit/styled.go	2018-01-18 02:03:11.000000000 +0000
+++ 0.12+ds1-1/edit/styled.go	1970-01-01 00:00:00.000000000 +0000
@@ -1,17 +0,0 @@
-package edit
-
-import (
-	"github.com/elves/elvish/edit/ui"
-	"github.com/elves/elvish/eval"
-	"github.com/elves/elvish/eval/types"
-)
-
-func styled(ec *eval.Frame, args []types.Value, opts map[string]types.Value) {
-	var textv, stylev types.String
-	eval.ScanArgs(args, &textv, &stylev)
-	text, style := string(textv), string(stylev)
-	eval.TakeNoOpt(opts)
-
-	out := ec.OutputChan()
-	out <- &ui.Styled{text, ui.StylesFromString(style)}
-}
diff -pruN 0.11+ds1-5/edit/style.go 0.12+ds1-1/edit/style.go
--- 0.11+ds1-5/edit/style.go	2018-01-18 02:03:11.000000000 +0000
+++ 0.12+ds1-1/edit/style.go	1970-01-01 00:00:00.000000000 +0000
@@ -1,26 +0,0 @@
-package edit
-
-import (
-	"github.com/elves/elvish/edit/ui"
-)
-
-var styleForCompilerError = ui.Styles{"white", "bg-red"}
-
-// Styles for UI.
-var (
-	//styleForPrompt           = ""
-	//styleForRPrompt          = "inverse"
-	styleForCompleted        = ui.Styles{"underlined"}
-	styleForCompletedHistory = ui.Styles{"underlined"}
-	styleForMode             = ui.Styles{"bold", "lightgray", "bg-magenta"}
-	styleForTip              = ui.Styles{}
-	styleForFilter           = ui.Styles{"underlined"}
-	styleForSelected         = ui.Styles{"inverse"}
-	styleForScrollBarArea    = ui.Styles{"magenta"}
-	styleForScrollBarThumb   = ui.Styles{"magenta", "inverse"}
-
-	// Use default style for completion listing
-	styleForCompletion = ui.Styles{}
-	// Use inverse style for selected completion entry
-	styleForSelectedCompletion = ui.Styles{"inverse"}
-)
diff -pruN 0.11+ds1-5/edit/tty/reader_unix.go 0.12+ds1-1/edit/tty/reader_unix.go
--- 0.11+ds1-5/edit/tty/reader_unix.go	2018-01-18 02:03:11.000000000 +0000
+++ 0.12+ds1-1/edit/tty/reader_unix.go	2018-07-11 02:32:25.000000000 +0000
@@ -273,9 +273,8 @@ func (rd *reader) readOne(r rune) (event
 				event = KeyEvent{'o', ui.Alt}
 				return
 			}
-			r, ok := g3Seq[r]
+			k, ok := g3Seq[r]
 			if ok {
-				k := KeyEvent{r, 0}
 				if hasTwoLeadingESC {
 					k.Mod |= ui.Alt
 				}
@@ -317,51 +316,96 @@ func ctrlModify(r rune) ui.Key {
 	return ui.Key{r, 0}
 }
 
+// Tables for key sequences. Comments document which terminal emulators are
+// known to generate which sequences. The terminal emulators tested are
+// categorized into xterm (including actual xterm, libvte-based terminals,
+// Konsole and Terminal.app unless otherwise noted), urxvt, tmux.
+
 // G3-style key sequences: \eO followed by exactly one character. For instance,
-// \eOP is F1.
-var g3Seq = map[rune]rune{
-	'A': ui.Up, 'B': ui.Down, 'C': ui.Right, 'D': ui.Left,
-
-	// F1-F4: xterm, libvte and tmux
-	'P': ui.F1, 'Q': ui.F2,
-	'R': ui.F3, 'S': ui.F4,
-
-	// Home and End: libvte
-	'H': ui.Home, 'F': ui.End,
-}
-
-// Tables for CSI-style key sequences, which are \e[ followed by a list of
-// semicolon-delimited numeric arguments, before being concluded by a
-// non-numeric, non-semicolon rune.
-
-// CSI-style key sequences that can be identified based on the ending rune. For
-// instance, \e[A is Up.
-var keyByLast = map[rune]ui.Key{
-	'A': {ui.Up, 0}, 'B': {ui.Down, 0},
-	'C': {ui.Right, 0}, 'D': {ui.Left, 0},
+// \eOP is F1. These are pretty limited in that they cannot be extended to
+// support modifier keys, other than a leading \e for Alt (e.g. \e\eOP is
+// Alt-F1). Terminals that send G3-style key sequences typically switch to
+// sending a CSI-style key sequence when a non-Alt modifier key is pressed.
+var g3Seq = map[rune]ui.Key{
+	// xterm, tmux -- only in Vim, depends on termios setting?
+	// NOTE(xiaq): According to urxvt's manpage, \eO[ABCD] sequences are used for
+	// Ctrl-Shift-modified arrow keys; however, this doesn't seem to be true for
+	// urxvt 9.22 packaged by Debian; those keys simply send the same sequence
+	// as Ctrl-modified keys (\eO[abcd]).
+	'A': {ui.Up, 0}, 'B': {ui.Down, 0}, 'C': {ui.Right, 0}, 'D': {ui.Left, 0},
+	'H': {ui.Home, 0}, 'F': {ui.End, 0}, 'M': {ui.Insert, 0},
+	// urxvt
+	'a': {ui.Up, ui.Ctrl}, 'b': {ui.Down, ui.Ctrl},
+	'c': {ui.Right, ui.Ctrl}, 'd': {ui.Left, ui.Ctrl},
+	// xterm, urxvt, tmux
+	'P': {ui.F1, 0}, 'Q': {ui.F2, 0}, 'R': {ui.F3, 0}, 'S': {ui.F4, 0},
+}
+
+// Tables for CSI-style key sequences. A CSI sequence is \e[ followed by zero or
+// more numerical arguments (separated by semicolons), ending in a non-numeric,
+// non-semicolon rune. They are used for many purposes, and CSI-style key
+// sequences are a subset of them.
+//
+// There are several variants of CSI-style key sequences; see comments above the
+// respective tables. In all variants, modifier keys are encoded in numerical
+// arguments; see xtermModify. Note that although the set of possible sequences
+// make it possible to express a very complete set of key combinations, they are
+// not always sent by terminals. For instance, many (if not most) terminals will
+// send the same sequence for Up when Shift-Up is pressed, even if Shift-Up is
+// expressible using the escape sequences described below.
+
+// CSI-style key sequences identified by the last rune. For instance, \e[A is
+// Up. When modified, two numerical arguments are added, the first always beging
+// 1 and the second identifying the modifier. For instance, \e1;5A is Ctrl-Up.
+var csiSeqByLast = map[rune]ui.Key{
+	// xterm, urxvt, tmux
+	'A': {ui.Up, 0}, 'B': {ui.Down, 0}, 'C': {ui.Right, 0}, 'D': {ui.Left, 0},
+	// urxvt
+	'a': {ui.Up, ui.Shift}, 'b': {ui.Down, ui.Shift},
+	'c': {ui.Right, ui.Shift}, 'd': {ui.Left, ui.Shift},
+	// xterm (Terminal.app only sends those in alternate screen)
 	'H': {ui.Home, 0}, 'F': {ui.End, 0},
+	// xterm, urxvt, tmux
 	'Z': {ui.Tab, ui.Shift},
 }
 
-// CSI-style key sequences ending with '~' and can be identified based on the
-// only number argument. For instance, \e[~ is Home. When they are
-// modified, they take two arguments, first being 1 and second identifying the
-// modifier (see xtermModify). For instance, \e[1;4~ is Shift-Alt-Home.
-var keyByNum0 = map[int]rune{
-	1: ui.Home, 2: ui.Insert, 3: ui.Delete, 4: ui.End,
+// CSI-style key sequences ending with '~' and identified by one numerical
+// argument. For instance, \e[3~ is Delete. When modified, an additional
+// argument identifies the modifier; for instance, \e[3;5~ is Ctrl-Delete.
+//
+// An alternative encoding of the modifier key, only known to be used by urxvt
+// (or for that matter, likely also rxvt) is to change the last rune: '$' for
+// Shift, '^' for Ctrl, and '@' for Ctrl+Shift. The numeric argument is kept
+// unchanged. For instance, \e[3^ is Ctrl-Delete.
+var csiSeqTilde = map[int]rune{
+	// tmux (NOTE: urxvt uses the pair for Find/Select)
+	1: ui.Home, 4: ui.End,
+	// xterm (Terminal.app sends ^M for Fn+Enter), urxvt, tmux
+	2: ui.Insert,
+	// xterm, urxvt, tmux
+	3: ui.Delete,
+	// xterm (Terminal.app only sends those in alternate screen), urxvt, tmux
+	// NOTE: called Prior/Next in urxvt manpage
 	5: ui.PageUp, 6: ui.PageDown,
+	// urxvt
+	7: ui.Home, 8: ui.End,
+	// urxvt
 	11: ui.F1, 12: ui.F2, 13: ui.F3, 14: ui.F4,
+	// xterm, urxvt, tmux
+	// NOTE: 16 and 22 are unused
 	15: ui.F5, 17: ui.F6, 18: ui.F7, 19: ui.F8,
 	20: ui.F9, 21: ui.F10, 23: ui.F11, 24: ui.F12,
 }
 
-// CSI-style key sequences ending with '~', with 27 as the first numeric
-// argument. For instance, \e[27;9~ is Tab.
+// CSI-style key sequences ending with '~', with the first argument always 27,
+// the second argument identifying the modifier, and the third argument
+// identifying the key. For instance, \e[27;5;9~ is Ctrl-Tab.
 //
-// The list is taken blindly from tmux source xterm-keys.c. I don't have a
-// keyboard-terminal combination that generate such sequences, but assumably
-// some PC keyboard with a numpad can.
-var keyByNum2 = map[int]rune{
+// NOTE(xiaq): The list is taken blindly from xterm-keys.c in the tmux source
+// tree. I do not have a keyboard-terminal combination that generate such
+// sequences, but assumably they are generated by some terminals for numpad
+// inputs.
+var csiSeqTilde27 = map[int]rune{
 	9: '\t', 13: '\r',
 	33: '!', 35: '#', 39: '\'', 40: '(', 41: ')', 43: '+', 44: ',', 45: '-',
 	46: '.',
@@ -370,9 +414,10 @@ var keyByNum2 = map[int]rune{
 	58: ':', 59: ';', 60: '<', 61: '=', 62: '>', 63: ';',
 }
 
-// parseCSI parses a CSI-style key sequence.
+// parseCSI parses a CSI-style key sequence. See comments above for all the 3
+// variants this function handles.
 func parseCSI(nums []int, last rune, seq string) ui.Key {
-	if k, ok := keyByLast[last]; ok {
+	if k, ok := csiSeqByLast[last]; ok {
 		if len(nums) == 0 {
 			// Unmodified: \e[A (Up)
 			return k
@@ -384,23 +429,40 @@ func parseCSI(nums []int, last rune, seq
 		}
 	}
 
-	if last == '~' {
+	switch last {
+	case '~':
 		if len(nums) == 1 || len(nums) == 2 {
-			if r, ok := keyByNum0[nums[0]]; ok {
+			if r, ok := csiSeqTilde[nums[0]]; ok {
 				k := ui.Key{r, 0}
 				if len(nums) == 1 {
-					// Unmodified: \e[5~ (PageUp)
+					// Unmodified: \e[5~ (e.g. PageUp)
 					return k
 				}
-				// Modified: \e[5;5~ (Ctrl-PageUp)
+				// Modified: \e[5;5~ (e.g. Ctrl-PageUp)
 				return xtermModify(k, nums[1], seq)
 			}
 		} else if len(nums) == 3 && nums[0] == 27 {
-			if r, ok := keyByNum2[nums[2]]; ok {
+			if r, ok := csiSeqTilde27[nums[2]]; ok {
 				k := ui.Key{r, 0}
 				return xtermModify(k, nums[1], seq)
 			}
 		}
+	case '$', '^', '@':
+		// Modified by urxvt; see comment above csiSeqTilde.
+		if len(nums) == 1 {
+			if r, ok := csiSeqTilde[nums[0]]; ok {
+				var mod ui.Mod
+				switch last {
+				case '$':
+					mod = ui.Shift
+				case '^':
+					mod = ui.Ctrl
+				case '@':
+					mod = ui.Shift | ui.Ctrl
+				}
+				return ui.Key{r, mod}
+			}
+		}
 	}
 
 	return ui.Key{}
diff -pruN 0.11+ds1-5/edit/tty/setup.go 0.12+ds1-1/edit/tty/setup.go
--- 0.11+ds1-5/edit/tty/setup.go	2018-01-18 02:03:11.000000000 +0000
+++ 0.12+ds1-1/edit/tty/setup.go	2018-07-11 02:32:25.000000000 +0000
@@ -83,6 +83,12 @@ func restoreVT(out *os.File) error {
 	}
 	// Disable bracketed paste.
 	s += "\033[?2004l"
+	// Move the cursor to the first row, even if we haven't written anything
+	// visible. This is because the terminal driver might not be smart enough to
+	// recognize some escape sequences as invisible and wrongly assume that we
+	// are not in the first column, which can mess up with tabs. See
+	// https://github.com/elves/elvish/issues/629 for an example.
+	s += "\r"
 	_, err := out.WriteString(s)
 	return err
 }
diff -pruN 0.11+ds1-5/edit/tty/setup_unix_test.go 0.12+ds1-1/edit/tty/setup_unix_test.go
--- 0.11+ds1-5/edit/tty/setup_unix_test.go	2018-01-18 02:03:11.000000000 +0000
+++ 0.12+ds1-1/edit/tty/setup_unix_test.go	2018-07-11 02:32:25.000000000 +0000
@@ -1,4 +1,4 @@
-// +build !windows,plan9
+// +build !windows,!plan9
 
 package tty
 
@@ -11,12 +11,12 @@ import (
 func TestSetupTerminal(t *testing.T) {
 	pty, tty, err := pty.Open()
 	if err != nil {
-		t.Errorf("cannot open pty for testing setupTerminal")
+		t.Skip("cannot open pty for testing setupTerminal")
 	}
 	defer pty.Close()
 	defer tty.Close()
 
-	_, err = setup(tty)
+	_, err = setup(tty, tty)
 	if err != nil {
 		t.Errorf("setupTerminal returns an error")
 	}
diff -pruN 0.11+ds1-5/edit/ui/buffer_builder.go 0.12+ds1-1/edit/ui/buffer_builder.go
--- 0.11+ds1-5/edit/ui/buffer_builder.go	1970-01-01 00:00:00.000000000 +0000
+++ 0.12+ds1-1/edit/ui/buffer_builder.go	2018-07-11 02:32:25.000000000 +0000
@@ -0,0 +1,181 @@
+package ui
+
+import (
+	"strings"
+
+	"github.com/elves/elvish/util"
+)
+
+// BufferBuilder supports building of Buffer.
+type BufferBuilder struct {
+	Width, Col, Indent int
+	// EagerWrap controls whether to wrap line as soon as the cursor reaches the
+	// right edge of the terminal. This is not often desirable as it creates
+	// unneessary line breaks, but is is useful when echoing the user input.
+	// will otherwise
+	EagerWrap bool
+	// Lines the content of the buffer.
+	Lines [][]Cell
+	// Dot is what the user perceives as the cursor.
+	Dot Pos
+}
+
+// NewBufferBuilder makes a new BufferBuilder, initially with one empty line.
+func NewBufferBuilder(width int) *BufferBuilder {
+	return &BufferBuilder{Width: width, Lines: [][]Cell{make([]Cell, 0, width)}}
+}
+
+func (bb *BufferBuilder) Cursor() Pos {
+	return Pos{len(bb.Lines) - 1, bb.Col}
+}
+
+// Buffer returns a Buffer built by the BufferBuilder.
+func (bb *BufferBuilder) Buffer() *Buffer {
+	return NewBuffer(bb.Width).SetLines(bb.Lines...).SetDot(bb.Dot)
+}
+
+func (bb *BufferBuilder) SetIndent(indent int) *BufferBuilder {
+	bb.Indent = indent
+	return bb
+}
+
+func (bb *BufferBuilder) SetEagerWrap(v bool) *BufferBuilder {
+	bb.EagerWrap = v
+	return bb
+}
+
+func (bb *BufferBuilder) SetLines(lines ...[]Cell) *BufferBuilder {
+	bb.Lines = lines
+	bb.Col = CellsWidth(lines[len(lines)-1])
+	return bb
+}
+
+func (bb *BufferBuilder) SetDot(dot Pos) *BufferBuilder {
+	bb.Dot = dot
+	return bb
+}
+
+func (bb *BufferBuilder) SetDotToCursor() *BufferBuilder {
+	return bb.SetDot(bb.Cursor())
+}
+
+func (bb *BufferBuilder) appendLine() {
+	bb.Lines = append(bb.Lines, make([]Cell, 0, bb.Width))
+	bb.Col = 0
+}
+
+func (bb *BufferBuilder) appendCell(c Cell) {
+	n := len(bb.Lines)
+	bb.Lines[n-1] = append(bb.Lines[n-1], c)
+	bb.Col += int(c.Width)
+}
+
+// Newline starts a newline.
+func (bb *BufferBuilder) Newline() {
+	bb.appendLine()
+
+	if bb.Indent > 0 {
+		for i := 0; i < bb.Indent; i++ {
+			bb.appendCell(Cell{Text: " ", Width: 1})
+		}
+	}
+}
+
+var styleForControlChar = Styles{"inverse"}
+
+// Write writes a single rune to a buffer, wrapping the line when needed. If the
+// rune is a control character, it will be written using the caret notation
+// (like ^X) and gets the additional style of styleForControlChar.
+func (bb *BufferBuilder) Write(r rune, style string) *BufferBuilder {
+	if r == '\n' {
+		bb.Newline()
+		return bb
+	}
+	wd := util.Wcwidth(r)
+	c := Cell{string(r), byte(wd), style}
+	if r < 0x20 || r == 0x7f {
+		wd = 2
+		if style != "" {
+			style = style + ";" + styleForControlChar.String()
+		} else {
+			style = styleForControlChar.String()
+		}
+		c = Cell{"^" + string(r^0x40), 2, style}
+	}
+
+	if bb.Col+wd > bb.Width {
+		bb.Newline()
+		bb.appendCell(c)
+	} else {
+		bb.appendCell(c)
+		if bb.Col == bb.Width && bb.EagerWrap {
+			bb.Newline()
+		}
+	}
+	return bb
+}
+
+// WriteString writes a string to a buffer, with one style.
+func (bb *BufferBuilder) WriteString(text, style string) *BufferBuilder {
+	for _, r := range text {
+		bb.Write(r, style)
+	}
+	return bb
+}
+
+// WriteSpaces writes w spaces.
+func (bb *BufferBuilder) WriteSpaces(w int, style string) *BufferBuilder {
+	return bb.WriteString(strings.Repeat(" ", w), style)
+}
+
+// WriteStyleds writes a slice of styled structs.
+func (bb *BufferBuilder) WriteStyleds(ss []*Styled) *BufferBuilder {
+	for _, s := range ss {
+		bb.WriteString(s.Text, s.Styles.String())
+	}
+	return bb
+}
+
+// Extend adds all lines from b2 to the bottom of this buffer. If moveDot is
+// true, the dot is updated to match the dot of b2.
+func (bb *BufferBuilder) Extend(b2 *Buffer, moveDot bool) *BufferBuilder {
+	if b2 != nil && b2.Lines != nil {
+		if moveDot {
+			bb.Dot.Line = b2.Dot.Line + len(bb.Lines)
+			bb.Dot.Col = b2.Dot.Col
+		}
+		bb.Lines = append(bb.Lines, b2.Lines...)
+		bb.Col = b2.Col
+	}
+	return bb
+}
+
+// ExtendRight extends bb to the right. It pads each line in bb to be at least of
+// width w and appends the corresponding line in b2 to it, making new lines in bb
+// when b2 has more lines than bb.
+// BUG(xiaq): after calling ExtendRight, the widths of some lines can exceed
+// bb.width.
+func (bb *BufferBuilder) ExtendRight(b2 *Buffer, w int) *BufferBuilder {
+	i := 0
+	for ; i < len(bb.Lines) && i < len(b2.Lines); i++ {
+		if w0 := CellsWidth(bb.Lines[i]); w0 < w {
+			bb.Lines[i] = append(bb.Lines[i], makeSpacing(w-w0)...)
+		}
+		bb.Lines[i] = append(bb.Lines[i], b2.Lines[i]...)
+	}
+	for ; i < len(b2.Lines); i++ {
+		row := append(makeSpacing(w), b2.Lines[i]...)
+		bb.Lines = append(bb.Lines, row)
+	}
+	bb.Col = CellsWidth(bb.Lines[len(bb.Lines)-1])
+	return bb
+}
+
+func makeSpacing(n int) []Cell {
+	s := make([]Cell, n)
+	for i := 0; i < n; i++ {
+		s[i].Text = " "
+		s[i].Width = 1
+	}
+	return s
+}
diff -pruN 0.11+ds1-5/edit/ui/buffer_builder_test.go 0.12+ds1-1/edit/ui/buffer_builder_test.go
--- 0.11+ds1-5/edit/ui/buffer_builder_test.go	1970-01-01 00:00:00.000000000 +0000
+++ 0.12+ds1-1/edit/ui/buffer_builder_test.go	2018-07-11 02:32:25.000000000 +0000
@@ -0,0 +1,186 @@
+package ui
+
+import (
+	"reflect"
+	"testing"
+)
+
+var bufferBuilderWritesTests = []struct {
+	bb    *BufferBuilder
+	text  string
+	style string
+	want  *Buffer
+}{
+	// Writing nothing.
+	{NewBufferBuilder(10), "", "", NewBuffer(10)},
+	// Writing a single rune.
+	{NewBufferBuilder(10), "a", "1", NewBuffer(10).SetLines([]Cell{{"a", 1, "1"}})},
+	// Writing control character.
+	{NewBufferBuilder(10), "\033", "",
+		NewBuffer(10).SetLines(
+			[]Cell{{"^[", 2, styleForControlChar.String()}},
+		)},
+	// Writing styled control character.
+	{NewBufferBuilder(10), "a\033b", "1",
+		NewBuffer(10).SetLines(
+			[]Cell{
+				{"a", 1, "1"},
+				{"^[", 2, "1;" + styleForControlChar.String()},
+				{"b", 1, "1"},
+			},
+		)},
+	// Writing text containing a newline.
+	{NewBufferBuilder(10), "a\nb", "1",
+		NewBuffer(10).SetLines(
+			[]Cell{{"a", 1, "1"}}, []Cell{{"b", 1, "1"}},
+		)},
+	// Writing text containing a newline when there is indent.
+	{NewBufferBuilder(10).SetIndent(2), "a\nb", "1",
+		NewBuffer(10).SetLines(
+			[]Cell{{"a", 1, "1"}},
+			[]Cell{{" ", 1, ""}, {" ", 1, ""}, {"b", 1, "1"}},
+		)},
+	// Writing long text that triggers wrapping.
+	{NewBufferBuilder(4), "aaaab", "1",
+		NewBuffer(4).SetLines(
+			[]Cell{{"a", 1, "1"}, {"a", 1, "1"}, {"a", 1, "1"}, {"a", 1, "1"}},
+			[]Cell{{"b", 1, "1"}},
+		)},
+	// Writing long text that triggers wrapping when there is indent.
+	{NewBufferBuilder(4).SetIndent(2), "aaaab", "1",
+		NewBuffer(4).SetLines(
+			[]Cell{{"a", 1, "1"}, {"a", 1, "1"}, {"a", 1, "1"}, {"a", 1, "1"}},
+			[]Cell{{" ", 1, ""}, {" ", 1, ""}, {"b", 1, "1"}},
+		)},
+	// Writing long text that triggers eager wrapping.
+	{NewBufferBuilder(4).SetIndent(2).SetEagerWrap(true), "aaaa", "1",
+		NewBuffer(4).SetLines(
+			[]Cell{{"a", 1, "1"}, {"a", 1, "1"}, {"a", 1, "1"}, {"a", 1, "1"}},
+			[]Cell{{" ", 1, ""}, {" ", 1, ""}},
+		)},
+}
+
+// TestBufferWrites tests BufferBuilder.Writes by calling Writes on a
+// BufferBuilder and see if the built Buffer matches what is expected.
+func TestBufferBuilderWrites(t *testing.T) {
+	for _, test := range bufferBuilderWritesTests {
+		bb := test.bb
+		bb.WriteString(test.text, test.style)
+		buf := bb.Buffer()
+		if !reflect.DeepEqual(buf, test.want) {
+			t.Errorf("buf.writes(%q, %q) makes it %v, want %v",
+				test.text, test.style, buf, test.want)
+		}
+	}
+}
+
+var bufferBuilderExtendTests = []struct {
+	bb      *BufferBuilder
+	buf2    *Buffer
+	moveDot bool
+	want    *Buffer
+}{
+	{
+		NewBufferBuilder(10).SetLines(
+			[]Cell{{"a", 1, ""}}, []Cell{{"b", 1, ""}}),
+		NewBuffer(11).SetLines([]Cell{{"c", 1, ""}}, []Cell{{"d", 1, ""}}),
+		false,
+		NewBuffer(10).SetLines(
+			[]Cell{{"a", 1, ""}}, []Cell{{"b", 1, ""}},
+			[]Cell{{"c", 1, ""}}, []Cell{{"d", 1, ""}}),
+	},
+	// Moving dot.
+	{
+		NewBufferBuilder(10).SetLines(
+			[]Cell{{"a", 1, ""}}, []Cell{{"b", 1, ""}}),
+		NewBuffer(11).SetLines(
+			[]Cell{{"c", 1, ""}}, []Cell{{"d", 1, ""}},
+		).SetDot(Pos{1, 1}),
+		true,
+		NewBuffer(10).SetLines(
+			[]Cell{{"a", 1, ""}}, []Cell{{"b", 1, ""}},
+			[]Cell{{"c", 1, ""}}, []Cell{{"d", 1, ""}},
+		).SetDot(Pos{3, 1}),
+	},
+}
+
+func TestBufferBuilderExtend(t *testing.T) {
+	for _, test := range bufferBuilderExtendTests {
+		bb := test.bb
+		bb.Extend(test.buf2, test.moveDot)
+		b := bb.Buffer()
+		if !reflect.DeepEqual(b, test.want) {
+			t.Errorf("buf.extend(%v, %v) makes it %v, want %v",
+				test.buf2, test.moveDot, b, test.want)
+		}
+	}
+}
+
+var bufferBuilderExtendRightTests = []struct {
+	bb   *BufferBuilder
+	buf2 *Buffer
+	w    int
+	want *Buffer
+}{
+	// No padding, equal height.
+	{
+		NewBufferBuilder(10).SetLines([]Cell{{"a", 1, ""}}, []Cell{}),
+		NewBuffer(11).SetLines([]Cell{{"c", 1, ""}}, []Cell{{"d", 1, ""}}),
+		0,
+		NewBuffer(10).SetLines(
+			[]Cell{{"a", 1, ""}, {"c", 1, ""}},
+			[]Cell{{"d", 1, ""}},
+		),
+	},
+	// With padding.
+	{
+		NewBufferBuilder(10).SetLines(
+			[]Cell{{"a", 1, ""}}, []Cell{{"b", 1, ""}}),
+		NewBuffer(11).SetLines([]Cell{{"c", 1, ""}}, []Cell{{"d", 1, ""}}),
+		2,
+		NewBuffer(10).SetLines(
+			[]Cell{{"a", 1, ""}, {" ", 1, ""}, {"c", 1, ""}},
+			[]Cell{{"b", 1, ""}, {" ", 1, ""}, {"d", 1, ""}},
+		),
+	},
+	// buf is higher.
+	{
+		NewBufferBuilder(10).SetLines(
+			[]Cell{{"a", 1, ""}},
+			[]Cell{{"b", 1, ""}},
+			[]Cell{{"x", 1, ""}},
+		),
+		NewBuffer(11).SetLines([]Cell{{"c", 1, ""}}, []Cell{{"d", 1, ""}}),
+		1,
+		NewBuffer(10).SetLines(
+			[]Cell{{"a", 1, ""}, {"c", 1, ""}},
+			[]Cell{{"b", 1, ""}, {"d", 1, ""}},
+			[]Cell{{"x", 1, ""}},
+		),
+	},
+	// buf2 is higher.
+	{
+		NewBufferBuilder(10).SetLines(
+			[]Cell{{"a", 1, ""}}, []Cell{{"b", 1, ""}}),
+		NewBuffer(11).SetLines(
+			[]Cell{{"c", 1, ""}}, []Cell{{"d", 1, ""}}, []Cell{{"e", 1, ""}}),
+		1,
+		NewBuffer(10).SetLines(
+			[]Cell{{"a", 1, ""}, {"c", 1, ""}},
+			[]Cell{{"b", 1, ""}, {"d", 1, ""}},
+			[]Cell{{" ", 1, ""}, {"e", 1, ""}},
+		),
+	},
+}
+
+func TestBufferBuilderExtendRight(t *testing.T) {
+	for _, test := range bufferBuilderExtendRightTests {
+		bb := test.bb
+		bb.ExtendRight(test.buf2, test.w)
+		b := bb.Buffer()
+		if !reflect.DeepEqual(b, test.want) {
+			t.Errorf("buf.extendRight(%v, %v) makes it %v, want %v",
+				test.buf2, test.w, b, test.want)
+		}
+	}
+}
diff -pruN 0.11+ds1-5/edit/ui/buffer.go 0.12+ds1-1/edit/ui/buffer.go
--- 0.11+ds1-5/edit/ui/buffer.go	2018-01-18 02:03:11.000000000 +0000
+++ 0.12+ds1-1/edit/ui/buffer.go	2018-07-11 02:32:25.000000000 +0000
@@ -1,11 +1,5 @@
 package ui
 
-import (
-	"strings"
-
-	"github.com/elves/elvish/util"
-)
-
 // Cell is an indivisible unit on the screen. It is not necessarily 1 column
 // wide.
 type Cell struct {
@@ -50,12 +44,7 @@ func CompareCells(r1, r2 []Cell) (bool,
 // exactly match the terminal's idea of the width of characters (wcwidth) and
 // where to insert soft carriage returns, so there could be bugs.
 type Buffer struct {
-	Width, Col, Indent int
-	// EagerWrap controls whether to wrap line as soon as the cursor reaches the
-	// right edge of the terminal. This is not often desirable as it creates
-	// unneessary line breaks, but is is useful when echoing the user input.
-	// will otherwise
-	EagerWrap bool
+	Width, Col int
 	// Lines the content of the buffer.
 	Lines [][]Cell
 	// Dot is what the user perceives as the cursor.
@@ -67,16 +56,6 @@ func NewBuffer(width int) *Buffer {
 	return &Buffer{Width: width, Lines: [][]Cell{make([]Cell, 0, width)}}
 }
 
-func (b *Buffer) SetIndent(indent int) *Buffer {
-	b.Indent = indent
-	return b
-}
-
-func (b *Buffer) SetEagerWrap(v bool) *Buffer {
-	b.EagerWrap = v
-	return b
-}
-
 func (b *Buffer) SetLines(lines ...[]Cell) *Buffer {
 	b.Lines = lines
 	b.Col = CellsWidth(lines[len(lines)-1])
@@ -103,84 +82,6 @@ func BuffersHeight(bufs ...*Buffer) (l i
 	return
 }
 
-// Low level buffer mutations.
-
-func (b *Buffer) appendLine() {
-	b.Lines = append(b.Lines, make([]Cell, 0, b.Width))
-	b.Col = 0
-}
-
-func (b *Buffer) appendCell(c Cell) {
-	n := len(b.Lines)
-	b.Lines[n-1] = append(b.Lines[n-1], c)
-	b.Col += int(c.Width)
-}
-
-// High-level buffer mutations.
-
-// Newline starts a newline.
-func (b *Buffer) Newline() {
-	b.appendLine()
-
-	if b.Indent > 0 {
-		for i := 0; i < b.Indent; i++ {
-			b.appendCell(Cell{Text: " ", Width: 1})
-		}
-	}
-}
-
-var styleForControlChar = Styles{"inverse"}
-
-// Write writes a single rune to a buffer, wrapping the line when needed. If the
-// rune is a control character, it will be written using the caret notation
-// (like ^X) and gets the additional style of styleForControlChar.
-func (b *Buffer) Write(r rune, style string) {
-	if r == '\n' {
-		b.Newline()
-		return
-	}
-	wd := util.Wcwidth(r)
-	c := Cell{string(r), byte(wd), style}
-	if r < 0x20 || r == 0x7f {
-		wd = 2
-		if style != "" {
-			style = style + ";" + styleForControlChar.String()
-		} else {
-			style = styleForControlChar.String()
-		}
-		c = Cell{"^" + string(r^0x40), 2, style}
-	}
-
-	if b.Col+wd > b.Width {
-		b.Newline()
-		b.appendCell(c)
-	} else {
-		b.appendCell(c)
-		if b.Col == b.Width && b.EagerWrap {
-			b.Newline()
-		}
-	}
-}
-
-// WriteString writes a string to a buffer, with one style.
-func (b *Buffer) WriteString(text, style string) {
-	for _, r := range text {
-		b.Write(r, style)
-	}
-}
-
-// WriteSpaces writes w spaces.
-func (b *Buffer) WriteSpaces(w int, style string) {
-	b.WriteString(strings.Repeat(" ", w), style)
-}
-
-// WriteStyleds writes a slice of styled structs.
-func (b *Buffer) WriteStyleds(ss []*Styled) {
-	for _, s := range ss {
-		b.WriteString(s.Text, s.Styles.String())
-	}
-}
-
 // TrimToLines trims a buffer to the lines [low, high).
 func (b *Buffer) TrimToLines(low, high int) {
 	for i := 0; i < low; i++ {
@@ -195,45 +96,3 @@ func (b *Buffer) TrimToLines(low, high i
 		b.Dot.Line = 0
 	}
 }
-
-// Extend adds all lines from b2 to the bottom of this buffer. If moveDot is
-// true, the dot is updated to match the dot of b2.
-func (b *Buffer) Extend(b2 *Buffer, moveDot bool) {
-	if b2 != nil && b2.Lines != nil {
-		if moveDot {
-			b.Dot.Line = b2.Dot.Line + len(b.Lines)
-			b.Dot.Col = b2.Dot.Col
-		}
-		b.Lines = append(b.Lines, b2.Lines...)
-		b.Col = b2.Col
-	}
-}
-
-// ExtendRight extends b to the right. It pads each line in b to be at least of
-// width w and appends the corresponding line in b2 to it, making new lines in b
-// when b2 has more lines than b.
-// BUG(xiaq): after calling ExtendRight, the widths of some lines can exceed
-// b.width.
-func (b *Buffer) ExtendRight(b2 *Buffer, w int) {
-	i := 0
-	for ; i < len(b.Lines) && i < len(b2.Lines); i++ {
-		if w0 := CellsWidth(b.Lines[i]); w0 < w {
-			b.Lines[i] = append(b.Lines[i], makeSpacing(w-w0)...)
-		}
-		b.Lines[i] = append(b.Lines[i], b2.Lines[i]...)
-	}
-	for ; i < len(b2.Lines); i++ {
-		row := append(makeSpacing(w), b2.Lines[i]...)
-		b.Lines = append(b.Lines, row)
-	}
-	b.Col = CellsWidth(b.Lines[len(b.Lines)-1])
-}
-
-func makeSpacing(n int) []Cell {
-	s := make([]Cell, n)
-	for i := 0; i < n; i++ {
-		s[i].Text = " "
-		s[i].Width = 1
-	}
-	return s
-}
diff -pruN 0.11+ds1-5/edit/ui/buffer_test.go 0.12+ds1-1/edit/ui/buffer_test.go
--- 0.11+ds1-5/edit/ui/buffer_test.go	2018-01-18 02:03:11.000000000 +0000
+++ 0.12+ds1-1/edit/ui/buffer_test.go	2018-07-11 02:32:25.000000000 +0000
@@ -109,74 +109,6 @@ func TestBuffersHeight(t *testing.T) {
 	}
 }
 
-var bufferWritesTests = []struct {
-	buf   *Buffer
-	text  string
-	style string
-	want  *Buffer
-}{
-	// Writing nothing.
-	{NewBuffer(10), "", "", NewBuffer(10)},
-	// Writing a single rune.
-	{NewBuffer(10), "a", "1", NewBuffer(10).SetLines([]Cell{{"a", 1, "1"}})},
-	// Writing control character.
-	{NewBuffer(10), "\033", "",
-		NewBuffer(10).SetLines(
-			[]Cell{{"^[", 2, styleForControlChar.String()}},
-		)},
-	// Writing styled control character.
-	{NewBuffer(10), "a\033b", "1",
-		NewBuffer(10).SetLines(
-			[]Cell{
-				{"a", 1, "1"},
-				{"^[", 2, "1;" + styleForControlChar.String()},
-				{"b", 1, "1"},
-			},
-		)},
-	// Writing text containing a newline.
-	{NewBuffer(10), "a\nb", "1",
-		NewBuffer(10).SetLines(
-			[]Cell{{"a", 1, "1"}}, []Cell{{"b", 1, "1"}},
-		)},
-	// Writing text containing a newline when there is indent.
-	{NewBuffer(10).SetIndent(2), "a\nb", "1",
-		NewBuffer(10).SetIndent(2).SetLines(
-			[]Cell{{"a", 1, "1"}},
-			[]Cell{{" ", 1, ""}, {" ", 1, ""}, {"b", 1, "1"}},
-		)},
-	// Writing long text that triggers wrapping.
-	{NewBuffer(4), "aaaab", "1",
-		NewBuffer(4).SetLines(
-			[]Cell{{"a", 1, "1"}, {"a", 1, "1"}, {"a", 1, "1"}, {"a", 1, "1"}},
-			[]Cell{{"b", 1, "1"}},
-		)},
-	// Writing long text that triggers wrapping when there is indent.
-	{NewBuffer(4).SetIndent(2), "aaaab", "1",
-		NewBuffer(4).SetIndent(2).SetLines(
-			[]Cell{{"a", 1, "1"}, {"a", 1, "1"}, {"a", 1, "1"}, {"a", 1, "1"}},
-			[]Cell{{" ", 1, ""}, {" ", 1, ""}, {"b", 1, "1"}},
-		)},
-	// Writing long text that triggers eager wrapping.
-	{NewBuffer(4).SetIndent(2).SetEagerWrap(true), "aaaa", "1",
-		NewBuffer(4).SetIndent(2).SetEagerWrap(true).SetLines(
-			[]Cell{{"a", 1, "1"}, {"a", 1, "1"}, {"a", 1, "1"}, {"a", 1, "1"}},
-			[]Cell{{" ", 1, ""}, {" ", 1, ""}},
-		)},
-}
-
-// TestBufferWrites tests buffer.writes by calling writes on a buffer and see if
-// the buffer matches what is expected.
-func TestBufferWrites(t *testing.T) {
-	for _, test := range bufferWritesTests {
-		b := test.buf
-		b.WriteString(test.text, test.style)
-		if !reflect.DeepEqual(b, test.want) {
-			t.Errorf("buf.writes(%q, %q) makes it %v, want %v",
-				test.text, test.style, b, test.want)
-		}
-	}
-}
-
 var bufferTrimToLinesTests = []struct {
 	buf  *Buffer
 	low  int
@@ -224,110 +156,3 @@ func TestBufferTrimToLines(t *testing.T)
 		}
 	}
 }
-
-var bufferExtendTests = []struct {
-	buf     *Buffer
-	buf2    *Buffer
-	moveDot bool
-	want    *Buffer
-}{
-	{
-		NewBuffer(10).SetLines([]Cell{{"a", 1, ""}}, []Cell{{"b", 1, ""}}),
-		NewBuffer(11).SetLines([]Cell{{"c", 1, ""}}, []Cell{{"d", 1, ""}}),
-		false,
-		NewBuffer(10).SetLines(
-			[]Cell{{"a", 1, ""}}, []Cell{{"b", 1, ""}},
-			[]Cell{{"c", 1, ""}}, []Cell{{"d", 1, ""}},
-		),
-	},
-	// Moving dot.
-	{
-		NewBuffer(10).SetLines([]Cell{{"a", 1, ""}}, []Cell{{"b", 1, ""}}),
-		NewBuffer(11).SetLines(
-			[]Cell{{"c", 1, ""}}, []Cell{{"d", 1, ""}},
-		).SetDot(Pos{1, 1}),
-		true,
-		NewBuffer(10).SetLines(
-			[]Cell{{"a", 1, ""}}, []Cell{{"b", 1, ""}},
-			[]Cell{{"c", 1, ""}}, []Cell{{"d", 1, ""}},
-		).SetDot(Pos{3, 1}),
-	},
-}
-
-func TestExtend(t *testing.T) {
-	for _, test := range bufferExtendTests {
-		b := test.buf
-		b.Extend(test.buf2, test.moveDot)
-		if !reflect.DeepEqual(b, test.want) {
-			t.Errorf("buf.extend(%v, %v) makes it %v, want %v",
-				test.buf2, test.moveDot, b, test.want)
-		}
-	}
-}
-
-var bufferExtendRightTests = []struct {
-	buf  *Buffer
-	buf2 *Buffer
-	w    int
-	want *Buffer
-}{
-	// No padding, equal height.
-	{
-		NewBuffer(10).SetLines([]Cell{{"a", 1, ""}}, []Cell{}),
-		NewBuffer(11).SetLines([]Cell{{"c", 1, ""}}, []Cell{{"d", 1, ""}}),
-		0,
-		NewBuffer(10).SetLines(
-			[]Cell{{"a", 1, ""}, {"c", 1, ""}},
-			[]Cell{{"d", 1, ""}},
-		),
-	},
-	// With padding.
-	{
-		NewBuffer(10).SetLines([]Cell{{"a", 1, ""}}, []Cell{{"b", 1, ""}}),
-		NewBuffer(11).SetLines([]Cell{{"c", 1, ""}}, []Cell{{"d", 1, ""}}),
-		2,
-		NewBuffer(10).SetLines(
-			[]Cell{{"a", 1, ""}, {" ", 1, ""}, {"c", 1, ""}},
-			[]Cell{{"b", 1, ""}, {" ", 1, ""}, {"d", 1, ""}},
-		),
-	},
-	// buf is higher.
-	{
-		NewBuffer(10).SetLines(
-			[]Cell{{"a", 1, ""}},
-			[]Cell{{"b", 1, ""}},
-			[]Cell{{"x", 1, ""}},
-		),
-		NewBuffer(11).SetLines([]Cell{{"c", 1, ""}}, []Cell{{"d", 1, ""}}),
-		1,
-		NewBuffer(10).SetLines(
-			[]Cell{{"a", 1, ""}, {"c", 1, ""}},
-			[]Cell{{"b", 1, ""}, {"d", 1, ""}},
-			[]Cell{{"x", 1, ""}},
-		),
-	},
-	// buf2 is higher.
-	{
-		NewBuffer(10).SetLines([]Cell{{"a", 1, ""}}, []Cell{{"b", 1, ""}}),
-		NewBuffer(11).SetLines(
-			[]Cell{{"c", 1, ""}}, []Cell{{"d", 1, ""}}, []Cell{{"e", 1, ""}},
-		),
-		1,
-		NewBuffer(10).SetLines(
-			[]Cell{{"a", 1, ""}, {"c", 1, ""}},
-			[]Cell{{"b", 1, ""}, {"d", 1, ""}},
-			[]Cell{{" ", 1, ""}, {"e", 1, ""}},
-		),
-	},
-}
-
-func TestExtendRight(t *testing.T) {
-	for _, test := range bufferExtendRightTests {
-		b := test.buf
-		b.ExtendRight(test.buf2, test.w)
-		if !reflect.DeepEqual(b, test.want) {
-			t.Errorf("buf.extendRight(%v, %v) makes it %v, want %v",
-				test.buf2, test.w, b, test.want)
-		}
-	}
-}
diff -pruN 0.11+ds1-5/edit/ui/key.go 0.12+ds1-1/edit/ui/key.go
--- 0.11+ds1-5/edit/ui/key.go	2018-01-18 02:03:11.000000000 +0000
+++ 0.12+ds1-1/edit/ui/key.go	2018-07-11 02:32:25.000000000 +0000
@@ -6,8 +6,6 @@ import (
 	"fmt"
 	"strings"
 
-	"github.com/elves/elvish/eval"
-	"github.com/elves/elvish/eval/types"
 	"github.com/elves/elvish/parse"
 	"github.com/elves/elvish/util"
 	"github.com/xiaq/persistent/hash"
@@ -214,11 +212,11 @@ func parseKey(s string) (Key, error) {
 
 // ToKey converts an Elvish Value to a Key. If the passed Value is not Key or
 // String, it throws an error.
-func ToKey(k types.Value) Key {
+func ToKey(k interface{}) Key {
 	switch k := k.(type) {
 	case Key:
 		return k
-	case types.String:
+	case string:
 		key, err := parseKey(string(k))
 		if err != nil {
 			util.Throw(err)
@@ -230,15 +228,6 @@ func ToKey(k types.Value) Key {
 	}
 }
 
-// KeyBuiltin implements the edit:key builtin.
-func KeyBuiltin(ec *eval.Frame, args []types.Value, opts map[string]types.Value) {
-	var s types.String
-	eval.ScanArgs(args, &s)
-	eval.TakeNoOpt(opts)
-
-	ec.OutputChan() <- ToKey(s)
-}
-
 // Keys implements sort.Interface.
 type Keys []Key
 
diff -pruN 0.11+ds1-5/edit/ui/render.go 0.12+ds1-1/edit/ui/render.go
--- 0.11+ds1-5/edit/ui/render.go	2018-01-18 02:03:11.000000000 +0000
+++ 0.12+ds1-1/edit/ui/render.go	2018-07-11 02:32:25.000000000 +0000
@@ -3,7 +3,7 @@ package ui
 // Renderer wraps the Render method.
 type Renderer interface {
 	// Render renders onto a Buffer.
-	Render(b *Buffer)
+	Render(bb *BufferBuilder)
 }
 
 // Render creates a new Buffer with the given width, and lets a Renderer render
@@ -12,7 +12,43 @@ func Render(r Renderer, width int) *Buff
 	if r == nil {
 		return nil
 	}
-	b := NewBuffer(width)
-	r.Render(b)
-	return b
+	bb := NewBufferBuilder(width)
+	r.Render(bb)
+	return bb.Buffer()
+}
+
+// NewModeLineRenderer returns a Renderer for a mode line.
+func NewModeLineRenderer(title, filter string) Renderer {
+	return modeLineRenderer{title, filter}
+}
+
+type modeLineRenderer struct {
+	title  string
+	filter string
+}
+
+func (ml modeLineRenderer) Render(bb *BufferBuilder) {
+	bb.WriteString(ml.title, styleForMode.String())
+	bb.WriteSpaces(1, "")
+	bb.WriteString(ml.filter, styleForFilter.String())
+	bb.Dot = bb.Cursor()
+}
+
+func NewModeLineWithScrollBarRenderer(base Renderer, n, low, high int) Renderer {
+	return &modeLineWithScrollBarRenderer{base, n, low, high}
+}
+
+type modeLineWithScrollBarRenderer struct {
+	base         Renderer
+	n, low, high int
+}
+
+func (ml modeLineWithScrollBarRenderer) Render(bb *BufferBuilder) {
+	ml.base.Render(bb)
+
+	scrollbarWidth := bb.Width - CellsWidth(bb.Lines[len(bb.Lines)-1]) - 2
+	if scrollbarWidth >= 3 {
+		bb.WriteSpaces(1, "")
+		writeHorizontalScrollbar(bb, ml.n, ml.low, ml.high, scrollbarWidth)
+	}
 }
diff -pruN 0.11+ds1-5/edit/ui/render_test.go 0.12+ds1-1/edit/ui/render_test.go
--- 0.11+ds1-5/edit/ui/render_test.go	2018-01-18 02:03:11.000000000 +0000
+++ 0.12+ds1-1/edit/ui/render_test.go	2018-07-11 02:32:25.000000000 +0000
@@ -5,8 +5,8 @@ import "testing"
 type dummyRenderer struct {
 }
 
-func (dummyRenderer) Render(b *Buffer) {
-	b.WriteString("xy", "1")
+func (dummyRenderer) Render(bb *BufferBuilder) {
+	bb.WriteString("xy", "1")
 }
 
 func TestRender(t *testing.T) {
diff -pruN 0.11+ds1-5/edit/ui/scrollbar.go 0.12+ds1-1/edit/ui/scrollbar.go
--- 0.11+ds1-5/edit/ui/scrollbar.go	1970-01-01 00:00:00.000000000 +0000
+++ 0.12+ds1-1/edit/ui/scrollbar.go	2018-07-11 02:32:25.000000000 +0000
@@ -0,0 +1,27 @@
+package ui
+
+func writeHorizontalScrollbar(bb *BufferBuilder, n, low, high, width int) {
+	slow, shigh := findScrollInterval(n, low, high, width)
+	for i := 0; i < width; i++ {
+		if slow <= i && i < shigh {
+			bb.Write(' ', styleForScrollBarThumb.String())
+		} else {
+			bb.Write('━', styleForScrollBarArea.String())
+		}
+	}
+}
+
+func findScrollInterval(n, low, high, height int) (int, int) {
+	f := func(i int) int {
+		return int(float64(i)/float64(n)*float64(height) + 0.5)
+	}
+	scrollLow, scrollHigh := f(low), f(high)
+	if scrollLow == scrollHigh {
+		if scrollHigh == high {
+			scrollLow--
+		} else {
+			scrollHigh++
+		}
+	}
+	return scrollLow, scrollHigh
+}
diff -pruN 0.11+ds1-5/edit/ui/styled.go 0.12+ds1-1/edit/ui/styled.go
--- 0.11+ds1-5/edit/ui/styled.go	2018-01-18 02:03:11.000000000 +0000
+++ 0.12+ds1-1/edit/ui/styled.go	2018-07-11 02:32:25.000000000 +0000
@@ -1,10 +1,13 @@
 package ui
 
 import (
+	"fmt"
 	"strings"
 
+	"github.com/elves/elvish/eval/vals"
 	"github.com/elves/elvish/parse"
 	"github.com/xiaq/persistent/hash"
+	"github.com/xiaq/persistent/vector"
 )
 
 // Styled is a piece of text with style.
@@ -85,7 +88,31 @@ func (s *Styled) String() string {
 }
 
 func (s *Styled) Repr(indent int) string {
-	return "(le:styled " + parse.Quote(s.Text) + " " + parse.Quote(s.Styles.String()) + ")"
+	quotedStyles := make([]string, len(s.Styles))
+	for i, st := range s.Styles {
+		quotedStyles[i] = parse.Quote(st)
+	}
+	return fmt.Sprintf("(edit:styled %s [%s]",
+		parse.Quote(s.Text), strings.Join(quotedStyles, " "))
+}
+
+func (s *Styled) Index(k interface{}) (interface{}, bool) {
+	switch k {
+	case "text":
+		return s.Text, true
+	case "styles":
+		li := vector.Empty
+		for _, st := range s.Styles {
+			li = li.Cons(st)
+		}
+		return li, true
+	default:
+		return nil, false
+	}
+}
+
+func (s *Styled) IterateKeys(f func(interface{}) bool) {
+	vals.Feed(f, "text", "styles")
 }
 
 type Styles []string
diff -pruN 0.11+ds1-5/edit/ui/style.go 0.12+ds1-1/edit/ui/style.go
--- 0.11+ds1-5/edit/ui/style.go	1970-01-01 00:00:00.000000000 +0000
+++ 0.12+ds1-1/edit/ui/style.go	2018-07-11 02:32:25.000000000 +0000
@@ -0,0 +1,9 @@
+package ui
+
+var (
+	styleForMode   = Styles{"bold", "lightgray", "bg-magenta"}
+	styleForFilter = Styles{"underlined"}
+
+	styleForScrollBarArea  = Styles{"magenta"}
+	styleForScrollBarThumb = Styles{"magenta", "inverse"}
+)
diff -pruN 0.11+ds1-5/eval/boilerplate.go 0.12+ds1-1/eval/boilerplate.go
--- 0.11+ds1-5/eval/boilerplate.go	2018-01-18 02:03:11.000000000 +0000
+++ 0.12+ds1-1/eval/boilerplate.go	2018-07-11 02:32:25.000000000 +0000
@@ -132,32 +132,6 @@ func (cp *compiler) listOps(ns []*parse.
 	return ops
 }
 
-func (cp *compiler) exceptionCaptureOp(n *parse.Chunk) ValuesOp {
-	cp.compiling(n)
-	return ValuesOp{cp.exceptionCapture(n), n.Begin(), n.End()}
-}
-
-func (cp *compiler) exceptionCaptureOps(ns []*parse.Chunk) []ValuesOp {
-	ops := make([]ValuesOp, len(ns))
-	for i, n := range ns {
-		ops[i] = cp.exceptionCaptureOp(n)
-	}
-	return ops
-}
-
-func (cp *compiler) outputCaptureOp(n *parse.Primary) ValuesOp {
-	cp.compiling(n)
-	return ValuesOp{cp.outputCapture(n), n.Begin(), n.End()}
-}
-
-func (cp *compiler) outputCaptureOps(ns []*parse.Primary) []ValuesOp {
-	ops := make([]ValuesOp, len(ns))
-	for i, n := range ns {
-		ops[i] = cp.outputCaptureOp(n)
-	}
-	return ops
-}
-
 func (cp *compiler) lambdaOp(n *parse.Primary) ValuesOp {
 	cp.compiling(n)
 	return ValuesOp{cp.lambda(n), n.Begin(), n.End()}
diff -pruN 0.11+ds1-5/eval/builtin_fn_cmd.go 0.12+ds1-1/eval/builtin_fn_cmd.go
--- 0.11+ds1-5/eval/builtin_fn_cmd.go	2018-01-18 02:03:11.000000000 +0000
+++ 0.12+ds1-1/eval/builtin_fn_cmd.go	2018-07-11 02:32:25.000000000 +0000
@@ -5,8 +5,6 @@ import (
 	"fmt"
 	"os"
 	"os/exec"
-
-	"github.com/elves/elvish/eval/types"
 )
 
 // Command and process control.
@@ -14,79 +12,50 @@ import (
 var ErrNotInSameGroup = errors.New("not in the same process group")
 
 func init() {
-	addToBuiltinFns([]*BuiltinFn{
+	addBuiltinFns(map[string]interface{}{
 		// Command resolution
-		{"resolve", resolveFn},
-		{"external", external},
-		{"has-external", hasExternal},
-		{"search-external", searchExternal},
+		"external":        external,
+		"has-external":    hasExternal,
+		"search-external": searchExternal,
 
 		// Process control
-		{"fg", fg},
-		{"exec", execFn},
-		{"exit", exit},
+		"fg":   fg,
+		"exec": execFn,
+		"exit": exit,
 	})
 }
 
-func resolveFn(ec *Frame, args []types.Value, opts map[string]types.Value) {
-	var cmd types.String
-	ScanArgs(args, &cmd)
-	TakeNoOpt(opts)
-
-	out := ec.ports[1].Chan
-	out <- resolve(string(cmd), ec)
+func external(cmd string) ExternalCmd {
+	return ExternalCmd{cmd}
 }
 
-func external(ec *Frame, args []types.Value, opts map[string]types.Value) {
-	var cmd types.String
-	ScanArgs(args, &cmd)
-	TakeNoOpt(opts)
-
-	ec.OutputChan() <- ExternalCmd{string(cmd)}
+func hasExternal(cmd string) bool {
+	_, err := exec.LookPath(cmd)
+	return err == nil
 }
 
-func hasExternal(ec *Frame, args []types.Value, opts map[string]types.Value) {
-	var cmd types.String
-	ScanArgs(args, &cmd)
-	TakeNoOpt(opts)
-
-	_, err := exec.LookPath(string(cmd))
-	ec.OutputChan() <- types.Bool(err == nil)
+func searchExternal(cmd string) (string, error) {
+	return exec.LookPath(cmd)
 }
 
-func searchExternal(ec *Frame, args []types.Value, opts map[string]types.Value) {
-	var cmd types.String
-	ScanArgs(args, &cmd)
-	TakeNoOpt(opts)
-
-	path, err := exec.LookPath(string(cmd))
-	maybeThrow(err)
-
-	out := ec.ports[1].Chan
-	out <- types.String(path)
-}
-
-func exit(ec *Frame, args []types.Value, opts map[string]types.Value) {
-	var codes []int
-	ScanArgsVariadic(args, &codes)
-	TakeNoOpt(opts)
-
-	doexit := func(i int) {
-		preExit(ec)
-		os.Exit(i)
-	}
+func exit(fm *Frame, codes ...int) error {
+	code := 0
 	switch len(codes) {
 	case 0:
-		doexit(0)
 	case 1:
-		doexit(codes[0])
+		code = codes[0]
 	default:
-		throw(ErrArgs)
+		return ErrArgs
 	}
+
+	preExit(fm)
+	os.Exit(code)
+	// Does not return
+	panic("os.Exit returned")
 }
 
-func preExit(ec *Frame) {
-	err := ec.DaemonClient.Close()
+func preExit(fm *Frame) {
+	err := fm.DaemonClient.Close()
 	if err != nil {
 		fmt.Fprintln(os.Stderr, err)
 	}
@@ -94,6 +63,6 @@ func preExit(ec *Frame) {
 
 var errNotSupportedOnWindows = errors.New("not supported on Windows")
 
-func notSupportedOnWindows(ec *Frame, args []types.Value, opts map[string]types.Value) {
-	throw(errNotSupportedOnWindows)
+func notSupportedOnWindows() error {
+	return errNotSupportedOnWindows
 }
diff -pruN 0.11+ds1-5/eval/builtin_fn_cmd_test.go 0.12+ds1-1/eval/builtin_fn_cmd_test.go
--- 0.11+ds1-5/eval/builtin_fn_cmd_test.go	2018-01-18 02:03:11.000000000 +0000
+++ 0.12+ds1-1/eval/builtin_fn_cmd_test.go	2018-07-11 02:32:25.000000000 +0000
@@ -3,5 +3,5 @@ package eval
 import "testing"
 
 func TestBuiltinFnCmd(t *testing.T) {
-	runTests(t, []Test{})
+	Test(t /* TODO: Add test cases */)
 }
diff -pruN 0.11+ds1-5/eval/builtin_fn_cmd_unix.go 0.12+ds1-1/eval/builtin_fn_cmd_unix.go
--- 0.11+ds1-5/eval/builtin_fn_cmd_unix.go	2018-01-18 02:03:11.000000000 +0000
+++ 0.12+ds1-1/eval/builtin_fn_cmd_unix.go	2018-07-11 02:32:25.000000000 +0000
@@ -8,54 +8,53 @@ import (
 	"strconv"
 	"syscall"
 
-	"github.com/elves/elvish/eval/types"
+	"github.com/elves/elvish/eval/vals"
 	"github.com/elves/elvish/sys"
 )
 
-func execFn(ec *Frame, args []types.Value, opts map[string]types.Value) {
-	TakeNoOpt(opts)
-
+func execFn(fm *Frame, args ...interface{}) error {
 	var argstrings []string
 	if len(args) == 0 {
 		argstrings = []string{"elvish"}
 	} else {
 		argstrings = make([]string, len(args))
 		for i, a := range args {
-			argstrings[i] = types.ToString(a)
+			argstrings[i] = vals.ToString(a)
 		}
 	}
 
 	var err error
 	argstrings[0], err = exec.LookPath(argstrings[0])
-	maybeThrow(err)
+	if err != nil {
+		return err
+	}
 
-	preExit(ec)
+	preExit(fm)
 
-	err = syscall.Exec(argstrings[0], argstrings, os.Environ())
-	maybeThrow(err)
+	return syscall.Exec(argstrings[0], argstrings, os.Environ())
 }
 
-func fg(ec *Frame, args []types.Value, opts map[string]types.Value) {
-	var pids []int
-	ScanArgsVariadic(args, &pids)
-	TakeNoOpt(opts)
-
+func fg(pids ...int) error {
 	if len(pids) == 0 {
-		throw(ErrArgs)
+		return ErrArgs
 	}
 	var thepgid int
 	for i, pid := range pids {
 		pgid, err := syscall.Getpgid(pid)
-		maybeThrow(err)
+		if err != nil {
+			return err
+		}
 		if i == 0 {
 			thepgid = pgid
 		} else if pgid != thepgid {
-			throw(ErrNotInSameGroup)
+			return ErrNotInSameGroup
 		}
 	}
 
 	err := sys.Tcsetpgrp(0, thepgid)
-	maybeThrow(err)
+	if err != nil {
+		return err
+	}
 
 	errors := make([]*Exception, len(pids))
 
@@ -81,5 +80,5 @@ func fg(ec *Frame, args []types.Value, o
 		}
 	}
 
-	maybeThrow(ComposeExceptionsFromPipeline(errors))
+	return ComposeExceptionsFromPipeline(errors)
 }
diff -pruN 0.11+ds1-5/eval/builtin_fn_container.go 0.12+ds1-1/eval/builtin_fn_container.go
--- 0.11+ds1-5/eval/builtin_fn_container.go	2018-01-18 02:03:11.000000000 +0000
+++ 0.12+ds1-1/eval/builtin_fn_container.go	2018-07-11 02:32:25.000000000 +0000
@@ -4,144 +4,125 @@ import (
 	"errors"
 	"fmt"
 	"io"
-	"strconv"
 
-	"github.com/elves/elvish/eval/types"
-	"github.com/elves/elvish/util"
+	"github.com/elves/elvish/eval/vals"
+	"github.com/elves/elvish/eval/vars"
+	"github.com/xiaq/persistent/hashmap"
 )
 
 // Sequence, list and maps.
 
 func init() {
-	addToBuiltinFns([]*BuiltinFn{
-		{"ns", nsFn},
+	addBuiltinFns(map[string]interface{}{
+		"ns": nsFn,
 
-		{"range", rangeFn},
-		{"repeat", repeat},
-		{"explode", explode},
+		"range":   rangeFn,
+		"repeat":  repeat,
+		"explode": explode,
 
-		{"assoc", assoc},
-		{"dissoc", dissoc},
+		"assoc":  assoc,
+		"dissoc": dissoc,
 
-		{"all", all},
-		{"take", take},
-		{"drop", drop},
+		"all": all,
 
-		{"has-key", hasKey},
-		{"has-value", hasValue},
+		"has-key":   hasKey,
+		"has-value": hasValue,
 
-		{"count", count},
+		"take":  take,
+		"drop":  drop,
+		"count": count,
 
-		{"keys", keys},
+		"keys": keys,
 	})
 }
 
-func nsFn(ec *Frame, args []types.Value, opts map[string]types.Value) {
-	TakeNoArg(args)
-	TakeNoOpt(opts)
+var errKeyMustBeString = errors.New("key must be string")
 
-	ec.OutputChan() <- make(Ns)
+func nsFn(m hashmap.Map) (Ns, error) {
+	ns := make(Ns)
+	for it := m.Iterator(); it.HasElem(); it.Next() {
+		k, v := it.Elem()
+		kstring, ok := k.(string)
+		if !ok {
+			return nil, errKeyMustBeString
+		}
+		ns[kstring] = vars.NewAnyWithInit(v)
+	}
+	return ns, nil
 }
 
-func rangeFn(ec *Frame, args []types.Value, opts map[string]types.Value) {
-	var step float64
-	ScanOpts(opts, OptToScan{"step", &step, types.String("1")})
+func rangeFn(fm *Frame, rawOpts RawOptions, args ...float64) error {
+	opts := struct{ Step float64 }{1}
+	rawOpts.Scan(&opts)
 
 	var lower, upper float64
-	var err error
 
 	switch len(args) {
 	case 1:
-		upper, err = toFloat(args[0])
-		maybeThrow(err)
+		upper = args[0]
 	case 2:
-		lower, err = toFloat(args[0])
-		maybeThrow(err)
-		upper, err = toFloat(args[1])
-		maybeThrow(err)
+		lower, upper = args[0], args[1]
 	default:
-		throw(ErrArgs)
+		return ErrArgs
 	}
 
-	out := ec.ports[1].Chan
-	for f := lower; f < upper; f += step {
-		out <- floatToString(f)
+	out := fm.ports[1].Chan
+	for f := lower; f < upper; f += opts.Step {
+		out <- vals.FromGo(f)
 	}
+	return nil
 }
 
-func repeat(ec *Frame, args []types.Value, opts map[string]types.Value) {
-	var (
-		n int
-		v types.Value
-	)
-	ScanArgs(args, &n, &v)
-	TakeNoOpt(opts)
-
-	out := ec.OutputChan()
+func repeat(fm *Frame, n int, v interface{}) {
+	out := fm.OutputChan()
 	for i := 0; i < n; i++ {
 		out <- v
 	}
 }
 
 // explode puts each element of the argument.
-func explode(ec *Frame, args []types.Value, opts map[string]types.Value) {
-	var v types.IteratorValue
-	ScanArgs(args, &v)
-	TakeNoOpt(opts)
-
-	out := ec.ports[1].Chan
-	v.Iterate(func(e types.Value) bool {
+func explode(fm *Frame, v interface{}) error {
+	out := fm.ports[1].Chan
+	return vals.Iterate(v, func(e interface{}) bool {
 		out <- e
 		return true
 	})
 }
 
-func assoc(ec *Frame, args []types.Value, opts map[string]types.Value) {
-	var (
-		a    types.Assocer
-		k, v types.Value
-	)
-	ScanArgs(args, &a, &k, &v)
-	TakeNoOpt(opts)
-	ec.OutputChan() <- a.Assoc(k, v)
-}
-
-func dissoc(ec *Frame, args []types.Value, opts map[string]types.Value) {
-	var (
-		a types.Dissocer
-		k types.Value
-	)
-	ScanArgs(args, &a, &k)
-	TakeNoOpt(opts)
-	ec.OutputChan() <- a.Dissoc(k)
-}
-
-func all(ec *Frame, args []types.Value, opts map[string]types.Value) {
-	TakeNoArg(args)
-	TakeNoOpt(opts)
+func assoc(a, k, v interface{}) (interface{}, error) {
+	return vals.Assoc(a, k, v)
+}
+
+var errCannotDissoc = errors.New("cannot dissoc")
 
+func dissoc(a, k interface{}) (interface{}, error) {
+	a2 := vals.Dissoc(a, k)
+	if a2 == nil {
+		return nil, errCannotDissoc
+	}
+	return a2, nil
+}
+
+func all(fm *Frame) error {
 	valuesDone := make(chan struct{})
 	go func() {
-		for input := range ec.ports[0].Chan {
-			ec.ports[1].Chan <- input
+		for input := range fm.ports[0].Chan {
+			fm.ports[1].Chan <- input
 		}
 		close(valuesDone)
 	}()
-	_, err := io.Copy(ec.ports[1].File, ec.ports[0].File)
+	_, err := io.Copy(fm.ports[1].File, fm.ports[0].File)
 	<-valuesDone
 	if err != nil {
-		throwf("cannot copy byte input: %s", err)
+		return fmt.Errorf("cannot copy byte input: %s", err)
 	}
+	return nil
 }
 
-func take(ec *Frame, args []types.Value, opts map[string]types.Value) {
-	var n int
-	iterate := ScanArgsOptionalInput(ec, args, &n)
-	TakeNoOpt(opts)
-
-	out := ec.ports[1].Chan
+func take(fm *Frame, n int, inputs Inputs) {
+	out := fm.ports[1].Chan
 	i := 0
-	iterate(func(v types.Value) {
+	inputs(func(v interface{}) {
 		if i < n {
 			out <- v
 		}
@@ -149,14 +130,10 @@ func take(ec *Frame, args []types.Value,
 	})
 }
 
-func drop(ec *Frame, args []types.Value, opts map[string]types.Value) {
-	var n int
-	iterate := ScanArgsOptionalInput(ec, args, &n)
-	TakeNoOpt(opts)
-
-	out := ec.ports[1].Chan
+func drop(fm *Frame, n int, inputs Inputs) {
+	out := fm.ports[1].Chan
 	i := 0
-	iterate(func(v types.Value) {
+	inputs(func(v interface{}) {
 		if i >= n {
 			out <- v
 		}
@@ -164,95 +141,83 @@ func drop(ec *Frame, args []types.Value,
 	})
 }
 
-func hasValue(ec *Frame, args []types.Value, opts map[string]types.Value) {
-	TakeNoOpt(opts)
-
-	var container, value types.Value
-	var found bool
-
-	ScanArgs(args, &container, &value)
-
+func hasValue(container, value interface{}) (bool, error) {
 	switch container := container.(type) {
-	case types.Iterator:
-		container.Iterate(func(v types.Value) bool {
+	case hashmap.Map:
+		for it := container.Iterator(); it.HasElem(); it.Next() {
+			_, v := it.Elem()
+			if vals.Equal(v, value) {
+				return true, nil
+			}
+		}
+		return false, nil
+	default:
+		var found bool
+		err := vals.Iterate(container, func(v interface{}) bool {
 			found = (v == value)
 			return !found
 		})
-	case types.MapLike:
-		container.IterateKey(func(v types.Value) bool {
-			found = (container.IndexOne(v) == value)
-			return !found
-		})
-	default:
-		throw(fmt.Errorf("argument of type '%s' is not iterable", container.Kind()))
+		return found, err
 	}
-
-	ec.ports[1].Chan <- types.Bool(found)
 }
 
-func hasKey(ec *Frame, args []types.Value, opts map[string]types.Value) {
-	TakeNoOpt(opts)
-
-	var container, key types.Value
-	var found bool
-
-	ScanArgs(args, &container, &key)
-
+func hasKey(container, key interface{}) (bool, error) {
 	switch container := container.(type) {
-	case types.HasKeyer:
-		found = container.HasKey(key)
-	case types.Lener:
-		// XXX(xiaq): Not all types that implement Lener have numerical indices
-		err := util.PCall(func() {
-			types.ParseAndFixListIndex(types.ToString(key), container.Len())
-		})
-		found = (err == nil)
+	case hashmap.Map:
+		return hashmap.HasKey(container, key), nil
 	default:
-		throw(fmt.Errorf("couldn't get key or index of type '%s'", container.Kind()))
+		if len := vals.Len(container); len >= 0 {
+			// XXX(xiaq): Not all types that implement Lener have numerical indices
+			_, err := vals.ConvertListIndex(key, len)
+			return err == nil, nil
+		} else {
+			var found bool
+			err := vals.IterateKeys(container, func(k interface{}) bool {
+				if key == k {
+					found = true
+				}
+				return !found
+			})
+			if err == nil {
+				return found, nil
+			}
+		}
+		return false, fmt.Errorf("couldn't get key or index of type '%s'", vals.Kind(container))
 	}
-
-	ec.ports[1].Chan <- types.Bool(found)
 }
 
-func count(ec *Frame, args []types.Value, opts map[string]types.Value) {
-	TakeNoOpt(opts)
-
+func count(fm *Frame, args ...interface{}) (int, error) {
 	var n int
 	switch len(args) {
 	case 0:
 		// Count inputs.
-		ec.IterateInputs(func(types.Value) {
+		fm.IterateInputs(func(interface{}) {
 			n++
 		})
 	case 1:
 		// Get length of argument.
 		v := args[0]
-		if lener, ok := v.(types.Lener); ok {
-			n = lener.Len()
-		} else if iterator, ok := v.(types.Iterator); ok {
-			iterator.Iterate(func(types.Value) bool {
+		if len := vals.Len(v); len >= 0 {
+			n = len
+		} else {
+			err := vals.Iterate(v, func(interface{}) bool {
 				n++
 				return true
 			})
-		} else {
-			throw(fmt.Errorf("cannot get length of a %s", v.Kind()))
+			if err != nil {
+				return 0, fmt.Errorf("cannot get length of a %s", vals.Kind(v))
+			}
 		}
 	default:
-		throw(errors.New("want 0 or 1 argument"))
+		return 0, errors.New("want 0 or 1 argument")
 	}
-	ec.ports[1].Chan <- types.String(strconv.Itoa(n))
+	return n, nil
 }
 
-func keys(ec *Frame, args []types.Value, opts map[string]types.Value) {
-	TakeNoOpt(opts)
-
-	var iter types.IterateKeyer
-	ScanArgs(args, &iter)
-
-	out := ec.ports[1].Chan
-
-	iter.IterateKey(func(v types.Value) bool {
-		out <- v
+func keys(fm *Frame, v interface{}) error {
+	out := fm.ports[1].Chan
+	return vals.IterateKeys(v, func(k interface{}) bool {
+		out <- k
 		return true
 	})
 }
diff -pruN 0.11+ds1-5/eval/builtin_fn_container_test.go 0.12+ds1-1/eval/builtin_fn_container_test.go
--- 0.11+ds1-5/eval/builtin_fn_container_test.go	2018-01-18 02:03:11.000000000 +0000
+++ 0.12+ds1-1/eval/builtin_fn_container_test.go	2018-07-11 02:32:25.000000000 +0000
@@ -3,45 +3,45 @@ package eval
 import "testing"
 
 func TestBuiltinFnContainer(t *testing.T) {
-	runTests(t, []Test{
-		{`range 3`, want{out: strs("0", "1", "2")}},
-		{`range 1 3`, want{out: strs("1", "2")}},
-		{`range 0 10 &step=3`, want{out: strs("0", "3", "6", "9")}},
-		{`repeat 4 foo`, want{out: strs("foo", "foo", "foo", "foo")}},
-		{`explode [foo bar]`, want{out: strs("foo", "bar")}},
-
-		{`put (assoc [0] 0 zero)[0]`, want{out: strs("zero")}},
-		{`put (assoc [&] k v)[k]`, want{out: strs("v")}},
-		{`put (assoc [&k=v] k v2)[k]`, want{out: strs("v2")}},
-		{`has-key (dissoc [&k=v] k) k`, wantFalse},
-
-		{`put foo bar | all`, want{out: strs("foo", "bar")}},
-		{`echo foobar | all`, want{bytesOut: []byte("foobar\n")}},
-		{`{ put foo bar; echo foobar } | all`,
-			want{out: strs("foo", "bar"), bytesOut: []byte("foobar\n")}},
-		{`range 100 | take 2`, want{out: strs("0", "1")}},
-		{`range 100 | drop 98`, want{out: strs("98", "99")}},
-
-		{`has-key [foo bar] 0`, wantTrue},
-		{`has-key [foo bar] 0:1`, wantTrue},
-		{`has-key [foo bar] 0:20`, wantFalse},
-		{`has-key [&lorem=ipsum &foo=bar] lorem`, wantTrue},
-		{`has-key [&lorem=ipsum &foo=bar] loremwsq`, wantFalse},
-		{`has-value [&lorem=ipsum &foo=bar] lorem`, wantFalse},
-		{`has-value [&lorem=ipsum &foo=bar] bar`, wantTrue},
-		{`has-value [foo bar] bar`, wantTrue},
-		{`has-value [foo bar] badehose`, wantFalse},
-		{`has-value "foo" o`, wantTrue},
-		{`has-value "foo" d`, wantFalse},
-
-		{`range 100 | count`, want{out: strs("100")}},
-		{`count [(range 100)]`, want{out: strs("100")}},
-		{`count 1 2 3`, want{err: errAny}},
+	Test(t,
+		That(`range 3`).Puts("0", "1", "2"),
+		That(`range 1 3`).Puts("1", "2"),
+		That(`range 0 10 &step=3`).Puts("0", "3", "6", "9"),
+		That(`repeat 4 foo`).Puts("foo", "foo", "foo", "foo"),
+		That(`explode [foo bar]`).Puts("foo", "bar"),
+
+		That(`put (assoc [0] 0 zero)[0]`).Puts("zero"),
+		That(`put (assoc [&] k v)[k]`).Puts("v"),
+		That(`put (assoc [&k=v] k v2)[k]`).Puts("v2"),
+		That(`has-key (dissoc [&k=v] k) k`).Puts(false),
+
+		That(`put foo bar | all`).Puts("foo", "bar"),
+		That(`echo foobar | all`).Prints("foobar\n"),
+		That(`{ put foo bar; echo foobar } | all`).Puts(
+			"foo", "bar").Prints("foobar\n"),
+		That(`range 100 | take 2`).Puts("0", "1"),
+		That(`range 100 | drop 98`).Puts("98", "99"),
+
+		That(`has-key [foo bar] 0`).Puts(true),
+		That(`has-key [foo bar] 0:1`).Puts(true),
+		That(`has-key [foo bar] 0:20`).Puts(false),
+		That(`has-key [&lorem=ipsum &foo=bar] lorem`).Puts(true),
+		That(`has-key [&lorem=ipsum &foo=bar] loremwsq`).Puts(false),
+		That(`has-value [&lorem=ipsum &foo=bar] lorem`).Puts(false),
+		That(`has-value [&lorem=ipsum &foo=bar] bar`).Puts(true),
+		That(`has-value [foo bar] bar`).Puts(true),
+		That(`has-value [foo bar] badehose`).Puts(false),
+		That(`has-value "foo" o`).Puts(true),
+		That(`has-value "foo" d`).Puts(false),
+
+		That(`range 100 | count`).Puts("100"),
+		That(`count [(range 100)]`).Puts("100"),
+		That(`count 1 2 3`).Errors(),
 
-		{`keys [&]`, wantNothing},
-		{`keys [&a=foo]`, want{out: strs("a")}},
+		That(`keys [&]`).DoesNothing(),
+		That(`keys [&a=foo]`).Puts("a"),
 		// Windows does not have an external sort command. Disabled until we have a
 		// builtin sort command.
-		// {`keys [&a=foo &b=bar] | each echo | sort | each $put~`, want{out: strs("a", "b")}},
-	})
+		// That(`keys [&a=foo &b=bar] | each echo | sort | each $put~`).Puts("a", "b"),
+	)
 }
diff -pruN 0.11+ds1-5/eval/builtin_fn_debug.go 0.12+ds1-1/eval/builtin_fn_debug.go
--- 0.11+ds1-5/eval/builtin_fn_debug.go	1970-01-01 00:00:00.000000000 +0000
+++ 0.12+ds1-1/eval/builtin_fn_debug.go	2018-07-11 02:32:25.000000000 +0000
@@ -0,0 +1,38 @@
+package eval
+
+import (
+	"runtime"
+
+	"github.com/elves/elvish/util"
+)
+
+func init() {
+	addBuiltinFns(map[string]interface{}{
+		"src":    src,
+		"-gc":    _gc,
+		"-stack": _stack,
+		"-log":   _log,
+	})
+}
+
+func src(fm *Frame) *Source {
+	return fm.srcMeta
+}
+
+func _gc() {
+	runtime.GC()
+}
+
+func _stack(fm *Frame) {
+	out := fm.ports[1].File
+	// XXX dup with main.go
+	buf := make([]byte, 1024)
+	for runtime.Stack(buf, true) == cap(buf) {
+		buf = make([]byte, cap(buf)*2)
+	}
+	out.Write(buf)
+}
+
+func _log(fname string) error {
+	return util.SetOutputFile(fname)
+}
diff -pruN 0.11+ds1-5/eval/builtin_fn_env.go 0.12+ds1-1/eval/builtin_fn_env.go
--- 0.11+ds1-5/eval/builtin_fn_env.go	1970-01-01 00:00:00.000000000 +0000
+++ 0.12+ds1-1/eval/builtin_fn_env.go	2018-07-11 02:32:25.000000000 +0000
@@ -0,0 +1,30 @@
+package eval
+
+import (
+	"errors"
+	"os"
+)
+
+var errNonExistentEnvVar = errors.New("non-existent environment variable")
+
+func init() {
+	addBuiltinFns(map[string]interface{}{
+		"has-env":   hasEnv,
+		"get-env":   getEnv,
+		"set-env":   os.Setenv,
+		"unset-env": os.Unsetenv,
+	})
+}
+
+func hasEnv(key string) bool {
+	_, ok := os.LookupEnv(key)
+	return ok
+}
+
+func getEnv(key string) (string, error) {
+	value, ok := os.LookupEnv(key)
+	if !ok {
+		return "", errNonExistentEnvVar
+	}
+	return value, nil
+}
diff -pruN 0.11+ds1-5/eval/builtin_fn_env_test.go 0.12+ds1-1/eval/builtin_fn_env_test.go
--- 0.11+ds1-5/eval/builtin_fn_env_test.go	1970-01-01 00:00:00.000000000 +0000
+++ 0.12+ds1-1/eval/builtin_fn_env_test.go	2018-07-11 02:32:25.000000000 +0000
@@ -0,0 +1,28 @@
+package eval
+
+import (
+	"os"
+	"testing"
+
+	"github.com/elves/elvish/eval/vals"
+)
+
+func TestBuiltinFnEnv(t *testing.T) {
+	oldpath := os.Getenv("PATH")
+	listSep := string(os.PathListSeparator)
+	Test(t,
+		That(`get-env var`).ErrorsWith(errNonExistentEnvVar),
+		That(`set-env var test1`),
+		That(`get-env var`).Puts("test1"),
+		That(`put $E:var`).Puts("test1"),
+		That(`set-env var test2`),
+		That(`get-env var`).Puts("test2"),
+		That(`put $E:var`).Puts("test2"),
+
+		That(`set-env PATH /test-path`),
+		That(`put $paths`).Puts(vals.MakeList("/test-path")),
+		That(`paths = [/test-path2 $@paths]`),
+		That(`get-env PATH`).Puts("/test-path2"+listSep+"/test-path"),
+	)
+	os.Setenv("PATH", oldpath)
+}
diff -pruN 0.11+ds1-5/eval/builtin_fn_flow.go 0.12+ds1-1/eval/builtin_fn_flow.go
--- 0.11+ds1-5/eval/builtin_fn_flow.go	2018-01-18 02:03:11.000000000 +0000
+++ 0.12+ds1-1/eval/builtin_fn_flow.go	2018-07-11 02:32:25.000000000 +0000
@@ -4,67 +4,56 @@ import (
 	"errors"
 	"sync"
 
-	"github.com/elves/elvish/eval/types"
+	"github.com/elves/elvish/util"
 )
 
 // Flow control.
 
 func init() {
-	addToBuiltinFns([]*BuiltinFn{
-		{"run-parallel", runParallel},
-
-		// Iterations.
-		{"each", each},
-		{"peach", peach},
-
+	addBuiltinFns(map[string]interface{}{
+		"run-parallel": runParallel,
 		// Exception and control
-		{"fail", fail},
-		{"multi-error", multiErrorFn},
-		{"return", returnFn},
-		{"break", breakFn},
-		{"continue", continueFn},
+		"fail":        fail,
+		"multi-error": multiErrorFn,
+		"return":      returnFn,
+		"break":       breakFn,
+		"continue":    continueFn,
+		// Iterations.
+		"each":  each,
+		"peach": peach,
 	})
 }
 
-func runParallel(ec *Frame, args []types.Value, opts map[string]types.Value) {
-	var functions []Fn
-	ScanArgsVariadic(args, &functions)
-	TakeNoOpt(opts)
-
+func runParallel(fm *Frame, functions ...Callable) error {
 	var waitg sync.WaitGroup
 	waitg.Add(len(functions))
 	exceptions := make([]*Exception, len(functions))
 	for i, function := range functions {
-		go func(ec *Frame, function Fn, exception **Exception) {
-			err := ec.PCall(function, NoArgs, NoOpts)
+		go func(fm2 *Frame, function Callable, exception **Exception) {
+			err := fm2.Call(function, NoArgs, NoOpts)
 			if err != nil {
 				*exception = err.(*Exception)
 			}
 			waitg.Done()
-		}(ec.fork("[run-parallel function]"), function, &exceptions[i])
+		}(fm.fork("[run-parallel function]"), function, &exceptions[i])
 	}
 
 	waitg.Wait()
-	maybeThrow(ComposeExceptionsFromPipeline(exceptions))
+	return ComposeExceptionsFromPipeline(exceptions)
 }
 
 // each takes a single closure and applies it to all input values.
-func each(ec *Frame, args []types.Value, opts map[string]types.Value) {
-	var f Fn
-	iterate := ScanArgsOptionalInput(ec, args, &f)
-	TakeNoOpt(opts)
-
+func each(fm *Frame, f Callable, inputs Inputs) error {
 	broken := false
-	iterate(func(v types.Value) {
+	var err error
+	inputs(func(v interface{}) {
 		if broken {
 			return
 		}
-		// NOTE We don't have the position range of the closure in the source.
-		// Ideally, it should be kept in the Closure itself.
-		newec := ec.fork("closure of each")
-		newec.ports[0] = DevNullClosedChan
-		ex := newec.PCall(f, []types.Value{v}, NoOpts)
-		ClosePorts(newec.ports)
+		newFm := fm.fork("closure of each")
+		newFm.ports[0] = DevNullClosedChan
+		ex := newFm.Call(f, []interface{}{v}, NoOpts)
+		newFm.Close()
 
 		if ex != nil {
 			switch ex.(*Exception).Cause {
@@ -73,33 +62,29 @@ func each(ec *Frame, args []types.Value,
 			case Break:
 				broken = true
 			default:
-				throw(ex)
+				broken = true
+				err = ex
 			}
 		}
 	})
+	return err
 }
 
 // peach takes a single closure and applies it to all input values in parallel.
-func peach(ec *Frame, args []types.Value, opts map[string]types.Value) {
-	var f Fn
-	iterate := ScanArgsOptionalInput(ec, args, &f)
-	TakeNoOpt(opts)
-
+func peach(fm *Frame, f Callable, inputs Inputs) error {
 	var w sync.WaitGroup
 	broken := false
 	var err error
-	iterate(func(v types.Value) {
+	inputs(func(v interface{}) {
 		if broken || err != nil {
 			return
 		}
 		w.Add(1)
 		go func() {
-			// NOTE We don't have the position range of the closure in the source.
-			// Ideally, it should be kept in the Closure itself.
-			newec := ec.fork("closure of each")
-			newec.ports[0] = DevNullClosedChan
-			ex := newec.PCall(f, []types.Value{v}, NoOpts)
-			ClosePorts(newec.ports)
+			newFm := fm.fork("closure of peach")
+			newFm.ports[0] = DevNullClosedChan
+			ex := newFm.Call(f, []interface{}{v}, NoOpts)
+			newFm.Close()
 
 			if ex != nil {
 				switch ex.(*Exception).Cause {
@@ -108,49 +93,33 @@ func peach(ec *Frame, args []types.Value
 				case Break:
 					broken = true
 				default:
-					err = ex
+					broken = true
+					err = util.Errors(err, ex)
 				}
 			}
 			w.Done()
 		}()
 	})
 	w.Wait()
-	maybeThrow(err)
+	return err
 }
 
-func fail(ec *Frame, args []types.Value, opts map[string]types.Value) {
-	var msg types.String
-	ScanArgs(args, &msg)
-	TakeNoOpt(opts)
-
-	throw(errors.New(string(msg)))
+func fail(msg string) error {
+	return errors.New(msg)
 }
 
-func multiErrorFn(ec *Frame, args []types.Value, opts map[string]types.Value) {
-	var excs []*Exception
-	ScanArgsVariadic(args, &excs)
-	TakeNoOpt(opts)
-
-	throw(PipelineError{excs})
+func multiErrorFn(excs ...*Exception) error {
+	return PipelineError{excs}
 }
 
-func returnFn(ec *Frame, args []types.Value, opts map[string]types.Value) {
-	TakeNoArg(args)
-	TakeNoOpt(opts)
-
-	throw(Return)
+func returnFn() error {
+	return Return
 }
 
-func breakFn(ec *Frame, args []types.Value, opts map[string]types.Value) {
-	TakeNoArg(args)
-	TakeNoOpt(opts)
-
-	throw(Break)
+func breakFn() error {
+	return Break
 }
 
-func continueFn(ec *Frame, args []types.Value, opts map[string]types.Value) {
-	TakeNoArg(args)
-	TakeNoOpt(opts)
-
-	throw(Continue)
+func continueFn() error {
+	return Continue
 }
diff -pruN 0.11+ds1-5/eval/builtin_fn_flow_test.go 0.12+ds1-1/eval/builtin_fn_flow_test.go
--- 0.11+ds1-5/eval/builtin_fn_flow_test.go	2018-01-18 02:03:11.000000000 +0000
+++ 0.12+ds1-1/eval/builtin_fn_flow_test.go	2018-07-11 02:32:25.000000000 +0000
@@ -3,20 +3,20 @@ package eval
 import "testing"
 
 func TestBuiltinFnFlow(t *testing.T) {
-	runTests(t, []Test{
-		{`run-parallel { put lorem } { echo ipsum }`,
-			want{out: strs("lorem"), bytesOut: []byte("ipsum\n")}},
+	Test(t,
+		That(`run-parallel { put lorem } { echo ipsum }`).
+			Puts("lorem").Prints("ipsum\n"),
 
-		{`put 1 233 | each $put~`, want{out: strs("1", "233")}},
-		{`echo "1\n233" | each $put~`, want{out: strs("1", "233")}},
-		{`each $put~ [1 233]`, want{out: strs("1", "233")}},
-		{`range 10 | each [x]{ if (== $x 4) { break }; put $x }`,
-			want{out: strs("0", "1", "2", "3")}},
-		{`range 10 | each [x]{ if (== $x 4) { fail haha }; put $x }`,
-			want{out: strs("0", "1", "2", "3"), err: errAny}},
+		That(`put 1 233 | each $put~`).Puts("1", "233"),
+		That(`echo "1\n233" | each $put~`).Puts("1", "233"),
+		That(`each $put~ [1 233]`).Puts("1", "233"),
+		That(`range 10 | each [x]{ if (== $x 4) { break }; put $x }`).
+			Puts("0", "1", "2", "3"),
+		That(`range 10 | each [x]{ if (== $x 4) { fail haha }; put $x }`).
+			Puts("0", "1", "2", "3").Errors(),
 		// TODO: test peach
 
-		{`fail haha`, want{err: errAny}},
-		{`return`, want{err: Return}},
-	})
+		That(`fail haha`).Errors(),
+		That(`return`).ErrorsWith(Return),
+	)
 }
diff -pruN 0.11+ds1-5/eval/builtin_fn_fs.go 0.12+ds1-1/eval/builtin_fn_fs.go
--- 0.11+ds1-5/eval/builtin_fn_fs.go	2018-01-18 02:03:11.000000000 +0000
+++ 0.12+ds1-1/eval/builtin_fn_fs.go	2018-07-11 02:32:25.000000000 +0000
@@ -5,7 +5,7 @@ import (
 	"os"
 	"path/filepath"
 
-	"github.com/elves/elvish/eval/types"
+	"github.com/elves/elvish/eval/vals"
 	"github.com/elves/elvish/store/storedefs"
 	"github.com/elves/elvish/util"
 )
@@ -15,120 +15,66 @@ import (
 var ErrStoreNotConnected = errors.New("store not connected")
 
 func init() {
-	addToBuiltinFns([]*BuiltinFn{
+	addBuiltinFns(map[string]interface{}{
 		// Directory
-		{"cd", cd},
-		{"dir-history", dirs},
+		"cd":          cd,
+		"dir-history": dirs,
 
 		// Path
-		{"path-abs", WrapStringToStringError(filepath.Abs)},
-		{"path-base", WrapStringToString(filepath.Base)},
-		{"path-clean", WrapStringToString(filepath.Clean)},
-		{"path-dir", WrapStringToString(filepath.Dir)},
-		{"path-ext", WrapStringToString(filepath.Ext)},
-
-		{"eval-symlinks", WrapStringToStringError(filepath.EvalSymlinks)},
-		{"tilde-abbr", tildeAbbr},
+		"path-abs":      filepath.Abs,
+		"path-base":     filepath.Base,
+		"path-clean":    filepath.Clean,
+		"path-dir":      filepath.Dir,
+		"path-ext":      filepath.Ext,
+		"eval-symlinks": filepath.EvalSymlinks,
+		"tilde-abbr":    tildeAbbr,
 
 		// File types
-		{"-is-dir", isDir},
+		"-is-dir": isDir,
 	})
 }
 
-func WrapStringToString(f func(string) string) BuiltinFnImpl {
-	return func(ec *Frame, args []types.Value, opts map[string]types.Value) {
-		TakeNoOpt(opts)
-		s := mustGetOneString(args)
-		ec.ports[1].Chan <- types.String(f(s))
-	}
-}
-
-func WrapStringToStringError(f func(string) (string, error)) BuiltinFnImpl {
-	return func(ec *Frame, args []types.Value, opts map[string]types.Value) {
-		TakeNoOpt(opts)
-		s := mustGetOneString(args)
-		result, err := f(s)
-		maybeThrow(err)
-		ec.ports[1].Chan <- types.String(result)
-	}
-}
-
-var errMustBeOneString = errors.New("must be one string argument")
-
-func mustGetOneString(args []types.Value) string {
-	if len(args) != 1 {
-		throw(errMustBeOneString)
-	}
-	s, ok := args[0].(types.String)
-	if !ok {
-		throw(errMustBeOneString)
-	}
-	return string(s)
-}
-
-func cd(ec *Frame, args []types.Value, opts map[string]types.Value) {
-	TakeNoOpt(opts)
-
+func cd(fm *Frame, args ...string) error {
 	var dir string
-	if len(args) == 0 {
+	switch len(args) {
+	case 0:
 		dir = mustGetHome("")
-	} else if len(args) == 1 {
-		dir = types.ToString(args[0])
-	} else {
-		throw(ErrArgs)
+	case 1:
+		dir = args[0]
+	default:
+		return ErrArgs
 	}
 
-	cdInner(dir, ec)
-}
-
-func cdInner(dir string, ec *Frame) {
-	maybeThrow(Chdir(dir, ec.DaemonClient))
+	return fm.Chdir(dir)
 }
 
-var dirDescriptor = types.NewStructDescriptor("path", "score")
+var dirDescriptor = vals.NewStructDescriptor("path", "score")
 
-func newDirStruct(path string, score float64) *types.Struct {
-	return types.NewStruct(dirDescriptor,
-		[]types.Value{types.String(path), floatToString(score)})
+func newDirStruct(path string, score float64) *vals.Struct {
+	return vals.NewStruct(dirDescriptor,
+		[]interface{}{path, vals.FromGo(score)})
 }
 
-func dirs(ec *Frame, args []types.Value, opts map[string]types.Value) {
-	TakeNoArg(args)
-	TakeNoOpt(opts)
-
-	if ec.DaemonClient == nil {
-		throw(ErrStoreNotConnected)
+func dirs(fm *Frame) error {
+	if fm.DaemonClient == nil {
+		return ErrStoreNotConnected
 	}
-	dirs, err := ec.DaemonClient.Dirs(storedefs.NoBlacklist)
+	dirs, err := fm.DaemonClient.Dirs(storedefs.NoBlacklist)
 	if err != nil {
-		throw(errors.New("store error: " + err.Error()))
+		return err
 	}
-	out := ec.ports[1].Chan
+	out := fm.ports[1].Chan
 	for _, dir := range dirs {
 		out <- newDirStruct(dir.Path, dir.Score)
 	}
+	return nil
 }
 
-func tildeAbbr(ec *Frame, args []types.Value, opts map[string]types.Value) {
-	var pathv types.String
-	ScanArgs(args, &pathv)
-	path := string(pathv)
-	TakeNoOpt(opts)
-
-	out := ec.ports[1].Chan
-	out <- types.String(util.TildeAbbr(path))
-}
-
-func isDir(ec *Frame, args []types.Value, opts map[string]types.Value) {
-	var pathv types.String
-	ScanArgs(args, &pathv)
-	path := string(pathv)
-	TakeNoOpt(opts)
-
-	ec.OutputChan() <- types.Bool(isDirInner(path))
+func tildeAbbr(path string) string {
+	return util.TildeAbbr(path)
 }
 
-func isDirInner(path string) bool {
+func isDir(path string) bool {
 	fi, err := os.Stat(path)
 	return err == nil && fi.Mode().IsDir()
 }
diff -pruN 0.11+ds1-5/eval/builtin_fn_fs_test.go 0.12+ds1-1/eval/builtin_fn_fs_test.go
--- 0.11+ds1-5/eval/builtin_fn_fs_test.go	2018-01-18 02:03:11.000000000 +0000
+++ 0.12+ds1-1/eval/builtin_fn_fs_test.go	2018-07-11 02:32:25.000000000 +0000
@@ -3,16 +3,23 @@ package eval
 import (
 	"path/filepath"
 	"testing"
+
+	"github.com/elves/elvish/parse"
 )
 
 func TestBuiltinFnFS(t *testing.T) {
 	pathSep := string(filepath.Separator)
-	runTests(t, []Test{
-		{`path-base a/b/c.png`, want{out: strs("c.png")}},
-		{`tilde-abbr $E:HOME` + pathSep + `foobar`,
-			want{out: strs("~" + pathSep + "foobar")}},
+	InTempHome(func(tmpHome string) {
+		MustMkdirAll("dir", 0700)
+		MustCreateEmpty("file")
+
+		Test(t,
+			That(`path-base a/b/c.png`).Puts("c.png"),
+			That("tilde-abbr "+parse.Quote(filepath.Join(tmpHome, "foobar"))).
+				Puts("~"+pathSep+"foobar"),
 
-		{`-is-dir ~/dir`, wantTrue}, // see testmain_test.go for setup
-		{`-is-dir ~/lorem`, wantFalse},
+			That(`-is-dir ~/dir`).Puts(true),
+			That(`-is-dir ~/file`).Puts(false),
+		)
 	})
 }
diff -pruN 0.11+ds1-5/eval/builtin_fn.go 0.12+ds1-1/eval/builtin_fn.go
--- 0.11+ds1-5/eval/builtin_fn.go	2018-01-18 02:03:11.000000000 +0000
+++ 0.12+ds1-1/eval/builtin_fn.go	2018-07-11 02:32:25.000000000 +0000
@@ -1,264 +1,198 @@
 package eval
 
-// Builtin functions.
-
 import (
 	"errors"
 	"fmt"
-	"math/rand"
-	"net"
-	"path/filepath"
-	"runtime"
-	"time"
+	"reflect"
 	"unsafe"
 
-	"github.com/elves/elvish/eval/types"
-	"github.com/elves/elvish/util"
+	"github.com/elves/elvish/eval/vals"
 	"github.com/xiaq/persistent/hash"
 )
 
-// BuiltinFn is a builtin function.
+var ErrArgs = errors.New("args error")
+
+// BuiltinFn uses reflection to wrap arbitrary Go functions into Elvish
+// functions.
+//
+// Parameters are passed following these rules:
+//
+// 1. If the first parameter of function has type *Frame, it gets the current
+// call frame.
+//
+// 2. If (possibly after a *Frame parameter) the first parameter has type
+// RawOptions, it gets a map of options. If the function has not declared an
+// RawOptions parameter but is passed options, an error is thrown.
+//
+// 3. If the last parameter is non-variadic and has type Inputs, it represents
+// an optional parameter that contains the input to this function. If the
+// argument is not supplied, the input channel of the Frame will be used to
+// supply the inputs.
+//
+// 4. Other parameters are converted using elvToGo.
+//
+// Return values go to the channel part of the stdout port, after being
+// converted using goToElv. If the last return value has type error and is not
+// nil, it is turned into an exception and no ouputting happens. If the last
+// return value is a nil error, it is ignored.
 type BuiltinFn struct {
-	Name string
-	Impl BuiltinFnImpl
+	name string
+	impl interface{}
+
+	// Type information of impl.
+
+	frame   bool
+	options bool
+	inputs  bool
+	// Type of "normal" (non-Frame, non-Options, non-variadic) arguments.
+	normalArgs []reflect.Type
+	// Type of variadic arguments, nil if function is non-variadic
+	variadicArg reflect.Type
 }
 
-type BuiltinFnImpl func(*Frame, []types.Value, map[string]types.Value)
+var _ Callable = &BuiltinFn{}
+
+type (
+	Inputs func(func(interface{}))
+)
+
+var (
+	frameType      = reflect.TypeOf((*Frame)(nil))
+	rawOptionsType = reflect.TypeOf(RawOptions(nil))
+	inputsType     = reflect.TypeOf(Inputs(nil))
+)
 
-var _ Fn = &BuiltinFn{}
+// NewBuiltinFn creates a new ReflectBuiltinFn instance.
+func NewBuiltinFn(name string, impl interface{}) *BuiltinFn {
+	implType := reflect.TypeOf(impl)
+	b := &BuiltinFn{name: name, impl: impl}
+
+	i := 0
+	if i < implType.NumIn() && implType.In(i) == frameType {
+		b.frame = true
+		i++
+	}
+	if i < implType.NumIn() && implType.In(i) == rawOptionsType {
+		b.options = true
+		i++
+	}
+	for ; i < implType.NumIn(); i++ {
+		paramType := implType.In(i)
+		if i == implType.NumIn()-1 {
+			if implType.IsVariadic() {
+				b.variadicArg = paramType.Elem()
+				break
+			} else if paramType == inputsType {
+				b.inputs = true
+				break
+			}
+		}
+		b.normalArgs = append(b.normalArgs, paramType)
+	}
+	return b
+}
 
 // Kind returns "fn".
 func (*BuiltinFn) Kind() string {
 	return "fn"
 }
 
-// Equal compares based on identity.
+// Equal compares identity.
 func (b *BuiltinFn) Equal(rhs interface{}) bool {
 	return b == rhs
 }
 
+// Hash hashes the address.
 func (b *BuiltinFn) Hash() uint32 {
 	return hash.Pointer(unsafe.Pointer(b))
 }
 
-// Repr returns an opaque representation "<builtin xxx>".
+// Repr returns an opaque representation "<builtin $name>".
 func (b *BuiltinFn) Repr(int) string {
-	return "<builtin " + b.Name + ">"
-}
-
-// Call calls a builtin function.
-func (b *BuiltinFn) Call(ec *Frame, args []types.Value, opts map[string]types.Value) {
-	b.Impl(ec, args, opts)
-}
-
-var builtinFns []*BuiltinFn
-
-func addToBuiltinFns(moreFns []*BuiltinFn) {
-	builtinFns = append(builtinFns, moreFns...)
-}
-
-// Builtins that have not been put into their own groups go here.
-
-var ErrArgs = errors.New("args error")
-
-func init() {
-	addToBuiltinFns([]*BuiltinFn{
-		{"nop", nop},
-
-		{"kind-of", kindOf},
-
-		{"bool", boolFn},
-		{"not", not},
-		{"is", is},
-		{"eq", eq},
-		{"not-eq", notEq},
-
-		{"constantly", constantly},
-
-		{"-source", source},
-
-		// Time
-		{"esleep", sleep},
-		{"-time", _time},
-
-		// Debugging
-		{"src", src},
-		{"-gc", _gc},
-		{"-stack", _stack},
-		{"-log", _log},
-
-		{"-ifaddrs", _ifaddrs},
-	})
-	// For rand and randint.
-	rand.Seed(time.Now().UTC().UnixNano())
-}
-
-func nop(ec *Frame, args []types.Value, opts map[string]types.Value) {
-}
-
-func kindOf(ec *Frame, args []types.Value, opts map[string]types.Value) {
-	TakeNoOpt(opts)
-	out := ec.ports[1].Chan
-	for _, a := range args {
-		out <- types.String(a.Kind())
-	}
+	return "<builtin " + b.name + ">"
 }
 
-func boolFn(ec *Frame, args []types.Value, opts map[string]types.Value) {
-	var v types.Value
-	ScanArgs(args, &v)
-	TakeNoOpt(opts)
+// error(nil) is treated as nil by reflect.TypeOf, so we first get the type of
+// *error and use Elem to obtain type of error.
+var errorType = reflect.TypeOf((*error)(nil)).Elem()
 
-	ec.OutputChan() <- types.Bool(types.ToBool(v))
-}
-
-func not(ec *Frame, args []types.Value, opts map[string]types.Value) {
-	var v types.Value
-	ScanArgs(args, &v)
-	TakeNoOpt(opts)
-
-	ec.OutputChan() <- types.Bool(!types.ToBool(v))
-}
+var errNoOptions = errors.New("function does not accept any options")
 
-func is(ec *Frame, args []types.Value, opts map[string]types.Value) {
-	TakeNoOpt(opts)
-	result := true
-	for i := 0; i+1 < len(args); i++ {
-		if args[i] != args[i+1] {
-			result = false
-			break
+// Call calls the implementation using reflection.
+func (b *BuiltinFn) Call(f *Frame, args []interface{}, opts map[string]interface{}) error {
+	if b.variadicArg != nil {
+		if len(args) < len(b.normalArgs) {
+			return fmt.Errorf("want %d or more arguments, got %d",
+				len(b.normalArgs), len(args))
 		}
-	}
-	ec.OutputChan() <- types.Bool(result)
-}
-
-func eq(ec *Frame, args []types.Value, opts map[string]types.Value) {
-	TakeNoOpt(opts)
-	result := true
-	for i := 0; i+1 < len(args); i++ {
-		if !args[i].Equal(args[i+1]) {
-			result = false
-			break
+	} else if b.inputs {
+		if len(args) != len(b.normalArgs) && len(args) != len(b.normalArgs)+1 {
+			return fmt.Errorf("want %d or %d arguments, got %d",
+				len(b.normalArgs), len(b.normalArgs)+1, len(args))
 		}
+	} else if len(args) != len(b.normalArgs) {
+		return fmt.Errorf("want %d arguments, got %d", len(b.normalArgs), len(args))
 	}
-	ec.OutputChan() <- types.Bool(result)
-}
-
-func notEq(ec *Frame, args []types.Value, opts map[string]types.Value) {
-	TakeNoOpt(opts)
-	result := true
-	for i := 0; i+1 < len(args); i++ {
-		if args[i].Equal(args[i+1]) {
-			result = false
-			break
-		}
+	if !b.options && len(opts) > 0 {
+		return errNoOptions
 	}
-	ec.OutputChan() <- types.Bool(result)
-}
-
-func constantly(ec *Frame, args []types.Value, opts map[string]types.Value) {
-	TakeNoOpt(opts)
 
-	out := ec.ports[1].Chan
-	// XXX Repr of this fn is not right
-	out <- &BuiltinFn{
-		"created by constantly",
-		func(ec *Frame, a []types.Value, o map[string]types.Value) {
-			TakeNoOpt(o)
-			if len(a) != 0 {
-				throw(ErrArgs)
-			}
-			out := ec.ports[1].Chan
-			for _, v := range args {
-				out <- v
-			}
-		},
+	var in []reflect.Value
+	if b.frame {
+		in = append(in, reflect.ValueOf(f))
+	}
+	if b.options {
+		in = append(in, reflect.ValueOf(opts))
+	}
+	for i, arg := range args {
+		var typ reflect.Type
+		if i < len(b.normalArgs) {
+			typ = b.normalArgs[i]
+		} else if b.variadicArg != nil {
+			typ = b.variadicArg
+		} else if b.inputs {
+			break // Handled after the loop
+		} else {
+			panic("impossible")
+		}
+		ptr := reflect.New(typ)
+		err := vals.ScanToGo(arg, ptr.Interface())
+		if err != nil {
+			return fmt.Errorf("wrong type of %d'th argument: %v", i+1, err)
+		}
+		in = append(in, ptr.Elem())
 	}
-}
-
-func source(ec *Frame, args []types.Value, opts map[string]types.Value) {
-	var argFname types.String
-	ScanArgs(args, &argFname)
-	ScanOpts(opts)
-
-	fname := string(argFname)
-	abs, err := filepath.Abs(fname)
-	maybeThrow(err)
-
-	maybeThrow(ec.Source(fname, abs))
-}
 
-func sleep(ec *Frame, args []types.Value, opts map[string]types.Value) {
-	var t float64
-	ScanArgs(args, &t)
-	TakeNoOpt(opts)
-
-	d := time.Duration(float64(time.Second) * t)
-	select {
-	case <-ec.Interrupts():
-		throw(ErrInterrupted)
-	case <-time.After(d):
+	if b.inputs {
+		var inputs Inputs
+		if len(args) == len(b.normalArgs) {
+			inputs = Inputs(f.IterateInputs)
+		} else {
+			// Wrap an iterable argument in Inputs.
+			iterable := args[len(args)-1]
+			inputs = Inputs(func(f func(interface{})) {
+				err := vals.Iterate(iterable, func(v interface{}) bool {
+					f(v)
+					return true
+				})
+				maybeThrow(err)
+			})
+		}
+		in = append(in, reflect.ValueOf(inputs))
 	}
-}
-
-func _time(ec *Frame, args []types.Value, opts map[string]types.Value) {
-	var f Fn
-	ScanArgs(args, &f)
-	TakeNoOpt(opts)
-
-	t0 := time.Now()
-	f.Call(ec, NoArgs, NoOpts)
-	t1 := time.Now()
-
-	dt := t1.Sub(t0)
-	fmt.Fprintln(ec.ports[1].File, dt)
-}
-
-func src(ec *Frame, args []types.Value, opts map[string]types.Value) {
-	TakeNoArg(args)
-	TakeNoOpt(opts)
-
-	ec.OutputChan() <- ec.srcMeta
-}
 
-func _gc(ec *Frame, args []types.Value, opts map[string]types.Value) {
-	TakeNoArg(args)
-	TakeNoOpt(opts)
+	outs := reflect.ValueOf(b.impl).Call(in)
 
-	runtime.GC()
-}
-
-func _stack(ec *Frame, args []types.Value, opts map[string]types.Value) {
-	TakeNoArg(args)
-	TakeNoOpt(opts)
-
-	out := ec.ports[1].File
-	// XXX dup with main.go
-	buf := make([]byte, 1024)
-	for runtime.Stack(buf, true) == cap(buf) {
-		buf = make([]byte, cap(buf)*2)
+	if len(outs) > 0 && outs[len(outs)-1].Type() == errorType {
+		err := outs[len(outs)-1].Interface()
+		if err != nil {
+			return err.(error)
+		}
+		outs = outs[:len(outs)-1]
 	}
-	out.Write(buf)
-}
-
-func _log(ec *Frame, args []types.Value, opts map[string]types.Value) {
-	var fnamev types.String
-	ScanArgs(args, &fnamev)
-	fname := string(fnamev)
-	TakeNoOpt(opts)
-
-	maybeThrow(util.SetOutputFile(fname))
-}
 
-func _ifaddrs(ec *Frame, args []types.Value, opts map[string]types.Value) {
-	TakeNoArg(args)
-	TakeNoOpt(opts)
-
-	out := ec.ports[1].Chan
-
-	addrs, err := net.InterfaceAddrs()
-	maybeThrow(err)
-	for _, addr := range addrs {
-		out <- types.String(addr.String())
+	for _, out := range outs {
+		f.OutputChan() <- vals.FromGo(out.Interface())
 	}
+	return nil
 }
diff -pruN 0.11+ds1-5/eval/builtin_fn_io.go 0.12+ds1-1/eval/builtin_fn_io.go
--- 0.11+ds1-5/eval/builtin_fn_io.go	2018-01-18 02:03:11.000000000 +0000
+++ 0.12+ds1-1/eval/builtin_fn_io.go	2018-07-11 02:32:25.000000000 +0000
@@ -7,117 +7,97 @@ import (
 	"io/ioutil"
 	"os"
 
-	"github.com/elves/elvish/eval/types"
+	"github.com/elves/elvish/eval/vals"
 )
 
 // Input and output.
 
 func init() {
-	addToBuiltinFns([]*BuiltinFn{
+	addBuiltinFns(map[string]interface{}{
 		// Value output
-		{"put", put},
+		"put": put,
 
 		// Bytes output
-		{"print", print},
-		{"echo", echo},
-		{"pprint", pprint},
-		{"repr", repr},
+		"print":  print,
+		"echo":   echo,
+		"pprint": pprint,
+		"repr":   repr,
 
 		// Bytes to value
-		{"slurp", slurp},
-		{"from-lines", fromLines},
-		{"from-json", fromJSON},
+		"slurp":      slurp,
+		"from-lines": fromLines,
+		"from-json":  fromJSON,
 
 		// Value to bytes
-		{"to-lines", toLines},
-		{"to-json", toJSON},
+		"to-lines": toLines,
+		"to-json":  toJSON,
 
 		// File and pipe
-		{"fopen", fopen},
-		{"fclose", fclose},
-		{"pipe", pipe},
-		{"prclose", prclose},
-		{"pwclose", pwclose},
+		"fopen":   fopen,
+		"fclose":  fclose,
+		"pipe":    pipe,
+		"prclose": prclose,
+		"pwclose": pwclose,
 	})
 }
 
-func put(ec *Frame, args []types.Value, opts map[string]types.Value) {
-	TakeNoOpt(opts)
-	out := ec.ports[1].Chan
+func put(fm *Frame, args ...interface{}) {
+	out := fm.ports[1].Chan
 	for _, a := range args {
 		out <- a
 	}
 }
 
-func print(ec *Frame, args []types.Value, opts map[string]types.Value) {
-	var sepv types.String
-	ScanOpts(opts, OptToScan{"sep", &sepv, types.String(" ")})
+func print(fm *Frame, rawOpts RawOptions, args ...interface{}) {
+	opts := struct{ Sep string }{" "}
+	rawOpts.Scan(&opts)
 
-	out := ec.ports[1].File
-	sep := string(sepv)
+	out := fm.ports[1].File
 	for i, arg := range args {
 		if i > 0 {
-			out.WriteString(sep)
+			out.WriteString(opts.Sep)
 		}
-		out.WriteString(types.ToString(arg))
+		out.WriteString(vals.ToString(arg))
 	}
 }
 
-func echo(ec *Frame, args []types.Value, opts map[string]types.Value) {
-	print(ec, args, opts)
-	ec.ports[1].File.WriteString("\n")
+func echo(fm *Frame, opts RawOptions, args ...interface{}) {
+	print(fm, opts, args...)
+	fm.ports[1].File.WriteString("\n")
 }
 
-func pprint(ec *Frame, args []types.Value, opts map[string]types.Value) {
-	TakeNoOpt(opts)
-	out := ec.ports[1].File
+func pprint(fm *Frame, args ...interface{}) {
+	out := fm.ports[1].File
 	for _, arg := range args {
-		out.WriteString(arg.Repr(0))
+		out.WriteString(vals.Repr(arg, 0))
 		out.WriteString("\n")
 	}
 }
 
-func repr(ec *Frame, args []types.Value, opts map[string]types.Value) {
-	TakeNoOpt(opts)
-	out := ec.ports[1].File
+func repr(fm *Frame, args ...interface{}) {
+	out := fm.ports[1].File
 	for i, arg := range args {
 		if i > 0 {
 			out.WriteString(" ")
 		}
-		out.WriteString(arg.Repr(types.NoPretty))
+		out.WriteString(vals.Repr(arg, vals.NoPretty))
 	}
 	out.WriteString("\n")
 }
 
-func slurp(ec *Frame, args []types.Value, opts map[string]types.Value) {
-	TakeNoArg(args)
-	TakeNoOpt(opts)
-
-	in := ec.ports[0].File
-	out := ec.ports[1].Chan
-
-	all, err := ioutil.ReadAll(in)
-	maybeThrow(err)
-	out <- types.String(string(all))
+func slurp(fm *Frame) (string, error) {
+	b, err := ioutil.ReadAll(fm.ports[0].File)
+	return string(b), err
 }
 
-func fromLines(ec *Frame, args []types.Value, opts map[string]types.Value) {
-	TakeNoArg(args)
-	TakeNoOpt(opts)
-
-	in := ec.ports[0].File
-	out := ec.ports[1].Chan
-
-	linesToChan(in, out)
+func fromLines(fm *Frame) {
+	linesToChan(fm.ports[0].File, fm.ports[1].Chan)
 }
 
 // fromJSON parses a stream of JSON data into Value's.
-func fromJSON(ec *Frame, args []types.Value, opts map[string]types.Value) {
-	TakeNoArg(args)
-	TakeNoOpt(opts)
-
-	in := ec.ports[0].File
-	out := ec.ports[1].Chan
+func fromJSON(fm *Frame) error {
+	in := fm.ports[0].File
+	out := fm.ports[1].Chan
 
 	dec := json.NewDecoder(in)
 	var v interface{}
@@ -125,82 +105,53 @@ func fromJSON(ec *Frame, args []types.Va
 		err := dec.Decode(&v)
 		if err != nil {
 			if err == io.EOF {
-				return
+				return nil
 			}
-			throw(err)
+			return err
 		}
 		out <- FromJSONInterface(v)
 	}
+	return nil
 }
 
-func toLines(ec *Frame, args []types.Value, opts map[string]types.Value) {
-	iterate := ScanArgsOptionalInput(ec, args)
-	TakeNoOpt(opts)
-
-	out := ec.ports[1].File
+func toLines(fm *Frame, inputs Inputs) {
+	out := fm.ports[1].File
 
-	iterate(func(v types.Value) {
-		fmt.Fprintln(out, types.ToString(v))
+	inputs(func(v interface{}) {
+		fmt.Fprintln(out, vals.ToString(v))
 	})
 }
 
 // toJSON converts a stream of Value's to JSON data.
-func toJSON(ec *Frame, args []types.Value, opts map[string]types.Value) {
-	iterate := ScanArgsOptionalInput(ec, args)
-	TakeNoOpt(opts)
-
-	out := ec.ports[1].File
+func toJSON(fm *Frame, inputs Inputs) {
+	out := fm.ports[1].File
 
 	enc := json.NewEncoder(out)
-	iterate(func(v types.Value) {
+	inputs(func(v interface{}) {
 		err := enc.Encode(v)
 		maybeThrow(err)
 	})
 }
 
-func fopen(ec *Frame, args []types.Value, opts map[string]types.Value) {
-	var namev types.String
-	ScanArgs(args, &namev)
-	name := string(namev)
-	TakeNoOpt(opts)
-
+func fopen(fm *Frame, name string) (vals.File, error) {
 	// TODO support opening files for writing etc as well.
-	out := ec.ports[1].Chan
 	f, err := os.Open(name)
-	maybeThrow(err)
-	out <- types.File{f}
+	return vals.File{f}, err
 }
 
-func fclose(ec *Frame, args []types.Value, opts map[string]types.Value) {
-	var f types.File
-	ScanArgs(args, &f)
-	TakeNoOpt(opts)
-
-	maybeThrow(f.Inner.Close())
+func fclose(f vals.File) error {
+	return f.Inner.Close()
 }
 
-func pipe(ec *Frame, args []types.Value, opts map[string]types.Value) {
-	TakeNoArg(args)
-	TakeNoOpt(opts)
-
+func pipe() (vals.Pipe, error) {
 	r, w, err := os.Pipe()
-	out := ec.ports[1].Chan
-	maybeThrow(err)
-	out <- types.Pipe{r, w}
+	return vals.Pipe{r, w}, err
 }
 
-func prclose(ec *Frame, args []types.Value, opts map[string]types.Value) {
-	var p types.Pipe
-	ScanArgs(args, &p)
-	TakeNoOpt(opts)
-
-	maybeThrow(p.ReadEnd.Close())
+func prclose(p vals.Pipe) error {
+	return p.ReadEnd.Close()
 }
 
-func pwclose(ec *Frame, args []types.Value, opts map[string]types.Value) {
-	var p types.Pipe
-	ScanArgs(args, &p)
-	TakeNoOpt(opts)
-
-	maybeThrow(p.WriteEnd.Close())
+func pwclose(p vals.Pipe) error {
+	return p.WriteEnd.Close()
 }
diff -pruN 0.11+ds1-5/eval/builtin_fn_io_test.go 0.12+ds1-1/eval/builtin_fn_io_test.go
--- 0.11+ds1-5/eval/builtin_fn_io_test.go	2018-01-18 02:03:11.000000000 +0000
+++ 0.12+ds1-1/eval/builtin_fn_io_test.go	2018-07-11 02:32:25.000000000 +0000
@@ -3,35 +3,33 @@ package eval
 import (
 	"testing"
 
-	"github.com/elves/elvish/eval/types"
+	"github.com/elves/elvish/eval/vals"
 )
 
 func TestBuiltinFnIO(t *testing.T) {
-	runTests(t, []Test{
-		{`put foo bar`, want{out: strs("foo", "bar")}},
+	Test(t,
+		That(`put foo bar`).Puts("foo", "bar"),
 
-		{`print [foo bar]`, want{bytesOut: []byte("[foo bar]")}},
-		{`echo [foo bar]`, want{bytesOut: []byte("[foo bar]\n")}},
-		{`pprint [foo bar]`, want{bytesOut: []byte("[\n foo\n bar\n]\n")}},
-		NewTest(`repr foo bar ['foo bar']`).WantBytesOutString("foo bar ['foo bar']\n"),
+		That(`print [foo bar]`).Prints("[foo bar]"),
+		That(`print foo bar &sep=,`).Prints("foo,bar"),
+		That(`echo [foo bar]`).Prints("[foo bar]\n"),
+		That(`pprint [foo bar]`).Prints("[\n foo\n bar\n]\n"),
+		That(`repr foo bar ['foo bar']`).Prints("foo bar ['foo bar']\n"),
 
-		{`print "a\nb" | slurp`, want{out: strs("a\nb")}},
-		{`print "a\nb" | from-lines`, want{out: strs("a", "b")}},
-		{`print "a\nb\n" | from-lines`, want{out: strs("a", "b")}},
-		{`echo '{"k": "v", "a": [1, 2]}' '"foo"' | from-json`,
-			want{out: []types.Value{
-				types.MakeMap(map[types.Value]types.Value{
-					types.String("k"): types.String("v"),
-					types.String("a"): types.MakeList(strs("1", "2")...)}),
-				types.String("foo"),
-			}}},
-		{`echo 'invalid' | from-json`, want{err: errAny}},
+		That(`print "a\nb" | slurp`).Puts("a\nb"),
+		That(`print "a\nb" | from-lines`).Puts("a", "b"),
+		That(`print "a\nb\n" | from-lines`).Puts("a", "b"),
+		That(`echo '{"k": "v", "a": [1, 2]}' '"foo"' | from-json`).
+			Puts(vals.MakeMap(map[interface{}]interface{}{
+				"k": "v",
+				"a": vals.MakeList("1", "2")}),
+				"foo"),
+		That(`echo 'invalid' | from-json`).Errors(),
 
-		{`put "l\norem" ipsum | to-lines`,
-			want{bytesOut: []byte("l\norem\nipsum\n")}},
-		{`put [&k=v &a=[1 2]] foo | to-json`,
-			want{bytesOut: []byte(`{"a":["1","2"],"k":"v"}
+		That(`put "l\norem" ipsum | to-lines`).Prints("l\norem\nipsum\n"),
+		That(`put [&k=v &a=[1 2]] foo | to-json`).
+			Prints(`{"a":["1","2"],"k":"v"}
 "foo"
-`)}},
-	})
+`),
+	)
 }
diff -pruN 0.11+ds1-5/eval/builtin_fn_misc.go 0.12+ds1-1/eval/builtin_fn_misc.go
--- 0.11+ds1-5/eval/builtin_fn_misc.go	1970-01-01 00:00:00.000000000 +0000
+++ 0.12+ds1-1/eval/builtin_fn_misc.go	2018-07-11 02:32:25.000000000 +0000
@@ -0,0 +1,149 @@
+package eval
+
+// Builtin functions.
+
+import (
+	"fmt"
+	"io/ioutil"
+	"math/rand"
+	"net"
+	"path/filepath"
+	"time"
+	"unicode/utf8"
+
+	"github.com/elves/elvish/eval/vals"
+	"github.com/elves/elvish/parse"
+)
+
+// Builtins that have not been put into their own groups go here.
+
+func init() {
+	addBuiltinFns(map[string]interface{}{
+		"nop":        nop,
+		"kind-of":    kindOf,
+		"constantly": constantly,
+
+		"resolve": resolve,
+
+		"-source": source,
+
+		// Time
+		"esleep": sleep,
+		"-time":  _time,
+
+		"-ifaddrs": _ifaddrs,
+	})
+
+	// For rand and randint.
+	rand.Seed(time.Now().UTC().UnixNano())
+}
+
+func nop(opts RawOptions, args ...interface{}) {
+	// Do nothing
+}
+
+func kindOf(fm *Frame, args ...interface{}) {
+	out := fm.ports[1].Chan
+	for _, a := range args {
+		out <- vals.Kind(a)
+	}
+}
+
+func constantly(args ...interface{}) Callable {
+	// XXX Repr of this fn is not right
+	return NewBuiltinFn(
+		"created by constantly",
+		func(fm *Frame) {
+			out := fm.ports[1].Chan
+			for _, v := range args {
+				out <- v
+			}
+		},
+	)
+}
+
+func resolve(fm *Frame, head string) string {
+	// Emulate static resolution of a command head. This needs to be kept in
+	// sync with (*compiler).form.
+
+	_, special := builtinSpecials[head]
+	if special {
+		return "special"
+	} else {
+		explode, ns, name := ParseVariableRef(head)
+		if !explode && fm.ResolveVar(ns, name+FnSuffix) != nil {
+			return "$" + head + FnSuffix
+		} else {
+			return "(external " + parse.Quote(head) + ")"
+		}
+	}
+}
+
+func source(fm *Frame, fname string) error {
+	path, err := filepath.Abs(fname)
+	if err != nil {
+		return err
+	}
+	code, err := readFileUTF8(path)
+	if err != nil {
+		return err
+	}
+	n, err := parse.Parse(fname, code)
+	if err != nil {
+		return err
+	}
+	scriptGlobal := fm.local.static()
+	for name := range fm.up.static() {
+		scriptGlobal.set(name)
+	}
+	op, err := compile(fm.Builtin.static(),
+		scriptGlobal, n, NewScriptSource(fname, path, code))
+	if err != nil {
+		return err
+	}
+	return fm.Eval(op)
+}
+
+func readFileUTF8(fname string) (string, error) {
+	bytes, err := ioutil.ReadFile(fname)
+	if err != nil {
+		return "", err
+	}
+	if !utf8.Valid(bytes) {
+		return "", fmt.Errorf("%s: source is not valid UTF-8", fname)
+	}
+	return string(bytes), nil
+}
+
+func sleep(fm *Frame, t float64) error {
+	d := time.Duration(float64(time.Second) * t)
+	select {
+	case <-fm.Interrupts():
+		return ErrInterrupted
+	case <-time.After(d):
+		return nil
+	}
+}
+
+func _time(fm *Frame, f Callable) error {
+	t0 := time.Now()
+	err := f.Call(fm, NoArgs, NoOpts)
+	t1 := time.Now()
+
+	dt := t1.Sub(t0)
+	fmt.Fprintln(fm.ports[1].File, dt)
+
+	return err
+}
+
+func _ifaddrs(fm *Frame) error {
+	addrs, err := net.InterfaceAddrs()
+	if err != nil {
+		return err
+	}
+	out := fm.ports[1].Chan
+	for _, addr := range addrs {
+		out <- addr.String()
+	}
+	return nil
+}
diff -pruN 0.11+ds1-5/eval/builtin_fn_misc_test.go 0.12+ds1-1/eval/builtin_fn_misc_test.go
--- 0.11+ds1-5/eval/builtin_fn_misc_test.go	1970-01-01 00:00:00.000000000 +0000
+++ 0.12+ds1-1/eval/builtin_fn_misc_test.go	2018-07-11 02:32:25.000000000 +0000
@@ -0,0 +1,30 @@
+package eval
+
+import (
+	"testing"
+
+	"github.com/elves/elvish/util"
+)
+
+func TestBuiltinFnMisc(t *testing.T) {
+	Test(t,
+		That(`f = (mktemp elvXXXXXX); echo 'put x' > $f
+		      -source $f; rm $f`).Puts("x"),
+		That(`f = (mktemp elvXXXXXX); echo 'put $x' > $f
+		      fn p [x]{ -source $f }; p x; rm $f`).Puts("x"),
+	)
+}
+
+func TestResolve(t *testing.T) {
+	util.InTempDir(func(libdir string) {
+		MustWriteFile("mod.elv", []byte("fn func { }"), 0600)
+
+		TestWithSetup(t, func(ev *Evaler) { ev.SetLibDir(libdir) },
+			That("resolve for").Puts("special"),
+			That("resolve put").Puts("$put~"),
+			That("fn f { }; resolve f").Puts("$f~"),
+			That("use mod; resolve mod:func").Puts("$mod:func~"),
+			That("resolve cat").Puts("(external cat)"),
+		)
+	})
+}
diff -pruN 0.11+ds1-5/eval/builtin_fn_num.go 0.12+ds1-1/eval/builtin_fn_num.go
--- 0.11+ds1-5/eval/builtin_fn_num.go	2018-01-18 02:03:11.000000000 +0000
+++ 0.12+ds1-1/eval/builtin_fn_num.go	2018-07-11 02:32:25.000000000 +0000
@@ -3,170 +3,138 @@ package eval
 import (
 	"math"
 	"math/rand"
-	"strconv"
 
-	"github.com/elves/elvish/eval/types"
+	"github.com/elves/elvish/eval/vals"
 )
 
 // Numerical operations.
 
 func init() {
-	addToBuiltinFns([]*BuiltinFn{
+	addBuiltinFns(map[string]interface{}{
 		// Comparison
-		{"<",
-			wrapNumCompare(func(a, b float64) bool { return a < b })},
-		{"<=",
-			wrapNumCompare(func(a, b float64) bool { return a <= b })},
-		{"==",
-			wrapNumCompare(func(a, b float64) bool { return a == b })},
-		{"!=",
-			wrapNumCompare(func(a, b float64) bool { return a != b })},
-		{">",
-			wrapNumCompare(func(a, b float64) bool { return a > b })},
-		{">=",
-			wrapNumCompare(func(a, b float64) bool { return a >= b })},
+		"<":  lt,
+		"<=": le,
+		"==": eqNum,
+		"!=": ne,
+		">":  gt,
+		">=": ge,
 
 		// Arithmetics
-		{"+", plus},
-		{"-", minus},
-		{"*", times},
-		{"/", slash},
-		{"^", pow},
-		{"%", mod},
+		"+": plus,
+		"-": minus,
+		"*": times,
+		"/": slash,
+		"^": math.Pow,
+		"%": func(a, b int) int { return a % b },
 
 		// Random
-		{"rand", randFn},
-		{"randint", randint},
+		"rand":    rand.Float64,
+		"randint": randint,
 	})
 }
 
-func wrapNumCompare(cmp func(a, b float64) bool) BuiltinFnImpl {
-	return func(ec *Frame, args []types.Value, opts map[string]types.Value) {
-		TakeNoOpt(opts)
-		floats := make([]float64, len(args))
-		for i, a := range args {
-			f, err := toFloat(a)
-			maybeThrow(err)
-			floats[i] = f
-		}
-		result := true
-		for i := 0; i < len(floats)-1; i++ {
-			if !cmp(floats[i], floats[i+1]) {
-				result = false
-				break
-			}
-		}
-		ec.OutputChan() <- types.Bool(result)
+func lt(nums ...float64) bool {
+	for i := 0; i < len(nums)-1; i++ {
+		if !(nums[i] < nums[i+1]) {
+			return false
+		}
+	}
+	return true
+}
+
+func le(nums ...float64) bool {
+	for i := 0; i < len(nums)-1; i++ {
+		if !(nums[i] <= nums[i+1]) {
+			return false
+		}
+	}
+	return true
+}
+
+func eqNum(nums ...float64) bool {
+	for i := 0; i < len(nums)-1; i++ {
+		if !(nums[i] == nums[i+1]) {
+			return false
+		}
 	}
+	return true
 }
 
-func plus(ec *Frame, args []types.Value, opts map[string]types.Value) {
-	var nums []float64
-	ScanArgsVariadic(args, &nums)
-	TakeNoOpt(opts)
+func ne(nums ...float64) bool {
+	for i := 0; i < len(nums)-1; i++ {
+		if !(nums[i] != nums[i+1]) {
+			return false
+		}
+	}
+	return true
+}
+
+func gt(nums ...float64) bool {
+	for i := 0; i < len(nums)-1; i++ {
+		if !(nums[i] > nums[i+1]) {
+			return false
+		}
+	}
+	return true
+}
+
+func ge(nums ...float64) bool {
+	for i := 0; i < len(nums)-1; i++ {
+		if !(nums[i] >= nums[i+1]) {
+			return false
+		}
+	}
+	return true
+}
 
-	out := ec.ports[1].Chan
+func plus(nums ...float64) float64 {
 	sum := 0.0
 	for _, f := range nums {
 		sum += f
 	}
-	out <- floatToString(sum)
+	return sum
 }
 
-func minus(ec *Frame, args []types.Value, opts map[string]types.Value) {
-	var (
-		sum  float64
-		nums []float64
-	)
-	ScanArgsVariadic(args, &sum, &nums)
-	TakeNoOpt(opts)
-
-	out := ec.ports[1].Chan
+func minus(sum float64, nums ...float64) float64 {
 	if len(nums) == 0 {
 		// Unary -
-		sum = -sum
-	} else {
-		for _, f := range nums {
-			sum -= f
-		}
+		return -sum
 	}
-	out <- floatToString(sum)
+	for _, f := range nums {
+		sum -= f
+	}
+	return sum
 }
 
-func times(ec *Frame, args []types.Value, opts map[string]types.Value) {
-	var nums []float64
-	ScanArgsVariadic(args, &nums)
-	TakeNoOpt(opts)
-
-	out := ec.ports[1].Chan
+func times(nums ...float64) float64 {
 	prod := 1.0
 	for _, f := range nums {
 		prod *= f
 	}
-	out <- floatToString(prod)
+	return prod
 }
 
-func slash(ec *Frame, args []types.Value, opts map[string]types.Value) {
-	TakeNoOpt(opts)
+func slash(fm *Frame, args ...float64) error {
 	if len(args) == 0 {
 		// cd /
-		cdInner("/", ec)
-		return
+		return fm.Chdir("/")
 	}
 	// Division
-	divide(ec, args, opts)
+	divide(fm, args[0], args[1:]...)
+	return nil
 }
 
-func divide(ec *Frame, args []types.Value, opts map[string]types.Value) {
-	var (
-		prod float64
-		nums []float64
-	)
-	ScanArgsVariadic(args, &prod, &nums)
-	TakeNoOpt(opts)
-
-	out := ec.ports[1].Chan
+func divide(fm *Frame, prod float64, nums ...float64) {
+	out := fm.ports[1].Chan
 	for _, f := range nums {
 		prod /= f
 	}
-	out <- floatToString(prod)
-}
-
-func pow(ec *Frame, args []types.Value, opts map[string]types.Value) {
-	var b, p float64
-	ScanArgs(args, &b, &p)
-	TakeNoOpt(opts)
-
-	out := ec.ports[1].Chan
-	out <- floatToString(math.Pow(b, p))
-}
-
-func mod(ec *Frame, args []types.Value, opts map[string]types.Value) {
-	var a, b int
-	ScanArgs(args, &a, &b)
-	TakeNoOpt(opts)
-
-	out := ec.ports[1].Chan
-	out <- types.String(strconv.Itoa(a % b))
+	out <- vals.FromGo(prod)
 }
 
-func randFn(ec *Frame, args []types.Value, opts map[string]types.Value) {
-	TakeNoArg(args)
-	TakeNoOpt(opts)
-
-	out := ec.ports[1].Chan
-	out <- floatToString(rand.Float64())
-}
-
-func randint(ec *Frame, args []types.Value, opts map[string]types.Value) {
-	var low, high int
-	ScanArgs(args, &low, &high)
-	TakeNoOpt(opts)
-
+func randint(low, high int) (int, error) {
 	if low >= high {
-		throw(ErrArgs)
+		return 0, ErrArgs
 	}
-	out := ec.ports[1].Chan
-	i := low + rand.Intn(high-low)
-	out <- types.String(strconv.Itoa(i))
+	return low + rand.Intn(high-low), nil
 }
diff -pruN 0.11+ds1-5/eval/builtin_fn_num_test.go 0.12+ds1-1/eval/builtin_fn_num_test.go
--- 0.11+ds1-5/eval/builtin_fn_num_test.go	2018-01-18 02:03:11.000000000 +0000
+++ 0.12+ds1-1/eval/builtin_fn_num_test.go	2018-07-11 02:32:25.000000000 +0000
@@ -3,20 +3,28 @@ package eval
 import "testing"
 
 func TestBuiltinFnNum(t *testing.T) {
-	runTests(t, []Test{
-		{`== 1 1.0`, wantTrue},
-		{`== 10 0xa`, wantTrue},
-		{`== a a`, want{err: errAny}},
-		{`> 0x10 1`, wantTrue},
+	Test(t,
+		That("< 1 2 3").Puts(true),
+		That("< 1 3 2").Puts(false),
+		That("<= 1 1 2").Puts(true),
+		That("<= 1 2 1").Puts(false),
+		That("== 1 1 1").Puts(true),
+		That("== 1 2 1").Puts(false),
+		That("!= 1 2 1").Puts(true),
+		That("!= 1 1 2").Puts(false),
+		That("> 3 2 1").Puts(true),
+		That("> 3 1 2").Puts(false),
+		That(">= 3 3 2").Puts(true),
+		That(">= 3 2 3").Puts(false),
 
 		// TODO test more edge cases
-		{"+ 233100 233", want{out: strs("233333")}},
-		{"- 233333 233100", want{out: strs("233")}},
-		{"- 233", want{out: strs("-233")}},
-		{"* 353 661", want{out: strs("233333")}},
-		{"/ 233333 353", want{out: strs("661")}},
-		{"/ 1 0", want{out: strs("+Inf")}},
-		{"^ 16 2", want{out: strs("256")}},
-		{"% 23 7", want{out: strs("2")}},
-	})
+		That("+ 233100 233").Puts("233333"),
+		That("- 233333 233100").Puts("233"),
+		That("- 233").Puts("-233"),
+		That("* 353 661").Puts("233333"),
+		That("/ 233333 353").Puts("661"),
+		That("/ 1 0").Puts("+Inf"),
+		That("^ 16 2").Puts("256"),
+		That("% 23 7").Puts("2"),
+	)
 }
diff -pruN 0.11+ds1-5/eval/builtin_fn_pred.go 0.12+ds1-1/eval/builtin_fn_pred.go
--- 0.11+ds1-5/eval/builtin_fn_pred.go	1970-01-01 00:00:00.000000000 +0000
+++ 0.12+ds1-1/eval/builtin_fn_pred.go	2018-07-11 02:32:25.000000000 +0000
@@ -0,0 +1,46 @@
+package eval
+
+import "github.com/elves/elvish/eval/vals"
+
+// Basic predicate commands.
+
+func init() {
+	addBuiltinFns(map[string]interface{}{
+		"bool":   vals.Bool,
+		"not":    not,
+		"is":     is,
+		"eq":     eq,
+		"not-eq": notEq,
+	})
+}
+
+func not(v interface{}) bool {
+	return !vals.Bool(v)
+}
+
+func is(args ...interface{}) bool {
+	for i := 0; i+1 < len(args); i++ {
+		if args[i] != args[i+1] {
+			return false
+		}
+	}
+	return true
+}
+
+func eq(args ...interface{}) bool {
+	for i := 0; i+1 < len(args); i++ {
+		if !vals.Equal(args[i], args[i+1]) {
+			return false
+		}
+	}
+	return true
+}
+
+func notEq(args ...interface{}) bool {
+	for i := 0; i+1 < len(args); i++ {
+		if vals.Equal(args[i], args[i+1]) {
+			return false
+		}
+	}
+	return true
+}
diff -pruN 0.11+ds1-5/eval/builtin_fn_pred_test.go 0.12+ds1-1/eval/builtin_fn_pred_test.go
--- 0.11+ds1-5/eval/builtin_fn_pred_test.go	1970-01-01 00:00:00.000000000 +0000
+++ 0.12+ds1-1/eval/builtin_fn_pred_test.go	2018-07-11 02:32:25.000000000 +0000
@@ -0,0 +1,36 @@
+package eval
+
+import (
+	"testing"
+)
+
+func TestBool(t *testing.T) {
+	Test(t,
+		That(`bool $true`).Puts(true),
+		That(`bool a`).Puts(true),
+		That(`bool [a]`).Puts(true),
+		// "Empty" values are also true in Elvish
+		That(`bool []`).Puts(true),
+		That(`bool [&]`).Puts(true),
+		That(`bool 0`).Puts(true),
+		That(`bool ""`).Puts(true),
+		// Only errors and $false are false
+		That(`bool ?(fail x)`).Puts(false),
+		That(`bool $false`).Puts(false),
+
+		That(`not $false`).Puts(true),
+		That(`not ?(fail x)`).Puts(true),
+		That(`not $true`).Puts(false),
+		That(`not 0`).Puts(false),
+
+		That(`is 1 1`).Puts(true),
+		That(`is a b`).Puts(false),
+		That(`is [] []`).Puts(true),
+		That(`is [1] [1]`).Puts(false),
+		That(`eq 1 1`).Puts(true),
+		That(`eq a b`).Puts(false),
+		That(`eq [] []`).Puts(true),
+		That(`eq [1] [1]`).Puts(true),
+		That(`not-eq a b`).Puts(true),
+	)
+}
diff -pruN 0.11+ds1-5/eval/builtin_fn_str.go 0.12+ds1-1/eval/builtin_fn_str.go
--- 0.11+ds1-5/eval/builtin_fn_str.go	2018-01-18 02:03:11.000000000 +0000
+++ 0.12+ds1-1/eval/builtin_fn_str.go	2018-07-11 02:32:25.000000000 +0000
@@ -7,7 +7,7 @@ import (
 	"strconv"
 	"strings"
 
-	"github.com/elves/elvish/eval/types"
+	"github.com/elves/elvish/eval/vals"
 	"github.com/elves/elvish/util"
 )
 
@@ -16,124 +16,82 @@ import (
 var ErrInput = errors.New("input error")
 
 func init() {
-	addToBuiltinFns([]*BuiltinFn{
-		{"<s",
-			wrapStrCompare(func(a, b string) bool { return a < b })},
-		{"<=s",
-			wrapStrCompare(func(a, b string) bool { return a <= b })},
-		{"==s",
-			wrapStrCompare(func(a, b string) bool { return a == b })},
-		{"!=s",
-			wrapStrCompare(func(a, b string) bool { return a != b })},
-		{">s",
-			wrapStrCompare(func(a, b string) bool { return a > b })},
-		{">=s",
-			wrapStrCompare(func(a, b string) bool { return a >= b })},
-
-		{"to-string", toString},
-
-		{"joins", joins},
-		{"splits", splits},
-		{"replaces", replaces},
-
-		{"ord", ord},
-		{"base", base},
+	addBuiltinFns(map[string]interface{}{
+		"<s":  func(a, b string) bool { return a < b },
+		"<=s": func(a, b string) bool { return a <= b },
+		"==s": func(a, b string) bool { return a == b },
+		"!=s": func(a, b string) bool { return a != b },
+		">s":  func(a, b string) bool { return a > b },
+		">=s": func(a, b string) bool { return a >= b },
+
+		"to-string": toString,
+
+		"ord":  ord,
+		"base": base,
+
+		"wcswidth":          util.Wcswidth,
+		"-override-wcwidth": util.OverrideWcwidth,
+
+		"has-prefix": strings.HasPrefix,
+		"has-suffix": strings.HasSuffix,
+
+		"joins":    joins,
+		"splits":   splits,
+		"replaces": replaces,
 
-		{"wcswidth", wcswidth},
-		{"-override-wcwidth", overrideWcwidth},
-
-		{"has-prefix", hasPrefix},
-		{"has-suffix", hasSuffix},
-
-		{"eawk", eawk},
+		"eawk": eawk,
 	})
 }
 
-func wrapStrCompare(cmp func(a, b string) bool) BuiltinFnImpl {
-	return func(ec *Frame, args []types.Value, opts map[string]types.Value) {
-		TakeNoOpt(opts)
-		for _, a := range args {
-			if _, ok := a.(types.String); !ok {
-				throw(ErrArgs)
-			}
-		}
-		result := true
-		for i := 0; i < len(args)-1; i++ {
-			if !cmp(string(args[i].(types.String)), string(args[i+1].(types.String))) {
-				result = false
-				break
-			}
-		}
-		ec.OutputChan() <- types.Bool(result)
-	}
-}
-
 // toString converts all arguments to strings.
-func toString(ec *Frame, args []types.Value, opts map[string]types.Value) {
-	TakeNoOpt(opts)
-	out := ec.OutputChan()
+func toString(fm *Frame, args ...interface{}) {
+	out := fm.OutputChan()
 	for _, a := range args {
-		out <- types.String(types.ToString(a))
+		out <- vals.ToString(a)
 	}
 }
 
 // joins joins all input strings with a delimiter.
-func joins(ec *Frame, args []types.Value, opts map[string]types.Value) {
-	var sepv types.String
-	iterate := ScanArgsOptionalInput(ec, args, &sepv)
-	sep := string(sepv)
-	TakeNoOpt(opts)
-
+func joins(sep string, inputs Inputs) string {
 	var buf bytes.Buffer
-	iterate(func(v types.Value) {
-		if s, ok := v.(types.String); ok {
-			if buf.Len() > 0 {
+	first := true
+	inputs(func(v interface{}) {
+		if s, ok := v.(string); ok {
+			if first {
+				first = false
+			} else {
 				buf.WriteString(sep)
 			}
-			buf.WriteString(string(s))
+			buf.WriteString(s)
 		} else {
-			throwf("join wants string input, got %s", v.Kind())
+			throwf("join wants string input, got %s", vals.Kind(v))
 		}
 	})
-	out := ec.ports[1].Chan
-	out <- types.String(buf.String())
+	return buf.String()
 }
 
 // splits splits an argument strings by a delimiter and writes all pieces.
-func splits(ec *Frame, args []types.Value, opts map[string]types.Value) {
-	var (
-		s, sep types.String
-		optMax int
-	)
-	ScanArgs(args, &sep, &s)
-	ScanOpts(opts, OptToScan{"max", &optMax, types.String("-1")})
+func splits(fm *Frame, rawOpts RawOptions, sep, s string) {
+	opts := struct{ Max int }{-1}
+	rawOpts.Scan(&opts)
 
-	out := ec.ports[1].Chan
-	parts := strings.SplitN(string(s), string(sep), optMax)
+	out := fm.ports[1].Chan
+	parts := strings.SplitN(s, sep, opts.Max)
 	for _, p := range parts {
-		out <- types.String(p)
+		out <- p
 	}
 }
 
-func replaces(ec *Frame, args []types.Value, opts map[string]types.Value) {
-	var (
-		old, repl, s types.String
-		optMax       int
-	)
-	ScanArgs(args, &old, &repl, &s)
-	ScanOpts(opts, OptToScan{"max", &optMax, types.String("-1")})
-
-	ec.ports[1].Chan <- types.String(strings.Replace(string(s), string(old), string(repl), optMax))
-}
-
-func ord(ec *Frame, args []types.Value, opts map[string]types.Value) {
-	var s types.String
-	ScanArgs(args, &s)
-	TakeNoOpt(opts)
+func replaces(rawOpts RawOptions, old, repl, s string) string {
+	opts := struct{ Max int }{-1}
+	rawOpts.Scan(&opts)
+	return strings.Replace(s, old, repl, opts.Max)
+}
 
-	out := ec.ports[1].Chan
+func ord(fm *Frame, s string) {
+	out := fm.ports[1].Chan
 	for _, r := range s {
-		out <- types.String("0x" + strconv.FormatInt(int64(r), 16))
+		out <- "0x" + strconv.FormatInt(int64(r), 16)
 	}
 }
 
@@ -141,61 +99,16 @@ func ord(ec *Frame, args []types.Value,
 // greater than 36.
 var ErrBadBase = errors.New("bad base")
 
-func base(ec *Frame, args []types.Value, opts map[string]types.Value) {
-	var (
-		b    int
-		nums []int
-	)
-	ScanArgsVariadic(args, &b, &nums)
-	TakeNoOpt(opts)
-
+func base(fm *Frame, b int, nums ...int) error {
 	if b < 2 || b > 36 {
-		throw(ErrBadBase)
+		return ErrBadBase
 	}
 
-	out := ec.ports[1].Chan
-
+	out := fm.ports[1].Chan
 	for _, num := range nums {
-		out <- types.String(strconv.FormatInt(int64(num), b))
+		out <- strconv.FormatInt(int64(num), b)
 	}
-}
-
-func wcswidth(ec *Frame, args []types.Value, opts map[string]types.Value) {
-	var s types.String
-	ScanArgs(args, &s)
-	TakeNoOpt(opts)
-
-	out := ec.ports[1].Chan
-	out <- types.String(strconv.Itoa(util.Wcswidth(string(s))))
-}
-
-func overrideWcwidth(ec *Frame, args []types.Value, opts map[string]types.Value) {
-	var (
-		s types.String
-		w int
-	)
-	ScanArgs(args, &s, &w)
-	TakeNoOpt(opts)
-
-	r, err := toRune(s)
-	maybeThrow(err)
-	util.OverrideWcwidth(r, w)
-}
-
-func hasPrefix(ec *Frame, args []types.Value, opts map[string]types.Value) {
-	var s, prefix types.String
-	ScanArgs(args, &s, &prefix)
-	TakeNoOpt(opts)
-
-	ec.OutputChan() <- types.Bool(strings.HasPrefix(string(s), string(prefix)))
-}
-
-func hasSuffix(ec *Frame, args []types.Value, opts map[string]types.Value) {
-	var s, suffix types.String
-	ScanArgs(args, &s, &suffix)
-	TakeNoOpt(opts)
-
-	ec.OutputChan() <- types.Bool(strings.HasSuffix(string(s), string(suffix)))
+	return nil
 }
 
 var eawkWordSep = regexp.MustCompile("[ \t]+")
@@ -205,29 +118,28 @@ var eawkWordSep = regexp.MustCompile("[
 // stripping the line and splitting the line by whitespaces. The function may
 // call break and continue. Overall this provides a similar functionality to
 // awk, hence the name.
-func eawk(ec *Frame, args []types.Value, opts map[string]types.Value) {
-	var f Fn
-	iterate := ScanArgsOptionalInput(ec, args, &f)
-	TakeNoOpt(opts)
-
+func eawk(fm *Frame, f Callable, inputs Inputs) error {
 	broken := false
-	iterate(func(v types.Value) {
+	var err error
+	inputs(func(v interface{}) {
 		if broken {
 			return
 		}
-		line, ok := v.(types.String)
+		line, ok := v.(string)
 		if !ok {
-			throw(ErrInput)
+			broken = true
+			err = ErrInput
+			return
 		}
-		args := []types.Value{line}
-		for _, field := range eawkWordSep.Split(strings.Trim(string(line), " \t"), -1) {
-			args = append(args, types.String(field))
+		args := []interface{}{line}
+		for _, field := range eawkWordSep.Split(strings.Trim(line, " \t"), -1) {
+			args = append(args, field)
 		}
 
-		newec := ec.fork("fn of eawk")
-		// TODO: Close port 0 of newec.
-		ex := newec.PCall(f, args, NoOpts)
-		ClosePorts(newec.ports)
+		newFm := fm.fork("fn of eawk")
+		// TODO: Close port 0 of newFm.
+		ex := newFm.Call(f, args, NoOpts)
+		newFm.Close()
 
 		if ex != nil {
 			switch ex.(*Exception).Cause {
@@ -236,8 +148,10 @@ func eawk(ec *Frame, args []types.Value,
 			case Break:
 				broken = true
 			default:
-				throw(ex)
+				broken = true
+				err = ex
 			}
 		}
 	})
+	return err
 }
diff -pruN 0.11+ds1-5/eval/builtin_fn_str_test.go 0.12+ds1-1/eval/builtin_fn_str_test.go
--- 0.11+ds1-5/eval/builtin_fn_str_test.go	2018-01-18 02:03:11.000000000 +0000
+++ 0.12+ds1-1/eval/builtin_fn_str_test.go	2018-07-11 02:32:25.000000000 +0000
@@ -3,30 +3,32 @@ package eval
 import "testing"
 
 func TestBuiltinFnStr(t *testing.T) {
-	runTests(t, []Test{
-		{`==s haha haha`, wantTrue},
-		{`==s 10 10.0`, wantFalse},
-		{`<s a b`, wantTrue},
-		{`<s 2 10`, wantFalse},
+	Test(t,
+		That(`==s haha haha`).Puts(true),
+		That(`==s 10 10.0`).Puts(false),
+		That(`<s a b`).Puts(true),
+		That(`<s 2 10`).Puts(false),
 
-		{`joins : [/usr /bin /tmp]`, want{out: strs("/usr:/bin:/tmp")}},
-		{`splits : /usr:/bin:/tmp`, want{out: strs("/usr", "/bin", "/tmp")}},
-		{`splits : /usr:/bin:/tmp &max=2`, want{out: strs("/usr", "/bin:/tmp")}},
-		{`replaces : / ":usr:bin:tmp"`, want{out: strs("/usr/bin/tmp")}},
-		{`replaces &max=2 : / :usr:bin:tmp`, want{out: strs("/usr/bin:tmp")}},
+		That(`joins : [/usr /bin /tmp]`).Puts("/usr:/bin:/tmp"),
+		That(`joins : ['' a '']`).Puts(":a:"),
+		That(`splits : /usr:/bin:/tmp`).Puts("/usr", "/bin", "/tmp"),
+		That(`splits : /usr:/bin:/tmp &max=2`).Puts("/usr", "/bin:/tmp"),
+		That(`replaces : / ":usr:bin:tmp"`).Puts("/usr/bin/tmp"),
+		That(`replaces &max=2 : / :usr:bin:tmp`).Puts("/usr/bin:tmp"),
 
-		{`ord a`, want{out: strs("0x61")}},
-		{`base 16 42 233`, want{out: strs("2a", "e9")}},
-		{`base 1 1`, want{err: errAny}},   // no base-1
-		{`base 37 10`, want{err: errAny}}, // no letter for base-37
-		{`wcswidth 你好`, want{out: strs("4")}},
-		{`-override-wcwidth x 10; wcswidth 1x2x; -override-wcwidth x 1`, want{out: strs("22")}},
+		That(`ord a`).Puts("0x61"),
+		That(`base 16 42 233`).Puts("2a", "e9"),
+		That(`base 1 1`).Errors(),   // no base-1
+		That(`base 37 10`).Errors(), // no letter for base-37
+		That(`wcswidth 你好`).Puts("4"),
+		That(`-override-wcwidth x 10; wcswidth 1x2x; -override-wcwidth x 1`).
+			Puts("22"),
 
-		{`has-prefix golang go`, wantTrue},
-		{`has-prefix golang x`, wantFalse},
-		{`has-suffix golang x`, wantFalse},
+		That(`has-prefix golang go`).Puts(true),
+		That(`has-prefix golang x`).Puts(false),
+		That(`has-suffix golang x`).Puts(false),
 
-		{`echo "  ax  by cz  \n11\t22 33" | eawk [@a]{ put $a[-1] }`,
-			want{out: strs("cz", "33")}},
-	})
+		That(`echo "  ax  by cz  \n11\t22 33" | eawk [@a]{ put $a[-1] }`).
+			Puts("cz", "33"),
+	)
 }
diff -pruN 0.11+ds1-5/eval/builtin_fn_styled.go 0.12+ds1-1/eval/builtin_fn_styled.go
--- 0.11+ds1-5/eval/builtin_fn_styled.go	1970-01-01 00:00:00.000000000 +0000
+++ 0.12+ds1-1/eval/builtin_fn_styled.go	2018-07-11 02:32:25.000000000 +0000
@@ -0,0 +1,100 @@
+package eval
+
+import (
+	"errors"
+	"fmt"
+
+	"github.com/elves/elvish/eval/vals"
+	"github.com/elves/elvish/styled"
+)
+
+var errStyledSegmentArgType = errors.New("argument to styled-segment must be a string or a styled segment")
+
+func init() {
+	addBuiltinFns(map[string]interface{}{
+		"styled-segment": styledSegment,
+		"styled":         Styled,
+	})
+}
+
+// Turns a string or styled Segment into a new styled Segment with the attributes
+// from the supplied options applied to it. If the input is already a Segment its
+// attributes are copied and modified.
+func styledSegment(options RawOptions, input interface{}) (*styled.Segment, error) {
+	var text string
+	var style styled.Style
+
+	switch input := input.(type) {
+	case string:
+		text = input
+	case *styled.Segment:
+		text = input.Text
+		style = input.Style
+	default:
+		return nil, errStyledSegmentArgType
+	}
+
+	if err := style.ImportFromOptions(options); err != nil {
+		return nil, err
+	}
+
+	return &styled.Segment{
+		Text:  text,
+		Style: style,
+	}, nil
+}
+
+// Styled turns a string, a styled Segment or a styled Text into a styled Text.
+// This is done by applying a range of transformers to the input.
+func Styled(fm *Frame, input interface{}, transformers ...interface{}) (*styled.Text, error) {
+	var text styled.Text
+
+	switch input := input.(type) {
+	case string:
+		text = styled.Text{styled.Segment{
+			Text:  input,
+			Style: styled.Style{},
+		}}
+	case *styled.Segment:
+		text = styled.Text{*input}
+	case *styled.Text:
+		text = *input
+	default:
+		return nil, fmt.Errorf("expected string, styled segment or styled text; got %s", vals.Kind(input))
+	}
+
+	for _, transformer := range transformers {
+		switch transformer := transformer.(type) {
+		case string:
+			transformerFn := styled.FindTransformer(transformer)
+			if transformerFn == nil {
+				return nil, fmt.Errorf("'%s' is no valid style transformer", transformer)
+			}
+
+			for i, segment := range text {
+				text[i] = transformerFn(segment)
+			}
+
+		case Callable:
+			for i, segment := range text {
+				vs, err := fm.CaptureOutput(transformer, []interface{}{&segment}, NoOpts)
+				if err != nil {
+					return nil, err
+				}
+
+				if n := len(vs); n != 1 {
+					return nil, fmt.Errorf("style transformers must return a single styled segment; got %d", n)
+				} else if transformedSegment, ok := vs[0].(*styled.Segment); !ok {
+					return nil, fmt.Errorf("style transformers must return a styled segment; got %s", vals.Kind(vs[0]))
+				} else {
+					text[i] = *transformedSegment
+				}
+			}
+
+		default:
+			return nil, fmt.Errorf("need string or callable; got %s", vals.Kind(transformer))
+		}
+	}
+
+	return &text, nil
+}
diff -pruN 0.11+ds1-5/eval/builtin_fn_styled_test.go 0.12+ds1-1/eval/builtin_fn_styled_test.go
--- 0.11+ds1-5/eval/builtin_fn_styled_test.go	1970-01-01 00:00:00.000000000 +0000
+++ 0.12+ds1-1/eval/builtin_fn_styled_test.go	2018-07-11 02:32:25.000000000 +0000
@@ -0,0 +1,109 @@
+package eval
+
+import (
+	"errors"
+	"testing"
+)
+
+func TestStyledString(t *testing.T) {
+	Test(t,
+		That("print (styled abc hopefully-never-exists)").ErrorsWith(errors.New("'hopefully-never-exists' is no valid style transformer")),
+		That("print (styled abc bold)").Prints("\033[1mabc\033[m"),
+		That("print (styled abc red cyan)").Prints("\033[36mabc\033[m"),
+		That("print (styled abc bg-green)").Prints("\033[42mabc\033[m"),
+		That("print (styled abc no-dim)").Prints("abc"),
+	)
+}
+
+func TestStyledSegment(t *testing.T) {
+	Test(t,
+		That("print (styled (styled-segment abc &fg-color=cyan) bold)").
+			Prints("\033[1;36mabc\033[m"),
+		That("print (styled (styled-segment (styled-segment abc &fg-color=magenta) &dim=$true) cyan)").
+			Prints("\033[2;36mabc\033[m"),
+		That("print (styled (styled-segment abc &inverse=$true) inverse)").
+			Prints("\033[7mabc\033[m"),
+		That("print (styled (styled-segment abc) toggle-inverse)").
+			Prints("\033[7mabc\033[m"),
+		That("print (styled (styled-segment abc &inverse=$true) no-inverse)").
+			Prints("abc"),
+		That("print (styled (styled-segment abc &inverse=$true) toggle-inverse)").
+			Prints("abc"),
+	)
+}
+
+func TestStyledText(t *testing.T) {
+	Test(t,
+		That("print (styled (styled abc red) blue)").
+			Prints("\033[34mabc\033[m"),
+		That("print (styled (styled abc italic) red)").
+			Prints("\033[3;31mabc\033[m"),
+		That("print (styled (styled abc inverse) inverse)").
+			Prints("\033[7mabc\033[m"),
+		That("print (styled (styled abc inverse) no-inverse)").Prints("abc"),
+		That("print (styled (styled abc inverse) toggle-inverse)").Prints("abc"),
+		That("print (styled (styled abc inverse) toggle-inverse toggle-inverse)").Prints("\033[7mabc\033[m"),
+	)
+}
+
+func TestStyledConcat(t *testing.T) {
+	Test(t,
+		// string+segment
+		That("print abc(styled-segment abc &fg-color=red)").Prints("abc\033[31mabc\033[m"),
+		// segment+string
+		That("print (styled-segment abc &fg-color=red)abc").Prints("\033[31mabc\033[mabc"),
+		// segment+segment
+		That("print (styled-segment abc &bg-color=red)(styled-segment abc &fg-color=red)").Prints("\033[41mabc\033[m\033[31mabc\033[m"),
+		// segment+text
+		That("print (styled-segment abc &underlined=$true)(styled abc lightcyan)").Prints("\033[4mabc\033[m\033[96mabc\033[m"),
+	)
+
+	Test(t,
+		// string+text
+		That("print abc(styled abc blink)").Prints("abc\033[5mabc\033[m"),
+		// text+string
+		That("print (styled abc blink)abc").Prints("\033[5mabc\033[mabc"),
+		// text+segment
+		That("print (styled abc inverse)(styled-segment abc &bg-color=white)").Prints("\033[7mabc\033[m\033[107mabc\033[m"),
+		// text+text
+		That("print (styled abc bold)(styled abc dim)").Prints("\033[1mabc\033[m\033[2mabc\033[m"),
+	)
+}
+
+func TestFunctionalStyleTransformers(t *testing.T) {
+	// lambda
+	Test(t,
+		That("print (styled abc [s]{ put $s })").Prints("abc"),
+		That("print (styled abc [s]{ styled-segment $s &bold=$true &italic=$false })").Prints("\033[1mabc\033[m"),
+		That("print (styled abc italic [s]{ styled-segment $s &bold=$true &italic=$false })").Prints("\033[1mabc\033[m"),
+	)
+
+	// fn
+	Test(t,
+		That("fn f [s]{ put $s }; print (styled abc $f~)").Prints("abc"),
+		That("fn f [s]{ styled-segment $s &bold=$true &italic=$false }; print (styled abc $f~)").Prints("\033[1mabc\033[m"),
+		That("fn f [s]{ styled-segment $s &bold=$true &italic=$false }; print (styled abc italic $f~)").Prints("\033[1mabc\033[m"),
+	)
+
+	// var
+	Test(t,
+		That("f = [s]{ put $s }; print (styled abc $f)").Prints("abc"),
+		That("f = [s]{ styled-segment $s &bold=$true &italic=$false }; print (styled abc $f)").Prints("\033[1mabc\033[m"),
+		That("f = [s]{ styled-segment $s &bold=$true &italic=$false }; print (styled abc italic $f)").Prints("\033[1mabc\033[m"),
+	)
+}
+
+func TestStyledIndexing(t *testing.T) {
+	Test(t,
+		That("put (styled-segment abc &italic=$true &fg-color=red)[bold]").Puts(false),
+		That("put (styled-segment abc &italic=$true &fg-color=red)[italic]").Puts(true),
+		That("put (styled-segment abc &italic=$true &fg-color=red)[fg-color]").Puts("red"),
+	)
+
+	Test(t,
+		That("put (styled abc red)[0][bold]").Puts(false),
+		That("put (styled abc red)[0][bg-color]").Puts("default"),
+		That("t = (styled-segment abc &underlined=$true)(styled abc lightcyan); put $t[1][fg-color]").Puts("lightcyan"),
+		That("t = (styled-segment abc &underlined=$true)(styled abc lightcyan); put $t[1][underlined]").Puts(false),
+	)
+}
diff -pruN 0.11+ds1-5/eval/builtin_fn_test.go 0.12+ds1-1/eval/builtin_fn_test.go
--- 0.11+ds1-5/eval/builtin_fn_test.go	2018-01-18 02:03:11.000000000 +0000
+++ 0.12+ds1-1/eval/builtin_fn_test.go	2018-07-11 02:32:25.000000000 +0000
@@ -1,52 +1,195 @@
 package eval
 
-import "testing"
-
-var builtinFnTests = []Test{
-	NewTest("kind-of $nop~").WantOutStrings("fn"),
-	NewTest("eq $nop~ { }").WantOutBools(false),
-	NewTest("put [&$nop~= foo][$nop~]").WantOutStrings("foo"),
-	NewTest("repr $nop~").WantBytesOutString("<builtin nop>\n"),
-
-	{"nop", wantNothing},
-	{"nop a b", wantNothing},
-	{"nop &k=v", wantNothing},
-	{"nop a b &k=v", wantNothing},
-
-	{"kind-of bare 'str' [] [&] []{ }",
-		want{out: strs("string", "string", "list", "map", "fn")}},
-
-	{`bool $true`, wantTrue},
-	{`bool a`, wantTrue},
-	{`bool [a]`, wantTrue},
-	// "Empty" values are also true in Elvish
-	{`bool []`, wantTrue},
-	{`bool [&]`, wantTrue},
-	{`bool 0`, wantTrue},
-	{`bool ""`, wantTrue},
-	// Only errors and $false are false
-	{`bool ?(fail x)`, wantFalse},
-	{`bool $false`, wantFalse},
-
-	{`not $false`, wantTrue},
-	{`not ?(fail x)`, wantTrue},
-	{`not $true`, wantFalse},
-	{`not 0`, wantFalse},
-
-	{`is 1 1`, wantTrue},
-	{`is a b`, wantFalse},
-	{`is [] []`, wantTrue},
-	{`is [1] [1]`, wantFalse},
-	{`eq 1 1`, wantTrue},
-	{`eq a b`, wantFalse},
-	{`eq [] []`, wantTrue},
-	{`eq [1] [1]`, wantTrue},
-	{`not-eq a b`, wantTrue},
-
-	{`f=(constantly foo); $f; $f`, want{out: strs("foo", "foo")}},
-	{`(constantly foo) bad`, want{err: errAny}},
-}
-
-func TestBuiltinFn(t *testing.T) {
-	runTests(t, builtinFnTests)
+import (
+	"errors"
+	"testing"
+
+	"github.com/elves/elvish/eval/vals"
+)
+
+func TestReflectBuiltinFnCall(t *testing.T) {
+	theFrame := new(Frame)
+	theOptions := map[string]interface{}{}
+
+	var f Callable
+	callGood := func(fm *Frame, args []interface{}, opts map[string]interface{}) {
+		err := f.Call(fm, args, opts)
+		if err != nil {
+			t.Errorf("Failed to call f: %v", err)
+		}
+	}
+	callBad := func(fm *Frame, args []interface{}, opts map[string]interface{}) {
+		err := f.Call(fm, args, opts)
+		if err == nil {
+			t.Errorf("Calling f didn't return error")
+		}
+	}
+
+	// *Frame parameter gets the Frame.
+	f = NewBuiltinFn("f", func(f *Frame) {
+		if f != theFrame {
+			t.Errorf("*Frame parameter doesn't get current frame")
+		}
+	})
+	callGood(theFrame, nil, theOptions)
+
+	// Options parameter gets options.
+	f = NewBuiltinFn("f", func(opts RawOptions) {
+		if opts["foo"] != "bar" {
+			t.Errorf("Options parameter doesn't get options")
+		}
+	})
+	callGood(theFrame, nil, RawOptions{"foo": "bar"})
+
+	// Combination of Frame and Options.
+	f = NewBuiltinFn("f", func(f *Frame, opts RawOptions) {
+		if f != theFrame {
+			t.Errorf("*Frame parameter doesn't get current frame")
+		}
+		if opts["foo"] != "bar" {
+			t.Errorf("Options parameter doesn't get options")
+		}
+	})
+	callGood(theFrame, nil, RawOptions{"foo": "bar"})
+
+	// Argument passing.
+	f = NewBuiltinFn("f", func(x, y string) {
+		if x != "lorem" {
+			t.Errorf("Argument x not passed")
+		}
+		if y != "ipsum" {
+			t.Errorf("Argument y not passed")
+		}
+	})
+	callGood(theFrame, []interface{}{"lorem", "ipsum"}, theOptions)
+
+	// Variadic arguments.
+	f = NewBuiltinFn("f", func(x ...string) {
+		if len(x) != 2 || x[0] != "lorem" || x[1] != "ipsum" {
+			t.Errorf("Variadic argument not passed")
+		}
+	})
+	callGood(theFrame, []interface{}{"lorem", "ipsum"}, theOptions)
+
+	// Conversion into int and float64.
+	f = NewBuiltinFn("f", func(i int, f float64) {
+		if i != 314 {
+			t.Errorf("Integer argument i not passed")
+		}
+		if f != 1.25 {
+			t.Errorf("Float argument f not passed")
+		}
+	})
+	callGood(theFrame, []interface{}{"314", "1.25"}, theOptions)
+
+	// Conversion of supplied inputs.
+	f = NewBuiltinFn("f", func(i Inputs) {
+		var values []interface{}
+		i(func(x interface{}) {
+			values = append(values, x)
+		})
+		if len(values) != 2 || values[0] != "foo" || values[1] != "bar" {
+			t.Errorf("Inputs parameter didn't get supplied inputs")
+		}
+	})
+	callGood(theFrame, []interface{}{vals.MakeList("foo", "bar")}, theOptions)
+
+	// Conversion of implicit inputs.
+	inFrame := &Frame{ports: make([]*Port, 3)}
+	ch := make(chan interface{}, 10)
+	ch <- "foo"
+	ch <- "bar"
+	close(ch)
+	inFrame.ports[0] = &Port{Chan: ch}
+	f = NewBuiltinFn("f", func(i Inputs) {
+		var values []interface{}
+		i(func(x interface{}) {
+			values = append(values, x)
+		})
+		if len(values) != 2 || values[0] != "foo" || values[1] != "bar" {
+			t.Errorf("Inputs parameter didn't get implicit inputs")
+		}
+	})
+	callGood(inFrame, []interface{}{vals.MakeList("foo", "bar")}, theOptions)
+
+	// Outputting of return values.
+	outFrame := &Frame{ports: make([]*Port, 3)}
+	ch = make(chan interface{}, 10)
+	outFrame.ports[1] = &Port{Chan: ch}
+	f = NewBuiltinFn("f", func() string { return "ret" })
+	callGood(outFrame, nil, theOptions)
+	select {
+	case ret := <-ch:
+		if ret != "ret" {
+			t.Errorf("Output is not the same as return value")
+		}
+	default:
+		t.Errorf("Return value is not outputted")
+	}
+
+	// Conversion of return values.
+	f = NewBuiltinFn("f", func() int { return 314 })
+	callGood(outFrame, nil, theOptions)
+	select {
+	case ret := <-ch:
+		if ret != "314" {
+			t.Errorf("Return value is not converted to string")
+		}
+	default:
+		t.Errorf("Return value is not outputted")
+	}
+
+	// Passing of error return value.
+	theError := errors.New("the error")
+	f = NewBuiltinFn("f", func() (string, error) {
+		return "x", theError
+	})
+	if f.Call(outFrame, nil, theOptions) != theError {
+		t.Errorf("Returned error is not passed")
+	}
+	select {
+	case <-ch:
+		t.Errorf("Return value is outputted when error is not nil")
+	default:
+	}
+
+	// Too many arguments.
+	f = NewBuiltinFn("f", func() {
+		t.Errorf("Function called when there are too many arguments")
+	})
+	callBad(theFrame, []interface{}{"x"}, theOptions)
+
+	// Too few arguments.
+	f = NewBuiltinFn("f", func(x string) {
+		t.Errorf("Function called when there are too few arguments")
+	})
+	callBad(theFrame, nil, theOptions)
+	f = NewBuiltinFn("f", func(x string, y ...string) {
+		t.Errorf("Function called when there are too few arguments")
+	})
+	callBad(theFrame, nil, theOptions)
+
+	// Options when the function does not accept options.
+	f = NewBuiltinFn("f", func() {
+		t.Errorf("Function called when there are extra options")
+	})
+	callBad(theFrame, nil, RawOptions{"foo": "bar"})
+
+	// Wrong argument type.
+	f = NewBuiltinFn("f", func(x string) {
+		t.Errorf("Function called when arguments have wrong type")
+	})
+	callBad(theFrame, []interface{}{1}, theOptions)
+
+	// Wrong argument type: cannot convert to int.
+	f = NewBuiltinFn("f", func(x int) {
+		t.Errorf("Function called when arguments have wrong type")
+	})
+	callBad(theFrame, []interface{}{"x"}, theOptions)
+
+	// Wrong argument type: cannot convert to float64.
+	f = NewBuiltinFn("f", func(x float64) {
+		t.Errorf("Function called when arguments have wrong type")
+	})
+	callBad(theFrame, []interface{}{"x"}, theOptions)
 }
diff -pruN 0.11+ds1-5/eval/builtin_ns.go 0.12+ds1-1/eval/builtin_ns.go
--- 0.11+ds1-5/eval/builtin_ns.go	2018-01-18 02:03:11.000000000 +0000
+++ 0.12+ds1-1/eval/builtin_ns.go	2018-07-11 02:32:25.000000000 +0000
@@ -2,34 +2,21 @@ package eval
 
 import (
 	"strconv"
-	"strings"
 	"syscall"
 
-	"github.com/elves/elvish/eval/types"
-	"github.com/elves/elvish/eval/vartypes"
+	"github.com/elves/elvish/eval/vars"
 )
 
-func makeBuiltinNs() Ns {
-	ns := Ns{
-		"_":     vartypes.NewBlackhole(),
-		"pid":   vartypes.NewRo(types.String(strconv.Itoa(syscall.Getpid()))),
-		"ok":    vartypes.NewRo(OK),
-		"true":  vartypes.NewRo(types.Bool(true)),
-		"false": vartypes.NewRo(types.Bool(false)),
-		"paths": &EnvList{envName: "PATH"},
-		"pwd":   PwdVariable{},
-	}
-	AddBuiltinFns(ns, builtinFns...)
-	return ns
+var builtinNs = Ns{
+	"_":     vars.NewBlackhole(),
+	"pid":   vars.NewRo(strconv.Itoa(syscall.Getpid())),
+	"ok":    vars.NewRo(OK),
+	"true":  vars.NewRo(true),
+	"false": vars.NewRo(false),
+	"paths": &EnvList{envName: "PATH"},
+	"pwd":   PwdVariable{},
 }
 
-// AddBuiltinFns adds builtin functions to a namespace.
-func AddBuiltinFns(ns Ns, fns ...*BuiltinFn) {
-	for _, b := range fns {
-		name := b.Name
-		if i := strings.IndexRune(b.Name, ':'); i != -1 {
-			name = b.Name[i+1:]
-		}
-		ns[name+FnSuffix] = vartypes.NewRo(b)
-	}
+func addBuiltinFns(fns map[string]interface{}) {
+	builtinNs.AddBuiltinFns("", fns)
 }
diff -pruN 0.11+ds1-5/eval/builtin_special.go 0.12+ds1-1/eval/builtin_special.go
--- 0.11+ds1-5/eval/builtin_special.go	2018-01-18 02:03:11.000000000 +0000
+++ 0.12+ds1-1/eval/builtin_special.go	2018-07-11 02:32:25.000000000 +0000
@@ -23,17 +23,14 @@ import (
 	"path/filepath"
 	"strings"
 
-	"github.com/elves/elvish/eval/types"
-	"github.com/elves/elvish/eval/vartypes"
+	"github.com/elves/elvish/eval/vals"
+	"github.com/elves/elvish/eval/vars"
 	"github.com/elves/elvish/parse"
 )
 
-type compileBuiltin func(*compiler, *parse.Form) OpFunc
+type compileBuiltin func(*compiler, *parse.Form) OpBody
 
 var (
-	// ErrNoLibDir is thrown by "use" when the Evaler does not have a library
-	// directory.
-	ErrNoLibDir = errors.New("Evaler does not have a lib directory")
 	// ErrRelativeUseNotFromMod is thrown by "use" when relative use is used
 	// not from a module
 	ErrRelativeUseNotFromMod = errors.New("Relative use not from module")
@@ -69,7 +66,7 @@ func init() {
 const delArgMsg = "arguments to del must be variable or variable elements"
 
 // DelForm = 'del' { VariablePrimary }
-func compileDel(cp *compiler, fn *parse.Form) OpFunc {
+func compileDel(cp *compiler, fn *parse.Form) OpBody {
 	var ops []Op
 	for _, cn := range fn.Args {
 		cp.compiling(cn)
@@ -87,12 +84,12 @@ func compileDel(cp *compiler, fn *parse.
 			continue
 		}
 
-		explode, ns, name := ParseVariable(head.Value)
+		explode, ns, name := ParseVariableRef(head.Value)
 		if explode {
 			cp.errorf("arguments to del may be have a leading @")
 			continue
 		}
-		var f OpFunc
+		var f OpBody
 		if len(indicies) == 0 {
 			switch ns {
 			case "", "local":
@@ -101,9 +98,9 @@ func compileDel(cp *compiler, fn *parse.
 					continue
 				}
 				cp.thisScope().del(name)
-				f = newDelLocalVariableOp(name)
+				f = delLocalVarOp{name}
 			case "E":
-				f = newDelEnvVariableOp(name)
+				f = delEnvVarOp{name}
 			default:
 				cp.errorf("only variables in local: or E: can be deleted")
 				continue
@@ -117,63 +114,65 @@ func compileDel(cp *compiler, fn *parse.
 		}
 		ops = append(ops, Op{f, cn.Begin(), cn.End()})
 	}
-	return func(f *Frame) {
-		for _, op := range ops {
-			op.Exec(f)
-		}
-	}
+	return seqOp{ops}
+}
+
+type delLocalVarOp struct{ name string }
+
+func (op delLocalVarOp) Invoke(fm *Frame) error {
+	delete(fm.local, op.name)
+	return nil
 }
 
-func newDelLocalVariableOp(name string) OpFunc {
-	return func(f *Frame) { delete(f.local, name) }
+type delEnvVarOp struct{ name string }
+
+func (op delEnvVarOp) Invoke(*Frame) error {
+	return os.Unsetenv(op.name)
 }
 
-func newDelElementOp(ns, name string, begin, headEnd int, indexOps []ValuesOp) OpFunc {
+func newDelElementOp(ns, name string, begin, headEnd int, indexOps []ValuesOp) OpBody {
 	ends := make([]int, len(indexOps)+1)
 	ends[0] = headEnd
 	for i, op := range indexOps {
 		ends[i+1] = op.End
 	}
-	return func(f *Frame) {
-		var indicies []types.Value
-		for _, indexOp := range indexOps {
-			indexValues := indexOp.Exec(f)
-			if len(indexValues) != 1 {
-				f.errorpf(indexOp.Begin, indexOp.End, "index must evaluate to a single value in argument to del")
-			}
-			indicies = append(indicies, indexValues[0])
-		}
-		err := vartypes.DelElement(f.ResolveVar(ns, name), indicies)
-		if err != nil {
-			if level := vartypes.GetElementErrorLevel(err); level >= 0 {
-				f.errorpf(begin, ends[level], "%s", err.Error())
-			}
-			throw(err)
-		}
-	}
+	return &delElemOp{ns, name, indexOps, begin, ends}
 }
 
-func newDelEnvVariableOp(name string) OpFunc {
-	return func(*Frame) {
-		maybeThrow(os.Unsetenv(name))
-	}
+type delElemOp struct {
+	ns       string
+	name     string
+	indexOps []ValuesOp
+	begin    int
+	ends     []int
 }
 
-// makeFnOp wraps an op such that a return is converted to an ok.
-func makeFnOp(op Op) Op {
-	return Op{func(ec *Frame) {
-		err := ec.PEval(op)
-		if err != nil && err.(*Exception).Cause != Return {
-			// rethrow
-			throw(err)
+func (op *delElemOp) Invoke(fm *Frame) error {
+	var indicies []interface{}
+	for _, indexOp := range op.indexOps {
+		indexValues, err := indexOp.Exec(fm)
+		if err != nil {
+			return err
+		}
+		if len(indexValues) != 1 {
+			fm.errorpf(indexOp.Begin, indexOp.End, "index must evaluate to a single value in argument to del")
 		}
-	}, op.Begin, op.End}
+		indicies = append(indicies, indexValues[0])
+	}
+	err := vars.DelElement(fm.ResolveVar(op.ns, op.name), indicies)
+	if err != nil {
+		if level := vars.ElementErrorLevel(err); level >= 0 {
+			fm.errorpf(op.begin, op.ends[level], "%s", err.Error())
+		}
+		return err
+	}
+	return nil
 }
 
 // FnForm = 'fn' StringPrimary LambdaPrimary
 //
 // fn f []{foobar} is a shorthand for set '&'f = []{foobar}.
-func compileFn(cp *compiler, fn *parse.Form) OpFunc {
+func compileFn(cp *compiler, fn *parse.Form) OpBody {
 	args := cp.walkArgs(fn)
 	nameNode := args.next()
 	varName := mustString(cp, nameNode, "must be a literal string") + FnSuffix
@@ -183,20 +182,45 @@ func compileFn(cp *compiler, fn *parse.F
 	cp.registerVariableSetQname(":" + varName)
 	op := cp.lambda(bodyNode)
 
-	return func(ec *Frame) {
-		// Initialize the function variable with the builtin nop
-		// function. This step allows the definition of recursive
-		// functions; the actual function will never be called.
-		ec.local[varName] = vartypes.NewPtr(&BuiltinFn{"<shouldn't be called>", nop})
-		closure := op(ec)[0].(*Closure)
-		closure.Op = makeFnOp(closure.Op)
-		err := ec.local[varName].Set(closure)
-		maybeThrow(err)
+	return fnOp{varName, op}
+}
+
+type fnOp struct {
+	varName  string
+	lambdaOp ValuesOpBody
+}
+
+func (op fnOp) Invoke(fm *Frame) error {
+	// Initialize the function variable with the builtin nop function. This step
+	// allows the definition of recursive functions; the actual function will
+	// never be called.
+	fm.local[op.varName] = vars.NewAnyWithInit(NewBuiltinFn("<shouldn't be called>", nop))
+	values, err := op.lambdaOp.Invoke(fm)
+	if err != nil {
+		return err
+	}
+	closure := values[0].(*Closure)
+	closure.Op = wrapFn(closure.Op)
+	return fm.local[op.varName].Set(closure)
+}
+
+func wrapFn(op Op) Op {
+	return Op{fnWrap{op}, op.Begin, op.End}
+}
+
+type fnWrap struct{ wrapped Op }
+
+func (op fnWrap) Invoke(fm *Frame) error {
+	err := fm.Eval(op.wrapped)
+	if err != nil && err.(*Exception).Cause != Return {
+		// rethrow
+		return err
 	}
+	return nil
 }
 
 // UseForm = 'use' StringPrimary
-func compileUse(cp *compiler, fn *parse.Form) OpFunc {
+func compileUse(cp *compiler, fn *parse.Form) OpBody {
 	if len(fn.Args) == 0 {
 		end := fn.Head.End()
 		cp.errorpf(end, end, "lack module name")
@@ -211,126 +235,153 @@ func compileUse(cp *compiler, fn *parse.
 	modpath := strings.Replace(spec, ":", "/", -1)
 	cp.thisScope().set(modname + NsSuffix)
 
-	return func(ec *Frame) {
-		use(ec, modname, modpath)
-	}
+	return useOp{modname, modpath}
+}
+
+type useOp struct{ modname, modpath string }
+
+func (op useOp) Invoke(fm *Frame) error {
+	return use(fm, op.modname, op.modpath)
 }
 
-func use(ec *Frame, modname, modpath string) {
+func use(fm *Frame, modname, modpath string) error {
 	resolvedPath := ""
 	if strings.HasPrefix(modpath, "./") || strings.HasPrefix(modpath, "../") {
-		if ec.srcMeta.typ != SrcModule {
-			throw(ErrRelativeUseNotFromMod)
+		if fm.srcMeta.typ != SrcModule {
+			return ErrRelativeUseNotFromMod
 		}
 		// Resolve relative modpath.
-		resolvedPath = filepath.Clean(filepath.Dir(ec.srcMeta.name) + "/" + modpath)
+		resolvedPath = filepath.Clean(filepath.Dir(fm.srcMeta.name) + "/" + modpath)
 	} else {
 		resolvedPath = filepath.Clean(modpath)
 	}
 	if strings.HasPrefix(resolvedPath, "../") {
-		throw(ErrRelativeUseGoesOutsideLib)
+		return ErrRelativeUseGoesOutsideLib
 	}
 
 	// Put the just loaded module into local scope.
-	ec.local[modname+NsSuffix] = vartypes.NewPtr(loadModule(ec, resolvedPath))
+	ns, err := loadModule(fm, resolvedPath)
+	if err != nil {
+		return err
+	}
+	fm.local.AddNs(modname, ns)
+	return nil
 }
 
-func loadModule(ec *Frame, name string) Ns {
-	if ns, ok := ec.Evaler.modules[name]; ok {
+func loadModule(fm *Frame, name string) (Ns, error) {
+	if ns, ok := fm.Evaler.modules[name]; ok {
 		// Module already loaded.
-		return ns
+		return ns, nil
 	}
 
 	// Load the source.
-	var path, code string
-
-	if ec.libDir == "" {
-		throw(ErrNoLibDir)
+	src, err := getModuleSource(fm.Evaler, name)
+	if err != nil {
+		return nil, err
 	}
 
-	path = filepath.Join(ec.libDir, name+".elv")
-	if _, err := os.Stat(path); os.IsNotExist(err) {
-		// File does not exist. Try loading from the table of builtin
-		// modules.
-		var ok bool
-		if code, ok = ec.bundled[name]; ok {
-			// Source is loaded. Do nothing more.
-			path = "<builtin module>"
-		} else {
-			throw(fmt.Errorf("cannot load %s: %s does not exist", name, path))
-		}
-	} else {
-		// File exists. Load it.
-		code, err = readFileUTF8(path)
-		maybeThrow(err)
+	n, err := parse.Parse(name, src.code)
+	if err != nil {
+		return nil, err
 	}
 
-	n, err := parse.Parse(name, code)
-	maybeThrow(err)
-
 	// Make an empty scope to evaluate the module in.
-	meta := NewModuleSource(name, path, code)
 	modGlobal := Ns{}
 
-	newEc := &Frame{
-		ec.Evaler, meta,
+	newFm := &Frame{
+		fm.Evaler, src,
 		modGlobal, make(Ns),
-		ec.ports,
-		0, len(code), ec.addTraceback(), false,
+		fm.ports,
+		0, len(src.code), fm.addTraceback(), false,
 	}
 
-	op, err := newEc.Compile(n, meta)
-	maybeThrow(err)
+	op, err := compile(newFm.Builtin.static(), modGlobal.static(), n, src)
+	if err != nil {
+		return nil, err
+	}
 
-	// Load the namespace before executing. This avoids mutual and self use's to
-	// result in an infinite recursion.
-	ec.Evaler.modules[name] = modGlobal
-	err = newEc.PEval(op)
+	// Load the namespace before executing. This prevent circular "use"es from
+	// resulting in an infinite recursion.
+	fm.Evaler.modules[name] = modGlobal
+	err = newFm.Eval(op)
 	if err != nil {
 		// Unload the namespace.
-		delete(ec.modules, name)
-		throw(err)
+		delete(fm.modules, name)
+		return nil, err
 	}
-	return modGlobal
+	return modGlobal, nil
+}
+
+func getModuleSource(ev *Evaler, name string) (*Source, error) {
+	// First try loading from file.
+	path := filepath.Join(ev.libDir, name+".elv")
+	if ev.libDir != "" {
+		_, err := os.Stat(path)
+		if err == nil {
+			code, err := readFileUTF8(path)
+			if err != nil {
+				return nil, err
+			}
+			return NewModuleSource(name, path, code), nil
+		} else if !os.IsNotExist(err) {
+			return nil, err
+		}
+	}
+
+	// Try loading bundled module.
+	if code, ok := ev.bundled[name]; ok {
+		return NewModuleSource(name, "", code), nil
+	}
+
+	return nil, fmt.Errorf("cannot load %s: %s does not exist", name, path)
 }
 
 // compileAnd compiles the "and" special form.
+//
 // The and special form evaluates arguments until a false-ish values is found
 // and outputs it; the remaining arguments are not evaluated. If there are no
 // false-ish values, the last value is output. If there are no arguments, it
 // outputs $true, as if there is a hidden $true before actual arguments.
-func compileAnd(cp *compiler, fn *parse.Form) OpFunc {
-	return compileAndOr(cp, fn, true, false)
+func compileAnd(cp *compiler, fn *parse.Form) OpBody {
+	return &andOrOp{cp.compoundOps(fn.Args), true, false}
 }
 
 // compileOr compiles the "or" special form.
+//
 // The or special form evaluates arguments until a true-ish values is found and
 // outputs it; the remaining arguments are not evaluated. If there are no
 // true-ish values, the last value is output. If there are no arguments, it
 // outputs $false, as if there is a hidden $false before actual arguments.
-func compileOr(cp *compiler, fn *parse.Form) OpFunc {
-	return compileAndOr(cp, fn, false, true)
+func compileOr(cp *compiler, fn *parse.Form) OpBody {
+	return &andOrOp{cp.compoundOps(fn.Args), false, true}
 }
 
-func compileAndOr(cp *compiler, fn *parse.Form, init, stopAt bool) OpFunc {
-	argOps := cp.compoundOps(fn.Args)
-	return func(ec *Frame) {
-		var lastValue types.Value = types.Bool(init)
-		for _, op := range argOps {
-			values := op.Exec(ec)
-			for _, value := range values {
-				if types.ToBool(value) == stopAt {
-					ec.OutputChan() <- value
-					return
-				}
-				lastValue = value
+type andOrOp struct {
+	argOps []ValuesOp
+	init   bool
+	stopAt bool
+}
+
+func (op *andOrOp) Invoke(fm *Frame) error {
+	var lastValue interface{} = vals.Bool(op.init)
+	for _, argOp := range op.argOps {
+		values, err := argOp.Exec(fm)
+		if err != nil {
+			return err
+		}
+		for _, value := range values {
+			if vals.Bool(value) == op.stopAt {
+				fm.OutputChan() <- value
+				return nil
 			}
+			lastValue = value
 		}
-		ec.OutputChan() <- lastValue
 	}
+	fm.OutputChan() <- lastValue
+	return nil
 }
 
-func compileIf(cp *compiler, fn *parse.Form) OpFunc {
+func compileIf(cp *compiler, fn *parse.Form) OpBody {
 	args := cp.walkArgs(fn)
 	var condNodes []*parse.Compound
 	var bodyNodes []*parse.Primary
@@ -351,57 +402,76 @@ func compileIf(cp *compiler, fn *parse.F
 		elseOp = cp.primaryOp(elseNode)
 	}
 
-	return func(ec *Frame) {
-		bodies := make([]Callable, len(bodyOps))
-		for i, bodyOp := range bodyOps {
-			bodies[i] = bodyOp.execlambdaOp(ec)
-		}
-		else_ := elseOp.execlambdaOp(ec)
-		for i, condOp := range condOps {
-			if allTrue(condOp.Exec(ec.fork("if cond"))) {
-				bodies[i].Call(ec.fork("if body"), NoArgs, NoOpts)
-				return
-			}
+	return &ifOp{condOps, bodyOps, elseOp}
+}
+
+type ifOp struct {
+	condOps []ValuesOp
+	bodyOps []ValuesOp
+	elseOp  ValuesOp
+}
+
+func (op *ifOp) Invoke(fm *Frame) error {
+	bodies := make([]Callable, len(op.bodyOps))
+	for i, bodyOp := range op.bodyOps {
+		bodies[i] = bodyOp.execlambdaOp(fm)
+	}
+	else_ := op.elseOp.execlambdaOp(fm)
+	for i, condOp := range op.condOps {
+		condValues, err := condOp.Exec(fm.fork("if cond"))
+		if err != nil {
+			return err
 		}
-		if elseOp.Func != nil {
-			else_.Call(ec.fork("if else"), NoArgs, NoOpts)
+		if allTrue(condValues) {
+			return bodies[i].Call(fm.fork("if body"), NoArgs, NoOpts)
 		}
 	}
+	if op.elseOp.Body != nil {
+		return else_.Call(fm.fork("if else"), NoArgs, NoOpts)
+	}
+	return nil
 }
 
-func compileWhile(cp *compiler, fn *parse.Form) OpFunc {
+func compileWhile(cp *compiler, fn *parse.Form) OpBody {
 	args := cp.walkArgs(fn)
 	condNode := args.next()
 	bodyNode := args.nextMustLambda()
 	args.mustEnd()
 
-	condOp := cp.compoundOp(condNode)
-	bodyOp := cp.primaryOp(bodyNode)
+	return &whileOp{cp.compoundOp(condNode), cp.primaryOp(bodyNode)}
+}
+
+type whileOp struct {
+	condOp, bodyOp ValuesOp
+}
 
-	return func(ec *Frame) {
-		body := bodyOp.execlambdaOp(ec)
+func (op *whileOp) Invoke(fm *Frame) error {
+	body := op.bodyOp.execlambdaOp(fm)
 
-		for {
-			cond := condOp.Exec(ec.fork("while cond"))
-			if !allTrue(cond) {
+	for {
+		condValues, err := op.condOp.Exec(fm.fork("while cond"))
+		if err != nil {
+			return err
+		}
+		if !allTrue(condValues) {
+			break
+		}
+		err = fm.fork("while").Call(body, NoArgs, NoOpts)
+		if err != nil {
+			exc := err.(*Exception)
+			if exc.Cause == Continue {
+				// do nothing
+			} else if exc.Cause == Break {
 				break
-			}
-			err := ec.fork("while").PCall(body, NoArgs, NoOpts)
-			if err != nil {
-				exc := err.(*Exception)
-				if exc.Cause == Continue {
-					// do nothing
-				} else if exc.Cause == Break {
-					continue
-				} else {
-					throw(err)
-				}
+			} else {
+				return err
 			}
 		}
 	}
+	return nil
 }
 
-func compileFor(cp *compiler, fn *parse.Form) OpFunc {
+func compileFor(cp *compiler, fn *parse.Form) OpBody {
 	args := cp.walkArgs(fn)
 	varNode := args.next()
 	iterNode := args.next()
@@ -410,7 +480,7 @@ func compileFor(cp *compiler, fn *parse.
 	args.mustEnd()
 
 	varOp, restOp := cp.lvaluesOp(varNode.Indexings[0])
-	if restOp.Func != nil {
+	if restOp.Body != nil {
 		cp.errorpf(restOp.Begin, restOp.End, "rest not allowed")
 	}
 
@@ -421,44 +491,67 @@ func compileFor(cp *compiler, fn *parse.
 		elseOp = cp.primaryOp(elseNode)
 	}
 
-	return func(ec *Frame) {
-		variables := varOp.Exec(ec)
-		if len(variables) != 1 {
-			ec.errorpf(varOp.Begin, varOp.End, "only one variable allowed")
-		}
-		variable := variables[0]
-
-		iterable := ec.ExecAndUnwrap("value being iterated", iterOp).One().Iterable()
-
-		body := bodyOp.execlambdaOp(ec)
-		elseBody := elseOp.execlambdaOp(ec)
-
-		iterated := false
-		iterable.Iterate(func(v types.Value) bool {
-			iterated = true
-			err := variable.Set(v)
-			maybeThrow(err)
-			err = ec.fork("for").PCall(body, NoArgs, NoOpts)
-			if err != nil {
-				exc := err.(*Exception)
-				if exc.Cause == Continue {
-					// do nothing
-				} else if exc.Cause == Break {
-					return false
-				} else {
-					throw(err)
-				}
-			}
-			return true
-		})
+	return &forOp{varOp, iterOp, bodyOp, elseOp}
+}
 
-		if !iterated && elseBody != nil {
-			elseBody.Call(ec.fork("for else"), NoArgs, NoOpts)
+type forOp struct {
+	varOp  LValuesOp
+	iterOp ValuesOp
+	bodyOp ValuesOp
+	elseOp ValuesOp
+}
+
+func (op *forOp) Invoke(fm *Frame) error {
+	variables, err := op.varOp.Exec(fm)
+	if err != nil {
+		return err
+	}
+	if len(variables) != 1 {
+		fm.errorpf(op.varOp.Begin, op.varOp.End, "only one variable allowed")
+	}
+	variable := variables[0]
+	iterable := fm.ExecAndUnwrap("value being iterated", op.iterOp).One().Any()
+
+	body := op.bodyOp.execlambdaOp(fm)
+	elseBody := op.elseOp.execlambdaOp(fm)
+
+	iterated := false
+	var errElement error
+	errIterate := vals.Iterate(iterable, func(v interface{}) bool {
+		iterated = true
+		err := variable.Set(v)
+		if err != nil {
+			errElement = err
+			return false
 		}
+		err = fm.fork("for").Call(body, NoArgs, NoOpts)
+		if err != nil {
+			exc := err.(*Exception)
+			if exc.Cause == Continue {
+				// do nothing
+			} else if exc.Cause == Break {
+				return false
+			} else {
+				errElement = err
+				return false
+			}
+		}
+		return true
+	})
+	if errIterate != nil {
+		return errIterate
+	}
+	if errElement != nil {
+		return errElement
+	}
+
+	if !iterated && elseBody != nil {
+		return elseBody.Call(fm.fork("for else"), NoArgs, NoOpts)
 	}
+	return nil
 }
 
-func compileTry(cp *compiler, fn *parse.Form) OpFunc {
+func compileTry(cp *compiler, fn *parse.Form) OpBody {
 	logger.Println("compiling try")
 	args := cp.walkArgs(fn)
 	bodyNode := args.nextMustLambda()
@@ -485,7 +578,7 @@ func compileTry(cp *compiler, fn *parse.
 	if exceptVarNode != nil {
 		var restOp LValuesOp
 		exceptVarOp, restOp = cp.lvaluesOp(exceptVarNode)
-		if restOp.Func != nil {
+		if restOp.Body != nil {
 			cp.errorpf(restOp.Begin, restOp.End, "may not use @rest in except variable")
 		}
 	}
@@ -499,56 +592,76 @@ func compileTry(cp *compiler, fn *parse.
 		finallyOp = cp.primaryOp(finallyNode)
 	}
 
-	return func(ec *Frame) {
-		body := bodyOp.execlambdaOp(ec)
-		exceptVar := exceptVarOp.execMustOne(ec)
-		except := exceptOp.execlambdaOp(ec)
-		else_ := elseOp.execlambdaOp(ec)
-		finally := finallyOp.execlambdaOp(ec)
+	return &tryOp{bodyOp, exceptVarOp, exceptOp, elseOp, finallyOp}
+}
 
-		err := ec.fork("try body").PCall(body, NoArgs, NoOpts)
-		if err != nil {
-			if except != nil {
-				if exceptVar != nil {
-					err := exceptVar.Set(err.(*Exception))
-					maybeThrow(err)
+type tryOp struct {
+	bodyOp      ValuesOp
+	exceptVarOp LValuesOp
+	exceptOp    ValuesOp
+	elseOp      ValuesOp
+	finallyOp   ValuesOp
+}
+
+func (op *tryOp) Invoke(fm *Frame) error {
+	body := op.bodyOp.execlambdaOp(fm)
+	exceptVar := op.exceptVarOp.execMustOne(fm)
+	except := op.exceptOp.execlambdaOp(fm)
+	else_ := op.elseOp.execlambdaOp(fm)
+	finally := op.finallyOp.execlambdaOp(fm)
+
+	err := fm.fork("try body").Call(body, NoArgs, NoOpts)
+	if err != nil {
+		if except != nil {
+			if exceptVar != nil {
+				err := exceptVar.Set(err.(*Exception))
+				if err != nil {
+					return err
 				}
-				err = ec.fork("try except").PCall(except, NoArgs, NoOpts)
-			}
-		} else {
-			if else_ != nil {
-				err = ec.fork("try else").PCall(else_, NoArgs, NoOpts)
 			}
+			err = fm.fork("try except").Call(except, NoArgs, NoOpts)
 		}
-		if finally != nil {
-			finally.Call(ec.fork("try finally"), NoArgs, NoOpts)
+	} else {
+		if else_ != nil {
+			err = fm.fork("try else").Call(else_, NoArgs, NoOpts)
 		}
-		if err != nil {
-			throw(err)
+	}
+	if finally != nil {
+		errFinally := finally.Call(fm.fork("try finally"), NoArgs, NoOpts)
+		if errFinally != nil {
+			// TODO: If err is not nil, this discards err. Use something similar
+			// to pipeline exception to expose both.
+			return errFinally
 		}
 	}
+	return err
 }
 
 // execLambdaOp executes a ValuesOp that is known to yield a lambda and returns
 // the lambda. If the ValuesOp is empty, it returns a nil.
-func (op ValuesOp) execlambdaOp(ec *Frame) Callable {
-	if op.Func == nil {
+func (op ValuesOp) execlambdaOp(fm *Frame) Callable {
+	if op.Body == nil {
 		return nil
 	}
 
-	return op.Exec(ec)[0].(Callable)
+	values, err := op.Exec(fm)
+	if err != nil {
+		panic("must not be erroneous")
+	}
+	return values[0].(Callable)
 }
 
 // execMustOne executes the LValuesOp and raises an exception if it does not
 // evaluate to exactly one Variable. If the given LValuesOp is empty, it returns
 // nil.
-func (op LValuesOp) execMustOne(ec *Frame) vartypes.Variable {
-	if op.Func == nil {
+func (op LValuesOp) execMustOne(fm *Frame) vars.Var {
+	if op.Body == nil {
 		return nil
 	}
-	variables := op.Exec(ec)
+	variables, err := op.Exec(fm)
+	maybeThrow(err)
 	if len(variables) != 1 {
-		ec.errorpf(op.Begin, op.End, "should be one variable")
+		fm.errorpf(op.Begin, op.End, "should be one variable")
 	}
 	return variables[0]
 }
diff -pruN 0.11+ds1-5/eval/builtin_special_test.go 0.12+ds1-1/eval/builtin_special_test.go
--- 0.11+ds1-5/eval/builtin_special_test.go	2018-01-18 02:03:11.000000000 +0000
+++ 0.12+ds1-1/eval/builtin_special_test.go	2018-07-11 02:32:25.000000000 +0000
@@ -1,73 +1,109 @@
 package eval
 
-import "testing"
+import (
+	"errors"
+	"path/filepath"
+	"strings"
+	"testing"
 
-var builtinSpecialTests = []Test{
-	// del
-	NewTest("x = [&k=v &k2=v2]; del x[k2]; keys $x").WantOutStrings("k"),
-	NewTest("x = [[&k=v &k2=v2]]; del x[0][k2]; keys $x[0]").WantOutStrings("k"),
-
-	// if
-	{"if true { put then }", want{out: strs("then")}},
-	{"if $false { put then } else { put else }", want{out: strs("else")}},
-	{"if $false { put 1 } elif $false { put 2 } else { put 3 }",
-		want{out: strs("3")}},
-	{"if $false { put 2 } elif true { put 2 } else { put 3 }",
-		want{out: strs("2")}},
-
-	// try
-	{"try { nop } except { put bad } else { put good }", want{out: strs("good")}},
-	{"try { e:false } except - { put bad } else { put good }", want{out: strs("bad")}},
-
-	// while
-	{"x=0; while (< $x 4) { put $x; x=(+ $x 1) }",
-		want{out: strs("0", "1", "2", "3")}},
-
-	// for
-	{"for x [tempora mores] { put 'O '$x }",
-		want{out: strs("O tempora", "O mores")}},
-	// break
-	{"for x [a] { break } else { put $x }", wantNothing},
-	// else
-	{"for x [a] { put $x } else { put $x }", want{out: strs("a")}},
-	// continue
-	{"for x [a b] { put $x; continue; put $x; }", want{out: strs("a", "b")}},
-
-	// fn.
-	{"fn f [x]{ put x=$x'.' }; f lorem; f ipsum",
-		want{out: strs("x=lorem.", "x=ipsum.")}},
-	// return.
-	{"fn f []{ put a; return; put b }; f", want{out: strs("a")}},
-
-	// Modules (see setup_datadir_test.go for setup)
-	// "use" imports a module.
-	{`use lorem; put $lorem:name`, want{out: strs("lorem")}},
-	// imports are lexically scoped
-	// TODO: Support testing for compilation error
-	// {`{ use lorem }; put $lorem:name`, want{err: errAny}},
-
-	// use of imported variable is captured in upvalue
-	{`({ use lorem; put { { put $lorem:name } } })`, want{out: strs("lorem")}},
-	// use of imported function is also captured in upvalue
-	{`{ use lorem; { lorem:put-name } }`, want{out: strs("lorem")}},
-
-	// multi-level module names
-	{`use a:b:c:d; put $a:b:c:d:name`, want{out: strs("a/b/c/d")}},
-	// shortening module names by using slashes for some path prefix
-	{`use a:b/c:d; put $c:d:name`, want{out: strs("a/b/c/d")}},
-	// importing the same module under different names
-	{`use a/b/c/d; use a/b/c:d; eq $d:name $c:d:name`, wantTrue},
-	// module is cached after first use
-	{`use has/init; use has:init`, want{out: strs("has/init")}},
-	// overriding module
-	{`use d; put $d:name; use a/b/c/d; put $d:name`,
-		want{out: strs("d", "a/b/c/d")}},
-	// relative uses
-	{`use a/b/c/x; put $x:d $x:lorem`, want{out: strs("a/b/c/d", "lorem")}},
+	"github.com/elves/elvish/util"
+)
 
-	// TODO: Test module namespace
+func TestBuiltinSpecial(t *testing.T) {
+	Test(t,
+		// del
+		That("x = [&k=v &k2=v2]; del x[k2]; keys $x").Puts("k"),
+		That("x = [[&k=v &k2=v2]]; del x[0][k2]; keys $x[0]").Puts("k"),
+
+		// if
+		That("if true { put then }").Puts("then"),
+		That("if $false { put then } else { put else }").Puts("else"),
+		That("if $false { put 1 } elif $false { put 2 } else { put 3 }").
+			Puts("3"),
+		That("if $false { put 2 } elif true { put 2 } else { put 3 }").Puts("2"),
+
+		// try
+		That("try { nop } except { put bad } else { put good }").Puts("good"),
+		That("try { e:false } except - { put bad } else { put good }").
+			Puts("bad"),
+		That("try { fail tr }").ErrorsWith(errors.New("tr")),
+		That("try { fail tr } finally { put final }").
+			Puts("final").ErrorsWith(errors.New("tr")),
+		That("try { fail tr } except { fail ex } finally { put final }").
+			Puts("final").ErrorsWith(errors.New("ex")),
+		That("try { fail tr } except { put ex } finally { fail final }").
+			Puts("ex").ErrorsWith(errors.New("final")),
+		That("try { fail tr } except { fail ex } finally { fail final }").
+			ErrorsWith(errors.New("final")),
+
+		// while
+		That("x=0; while (< $x 4) { put $x; x=(+ $x 1) }").
+			Puts("0", "1", "2", "3"),
+		That("x = 0; while (< $x 4) { put $x; break }").Puts("0"),
+		That("x = 0; while (< $x 4) { fail haha }").Errors(),
+
+		// for
+		That("for x [tempora mores] { put 'O '$x }").
+			Puts("O tempora", "O mores"),
+		// break
+		That("for x [a] { break } else { put $x }").DoesNothing(),
+		// else
+		That("for x [a] { put $x } else { put $x }").Puts("a"),
+		// continue
+		That("for x [a b] { put $x; continue; put $x; }").Puts("a", "b"),
+
+		// fn.
+		That("fn f [x]{ put x=$x'.' }; f lorem; f ipsum").
+			Puts("x=lorem.", "x=ipsum."),
+		// return.
+		That("fn f []{ put a; return; put b }; f").Puts("a"),
+	)
 }
 
-func TestBuiltinSpecial(t *testing.T) {
-	runTests(t, builtinSpecialTests)
+var useTests = []TestCase{}
+
+func TestUse(t *testing.T) {
+	util.InTempDir(func(libdir string) {
+		MustMkdirAll(filepath.Join("a", "b", "c"), 0700)
+
+		writeMod := func(name, content string) {
+			fname := filepath.Join(strings.Split(name, "/")...) + ".elv"
+			MustWriteFile(fname, []byte(content), 0600)
+		}
+		writeMod("has-init", "put has-init")
+		writeMod("put-x", "put $x")
+		writeMod("lorem", "name = lorem; fn put-name { put $name }")
+		writeMod("d", "name = d")
+		writeMod("a/b/c/d", "name = a/b/c/d")
+		writeMod("a/b/c/x",
+			"use ./d; d = $d:name; use ../../../lorem; lorem = $lorem:name")
+
+		TestWithSetup(t, func(ev *Evaler) { ev.SetLibDir(libdir) },
+			That(`use lorem; put $lorem:name`).Puts("lorem"),
+			// imports are lexically scoped
+			// TODO: Support testing for compilation error
+			// That(`{ use lorem }; put $lorem:name`).ErrorsAny(),
+
+			// use of imported variable is captured in upvalue
+			That(`({ use lorem; put { { put $lorem:name } } })`).Puts("lorem"),
+			// use of imported function is also captured in upvalue
+			That(`{ use lorem; { lorem:put-name } }`).Puts("lorem"),
+
+			// use of a nested module
+			That(`use a:b/c/d; put $d:name`).Puts("a/b/c/d"),
+			// module is cached after first use
+			That(`use has-init; use has-init`).Puts("has-init"),
+			// overriding module
+			That(`use d; put $d:name; use a/b/c/d; put $d:name`).Puts(
+				"d", "a/b/c/d"),
+			// relative uses
+			That(`use a/b/c/x; put $x:d $x:lorem`).Puts("a/b/c/d", "lorem"),
+
+			// Variables defined in the default global scope is invisible from
+			// modules
+			That("x = foo; use put-x").Errors(),
+
+			// TODO: Test module namespace
+		)
+	})
 }
diff -pruN 0.11+ds1-5/eval/bundled/binding.elv.go 0.12+ds1-1/eval/bundled/binding.elv.go
--- 0.11+ds1-5/eval/bundled/binding.elv.go	2018-01-18 02:03:11.000000000 +0000
+++ 0.12+ds1-1/eval/bundled/binding.elv.go	2018-07-11 02:32:25.000000000 +0000
@@ -54,7 +54,6 @@ fn install {
         &Default=  $edit:history:default~
         &Up=       $edit:history:up~
         &Down=     $edit:history:down-or-quit~
-        &Ctrl-R=   $edit:history:switch-to-histlist~
         &'Ctrl-['= $edit:insert:start~
     ])
 
@@ -93,7 +92,7 @@ fn install {
     edit:location:binding = (edit:binding-table [&])
 
     edit:lastcmd:binding = (edit:binding-table [
-        &Default= $edit:lastcmd:alt-default~
+        &Alt-1= $edit:lastcmd:accept-line~
     ])
 
     edit:navigation:binding = (edit:binding-table [
@@ -116,3 +115,5 @@ fn install {
     edit:narrow:binding = (edit:binding-table [&])
 }
 `
+
+// vi: se ft=elvish et:
diff -pruN 0.11+ds1-5/eval/bundled/epm.elv.go 0.12+ds1-1/eval/bundled/epm.elv.go
--- 0.11+ds1-5/eval/bundled/epm.elv.go	2018-01-18 02:03:11.000000000 +0000
+++ 0.12+ds1-1/eval/bundled/epm.elv.go	2018-07-11 02:32:25.000000000 +0000
@@ -1,111 +1,353 @@
 package bundled
 
-const epmElv = `
+const epmElv = `use re
+
+# Verbosity configuration
+debug-mode = $false
+
+# Configuration for common domains
+-default-domain-config = [
+  &"github.com"= [
+    &method= git
+    &protocol= https
+    &levels= 2
+  ]
+  &"bitbucket.org"= [
+    &method= git
+    &protocol= https
+    &levels= 2
+  ]
+  &"gitlab.com"= [
+    &method= git
+    &protocol= https
+    &levels= 2
+  ]
+]
+
+# Internal configuration
 -data-dir = ~/.elvish
--installed = $-data-dir/epm-installed
+-lib-dir = $-data-dir/lib
 
-fn -info [text]{
-    print (edit:styled '=> ' green)
+# General utility functions
+
+fn -debug [text]{
+  if $debug-mode {
+    print (styled '=> ' blue)
     echo $text
+  }
+}
+
+fn -info [text]{
+  print (styled '=> ' green)
+  echo $text
+}
+
+fn -warn [text]{
+  print (styled '=> ' yellow)
+  echo $text
 }
 
 fn -error [text]{
-    print (edit:styled '=> ' red)
-    echo $text
+  print (styled '=> ' red)
+  echo $text
 }
 
-fn add-installed [@pkgs]{
-    if (eq $pkgs []) {
-        -error 'Must specify at least one package.'
-        return
-    }
-    for pkg $pkgs {
-        echo $pkg >> $-installed
-    }
+fn dest [pkg]{
+  put $-lib-dir/$pkg
 }
 
-fn -get-url [pkg]{
-    put https://$pkg
+fn is-installed [pkg]{
+  bool ?(test -e (dest $pkg))
 }
 
-fn -install-one [pkg]{
-    dest = $-data-dir/lib/$pkg
-    if ?(test -e $dest) {
-        -error 'Package '$pkg' already exists locally.'
-        return
-    }
-    -info 'Installing '$pkg
-    mkdir -p $dest
-    git clone (-get-url $pkg) $dest
-    add-installed $pkg
+fn -package-domain [pkg]{
+  splits &max=2 / $pkg | take 1
 }
 
-fn install [@pkgs]{
-    if (eq $pkgs []) {
-        -error 'Must specify at least one package.'
-        return
-    }
-    for pkg $pkgs {
-        -install-one $pkg
-    }
+fn -package-without-domain [pkg]{
+  splits &max=2 / $pkg | drop 1 | joins ''
 }
 
-fn installed {
-    if ?(test -f $-installed) {
-        cat $-installed
-    }
+# Merge two maps
+fn -merge [a b]{
+  keys $b | each [k]{ a[$k] = $b[$k] }
+  put $a
 }
 
-fn -upgrade-one [pkg]{
-    dest = $-data-dir/lib/$pkg
-    if (not ?(test -d $dest)) {
-        -error 'Package '$pkg' not installed locally.'
-        return
-    }
-    -info 'Upgrading package '$pkg
-    git -C $dest pull
+# Uppercase first letter of a string
+fn -first-upper [s]{
+  put (echo $s[0] | tr '[:lower:]' '[:upper:]')$s[(count $s[0]):]
 }
 
-fn upgrade [@pkgs]{
-    if (eq $pkgs []) {
-        pkgs = [(installed)]
-        -info 'Upgrading all installed packages'
+# Expand tilde at the beginning of a string to the home dir
+fn -tilde-expand [p]{
+  re:replace "^~" $E:HOME $p
+}
+
+# Known method handlers. Each entry is indexed by method name (the
+# value of the "method" key in the domain configs), and must contain
+# two keys: install and upgrade, each one must be a closure that
+# received two arguments: package name and the domain config entry
+#
+# - Method 'git' requires the key 'protocol' in the domain config,
+#   which has to be 'http' or 'https'
+# - Method 'rsync' requires the key 'location' in the domain config,
+#   which has to contain the directory where the domain files are
+#   stored. It can be any source location understood by the rsync
+#   command.
+-method-handler = [
+  &git= [
+    &src= [pkg dom-cfg]{
+      put $dom-cfg[protocol]"://"$pkg
     }
-    for pkg $pkgs {
-        -upgrade-one $pkg
+
+    &install= [pkg dom-cfg]{
+      dest = (dest $pkg)
+      -info "Installing "$pkg
+      mkdir -p $dest
+      git clone ($-method-handler[git][src] $pkg $dom-cfg) $dest
     }
-}
 
-fn -uninstall-one [pkg]{
-    installed-pkgs = [(installed)]
-    if (not (has-value $installed-pkgs $pkg)) {
-        -error 'Package '$pkg' is not registered as installed.'
-        return
+    &upgrade= [pkg dom-cfg]{
+      dest = (dest $pkg)
+      -info "Updating "$pkg
+      try {
+        git -C $dest pull
+      } except _ {
+          -error "Something failed, please check error above and retry."
+      }
+    }
+  ]
+
+  &rsync= [
+    &src= [pkg dom-cfg]{
+      put (-tilde-expand $dom-cfg[location])/(-package-without-domain $pkg)/
+    }
+
+    &install= [pkg dom-cfg]{
+      dest = (dest $pkg)
+      pkgd = (-package-without-domain $pkg)
+      -info "Installing "$pkg
+      rsync -av ($-method-handler[rsync][src] $pkg $dom-cfg) $dest
     }
-    dest = $-data-dir/lib/$pkg
-    if (not ?(test -d $dest)) {
-        -error 'Package '$pkg' does not exist locally.'
+
+    &upgrade= [pkg dom-cfg]{
+      dest = (dest $pkg)
+      pkgd = (-package-without-domain $pkg)
+      if (not (is-installed $pkg)) {
+        -error "Package "$pkg" is not installed."
         return
+      }
+      -info "Updating "$pkg
+      rsync -av ($-method-handler[rsync][src] $pkg $dom-cfg) $dest
+    }
+  ]
+]
+
+# Return the filename of the domain config file for the given domain
+# (regardless of whether it exists)
+fn -domain-config-file [dom]{
+  put $-lib-dir/$dom/epm-domain.cfg
+}
+
+# Return the filename of the metadata file for the given package
+# (regardless of whether it exists)
+fn -package-metadata-file [pkg]{
+  put (dest $pkg)/metadata.json
+}
+
+fn -write-domain-config [dom]{
+  cfgfile = (-domain-config-file $dom)
+  mkdir -p (dirname $cfgfile)
+  if (has-key $-default-domain-config $dom) {
+    put $-default-domain-config[$dom] | to-json > $cfgfile
+  } else {
+    -error "No default config exists for domain "$dom"."
+  }
+}
+
+# Returns the domain config for a given domain, as a JSON data
+# structure. If the file does not exist but we have a built-in
+# definition, then we return the default. Otherwise we return $false,
+# so the result can always be checked with 'if'.
+fn -domain-config [dom]{
+  cfgfile = (-domain-config-file $dom)
+  cfg = $false
+  if ?(test -f $cfgfile) {
+    # If the config file exists, read it...
+    cfg = (cat $cfgfile | from-json)
+    -debug "Read domain config for "$dom": "(to-string $cfg)
+  } else {
+    # ...otherwise check if we have a default config for the domain, and save it
+    if (has-key $-default-domain-config $dom) {
+      cfg = $-default-domain-config[$dom]
+      -debug "No existing config for "$dom", using the default: "(to-string $cfg)
+    } else {
+      -debug "No existing config for "$dom" and no default available."
+    }
+  }
+  put $cfg
+}
+
+
+# Return the method by which a package is installed
+fn -package-method [pkg]{
+  dom = (-package-domain $pkg)
+  cfg = (-domain-config $dom)
+  if $cfg {
+    put $cfg[method]
+  } else {
+    put $false
+  }
+}
+
+# Invoke package operations defined in $-method-handler above
+fn -package-op [pkg what]{
+  dom = (-package-domain $pkg)
+  cfg = (-domain-config $dom)
+  if $cfg {
+    method = $cfg[method]
+    if (has-key $-method-handler $method) {
+      if (has-key $-method-handler[$method] $what) {
+        $-method-handler[$method][$what] $pkg $cfg
+      } else {
+        fail "Unknown operation '"$what"' for package "$pkg
+      }
+    } else {
+      fail "Unknown method '"$method"', specified in in config file "(-domain-config-file $dom)
+    }
+  } else {
+    -error "No config for domain '"$dom"'."
+  }
+}
+
+# Uninstall a single package by removing its directory
+fn -uninstall-package [pkg]{
+  if (not (is-installed $pkg)) {
+    -error "Package "$pkg" is not installed."
+    return
+  }
+  dest = (dest $pkg)
+  -info "Removing package "$pkg
+  rm -rf $dest
+}
+
+######################################################################
+# Main user-facing functions
+
+# Read and parse the package metadata, if it exists
+fn metadata [pkg]{
+  # Base metadata attributes
+  res = [
+    &name= $pkg
+    &method= (-package-method $pkg)
+    &src= (-package-op $pkg src)
+    &dst= (dest $pkg)
+    &installed= (is-installed $pkg)
+  ]
+  # Merge with package-specified attributes, if any
+  file = (-package-metadata-file $pkg)
+  if (and (is-installed $pkg) ?(test -f $file)) {
+    res = (-merge (cat $file | from-json) $res)
+  }
+  put $res
+}
+
+# Print out information about a package
+fn query [pkg]{
+  data = (metadata $pkg)
+  special-keys = [name method installed src dst]
+  echo (styled "Package "$data[name] cyan)
+  if $data[installed] {
+    echo (styled "Installed at "$data[dst] green)
+  } else {
+    echo (styled "Not installed" red)
+  }
+  echo (styled "Source:" blue) $data[method] $data[src]
+  keys $data | each [key]{
+    if (not (has-value $special-keys $key)) {
+      val = $data[$key]
+      if (eq (kind-of $val) list) {
+        val = (joins ", " $val)
+      }
+      echo (styled (-first-upper $key)":" blue) $val
     }
-    -info 'Removing package '$pkg
-    rm -rf $dest
-    # issue #486
-    {
-        for installed $installed-pkgs {
-            if (not-eq $installed $pkg) {
-                echo $installed
-            }
-        }
-    } > $-installed
+  }
 }
 
-fn uninstall [@pkgs]{
-    if (eq $pkgs []) {
-        -error 'Must specify at least one package.'
-        return
+# List installed packages
+fn installed {
+  put $-lib-dir/*[nomatch-ok] | each [dir]{
+    dom = (re:replace &literal $-lib-dir/ '' $dir)
+    cfg = (-domain-config $dom)
+    # Only list domains for which we know the config, so that the user
+    # can have his own non-package directories under ~/.elvish/lib
+    # without conflicts.
+    if $cfg {
+      lvl = $cfg[levels]
+      pat = '^\Q'$-lib-dir'/\E('(repeat (+ $lvl 1) '[^/]+' | joins '/')')/$'
+      put (each [d]{ re:find $pat $d } [ $-lib-dir/$dom/**[nomatch-ok]/ ] )[groups][1][text]
+    }
+  }
+}
+
+# epm:list is an alias for epm:installed
+fn list { installed }
+
+# Install and upgrade are method-specific, so we call the
+# corresponding functions using -package-op
+fn install [&silent-if-installed=$false @pkgs]{
+  if (eq $pkgs []) {
+    -error "You must specify at least one package."
+    return
+  }
+  for pkg $pkgs {
+    if (is-installed $pkg) {
+      if (not $silent-if-installed) {
+        -info "Package "$pkg" is already installed."
+      }
+    } else {
+      -package-op $pkg install
+      # Check if there are any dependencies to install
+      metadata = (metadata $pkg)
+      if (has-key $metadata dependencies) {
+        deps = $metadata[dependencies]
+        -info "Installing dependencies: "(joins " " $deps)
+        # If the installation of dependencies fails, uninstall the
+        # target package (leave any already-installed dependencies in
+        # place)
+        try {
+          install $@deps
+        } except e {
+          -error "Dependency installation failed. Uninstalling "$pkg", please check the errors above and try again."
+          -uninstall-package $pkg
+        }
+      }
     }
-    for pkg $pkgs {
-        -uninstall-one $pkg
+  }
+}
+
+fn upgrade [@pkgs]{
+  if (eq $pkgs []) {
+    pkgs = [(installed)]
+    -info 'Upgrading all installed packages'
+  }
+  for pkg $pkgs {
+    if (not (is-installed $pkg)) {
+      -error "Package "$pkg" is not installed."
+    } else {
+      -package-op $pkg upgrade
     }
+  }
 }
-`
+
+# Uninstall is the same for everyone, just remove the directory
+fn uninstall [@pkgs]{
+  if (eq $pkgs []) {
+    -error 'You must specify at least one package.'
+    return
+  }
+  for pkg $pkgs {
+    -uninstall-package $pkg
+  }
+}`
diff -pruN 0.11+ds1-5/eval/bundled/narrow.elv.go 0.12+ds1-1/eval/bundled/narrow.elv.go
--- 0.11+ds1-5/eval/bundled/narrow.elv.go	2018-01-18 02:03:11.000000000 +0000
+++ 0.12+ds1-1/eval/bundled/narrow.elv.go	2018-07-11 02:32:25.000000000 +0000
@@ -15,7 +15,7 @@ const narrowElv = `
 # but different keys can be specified with the options. To disable a
 # binding, specify its key as "".
 # Example:
-#   narrow:bind-trigger-keys &location=Alt-l &lastcmd="" 
+#   narrow:bind-trigger-keys &location=Alt-l &lastcmd=""
 
 # Hooks
 # Each hook variable is an array which must contain lambdas, all of
@@ -36,7 +36,7 @@ fn location {
 	      score = (splits . $arg[score] | take 1)
         put [
           &content=$arg[path]
-          &display=$score" "$arg[path]
+          &display=$score" "(tilde-abbr $arg[path])
 	      ]
   })]
 
@@ -53,7 +53,7 @@ fn history {
   candidates = [(edit:command-history | each [arg]{
         put [
 	        &content=$arg[cmd]
-	        &display=$arg[id]" "(replaces "\t" " " (replaces "\n" " " $arg[cmd]))
+	        &display=$arg[id]" "$arg[cmd]
         ]
   })]
 
diff -pruN 0.11+ds1-5/eval/chdir.go 0.12+ds1-1/eval/chdir.go
--- 0.11+ds1-5/eval/chdir.go	2018-01-18 02:03:11.000000000 +0000
+++ 0.12+ds1-1/eval/chdir.go	2018-07-11 02:32:25.000000000 +0000
@@ -12,25 +12,29 @@ type AddDirer interface {
 
 // Chdir changes the current directory. On success it also updates the PWD
 // environment variable and records the new directory in the directory history.
-// It returns nil as long as the directory changing part succeeds.
-func Chdir(path string, store AddDirer) error {
+// It runs the functions in beforeChdir immediately before changing the
+// directory, and the functions in afterChdir immediately after (if chdir was
+// successful). It returns nil as long as the directory changing part succeeds.
+func (ev *Evaler) Chdir(path string) error {
+	for _, hook := range ev.beforeChdir {
+		hook(path)
+	}
+
 	err := os.Chdir(path)
 	if err != nil {
 		return err
 	}
+
+	for _, hook := range ev.afterChdir {
+		hook(path)
+	}
+
 	pwd, err := os.Getwd()
 	if err != nil {
 		logger.Println("getwd after cd:", err)
 		return nil
 	}
 	os.Setenv("PWD", pwd)
-	if store != nil {
-		go func() {
-			err := store.AddDir(pwd, 1)
-			if err != nil {
-				logger.Println("Failed to save dir to history:", err)
-			}
-		}()
-	}
+
 	return nil
 }
diff -pruN 0.11+ds1-5/eval/chdir_test.go 0.12+ds1-1/eval/chdir_test.go
--- 0.11+ds1-5/eval/chdir_test.go	2018-01-18 02:03:11.000000000 +0000
+++ 0.12+ds1-1/eval/chdir_test.go	2018-07-11 02:32:25.000000000 +0000
@@ -1,11 +1,10 @@
 package eval
 
 import (
-	"errors"
 	"os"
 	"testing"
-	"time"
 
+	"github.com/elves/elvish/parse"
 	"github.com/elves/elvish/util"
 )
 
@@ -16,59 +15,61 @@ func (t testAddDirer) AddDir(dir string,
 }
 
 func TestChdir(t *testing.T) {
-	util.WithTempDir(func(destDir string) {
-		pwd, err := os.Getwd()
-		if err != nil {
-			panic(err)
-		}
-		defer func() {
-			err = os.Chdir(pwd)
-			if err != nil {
-				panic(err)
-			}
-		}()
-
-		chanAddedDir := make(chan string)
-		testAddDirer := testAddDirer(func(dir string, weight float64) error {
-			chanAddedDir <- dir
-			// Error returned from here should not affect the return value of
-			// Chdir
-			return errors.New("fake error")
-		})
+	inWithTempDir(func(pwd, dst string) {
+		ev := NewEvaler()
+
+		argDirInBefore, argDirInAfter := "", ""
+		ev.AddBeforeChdir(func(dir string) { argDirInBefore = dir })
+		ev.AddAfterChdir(func(dir string) { argDirInAfter = dir })
 
-		err = Chdir(destDir, testAddDirer)
+		err := ev.Chdir(dst)
 
 		if err != nil {
 			t.Errorf("Chdir => error %v", err)
 		}
-
-		if env := os.Getenv("PWD"); env != destDir {
-			t.Errorf("$PWD is %q after Chdir, want %q", env, destDir)
+		if envPwd := os.Getenv("PWD"); envPwd != dst {
+			t.Errorf("$PWD is %q after Chdir, want %q", envPwd, dst)
 		}
 
-		select {
-		case addedDir := <-chanAddedDir:
-			if addedDir != destDir {
-				t.Errorf("Chdir called AddDir %q, want %q", addedDir[0], destDir)
-			}
-		case <-time.After(100 * time.Millisecond):
-			t.Errorf("Chdir did not call AddDir within 100ms")
+		if argDirInBefore != dst {
+			t.Errorf("Chdir called before-hook with %q, want %q",
+				argDirInBefore, dst)
+		}
+		if argDirInAfter != dst {
+			t.Errorf("Chdir called before-hook with %q, want %q",
+				argDirInAfter, dst)
 		}
 	})
 }
 
-const badDir = "/i/dont/exist"
+func TestChdirElvishHooks(t *testing.T) {
+	inWithTempDir(func(pwd, dst string) {
+		Test(t,
+			That(`
+			dir-in-before dir-in-after = '' ''
+			@before-chdir = [dst]{ dir-in-before = $dst }
+			@after-chdir  = [dst]{ dir-in-after  = $dst }
+			cd `+parse.Quote(dst)+`
+			put $dir-in-before $dir-in-after
+			`).Puts(dst, dst),
+		)
+	})
+}
 
 func TestChdirError(t *testing.T) {
-	testAddDirer := testAddDirer(func(dir string, weight float64) error {
-		t.Errorf("Chdir called AddDir when os.Chdir errors")
-		return nil
+	util.InTempDir(func(pwd string) {
+		ev := NewEvaler()
+		err := ev.Chdir("i/dont/exist")
+		if err == nil {
+			t.Errorf("Chdir => no error when dir does not exist")
+		}
+	})
+}
+
+func inWithTempDir(f func(pwd, other string)) {
+	util.InTempDir(func(pwd string) {
+		util.WithTempDir(func(other string) {
+			f(pwd, other)
+		})
 	})
-	if _, err := os.Stat(badDir); err == nil {
-		panic(badDir + " exists")
-	}
-	err := Chdir(badDir, testAddDirer)
-	if err == nil {
-		t.Errorf("Chdir => no error when dir does not exist")
-	}
 }
diff -pruN 0.11+ds1-5/eval/closure.go 0.12+ds1-1/eval/closure.go
--- 0.11+ds1-5/eval/closure.go	2018-01-18 02:03:11.000000000 +0000
+++ 0.12+ds1-1/eval/closure.go	2018-07-11 02:32:25.000000000 +0000
@@ -5,8 +5,8 @@ import (
 	"fmt"
 	"unsafe"
 
-	"github.com/elves/elvish/eval/types"
-	"github.com/elves/elvish/eval/vartypes"
+	"github.com/elves/elvish/eval/vals"
+	"github.com/elves/elvish/eval/vars"
 	"github.com/elves/elvish/parse"
 	"github.com/xiaq/persistent/hash"
 )
@@ -21,13 +21,15 @@ type Closure struct {
 	// The name for the rest argument. If empty, the function has fixed arity.
 	RestArg     string
 	OptNames    []string
-	OptDefaults []types.Value
+	OptDefaults []interface{}
 	Op          Op
 	Captured    Ns
 	SrcMeta     *Source
+	DefBegint   int
+	DefEnd      int
 }
 
-var _ Fn = &Closure{}
+var _ Callable = &Closure{}
 
 // Kind returns "fn".
 func (*Closure) Kind() string {
@@ -48,15 +50,40 @@ func (c *Closure) Repr(int) string {
 	return fmt.Sprintf("<closure %p>", c)
 }
 
+func (c *Closure) Index(k interface{}) (interface{}, bool) {
+	switch k {
+	case "arg-names":
+		return vals.MakeStringList(c.ArgNames...), true
+	case "rest-arg":
+		return c.RestArg, true
+	case "opt-names":
+		return vals.MakeStringList(c.OptNames...), true
+	case "opt-defaults":
+		return vals.MakeList(c.OptDefaults...), true
+	case "body":
+		return c.SrcMeta.code[c.Op.Begin:c.Op.End], true
+	case "def":
+		return c.SrcMeta.code[c.DefBegint:c.DefEnd], true
+	case "src":
+		return c.SrcMeta, true
+	}
+	return nil, false
+}
+
+func (c *Closure) IterateKeys(f func(interface{}) bool) {
+	vals.Feed(f, "arg-names", "rest-arg",
+		"opt-names", "opt-defaults", "body", "def", "src")
+}
+
 // Call calls a closure.
-func (c *Closure) Call(ec *Frame, args []types.Value, opts map[string]types.Value) {
+func (c *Closure) Call(fm *Frame, args []interface{}, opts map[string]interface{}) error {
 	if c.RestArg != "" {
 		if len(c.ArgNames) > len(args) {
-			throwf("need %d or more arguments, got %d", len(c.ArgNames), len(args))
+			return fmt.Errorf("need %d or more arguments, got %d", len(c.ArgNames), len(args))
 		}
 	} else {
 		if len(c.ArgNames) != len(args) {
-			throwf("need %d arguments, got %d", len(c.ArgNames), len(args))
+			return fmt.Errorf("need %d arguments, got %d", len(c.ArgNames), len(args))
 		}
 	}
 
@@ -66,19 +93,19 @@ func (c *Closure) Call(ec *Frame, args [
 
 	// Make upvalue namespace and capture variables.
 	// TODO(xiaq): Is it safe to simply assign ec.up = c.Captured?
-	ec.up = make(Ns)
+	fm.up = make(Ns)
 	for name, variable := range c.Captured {
-		ec.up[name] = variable
+		fm.up[name] = variable
 	}
 
 	// Populate local scope with arguments, possibly a rest argument, and
 	// options.
-	ec.local = make(Ns)
+	fm.local = make(Ns)
 	for i, name := range c.ArgNames {
-		ec.local[name] = vartypes.NewPtr(args[i])
+		fm.local[name] = vars.NewAnyWithInit(args[i])
 	}
 	if c.RestArg != "" {
-		ec.local[c.RestArg] = vartypes.NewPtr(types.MakeList(args[len(c.ArgNames):]...))
+		fm.local[c.RestArg] = vars.NewAnyWithInit(vals.MakeList(args[len(c.ArgNames):]...))
 	}
 	optUsed := make(map[string]struct{})
 	for i, name := range c.OptNames {
@@ -88,17 +115,17 @@ func (c *Closure) Call(ec *Frame, args [
 		} else {
 			v = c.OptDefaults[i]
 		}
-		ec.local[name] = vartypes.NewPtr(v)
+		fm.local[name] = vars.NewAnyWithInit(v)
 	}
 	for name := range opts {
 		_, used := optUsed[name]
 		if !used {
-			throwf("unknown option %s", parse.Quote(name))
+			return fmt.Errorf("unknown option %s", parse.Quote(name))
 		}
 	}
 
-	ec.traceback = ec.addTraceback()
+	fm.traceback = fm.addTraceback()
 
-	ec.srcMeta = c.SrcMeta
-	c.Op.Exec(ec)
+	fm.srcMeta = c.SrcMeta
+	return c.Op.Exec(fm)
 }
diff -pruN 0.11+ds1-5/eval/closure_test.go 0.12+ds1-1/eval/closure_test.go
--- 0.11+ds1-5/eval/closure_test.go	2018-01-18 02:03:11.000000000 +0000
+++ 0.12+ds1-1/eval/closure_test.go	2018-07-11 02:32:25.000000000 +0000
@@ -3,13 +3,22 @@ package eval
 import "testing"
 
 func TestClosure(t *testing.T) {
-	runTests(t, []Test{
-		NewTest("kind-of { }").WantOutStrings("fn"),
-		NewTest("eq { } { }").WantOutBools(false),
-		NewTest("x = { }; put [&$x= foo][$x]").WantOutStrings("foo"),
-		NewTest("[x]{ } a b").WantAnyErr(),
-		NewTest("[x y]{ } a").WantAnyErr(),
-		NewTest("[x y @rest]{ } a").WantAnyErr(),
-		NewTest("[]{ } &k=v").WantAnyErr(),
-	})
+	Test(t,
+		That("kind-of { }").Puts("fn"),
+		That("eq { } { }").Puts(false),
+		That("x = { }; put [&$x= foo][$x]").Puts("foo"),
+		That("[x]{ } a b").Errors(),
+		That("[x y]{ } a").Errors(),
+		That("[x y @rest]{ } a").Errors(),
+		That("[]{ } &k=v").Errors(),
+
+		That("explode [a b]{ }[arg-names]").Puts("a", "b"),
+		That("put [@r]{ }[rest-arg]").Puts("r"),
+		That("explode [&opt=def]{ }[opt-names]").Puts("opt"),
+		That("explode [&opt=def]{ }[opt-defaults]").Puts("def"),
+		That("put { body }[body]").Puts(" body "),
+		That("put [x @y]{ body }[def]").Puts("[x @y]{ body }"),
+		That("put { body }[src][code]").
+			Puts("put { body }[src][code]"),
+	)
 }
diff -pruN 0.11+ds1-5/eval/compile_lvalue.go 0.12+ds1-1/eval/compile_lvalue.go
--- 0.11+ds1-5/eval/compile_lvalue.go	2018-01-18 02:03:11.000000000 +0000
+++ 0.12+ds1-1/eval/compile_lvalue.go	2018-07-11 02:32:25.000000000 +0000
@@ -2,30 +2,32 @@ package eval
 
 import (
 	"errors"
+	"fmt"
 	"strings"
 
-	"github.com/elves/elvish/eval/types"
-	"github.com/elves/elvish/eval/vartypes"
+	"github.com/elves/elvish/eval/vars"
 	"github.com/elves/elvish/parse"
 )
 
-// LValuesOp is an operation on an EvalCtx that produce Variable's.
+// LValuesOp is an operation on an Frame that produce Variable's.
 type LValuesOp struct {
-	Func       LValuesOpFunc
+	Body       LValuesOpBody
 	Begin, End int
 }
 
-// LValuesOpFunc is the body of an LValuesOp.
-type LValuesOpFunc func(*Frame) []vartypes.Variable
+// LValuesOpBody is the body of an LValuesOp.
+type LValuesOpBody interface {
+	Invoke(*Frame) ([]vars.Var, error)
+}
 
 // Exec executes an LValuesOp, producing Variable's.
-func (op LValuesOp) Exec(ec *Frame) []vartypes.Variable {
+func (op LValuesOp) Exec(fm *Frame) ([]vars.Var, error) {
 	// Empty value is considered to generate no lvalues.
-	if op.Func == nil {
-		return []vartypes.Variable{}
+	if op.Body == nil {
+		return []vars.Var{}, nil
 	}
-	ec.begin, ec.end = op.Begin, op.End
-	return op.Func(ec)
+	fm.begin, fm.end = op.Begin, op.End
+	return op.Body.Invoke(fm)
 }
 
 // lvaluesOp compiles lvalues, returning the fixed part and, optionally a rest
@@ -52,9 +54,9 @@ func (cp *compiler) lvaluesOp(n *parse.I
 }
 
 func (cp *compiler) lvaluesMulti(nodes []*parse.Compound) (LValuesOp, LValuesOp) {
-	opFuncs := make([]LValuesOpFunc, len(nodes))
+	opFuncs := make([]LValuesOpBody, len(nodes))
 	var restNode *parse.Indexing
-	var restOpFunc LValuesOpFunc
+	var restOpFunc LValuesOpBody
 
 	// Compile each spec inside the brace.
 	fixedEnd := 0
@@ -87,54 +89,23 @@ func (cp *compiler) lvaluesMulti(nodes [
 	var op LValuesOp
 	// If there is still anything left in opFuncs, make LValuesOp for the fixed part.
 	if len(opFuncs) > 0 {
-		op = LValuesOp{func(ec *Frame) []vartypes.Variable {
-			var variables []vartypes.Variable
-			for _, opFunc := range opFuncs {
-				variables = append(variables, opFunc(ec)...)
-			}
-			return variables
-		}, nodes[0].Begin(), fixedEnd}
+		op = LValuesOp{seqLValuesOpBody{opFuncs}, nodes[0].Begin(), fixedEnd}
 	}
 
 	return op, restOp
 }
 
-func (cp *compiler) lvalueBase(n *parse.Indexing, msg string) (bool, LValuesOpFunc) {
+func (cp *compiler) lvalueBase(n *parse.Indexing, msg string) (bool, LValuesOpBody) {
 	qname := cp.literal(n.Head, msg)
-	explode, ns, name := ParseVariable(qname)
+	explode, ns, name := ParseVariableRef(qname)
 	if len(n.Indicies) == 0 {
-		return explode, cp.lvalueVariable(ns, name)
+		cp.registerVariableSet(ns, name)
+		return explode, varOp{ns, name}
 	}
 	return explode, cp.lvalueElement(ns, name, n)
 }
 
-func (cp *compiler) lvalueVariable(ns, name string) LValuesOpFunc {
-	cp.registerVariableSet(ns, name)
-
-	return func(ec *Frame) []vartypes.Variable {
-		variable := ec.ResolveVar(ns, name)
-		if variable == nil {
-			if ns == "" || ns == "local" {
-				// New variable.
-				// XXX We depend on the fact that this variable will
-				// immeidately be set.
-				if strings.HasSuffix(name, FnSuffix) {
-					variable = vartypes.NewValidatedPtr(nil, ShouldBeFn)
-				} else if strings.HasSuffix(name, NsSuffix) {
-					variable = vartypes.NewValidatedPtr(nil, ShouldBeNs)
-				} else {
-					variable = vartypes.NewPtr(nil)
-				}
-				ec.local[name] = variable
-			} else {
-				throwf("new variables can only be created in local scope")
-			}
-		}
-		return []vartypes.Variable{variable}
-	}
-}
-
-func (cp *compiler) lvalueElement(ns, name string, n *parse.Indexing) LValuesOpFunc {
+func (cp *compiler) lvalueElement(ns, name string, n *parse.Indexing) LValuesOpBody {
 	cp.registerVariableGet(ns, name)
 
 	begin, end := n.Begin(), n.End()
@@ -146,30 +117,88 @@ func (cp *compiler) lvalueElement(ns, na
 
 	indexOps := cp.arrayOps(n.Indicies)
 
-	return func(ec *Frame) []vartypes.Variable {
-		variable := ec.ResolveVar(ns, name)
-		if variable == nil {
-			throwf("variable $%s:%s does not exist, compiler bug", ns, name)
+	return &elemOp{ns, name, indexOps, begin, end, ends}
+}
+
+type seqLValuesOpBody struct {
+	ops []LValuesOpBody
+}
+
+func (op seqLValuesOpBody) Invoke(fm *Frame) ([]vars.Var, error) {
+	var variables []vars.Var
+	for _, op := range op.ops {
+		moreVariables, err := op.Invoke(fm)
+		if err != nil {
+			return nil, err
 		}
+		variables = append(variables, moreVariables...)
+	}
+	return variables, nil
+}
 
-		indicies := make([]types.Value, len(indexOps))
-		for i, op := range indexOps {
-			values := op.Exec(ec)
-			// TODO: Implement multi-indexing.
-			if len(values) != 1 {
-				throw(errors.New("multi indexing not implemented"))
+type varOp struct {
+	ns, name string
+}
+
+func (op varOp) Invoke(fm *Frame) ([]vars.Var, error) {
+	variable := fm.ResolveVar(op.ns, op.name)
+	if variable == nil {
+		if op.ns == "" || op.ns == "local" {
+			// New variable.
+			// XXX We depend on the fact that this variable will
+			// immeidately be set.
+			if strings.HasSuffix(op.name, FnSuffix) {
+				val := Callable(nil)
+				variable = vars.FromPtr(&val)
+			} else if strings.HasSuffix(op.name, NsSuffix) {
+				val := Ns(nil)
+				variable = vars.FromPtr(&val)
+			} else {
+				variable = vars.NewAnyWithInit(nil)
 			}
-			indicies[i] = values[0]
+			fm.local[op.name] = variable
+		} else {
+			return nil, fmt.Errorf("new variables can only be created in local scope")
 		}
-		elemVar, err := vartypes.MakeElement(variable, indicies)
+	}
+	return []vars.Var{variable}, nil
+}
+
+type elemOp struct {
+	ns       string
+	name     string
+	indexOps []ValuesOp
+	begin    int
+	end      int
+	ends     []int
+}
+
+func (op *elemOp) Invoke(fm *Frame) ([]vars.Var, error) {
+	variable := fm.ResolveVar(op.ns, op.name)
+	if variable == nil {
+		return nil, fmt.Errorf("variable $%s:%s does not exist, compiler bug", op.ns, op.name)
+	}
+
+	indicies := make([]interface{}, len(op.indexOps))
+	for i, op := range op.indexOps {
+		values, err := op.Exec(fm)
 		if err != nil {
-			level := vartypes.GetElementErrorLevel(err)
-			if level < 0 {
-				ec.errorpf(begin, end, "%s", err)
-			} else {
-				ec.errorpf(begin, ends[level], "%s", err)
-			}
+			return nil, err
+		}
+		// TODO: Implement multi-indexing.
+		if len(values) != 1 {
+			return nil, errors.New("multi indexing not implemented")
+		}
+		indicies[i] = values[0]
+	}
+	elemVar, err := vars.MakeElement(variable, indicies)
+	if err != nil {
+		level := vars.ElementErrorLevel(err)
+		if level < 0 {
+			fm.errorpf(op.begin, op.end, "%s", err)
+		} else {
+			fm.errorpf(op.begin, op.ends[level], "%s", err)
 		}
-		return []vartypes.Variable{elemVar}
 	}
+	return []vars.Var{elemVar}, nil
 }
diff -pruN 0.11+ds1-5/eval/compile_lvalue_test.go 0.12+ds1-1/eval/compile_lvalue_test.go
--- 0.11+ds1-5/eval/compile_lvalue_test.go	2018-01-18 02:03:11.000000000 +0000
+++ 0.12+ds1-1/eval/compile_lvalue_test.go	2018-07-11 02:32:25.000000000 +0000
@@ -3,10 +3,10 @@ package eval
 import "testing"
 
 func TestAssignment(t *testing.T) {
-	runTests(t, []Test{
-		NewTest("x = a; put $x").WantOutStrings("a"),
-		NewTest("x = [a]; x[0] = b; put $x[0]").WantOutStrings("b"),
-		NewTest("x = a; { x = b }; put $x").WantOutStrings("b"),
-		NewTest("x = [a]; { x[0] = b }; put $x[0]").WantOutStrings("b"),
-	})
+	Test(t,
+		That("x = a; put $x").Puts("a"),
+		That("x = [a]; x[0] = b; put $x[0]").Puts("b"),
+		That("x = a; { x = b }; put $x").Puts("b"),
+		That("x = [a]; { x[0] = b }; put $x[0]").Puts("b"),
+	)
 }
diff -pruN 0.11+ds1-5/eval/compile_op.go 0.12+ds1-1/eval/compile_op.go
--- 0.11+ds1-5/eval/compile_op.go	2018-01-18 02:03:11.000000000 +0000
+++ 0.12+ds1-1/eval/compile_op.go	2018-07-11 02:32:25.000000000 +0000
@@ -2,155 +2,169 @@ package eval
 
 import (
 	"errors"
+	"fmt"
 	"os"
 	"sync"
 
-	"github.com/elves/elvish/eval/types"
-	"github.com/elves/elvish/eval/vartypes"
+	"github.com/elves/elvish/eval/vals"
+	"github.com/elves/elvish/eval/vars"
 	"github.com/elves/elvish/parse"
+	"github.com/elves/elvish/util"
+	"github.com/xiaq/persistent/hashmap"
 )
 
-// Op is an operation on an EvalCtx.
+// Op is an operation on an Frame.
 type Op struct {
-	Func       OpFunc
+	Body       OpBody
 	Begin, End int
 }
 
-// OpFunc is the body of an Op.
-type OpFunc func(*Frame)
+// OpBody is the body of an Op.
+type OpBody interface {
+	Invoke(*Frame) error
+}
 
 // Exec executes an Op.
-func (op Op) Exec(ec *Frame) {
-	ec.begin, ec.end = op.Begin, op.End
-	op.Func(ec)
-}
-
-func (cp *compiler) chunk(n *parse.Chunk) OpFunc {
-	ops := cp.pipelineOps(n.Pipelines)
-
-	return func(ec *Frame) {
-		for _, op := range ops {
-			op.Exec(ec)
-		}
-		// Check for interrupts after the chunk.
-		// We also check for interrupts before each pipeline, so there is no
-		// need to check it before the chunk or after each pipeline.
-		ec.CheckInterrupts()
-	}
+func (op Op) Exec(fm *Frame) error {
+	fm.begin, fm.end = op.Begin, op.End
+	return op.Body.Invoke(fm)
 }
 
-const pipelineChanBufferSize = 32
-
-func (cp *compiler) pipeline(n *parse.Pipeline) OpFunc {
-	ops := cp.formOps(n.Forms)
+func (cp *compiler) chunk(n *parse.Chunk) OpBody {
+	return chunkOp{cp.pipelineOps(n.Pipelines)}
+}
 
-	return func(ec *Frame) {
-		ec.CheckInterrupts()
+type chunkOp struct {
+	subops []Op
+}
 
-		bg := n.Background
-		if bg {
-			ec = ec.fork("background job " + n.SourceText())
-			ec.intCh = nil
-			ec.background = true
-
-			if ec.Editor != nil {
-				// TODO: Redirect output in interactive mode so that the line
-				// editor does not get messed up.
-			}
+func (op chunkOp) Invoke(fm *Frame) error {
+	for _, subop := range op.subops {
+		err := subop.Exec(fm)
+		if err != nil {
+			return err
 		}
+	}
+	// Check for interrupts after the chunk.
+	// We also check for interrupts before each pipeline, so there is no
+	// need to check it before the chunk or after each pipeline.
+	if fm.IsInterrupted() {
+		return ErrInterrupted
+	}
+	return nil
+}
 
-		nforms := len(ops)
-
-		var wg sync.WaitGroup
-		wg.Add(nforms)
-		errors := make([]*Exception, nforms)
-
-		var nextIn *Port
-
-		// For each form, create a dedicated evalCtx and run asynchronously
-		for i, op := range ops {
-			hasChanInput := i > 0
-			newEc := ec.fork("[form op]")
-			if i > 0 {
-				newEc.ports[0] = nextIn
-			}
-			if i < nforms-1 {
-				// Each internal port pair consists of a (byte) pipe pair and a
-				// channel.
-				// os.Pipe sets O_CLOEXEC, which is what we want.
-				reader, writer, e := os.Pipe()
-				if e != nil {
-					throwf("failed to create pipe: %s", e)
-				}
-				ch := make(chan types.Value, pipelineChanBufferSize)
-				newEc.ports[1] = &Port{
-					File: writer, Chan: ch, CloseFile: true, CloseChan: true}
-				nextIn = &Port{
-					File: reader, Chan: ch, CloseFile: true, CloseChan: false}
-			}
-			thisOp := op
-			thisError := &errors[i]
-			go func() {
-				err := newEc.PEval(thisOp)
-				// Logger.Printf("closing ports of %s", newEc.context)
-				ClosePorts(newEc.ports)
-				if err != nil {
-					*thisError = err.(*Exception)
-				}
-				wg.Done()
-				if hasChanInput {
-					// If the command has channel input, drain it. This
-					// mitigates the effect of erroneous pipelines like
-					// "range 100 | cat"; without draining the pipeline will
-					// lock up.
-					for range newEc.ports[0].Chan {
-					}
-				}
-			}()
-		}
+func (cp *compiler) pipeline(n *parse.Pipeline) OpBody {
+	return &pipelineOp{n.Background, n.SourceText(), cp.formOps(n.Forms)}
+}
 
-		if bg {
-			// Background job, wait for form termination asynchronously.
-			go func() {
-				wg.Wait()
-				msg := "job " + n.SourceText() + " finished"
-				err := ComposeExceptionsFromPipeline(errors)
-				if err != nil {
-					msg += ", errors = " + err.Error()
-				}
-				if ec.Editor != nil {
-					m := ec.Editor.ActiveMutex()
-					m.Lock()
-					defer m.Unlock()
-
-					if ec.Editor.Active() {
-						ec.Editor.Notify("%s", msg)
-					} else {
-						ec.ports[2].File.WriteString(msg + "\n")
-					}
+type pipelineOp struct {
+	bg     bool
+	source string
+	subops []Op
+}
+
+const pipelineChanBufferSize = 32
+
+func (op *pipelineOp) Invoke(fm *Frame) error {
+	if fm.IsInterrupted() {
+		return ErrInterrupted
+	}
+
+	if op.bg {
+		fm = fm.fork("background job" + op.source)
+		fm.intCh = nil
+		fm.background = true
+		fm.Evaler.numBgJobs++
+
+		if fm.Editor != nil {
+			// TODO: Redirect output in interactive mode so that the line
+			// editor does not get messed up.
+		}
+	}
+
+	nforms := len(op.subops)
+
+	var wg sync.WaitGroup
+	wg.Add(nforms)
+	errors := make([]*Exception, nforms)
+
+	var nextIn *Port
+
+	// For each form, create a dedicated evalCtx and run asynchronously
+	for i, op := range op.subops {
+		hasChanInput := i > 0
+		newFm := fm.fork("[form op]")
+		if i > 0 {
+			newFm.ports[0] = nextIn
+		}
+		if i < nforms-1 {
+			// Each internal port pair consists of a (byte) pipe pair and a
+			// channel.
+			// os.Pipe sets O_CLOEXEC, which is what we want.
+			reader, writer, e := os.Pipe()
+			if e != nil {
+				return fmt.Errorf("failed to create pipe: %s", e)
+			}
+			ch := make(chan interface{}, pipelineChanBufferSize)
+			newFm.ports[1] = &Port{
+				File: writer, Chan: ch, CloseFile: true, CloseChan: true}
+			nextIn = &Port{
+				File: reader, Chan: ch, CloseFile: true, CloseChan: false}
+		}
+		thisOp := op
+		thisError := &errors[i]
+		go func() {
+			err := newFm.Eval(thisOp)
+			newFm.Close()
+			if err != nil {
+				*thisError = err.(*Exception)
+			}
+			wg.Done()
+			if hasChanInput {
+				// If the command has channel input, drain it. This
+				// mitigates the effect of erroneous pipelines like
+				// "range 100 | cat"; without draining the pipeline will
+				// lock up.
+				for range newFm.ports[0].Chan {
+				}
+			}
+		}()
+	}
+
+	if op.bg {
+		// Background job, wait for form termination asynchronously.
+		go func() {
+			wg.Wait()
+			fm.Evaler.numBgJobs--
+			msg := "job " + op.source + " finished"
+			err := ComposeExceptionsFromPipeline(errors)
+			if err != nil {
+				msg += ", errors = " + err.Error()
+			}
+			if fm.Evaler.notifyBgJobSuccess || err != nil {
+				if fm.Editor != nil {
+					fm.Editor.Notify("%s", msg)
 				} else {
-					ec.ports[2].File.WriteString(msg + "\n")
+					fm.ports[2].File.WriteString(msg + "\n")
 				}
-			}()
-		} else {
-			wg.Wait()
-			maybeThrow(ComposeExceptionsFromPipeline(errors))
-		}
+			}
+		}()
+		return nil
+	} else {
+		wg.Wait()
+		return ComposeExceptionsFromPipeline(errors)
 	}
 }
 
-func (cp *compiler) form(n *parse.Form) OpFunc {
+func (cp *compiler) form(n *parse.Form) OpBody {
 	var saveVarsOps []LValuesOp
 	var assignmentOps []Op
 	if len(n.Assignments) > 0 {
 		assignmentOps = cp.assignmentOps(n.Assignments)
 		if n.Head == nil && n.Vars == nil {
 			// Permanent assignment.
-			return func(ec *Frame) {
-				for _, op := range assignmentOps {
-					op.Exec(ec)
-				}
-			}
+			return seqOp{assignmentOps}
 		}
 		for _, a := range n.Assignments {
 			v, r := cp.lvaluesOp(a.Left)
@@ -162,7 +176,7 @@ func (cp *compiler) form(n *parse.Form)
 	// Depending on the type of the form, exactly one of the three below will be
 	// set.
 	var (
-		specialOpFunc  OpFunc
+		specialOpFunc  OpBody
 		headOp         ValuesOp
 		spaceyAssignOp Op
 	)
@@ -178,16 +192,14 @@ func (cp *compiler) form(n *parse.Form)
 				// Special form.
 				specialOpFunc = compileForm(cp, n)
 			} else {
-				var headOpFunc ValuesOpFunc
-				explode, ns, name := ParseVariable(headStr)
+				var headOpFunc ValuesOpBody
+				explode, ns, name := ParseVariableRef(headStr)
 				if !explode && cp.registerVariableGet(ns, name+FnSuffix) {
 					// $head~ resolves.
-					headOpFunc = variable(headStr + FnSuffix)
+					headOpFunc = variableOp{false, ns, name + FnSuffix}
 				} else {
 					// Fall back to $e:head~.
-					headOpFunc = func(f *Frame) []types.Value {
-						return []types.Value{ExternalCmd{headStr}}
-					}
+					headOpFunc = literalValues(ExternalCmd{headStr})
 				}
 				headOp = ValuesOp{headOpFunc, n.Head.Begin(), n.Head.End()}
 			}
@@ -199,22 +211,26 @@ func (cp *compiler) form(n *parse.Form)
 	} else {
 		// Assignment form.
 		varsOp, restOp := cp.lvaluesMulti(n.Vars)
+		// This cannot be replaced with newSeqValuesOp as it depends on the fact
+		// that argOps will be changed later.
 		argsOp := ValuesOp{
-			func(ec *Frame) []types.Value {
-				var vs []types.Value
+			funcValuesOp(func(fm *Frame) ([]interface{}, error) {
+				var values []interface{}
 				for _, op := range argOps {
-					vs = append(vs, op.Exec(ec)...)
+					moreValues, err := op.Exec(fm)
+					if err != nil {
+						return nil, err
+					}
+					values = append(values, moreValues...)
 				}
-				return vs
-			},
-			-1, -1,
-		}
+				return values, nil
+			}), -1, -1}
 		if len(argOps) > 0 {
 			argsOp.Begin = argOps[0].Begin
 			argsOp.End = argOps[len(argOps)-1].End
 		}
 		spaceyAssignOp = Op{
-			makeAssignmentOpFunc(varsOp, restOp, argsOp),
+			&assignmentOp{varsOp, restOp, argsOp},
 			n.Begin(), argsOp.End,
 		}
 	}
@@ -224,164 +240,216 @@ func (cp *compiler) form(n *parse.Form)
 	redirOps := cp.redirOps(n.Redirs)
 	// TODO: n.ErrorRedir
 
-	begin, end := n.Begin(), n.End()
+	return &formOp{saveVarsOps, assignmentOps, redirOps, specialOpFunc, headOp, argOps, optsOp, spaceyAssignOp, n.Begin(), n.End()}
+}
+
+type formOp struct {
+	saveVarsOps    []LValuesOp
+	assignmentOps  []Op
+	redirOps       []Op
+	specialOpBody  OpBody
+	headOp         ValuesOp
+	argOps         []ValuesOp
+	optsOp         ValuesOpBody
+	spaceyAssignOp Op
+	begin, end     int
+}
+
+func (op *formOp) Invoke(fm *Frame) (errRet error) {
 	// ec here is always a subevaler created in compiler.pipeline, so it can
 	// be safely modified.
-	return func(ec *Frame) {
-		// Temporary assignment.
-		if len(saveVarsOps) > 0 {
-			// There is a temporary assignment.
-			// Save variables.
-			var saveVars []vartypes.Variable
-			var saveVals []types.Value
-			for _, op := range saveVarsOps {
-				saveVars = append(saveVars, op.Exec(ec)...)
-			}
+
+	// Temporary assignment.
+	if len(op.saveVarsOps) > 0 {
+		// There is a temporary assignment.
+		// Save variables.
+		var saveVars []vars.Var
+		var saveVals []interface{}
+		for _, op := range op.saveVarsOps {
+			moreSaveVars, err := op.Exec(fm)
+			if err != nil {
+				return err
+			}
+			saveVars = append(saveVars, moreSaveVars...)
+		}
+		for i, v := range saveVars {
+			// XXX(xiaq): If the variable to save is a elemVariable, save
+			// the outermost variable instead.
+			if u := vars.HeadOfElement(v); u != nil {
+				v = u
+				saveVars[i] = v
+			}
+			val := v.Get()
+			saveVals = append(saveVals, val)
+			logger.Printf("saved %s = %s", v, val)
+		}
+		// Do assignment.
+		for _, subop := range op.assignmentOps {
+			err := subop.Exec(fm)
+			if err != nil {
+				return err
+			}
+		}
+		// Defer variable restoration. Will be executed even if an error
+		// occurs when evaling other part of the form.
+		defer func() {
 			for i, v := range saveVars {
-				// XXX(xiaq): If the variable to save is a elemVariable, save
-				// the outermost variable instead.
-				if u := vartypes.GetHeadOfElement(v); u != nil {
-					v = u
-					saveVars[i] = v
+				val := saveVals[i]
+				if val == nil {
+					// XXX Old value is nonexistent. We should delete the
+					// variable. However, since the compiler now doesn't delete
+					// it, we don't delete it in the evaler either.
+					val = ""
 				}
-				val := v.Get()
-				saveVals = append(saveVals, val)
-				logger.Printf("saved %s = %s", v, val)
-			}
-			// Do assignment.
-			for _, op := range assignmentOps {
-				op.Exec(ec)
-			}
-			// Defer variable restoration. Will be executed even if an error
-			// occurs when evaling other part of the form.
-			defer func() {
-				for i, v := range saveVars {
-					val := saveVals[i]
-					if val == nil {
-						// XXX Old value is nonexistent. We should delete the
-						// variable. However, since the compiler now doesn't delete
-						// it, we don't delete it in the evaler either.
-						val = types.String("")
-					}
-					err := v.Set(val)
-					maybeThrow(err)
-					logger.Printf("restored %s = %s", v, val)
+				err := v.Set(val)
+				if err != nil {
+					errRet = err
 				}
-			}()
+				logger.Printf("restored %s = %s", v, val)
+			}
+		}()
+	}
+
+	// redirs
+	for _, redirOp := range op.redirOps {
+		err := redirOp.Exec(fm)
+		if err != nil {
+			return err
 		}
+	}
+
+	if op.specialOpBody != nil {
+		return op.specialOpBody.Invoke(fm)
+	}
+	var headFn Callable
+	var args []interface{}
+	if op.headOp.Body != nil {
+		// head
+		headFn = fm.ExecAndUnwrap("head of command", op.headOp).One().Callable()
 
-		// redirs
-		for _, redirOp := range redirOps {
-			redirOp.Exec(ec)
+		// args
+		for _, argOp := range op.argOps {
+			moreArgs, err := argOp.Exec(fm)
+			if err != nil {
+				return err
+			}
+			args = append(args, moreArgs...)
 		}
+	}
 
-		if specialOpFunc != nil {
-			specialOpFunc(ec)
+	// opts
+	// XXX This conversion should be avoided.
+	optValues, err := op.optsOp.Invoke(fm)
+	if err != nil {
+		return err
+	}
+	opts := optValues[0].(hashmap.Map)
+	convertedOpts := make(map[string]interface{})
+	for it := opts.Iterator(); it.HasElem(); it.Next() {
+		k, v := it.Elem()
+		if ks, ok := k.(string); ok {
+			convertedOpts[ks] = v
 		} else {
-			var headFn Callable
-			var args []types.Value
-			if headOp.Func != nil {
-				// head
-				headFn = ec.ExecAndUnwrap("head of command", headOp).One().Callable()
-
-				// args
-				for _, argOp := range argOps {
-					args = append(args, argOp.Exec(ec)...)
-				}
-			}
+			return fmt.Errorf("Option key must be string, got %s", vals.Kind(k))
+		}
+	}
 
-			// opts
-			// XXX This conversion should be avoided.
-			opts := optsOp(ec)[0].(types.Map)
-			convertedOpts := make(map[string]types.Value)
-			opts.IteratePair(func(k, v types.Value) bool {
-				if ks, ok := k.(types.String); ok {
-					convertedOpts[string(ks)] = v
-				} else {
-					throwf("Option key must be string, got %s", k.Kind())
-				}
-				return true
-			})
-			ec.begin, ec.end = begin, end
+	fm.begin, fm.end = op.begin, op.end
 
-			if headFn != nil {
-				headFn.Call(ec, args, convertedOpts)
-			} else {
-				spaceyAssignOp.Exec(ec)
-			}
-		}
+	if headFn != nil {
+		return headFn.Call(fm, args, convertedOpts)
+	} else {
+		return op.spaceyAssignOp.Exec(fm)
 	}
 }
 
-func allTrue(vs []types.Value) bool {
+func allTrue(vs []interface{}) bool {
 	for _, v := range vs {
-		if !types.ToBool(v) {
+		if !vals.Bool(v) {
 			return false
 		}
 	}
 	return true
 }
 
-func (cp *compiler) assignment(n *parse.Assignment) OpFunc {
+func (cp *compiler) assignment(n *parse.Assignment) OpBody {
 	variablesOp, restOp := cp.lvaluesOp(n.Left)
 	valuesOp := cp.compoundOp(n.Right)
-	return makeAssignmentOpFunc(variablesOp, restOp, valuesOp)
+	return &assignmentOp{variablesOp, restOp, valuesOp}
 }
 
-// ErrMoreThanOneRest is thrown when the LHS of an assignment contains more than
-// one rest variables.
+// ErrMoreThanOneRest is returned when the LHS of an assignment contains more
+// than one rest variables.
 var ErrMoreThanOneRest = errors.New("more than one @ lvalue")
 
-func makeAssignmentOpFunc(variablesOp, restOp LValuesOp, valuesOp ValuesOp) OpFunc {
-	return func(ec *Frame) {
-		variables := variablesOp.Exec(ec)
-		rest := restOp.Exec(ec)
-
-		// If any LHS ends up being nil, assign an empty string to all of them.
-		//
-		// This is to fix #176, which only happens in the top level of REPL; in
-		// other cases, a failure in the evaluation of the RHS causes this
-		// level to fail, making the variables unaccessible.
-		//
-		// XXX(xiaq): Should think about how to get rid of this.
-		defer fixNilVariables(variables)
-		defer fixNilVariables(rest)
-
-		values := valuesOp.Exec(ec)
-
-		if len(rest) > 1 {
-			throw(ErrMoreThanOneRest)
-		}
-		if len(rest) == 1 {
-			if len(variables) > len(values) {
-				throw(ErrArityMismatch)
-			}
-		} else {
-			if len(variables) != len(values) {
-				throw(ErrArityMismatch)
-			}
+type assignmentOp struct {
+	variablesOp LValuesOp
+	restOp      LValuesOp
+	valuesOp    ValuesOp
+}
+
+func (op *assignmentOp) Invoke(fm *Frame) (errRet error) {
+	variables, err := op.variablesOp.Exec(fm)
+	if err != nil {
+		return err
+	}
+	rest, err := op.restOp.Exec(fm)
+	if err != nil {
+		return err
+	}
+
+	// If any LHS ends up being nil, assign an empty string to all of them.
+	//
+	// This is to fix #176, which only happens in the top level of REPL; in
+	// other cases, a failure in the evaluation of the RHS causes this
+	// level to fail, making the variables unaccessible.
+	//
+	// XXX(xiaq): Should think about how to get rid of this.
+	defer fixNilVariables(variables, &errRet)
+	defer fixNilVariables(rest, &errRet)
+
+	values, err := op.valuesOp.Exec(fm)
+	if err != nil {
+		return err
+	}
+
+	if len(rest) > 1 {
+		return ErrMoreThanOneRest
+	}
+	if len(rest) == 1 {
+		if len(variables) > len(values) {
+			return ErrArityMismatch
+		}
+	} else {
+		if len(variables) != len(values) {
+			return ErrArityMismatch
 		}
+	}
 
-		for i, variable := range variables {
-			err := variable.Set(values[i])
-			maybeThrow(err)
+	for i, variable := range variables {
+		err := variable.Set(values[i])
+		if err != nil {
+			return err
 		}
+	}
 
-		if len(rest) == 1 {
-			err := rest[0].Set(types.MakeList(values[len(variables):]...))
-			maybeThrow(err)
+	if len(rest) == 1 {
+		err := rest[0].Set(vals.MakeList(values[len(variables):]...))
+		if err != nil {
+			return err
 		}
 	}
+	return nil
 }
 
-func fixNilVariables(vs []vartypes.Variable) {
+func fixNilVariables(vs []vars.Var, perr *error) {
 	for _, v := range vs {
-		if vartypes.IsBlackhole(v) {
+		if vars.IsBlackhole(v) {
 			continue
 		}
 		if v.Get() == nil {
-			err := v.Set(types.String(""))
-			maybeThrow(err)
+			err := v.Set("")
+			*perr = util.Errors(*perr, err)
 		}
 	}
 }
@@ -400,84 +468,108 @@ func (cp *compiler) literal(n *parse.Pri
 const defaultFileRedirPerm = 0644
 
 // redir compiles a Redir into a op.
-func (cp *compiler) redir(n *parse.Redir) OpFunc {
+func (cp *compiler) redir(n *parse.Redir) OpBody {
 	var dstOp ValuesOp
 	if n.Left != nil {
 		dstOp = cp.compoundOp(n.Left)
 	}
-	srcOp := cp.compoundOp(n.Right)
-	sourceIsFd := n.RightIsFd
-	mode := n.Mode
-	flag := makeFlag(mode)
+	flag := makeFlag(n.Mode)
 	if flag == -1 {
 		// TODO: Record and get redirection sign position
 		cp.errorf("bad redirection sign")
 	}
+	return &redirOp{dstOp, cp.compoundOp(n.Right), n.RightIsFd, n.Mode, flag}
+}
 
-	return func(ec *Frame) {
-		var dst int
-		if dstOp.Func == nil {
-			// use default dst fd
-			switch mode {
-			case parse.Read:
-				dst = 0
-			case parse.Write, parse.ReadWrite, parse.Append:
-				dst = 1
-			default:
-				// XXX should report parser bug
-				panic("bad RedirMode; parser bug")
-			}
-		} else {
-			// dst must be a valid fd
-			dst = ec.ExecAndUnwrap("Fd", dstOp).One().NonNegativeInt()
+type redirOp struct {
+	dstOp   ValuesOp
+	srcOp   ValuesOp
+	srcIsFd bool
+	mode    parse.RedirMode
+	flag    int
+}
+
+func (op *redirOp) Invoke(fm *Frame) error {
+	var dst int
+	if op.dstOp.Body == nil {
+		// use default dst fd
+		switch op.mode {
+		case parse.Read:
+			dst = 0
+		case parse.Write, parse.ReadWrite, parse.Append:
+			dst = 1
+		default:
+			panic("bad RedirMode; parser bug")
 		}
+	} else {
+		// dst must be a valid fd
+		dst = fm.ExecAndUnwrap("Fd", op.dstOp).One().NonNegativeInt()
+	}
 
-		ec.growPorts(dst + 1)
-		// Logger.Printf("closing old port %d of %s", dst, ec.context)
-		ec.ports[dst].Close()
-
-		srcUnwrap := ec.ExecAndUnwrap("redirection source", srcOp).One()
-		if sourceIsFd {
-			src := srcUnwrap.FdOrClose()
-			if src == -1 {
-				// close
-				ec.ports[dst] = &Port{}
-			} else {
-				ec.ports[dst] = ec.ports[src].Fork()
-			}
+	fm.growPorts(dst + 1)
+	// Logger.Printf("closing old port %d of %s", dst, ec.context)
+	fm.ports[dst].Close()
+
+	srcUnwrap := fm.ExecAndUnwrap("redirection source", op.srcOp).One()
+	if op.srcIsFd {
+		src := srcUnwrap.FdOrClose()
+		if src == -1 {
+			// close
+			fm.ports[dst] = &Port{}
 		} else {
-			switch src := srcUnwrap.Any().(type) {
-			case types.String:
-				f, err := os.OpenFile(string(src), flag, defaultFileRedirPerm)
-				if err != nil {
-					throwf("failed to open file %s: %s", src.Repr(types.NoPretty), err)
-				}
-				ec.ports[dst] = &Port{
-					File: f, Chan: BlackholeChan,
-					CloseFile: true,
-				}
-			case types.File:
-				ec.ports[dst] = &Port{
-					File: src.Inner, Chan: BlackholeChan,
-					CloseFile: false,
-				}
-			case types.Pipe:
-				var f *os.File
-				switch mode {
-				case parse.Read:
-					f = src.ReadEnd
-				case parse.Write:
-					f = src.WriteEnd
-				default:
-					cp.errorf("can only use < or > with pipes")
-				}
-				ec.ports[dst] = &Port{
-					File: f, Chan: BlackholeChan,
-					CloseFile: false,
-				}
+			fm.ports[dst] = fm.ports[src].Fork()
+		}
+	} else {
+		switch src := srcUnwrap.Any().(type) {
+		case string:
+			f, err := os.OpenFile(src, op.flag, defaultFileRedirPerm)
+			if err != nil {
+				return fmt.Errorf("failed to open file %s: %s", vals.Repr(src, vals.NoPretty), err)
+			}
+			fm.ports[dst] = &Port{
+				File: f, Chan: BlackholeChan,
+				CloseFile: true,
+			}
+		case vals.File:
+			fm.ports[dst] = &Port{
+				File: src.Inner, Chan: BlackholeChan,
+				CloseFile: false,
+			}
+		case vals.Pipe:
+			var f *os.File
+			switch op.mode {
+			case parse.Read:
+				f = src.ReadEnd
+			case parse.Write:
+				f = src.WriteEnd
 			default:
-				srcUnwrap.error("string or file", "%s", src.Kind())
+				return errors.New("can only use < or > with pipes")
 			}
+			fm.ports[dst] = &Port{
+				File: f, Chan: BlackholeChan,
+				CloseFile: false,
+			}
+		default:
+			srcUnwrap.error("string or file", "%s", vals.Kind(src))
+		}
+	}
+	return nil
+}
+
+type seqOp struct{ subops []Op }
+
+func (op seqOp) Invoke(fm *Frame) error {
+	for _, subop := range op.subops {
+		err := subop.Exec(fm)
+		if err != nil {
+			return err
 		}
 	}
+	return nil
+}
+
+type funcOp func(*Frame) error
+
+func (op funcOp) Invoke(fm *Frame) error {
+	return op(fm)
 }
diff -pruN 0.11+ds1-5/eval/compile_op_test.go 0.12+ds1-1/eval/compile_op_test.go
--- 0.11+ds1-5/eval/compile_op_test.go	2018-01-18 02:03:11.000000000 +0000
+++ 0.12+ds1-1/eval/compile_op_test.go	2018-07-11 02:32:25.000000000 +0000
@@ -2,84 +2,79 @@ package eval
 
 import "testing"
 
-var opTests = []Test{
-	// Chunks
-	// ------
-
-	// Empty chunk
-	{"", wantNothing},
-	// Outputs of pipelines in a chunk are concatenated
-	{"put x; put y; put z", want{out: strs("x", "y", "z")}},
-	// A failed pipeline cause the whole chunk to fail
-	{"put a; e:false; put b", want{out: strs("a"), err: errAny}},
-
-	// Pipelines
-	// ---------
-
-	// Pure byte pipeline
-	{`echo "Albert\nAllan\nAlbraham\nBerlin" | sed s/l/1/g | grep e`,
-		want{bytesOut: []byte("A1bert\nBer1in\n")}},
-	// Pure channel pipeline
-	{`put 233 42 19 | each [x]{+ $x 10}`, want{out: strs("243", "52", "29")}},
-	// Pipeline draining.
-	{`range 100 | put x`, want{out: strs("x")}},
-	// TODO: Add a useful hybrid pipeline sample
-
-	// Assignments
-	// -----------
-
-	// List element assignment
-	{"li=[foo bar]; li[0]=233; put $@li",
-		want{out: strs("233", "bar")}},
-	// Map element assignment
-	{"di=[&k=v]; di[k]=lorem; di[k2]=ipsum; put $di[k] $di[k2]",
-		want{out: strs("lorem", "ipsum")}},
-	{"d=[&a=[&b=v]]; put $d[a][b]; d[a][b]=u; put $d[a][b]",
-		want{out: strs("v", "u")}},
-	// Multi-assignments.
-	{"{a,b}=(put a b); put $a $b", want{out: strs("a", "b")}},
-	{"@a=(put a b); put $@a", want{out: strs("a", "b")}},
-	{"{a,@b}=(put a b c); put $@b", want{out: strs("b", "c")}},
-	//{"di=[&]; di[a b]=(put a b); put $di[a] $di[b]", want{out: strs("a", "b")}},
-
-	// Temporary assignment.
-	{"a=alice b=bob; {a,@b}=(put amy ben) put $a $@b; put $a $b",
-		want{out: strs("amy", "ben", "alice", "bob")}},
-	// Temporary assignment of list element.
-	{"l = [a]; l[0]=x put $l[0]; put $l[0]", want{out: strs("x", "a")}},
-	// Temporary assignment of map element.
-	{"m = [&k=v]; m[k]=v2 put $m[k]; put $m[k]", want{out: strs("v2", "v")}},
-	// Temporary assignment before special form.
-	{"li=[foo bar] for x $li { put $x }", want{out: strs("foo", "bar")}},
-
-	// Spacey assignment.
-	{"a @b = 2 3 foo; put $a $b[1]", want{out: strs("2", "foo")}},
-	// Spacey assignment with temporary assignment
-	{"x = 1; x=2 y = (+ 1 $x); put $x $y", want{out: strs("1", "3")}},
+func TestOp(t *testing.T) {
+	Test(t,
+		// Chunks
+		// ------
+
+		// Empty chunk
+		That("").DoesNothing(),
+		// Outputs of pipelines in a chunk are concatenated
+		That("put x; put y; put z").Puts("x", "y", "z"),
+		// A failed pipeline cause the whole chunk to fail
+		That("put a; e:false; put b").Puts("a").Errors(),
+
+		// Pipelines
+		// ---------
+
+		// Pure byte pipeline
+		That(`echo "Albert\nAllan\nAlbraham\nBerlin" | sed s/l/1/g | grep e`).
+			Prints("A1bert\nBer1in\n"),
+		// Pure channel pipeline
+		That(`put 233 42 19 | each [x]{+ $x 10}`).Puts("243", "52", "29"),
+		// Pipeline draining.
+		That(`range 100 | put x`).Puts("x"),
+		// TODO: Add a useful hybrid pipeline sample
+
+		// Assignments
+		// -----------
+
+		// List element assignment
+		That("li=[foo bar]; li[0]=233; put $@li").Puts("233", "bar"),
+		// Map element assignment
+		That("di=[&k=v]; di[k]=lorem; di[k2]=ipsum; put $di[k] $di[k2]").
+			Puts("lorem", "ipsum"),
+		That("d=[&a=[&b=v]]; put $d[a][b]; d[a][b]=u; put $d[a][b]").
+			Puts("v", "u"),
+		// Multi-assignments.
+		That("{a,b}=(put a b); put $a $b").Puts("a", "b"),
+		That("@a=(put a b); put $@a").Puts("a", "b"),
+		That("{a,@b}=(put a b c); put $@b").Puts("b", "c"),
+		//That("di=[&]; di[a b]=(put a b); put $di[a] $di[b]").Puts("a", "b"),
+
+		// Temporary assignment.
+		That("a=alice b=bob; {a,@b}=(put amy ben) put $a $@b; put $a $b").
+			Puts("amy", "ben", "alice", "bob"),
+		// Temporary assignment of list element.
+		That("l = [a]; l[0]=x put $l[0]; put $l[0]").Puts("x", "a"),
+		// Temporary assignment of map element.
+		That("m = [&k=v]; m[k]=v2 put $m[k]; put $m[k]").Puts("v2", "v"),
+		// Temporary assignment before special form.
+		That("li=[foo bar] for x $li { put $x }").Puts("foo", "bar"),
+
+		// Spacey assignment.
+		That("a @b = 2 3 foo; put $a $b[1]").Puts("2", "foo"),
+		// Spacey assignment with temporary assignment
+		That("x = 1; x=2 y = (+ 1 $x); put $x $y").Puts("1", "3"),
 
-	// Redirections
-	// ------------
+		// Redirections
+		// ------------
 
-	{"f=(mktemp elvXXXXXX); echo 233 > $f; cat < $f; rm $f",
-		want{bytesOut: []byte("233\n")}},
+		That("f=(mktemp elvXXXXXX); echo 233 > $f; cat < $f; rm $f").
+			Prints("233\n"),
 
-	// Redirections from special form.
-	{`f = (mktemp elvXXXXXX);
+		// Redirections from special form.
+		That(`f = (mktemp elvXXXXXX);
 	for x [lorem ipsum] { echo $x } > $f
 	cat $f
-	rm $f`,
-		want{bytesOut: []byte("lorem\nipsum\n")}},
-
-	// Redirections from File object.
-	{`fname=(mktemp elvXXXXXX); echo haha > $fname;
-			f=(fopen $fname); cat <$f; fclose $f; rm $fname`,
-		want{bytesOut: []byte("haha\n")}},
-
-	// Redirections from Pipe object.
-	{`p=(pipe); echo haha > $p; pwclose $p; cat < $p; prclose $p`,
-		want{bytesOut: []byte("haha\n")}},
-}
+	rm $f`).Prints("lorem\nipsum\n"),
 
-func TestOp(t *testing.T) {
-	runTests(t, opTests)
+		// Redirections from File object.
+		That(`fname=(mktemp elvXXXXXX); echo haha > $fname;
+			f=(fopen $fname); cat <$f; fclose $f; rm $fname`).Prints("haha\n"),
+
+		// Redirections from Pipe object.
+		That(`p=(pipe); echo haha > $p; pwclose $p; cat < $p; prclose $p`).
+			Prints("haha\n"),
+	)
 }
diff -pruN 0.11+ds1-5/eval/compiler.go 0.12+ds1-1/eval/compiler.go
--- 0.11+ds1-5/eval/compiler.go	2018-01-18 02:03:11.000000000 +0000
+++ 0.12+ds1-1/eval/compiler.go	2018-07-11 02:32:25.000000000 +0000
@@ -4,7 +4,7 @@ package eval
 
 import (
 	"fmt"
-	"strconv"
+	"strings"
 
 	"github.com/elves/elvish/parse"
 	"github.com/elves/elvish/util"
@@ -37,7 +37,7 @@ func (cp *compiler) compiling(n parse.No
 
 func (cp *compiler) errorpf(begin, end int, format string, args ...interface{}) {
 	throw(&CompilationError{fmt.Sprintf(format, args...),
-		*util.NewSourceRange(cp.srcMeta.describePath(), cp.srcMeta.code, begin, end, nil)})
+		*util.NewSourceRange(cp.srcMeta.describePath(), cp.srcMeta.code, begin, end)})
 }
 
 func (cp *compiler) errorf(format string, args ...interface{}) {
@@ -60,7 +60,7 @@ func (cp *compiler) popScope() {
 }
 
 func (cp *compiler) registerVariableGetQname(qname string) bool {
-	_, ns, name := ParseVariable(qname)
+	_, ns, name := ParseVariableRef(qname)
 	return cp.registerVariableGet(ns, name)
 }
 
@@ -68,23 +68,21 @@ func (cp *compiler) registerVariableGet(
 	switch ns {
 	case "", "local", "up":
 		// Handled below
-	case "e", "E", "shared":
+	case "e", "E":
 		return true
 	default:
 		return cp.registerModAccess(ns)
 	}
-	_, err := strconv.Atoi(name)
-	isnum := err == nil
 	// Find in local scope
 	if ns == "" || ns == "local" {
-		if cp.thisScope().has(name) || isnum {
+		if cp.thisScope().has(name) {
 			return true
 		}
 	}
 	// Find in upper scopes
 	if ns == "" || ns == "up" {
 		for i := len(cp.scopes) - 2; i >= 0; i-- {
-			if cp.scopes[i].has(name) || isnum {
+			if cp.scopes[i].has(name) {
 				// Existing name: record capture and return.
 				cp.capture.set(name)
 				return true
@@ -101,7 +99,7 @@ func (cp *compiler) registerVariableGet(
 }
 
 func (cp *compiler) registerVariableSetQname(qname string) bool {
-	_, ns, name := ParseVariable(qname)
+	_, ns, name := ParseVariableRef(qname)
 	return cp.registerVariableSet(ns, name)
 }
 
@@ -138,7 +136,7 @@ func (cp *compiler) registerVariableSet(
 		// New name. Register on this scope!
 		cp.thisScope().set(name)
 		return true
-	case "e", "E", "shared":
+	case "e", "E":
 		// Special namespaces, do nothing
 		return true
 	default:
@@ -147,5 +145,8 @@ func (cp *compiler) registerVariableSet(
 }
 
 func (cp *compiler) registerModAccess(name string) bool {
+	if strings.ContainsRune(name, ':') {
+		name = name[:strings.IndexByte(name, ':')]
+	}
 	return cp.registerVariableGet("", name+NsSuffix)
 }
diff -pruN 0.11+ds1-5/eval/compile_value.go 0.12+ds1-1/eval/compile_value.go
--- 0.11+ds1-5/eval/compile_value.go	2018-01-18 02:03:11.000000000 +0000
+++ 0.12+ds1-1/eval/compile_value.go	2018-07-11 02:32:25.000000000 +0000
@@ -10,30 +10,32 @@ import (
 	"strings"
 	"sync"
 
-	"github.com/elves/elvish/eval/types"
+	"github.com/elves/elvish/eval/vals"
 	"github.com/elves/elvish/glob"
 	"github.com/elves/elvish/parse"
-	"github.com/xiaq/persistent/hashmap"
+	"github.com/elves/elvish/util"
 )
 
 var outputCaptureBufferSize = 16
 
-// ValuesOp is an operation on an EvalCtx that produce Value's.
+// ValuesOp is an operation on an Frame that produce Value's.
 type ValuesOp struct {
-	Func       ValuesOpFunc
+	Body       ValuesOpBody
 	Begin, End int
 }
 
-// ValuesOpFunc is the body of ValuesOp.
-type ValuesOpFunc func(*Frame) []types.Value
+// ValuesOpBody is the body of ValuesOp.
+type ValuesOpBody interface {
+	Invoke(*Frame) ([]interface{}, error)
+}
 
 // Exec executes a ValuesOp and produces Value's.
-func (op ValuesOp) Exec(ec *Frame) []types.Value {
-	ec.begin, ec.end = op.Begin, op.End
-	return op.Func(ec)
+func (op ValuesOp) Exec(fm *Frame) ([]interface{}, error) {
+	fm.begin, fm.end = op.Begin, op.End
+	return op.Body.Invoke(fm)
 }
 
-func (cp *compiler) compound(n *parse.Compound) ValuesOpFunc {
+func (cp *compiler) compound(n *parse.Compound) ValuesOpBody {
 	if len(n.Indexings) == 0 {
 		return literalStr("")
 	}
@@ -44,96 +46,85 @@ func (cp *compiler) compound(n *parse.Co
 	if n.Indexings[0].Head.Type == parse.Tilde {
 		// A lone ~.
 		if len(n.Indexings) == 1 {
-			return func(ec *Frame) []types.Value {
-				return []types.Value{types.String(mustGetHome(""))}
-			}
+			return funcValuesOp(func(fm *Frame) ([]interface{}, error) {
+				home, err := util.GetHome("")
+				if err != nil {
+					return nil, err
+				}
+				return []interface{}{home}, nil
+			})
 		}
 		tilde = true
 		indexings = indexings[1:]
 	}
 
-	ops := cp.indexingOps(indexings)
+	return compoundOp{tilde, cp.indexingOps(indexings)}
+}
 
-	return func(ec *Frame) []types.Value {
-		// Accumulator.
-		vs := ops[0].Exec(ec)
+type compoundOp struct {
+	tilde  bool
+	subops []ValuesOp
+}
 
-		// Logger.Printf("concatenating %v with %d more", vs, len(ops)-1)
+func (op compoundOp) Invoke(fm *Frame) ([]interface{}, error) {
+	// Accumulator.
+	vs, err := op.subops[0].Exec(fm)
+	if err != nil {
+		return nil, err
+	}
 
-		for _, op := range ops[1:] {
-			us := op.Exec(ec)
-			vs = outerProduct(vs, us, cat)
-			// Logger.Printf("with %v => %v", us, vs)
-		}
-		if tilde {
-			newvs := make([]types.Value, len(vs))
-			for i, v := range vs {
-				newvs[i] = doTilde(v)
-			}
-			vs = newvs
+	for _, subop := range op.subops[1:] {
+		us, err := subop.Exec(fm)
+		if err != nil {
+			return nil, err
 		}
-		hasGlob := false
-		for _, v := range vs {
-			if _, ok := v.(GlobPattern); ok {
-				hasGlob = true
-				break
-			}
+		vs, err = outerProduct(vs, us, vals.Concat)
+		if err != nil {
+			return nil, err
 		}
-		if hasGlob {
-			newvs := make([]types.Value, 0, len(vs))
-			for _, v := range vs {
-				if gp, ok := v.(GlobPattern); ok {
-					// Logger.Printf("globbing %v", gp)
-					newvs = append(newvs, doGlob(gp, ec.Interrupts())...)
-				} else {
-					newvs = append(newvs, v)
-				}
-			}
-			vs = newvs
+	}
+	if op.tilde {
+		newvs := make([]interface{}, len(vs))
+		for i, v := range vs {
+			newvs[i] = doTilde(v)
 		}
-		return vs
+		vs = newvs
 	}
-}
-
-func cat(lhs, rhs types.Value) types.Value {
-	switch lhs := lhs.(type) {
-	case types.String:
-		switch rhs := rhs.(type) {
-		case types.String:
-			return lhs + rhs
-		case GlobPattern:
-			segs := stringToSegments(string(lhs))
-			// We know rhs contains exactly one segment.
-			segs = append(segs, rhs.Segments[0])
-			return GlobPattern{glob.Pattern{segs, ""}, rhs.Flags, rhs.Buts}
+	hasGlob := false
+	for _, v := range vs {
+		if _, ok := v.(GlobPattern); ok {
+			hasGlob = true
+			break
 		}
-	case GlobPattern:
-		// NOTE Modifies lhs in place.
-		switch rhs := rhs.(type) {
-		case types.String:
-			lhs.append(stringToSegments(string(rhs))...)
-			return lhs
-		case GlobPattern:
-			// We know rhs contains exactly one segment.
-			lhs.append(rhs.Segments[0])
-			lhs.Flags |= rhs.Flags
-			lhs.Buts = append(lhs.Buts, rhs.Buts...)
-			return lhs
+	}
+	if hasGlob {
+		newvs := make([]interface{}, 0, len(vs))
+		for _, v := range vs {
+			if gp, ok := v.(GlobPattern); ok {
+				// Logger.Printf("globbing %v", gp)
+				newvs = append(newvs, doGlob(gp, fm.Interrupts())...)
+			} else {
+				newvs = append(newvs, v)
+			}
 		}
+		vs = newvs
 	}
-	throw(fmt.Errorf("unsupported concat: %s and %s", lhs.Kind(), rhs.Kind()))
-	panic("unreachable")
+	return vs, nil
 }
 
-func outerProduct(vs []types.Value, us []types.Value, f func(types.Value, types.Value) types.Value) []types.Value {
-	ws := make([]types.Value, len(vs)*len(us))
+func outerProduct(vs []interface{}, us []interface{}, f func(interface{}, interface{}) (interface{}, error)) ([]interface{}, error) {
+	ws := make([]interface{}, len(vs)*len(us))
 	nu := len(us)
 	for i, v := range vs {
 		for j, u := range us {
-			ws[i*nu+j] = f(v, u)
+			var err error
+			ws[i*nu+j], err = f(v, u)
+			if err != nil {
+				return nil, err
+			}
 		}
 	}
-	return ws
+	return ws, nil
 }
 
 // Errors thrown when globbing.
@@ -142,10 +133,10 @@ var (
 	ErrCannotDetermineUsername = errors.New("cannot determine user name from glob pattern")
 )
 
-func doTilde(v types.Value) types.Value {
+func doTilde(v interface{}) interface{} {
 	switch v := v.(type) {
-	case types.String:
-		s := string(v)
+	case string:
+		s := v
 		i := strings.Index(s, "/")
 		var uname, rest string
 		if i == -1 {
@@ -155,7 +146,7 @@ func doTilde(v types.Value) types.Value
 			rest = s[i+1:]
 		}
 		dir := mustGetHome(uname)
-		return types.String(path.Join(dir, rest))
+		return path.Join(dir, rest)
 	case GlobPattern:
 		if len(v.Segments) == 0 {
 			throw(ErrBadGlobPattern)
@@ -179,107 +170,78 @@ func doTilde(v types.Value) types.Value
 		}
 		return v
 	default:
-		throw(fmt.Errorf("tilde doesn't work on value of type %s", v.Kind()))
+		throw(fmt.Errorf("tilde doesn't work on value of type %s", vals.Kind(v)))
 		panic("unreachable")
 	}
 }
 
-func (cp *compiler) array(n *parse.Array) ValuesOpFunc {
-	return catValuesOps(cp.compoundOps(n.Compounds))
+func (cp *compiler) array(n *parse.Array) ValuesOpBody {
+	return seqValuesOp{cp.compoundOps(n.Compounds)}
 }
 
-func catValuesOps(ops []ValuesOp) ValuesOpFunc {
-	return func(ec *Frame) []types.Value {
-		// Use number of compound expressions as an estimation of the number
-		// of values
-		vs := make([]types.Value, 0, len(ops))
-		for _, op := range ops {
-			us := op.Exec(ec)
-			vs = append(vs, us...)
-		}
-		return vs
-	}
-}
-
-func (cp *compiler) indexing(n *parse.Indexing) ValuesOpFunc {
+func (cp *compiler) indexing(n *parse.Indexing) ValuesOpBody {
 	if len(n.Indicies) == 0 {
 		return cp.primary(n.Head)
 	}
 
-	headOp := cp.primaryOp(n.Head)
-	indexOps := cp.arrayOps(n.Indicies)
-
-	return func(ec *Frame) []types.Value {
-		vs := headOp.Exec(ec)
-		for _, indexOp := range indexOps {
-			indicies := indexOp.Exec(ec)
-			newvs := make([]types.Value, 0, len(vs)*len(indicies))
-			for _, v := range vs {
-				newvs = append(newvs, mustIndexer(v, ec).Index(indicies)...)
-			}
-			vs = newvs
-		}
-		return vs
-	}
-}
-
-func literalValues(v ...types.Value) ValuesOpFunc {
-	return func(e *Frame) []types.Value {
-		return v
-	}
+	return &indexingOp{cp.primaryOp(n.Head), cp.arrayOps(n.Indicies)}
 }
 
-func literalStr(text string) ValuesOpFunc {
-	return literalValues(types.String(text))
+type indexingOp struct {
+	headOp   ValuesOp
+	indexOps []ValuesOp
 }
 
-func variable(qname string) ValuesOpFunc {
-	explode, ns, name := ParseVariable(qname)
-	return func(ec *Frame) []types.Value {
-		variable := ec.ResolveVar(ns, name)
-		if variable == nil {
-			throwf("variable $%s not found", qname)
+func (op *indexingOp) Invoke(fm *Frame) ([]interface{}, error) {
+	vs, err := op.headOp.Exec(fm)
+	if err != nil {
+		return nil, err
+	}
+	for _, indexOp := range op.indexOps {
+		indicies, err := indexOp.Exec(fm)
+		if err != nil {
+			return nil, err
 		}
-		value := variable.Get()
-		if explode {
-			iterator, ok := value.(types.Iterator)
-			if !ok {
-				// Use qname[1:] to skip the leading "@"
-				throwf("variable $%s (kind %s) cannot be exploded", qname[1:], value.Kind())
+		newvs := make([]interface{}, 0, len(vs)*len(indicies))
+		for _, v := range vs {
+			for _, index := range indicies {
+				result, err := vals.Index(v, index)
+				if err != nil {
+					return nil, err
+				}
+				newvs = append(newvs, result)
 			}
-			return types.CollectFromIterator(iterator)
 		}
-		return []types.Value{value}
+		vs = newvs
 	}
+	return vs, nil
 }
 
-func (cp *compiler) primary(n *parse.Primary) ValuesOpFunc {
+func (cp *compiler) primary(n *parse.Primary) ValuesOpBody {
 	switch n.Type {
 	case parse.Bareword, parse.SingleQuoted, parse.DoubleQuoted:
 		return literalStr(n.Value)
 	case parse.Variable:
-		qname := n.Value
-		if !cp.registerVariableGetQname(qname) {
+		explode, ns, name := ParseVariableRef(n.Value)
+		if !cp.registerVariableGet(ns, name) {
 			cp.errorf("variable $%s not found", n.Value)
 		}
-		return variable(qname)
+		return &variableOp{explode, ns, name}
 	case parse.Wildcard:
 		seg, err := wildcardToSegment(n.SourceText())
 		if err != nil {
 			cp.errorf("%s", err)
 		}
-		vs := []types.Value{
+		vs := []interface{}{
 			GlobPattern{glob.Pattern{[]glob.Segment{seg}, ""}, 0, nil}}
-		return func(ec *Frame) []types.Value {
-			return vs
-		}
+		return literalValues(vs...)
 	case parse.Tilde:
 		cp.errorf("compiler bug: Tilde not handled in .compound")
 		return literalStr("~")
 	case parse.ExceptionCapture:
-		return cp.exceptionCapture(n.Chunk)
+		return exceptionCaptureOp{cp.chunkOp(n.Chunk)}
 	case parse.OutputCapture:
-		return cp.outputCapture(n)
+		return outputCaptureOp{cp.chunkOp(n.Chunk)}
 	case parse.List:
 		return cp.list(n)
 	case parse.Lambda:
@@ -294,43 +256,64 @@ func (cp *compiler) primary(n *parse.Pri
 	}
 }
 
-func (cp *compiler) list(n *parse.Primary) ValuesOpFunc {
-	// TODO(xiaq): Use Vector.Cons to build the list, instead of building a
-	// slice and converting to Vector.
-	op := catValuesOps(cp.compoundOps(n.Elements))
-	return func(ec *Frame) []types.Value {
-		return []types.Value{types.MakeList(op(ec)...)}
+type variableOp struct {
+	explode bool
+	ns      string
+	name    string
+}
+
+func (op variableOp) Invoke(fm *Frame) ([]interface{}, error) {
+	variable := fm.ResolveVar(op.ns, op.name)
+	if variable == nil {
+		return nil, fmt.Errorf("variable $%s:%s not found", op.ns, op.name)
 	}
+	value := variable.Get()
+	if op.explode {
+		return vals.Collect(value)
+	}
+	return []interface{}{value}, nil
+}
+
+func (cp *compiler) list(n *parse.Primary) ValuesOpBody {
+	return listOp{cp.compoundOps(n.Elements)}
 }
 
-func (cp *compiler) exceptionCapture(n *parse.Chunk) ValuesOpFunc {
-	op := cp.chunkOp(n)
-	return func(ec *Frame) []types.Value {
-		err := ec.PEval(op)
-		if err == nil {
-			return []types.Value{OK}
+type listOp struct{ subops []ValuesOp }
+
+func (op listOp) Invoke(fm *Frame) ([]interface{}, error) {
+	list := vals.EmptyList
+	for _, subop := range op.subops {
+		moreValues, err := subop.Exec(fm)
+		if err != nil {
+			return nil, err
+		}
+		for _, moreValue := range moreValues {
+			list = list.Cons(moreValue)
 		}
-		return []types.Value{err.(*Exception)}
 	}
+	return []interface{}{list}, nil
 }
 
-func (cp *compiler) outputCapture(n *parse.Primary) ValuesOpFunc {
-	op := cp.chunkOp(n.Chunk)
-	return func(ec *Frame) []types.Value {
-		return captureOutput(ec, op)
+type exceptionCaptureOp struct{ subop Op }
+
+func (op exceptionCaptureOp) Invoke(fm *Frame) ([]interface{}, error) {
+	err := fm.Eval(op.subop)
+	if err == nil {
+		return []interface{}{OK}, nil
 	}
+	return []interface{}{err.(*Exception)}, nil
 }
 
-func captureOutput(ec *Frame, op Op) []types.Value {
-	vs, err := pcaptureOutput(ec, op)
-	maybeThrow(err)
-	return vs
+type outputCaptureOp struct{ subop Op }
+
+func (op outputCaptureOp) Invoke(fm *Frame) ([]interface{}, error) {
+	return pcaptureOutput(fm, op.subop)
 }
 
-func pcaptureOutput(ec *Frame, op Op) ([]types.Value, error) {
-	vs := []types.Value{}
+func pcaptureOutput(fm *Frame, op Op) ([]interface{}, error) {
+	vs := []interface{}{}
 	var m sync.Mutex
-	valueCb := func(ch <-chan types.Value) {
+	valueCb := func(ch <-chan interface{}) {
 		for v := range ch {
 			m.Lock()
 			vs = append(vs, v)
@@ -342,7 +325,7 @@ func pcaptureOutput(ec *Frame, op Op) ([
 		for {
 			line, err := buffered.ReadString('\n')
 			if line != "" {
-				v := types.String(strings.TrimSuffix(line, "\n"))
+				v := strings.TrimSuffix(line, "\n")
 				m.Lock()
 				vs = append(vs, v)
 				m.Unlock()
@@ -356,20 +339,20 @@ func pcaptureOutput(ec *Frame, op Op) ([
 		}
 	}
 
-	err := pcaptureOutputInner(ec, op, valueCb, bytesCb)
+	err := pcaptureOutputInner(fm, op, valueCb, bytesCb)
 	return vs, err
 }
 
-func pcaptureOutputInner(ec *Frame, op Op, valuesCb func(<-chan types.Value), bytesCb func(*os.File)) error {
+func pcaptureOutputInner(fm *Frame, op Op, valuesCb func(<-chan interface{}), bytesCb func(*os.File)) error {
 
-	newEc := ec.fork("[output capture]")
+	newFm := fm.fork("[output capture]")
 
-	ch := make(chan types.Value, outputCaptureBufferSize)
+	ch := make(chan interface{}, outputCaptureBufferSize)
 	pipeRead, pipeWrite, err := os.Pipe()
 	if err != nil {
 		return fmt.Errorf("failed to create pipe: %v", err)
 	}
-	newEc.ports[1] = &Port{
+	newFm.ports[1] = &Port{
 		Chan: ch, CloseChan: true,
 		File: pipeWrite, CloseFile: true,
 	}
@@ -387,16 +370,16 @@ func pcaptureOutputInner(ec *Frame, op O
 		close(bytesCollected)
 	}()
 
-	err = newEc.PEval(op)
+	err = newFm.Eval(op)
 
-	ClosePorts(newEc.ports)
+	newFm.Close()
 	<-bytesCollected
 	<-chCollected
 
 	return err
 }
 
-func (cp *compiler) lambda(n *parse.Primary) ValuesOpFunc {
+func (cp *compiler) lambda(n *parse.Primary) ValuesOpBody {
 	// Parse signature.
 	var (
 		argNames      []string
@@ -409,7 +392,7 @@ func (cp *compiler) lambda(n *parse.Prim
 		argNames = make([]string, len(n.Elements))
 		for i, arg := range n.Elements {
 			qname := mustString(cp, arg, "argument name must be literal string")
-			explode, ns, name := ParseVariable(qname)
+			explode, ns, name := ParseVariableRef(qname)
 			if ns != "" {
 				cp.errorpf(arg.Begin(), arg.End(), "argument name must be unqualified")
 			}
@@ -432,7 +415,7 @@ func (cp *compiler) lambda(n *parse.Prim
 		optDefaultOps = make([]ValuesOp, len(n.MapPairs))
 		for i, opt := range n.MapPairs {
 			qname := mustString(cp, opt.Key, "option name must be literal string")
-			_, ns, name := ParseVariable(qname)
+			_, ns, name := ParseVariableRef(qname)
 			if ns != "" {
 				cp.errorpf(opt.Key.Begin(), opt.Key.End(), "option name must be unqualified")
 			}
@@ -458,9 +441,8 @@ func (cp *compiler) lambda(n *parse.Prim
 	for _, optName := range optNames {
 		thisScope.set(optName)
 	}
-	thisScope.set("opts")
 
-	op := cp.chunkOp(n.Chunk)
+	subop := cp.chunkOp(n.Chunk)
 
 	// XXX The fiddlings with cp.capture is error-prone.
 	capture := cp.capture
@@ -471,28 +453,40 @@ func (cp *compiler) lambda(n *parse.Prim
 		cp.registerVariableGetQname(name)
 	}
 
-	srcMeta := cp.srcMeta
+	return &lambdaOp{argNames, restArgName, optNames, optDefaultOps, capture, subop, cp.srcMeta, n.Begin(), n.End()}
+}
 
-	return func(ec *Frame) []types.Value {
-		evCapture := make(Ns)
-		for name := range capture {
-			evCapture[name] = ec.ResolveVar("", name)
-		}
-		optDefaults := make([]types.Value, len(optDefaultOps))
-		for i, op := range optDefaultOps {
-			defaultValue := ec.ExecAndUnwrap("option default value", op).One().Any()
-			optDefaults[i] = defaultValue
-		}
-		// XXX(xiaq): Capture uses.
-		return []types.Value{&Closure{argNames, restArgName, optNames, optDefaults, op, evCapture, srcMeta}}
+type lambdaOp struct {
+	argNames      []string
+	restArgName   string
+	optNames      []string
+	optDefaultOps []ValuesOp
+	capture       staticNs
+	subop         Op
+	srcMeta       *Source
+	defBegin      int
+	defEnd        int
+}
+
+func (op *lambdaOp) Invoke(fm *Frame) ([]interface{}, error) {
+	evCapture := make(Ns)
+	for name := range op.capture {
+		evCapture[name] = fm.ResolveVar("", name)
+	}
+	optDefaults := make([]interface{}, len(op.optDefaultOps))
+	for i, op := range op.optDefaultOps {
+		defaultValue := fm.ExecAndUnwrap("option default value", op).One().Any()
+		optDefaults[i] = defaultValue
 	}
+	// XXX(xiaq): Capture uses.
+	return []interface{}{&Closure{op.argNames, op.restArgName, op.optNames, optDefaults, op.subop, evCapture, op.srcMeta, op.defBegin, op.defEnd}}, nil
 }
 
-func (cp *compiler) map_(n *parse.Primary) ValuesOpFunc {
+func (cp *compiler) map_(n *parse.Primary) ValuesOpBody {
 	return cp.mapPairs(n.MapPairs)
 }
 
-func (cp *compiler) mapPairs(pairs []*parse.MapPair) ValuesOpFunc {
+func (cp *compiler) mapPairs(pairs []*parse.MapPair) ValuesOpBody {
 	npairs := len(pairs)
 	keysOps := make([]ValuesOp, npairs)
 	valuesOps := make([]ValuesOp, npairs)
@@ -501,32 +495,79 @@ func (cp *compiler) mapPairs(pairs []*pa
 		keysOps[i] = cp.compoundOp(pair.Key)
 		if pair.Value == nil {
 			p := pair.End()
-			valuesOps[i] = ValuesOp{literalValues(types.Bool(true)), p, p}
+			valuesOps[i] = ValuesOp{literalValues(true), p, p}
 		} else {
 			valuesOps[i] = cp.compoundOp(pairs[i].Value)
 		}
 		begins[i], ends[i] = pair.Begin(), pair.End()
 	}
-	return func(ec *Frame) []types.Value {
-		m := hashmap.Empty
-		for i := 0; i < npairs; i++ {
-			keys := keysOps[i].Exec(ec)
-			values := valuesOps[i].Exec(ec)
-			if len(keys) != len(values) {
-				ec.errorpf(begins[i], ends[i],
-					"%d keys but %d values", len(keys), len(values))
-			}
-			for j, key := range keys {
-				m = m.Assoc(key, values[j])
-			}
+	return &mapPairsOp{keysOps, valuesOps, begins, ends}
+}
+
+type mapPairsOp struct {
+	keysOps   []ValuesOp
+	valuesOps []ValuesOp
+	begins    []int
+	ends      []int
+}
+
+func (op *mapPairsOp) Invoke(fm *Frame) ([]interface{}, error) {
+	m := vals.EmptyMap
+	for i := range op.keysOps {
+		keys, err := op.keysOps[i].Exec(fm)
+		if err != nil {
+			return nil, err
+		}
+		values, err := op.valuesOps[i].Exec(fm)
+		if err != nil {
+			return nil, err
+		}
+		if len(keys) != len(values) {
+			fm.errorpf(op.begins[i], op.ends[i],
+				"%d keys but %d values", len(keys), len(values))
+		}
+		for j, key := range keys {
+			m = m.Assoc(key, values[j])
 		}
-		return []types.Value{types.NewMap(m)}
 	}
+	return []interface{}{m}, nil
 }
 
-func (cp *compiler) braced(n *parse.Primary) ValuesOpFunc {
+func (cp *compiler) braced(n *parse.Primary) ValuesOpBody {
 	ops := cp.compoundOps(n.Braced)
 	// TODO: n.IsRange
 	// isRange := n.IsRange
-	return catValuesOps(ops)
+	return seqValuesOp{ops}
+}
+
+type literalValuesOp struct{ values []interface{} }
+
+func (op literalValuesOp) Invoke(*Frame) ([]interface{}, error) {
+	return op.values, nil
+}
+
+func literalValues(v ...interface{}) ValuesOpBody {
+	return literalValuesOp{v}
 }
+
+func literalStr(text string) ValuesOpBody {
+	return literalValues(text)
+}
+
+type seqValuesOp struct{ subops []ValuesOp }
+
+func (op seqValuesOp) Invoke(fm *Frame) ([]interface{}, error) {
+	var values []interface{}
+	for _, subop := range op.subops {
+		moreValues, err := subop.Exec(fm)
+		if err != nil {
+			return nil, err
+		}
+		values = append(values, moreValues...)
+	}
+	return values, nil
+}
+
+type funcValuesOp func(*Frame) ([]interface{}, error)
+
+func (op funcValuesOp) Invoke(fm *Frame) ([]interface{}, error) { return op(fm) }
diff -pruN 0.11+ds1-5/eval/compile_value_test.go 0.12+ds1-1/eval/compile_value_test.go
--- 0.11+ds1-5/eval/compile_value_test.go	2018-01-18 02:03:11.000000000 +0000
+++ 0.12+ds1-1/eval/compile_value_test.go	2018-07-11 02:32:25.000000000 +0000
@@ -1,105 +1,161 @@
 package eval
 
 import (
+	"sort"
+	"strings"
 	"testing"
 
-	"github.com/elves/elvish/eval/types"
+	"github.com/elves/elvish/eval/vals"
+	"github.com/elves/elvish/util"
 )
 
-var valueTests = []Test{
-	// Compounding
-	// -----------
-	{"put {fi,elvi}sh{1.0,1.1}",
-		want{out: strs("fish1.0", "fish1.1", "elvish1.0", "elvish1.1")}},
-
-	// List, Map and Indexing
-	// ----------------------
-
-	{"echo [a b c] [&key=value] | each $put~",
-		want{out: strs("[a b c] [&key=value]")}},
-	{"put [a b c][2]", want{out: strs("c")}},
-	{"put [&key=value][key]", want{out: strs("value")}},
-
-	// String Literals
-	// ---------------
-	{`put 'such \"''literal'`, want{out: strs(`such \"'literal`)}},
-	{`put "much \n\033[31;1m$cool\033[m"`,
-		want{out: strs("much \n\033[31;1m$cool\033[m")}},
-
-	// Captures
-	// ---------
-
-	// Output capture
-	{"put (put lorem ipsum)", want{out: strs("lorem", "ipsum")}},
-	{"put (print \"lorem\nipsum\")", want{out: strs("lorem", "ipsum")}},
-
-	// Exception capture
-	{"bool ?(nop); bool ?(e:false)", want{out: bools(true, false)}},
-
-	// Variable Use
-	// ------------
-
-	// Compounding
-	{"x='SHELL'\nput 'WOW, SUCH '$x', MUCH COOL'\n",
-		want{out: strs("WOW, SUCH SHELL, MUCH COOL")}},
-	// Splicing
-	{"x=[elvish rules]; put $@x", want{out: strs("elvish", "rules")}},
-
-	// Wildcard; see testmain_test.go for FS setup
-	// -------------------------------------------
-
-	{"put *", want{out: strs(fileListing...)}},
-	{"put a/b/nonexistent*", want{err: ErrWildcardNoMatch}},
-	{"put a/b/nonexistent*[nomatch-ok]", wantNothing},
-
-	// Character set and range
-	{"put ?[set:ab]*", want{out: strs(getFilesWithPrefix("a", "b")...)}},
-	{"put ?[range:a-c]*", want{out: strs(getFilesWithPrefix("a", "b", "c")...)}},
-	{"put ?[range:a~c]*", want{out: strs(getFilesWithPrefix("a", "b")...)}},
-	{"put *[range:a-z]", want{out: strs("bar", "dir", "foo", "ipsum", "lorem")}},
-
-	// Exclusion
-	{"put *[but:foo but:lorem]", want{out: strs(getFilesBut("foo", "lorem")...)}},
-
-	// Tilde
-	// -----
-	{"h=$E:HOME; E:HOME=/foo; put ~ ~/src; E:HOME=$h",
-		want{out: strs("/foo", "/foo/src")}},
-
-	// Closure
-	// -------
-
-	{"[]{ }", wantNothing},
-	{"[x]{put $x} foo", want{out: strs("foo")}},
-
-	// Variable capture
-	{"x=lorem; []{x=ipsum}; put $x", want{out: strs("ipsum")}},
-	{"x=lorem; []{ put $x; x=ipsum }; put $x",
-		want{out: strs("lorem", "ipsum")}},
-
-	// Shadowing
-	{"x=ipsum; []{ local:x=lorem; put $x }; put $x",
-		want{out: strs("lorem", "ipsum")}},
-
-	// Shadowing by argument
-	{"x=ipsum; [x]{ put $x; x=BAD } lorem; put $x",
-		want{out: strs("lorem", "ipsum")}},
-
-	// Closure captures new local variables every time
-	{`fn f []{ x=0; put []{x=(+ $x 1)} []{put $x} }
-		      {inc1,put1}=(f); $put1; $inc1; $put1
-			  {inc2,put2}=(f); $put2; $inc2; $put2`,
-		want{out: strs("0", "1", "0", "1")}},
-
-	// Rest argument.
-	{"[x @xs]{ put $x $xs } a b c",
-		want{out: []types.Value{types.String("a"), types.MakeList(types.String("b"), types.String("c"))}}},
-	// Options.
-	{"[a &k=v]{ put $a $k } foo &k=bar", want{out: strs("foo", "bar")}},
-	// Option default value.
-	{"[a &k=v]{ put $a $k } foo", want{out: strs("foo", "v")}},
+func TestValue(t *testing.T) {
+	Test(t,
+		// Compounding
+		// -----------
+		That("put {fi,elvi}sh{1.0,1.1}").Puts(
+			"fish1.0", "fish1.1", "elvish1.0", "elvish1.1"),
+
+		// List, Map and Indexing
+		// ----------------------
+
+		That("echo [a b c] [&key=value] | each $put~").
+			Puts("[a b c] [&key=value]"),
+		That("put [a b c][2]").Puts("c"),
+		That("put [&key=value][key]").Puts("value"),
+
+		// String Literals
+		// ---------------
+		That(`put 'such \"''literal'`).Puts(`such \"'literal`),
+		That(`put "much \n\033[31;1m$cool\033[m"`).
+			Puts("much \n\033[31;1m$cool\033[m"),
+
+		// Captures
+		// ---------
+
+		// Output capture
+		That("put (put lorem ipsum)").Puts("lorem", "ipsum"),
+		That("put (print \"lorem\nipsum\")").Puts("lorem", "ipsum"),
+
+		// Exception capture
+		That("bool ?(nop); bool ?(e:false)").Puts(true, false),
+
+		// Variable Use
+		// ------------
+
+		// Compounding
+		That("x='SHELL'\nput 'WOW, SUCH '$x', MUCH COOL'\n").
+			Puts("WOW, SUCH SHELL, MUCH COOL"),
+		// Splicing
+		That("x=[elvish rules]; put $@x").Puts("elvish", "rules"),
+
+		// Tilde
+		// -----
+		That("h=$E:HOME; E:HOME=/foo; put ~ ~/src; E:HOME=$h").Puts("/foo", "/foo/src"),
+
+		// Closure
+		// -------
+
+		That("[]{ }").DoesNothing(),
+		That("[x]{put $x} foo").Puts("foo"),
+
+		// Variable capture
+		That("x=lorem; []{x=ipsum}; put $x").Puts("ipsum"),
+		That("x=lorem; []{ put $x; x=ipsum }; put $x").Puts("lorem", "ipsum"),
+
+		// Shadowing
+		That("x=ipsum; []{ local:x=lorem; put $x }; put $x").Puts("lorem", "ipsum"),
+
+		// Shadowing by argument
+		That("x=ipsum; [x]{ put $x; x=BAD } lorem; put $x").Puts("lorem", "ipsum"),
+
+		// Closure captures new local variables every time
+		That(`fn f []{ x=0; put []{x=(+ $x 1)} []{put $x} }
+		  {inc1,put1}=(f); $put1; $inc1; $put1
+		  {inc2,put2}=(f); $put2; $inc2; $put2`).Puts("0", "1", "0", "1"),
+
+		// Rest argument.
+		That("[x @xs]{ put $x $xs } a b c").Puts("a", vals.MakeList("b", "c")),
+		// Options.
+		That("[a &k=v]{ put $a $k } foo &k=bar").Puts("foo", "bar"),
+		// Option default value.
+		That("[a &k=v]{ put $a $k } foo").Puts("foo", "v"),
+	)
 }
 
-func TestValue(t *testing.T) {
-	runTests(t, valueTests)
+var (
+	filesToCreate = []string{
+		"a1", "a2", "a3", "a10",
+		"b1", "b2", "b3",
+		"c1", "c2",
+		"foo", "bar", "lorem", "ipsum",
+	}
+	dirsToCreate = []string{"dir", "dir2"}
+	fileListing  = getFileListing()
+)
+
+func getFileListing() []string {
+	var x []string
+	x = append(x, filesToCreate...)
+	x = append(x, dirsToCreate...)
+	sort.Strings(x)
+	return x
+}
+
+func getFilesWithPrefix(prefixes ...string) []string {
+	var x []string
+	for _, name := range fileListing {
+		for _, prefix := range prefixes {
+			if strings.HasPrefix(name, prefix) {
+				x = append(x, name)
+				break
+			}
+		}
+	}
+	sort.Strings(x)
+	return x
+}
+
+func getFilesBut(excludes ...string) []string {
+	var x []string
+	for _, name := range fileListing {
+		excluded := false
+		for _, exclude := range excludes {
+			if name == exclude {
+				excluded = true
+				break
+			}
+		}
+		if !excluded {
+			x = append(x, name)
+		}
+	}
+	sort.Strings(x)
+	return x
+}
+
+func TestWildcard(t *testing.T) {
+	util.InTempDir(func(string) {
+		for _, filename := range filesToCreate {
+			MustCreateEmpty(filename)
+		}
+		for _, dirname := range dirsToCreate {
+			MustMkdirAll(dirname, 0700)
+		}
+
+		Test(t,
+			That("put *").PutsStrings(fileListing),
+			That("put a/b/nonexistent*").ErrorsWith(ErrWildcardNoMatch),
+			That("put a/b/nonexistent*[nomatch-ok]").DoesNothing(),
+
+			// Character set and range
+			That("put ?[set:ab]*").PutsStrings(getFilesWithPrefix("a", "b")),
+			That("put ?[range:a-c]*").PutsStrings(getFilesWithPrefix("a", "b", "c")),
+			That("put ?[range:a~c]*").PutsStrings(getFilesWithPrefix("a", "b")),
+			That("put *[range:a-z]").Puts("bar", "dir", "foo", "ipsum", "lorem"),
+
+			// Exclusion
+			That("put *[but:foo][but:lorem]").PutsStrings(getFilesBut("foo", "lorem")),
+		)
+	})
 }
diff -pruN 0.11+ds1-5/eval/conversion.go 0.12+ds1-1/eval/conversion.go
--- 0.11+ds1-5/eval/conversion.go	2018-01-18 02:03:11.000000000 +0000
+++ 0.12+ds1-1/eval/conversion.go	1970-01-01 00:00:00.000000000 +0000
@@ -1,92 +0,0 @@
-package eval
-
-import (
-	"fmt"
-	"reflect"
-	"strconv"
-	"unicode/utf8"
-
-	"github.com/elves/elvish/eval/types"
-)
-
-// Conversion between Go value and Value.
-
-func toFloat(arg types.Value) (float64, error) {
-	if _, ok := arg.(types.String); !ok {
-		return 0, fmt.Errorf("must be string")
-	}
-	s := string(arg.(types.String))
-	num, err := strconv.ParseFloat(s, 64)
-	if err != nil {
-		num, err2 := strconv.ParseInt(s, 0, 64)
-		if err2 != nil {
-			return 0, err
-		}
-		return float64(num), nil
-	}
-	return num, nil
-}
-
-func floatToString(f float64) types.String {
-	return types.String(strconv.FormatFloat(f, 'g', -1, 64))
-}
-
-func toInt(arg types.Value) (int, error) {
-	arg, ok := arg.(types.String)
-	if !ok {
-		return 0, fmt.Errorf("must be string")
-	}
-	num, err := strconv.ParseInt(string(arg.(types.String)), 0, 0)
-	if err != nil {
-		return 0, err
-	}
-	return int(num), nil
-}
-
-func toRune(arg types.Value) (rune, error) {
-	ss, ok := arg.(types.String)
-	if !ok {
-		return -1, fmt.Errorf("must be string")
-	}
-	s := string(ss)
-	r, size := utf8.DecodeRuneInString(s)
-	if r == utf8.RuneError {
-		return -1, fmt.Errorf("string is not valid UTF-8")
-	}
-	if size != len(s) {
-		return -1, fmt.Errorf("string has multiple runes")
-	}
-	return r, nil
-}
-
-// scanValueToGo converts Value to Go data, depending on the type of the
-// destination.
-func scanValueToGo(src types.Value, dstPtr interface{}) {
-	switch dstPtr := dstPtr.(type) {
-	case *string:
-		s, ok := src.(types.String)
-		if !ok {
-			throwf("cannot convert %T to string", src)
-		}
-		*dstPtr = string(s)
-	case *int:
-		i, err := toInt(src)
-		maybeThrow(err)
-		*dstPtr = i
-	case *float64:
-		f, err := toFloat(src)
-		maybeThrow(err)
-		*dstPtr = f
-	default:
-		ptr := reflect.ValueOf(dstPtr)
-		if ptr.Kind() != reflect.Ptr {
-			throwf("internal bug: %T to ScanArgs, need pointer", dstPtr)
-		}
-		dstReflect := reflect.Indirect(ptr)
-		if reflect.TypeOf(src).ConvertibleTo(dstReflect.Type()) {
-			dstReflect.Set(reflect.ValueOf(src).Convert(dstReflect.Type()))
-		} else {
-			throwf("need %s argument, got %s", dstReflect.Type().Name(), src.Kind())
-		}
-	}
-}
diff -pruN 0.11+ds1-5/eval/daemon/daemon.go 0.12+ds1-1/eval/daemon/daemon.go
--- 0.11+ds1-5/eval/daemon/daemon.go	2018-01-18 02:03:11.000000000 +0000
+++ 0.12+ds1-1/eval/daemon/daemon.go	2018-07-11 02:32:25.000000000 +0000
@@ -7,8 +7,7 @@ import (
 
 	"github.com/elves/elvish/daemon"
 	"github.com/elves/elvish/eval"
-	"github.com/elves/elvish/eval/types"
-	"github.com/elves/elvish/eval/vartypes"
+	"github.com/elves/elvish/eval/vars"
 	daemonp "github.com/elves/elvish/program/daemon"
 	"github.com/elves/elvish/util"
 )
@@ -20,29 +19,23 @@ var errDontKnowHowToSpawnDaemon = errors
 // Ns makes the daemon: namespace.
 func Ns(daemon *daemon.Client, spawner *daemonp.Daemon) eval.Ns {
 	// Obtain process ID
-	daemonPid := func() types.Value {
+	getPid := func() interface{} {
 		pid, err := daemon.Pid()
 		if err != nil {
 			util.Throw(err)
 		}
-		return types.String(strconv.Itoa(pid))
+		return string(strconv.Itoa(pid))
 	}
 
-	daemonSpawn := func(ec *eval.Frame, args []types.Value, opts map[string]types.Value) {
-		eval.TakeNoArg(args)
-		eval.TakeNoOpt(opts)
+	spawn := func() error {
 		if spawner == nil {
-			util.Throw(errDontKnowHowToSpawnDaemon)
-		}
-		err := spawner.Spawn()
-		if err != nil {
-			util.Throw(err)
+			return errDontKnowHowToSpawnDaemon
 		}
+		return spawner.Spawn()
 	}
-	return eval.Ns{
-		"pid":  vartypes.NewRoCallback(daemonPid),
-		"sock": vartypes.NewRo(types.String(daemon.SockPath())),
 
-		"spawn" + eval.FnSuffix: vartypes.NewRo(&eval.BuiltinFn{"daemon:spawn", daemonSpawn}),
-	}
+	return eval.Ns{
+		"pid":  vars.FromGet(getPid),
+		"sock": vars.NewRo(string(daemon.SockPath())),
+	}.AddBuiltinFn("daemon:", "spawn", spawn)
 }
diff -pruN 0.11+ds1-5/eval/editor.go 0.12+ds1-1/eval/editor.go
--- 0.11+ds1-5/eval/editor.go	2018-01-18 02:03:11.000000000 +0000
+++ 0.12+ds1-1/eval/editor.go	2018-07-11 02:32:25.000000000 +0000
@@ -1,11 +1,7 @@
 package eval
 
-import "sync"
-
 // Editor is the interface that the line editor has to satisfy. It is needed so
 // that this package does not depend on the edit package.
 type Editor interface {
-	Active() bool
-	ActiveMutex() *sync.Mutex
 	Notify(string, ...interface{})
 }
diff -pruN 0.11+ds1-5/eval/env_list.go 0.12+ds1-1/eval/env_list.go
--- 0.11+ds1-5/eval/env_list.go	2018-01-18 02:03:11.000000000 +0000
+++ 0.12+ds1-1/eval/env_list.go	2018-07-11 02:32:25.000000000 +0000
@@ -6,8 +6,9 @@ import (
 	"strings"
 	"sync"
 
-	"github.com/elves/elvish/eval/types"
-	"github.com/elves/elvish/eval/vartypes"
+	"github.com/elves/elvish/eval/vals"
+	"github.com/elves/elvish/eval/vars"
+	"github.com/elves/elvish/util"
 	"github.com/xiaq/persistent/vector"
 )
 
@@ -32,15 +33,15 @@ type EnvList struct {
 	sync.RWMutex
 	envName    string
 	cacheFor   string
-	cacheValue types.Value
+	cacheValue interface{}
 }
 
 var (
-	_ vartypes.Variable = (*EnvList)(nil)
+	_ vars.Var = (*EnvList)(nil)
 )
 
 // Get returns a Value for an EnvPathList.
-func (envli *EnvList) Get() types.Value {
+func (envli *EnvList) Get() interface{} {
 	envli.Lock()
 	defer envli.Unlock()
 
@@ -51,37 +52,35 @@ func (envli *EnvList) Get() types.Value
 	envli.cacheFor = value
 	v := vector.Empty
 	for _, path := range strings.Split(value, pathListSeparator) {
-		v = v.Cons(types.String(path))
+		v = v.Cons(path)
 	}
-	envli.cacheValue = types.NewList(v)
+	envli.cacheValue = v
 	return envli.cacheValue
 }
 
 // Set sets an EnvPathList. The underlying environment variable is set.
-func (envli *EnvList) Set(v types.Value) error {
-	iterator, ok := v.(types.Iterator)
-	if !ok {
-		return ErrCanOnlyAssignList
-	}
-	var paths []string
-	var err error
-	iterator.Iterate(func(v types.Value) bool {
-		s, ok := v.(types.String)
+func (envli *EnvList) Set(v interface{}) error {
+	var (
+		paths      []string
+		errElement error
+	)
+	errIterate := vals.Iterate(v, func(v interface{}) bool {
+		s, ok := v.(string)
 		if !ok {
-			err = ErrPathMustBeString
+			errElement = ErrPathMustBeString
 			return false
 		}
-		path := string(s)
+		path := s
 		if strings.ContainsAny(path, forbiddenInPath) {
-			err = ErrPathCannotContainColonZero
+			errElement = ErrPathCannotContainColonZero
 			return false
 		}
-		paths = append(paths, string(s))
+		paths = append(paths, s)
 		return true
 	})
 
-	if err != nil {
-		return err
+	if errElement != nil || errIterate != nil {
+		return util.Errors(errElement, errIterate)
 	}
 
 	envli.Lock()
diff -pruN 0.11+ds1-5/eval/evaler_ports.go 0.12+ds1-1/eval/evaler_ports.go
--- 0.11+ds1-5/eval/evaler_ports.go	2018-01-18 02:03:11.000000000 +0000
+++ 0.12+ds1-1/eval/evaler_ports.go	1970-01-01 00:00:00.000000000 +0000
@@ -1,52 +0,0 @@
-package eval
-
-import (
-	"os"
-	"sync"
-
-	"github.com/elves/elvish/eval/types"
-)
-
-const (
-	stdoutChanSize = 32
-	stderrChanSize = 32
-)
-
-type evalerPorts struct {
-	ports      [3]*Port
-	relayeWait *sync.WaitGroup
-}
-
-func newEvalerPorts(stdin, stdout, stderr *os.File, prefix *string) evalerPorts {
-	stdoutChan := make(chan types.Value, stdoutChanSize)
-	stderrChan := make(chan types.Value, stderrChanSize)
-
-	var relayerWait sync.WaitGroup
-	relayerWait.Add(2)
-	go relayChanToFile(stdoutChan, stdout, prefix, &relayerWait)
-	go relayChanToFile(stderrChan, stderr, prefix, &relayerWait)
-
-	return evalerPorts{
-		[3]*Port{
-			{File: stdin, Chan: ClosedChan},
-			{File: stdout, Chan: stdoutChan, CloseChan: true},
-			{File: stderr, Chan: stderrChan, CloseChan: true},
-		},
-		&relayerWait,
-	}
-}
-
-func relayChanToFile(ch <-chan types.Value, file *os.File, prefix *string, w *sync.WaitGroup) {
-	for v := range ch {
-		file.WriteString(*prefix)
-		file.WriteString(v.Repr(initIndent))
-		file.WriteString("\n")
-	}
-	w.Done()
-}
-
-func (ep *evalerPorts) close() {
-	ep.ports[1].Close()
-	ep.ports[2].Close()
-	ep.relayeWait.Wait()
-}
diff -pruN 0.11+ds1-5/eval/evaler_ports_test.go 0.12+ds1-1/eval/evaler_ports_test.go
--- 0.11+ds1-5/eval/evaler_ports_test.go	2018-01-18 02:03:11.000000000 +0000
+++ 0.12+ds1-1/eval/evaler_ports_test.go	1970-01-01 00:00:00.000000000 +0000
@@ -1,55 +0,0 @@
-package eval
-
-import (
-	"io"
-	"io/ioutil"
-	"os"
-	"testing"
-
-	"github.com/elves/elvish/eval/types"
-)
-
-func TestEvalerPorts(t *testing.T) {
-	stdoutReader, stdout := mustPipe()
-	defer stdoutReader.Close()
-
-	stderrReader, stderr := mustPipe()
-	defer stderrReader.Close()
-
-	prefix := "> "
-	ep := newEvalerPorts(DevNull, stdout, stderr, &prefix)
-	ep.ports[1].Chan <- types.String("x")
-	ep.ports[1].Chan <- types.String("y")
-	ep.ports[2].Chan <- types.String("bad")
-	ep.ports[2].Chan <- types.String("err")
-	ep.close()
-	stdout.Close()
-	stderr.Close()
-
-	stdoutAll := mustReadAllString(stdoutReader)
-	wantStdoutAll := "> x\n> y\n"
-	if stdoutAll != wantStdoutAll {
-		t.Errorf("stdout is %q, want %q", stdoutAll, wantStdoutAll)
-	}
-	stderrAll := mustReadAllString(stderrReader)
-	wantStderrAll := "> bad\n> err\n"
-	if stderrAll != wantStderrAll {
-		t.Errorf("stderr is %q, want %q", stderrAll, wantStderrAll)
-	}
-}
-
-func mustPipe() (*os.File, *os.File) {
-	r, w, err := os.Pipe()
-	if err != nil {
-		panic(err)
-	}
-	return r, w
-}
-
-func mustReadAllString(r io.Reader) string {
-	b, err := ioutil.ReadAll(r)
-	if err != nil {
-		panic(err)
-	}
-	return string(b)
-}
diff -pruN 0.11+ds1-5/eval/eval.go 0.12+ds1-1/eval/eval.go
--- 0.11+ds1-5/eval/eval.go	2018-01-18 02:03:11.000000000 +0000
+++ 0.12+ds1-1/eval/eval.go	2018-07-11 02:32:25.000000000 +0000
@@ -4,20 +4,19 @@ package eval
 
 import (
 	"fmt"
-	"io/ioutil"
 	"os"
 	"os/signal"
 	"strings"
 	"syscall"
-	"unicode/utf8"
 
 	"github.com/elves/elvish/daemon"
 	"github.com/elves/elvish/eval/bundled"
-	"github.com/elves/elvish/eval/types"
-	"github.com/elves/elvish/eval/vartypes"
+	"github.com/elves/elvish/eval/vals"
+	"github.com/elves/elvish/eval/vars"
 	"github.com/elves/elvish/parse"
 	"github.com/elves/elvish/sys"
 	"github.com/elves/elvish/util"
+	"github.com/xiaq/persistent/hashmap"
 	"github.com/xiaq/persistent/vector"
 )
 
@@ -35,17 +34,23 @@ const (
 )
 
 const (
-	defaultValueOutIndicator = "▶ "
-	initIndent               = types.NoPretty
+	defaultValuePrefix        = "▶ "
+	defaultNotifyBgJobSuccess = true
+	defaultNumBgJobs          = 0
+	initIndent                = vals.NoPretty
 )
 
 // Evaler is used to evaluate elvish sources. It maintains runtime context
 // shared among all evalCtx instances.
 type Evaler struct {
 	evalerScopes
-	evalerPorts
-	DaemonClient *daemon.Client
-	modules      map[string]Ns
+	valuePrefix        string
+	notifyBgJobSuccess bool
+	numBgJobs          int
+	beforeChdir        []func(string)
+	afterChdir         []func(string)
+	DaemonClient       *daemon.Client
+	modules            map[string]Ns
 	// bundled modules
 	bundled map[string]string
 	Editor  Editor
@@ -60,9 +65,12 @@ type evalerScopes struct {
 
 // NewEvaler creates a new Evaler.
 func NewEvaler() *Evaler {
-	builtin := makeBuiltinNs()
+	builtin := builtinNs.Clone()
 
 	ev := &Evaler{
+		valuePrefix:        defaultValuePrefix,
+		notifyBgJobSuccess: defaultNotifyBgJobSuccess,
+		numBgJobs:          defaultNumBgJobs,
 		evalerScopes: evalerScopes{
 			Global:  make(Ns),
 			Builtin: builtin,
@@ -75,23 +83,62 @@ func NewEvaler() *Evaler {
 		intCh:   nil,
 	}
 
-	valueOutIndicator := defaultValueOutIndicator
-	ev.evalerPorts = newEvalerPorts(os.Stdin, os.Stdout, os.Stderr, &valueOutIndicator)
-	builtin["value-out-indicator"] = vartypes.NewString(&valueOutIndicator)
+	beforeChdirElvish, afterChdirElvish := vector.Empty, vector.Empty
+	ev.beforeChdir = append(ev.beforeChdir,
+		adaptChdirHook("before-chdir", ev, &beforeChdirElvish))
+	ev.afterChdir = append(ev.afterChdir,
+		adaptChdirHook("after-chdir", ev, &afterChdirElvish))
+	builtin["before-chdir"] = vars.FromPtr(&beforeChdirElvish)
+	builtin["after-chdir"] = vars.FromPtr(&afterChdirElvish)
+
+	builtin["value-out-indicator"] = vars.FromPtr(&ev.valuePrefix)
+	builtin["notify-bg-job-success"] = vars.FromPtr(&ev.notifyBgJobSuccess)
+	builtin["num-bg-jobs"] = vars.FromPtr(&ev.numBgJobs)
+	builtin["pwd"] = PwdVariable{ev}
 
 	return ev
 }
 
-// Close releases resources allocated when creating this Evaler.
-func (ev *Evaler) Close() {
-	ev.evalerPorts.close()
+func adaptChdirHook(name string, ev *Evaler, pfns *vector.Vector) func(string) {
+	return func(path string) {
+		stdPorts := newStdPorts(os.Stdin, os.Stdout, os.Stderr, ev.valuePrefix)
+		defer stdPorts.close()
+		for it := (*pfns).Iterator(); it.HasElem(); it.Next() {
+			fn, ok := it.Elem().(Callable)
+			if !ok {
+				fmt.Fprintln(os.Stderr, name, "hook must be callable")
+				continue
+			}
+			fm := NewTopFrame(ev,
+				NewInternalSource("["+name+" hook]"), stdPorts.ports[:])
+			err := fm.Call(fn, []interface{}{path}, NoOpts)
+			if err != nil {
+				// TODO: Stack trace
+				fmt.Fprintln(os.Stderr, err)
+			}
+		}
+	}
+}
+
+// Close releases resources allocated when creating this Evaler. Currently this
+// does nothing and always returns a nil error.
+func (ev *Evaler) Close() error {
+	return nil
+}
+
+// AddBeforeChdir adds a function to run before changing directory.
+func (ev *Evaler) AddBeforeChdir(f func(string)) {
+	ev.beforeChdir = append(ev.beforeChdir, f)
+}
+
+// AddAfterChdir adds a function to run after changing directory.
+func (ev *Evaler) AddAfterChdir(f func(string)) {
+	ev.afterChdir = append(ev.afterChdir, f)
 }
 
 // InstallDaemonClient installs a daemon client to the Evaler.
 func (ev *Evaler) InstallDaemonClient(client *daemon.Client) {
 	ev.DaemonClient = client
-	// XXX This is really brittle
-	ev.Builtin["pwd"] = PwdVariable{client}
 }
 
 // InstallModule installs a module to the Evaler so that it can be used with
@@ -109,9 +156,9 @@ func (ev *Evaler) InstallBundled(name, s
 func (ev *Evaler) SetArgs(args []string) {
 	v := vector.Empty
 	for _, arg := range args {
-		v = v.Cons(types.String(arg))
+		v = v.Cons(arg)
 	}
-	ev.Builtin["args"] = vartypes.NewRo(types.NewList(v))
+	ev.Builtin["args"] = vars.NewRo(v)
 }
 
 // SetLibDir sets the library directory, in which external modules are to be
@@ -125,31 +172,26 @@ func searchPaths() []string {
 }
 
 // growPorts makes the size of ec.ports at least n, adding nil's if necessary.
-func (ec *Frame) growPorts(n int) {
-	if len(ec.ports) >= n {
+func (fm *Frame) growPorts(n int) {
+	if len(fm.ports) >= n {
 		return
 	}
-	ports := ec.ports
-	ec.ports = make([]*Port, n)
-	copy(ec.ports, ports)
-}
-
-// eval evaluates a chunk node n. The supplied name and text are used in
-// diagnostic messages.
-func (ev *Evaler) eval(op Op, ports []*Port, src *Source) error {
-	ec := NewTopFrame(ev, src, ports)
-	return ec.PEval(op)
+	ports := fm.ports
+	fm.ports = make([]*Port, n)
+	copy(fm.ports, ports)
 }
 
-// Eval sets up the Evaler with standard ports and evaluates an Op. The supplied
-// name and text are used in diagnostic messages.
-func (ev *Evaler) Eval(op Op, src *Source) error {
-	return ev.EvalWithPorts(ev.ports[:], op, src)
+// EvalWithStdPorts sets up the Evaler with standard ports and evaluates an Op.
+// The supplied name and text are used in diagnostic messages.
+func (ev *Evaler) EvalWithStdPorts(op Op, src *Source) error {
+	stdPorts := newStdPorts(os.Stdin, os.Stdout, os.Stderr, ev.valuePrefix)
+	defer stdPorts.close()
+	return ev.Eval(op, stdPorts.ports[:], src)
 }
 
-// EvalWithPorts sets up the Evaler with the given ports and evaluates an Op.
+// Eval sets up the Evaler with the given ports and evaluates an Op.
 // The supplied name and text are used in diagnostic messages.
-func (ev *Evaler) EvalWithPorts(ports []*Port, op Op, src *Source) error {
+func (ev *Evaler) Eval(op Op, ports []*Port, src *Source) error {
 	// Ignore TTOU.
 	//
 	// When a subprocess in its own process group puts itself in the foreground,
@@ -206,14 +248,21 @@ func (ev *Evaler) EvalWithPorts(ports []
 	return err
 }
 
-// Compile compiles elvish code in the global scope. If the error is not nil, it
+// eval evaluates a chunk node n. The supplied name and text are used in
+// diagnostic messages.
+func (ev *Evaler) eval(op Op, ports []*Port, src *Source) error {
+	ec := NewTopFrame(ev, src, ports)
+	return ec.Eval(op)
+}
+
+// Compile compiles Elvish code in the global scope. If the error is not nil, it
 // always has type CompilationError.
 func (ev *Evaler) Compile(n *parse.Chunk, src *Source) (Op, error) {
 	return compile(ev.Builtin.static(), ev.Global.static(), n, src)
 }
 
-// SourceText evaluates a chunk of elvish source.
-func (ev *Evaler) SourceText(src *Source) error {
+// EvalSource evaluates a chunk of Elvish source.
+func (ev *Evaler) EvalSource(src *Source) error {
 	n, err := parse.Parse(src.name, src.code)
 	if err != nil {
 		return err
@@ -222,25 +271,44 @@ func (ev *Evaler) SourceText(src *Source
 	if err != nil {
 		return err
 	}
-	return ev.Eval(op, src)
+	return ev.EvalWithStdPorts(op, src)
 }
 
-func readFileUTF8(fname string) (string, error) {
-	bytes, err := ioutil.ReadFile(fname)
+// SourceRC evaluates a rc.elv file. It evaluates the file in the global
+// namespace. If the file defines a $-exports- map variable, the variable is
+// removed and its content is poured into the global namespace, not overriding
+// existing variables.
+func (ev *Evaler) SourceRC(src *Source) error {
+	n, err := parse.Parse(src.name, src.code)
 	if err != nil {
-		return "", err
-	}
-	if !utf8.Valid(bytes) {
-		return "", fmt.Errorf("%s: source is not valid UTF-8", fname)
+		return err
 	}
-	return string(bytes), nil
-}
-
-// Source evaluates the content of a file.
-func (ev *Evaler) Source(name, path string) error {
-	code, err := readFileUTF8(path)
+	op, err := ev.Compile(n, src)
 	if err != nil {
 		return err
 	}
-	return ev.SourceText(NewScriptSource(name, path, code))
+	errEval := ev.EvalWithStdPorts(op, src)
+	var errExports error
+	if ev.Global.HasName("-exports-") {
+		exports := ev.Global.PopName("-exports-").Get()
+		switch exports := exports.(type) {
+		case hashmap.Map:
+			for it := exports.Iterator(); it.HasElem(); it.Next() {
+				k, v := it.Elem()
+				if name, ok := k.(string); ok {
+					if !ev.Global.HasName(name) {
+						ev.Global.Add(name, vars.NewAnyWithInit(v))
+					}
+				} else {
+					errKey := fmt.Errorf("non-string key in $-exports-: %s",
+						vals.Repr(k, vals.NoPretty))
+					errExports = util.Errors(errExports, errKey)
+				}
+			}
+		default:
+			errExports = fmt.Errorf(
+				"$-exports- should be a map, got %s", vals.Kind(exports))
+		}
+	}
+	return util.Errors(errEval, errExports)
 }
diff -pruN 0.11+ds1-5/eval/eval_test.go 0.12+ds1-1/eval/eval_test.go
--- 0.11+ds1-5/eval/eval_test.go	2018-01-18 02:03:11.000000000 +0000
+++ 0.12+ds1-1/eval/eval_test.go	2018-07-11 02:32:25.000000000 +0000
@@ -6,70 +6,70 @@ import (
 	"syscall"
 	"testing"
 
-	"github.com/elves/elvish/eval/types"
+	"github.com/elves/elvish/eval/vals"
 )
 
 func TestBuiltinPid(t *testing.T) {
 	pid := strconv.Itoa(syscall.Getpid())
-	builtinPid := types.ToString(makeBuiltinNs()["pid"].Get())
+	builtinPid := vals.ToString(builtinNs["pid"].Get())
 	if builtinPid != pid {
 		t.Errorf(`ev.builtin["pid"] = %v, want %v`, builtinPid, pid)
 	}
 }
 
-var miscEvalTests = []Test{
-	// Pseudo-namespaces local: and up:
-	{"x=lorem; []{local:x=ipsum; put $up:x $local:x}",
-		want{out: strs("lorem", "ipsum")}},
-	{"x=lorem; []{up:x=ipsum; put $x}; put $x",
-		want{out: strs("ipsum", "ipsum")}},
-	// Pseudo-namespace E:
-	{"E:FOO=lorem; put $E:FOO", want{out: strs("lorem")}},
-	{"del E:FOO; put $E:FOO", want{out: strs("")}},
-}
-
 func TestMiscEval(t *testing.T) {
-	runTests(t, miscEvalTests)
+	Test(t,
+		// Pseudo-namespaces local: and up:
+		That("x=lorem; { local:x=ipsum; put $up:x $local:x }").Puts(
+			"lorem", "ipsum"),
+		That("x=lorem; { up:x=ipsum; put $x }; put $x").Puts("ipsum", "ipsum"),
+		// Pseudo-namespace E:
+		That("E:FOO=lorem; put $E:FOO").Puts("lorem"),
+		That("del E:FOO; put $E:FOO").Puts(""),
+	)
 }
 
 func TestMultipleEval(t *testing.T) {
 	texts := []string{"x=hello", "put $x"}
 	outs, _, err := evalAndCollect(t, NewEvaler(), texts, 1)
-	wanted := strs("hello")
+	wantOuts := []interface{}{"hello"}
 	if err != nil {
 		t.Errorf("eval %s => %v, want nil", texts, err)
 	}
-	if !reflect.DeepEqual(outs, wanted) {
-		t.Errorf("eval %s outputs %v, want %v", texts, outs, wanted)
+	if !reflect.DeepEqual(outs, wantOuts) {
+		t.Errorf("eval %s outputs %v, want %v", texts, outs, wantOuts)
 	}
 }
 
 func BenchmarkOutputCaptureOverhead(b *testing.B) {
-	op := Op{func(*Frame) {}, 0, 0}
+	op := Op{funcOp(func(*Frame) error { return nil }), 0, 0}
 	benchmarkOutputCapture(op, b.N)
 }
 
 func BenchmarkOutputCaptureValues(b *testing.B) {
-	op := Op{func(ec *Frame) {
-		ec.ports[1].Chan <- types.String("test")
-	}, 0, 0}
+	op := Op{funcOp(func(fm *Frame) error {
+		fm.ports[1].Chan <- "test"
+		return nil
+	}), 0, 0}
 	benchmarkOutputCapture(op, b.N)
 }
 
 func BenchmarkOutputCaptureBytes(b *testing.B) {
 	bytesToWrite := []byte("test")
-	op := Op{func(ec *Frame) {
-		ec.ports[1].File.Write(bytesToWrite)
-	}, 0, 0}
+	op := Op{funcOp(func(fm *Frame) error {
+		fm.ports[1].File.Write(bytesToWrite)
+		return nil
+	}), 0, 0}
 	benchmarkOutputCapture(op, b.N)
 }
 
 func BenchmarkOutputCaptureMixed(b *testing.B) {
 	bytesToWrite := []byte("test")
-	op := Op{func(ec *Frame) {
-		ec.ports[1].Chan <- types.Bool(false)
-		ec.ports[1].File.Write(bytesToWrite)
-	}, 0, 0}
+	op := Op{funcOp(func(fm *Frame) error {
+		fm.ports[1].Chan <- false
+		fm.ports[1].File.Write(bytesToWrite)
+		return nil
+	}), 0, 0}
 	benchmarkOutputCapture(op, b.N)
 }
 
diff -pruN 0.11+ds1-5/eval/exception.go 0.12+ds1-1/eval/exception.go
--- 0.11+ds1-5/eval/exception.go	2018-01-18 02:03:11.000000000 +0000
+++ 0.12+ds1-1/eval/exception.go	2018-07-11 02:32:25.000000000 +0000
@@ -8,7 +8,7 @@ import (
 	"syscall"
 	"unsafe"
 
-	"github.com/elves/elvish/eval/types"
+	"github.com/elves/elvish/eval/vals"
 	"github.com/elves/elvish/parse"
 	"github.com/elves/elvish/util"
 	"github.com/xiaq/persistent/hash"
@@ -19,7 +19,7 @@ import (
 // methods like (*Evaler)PEval.
 type Exception struct {
 	Cause     error
-	Traceback *util.SourceRange
+	Traceback *stackTrace
 }
 
 // OK is a pointer to the zero value of Exception, representing the absence of
@@ -41,13 +41,13 @@ func (exc *Exception) Pprint(indent stri
 	}
 	fmt.Fprintf(buf, "Exception: %s\n", causeDescription)
 
-	if exc.Traceback.Next == nil {
-		buf.WriteString(exc.Traceback.PprintCompact(indent))
+	if exc.Traceback.next == nil {
+		buf.WriteString(exc.Traceback.entry.PprintCompact(indent))
 	} else {
 		buf.WriteString(indent + "Traceback:")
-		for tb := exc.Traceback; tb != nil; tb = tb.Next {
+		for tb := exc.Traceback; tb != nil; tb = tb.next {
 			buf.WriteString("\n" + indent + "  ")
-			buf.WriteString(tb.Pprint(indent + "    "))
+			buf.WriteString(tb.entry.Pprint(indent + "    "))
 		}
 	}
 
@@ -72,7 +72,7 @@ func (exc *Exception) Repr(indent int) s
 	if exc.Cause == nil {
 		return "$ok"
 	}
-	if r, ok := exc.Cause.(types.Reprer); ok {
+	if r, ok := exc.Cause.(vals.Reprer); ok {
 		return r.Repr(indent)
 	}
 	return "?(fail " + parse.Quote(exc.Cause.Error()) + ")"
@@ -91,6 +91,20 @@ func (exc *Exception) Bool() bool {
 	return exc.Cause == nil
 }
 
+func (exc *Exception) Index(k interface{}) (interface{}, bool) {
+	// TODO: Access to Traceback
+	switch k {
+	case "cause":
+		return exc.Cause, true
+	default:
+		return nil, false
+	}
+}
+
+func (exc *Exception) IterateKeys(f func(interface{}) bool) {
+	vals.Feed(f, "cause")
+}
+
 // PipelineError represents the errors of pipelines, in which multiple commands
 // may error.
 type PipelineError struct {
diff -pruN 0.11+ds1-5/eval/exception_test.go 0.12+ds1-1/eval/exception_test.go
--- 0.11+ds1-5/eval/exception_test.go	2018-01-18 02:03:11.000000000 +0000
+++ 0.12+ds1-1/eval/exception_test.go	2018-07-11 02:32:25.000000000 +0000
@@ -3,7 +3,7 @@ package eval
 import "testing"
 
 func TestException(t *testing.T) {
-	runTests(t, []Test{
-		NewTest("kind-of ?(fail foo)").WantOutStrings("exception"),
-	})
+	Test(t,
+		That("kind-of ?(fail foo)").Puts("exception"),
+	)
 }
diff -pruN 0.11+ds1-5/eval/external_cmd.go 0.12+ds1-1/eval/external_cmd.go
--- 0.11+ds1-5/eval/external_cmd.go	2018-01-18 02:03:11.000000000 +0000
+++ 0.12+ds1-1/eval/external_cmd.go	2018-07-11 02:32:25.000000000 +0000
@@ -7,7 +7,7 @@ import (
 	"os/exec"
 	"syscall"
 
-	"github.com/elves/elvish/eval/types"
+	"github.com/elves/elvish/eval/vals"
 	"github.com/elves/elvish/parse"
 	"github.com/elves/elvish/util"
 	"github.com/xiaq/persistent/hash"
@@ -40,24 +40,23 @@ func (e ExternalCmd) Repr(int) string {
 }
 
 // Call calls an external command.
-func (e ExternalCmd) Call(ec *Frame, argVals []types.Value, opts map[string]types.Value) {
+func (e ExternalCmd) Call(fm *Frame, argVals []interface{}, opts map[string]interface{}) error {
 	if len(opts) > 0 {
-		throw(ErrExternalCmdOpts)
+		return ErrExternalCmdOpts
 	}
 	if util.DontSearch(e.Name) {
 		stat, err := os.Stat(e.Name)
 		if err == nil && stat.IsDir() {
 			// implicit cd
 			if len(argVals) > 0 {
-				throw(ErrCdNoArg)
+				return ErrCdNoArg
 			}
-			cdInner(e.Name, ec)
-			return
+			return fm.Chdir(e.Name)
 		}
 	}
 
-	files := make([]*os.File, len(ec.ports))
-	for i, port := range ec.ports {
+	files := make([]*os.File, len(fm.ports))
+	for i, port := range fm.ports {
 		files[i] = port.File
 	}
 
@@ -65,30 +64,29 @@ func (e ExternalCmd) Call(ec *Frame, arg
 	for i, a := range argVals {
 		// NOTE Maybe we should enfore string arguments instead of coercing all
 		// args into string
-		args[i+1] = types.ToString(a)
+		args[i+1] = vals.ToString(a)
 	}
 
 	path, err := exec.LookPath(e.Name)
 	if err != nil {
-		throw(err)
+		return err
 	}
 
 	args[0] = path
 
-	sys := makeSysProcAttr(ec.background)
+	sys := makeSysProcAttr(fm.background)
 	proc, err := os.StartProcess(path, args, &os.ProcAttr{Files: files, Sys: sys})
 
 	if err != nil {
-		throw(err)
+		return err
 	}
 
 	state, err := proc.Wait()
 
 	if err != nil {
-		throw(err)
-	} else {
-		maybeThrow(NewExternalCmdExit(e.Name, state.Sys().(syscall.WaitStatus), proc.Pid))
+		return err
 	}
+	return NewExternalCmdExit(e.Name, state.Sys().(syscall.WaitStatus), proc.Pid)
 }
 
 // EachExternal calls f for each name that can resolve to an external
diff -pruN 0.11+ds1-5/eval/frame.go 0.12+ds1-1/eval/frame.go
--- 0.11+ds1-5/eval/frame.go	2018-01-18 02:03:11.000000000 +0000
+++ 0.12+ds1-1/eval/frame.go	2018-07-11 02:32:25.000000000 +0000
@@ -7,7 +7,6 @@ import (
 	"strings"
 	"sync"
 
-	"github.com/elves/elvish/eval/types"
 	"github.com/elves/elvish/util"
 )
 
@@ -22,7 +21,7 @@ type Frame struct {
 	ports     []*Port
 
 	begin, end int
-	traceback  *util.SourceRange
+	traceback  *stackTrace
 
 	background bool
 }
@@ -37,38 +36,47 @@ func NewTopFrame(ev *Evaler, src *Source
 	}
 }
 
+// Close releases resources allocated for this frame. It always returns a nil
+// error. It may be called only once.
+func (fm *Frame) Close() error {
+	for _, port := range fm.ports {
+		port.Close()
+	}
+	return nil
+}
+
 // InputChan returns a channel from which input can be read.
-func (ec *Frame) InputChan() chan types.Value {
-	return ec.ports[0].Chan
+func (fm *Frame) InputChan() chan interface{} {
+	return fm.ports[0].Chan
 }
 
 // InputFile returns a file from which input can be read.
-func (ec *Frame) InputFile() *os.File {
-	return ec.ports[0].File
+func (fm *Frame) InputFile() *os.File {
+	return fm.ports[0].File
 }
 
 // OutputChan returns a channel onto which output can be written.
-func (ec *Frame) OutputChan() chan<- types.Value {
-	return ec.ports[1].Chan
+func (fm *Frame) OutputChan() chan<- interface{} {
+	return fm.ports[1].Chan
 }
 
 // OutputFile returns a file onto which output can be written.
-func (ec *Frame) OutputFile() *os.File {
-	return ec.ports[1].File
+func (fm *Frame) OutputFile() *os.File {
+	return fm.ports[1].File
 }
 
 // IterateInputs calls the passed function for each input element.
-func (ec *Frame) IterateInputs(f func(types.Value)) {
+func (fm *Frame) IterateInputs(f func(interface{})) {
 	var w sync.WaitGroup
-	inputs := make(chan types.Value)
+	inputs := make(chan interface{})
 
 	w.Add(2)
 	go func() {
-		linesToChan(ec.ports[0].File, inputs)
+		linesToChan(fm.ports[0].File, inputs)
 		w.Done()
 	}()
 	go func() {
-		for v := range ec.ports[0].Chan {
+		for v := range fm.ports[0].Chan {
 			inputs <- v
 		}
 		w.Done()
@@ -83,12 +91,12 @@ func (ec *Frame) IterateInputs(f func(ty
 	}
 }
 
-func linesToChan(r io.Reader, ch chan<- types.Value) {
+func linesToChan(r io.Reader, ch chan<- interface{}) {
 	filein := bufio.NewReader(r)
 	for {
 		line, err := filein.ReadString('\n')
 		if line != "" {
-			ch <- types.String(strings.TrimSuffix(line, "\n"))
+			ch <- strings.TrimSuffix(line, "\n")
 		}
 		if err != nil {
 			if err != io.EOF {
@@ -101,47 +109,70 @@ func linesToChan(r io.Reader, ch chan<-
 
 // fork returns a modified copy of ec. The ports are forked, and the name is
 // changed to the given value. Other fields are copied shallowly.
-func (ec *Frame) fork(name string) *Frame {
-	newPorts := make([]*Port, len(ec.ports))
-	for i, p := range ec.ports {
+func (fm *Frame) fork(name string) *Frame {
+	newPorts := make([]*Port, len(fm.ports))
+	for i, p := range fm.ports {
 		newPorts[i] = p.Fork()
 	}
 	return &Frame{
-		ec.Evaler, ec.srcMeta,
-		ec.local, ec.up,
+		fm.Evaler, fm.srcMeta,
+		fm.local, fm.up,
 		newPorts,
-		ec.begin, ec.end, ec.traceback, ec.background,
+		fm.begin, fm.end, fm.traceback, fm.background,
 	}
 }
 
-// PEval evaluates an op in a protected environment so that calls to errorf are
-// wrapped in an Error.
-func (ec *Frame) PEval(op Op) (err error) {
-	defer catch(&err, ec)
-	op.Exec(ec)
+// Eval evaluates an op. It does so in a protected environment so that
+// exceptions thrown are wrapped in an Error.
+func (fm *Frame) Eval(op Op) (err error) {
+	defer catch(&err, fm)
+	e := op.Exec(fm)
+	if e != nil {
+		if exc, ok := e.(*Exception); ok {
+			return exc
+		}
+		return fm.makeException(e)
+	}
 	return nil
 }
 
-func (ec *Frame) PCall(f Callable, args []types.Value, opts map[string]types.Value) (err error) {
-	defer catch(&err, ec)
-	f.Call(ec, args, opts)
+// Call calls a function with the given arguments and options. It does so in a
+// protected environment so that exceptions thrown are wrapped in an Error.
+func (fm *Frame) Call(f Callable, args []interface{}, opts map[string]interface{}) (err error) {
+	defer catch(&err, fm)
+	e := f.Call(fm, args, opts)
+	if e != nil {
+		if exc, ok := e.(*Exception); ok {
+			return exc
+		}
+		return fm.makeException(e)
+	}
 	return nil
 }
 
-func (ec *Frame) PCaptureOutput(f Callable, args []types.Value, opts map[string]types.Value) (vs []types.Value, err error) {
+// CaptureOutput calls a function with the given arguments and options,
+// capturing and returning the output. It does so in a protected environment so
+// that exceptions thrown are wrapped in an Error.
+func (fm *Frame) CaptureOutput(fn Callable, args []interface{}, opts map[string]interface{}) (vs []interface{}, err error) {
 	// XXX There is no source.
-	return pcaptureOutput(ec, Op{
-		func(newec *Frame) { f.Call(newec, args, opts) }, -1, -1})
+	opFunc := func(f *Frame) error {
+		return fn.Call(f, args, opts)
+	}
+	return pcaptureOutput(fm, Op{funcOp(opFunc), -1, -1})
 }
 
-func (ec *Frame) PCaptureOutputInner(f Callable, args []types.Value, opts map[string]types.Value, valuesCb func(<-chan types.Value), bytesCb func(*os.File)) error {
+// CallWithOutputCallback calls a function with the given arguments and options,
+// feeding the outputs to the given callbacks. It does so in a protected
+// environment so that exceptions thrown are wrapped in an Error.
+func (fm *Frame) CallWithOutputCallback(fn Callable, args []interface{}, opts map[string]interface{}, valuesCb func(<-chan interface{}), bytesCb func(*os.File)) error {
 	// XXX There is no source.
-	return pcaptureOutputInner(ec, Op{
-		func(newec *Frame) { f.Call(newec, args, opts) }, -1, -1},
-		valuesCb, bytesCb)
+	opFunc := func(f *Frame) error {
+		return fn.Call(f, args, opts)
+	}
+	return pcaptureOutputInner(fm, Op{funcOp(opFunc), -1, -1}, valuesCb, bytesCb)
 }
 
-func catch(perr *error, ec *Frame) {
+func catch(perr *error, fm *Frame) {
 	// NOTE: We have to duplicate instead of calling util.Catch here, since
 	// recover can only catch a panic when called directly from a deferred
 	// function.
@@ -152,7 +183,7 @@ func catch(perr *error, ec *Frame) {
 	if exc, ok := r.(util.Thrown); ok {
 		err := exc.Wrapped
 		if _, ok := err.(*Exception); !ok {
-			err = ec.makeException(err)
+			err = fm.makeException(err)
 		}
 		*perr = err
 	} else if r != nil {
@@ -161,20 +192,23 @@ func catch(perr *error, ec *Frame) {
 }
 
 // makeException turns an error into an Exception by adding traceback.
-func (ec *Frame) makeException(e error) *Exception {
-	return &Exception{e, ec.addTraceback()}
+func (fm *Frame) makeException(e error) *Exception {
+	return &Exception{e, fm.addTraceback()}
 }
 
-func (ec *Frame) addTraceback() *util.SourceRange {
-	return &util.SourceRange{
-		Name: ec.srcMeta.describePath(), Source: ec.srcMeta.code,
-		Begin: ec.begin, End: ec.end, Next: ec.traceback,
+func (fm *Frame) addTraceback() *stackTrace {
+	return &stackTrace{
+		entry: &util.SourceRange{
+			Name: fm.srcMeta.describePath(), Source: fm.srcMeta.code,
+			Begin: fm.begin, End: fm.end,
+		},
+		next: fm.traceback,
 	}
 }
 
 // errorpf stops the ec.eval immediately by panicking with a diagnostic message.
 // The panic is supposed to be caught by ec.eval.
-func (ec *Frame) errorpf(begin, end int, format string, args ...interface{}) {
-	ec.begin, ec.end = begin, end
+func (fm *Frame) errorpf(begin, end int, format string, args ...interface{}) {
+	fm.begin, fm.end = begin, end
 	throwf(format, args...)
 }
diff -pruN 0.11+ds1-5/eval/glob.go 0.12+ds1-1/eval/glob.go
--- 0.11+ds1-5/eval/glob.go	2018-01-18 02:03:11.000000000 +0000
+++ 0.12+ds1-1/eval/glob.go	2018-07-11 02:32:25.000000000 +0000
@@ -7,7 +7,7 @@ import (
 	"strings"
 	"unicode"
 
-	"github.com/elves/elvish/eval/types"
+	"github.com/elves/elvish/eval/vals"
 	"github.com/elves/elvish/glob"
 	"github.com/elves/elvish/parse"
 )
@@ -31,8 +31,8 @@ func (f GlobFlag) Has(g GlobFlag) bool {
 }
 
 var (
-	_ types.Value   = GlobPattern{}
-	_ types.Indexer = GlobPattern{}
+	_ interface{}     = GlobPattern{}
+	_ vals.ErrIndexer = GlobPattern{}
 )
 
 var (
@@ -74,57 +74,83 @@ func (gp GlobPattern) Repr(int) string {
 	return fmt.Sprintf("<GlobPattern%v>", gp)
 }
 
-func (gp GlobPattern) Index(modifiers []types.Value) []types.Value {
-	for _, value := range modifiers {
-		modifierv, ok := value.(types.String)
-		if !ok {
-			throw(ErrModifierMustBeString)
-		}
-		modifier := string(modifierv)
-		switch {
-		case modifier == "nomatch-ok":
-			gp.Flags |= NoMatchOK
-		case strings.HasPrefix(modifier, "but:"):
-			gp.Buts = append(gp.Buts, modifier[len("but:"):])
-		case modifier == "match-hidden":
-			lastSeg := gp.mustGetLastWildSeg()
-			gp.Segments[len(gp.Segments)-1] = glob.Wild{
-				lastSeg.Type, true, lastSeg.Matchers,
+func (gp GlobPattern) Index(k interface{}) (interface{}, error) {
+	modifierv, ok := k.(string)
+	if !ok {
+		return nil, ErrModifierMustBeString
+	}
+	modifier := modifierv
+	switch {
+	case modifier == "nomatch-ok":
+		gp.Flags |= NoMatchOK
+	case strings.HasPrefix(modifier, "but:"):
+		gp.Buts = append(gp.Buts, modifier[len("but:"):])
+	case modifier == "match-hidden":
+		lastSeg := gp.mustGetLastWildSeg()
+		gp.Segments[len(gp.Segments)-1] = glob.Wild{
+			lastSeg.Type, true, lastSeg.Matchers,
+		}
+	default:
+		if matcher, ok := runeMatchers[modifier]; ok {
+			gp.addMatcher(matcher)
+		} else if strings.HasPrefix(modifier, "set:") {
+			set := modifier[len("set:"):]
+			gp.addMatcher(func(r rune) bool {
+				return strings.ContainsRune(set, r)
+			})
+		} else if strings.HasPrefix(modifier, "range:") {
+			rangeExpr := modifier[len("range:"):]
+			badRangeExpr := fmt.Errorf("bad range modifier: %s", parse.Quote(rangeExpr))
+			runes := []rune(rangeExpr)
+			if len(runes) != 3 {
+				return nil, badRangeExpr
 			}
-		default:
-			if matcher, ok := runeMatchers[modifier]; ok {
-				gp.addMatcher(matcher)
-			} else if strings.HasPrefix(modifier, "set:") {
-				set := modifier[len("set:"):]
+			from, sep, to := runes[0], runes[1], runes[2]
+			switch sep {
+			case '-':
 				gp.addMatcher(func(r rune) bool {
-					return strings.ContainsRune(set, r)
+					return from <= r && r <= to
 				})
-			} else if strings.HasPrefix(modifier, "range:") {
-				rangeExpr := modifier[len("range:"):]
-				badRangeExpr := fmt.Errorf("bad range modifier: %s", parse.Quote(rangeExpr))
-				runes := []rune(rangeExpr)
-				if len(runes) != 3 {
-					throw(badRangeExpr)
-				}
-				from, sep, to := runes[0], runes[1], runes[2]
-				switch sep {
-				case '-':
-					gp.addMatcher(func(r rune) bool {
-						return from <= r && r <= to
-					})
-				case '~':
-					gp.addMatcher(func(r rune) bool {
-						return from <= r && r < to
-					})
-				default:
-					throw(badRangeExpr)
-				}
-			} else {
-				throw(fmt.Errorf("unknown modifier %s", modifierv.Repr(types.NoPretty)))
+			case '~':
+				gp.addMatcher(func(r rune) bool {
+					return from <= r && r < to
+				})
+			default:
+				return nil, badRangeExpr
 			}
+		} else {
+			return nil, fmt.Errorf("unknown modifier %s", vals.Repr(modifierv, vals.NoPretty))
 		}
 	}
-	return []types.Value{gp}
+	return gp, nil
+}
+
+func (gp GlobPattern) Concat(v interface{}) (interface{}, error) {
+	switch rhs := v.(type) {
+	case string:
+		gp.append(stringToSegments(rhs)...)
+		return gp, nil
+	case GlobPattern:
+		// We know rhs contains exactly one segment.
+		gp.append(rhs.Segments[0])
+		gp.Flags |= rhs.Flags
+		gp.Buts = append(gp.Buts, rhs.Buts...)
+		return gp, nil
+	}
+
+	return nil, vals.ErrConcatNotImplemented
+}
+
+func (gp GlobPattern) RConcat(v interface{}) (interface{}, error) {
+	switch lhs := v.(type) {
+	case string:
+		segs := stringToSegments(lhs)
+		// We know gp contains exactly one segment.
+		segs = append(segs, gp.Segments[0])
+		return GlobPattern{glob.Pattern{Segments: segs}, gp.Flags, gp.Buts}, nil
+	}
+
+	return nil, vals.ErrConcatNotImplemented
 }
 
 func (gp *GlobPattern) mustGetLastWildSeg() glob.Wild {
@@ -183,13 +209,13 @@ func stringToSegments(s string) []glob.S
 	return segs
 }
 
-func doGlob(gp GlobPattern, abort <-chan struct{}) []types.Value {
+func doGlob(gp GlobPattern, abort <-chan struct{}) []interface{} {
 	but := make(map[string]struct{})
 	for _, s := range gp.Buts {
 		but[s] = struct{}{}
 	}
 
-	vs := make([]types.Value, 0)
+	vs := make([]interface{}, 0)
 	if !gp.Glob(func(name string) bool {
 		select {
 		case <-abort:
@@ -198,7 +224,7 @@ func doGlob(gp GlobPattern, abort <-chan
 		default:
 		}
 		if _, b := but[name]; !b {
-			vs = append(vs, types.String(name))
+			vs = append(vs, name)
 		}
 		return true
 	}) {
diff -pruN 0.11+ds1-5/eval/interrupts.go 0.12+ds1-1/eval/interrupts.go
--- 0.11+ds1-5/eval/interrupts.go	2018-01-18 02:03:11.000000000 +0000
+++ 0.12+ds1-1/eval/interrupts.go	2018-07-11 02:32:25.000000000 +0000
@@ -3,18 +3,18 @@ package eval
 import "errors"
 
 // Interrupts returns a channel that is closed when an interrupt signal comes.
-func (ec *Frame) Interrupts() <-chan struct{} {
-	return ec.intCh
+func (fm *Frame) Interrupts() <-chan struct{} {
+	return fm.intCh
 }
 
 var ErrInterrupted = errors.New("interrupted")
 
-// CheckInterrupts checks whether there has been an interrupt, and throws
-// ErrInterrupted if that is the case
-func (ec *Frame) CheckInterrupts() {
+// IsInterrupted reports whether there has been an interrupt.
+func (fm *Frame) IsInterrupted() bool {
 	select {
-	case <-ec.Interrupts():
-		throw(ErrInterrupted)
+	case <-fm.Interrupts():
+		return true
 	default:
+		return false
 	}
 }
diff -pruN 0.11+ds1-5/eval/names.go 0.12+ds1-1/eval/names.go
--- 0.11+ds1-5/eval/names.go	2018-01-18 02:03:11.000000000 +0000
+++ 0.12+ds1-1/eval/names.go	1970-01-01 00:00:00.000000000 +0000
@@ -1,126 +0,0 @@
-package eval
-
-import (
-	"errors"
-	"os"
-	"strings"
-
-	"github.com/elves/elvish/eval/vartypes"
-)
-
-// Resolution and iteration of variables and namespaces.
-
-// ErrStoreUnconnected is thrown by ResolveVar when a shared: variable needs to
-// be resolved but the store is not connected.
-var ErrStoreUnconnected = errors.New("store unconnected")
-
-// EachVariableInTop calls the passed function for each variable name in
-// namespace ns that can be found from the top context.
-func (ev *evalerScopes) EachVariableInTop(ns string, f func(s string)) {
-	switch ns {
-	case "builtin":
-		for name := range ev.Builtin {
-			f(name)
-		}
-	case "":
-		for name := range ev.Global {
-			f(name)
-		}
-		for name := range ev.Builtin {
-			f(name)
-		}
-	case "e":
-		EachExternal(func(cmd string) {
-			f(cmd + FnSuffix)
-		})
-	case "E":
-		for _, s := range os.Environ() {
-			if i := strings.IndexByte(s, '='); i > 0 {
-				f(s[:i])
-			}
-		}
-	case "shared":
-		// TODO Add daemon RPC for enumerating shared variables.
-	default:
-		mod := ev.Global[ns+NsSuffix]
-		if mod == nil {
-			mod = ev.Builtin[ns+NsSuffix]
-		}
-		if mod != nil {
-			for name := range mod.Get().(Ns) {
-				f(name)
-			}
-		}
-	}
-}
-
-// EachModInTop calls the passed function for each module that can be found from
-// the top context.
-func (ev *evalerScopes) EachModInTop(f func(s string)) {
-	for name := range ev.Global {
-		if strings.HasSuffix(name, NsSuffix) {
-			f(name[:len(name)-len(NsSuffix)])
-		}
-	}
-	for name := range ev.Builtin {
-		if strings.HasSuffix(name, NsSuffix) {
-			f(name[:len(name)-len(NsSuffix)])
-		}
-	}
-}
-
-// EachNsInTop calls the passed function for each namespace that can be used
-// from the top context.
-func (ev *evalerScopes) EachNsInTop(f func(s string)) {
-	f("builtin")
-	f("e")
-	f("E")
-	f("shared")
-	ev.EachModInTop(f)
-}
-
-// ResolveVar resolves a variable. When the variable cannot be found, nil is
-// returned.
-func (ec *Frame) ResolveVar(ns, name string) vartypes.Variable {
-	switch ns {
-	case "local":
-		return ec.local[name]
-	case "up":
-		return ec.up[name]
-	case "builtin":
-		return ec.Builtin[name]
-	case "":
-		if v := ec.local[name]; v != nil {
-			return v
-		}
-		if v, ok := ec.up[name]; ok {
-			return v
-		}
-		return ec.Builtin[name]
-	case "e":
-		if strings.HasSuffix(name, FnSuffix) {
-			return vartypes.NewRo(ExternalCmd{name[:len(name)-len(FnSuffix)]})
-		}
-	case "E":
-		return vartypes.NewEnv(name)
-	case "shared":
-		if ec.DaemonClient == nil {
-			throw(ErrStoreUnconnected)
-		}
-		return sharedVariable{ec.DaemonClient, name}
-	default:
-		ns := ec.ResolveMod(ns)
-		if ns != nil {
-			return ns[name]
-		}
-	}
-	return nil
-}
-
-func (ec *Frame) ResolveMod(name string) Ns {
-	ns := ec.ResolveVar("", name+NsSuffix)
-	if ns == nil {
-		return nil
-	}
-	return ns.Get().(Ns)
-}
diff -pruN 0.11+ds1-5/eval/ns.go 0.12+ds1-1/eval/ns.go
--- 0.11+ds1-5/eval/ns.go	2018-01-18 02:03:11.000000000 +0000
+++ 0.12+ds1-1/eval/ns.go	2018-07-11 02:32:25.000000000 +0000
@@ -4,14 +4,18 @@ import (
 	"fmt"
 	"reflect"
 
-	"github.com/elves/elvish/eval/types"
-	"github.com/elves/elvish/eval/vartypes"
+	"github.com/elves/elvish/eval/vars"
 )
 
 // Ns is a map from names to variables.
-type Ns map[string]vartypes.Variable
+type Ns map[string]vars.Var
 
-var _ types.Value = Ns(nil)
+var _ interface{} = Ns(nil)
+
+// NewNs creates an empty namespace.
+func NewNs() Ns {
+	return make(Ns)
+}
 
 func (Ns) Kind() string {
 	return "ns"
@@ -32,6 +36,77 @@ func (ns Ns) Repr(int) string {
 	return fmt.Sprintf("<ns 0x%x>", addrOf(ns))
 }
 
+func (ns Ns) Index(k interface{}) (interface{}, bool) {
+	if kstring, ok := k.(string); ok {
+		if v, ok := ns[kstring]; ok {
+			return v.Get(), true
+		}
+	}
+	return nil, false
+}
+
+func (ns Ns) IterateKeys(f func(interface{}) bool) {
+	for k := range ns {
+		if !f(k) {
+			break
+		}
+	}
+}
+
+// HasName reports the namespace contains the given name.
+func (ns Ns) HasName(name string) bool {
+	_, ok := ns[name]
+	return ok
+}
+
+// PopName removes a name from the namespace and returns the variable it used to
+// contain.
+func (ns Ns) PopName(name string) vars.Var {
+	v := ns[name]
+	delete(ns, name)
+	return v
+}
+
+// Clone returns a shallow copy of the namespace.
+func (ns Ns) Clone() Ns {
+	ns2 := make(Ns)
+	for name, variable := range ns {
+		ns2[name] = variable
+	}
+	return ns2
+}
+
+// Add adds a variable to the namespace and returns the namespace itself.
+func (ns Ns) Add(name string, v vars.Var) Ns {
+	ns[name] = v
+	return ns
+}
+
+// AddFn adds a function to a namespace. It returns the namespace itself.
+func (ns Ns) AddFn(name string, v Callable) Ns {
+	return ns.Add(name+FnSuffix, vars.FromPtr(&v))
+}
+
+// AddNs adds a sub-namespace to a namespace. It returns the namespace itself.
+func (ns Ns) AddNs(name string, v Ns) Ns {
+	return ns.Add(name+NsSuffix, vars.FromPtr(&v))
+}
+
+// AddBuiltinFn adds a builtin function to a namespace. It returns the namespace
+// itself.
+func (ns Ns) AddBuiltinFn(nsName, name string, impl interface{}) Ns {
+	return ns.AddFn(name, NewBuiltinFn(nsName+name, impl))
+}
+
+// AddBuiltinFns adds builtin functions to a namespace. It returns the namespace
+// itself.
+func (ns Ns) AddBuiltinFns(nsName string, fns map[string]interface{}) Ns {
+	for name, impl := range fns {
+		ns.AddBuiltinFn(nsName, name, impl)
+	}
+	return ns
+}
+
 func addrOf(a interface{}) uintptr {
 	return reflect.ValueOf(a).Pointer()
 }
diff -pruN 0.11+ds1-5/eval/ns_test.go 0.12+ds1-1/eval/ns_test.go
--- 0.11+ds1-5/eval/ns_test.go	2018-01-18 02:03:11.000000000 +0000
+++ 0.12+ds1-1/eval/ns_test.go	2018-07-11 02:32:25.000000000 +0000
@@ -3,11 +3,18 @@ package eval
 import "testing"
 
 func TestNs(t *testing.T) {
-	runTests(t, []Test{
-		NewTest("kind-of (ns)").WantOutStrings("ns"),
+	Test(t,
+		That("kind-of (ns [&])").Puts("ns"),
 		// A Ns is only equal to itself
-		NewTest("ns = (ns); eq $ns $ns").WantOutBools(true),
-		NewTest("eq (ns) (ns)").WantOutBools(false),
-		NewTest("eq (ns) [&]").WantOutBools(false),
-	})
+		That("ns = (ns [&]); eq $ns $ns").Puts(true),
+		That("eq (ns [&]) (ns [&])").Puts(false),
+		That("eq (ns [&]) [&]").Puts(false),
+
+		That(`ns: = (ns [&a=b &x=y]); put $ns:a`).Puts("b"),
+		That(`ns: = (ns [&a=b &x=y]); put $ns:[a]`).Puts("b"),
+		// Test multi-key ns when sorting is possible
+		That(`keys (ns [&a=b])`).Puts("a"),
+		That(`has-key (ns [&a=b &x=y]) a`).Puts(true),
+		That(`has-key (ns [&a=b &x=y]) b`).Puts(false),
+	)
 }
diff -pruN 0.11+ds1-5/eval/port.go 0.12+ds1-1/eval/port.go
--- 0.11+ds1-5/eval/port.go	2018-01-18 02:03:11.000000000 +0000
+++ 0.12+ds1-1/eval/port.go	2018-07-11 02:32:25.000000000 +0000
@@ -2,14 +2,12 @@ package eval
 
 import (
 	"os"
-
-	"github.com/elves/elvish/eval/types"
 )
 
 // Port conveys data stream. It always consists of a byte band and a channel band.
 type Port struct {
 	File      *os.File
-	Chan      chan types.Value
+	Chan      chan interface{}
 	CloseFile bool
 	CloseChan bool
 }
@@ -28,25 +26,16 @@ func (p *Port) Close() {
 		p.File.Close()
 	}
 	if p.CloseChan {
-		// Logger.Printf("closing channel %v", p.Chan)
 		close(p.Chan)
 	}
 }
 
-// ClosePorts closes a list of Ports.
-func ClosePorts(ports []*Port) {
-	for _, port := range ports {
-		// Logger.Printf("closing port %d", i)
-		port.Close()
-	}
-}
-
 var (
 	// ClosedChan is a closed channel, suitable for use as placeholder channel input.
-	ClosedChan = make(chan types.Value)
+	ClosedChan = make(chan interface{})
 	// BlackholeChan is channel writes onto which disappear, suitable for use as
 	// placeholder channel output.
-	BlackholeChan = make(chan types.Value)
+	BlackholeChan = make(chan interface{})
 	// DevNull is /dev/null.
 	DevNull *os.File
 	// DevNullClosedInput is a port made up from DevNull and ClosedChan,
diff -pruN 0.11+ds1-5/eval/purely_eval.go 0.12+ds1-1/eval/purely_eval.go
--- 0.11+ds1-5/eval/purely_eval.go	2018-01-18 02:03:11.000000000 +0000
+++ 0.12+ds1-1/eval/purely_eval.go	2018-07-11 02:32:25.000000000 +0000
@@ -4,7 +4,6 @@ import (
 	"errors"
 	"strings"
 
-	"github.com/elves/elvish/eval/types"
 	"github.com/elves/elvish/parse"
 	"github.com/elves/elvish/util"
 )
@@ -36,8 +35,8 @@ func (ev *Evaler) PurelyEvalPartialCompo
 				return "", ErrImpure
 			}
 			v := ev.PurelyEvalPrimary(in.Head)
-			if s, ok := v.(types.String); ok {
-				head += string(s)
+			if s, ok := v.(string); ok {
+				head += s
 			} else {
 				return "", ErrImpure
 			}
@@ -68,12 +67,12 @@ func (ev *Evaler) PurelyEvalPartialCompo
 // If this cannot be done, it returns nil.
 //
 // Currently, only string literals and variables with no @ can be evaluated.
-func (ev *Evaler) PurelyEvalPrimary(pn *parse.Primary) types.Value {
+func (ev *Evaler) PurelyEvalPrimary(pn *parse.Primary) interface{} {
 	switch pn.Type {
 	case parse.Bareword, parse.SingleQuoted, parse.DoubleQuoted:
-		return types.String(pn.Value)
+		return pn.Value
 	case parse.Variable:
-		explode, ns, name := ParseVariable(pn.Value)
+		explode, ns, name := ParseVariableRef(pn.Value)
 		if explode {
 			return nil
 		}
diff -pruN 0.11+ds1-5/eval/pwd.go 0.12+ds1-1/eval/pwd.go
--- 0.11+ds1-5/eval/pwd.go	2018-01-18 02:03:11.000000000 +0000
+++ 0.12+ds1-1/eval/pwd.go	2018-07-11 02:32:25.000000000 +0000
@@ -3,28 +3,27 @@ package eval
 import (
 	"os"
 
-	"github.com/elves/elvish/eval/types"
-	"github.com/elves/elvish/eval/vartypes"
+	"github.com/elves/elvish/eval/vars"
 )
 
 // PwdVariable is a variable whose value always reflects the current working
 // directory. Setting it changes the current working directory.
 type PwdVariable struct {
-	store AddDirer
+	ev *Evaler
 }
 
-var _ vartypes.Variable = PwdVariable{}
+var _ vars.Var = PwdVariable{}
 
-func (PwdVariable) Get() types.Value {
+func (PwdVariable) Get() interface{} {
 	pwd, err := os.Getwd()
 	maybeThrow(err)
-	return types.String(pwd)
+	return pwd
 }
 
-func (pwd PwdVariable) Set(v types.Value) error {
-	path, ok := v.(types.String)
+func (pwd PwdVariable) Set(v interface{}) error {
+	path, ok := v.(string)
 	if !ok {
 		return ErrPathMustBeString
 	}
-	return Chdir(string(path), pwd.store)
+	return pwd.ev.Chdir(path)
 }
diff -pruN 0.11+ds1-5/eval/raw_options.go 0.12+ds1-1/eval/raw_options.go
--- 0.11+ds1-5/eval/raw_options.go	1970-01-01 00:00:00.000000000 +0000
+++ 0.12+ds1-1/eval/raw_options.go	2018-07-11 02:32:25.000000000 +0000
@@ -0,0 +1,54 @@
+package eval
+
+import (
+	"errors"
+	"reflect"
+
+	"github.com/elves/elvish/parse"
+	"github.com/elves/elvish/util"
+)
+
+type RawOptions map[string]interface{}
+
+// Scan takes a pointer to a struct and scan options into it. A field
+// named FieldName corresponds to the option named field-name, unless the field
+// has a explicit "name" tag.
+func (rawOpts RawOptions) Scan(ptr interface{}) {
+	ptrValue := reflect.ValueOf(ptr)
+	if ptrValue.Kind() != reflect.Ptr || ptrValue.Elem().Kind() != reflect.Struct {
+		throwf("internal bug: need struct ptr to scan options, got %T", ptr)
+	}
+	struc := ptrValue.Elem()
+
+	// fieldIdxForOpt maps option name to the index of field in struc.
+	fieldIdxForOpt := make(map[string]int)
+	for i := 0; i < struc.Type().NumField(); i++ {
+		// ignore unexported fields
+		if !struc.Field(i).CanSet() {
+			continue
+		}
+
+		f := struc.Type().Field(i)
+		optName := f.Tag.Get("name")
+		if optName == "" {
+			optName = util.CamelToDashed(f.Name)
+		}
+		fieldIdxForOpt[optName] = i
+	}
+
+	for k, v := range rawOpts {
+		fieldIdx, ok := fieldIdxForOpt[k]
+		if !ok {
+			throwf("unknown option %s", parse.Quote(k))
+		}
+		mustScanToGo(v, struc.Field(fieldIdx).Addr().Interface())
+	}
+}
+
+var ErrNoOptAccepted = errors.New("no option accepted")
+
+func TakeNoOpt(opts map[string]interface{}) {
+	if len(opts) > 0 {
+		throw(ErrNoOptAccepted)
+	}
+}
diff -pruN 0.11+ds1-5/eval/raw_options_test.go 0.12+ds1-1/eval/raw_options_test.go
--- 0.11+ds1-5/eval/raw_options_test.go	1970-01-01 00:00:00.000000000 +0000
+++ 0.12+ds1-1/eval/raw_options_test.go	2018-07-11 02:32:25.000000000 +0000
@@ -0,0 +1,30 @@
+package eval
+
+import "testing"
+
+type opts struct {
+	FooBar string
+	POSIX  bool `name:"posix"`
+	Min    int
+}
+
+var scanTests = []struct {
+	rawOpts  RawOptions
+	preScan  opts
+	postScan opts
+}{
+	{RawOptions{"foo-bar": "lorem ipsum"},
+		opts{}, opts{FooBar: "lorem ipsum"}},
+	{RawOptions{"posix": true},
+		opts{}, opts{POSIX: true}},
+}
+
+func TestScan(t *testing.T) {
+	for _, test := range scanTests {
+		opts := test.preScan
+		test.rawOpts.Scan(&opts)
+		if opts != test.postScan {
+			t.Errorf("Scan %v => %v, want %v", test.rawOpts, opts, test.postScan)
+		}
+	}
+}
diff -pruN 0.11+ds1-5/eval/re/match.go 0.12+ds1-1/eval/re/match.go
--- 0.11+ds1-5/eval/re/match.go	2018-01-18 02:03:11.000000000 +0000
+++ 0.12+ds1-1/eval/re/match.go	2018-07-11 02:32:25.000000000 +0000
@@ -3,27 +3,24 @@ package re
 import (
 	"strconv"
 
-	"github.com/elves/elvish/eval/types"
+	"github.com/elves/elvish/eval/vals"
 	"github.com/xiaq/persistent/vector"
 )
 
 var (
-	matchDescriptor    = types.NewStructDescriptor("text", "start", "end", "groups")
-	submatchDescriptor = types.NewStructDescriptor("text", "start", "end")
+	matchDescriptor    = vals.NewStructDescriptor("text", "start", "end", "groups")
+	submatchDescriptor = vals.NewStructDescriptor("text", "start", "end")
 )
 
-func newMatch(text string, start, end int, groups vector.Vector) *types.Struct {
-	return types.NewStruct(matchDescriptor, []types.Value{
-		types.String(text),
-		types.String(strconv.Itoa(start)),
-		types.String(strconv.Itoa(end)),
-		types.NewList(groups),
+func newMatch(text string, start, end int, groups vector.Vector) *vals.Struct {
+	return vals.NewStruct(matchDescriptor, []interface{}{
+		text, strconv.Itoa(start), strconv.Itoa(end), groups,
 	})
 }
 
-func newSubmatch(text string, start, end int) *types.Struct {
-	return types.NewStruct(submatchDescriptor, []types.Value{
-		types.String(text),
-		types.String(strconv.Itoa(start)),
-		types.String(strconv.Itoa(end))})
+func newSubmatch(text string, start, end int) *vals.Struct {
+	return vals.NewStruct(submatchDescriptor, []interface{}{
+		string(text),
+		string(strconv.Itoa(start)),
+		string(strconv.Itoa(end))})
 }
diff -pruN 0.11+ds1-5/eval/re/re.go 0.12+ds1-1/eval/re/re.go
--- 0.11+ds1-5/eval/re/re.go	2018-01-18 02:03:11.000000000 +0000
+++ 0.12+ds1-1/eval/re/re.go	2018-07-11 02:32:25.000000000 +0000
@@ -6,59 +6,41 @@ import (
 	"regexp"
 
 	"github.com/elves/elvish/eval"
-	"github.com/elves/elvish/eval/types"
+	"github.com/elves/elvish/eval/vals"
 	"github.com/elves/elvish/util"
 	"github.com/xiaq/persistent/vector"
 )
 
-func Ns() eval.Ns {
-	ns := eval.Ns{}
-	eval.AddBuiltinFns(ns, fns...)
-	return ns
-}
-
-var fns = []*eval.BuiltinFn{
-	{"quote", eval.WrapStringToString(regexp.QuoteMeta)},
-	{"match", match},
-	{"find", find},
-	{"replace", replace},
-	{"split", split},
-}
-
-func match(ec *eval.Frame, args []types.Value, opts map[string]types.Value) {
-	out := ec.OutputChan()
-	var (
-		argPattern types.String
-		argSource  types.String
-		optPOSIX   types.Bool
-	)
-	eval.ScanArgs(args, &argPattern, &argSource)
-	eval.ScanOpts(opts, eval.OptToScan{"posix", &optPOSIX, types.Bool(false)})
-
-	pattern := makePattern(argPattern, optPOSIX, types.Bool(false))
-	matched := pattern.MatchString(string(argSource))
-	out <- types.Bool(matched)
-}
+var Ns = eval.NewNs().AddBuiltinFns("re:", fns)
 
-func find(ec *eval.Frame, args []types.Value, opts map[string]types.Value) {
-	out := ec.OutputChan()
-	var (
-		argPattern types.String
-		argSource  types.String
-		optPOSIX   types.Bool
-		optLongest types.Bool
-		optMax     int
-	)
-	eval.ScanArgs(args, &argPattern, &argSource)
-	eval.ScanOpts(opts,
-		eval.OptToScan{"posix", &optPOSIX, types.Bool(false)},
-		eval.OptToScan{"longest", &optLongest, types.Bool(false)},
-		eval.OptToScan{"max", &optMax, types.String("-1")})
+var fns = map[string]interface{}{
+	"quote":   regexp.QuoteMeta,
+	"match":   match,
+	"find":    find,
+	"replace": replace,
+	"split":   split,
+}
+
+func match(rawOpts eval.RawOptions, argPattern, source string) bool {
+	opts := struct{ Posix bool }{}
+	rawOpts.Scan(&opts)
+
+	pattern := makePattern(argPattern, opts.Posix, false)
+	return pattern.MatchString(source)
+}
+
+func find(fm *eval.Frame, rawOpts eval.RawOptions, argPattern, source string) {
+	out := fm.OutputChan()
+	opts := struct {
+		Posix   bool
+		Longest bool
+		Max     int
+	}{Max: -1}
+	rawOpts.Scan(&opts)
 
-	pattern := makePattern(argPattern, optPOSIX, optLongest)
-	source := string(argSource)
+	pattern := makePattern(argPattern, opts.Posix, opts.Longest)
+	matches := pattern.FindAllSubmatchIndex([]byte(source), opts.Max)
 
-	matches := pattern.FindAllSubmatchIndex([]byte(argSource), optMax)
 	for _, match := range matches {
 		start, end := match[0], match[1]
 		groups := vector.Empty
@@ -76,83 +58,70 @@ func find(ec *eval.Frame, args []types.V
 	}
 }
 
-func replace(ec *eval.Frame, args []types.Value, opts map[string]types.Value) {
-	out := ec.OutputChan()
-	var (
-		argPattern types.String
-		argRepl    types.Value
-		argSource  types.String
-		optPOSIX   types.Bool
-		optLongest types.Bool
-		optLiteral types.Bool
-	)
-	eval.ScanArgs(args, &argPattern, &argRepl, &argSource)
-	eval.ScanOpts(opts,
-		eval.OptToScan{"posix", &optPOSIX, types.Bool(false)},
-		eval.OptToScan{"longest", &optLongest, types.Bool(false)},
-		eval.OptToScan{"literal", &optLiteral, types.Bool(false)})
-
-	pattern := makePattern(argPattern, optPOSIX, optLongest)
-
-	var result string
-	if optLiteral {
-		repl, ok := argRepl.(types.String)
+func replace(fm *eval.Frame, rawOpts eval.RawOptions, argPattern string, argRepl interface{}, source string) (string, error) {
+
+	opts := struct {
+		Posix   bool
+		Longest bool
+		Literal bool
+	}{}
+	rawOpts.Scan(&opts)
+
+	pattern := makePattern(argPattern, opts.Posix, opts.Longest)
+
+	if opts.Literal {
+		repl, ok := argRepl.(string)
 		if !ok {
-			throwf("replacement must be string when literal is set, got %s",
-				argRepl.Kind())
+			return "", fmt.Errorf(
+				"replacement must be string when literal is set, got %s",
+				vals.Kind(argRepl))
 		}
-		result = pattern.ReplaceAllLiteralString(string(argSource), string(repl))
+		return pattern.ReplaceAllLiteralString(source, repl), nil
 	} else {
 		switch repl := argRepl.(type) {
-		case types.String:
-			result = pattern.ReplaceAllString(string(argSource), string(repl))
-		case eval.Fn:
+		case string:
+			return pattern.ReplaceAllString(source, repl), nil
+		case eval.Callable:
 			replFunc := func(s string) string {
-				values, err := ec.PCaptureOutput(repl,
-					[]types.Value{types.String(s)}, eval.NoOpts)
+				values, err := fm.CaptureOutput(repl, []interface{}{s}, eval.NoOpts)
 				maybeThrow(err)
 				if len(values) != 1 {
 					throwf("replacement function must output exactly one value, got %d", len(values))
 				}
-				output, ok := values[0].(types.String)
+				output, ok := values[0].(string)
 				if !ok {
-					throwf("replacement function must output one string, got %s", values[0].Kind())
+					throwf("replacement function must output one string, got %s",
+						vals.Kind(values[0]))
 				}
-				return string(output)
+				return output
 			}
-			result = pattern.ReplaceAllStringFunc(string(argSource), replFunc)
+			return pattern.ReplaceAllStringFunc(source, replFunc), nil
 		default:
-			throwf("replacement must be string or function, got %s",
-				argRepl.Kind())
+			return "", fmt.Errorf(
+				"replacement must be string or function, got %s",
+				vals.Kind(argRepl))
 		}
 	}
-	out <- types.String(result)
 }
 
-func split(ec *eval.Frame, args []types.Value, opts map[string]types.Value) {
-	out := ec.OutputChan()
-	var (
-		argPattern types.String
-		argSource  types.String
-		optPOSIX   types.Bool
-		optLongest types.Bool
-		optMax     int
-	)
-	eval.ScanArgs(args, &argPattern, &argSource)
-	eval.ScanOpts(opts,
-		eval.OptToScan{"posix", &optPOSIX, types.Bool(false)},
-		eval.OptToScan{"longest", &optLongest, types.Bool(false)},
-		eval.OptToScan{"max", &optMax, types.String("-1")})
+func split(fm *eval.Frame, rawOpts eval.RawOptions, argPattern, source string) {
+	out := fm.OutputChan()
+	opts := struct {
+		Posix   bool
+		Longest bool
+		Max     int
+	}{Max: -1}
+	rawOpts.Scan(&opts)
 
-	pattern := makePattern(argPattern, optPOSIX, optLongest)
+	pattern := makePattern(argPattern, opts.Posix, opts.Longest)
 
-	pieces := pattern.Split(string(argSource), optMax)
+	pieces := pattern.Split(source, opts.Max)
 	for _, piece := range pieces {
-		out <- types.String(piece)
+		out <- piece
 	}
 }
 
-func makePattern(argPattern types.String, optPOSIX, optLongest types.Bool) *regexp.Regexp {
+func makePattern(argPattern string, optPOSIX, optLongest bool) *regexp.Regexp {
 	var (
 		pattern *regexp.Regexp
 		err     error
diff -pruN 0.11+ds1-5/eval/re/re_test.go 0.12+ds1-1/eval/re/re_test.go
--- 0.11+ds1-5/eval/re/re_test.go	2018-01-18 02:03:11.000000000 +0000
+++ 0.12+ds1-1/eval/re/re_test.go	2018-07-11 02:32:25.000000000 +0000
@@ -4,41 +4,36 @@ import (
 	"testing"
 
 	"github.com/elves/elvish/eval"
-	"github.com/elves/elvish/eval/vartypes"
 	"github.com/xiaq/persistent/vector"
 )
 
-var tests = []eval.Test{
-	eval.NewTest("re:match . xyz").WantOutBools(true),
-	eval.NewTest("re:match . ''").WantOutBools(false),
-	eval.NewTest("re:match '[a-z]' A").WantOutBools(false),
-
-	eval.NewTest("re:find . ab").WantOut(
-		newMatch("a", 0, 1, vector.Empty.Cons(newSubmatch("a", 0, 1))),
-		newMatch("b", 1, 2, vector.Empty.Cons(newSubmatch("b", 1, 2))),
-	),
-	eval.NewTest("re:find '[A-Z]([0-9])' 'A1 B2'").WantOut(
-		newMatch("A1", 0, 2,
-			vector.Empty.Cons(newSubmatch("A1", 0, 2)).Cons(newSubmatch("1", 1, 2))),
-		newMatch("B2", 3, 5,
-			vector.Empty.Cons(newSubmatch("B2", 3, 5)).Cons(newSubmatch("2", 4, 5))),
-	),
-
-	eval.NewTest("re:replace '(ba|z)sh' '${1}SH' 'bash and zsh'").WantOutStrings("baSH and zSH"),
-	eval.NewTest("re:replace &literal '(ba|z)sh' '$sh' 'bash and zsh'").WantOutStrings("$sh and $sh"),
-	eval.NewTest("re:replace '(ba|z)sh' [x]{ put [&bash=BaSh &zsh=ZsH][$x] } 'bash and zsh'").WantOutStrings("BaSh and ZsH"),
-
-	eval.NewTest("re:split : /usr/sbin:/usr/bin:/bin").WantOutStrings("/usr/sbin", "/usr/bin", "/bin"),
-	eval.NewTest("re:split &max=2 : /usr/sbin:/usr/bin:/bin").WantOutStrings("/usr/sbin", "/usr/bin:/bin"),
-
-	eval.NewTest("re:quote a.txt").WantOutStrings(`a\.txt`),
-	eval.NewTest("re:quote '(*)'").WantOutStrings(`\(\*\)`),
-}
-
 func TestRe(t *testing.T) {
-	eval.RunTests(t, tests, func() *eval.Evaler {
-		ev := eval.NewEvaler()
-		ev.Builtin["re"+eval.NsSuffix] = vartypes.NewRo(Ns())
-		return ev
-	})
+	setup := func(ev *eval.Evaler) { ev.Builtin.AddNs("re", Ns) }
+	That := eval.That
+	eval.TestWithSetup(t, setup,
+		That("re:match . xyz").Puts(true),
+		That("re:match . ''").Puts(false),
+		That("re:match '[a-z]' A").Puts(false),
+
+		That("re:find . ab").Puts(
+			newMatch("a", 0, 1, vector.Empty.Cons(newSubmatch("a", 0, 1))),
+			newMatch("b", 1, 2, vector.Empty.Cons(newSubmatch("b", 1, 2))),
+		),
+		That("re:find '[A-Z]([0-9])' 'A1 B2'").Puts(
+			newMatch("A1", 0, 2,
+				vector.Empty.Cons(newSubmatch("A1", 0, 2)).Cons(newSubmatch("1", 1, 2))),
+			newMatch("B2", 3, 5,
+				vector.Empty.Cons(newSubmatch("B2", 3, 5)).Cons(newSubmatch("2", 4, 5))),
+		),
+
+		That("re:replace '(ba|z)sh' '${1}SH' 'bash and zsh'").Puts("baSH and zSH"),
+		That("re:replace &literal '(ba|z)sh' '$sh' 'bash and zsh'").Puts("$sh and $sh"),
+		That("re:replace '(ba|z)sh' [x]{ put [&bash=BaSh &zsh=ZsH][$x] } 'bash and zsh'").Puts("BaSh and ZsH"),
+
+		That("re:split : /usr/sbin:/usr/bin:/bin").Puts("/usr/sbin", "/usr/bin", "/bin"),
+		That("re:split &max=2 : /usr/sbin:/usr/bin:/bin").Puts("/usr/sbin", "/usr/bin:/bin"),
+
+		That("re:quote a.txt").Puts(`a\.txt`),
+		That("re:quote '(*)'").Puts(`\(\*\)`),
+	)
 }
diff -pruN 0.11+ds1-5/eval/resolve.go 0.12+ds1-1/eval/resolve.go
--- 0.11+ds1-5/eval/resolve.go	2018-01-18 02:03:11.000000000 +0000
+++ 0.12+ds1-1/eval/resolve.go	2018-07-11 02:32:25.000000000 +0000
@@ -1,19 +1,146 @@
 package eval
 
-func resolve(s string, ec *Frame) Fn {
-	// NOTE: This needs to be kept in sync with the resolution algorithm used in
-	// (*compiler).form.
-
-	// Try variable
-	explode, ns, name := ParseVariable(string(s))
-	if !explode {
-		if v := ec.ResolveVar(ns, name+FnSuffix); v != nil {
-			if caller, ok := v.Get().(Fn); ok {
-				return caller
+import (
+	"os"
+	"strings"
+
+	"github.com/elves/elvish/eval/vars"
+)
+
+// Resolution and iteration of variables and namespaces.
+
+// EachVariableInTop calls the passed function for each variable name in
+// namespace ns that can be found from the top context.
+func (ev *evalerScopes) EachVariableInTop(ns string, f func(s string)) {
+	switch ns {
+	case "builtin":
+		for name := range ev.Builtin {
+			f(name)
+		}
+	case "":
+		for name := range ev.Global {
+			f(name)
+		}
+		for name := range ev.Builtin {
+			f(name)
+		}
+	case "e":
+		EachExternal(func(cmd string) {
+			f(cmd + FnSuffix)
+		})
+	case "E":
+		for _, s := range os.Environ() {
+			if i := strings.IndexByte(s, '='); i > 0 {
+				f(s[:i])
 			}
 		}
+	default:
+		segs := splitQName(ns + NsSuffix)
+		mod := ev.Global[segs[0]]
+		if mod == nil {
+			mod = ev.Builtin[segs[0]]
+		}
+		for _, seg := range segs[1:] {
+			if mod == nil {
+				return
+			}
+			mod = mod.Get().(Ns)[seg]
+		}
+		if mod != nil {
+			for name := range mod.Get().(Ns) {
+				f(name)
+			}
+		}
+	}
+}
+
+// EachNsInTop calls the passed function for each namespace that can be used
+// from the top context.
+func (ev *evalerScopes) EachNsInTop(f func(s string)) {
+	f("builtin")
+	f("e")
+	f("E")
+
+	for name := range ev.Global {
+		if strings.HasSuffix(name, NsSuffix) {
+			f(name[:len(name)-len(NsSuffix)])
+		}
+	}
+	for name := range ev.Builtin {
+		if strings.HasSuffix(name, NsSuffix) {
+			f(name[:len(name)-len(NsSuffix)])
+		}
+	}
+}
+
+// ResolveVar resolves a variable. When the variable cannot be found, nil is
+// returned.
+func (fm *Frame) ResolveVar(n, name string) vars.Var {
+	if n == "" {
+		return fm.resolveUnqualified(name)
+	}
+
+	// TODO: Let this function accept the fully qualified name.
+	segs := splitQName(n + ":" + name)
+
+	var ns Ns
+
+	switch segs[0] {
+	case "e:":
+		if len(segs) == 2 && strings.HasSuffix(segs[1], FnSuffix) {
+			return vars.NewRo(ExternalCmd{Name: segs[1][:len(segs[1])-len(FnSuffix)]})
+		}
+		return nil
+	case "E:":
+		if len(segs) == 2 {
+			return vars.NewEnv(segs[1])
+		}
+		return nil
+	case "local:":
+		ns = fm.local
+	case "up:":
+		ns = fm.up
+	case "builtin:":
+		ns = fm.Builtin
+	default:
+		v := fm.resolveUnqualified(segs[0])
+		if v == nil {
+			return nil
+		}
+		ns = v.Get().(Ns)
 	}
 
-	// External command
-	return ExternalCmd{string(s)}
+	for _, seg := range segs[1 : len(segs)-1] {
+		v := ns[seg]
+		if v == nil {
+			return nil
+		}
+		ns = v.Get().(Ns)
+	}
+	return ns[segs[len(segs)-1]]
+}
+
+func splitQName(qname string) []string {
+	i := 0
+	var segs []string
+	for i < len(qname) {
+		j := strings.IndexByte(qname[i:], ':')
+		if j == -1 {
+			segs = append(segs, qname[i:])
+			break
+		}
+		segs = append(segs, qname[i:i+j+1])
+		i += j + 1
+	}
+	return segs
+}
+
+func (fm *Frame) resolveUnqualified(name string) vars.Var {
+	if v, ok := fm.local[name]; ok {
+		return v
+	}
+	if v, ok := fm.up[name]; ok {
+		return v
+	}
+	return fm.Builtin[name]
 }
diff -pruN 0.11+ds1-5/eval/resolve_test.go 0.12+ds1-1/eval/resolve_test.go
--- 0.11+ds1-5/eval/resolve_test.go	1970-01-01 00:00:00.000000000 +0000
+++ 0.12+ds1-1/eval/resolve_test.go	2018-07-11 02:32:25.000000000 +0000
@@ -0,0 +1,18 @@
+package eval
+
+import (
+	"testing"
+
+	"github.com/elves/elvish/tt"
+)
+
+var splitQNameTests = tt.Table{
+	tt.Args("a").Rets([]string{"a"}),
+	tt.Args("a:b").Rets([]string{"a:", "b"}),
+	tt.Args("a:b:").Rets([]string{"a:", "b:"}),
+	tt.Args("a:b:c:d").Rets([]string{"a:", "b:", "c:", "d"}),
+}
+
+func TestSplitQName(t *testing.T) {
+	tt.Test(t, tt.Fn("splitQName", splitQName), splitQNameTests)
+}
diff -pruN 0.11+ds1-5/eval/scan.go 0.12+ds1-1/eval/scan.go
--- 0.11+ds1-5/eval/scan.go	2018-01-18 02:03:11.000000000 +0000
+++ 0.12+ds1-1/eval/scan.go	1970-01-01 00:00:00.000000000 +0000
@@ -1,134 +0,0 @@
-package eval
-
-// This file implements facilities for "scanning" arguments and options.
-
-import (
-	"reflect"
-
-	"github.com/elves/elvish/eval/types"
-	"github.com/elves/elvish/parse"
-	"github.com/elves/elvish/util"
-)
-
-// ScanArgs scans arguments into pointers to supported argument types. If the
-// arguments cannot be scanned, an error is thrown.
-func ScanArgs(src []types.Value, dstPtrs ...interface{}) {
-	if len(src) != len(dstPtrs) {
-		throwf("arity mistmatch: want %d arguments, got %d", len(dstPtrs), len(src))
-	}
-	for i, value := range src {
-		scanValueToGo(value, dstPtrs[i])
-	}
-}
-
-// ScanArgsVariadic is like ScanArgs, but the last element of args should be a
-// pointer to a slice, and the rest of arguments will be scanned into it.
-func ScanArgsVariadic(src []types.Value, dstPtrs ...interface{}) {
-	if len(src) < len(dstPtrs)-1 {
-		throwf("arity mistmatch: want at least %d arguments, got %d", len(dstPtrs)-1, len(src))
-	}
-	ScanArgs(src[:len(dstPtrs)-1], dstPtrs[:len(dstPtrs)-1]...)
-
-	// Scan the rest of arguments into a slice.
-	rest := src[len(dstPtrs)-1:]
-	restDst := reflect.ValueOf(dstPtrs[len(dstPtrs)-1])
-	if restDst.Kind() != reflect.Ptr || restDst.Elem().Kind() != reflect.Slice {
-		throwf("internal bug: %T to ScanArgsVariadic, need pointer to slice", dstPtrs[len(dstPtrs)-1])
-	}
-	scanned := reflect.MakeSlice(restDst.Elem().Type(), len(rest), len(rest))
-	for i, value := range rest {
-		scanValueToGo(value, scanned.Index(i).Addr().Interface())
-	}
-	reflect.Indirect(restDst).Set(scanned)
-}
-
-// ScanArgsOptionalInput is like ScanArgs, but the argument can contain an
-// optional iterable value at the end containing inputs to the function. The
-// return value is a function that iterates the iterable value if it exists, or
-// the input otherwise.
-func ScanArgsOptionalInput(ec *Frame, src []types.Value, dstArgs ...interface{}) func(func(types.Value)) {
-	switch len(src) {
-	case len(dstArgs):
-		ScanArgs(src, dstArgs...)
-		return ec.IterateInputs
-	case len(dstArgs) + 1:
-		ScanArgs(src[:len(dstArgs)], dstArgs...)
-		value := src[len(dstArgs)]
-		iterable, ok := value.(types.Iterator)
-		if !ok {
-			throwf("need iterable argument, got %s", value.Kind())
-		}
-		return func(f func(types.Value)) {
-			iterable.Iterate(func(v types.Value) bool {
-				f(v)
-				return true
-			})
-		}
-	default:
-		throwf("arity mistmatch: want %d or %d arguments, got %d", len(dstArgs), len(dstArgs)+1, len(src))
-		return nil
-	}
-}
-
-// OptToScan is a data structure for an option that is intended to be used in
-// ScanOpts.
-type OptToScan struct {
-	Name    string
-	Ptr     interface{}
-	Default types.Value
-}
-
-// ScanOpts scans options from a map.
-func ScanOpts(m map[string]types.Value, opts ...OptToScan) {
-	scanned := make(map[string]bool)
-	for _, opt := range opts {
-		a := opt.Ptr
-		value, ok := m[opt.Name]
-		if !ok {
-			value = opt.Default
-		}
-		scanValueToGo(value, a)
-		scanned[opt.Name] = true
-	}
-	for key := range m {
-		if !scanned[key] {
-			throwf("unknown option %s", parse.Quote(key))
-		}
-	}
-}
-
-// ScanOptsToStruct scan options from a map like ScanOpts except the destination
-// is a struct whose fields correspond to the options to be parsed. A field
-// named FieldName corresponds to the option named field-name, unless the field
-// has a explicit "name" tag.
-func ScanOptsToStruct(m map[string]types.Value, structPtr interface{}) {
-	ptrValue := reflect.ValueOf(structPtr)
-	if ptrValue.Kind() != reflect.Ptr || ptrValue.Elem().Kind() != reflect.Struct {
-		throwf("internal bug: need struct ptr for ScanOptsToStruct, got %T", structPtr)
-	}
-	struc := ptrValue.Elem()
-
-	// fieldIdxForOpt maps option name to the index of field in struc.
-	fieldIdxForOpt := make(map[string]int)
-	for i := 0; i < struc.Type().NumField(); i++ {
-		// ignore unexported fields
-		if !struc.Field(i).CanSet() {
-			continue
-		}
-
-		f := struc.Type().Field(i)
-		optName := f.Tag.Get("name")
-		if optName == "" {
-			optName = util.CamelToDashed(f.Name)
-		}
-		fieldIdxForOpt[optName] = i
-	}
-
-	for k, v := range m {
-		fieldIdx, ok := fieldIdxForOpt[k]
-		if !ok {
-			throwf("unknown option %s", parse.Quote(k))
-		}
-		scanValueToGo(v, struc.Field(fieldIdx).Addr().Interface())
-	}
-}
diff -pruN 0.11+ds1-5/eval/scan_test.go 0.12+ds1-1/eval/scan_test.go
--- 0.11+ds1-5/eval/scan_test.go	2018-01-18 02:03:11.000000000 +0000
+++ 0.12+ds1-1/eval/scan_test.go	1970-01-01 00:00:00.000000000 +0000
@@ -1,213 +0,0 @@
-package eval
-
-import (
-	"reflect"
-	"testing"
-
-	"github.com/elves/elvish/eval/types"
-	"github.com/elves/elvish/util"
-)
-
-// These are used as arguments to scanArg, since Go does not allow taking
-// address of literals.
-
-func intPtr() *int             { var x int; return &x }
-func intsPtr() *[]int          { var x []int; return &x }
-func float64Ptr() *float64     { var x float64; return &x }
-func stringPtr() *types.String { var x types.String; return &x }
-
-var scanArgsTestCases = []struct {
-	variadic bool
-	src      []types.Value
-	dstPtrs  []interface{}
-	want     []interface{}
-	bad      bool
-}{
-	// Scanning an int and a String
-	{
-		src:     []types.Value{types.String("20"), types.String("20")},
-		dstPtrs: []interface{}{intPtr(), stringPtr()},
-		want:    []interface{}{20, types.String("20")},
-	},
-	// Scanning a String and any number of ints (here 2)
-	{
-		variadic: true,
-		src:      []types.Value{types.String("a"), types.String("1"), types.String("2")},
-		dstPtrs:  []interface{}{stringPtr(), intsPtr()},
-		want:     []interface{}{types.String("a"), []int{1, 2}},
-	},
-	// Scanning a String and any number of ints (here 0)
-	{
-		variadic: true,
-		src:      []types.Value{types.String("a")},
-		dstPtrs:  []interface{}{stringPtr(), intsPtr()},
-		want:     []interface{}{types.String("a"), []int{}},
-	},
-
-	// Arity mismatch: too few arguments (non-variadic)
-	{
-		bad:     true,
-		src:     []types.Value{},
-		dstPtrs: []interface{}{stringPtr()},
-	},
-	// Arity mismatch: too few arguments (non-variadic)
-	{
-		bad:     true,
-		src:     []types.Value{types.String("")},
-		dstPtrs: []interface{}{stringPtr(), intPtr()},
-	},
-	// Arity mismatch: too many arguments (non-variadic)
-	{
-		bad:     true,
-		src:     []types.Value{types.String("1"), types.String("2")},
-		dstPtrs: []interface{}{stringPtr()},
-	},
-	// Type mismatch (nonvariadic)
-	{
-		bad:     true,
-		src:     []types.Value{types.String("x")},
-		dstPtrs: []interface{}{intPtr()},
-	},
-
-	// Arity mismatch: too few arguments (variadic)
-	{
-		bad:      true,
-		src:      []types.Value{},
-		dstPtrs:  []interface{}{stringPtr(), intsPtr()},
-		variadic: true,
-	},
-	// Type mismatch within rest arg
-	{
-		bad:      true,
-		src:      []types.Value{types.String("a"), types.String("1"), types.String("lorem")},
-		dstPtrs:  []interface{}{stringPtr(), intsPtr()},
-		variadic: true,
-	},
-}
-
-func TestScanArgs(t *testing.T) {
-	for _, tc := range scanArgsTestCases {
-		funcToTest := ScanArgs
-		if tc.variadic {
-			funcToTest = ScanArgsVariadic
-		}
-
-		if tc.bad {
-			if !util.ThrowsAny(func() { funcToTest(tc.src, tc.dstPtrs...) }) {
-				t.Errorf("ScanArgs(%v, %v) should throw an error", tc.src, tc.dstPtrs)
-			}
-			continue
-		}
-
-		funcToTest(tc.src, tc.dstPtrs...)
-		dsts := make([]interface{}, len(tc.dstPtrs))
-		for i, ptr := range tc.dstPtrs {
-			dsts[i] = indirect(ptr)
-		}
-		err := compareSlice(tc.want, dsts)
-		if err != nil {
-			t.Errorf("ScanArgs(%v) got %v, want %v", tc.src, dsts, tc.want)
-		}
-	}
-}
-
-var scanArgsBadTestCases = []struct {
-	variadic bool
-	src      []types.Value
-	dstPtrs  []interface{}
-}{}
-
-var scanOptsTestCases = []struct {
-	bad  bool
-	src  map[string]types.Value
-	opts []OptToScan
-	want []interface{}
-}{
-	{
-		src: map[string]types.Value{
-			"foo": types.String("bar"),
-		},
-		opts: []OptToScan{
-			{"foo", stringPtr(), types.String("haha")},
-		},
-		want: []interface{}{
-			types.String("bar"),
-		},
-	},
-	// Default values.
-	{
-		src: map[string]types.Value{
-			"foo": types.String("bar"),
-		},
-		opts: []OptToScan{
-			{"foo", stringPtr(), types.String("haha")},
-			{"lorem", stringPtr(), types.String("ipsum")},
-		},
-		want: []interface{}{
-			types.String("bar"),
-			types.String("ipsum"),
-		},
-	},
-
-	// Unknown option
-	{
-		bad: true,
-		src: map[string]types.Value{
-			"foo": types.String("bar"),
-		},
-		opts: []OptToScan{
-			{"lorem", stringPtr(), types.String("ipsum")},
-		},
-	},
-}
-
-func TestScanOpts(t *testing.T) {
-	for _, tc := range scanOptsTestCases {
-		if tc.bad {
-			if !util.ThrowsAny(func() { ScanOpts(tc.src, tc.opts...) }) {
-				t.Errorf("ScanOpts(%v, %v...) should throw an error",
-					tc.src, tc.opts)
-			}
-			continue
-		}
-
-		ScanOpts(tc.src, tc.opts...)
-		dsts := make([]interface{}, len(tc.opts))
-		for i, opt := range tc.opts {
-			dsts[i] = indirect(opt.Ptr)
-		}
-		err := compareSlice(tc.want, dsts)
-		if err != nil {
-			t.Errorf("ScanOpts(%v, %v) got %v, want %v", tc.src, tc.opts,
-				dsts, tc.want)
-		}
-	}
-}
-
-var scanArgTestCases = []struct {
-	source  types.Value
-	destPtr interface{}
-	want    interface{}
-}{
-	{types.String("20"), intPtr(), 20},
-	{types.String("0x20"), intPtr(), 0x20},
-	{types.String("20"), float64Ptr(), 20.0},
-	{types.String("23.33"), float64Ptr(), 23.33},
-	{types.String(""), stringPtr(), types.String("")},
-	{types.String("1"), stringPtr(), types.String("1")},
-	{types.String("2.33"), stringPtr(), types.String("2.33")},
-}
-
-func TestScanArg(t *testing.T) {
-	for _, tc := range scanArgTestCases {
-		scanValueToGo(tc.source, tc.destPtr)
-		if !equals(indirect(tc.destPtr), tc.want) {
-			t.Errorf("scanArg(%s) got %q, want %v", tc.source,
-				indirect(tc.destPtr), tc.want)
-		}
-	}
-}
-
-func indirect(i interface{}) interface{} {
-	return reflect.Indirect(reflect.ValueOf(i)).Interface()
-}
diff -pruN 0.11+ds1-5/eval/shared.go 0.12+ds1-1/eval/shared.go
--- 0.11+ds1-5/eval/shared.go	2018-01-18 02:03:11.000000000 +0000
+++ 0.12+ds1-1/eval/shared.go	1970-01-01 00:00:00.000000000 +0000
@@ -1,21 +0,0 @@
-package eval
-
-import (
-	"github.com/elves/elvish/daemon"
-	"github.com/elves/elvish/eval/types"
-)
-
-type sharedVariable struct {
-	store *daemon.Client
-	name  string
-}
-
-func (sv sharedVariable) Set(val types.Value) error {
-	return sv.store.SetSharedVar(sv.name, types.ToString(val))
-}
-
-func (sv sharedVariable) Get() types.Value {
-	value, err := sv.store.SharedVar(sv.name)
-	maybeThrow(err)
-	return types.String(value)
-}
diff -pruN 0.11+ds1-5/eval/source.go 0.12+ds1-1/eval/source.go
--- 0.11+ds1-5/eval/source.go	2018-01-18 02:03:11.000000000 +0000
+++ 0.12+ds1-1/eval/source.go	2018-07-11 02:32:25.000000000 +0000
@@ -4,7 +4,7 @@ import (
 	"fmt"
 	"strconv"
 
-	"github.com/elves/elvish/eval/types"
+	"github.com/elves/elvish/eval/vals"
 	"github.com/elves/elvish/parse"
 	"github.com/xiaq/persistent/hash"
 )
@@ -45,8 +45,7 @@ func (src *Source) describePath() string
 }
 
 var (
-	_ types.Value      = (*Source)(nil)
-	_ types.IndexOneer = (*Source)(nil)
+	_ vals.Indexer = (*Source)(nil)
 )
 
 func (src *Source) Kind() string {
@@ -70,20 +69,25 @@ func (src *Source) Repr(int) string {
 		src.typ, parse.Quote(src.name), parse.Quote(src.path))
 }
 
-func (src *Source) IndexOne(k types.Value) types.Value {
+func (src *Source) Index(k interface{}) (interface{}, bool) {
+	ret := ""
 	switch k {
-	case types.String("type"):
-		return types.String(src.typ.String())
-	case types.String("name"):
-		return types.String(src.name)
-	case types.String("path"):
-		return types.String(src.path)
-	case types.String("code"):
-		return types.String(src.code)
+	case "type":
+		ret = src.typ.String()
+	case "name":
+		ret = src.name
+	case "path":
+		ret = src.path
+	case "code":
+		ret = src.code
 	default:
-		throwf("No such key %s", k.Repr(types.NoPretty))
-		panic("unreachable")
+		return nil, false
 	}
+	return ret, true
+}
+
+func (src *Source) IterateKeys(f func(interface{}) bool) {
+	vals.Feed(f, "type", "name", "path", "code")
 }
 
 // SrcType records the type of a piece of source code.
diff -pruN 0.11+ds1-5/eval/stacktrace.go 0.12+ds1-1/eval/stacktrace.go
--- 0.11+ds1-5/eval/stacktrace.go	1970-01-01 00:00:00.000000000 +0000
+++ 0.12+ds1-1/eval/stacktrace.go	2018-07-11 02:32:25.000000000 +0000
@@ -0,0 +1,11 @@
+package eval
+
+import "github.com/elves/elvish/util"
+
+// stackTrace represents a stack trace as a linked list. Inner stacks appear
+// first. Since pipelines can call multiple functions in parallel, all the
+// stackTrace nodes form a DAG.
+type stackTrace struct {
+	entry *util.SourceRange
+	next  *stackTrace
+}
diff -pruN 0.11+ds1-5/eval/std_ports.go 0.12+ds1-1/eval/std_ports.go
--- 0.11+ds1-5/eval/std_ports.go	1970-01-01 00:00:00.000000000 +0000
+++ 0.12+ds1-1/eval/std_ports.go	2018-07-11 02:32:25.000000000 +0000
@@ -0,0 +1,52 @@
+package eval
+
+import (
+	"os"
+	"sync"
+
+	"github.com/elves/elvish/eval/vals"
+)
+
+const (
+	stdoutChanSize = 32
+	stderrChanSize = 32
+)
+
+type stdPorts struct {
+	ports       [3]*Port
+	relayerWait *sync.WaitGroup
+}
+
+func newStdPorts(stdin, stdout, stderr *os.File, prefix string) stdPorts {
+	stdoutChan := make(chan interface{}, stdoutChanSize)
+	stderrChan := make(chan interface{}, stderrChanSize)
+
+	relayerWait := new(sync.WaitGroup)
+	relayerWait.Add(2)
+	go relayChanToFile(stdoutChan, stdout, prefix, relayerWait)
+	go relayChanToFile(stderrChan, stderr, prefix, relayerWait)
+
+	return stdPorts{
+		[3]*Port{
+			{File: stdin, Chan: ClosedChan},
+			{File: stdout, Chan: stdoutChan, CloseChan: true},
+			{File: stderr, Chan: stderrChan, CloseChan: true},
+		},
+		relayerWait,
+	}
+}
+
+func relayChanToFile(ch <-chan interface{}, file *os.File, prefix string, w *sync.WaitGroup) {
+	for v := range ch {
+		file.WriteString(prefix)
+		file.WriteString(vals.Repr(v, initIndent))
+		file.WriteString("\n")
+	}
+	w.Done()
+}
+
+func (ep *stdPorts) close() {
+	ep.ports[1].Close()
+	ep.ports[2].Close()
+	ep.relayerWait.Wait()
+}
diff -pruN 0.11+ds1-5/eval/std_ports_test.go 0.12+ds1-1/eval/std_ports_test.go
--- 0.11+ds1-5/eval/std_ports_test.go	1970-01-01 00:00:00.000000000 +0000
+++ 0.12+ds1-1/eval/std_ports_test.go	2018-07-11 02:32:25.000000000 +0000
@@ -0,0 +1,53 @@
+package eval
+
+import (
+	"io"
+	"io/ioutil"
+	"os"
+	"testing"
+)
+
+func TestEvalerPorts(t *testing.T) {
+	stdoutReader, stdout := mustPipe()
+	defer stdoutReader.Close()
+
+	stderrReader, stderr := mustPipe()
+	defer stderrReader.Close()
+
+	prefix := "> "
+	ep := newStdPorts(DevNull, stdout, stderr, prefix)
+	ep.ports[1].Chan <- "x"
+	ep.ports[1].Chan <- "y"
+	ep.ports[2].Chan <- "bad"
+	ep.ports[2].Chan <- "err"
+	ep.close()
+	stdout.Close()
+	stderr.Close()
+
+	stdoutAll := mustReadAllString(stdoutReader)
+	wantStdoutAll := "> x\n> y\n"
+	if stdoutAll != wantStdoutAll {
+		t.Errorf("stdout is %q, want %q", stdoutAll, wantStdoutAll)
+	}
+	stderrAll := mustReadAllString(stderrReader)
+	wantStderrAll := "> bad\n> err\n"
+	if stderrAll != wantStderrAll {
+		t.Errorf("stderr is %q, want %q", stderrAll, wantStderrAll)
+	}
+}
+
+func mustPipe() (*os.File, *os.File) {
+	r, w, err := os.Pipe()
+	if err != nil {
+		panic(err)
+	}
+	return r, w
+}
+
+func mustReadAllString(r io.Reader) string {
+	b, err := ioutil.ReadAll(r)
+	if err != nil {
+		panic(err)
+	}
+	return string(b)
+}
diff -pruN 0.11+ds1-5/eval/store/store.go 0.12+ds1-1/eval/store/store.go
--- 0.11+ds1-5/eval/store/store.go	1970-01-01 00:00:00.000000000 +0000
+++ 0.12+ds1-1/eval/store/store.go	2018-07-11 02:32:25.000000000 +0000
@@ -0,0 +1,13 @@
+package store
+
+import (
+	"github.com/elves/elvish/eval"
+	"github.com/elves/elvish/store/storedefs"
+)
+
+func Ns(s storedefs.Store) eval.Ns {
+	return eval.NewNs().AddBuiltinFns("store:", map[string]interface{}{
+		"del-dir": s.DelDir,
+		"del-cmd": s.DelCmd,
+	})
+}
diff -pruN 0.11+ds1-5/eval/str/str.go 0.12+ds1-1/eval/str/str.go
--- 0.11+ds1-5/eval/str/str.go	1970-01-01 00:00:00.000000000 +0000
+++ 0.12+ds1-1/eval/str/str.go	2018-07-11 02:32:25.000000000 +0000
@@ -0,0 +1,39 @@
+// Package str exposes functionality from Go's strings package as an Elvish
+// module.
+package str
+
+import (
+	"strings"
+
+	"github.com/elves/elvish/eval"
+)
+
+var Ns = eval.NewNs().AddBuiltinFns("str:", fns)
+
+var fns = map[string]interface{}{
+	"compare":      strings.Compare,
+	"contains":     strings.Contains,
+	"contains-any": strings.ContainsAny,
+	"count":        strings.Count,
+	"equal-fold":   strings.EqualFold,
+	// TODO: Fields, FieldsFunc
+	"has-prefix": strings.HasPrefix,
+	"has-suffix": strings.HasSuffix,
+	"index":      strings.Index,
+	"index-any":  strings.IndexAny,
+	// TODO: IndexFunc, Join
+	"last-index": strings.LastIndex,
+	// TODO: LastIndexFunc, Map, Repeat, Replace, Split, SplitAfter
+	"title":    strings.Title,
+	"to-lower": strings.ToLower,
+	"to-title": strings.ToTitle,
+	"to-upper": strings.ToUpper,
+	// TODO: ToLowerSpecial, ToTitleSpecial, ToUpperSpecial
+	"trim":       strings.Trim,
+	"trim-left":  strings.TrimLeft,
+	"trim-right": strings.TrimRight,
+	// TODO: TrimLeft,Right}Func
+	"trim-space":  strings.TrimSpace,
+	"trim-prefix": strings.TrimPrefix,
+	"trim-suffix": strings.TrimSuffix,
+}
diff -pruN 0.11+ds1-5/eval/testmain_test.go 0.12+ds1-1/eval/testmain_test.go
--- 0.11+ds1-5/eval/testmain_test.go	2018-01-18 02:03:11.000000000 +0000
+++ 0.12+ds1-1/eval/testmain_test.go	1970-01-01 00:00:00.000000000 +0000
@@ -1,121 +0,0 @@
-package eval
-
-import (
-	"io/ioutil"
-	"os"
-	"path/filepath"
-	"sort"
-	"strings"
-	"testing"
-
-	"github.com/elves/elvish/util"
-)
-
-var (
-	filesToCreate = []string{
-		"a1", "a2", "a3", "a10",
-		"b1", "b2", "b3",
-		"c1", "c2",
-		"foo", "bar", "lorem", "ipsum",
-	}
-	dirsToCreate = []string{"dir", "dir2"}
-	fileListing  = getFileListing()
-)
-
-func getFileListing() []string {
-	var x []string
-	x = append(x, filesToCreate...)
-	x = append(x, dirsToCreate...)
-	sort.Strings(x)
-	return x
-}
-
-func getFilesWithPrefix(prefixes ...string) []string {
-	var x []string
-	for _, name := range fileListing {
-		for _, prefix := range prefixes {
-			if strings.HasPrefix(name, prefix) {
-				x = append(x, name)
-				break
-			}
-		}
-	}
-	sort.Strings(x)
-	return x
-}
-
-func getFilesBut(excludes ...string) []string {
-	var x []string
-	for _, name := range fileListing {
-		excluded := false
-		for _, exclude := range excludes {
-			if name == exclude {
-				excluded = true
-				break
-			}
-		}
-		if !excluded {
-			x = append(x, name)
-		}
-	}
-	sort.Strings(x)
-	return x
-}
-
-var mods = map[string]string{
-	"lorem":    "name = lorem; fn put-name { put $name }",
-	"d":        "name = d",
-	"a/b/c/d":  "name = a/b/c/d",
-	"a/b/c/x":  "use ./d; d = $d:name; use ../../../lorem; lorem = $lorem:name",
-	"has/init": "put has/init",
-}
-
-var libDir string
-
-func TestMain(m *testing.M) {
-	var exitCode int
-	util.InTempDir(func(tmpHome string) {
-		oldHome := os.Getenv("HOME")
-		os.Setenv("HOME", tmpHome)
-		defer os.Setenv("HOME", oldHome)
-
-		for _, filename := range filesToCreate {
-			file, err := os.Create(filename)
-			if err != nil {
-				panic(err)
-			}
-			file.Close()
-		}
-
-		for _, dirname := range dirsToCreate {
-			err := os.Mkdir(dirname, 0700)
-			if err != nil {
-				panic(err)
-			}
-		}
-
-		util.WithTempDir(func(dir string) {
-			libDir = dir
-
-			for mod, content := range mods {
-				fname := filepath.Join(libDir, mod+".elv")
-				os.MkdirAll(filepath.Dir(fname), 0700)
-				err := ioutil.WriteFile(fname, []byte(content), 0600)
-				if err != nil {
-					panic(err)
-				}
-			}
-
-			exitCode = m.Run()
-		})
-	})
-	os.Exit(exitCode)
-}
-
-func runTests(t *testing.T, tests []Test) {
-	RunTests(t, tests, func() *Evaler {
-		ev := NewEvaler()
-		ev.SetLibDir(libDir)
-		return ev
-	})
-}
diff -pruN 0.11+ds1-5/eval/testutils.go 0.12+ds1-1/eval/testutils.go
--- 0.11+ds1-5/eval/testutils.go	2018-01-18 02:03:11.000000000 +0000
+++ 0.12+ds1-1/eval/testutils.go	2018-07-11 02:32:25.000000000 +0000
@@ -1,6 +1,17 @@
-// Common testing utilities. This file does not file a _test.go suffix so that
-// it can be used from other packages that also want to test the modules they
-// implement (e.g. edit: and re:).
+// Framework for testing Elvish script. This file does not file a _test.go
+// suffix so that it can be used from other packages that also want to test the
+// modules they implement (e.g. edit: and re:).
+//
+// The entry point for the framework is the Test function, which accepts a
+// *testing.T and a variadic number of test cases. Test cases are constructed
+// using the That function followed by methods that add constraints on the test
+// case. Overall, a test looks like:
+//
+//     Test(t,
+//         That("put x").Puts("x"),
+//         That("echo x").Prints("x\n"))
+//
+// If some setup is needed, use the TestWithSetup function instead.
 
 package eval
 
@@ -8,112 +19,98 @@ import (
 	"bytes"
 	"errors"
 	"fmt"
+	"io/ioutil"
 	"os"
 	"reflect"
 	"testing"
 
-	"github.com/elves/elvish/eval/types"
+	"github.com/elves/elvish/eval/vals"
 	"github.com/elves/elvish/parse"
+	"github.com/elves/elvish/util"
 )
 
-// Test is a test case for TestEval.
-type Test struct {
+// TestCase is a test case for Test.
+type TestCase struct {
 	text string
 	want
 }
 
 type want struct {
-	out      []types.Value
+	out      []interface{}
 	bytesOut []byte
 	err      error
 }
 
 // A special value for want.err to indicate that any error, as long as not nil,
 // is OK
-var errAny = errors.New("")
+var errAny = errors.New("any error")
 
-var (
-	wantNothing = want{}
-	wantTrue    = want{out: bools(true)}
-	wantFalse   = want{out: bools(false)}
-)
-
-// Shorthands for values in want.out
-
-func strs(ss ...string) []types.Value {
-	vs := make([]types.Value, len(ss))
-	for i, s := range ss {
-		vs[i] = types.String(s)
-	}
-	return vs
-}
-
-func bools(bs ...bool) []types.Value {
-	vs := make([]types.Value, len(bs))
-	for i, b := range bs {
-		vs[i] = types.Bool(b)
-	}
-	return vs
-}
-
-// NewTest returns a new Test with the specified source code.
-func NewTest(text string) Test {
-	return Test{text: text}
+// The following functions and methods are used to build Test structs. They are
+// supposed to read like English, so a test that "put x" should put "x" reads:
+//
+// That("put x").Puts("x")
+
+// That returns a new Test with the specified source code.
+func That(text string) TestCase {
+	return TestCase{text: text}
+}
+
+// DoesNothing returns t unchanged. It is used to mark that a piece of code
+// should simply does nothing. In particular, it shouldn't have any output and
+// does not error.
+func (t TestCase) DoesNothing() TestCase {
+	return t
 }
 
-// WantOut returns an altered Test that requires the source code to produce the
+// Puts returns an altered Test that requires the source code to produce the
 // specified values in the value channel when evaluated.
-func (t Test) WantOut(vs ...types.Value) Test {
+func (t TestCase) Puts(vs ...interface{}) TestCase {
 	t.want.out = vs
 	return t
 }
 
-// WantOutStrings returns an altered Test that requires the source code to
-// produce the specified string values in the value channel when evaluated.
-func (t Test) WantOutStrings(ss ...string) Test {
-	return t.WantOut(strs(ss...)...)
-}
-
-// WantOutBools returns an altered Test that requires the source code to produce
-// the specified boolean values in the value channel when evaluated.
-func (t Test) WantOutBools(bs ...bool) Test {
-	return t.WantOut(bools(bs...)...)
+// Puts returns an altered Test that requires the source code to produce the
+// specified strings in the value channel when evaluated.
+func (t TestCase) PutsStrings(ss []string) TestCase {
+	t.want.out = make([]interface{}, len(ss))
+	for i, s := range ss {
+		t.want.out[i] = s
+	}
+	return t
 }
 
-// WantBytesOut returns an altered test that requires the source code to produce
+// Prints returns an altered test that requires the source code to produce
 // the specified output in the byte pipe when evaluated.
-func (t Test) WantBytesOut(b []byte) Test {
-	t.want.bytesOut = b
+func (t TestCase) Prints(s string) TestCase {
+	t.want.bytesOut = []byte(s)
 	return t
 }
 
-// WantBytesOutString is the same as WantBytesOut except that its argument is a
-// string.
-func (t Test) WantBytesOutString(s string) Test {
-	return t.WantBytesOut([]byte(s))
-}
-
-// WantErr returns an altered Test that requires the source code to result in
+// ErrorsWith returns an altered Test that requires the source code to result in
 // the specified error when evaluted.
-func (t Test) WantErr(err error) Test {
+func (t TestCase) ErrorsWith(err error) TestCase {
 	t.want.err = err
 	return t
 }
 
-// WantAnyErr returns an altered Test that requires the source code to result in
-// any error when evaluated.
-func (t Test) WantAnyErr() Test {
-	return t.WantErr(errAny)
-}
-
-// RunTests runs test cases. For each test case, a new Evaler is made by calling
-// makeEvaler.
-func RunTests(t *testing.T, evalTests []Test, makeEvaler func() *Evaler) {
-	for _, tt := range evalTests {
-		// fmt.Printf("eval %q\n", tt.text)
-
-		ev := makeEvaler()
-		defer ev.Close()
+// Errors returns an altered Test that requires the source code to result in any
+// error when evaluated.
+func (t TestCase) Errors() TestCase {
+	return t.ErrorsWith(errAny)
+}
+
+// Test runs test cases. For each test case, a new Evaler is created with
+// NewEvaler.
+func Test(t *testing.T, tests ...TestCase) {
+	TestWithSetup(t, func(*Evaler) {}, tests...)
+}
+
+// Test runs test cases. For each test case, a new Evaler is created with
+// NewEvaler and passed to the setup function.
+func TestWithSetup(t *testing.T, setup func(*Evaler), tests ...TestCase) {
+	for _, tt := range tests {
+		ev := NewEvaler()
+		setup(ev)
 		out, bytesOut, err := evalAndCollect(t, ev, []string{tt.text}, len(tt.want.out))
 
 		first := true
@@ -134,10 +131,12 @@ func RunTests(t *testing.T, evalTests []
 		if !matchErr(tt.want.err, err) {
 			errorf("got err=%v, want %v", err, tt.want.err)
 		}
+
+		ev.Close()
 	}
 }
 
-func evalAndCollect(t *testing.T, ev *Evaler, texts []string, chsize int) ([]types.Value, []byte, error) {
+func evalAndCollect(t *testing.T, ev *Evaler, texts []string, chsize int) ([]interface{}, []byte, error) {
 	// Collect byte output
 	bytesOut := []byte{}
 	pr, pw, _ := os.Pipe()
@@ -155,7 +154,7 @@ func evalAndCollect(t *testing.T, ev *Ev
 	}()
 
 	// Channel output
-	outs := []types.Value{}
+	outs := []interface{}{}
 
 	// Eval error. Only that of the last text is saved.
 	var ex error
@@ -166,7 +165,7 @@ func evalAndCollect(t *testing.T, ev *Ev
 
 		op := mustParseAndCompile(t, ev, src)
 
-		outCh := make(chan types.Value, chsize)
+		outCh := make(chan interface{}, chsize)
 		outDone := make(chan struct{})
 		go func() {
 			for v := range outCh {
@@ -205,11 +204,19 @@ func mustParseAndCompile(t *testing.T, e
 	return op
 }
 
-func matchOut(want, got []types.Value) bool {
+func matchOut(want, got []interface{}) bool {
 	if len(got) == 0 && len(want) == 0 {
 		return true
 	}
-	return reflect.DeepEqual(got, want)
+	if len(got) != len(want) {
+		return false
+	}
+	for i := range got {
+		if !vals.Equal(got[i], want[i]) {
+			return false
+		}
+	}
+	return true
 }
 
 func matchErr(want, got error) bool {
@@ -219,25 +226,41 @@ func matchErr(want, got error) bool {
 	return want == errAny || reflect.DeepEqual(got.(*Exception).Cause, want)
 }
 
-// compareValues compares two slices, using equals for each element.
-func compareSlice(wantValues, gotValues []interface{}) error {
-	if len(wantValues) != len(gotValues) {
-		return fmt.Errorf("want %d values, got %d",
-			len(wantValues), len(gotValues))
+// MustMkdirAll calls os.MkdirAll and panics if an error is returned. It is
+// mainly useful in tests.
+func MustMkdirAll(name string, perm os.FileMode) {
+	err := os.MkdirAll(name, perm)
+	if err != nil {
+		panic(err)
 	}
-	for i, want := range wantValues {
-		if !equals(want, gotValues[i]) {
-			return fmt.Errorf("want [%d] = %s, got %s", i, want, gotValues[i])
-		}
+}
+
+// MustCreateEmpty creates an empty file, and panics if an error occurs. It is
+// mainly useful in tests.
+func MustCreateEmpty(name string) {
+	file, err := os.Create(name)
+	if err != nil {
+		panic(err)
 	}
-	return nil
+	file.Close()
 }
 
-// equals compares two values. It uses Eq if want is a Value instance, or
-// reflect.DeepEqual otherwise.
-func equals(a, b interface{}) bool {
-	if aValue, ok := a.(types.Value); ok {
-		return aValue.Equal(b)
+// MustWriteFile calls ioutil.WriteFile and panics if an error occurs. It is
+// mainly useful in tests.
+func MustWriteFile(filename string, data []byte, perm os.FileMode) {
+	err := ioutil.WriteFile(filename, data, perm)
+	if err != nil {
+		panic(err)
 	}
-	return reflect.DeepEqual(a, b)
+}
+
+// InTempHome is like util.InTempDir, but it also sets HOME to the temporary
+// directory when f is called.
+func InTempHome(f func(string)) {
+	util.InTempDir(func(tmpHome string) {
+		oldHome := os.Getenv("HOME")
+		os.Setenv("HOME", tmpHome)
+		f(tmpHome)
+		os.Setenv("HOME", oldHome)
+	})
 }
diff -pruN 0.11+ds1-5/eval/types/bool.go 0.12+ds1-1/eval/types/bool.go
--- 0.11+ds1-5/eval/types/bool.go	2018-01-18 02:03:11.000000000 +0000
+++ 0.12+ds1-1/eval/types/bool.go	1970-01-01 00:00:00.000000000 +0000
@@ -1,41 +0,0 @@
-package types
-
-// Bool represents truthness.
-type Bool bool
-
-var _ Value = Bool(false)
-
-func (Bool) Kind() string {
-	return "bool"
-}
-
-func (b Bool) Equal(rhs interface{}) bool {
-	return b == rhs
-}
-
-func (b Bool) Hash() uint32 {
-	if b {
-		return 1
-	}
-	return 0
-}
-
-func (b Bool) Repr(int) string {
-	if b {
-		return "$true"
-	}
-	return "$false"
-}
-
-func (b Bool) Bool() bool {
-	return bool(b)
-}
-
-// ToBool converts a Value to bool. When the Value type implements Bool(), it
-// is used. Otherwise it is considered true.
-func ToBool(v Value) bool {
-	if b, ok := v.(Booler); ok {
-		return b.Bool()
-	}
-	return true
-}
diff -pruN 0.11+ds1-5/eval/types/bool_test.go 0.12+ds1-1/eval/types/bool_test.go
--- 0.11+ds1-5/eval/types/bool_test.go	2018-01-18 02:03:11.000000000 +0000
+++ 0.12+ds1-1/eval/types/bool_test.go	1970-01-01 00:00:00.000000000 +0000
@@ -1,19 +0,0 @@
-package types_test
-
-import (
-	"testing"
-
-	"github.com/elves/elvish/eval"
-)
-
-func TestBool(t *testing.T) {
-	eval.RunTests(t, []eval.Test{
-		eval.NewTest("kind-of $true").WantOutStrings("bool"),
-		eval.NewTest("eq $true $true").WantOutBools(true),
-		eval.NewTest("eq $true true").WantOutBools(false),
-		eval.NewTest("repr $true").WantBytesOutString("$true\n"),
-		eval.NewTest("repr $false").WantBytesOutString("$false\n"),
-		eval.NewTest("bool $true").WantOutBools(true),
-		eval.NewTest("bool $false").WantOutBools(false),
-	}, eval.NewEvaler)
-}
diff -pruN 0.11+ds1-5/eval/types/file.go 0.12+ds1-1/eval/types/file.go
--- 0.11+ds1-5/eval/types/file.go	2018-01-18 02:03:11.000000000 +0000
+++ 0.12+ds1-1/eval/types/file.go	1970-01-01 00:00:00.000000000 +0000
@@ -1,37 +0,0 @@
-package types
-
-import (
-	"fmt"
-	"os"
-
-	"github.com/elves/elvish/parse"
-	"github.com/xiaq/persistent/hash"
-)
-
-// File wraps a pointer to os.File.
-type File struct {
-	Inner *os.File
-}
-
-var _ Value = File{}
-
-// NewFile creates a new File value.
-func NewFile(inner *os.File) File {
-	return File{inner}
-}
-
-func (File) Kind() string {
-	return "file"
-}
-
-func (f File) Equal(rhs interface{}) bool {
-	return f == rhs
-}
-
-func (f File) Hash() uint32 {
-	return hash.UIntPtr(f.Inner.Fd())
-}
-
-func (f File) Repr(int) string {
-	return fmt.Sprintf("<file{%s %p}>", parse.Quote(f.Inner.Name()), f.Inner)
-}
diff -pruN 0.11+ds1-5/eval/types/list.go 0.12+ds1-1/eval/types/list.go
--- 0.11+ds1-5/eval/types/list.go	2018-01-18 02:03:11.000000000 +0000
+++ 0.12+ds1-1/eval/types/list.go	1970-01-01 00:00:00.000000000 +0000
@@ -1,168 +0,0 @@
-package types
-
-import (
-	"bytes"
-	"encoding/json"
-	"errors"
-	"strconv"
-	"strings"
-
-	"github.com/xiaq/persistent/vector"
-)
-
-// Error definitions.
-var (
-	ErrBadIndex        = errors.New("bad index")
-	ErrIndexOutOfRange = errors.New("index out of range")
-	ErrAssocWithSlice  = errors.New("assoc with slice not yet supported")
-)
-
-// List is a list of Value's.
-type List struct {
-	inner vector.Vector
-}
-
-// EmptyList is an empty list.
-var EmptyList = List{vector.Empty}
-
-// Make sure that List implements ListLike and Assocer at compile time.
-var (
-	_ ListLike = List{}
-	_ Assocer  = List{}
-)
-
-// MakeList creates a new List from values.
-func MakeList(vs ...Value) List {
-	vec := vector.Empty
-	for _, v := range vs {
-		vec = vec.Cons(v)
-	}
-	return List{vec}
-}
-
-// NewList creates a new List from an existing Vector.
-func NewList(vec vector.Vector) List {
-	return List{vec}
-}
-
-func (List) Kind() string {
-	return "list"
-}
-
-func (l List) Equal(rhs interface{}) bool {
-	return eqListLike(l, rhs)
-}
-
-func (l List) Hash() uint32 {
-	return hashListLike(l)
-}
-
-func (l List) Repr(indent int) string {
-	var b ListReprBuilder
-	b.Indent = indent
-	for it := l.inner.Iterator(); it.HasElem(); it.Next() {
-		v := it.Elem().(Value)
-		b.WriteElem(v.Repr(indent + 1))
-	}
-	return b.String()
-}
-
-func (l List) MarshalJSON() ([]byte, error) {
-	var buf bytes.Buffer
-	encoder := json.NewEncoder(&buf)
-	buf.WriteByte('[')
-	first := true
-	for it := l.inner.Iterator(); it.HasElem(); it.Next() {
-		if first {
-			first = false
-		} else {
-			buf.WriteByte(',')
-		}
-		err := encoder.Encode(it.Elem())
-		if err != nil {
-			return nil, err
-		}
-	}
-	buf.WriteByte(']')
-	return buf.Bytes(), nil
-}
-
-func (l List) Len() int {
-	return l.inner.Len()
-}
-
-func (l List) Iterate(f func(Value) bool) {
-	for it := l.inner.Iterator(); it.HasElem(); it.Next() {
-		v := it.Elem().(Value)
-		if !f(v) {
-			break
-		}
-	}
-}
-
-func (l List) IndexOne(idx Value) Value {
-	slice, i, j := ParseAndFixListIndex(ToString(idx), l.Len())
-	if slice {
-		return List{l.inner.SubVector(i, j)}
-	}
-	return l.inner.Nth(i).(Value)
-}
-
-func (l List) Assoc(idx, v Value) Value {
-	slice, i, _ := ParseAndFixListIndex(ToString(idx), l.Len())
-	if slice {
-		throw(ErrAssocWithSlice)
-	}
-	return List{l.inner.AssocN(i, v)}
-}
-
-// ParseAndFixListIndex parses a list index and returns whether the index is a
-// slice and "real" (-1 becomes n-1) indicies. It throws errors when the index
-// is invalid or out of range.
-func ParseAndFixListIndex(s string, n int) (bool, int, int) {
-	slice, i, j := parseListIndex(s, n)
-	if i < 0 {
-		i += n
-	}
-	if j < 0 {
-		j += n
-	}
-	if i < 0 || i >= n || (slice && (j < 0 || j > n || i > j)) {
-		throw(ErrIndexOutOfRange)
-	}
-	return slice, i, j
-}
-
-// ListIndex = Number |
-//             Number ':' Number
-func parseListIndex(s string, n int) (slice bool, i int, j int) {
-	atoi := func(a string) int {
-		i, err := strconv.Atoi(a)
-		if err != nil {
-			if err.(*strconv.NumError).Err == strconv.ErrRange {
-				throw(ErrIndexOutOfRange)
-			} else {
-				throw(ErrBadIndex)
-			}
-		}
-		return i
-	}
-
-	colon := strings.IndexRune(s, ':')
-	if colon == -1 {
-		// A single number
-		return false, atoi(s), 0
-	}
-	if s[:colon] == "" {
-		i = 0
-	} else {
-		i = atoi(s[:colon])
-	}
-	if s[colon+1:] == "" {
-		j = n
-	} else {
-		j = atoi(s[colon+1:])
-	}
-	// Two numbers
-	return true, i, j
-}
diff -pruN 0.11+ds1-5/eval/types/listlike.go 0.12+ds1-1/eval/types/listlike.go
--- 0.11+ds1-5/eval/types/listlike.go	2018-01-18 02:03:11.000000000 +0000
+++ 0.12+ds1-1/eval/types/listlike.go	1970-01-01 00:00:00.000000000 +0000
@@ -1,67 +0,0 @@
-package types
-
-import (
-	"bytes"
-	"strings"
-
-	"github.com/xiaq/persistent/hash"
-)
-
-type ListLike interface {
-	Lener
-	Iterator
-	IndexOneer
-}
-
-func eqListLike(lhs ListLike, r interface{}) bool {
-	rhs, ok := r.(ListLike)
-	if !ok {
-		return false
-	}
-	if lhs.Len() != rhs.Len() {
-		return false
-	}
-	return true
-}
-
-func hashListLike(l ListLike) uint32 {
-	h := hash.DJBInit
-	l.Iterate(func(v Value) bool {
-		h = hash.DJBCombine(h, v.Hash())
-		return true
-	})
-	return h
-}
-
-// ListReprBuilder helps to build Repr of list-like Values.
-type ListReprBuilder struct {
-	Indent int
-	buf    bytes.Buffer
-}
-
-func (b *ListReprBuilder) WriteElem(v string) {
-	if b.buf.Len() == 0 {
-		b.buf.WriteByte('[')
-	}
-	if b.Indent >= 0 {
-		// Pretty printing.
-		//
-		// Add a newline and indent+1 spaces, so that the
-		// starting & lines up with the first pair.
-		b.buf.WriteString("\n" + strings.Repeat(" ", b.Indent+1))
-	} else if b.buf.Len() > 1 {
-		b.buf.WriteByte(' ')
-	}
-	b.buf.WriteString(v)
-}
-
-func (b *ListReprBuilder) String() string {
-	if b.buf.Len() == 0 {
-		return "[]"
-	}
-	if b.Indent >= 0 {
-		b.buf.WriteString("\n" + strings.Repeat(" ", b.Indent))
-	}
-	b.buf.WriteByte(']')
-	return b.buf.String()
-}
diff -pruN 0.11+ds1-5/eval/types/list_test.go 0.12+ds1-1/eval/types/list_test.go
--- 0.11+ds1-5/eval/types/list_test.go	2018-01-18 02:03:11.000000000 +0000
+++ 0.12+ds1-1/eval/types/list_test.go	1970-01-01 00:00:00.000000000 +0000
@@ -1,140 +0,0 @@
-package types
-
-import (
-	"testing"
-
-	"github.com/elves/elvish/util"
-)
-
-var parseAndFixListIndexTests = []struct {
-	name string
-	// input
-	expr string
-	len  int
-	// output
-	shouldPanic, isSlice bool
-	begin, end           int
-}{
-	{
-		name: "stringIndex",
-		expr: "a", len: 0,
-		shouldPanic: true,
-	},
-	{
-		name: "floatIndex",
-		expr: "1.0", len: 0,
-		shouldPanic: true,
-	},
-	{
-		name: "emptyZeroIndex",
-		expr: "0", len: 0,
-		shouldPanic: true,
-	},
-	{
-		name: "emptyPosIndex",
-		expr: "1", len: 0,
-		shouldPanic: true,
-	},
-	{
-		name: "emptyNegIndex",
-		expr: "-1", len: 0,
-		shouldPanic: true,
-	},
-	{
-		name: "emptySliceAbbrevBoth",
-		expr: ":", len: 0,
-		shouldPanic: true,
-	},
-	{
-		name: "i<-n",
-		expr: "-2", len: 1,
-		shouldPanic: true,
-	},
-	{
-		name: "i=-n",
-		expr: "-1", len: 1,
-		begin: 0, end: 0,
-	},
-	{
-		name: "-n<i<0",
-		expr: "-1", len: 2,
-		begin: 1, end: 0,
-	},
-	{
-		name: "i=0",
-		expr: "0", len: 2,
-		begin: 0, end: 0,
-	},
-	{
-		name: "0<i<n",
-		expr: "1", len: 2,
-		begin: 1, end: 0,
-	},
-	{
-		name: "i=n",
-		expr: "1", len: 1,
-		shouldPanic: true,
-	},
-	{
-		name: "i>n",
-		expr: "2", len: 1,
-		shouldPanic: true,
-	},
-	{
-		name: "sliceAbbrevBoth",
-		expr: ":", len: 1,
-		isSlice: true, begin: 0, end: 1,
-	},
-	{
-		name: "sliceAbbrevBegin",
-		expr: ":1", len: 1,
-		isSlice: true, begin: 0, end: 1,
-	},
-	{
-		name: "sliceAbbrevEnd",
-		expr: "0:", len: 1,
-		isSlice: true, begin: 0, end: 1,
-	},
-	{
-		name: "sliceNegEnd",
-		expr: "0:-1", len: 1,
-		isSlice: true, begin: 0, end: 0,
-	},
-	{
-		name: "sliceBeginEqualEnd",
-		expr: "1:1", len: 2,
-		isSlice: true, begin: 1, end: 1,
-	},
-	{
-		name: "sliceBeginAboveEnd",
-		expr: "1:0", len: 2,
-		shouldPanic: true,
-	},
-}
-
-func TestParseAndFixListIndex(t *testing.T) {
-	checkEqual := func(name, value string, want, got interface{}) {
-		if want != got {
-			t.Errorf("%s value: [%s] want: [%v] got: [%v]",
-				name, value, want, got)
-		}
-	}
-
-	for _, item := range parseAndFixListIndexTests {
-		var (
-			isSlice    bool
-			begin, end int
-		)
-
-		if err := util.PCall(func() {
-			isSlice, begin, end = ParseAndFixListIndex(item.expr, item.len)
-		}); err != nil {
-			checkEqual(item.name, "shouldPanic", item.shouldPanic, err != nil)
-			continue
-		}
-		checkEqual(item.name, "isSlice", item.isSlice, isSlice)
-		checkEqual(item.name, "begin", item.begin, begin)
-		checkEqual(item.name, "end", item.end, end)
-	}
-
-}
diff -pruN 0.11+ds1-5/eval/types/map.go 0.12+ds1-1/eval/types/map.go
--- 0.11+ds1-5/eval/types/map.go	2018-01-18 02:03:11.000000000 +0000
+++ 0.12+ds1-1/eval/types/map.go	1970-01-01 00:00:00.000000000 +0000
@@ -1,130 +0,0 @@
-package types
-
-import (
-	"encoding/json"
-	"errors"
-
-	"github.com/xiaq/persistent/hashmap"
-)
-
-// Map is a map from string to Value.
-type Map struct {
-	inner hashmap.HashMap
-}
-
-var _ MapLike = Map{}
-
-// EmptyMap is an empty Map.
-var EmptyMap = Map{hashmap.Empty}
-
-// NewMap creates a new Map from an inner HashMap.
-func NewMap(inner hashmap.HashMap) Map {
-	return Map{inner}
-}
-
-// MakeMap converts a native Go map to Map.
-func MakeMap(m map[Value]Value) Map {
-	inner := hashmap.Empty
-	for k, v := range m {
-		inner = inner.Assoc(k, v)
-	}
-	return NewMap(inner)
-}
-
-func (Map) Kind() string {
-	return "map"
-}
-
-func (m Map) Equal(a interface{}) bool {
-	return m == a || EqMapLike(m, a)
-}
-
-func (m Map) Hash() uint32 {
-	return HashMapLike(m)
-}
-
-func (m Map) MarshalJSON() ([]byte, error) {
-	// TODO(xiaq): Replace with a more efficient implementation.
-	mm := map[string]Value{}
-	for it := m.inner.Iterator(); it.HasElem(); it.Next() {
-		k, v := it.Elem()
-		mm[ToString(k.(Value))] = v.(Value)
-	}
-	return json.Marshal(mm)
-}
-
-func (m Map) Repr(indent int) string {
-	var builder MapReprBuilder
-	builder.Indent = indent
-	for it := m.inner.Iterator(); it.HasElem(); it.Next() {
-		k, v := it.Elem()
-		builder.WritePair(k.(Value).Repr(indent+1), indent+2, v.(Value).Repr(indent+2))
-	}
-	return builder.String()
-}
-
-func (m Map) Len() int {
-	return m.inner.Len()
-}
-
-func (m Map) IndexOne(idx Value) Value {
-	v, ok := m.inner.Get(idx)
-	if !ok {
-		throw(errors.New("no such key: " + idx.Repr(NoPretty)))
-	}
-	return v.(Value)
-}
-
-func (m Map) Assoc(k, v Value) Value {
-	return Map{m.inner.Assoc(k, v)}
-}
-
-func (m Map) Dissoc(k Value) Value {
-	return Map{m.inner.Without(k)}
-}
-
-func (m Map) IterateKey(f func(Value) bool) {
-	for it := m.inner.Iterator(); it.HasElem(); it.Next() {
-		k, _ := it.Elem()
-		if !f(k.(Value)) {
-			break
-		}
-	}
-}
-
-func (m Map) IteratePair(f func(Value, Value) bool) {
-	for it := m.inner.Iterator(); it.HasElem(); it.Next() {
-		k, v := it.Elem()
-		if !f(k.(Value), v.(Value)) {
-			break
-		}
-	}
-}
-
-func (m Map) HasKey(k Value) bool {
-	_, ok := m.inner.Get(k)
-	return ok
-}
-
-// MapReprBuilder helps building the Repr of a Map. It is also useful for
-// implementing other Map-like values. The zero value of a MapReprBuilder is
-// ready to use.
-type MapReprBuilder struct {
-	ListReprBuilder
-}
-
-func (b *MapReprBuilder) WritePair(k string, indent int, v string) {
-	if indent > 0 {
-		b.WriteElem("&" + k + "=\t" + v)
-	} else {
-		b.WriteElem("&" + k + "=" + v)
-	}
-}
-
-func (b *MapReprBuilder) String() string {
-	s := b.ListReprBuilder.String()
-	if s == "[]" {
-		s = "[&]"
-	}
-	return s
-}
diff -pruN 0.11+ds1-5/eval/types/maplike.go 0.12+ds1-1/eval/types/maplike.go
--- 0.11+ds1-5/eval/types/maplike.go	2018-01-18 02:03:11.000000000 +0000
+++ 0.12+ds1-1/eval/types/maplike.go	1970-01-01 00:00:00.000000000 +0000
@@ -1,47 +0,0 @@
-package types
-
-import (
-	"github.com/xiaq/persistent/hash"
-)
-
-type MapLike interface {
-	Lener
-	IndexOneer
-	Assocer
-	HasKeyer
-	IterateKeyer
-	IteratePairer
-}
-
-type HasKeyer interface {
-	HasKey(k Value) bool
-}
-
-func EqMapLike(lhs MapLike, a interface{}) bool {
-	rhs, ok := a.(MapLike)
-	if !ok {
-		return false
-	}
-	if lhs.Len() != rhs.Len() {
-		return false
-	}
-	eq := true
-	lhs.IteratePair(func(k, v Value) bool {
-		if !v.Equal(rhs.IndexOne(k)) {
-			eq = false
-			return false
-		}
-		return true
-	})
-	return eq
-}
-
-func HashMapLike(m MapLike) uint32 {
-	h := hash.DJBInit
-	m.IteratePair(func(k, v Value) bool {
-		h = hash.DJBCombine(h, k.Hash())
-		h = hash.DJBCombine(h, v.Hash())
-		return true
-	})
-	return h
-}
diff -pruN 0.11+ds1-5/eval/types/pipe.go 0.12+ds1-1/eval/types/pipe.go
--- 0.11+ds1-5/eval/types/pipe.go	2018-01-18 02:03:11.000000000 +0000
+++ 0.12+ds1-1/eval/types/pipe.go	1970-01-01 00:00:00.000000000 +0000
@@ -1,40 +0,0 @@
-package types
-
-import (
-	"fmt"
-	"os"
-
-	"github.com/xiaq/persistent/hash"
-)
-
-// Pipe wraps a pair of pointers to os.File that are the two ends of the same
-// pipe.
-type Pipe struct {
-	ReadEnd, WriteEnd *os.File
-}
-
-var _ Value = Pipe{}
-
-// NewPipe creates a new Pipe value.
-func NewPipe(r, w *os.File) Pipe {
-	return Pipe{r, w}
-}
-
-func (Pipe) Kind() string {
-	return "pipe"
-}
-
-func (p Pipe) Equal(rhs interface{}) bool {
-	return p == rhs
-}
-
-func (p Pipe) Hash() uint32 {
-	h := hash.DJBInit
-	h = hash.DJBCombine(h, hash.UIntPtr(p.ReadEnd.Fd()))
-	h = hash.DJBCombine(h, hash.UIntPtr(p.WriteEnd.Fd()))
-	return h
-}
-
-func (p Pipe) Repr(int) string {
-	return fmt.Sprintf("<pipe{%v %v}>", p.ReadEnd.Fd(), p.WriteEnd.Fd())
-}
diff -pruN 0.11+ds1-5/eval/types/rat.go 0.12+ds1-1/eval/types/rat.go
--- 0.11+ds1-5/eval/types/rat.go	2018-01-18 02:03:11.000000000 +0000
+++ 0.12+ds1-1/eval/types/rat.go	1970-01-01 00:00:00.000000000 +0000
@@ -1,67 +0,0 @@
-package types
-
-import (
-	"errors"
-	"fmt"
-	"math/big"
-
-	"github.com/xiaq/persistent/hash"
-)
-
-var ErrOnlyStrOrRat = errors.New("only str or rat may be converted to rat")
-
-// Rat is a rational number.
-type Rat struct {
-	b *big.Rat
-}
-
-var _ Value = Rat{}
-
-func (Rat) Kind() string {
-	return "string"
-}
-
-func (r Rat) Equal(a interface{}) bool {
-	if r == a {
-		return true
-	}
-	r2, ok := a.(Rat)
-	if !ok {
-		return false
-	}
-	return r.b.Cmp(r2.b) == 0
-}
-
-func (r Rat) Hash() uint32 {
-	// TODO(xiaq): Use a more efficient implementation.
-	return hash.String(r.String())
-}
-
-func (r Rat) Repr(int) string {
-	return "(rat " + r.String() + ")"
-}
-
-func (r Rat) String() string {
-	if r.b.IsInt() {
-		return r.b.Num().String()
-	}
-	return r.b.String()
-}
-
-// ToRat converts a Value to rat. A str can be converted to a rat if it can be
-// parsed. A rat is returned as-is. Other types of values cannot be converted.
-func ToRat(v Value) (Rat, error) {
-	switch v := v.(type) {
-	case Rat:
-		return v, nil
-	case String:
-		r := big.Rat{}
-		_, err := fmt.Sscanln(string(v), &r)
-		if err != nil {
-			return Rat{}, fmt.Errorf("%s cannot be parsed as rat", v.Repr(NoPretty))
-		}
-		return Rat{&r}, nil
-	default:
-		return Rat{}, ErrOnlyStrOrRat
-	}
-}
diff -pruN 0.11+ds1-5/eval/types/string.go 0.12+ds1-1/eval/types/string.go
--- 0.11+ds1-5/eval/types/string.go	2018-01-18 02:03:11.000000000 +0000
+++ 0.12+ds1-1/eval/types/string.go	1970-01-01 00:00:00.000000000 +0000
@@ -1,81 +0,0 @@
-package types
-
-import (
-	"errors"
-	"unicode/utf8"
-
-	"github.com/elves/elvish/parse"
-	"github.com/xiaq/persistent/hash"
-)
-
-// String is just a string.
-type String string
-
-var (
-	_ Value    = String("")
-	_ ListLike = String("")
-)
-
-var ErrReplacementMustBeString = errors.New("replacement must be string")
-
-func (String) Kind() string {
-	return "string"
-}
-
-func (s String) Repr(int) string {
-	return parse.Quote(string(s))
-}
-
-func (s String) Equal(rhs interface{}) bool {
-	return s == rhs
-}
-
-func (s String) Hash() uint32 {
-	return hash.String(string(s))
-}
-
-func (s String) String() string {
-	return string(s)
-}
-
-func (s String) Len() int {
-	return len(string(s))
-}
-
-func (s String) IndexOne(idx Value) Value {
-	i, j := s.index(idx)
-	return s[i:j]
-}
-
-func (s String) Assoc(idx, v Value) Value {
-	i, j := s.index(idx)
-	repl, ok := v.(String)
-	if !ok {
-		throw(ErrReplacementMustBeString)
-	}
-	return s[:i] + repl + s[j:]
-}
-
-func (s String) index(idx Value) (int, int) {
-	slice, i, j := ParseAndFixListIndex(ToString(idx), len(s))
-	r, size := utf8.DecodeRuneInString(string(s[i:]))
-	if r == utf8.RuneError {
-		throw(ErrBadIndex)
-	}
-	if slice {
-		if r, _ := utf8.DecodeLastRuneInString(string(s[:j])); r == utf8.RuneError {
-			throw(ErrBadIndex)
-		}
-		return i, j
-	}
-	return i, i + size
-}
-
-func (s String) Iterate(f func(v Value) bool) {
-	for _, r := range s {
-		b := f(String(string(r)))
-		if !b {
-			break
-		}
-	}
-}
diff -pruN 0.11+ds1-5/eval/types/struct.go 0.12+ds1-1/eval/types/struct.go
--- 0.11+ds1-5/eval/types/struct.go	2018-01-18 02:03:11.000000000 +0000
+++ 0.12+ds1-1/eval/types/struct.go	1970-01-01 00:00:00.000000000 +0000
@@ -1,150 +0,0 @@
-package types
-
-import (
-	"bytes"
-	"encoding/json"
-	"errors"
-	"fmt"
-
-	"github.com/elves/elvish/parse"
-)
-
-var (
-	ErrIndexMustBeString = errors.New("index must be string")
-)
-
-// Struct is like a Map with fixed keys.
-type Struct struct {
-	descriptor *StructDescriptor
-	fields     []Value
-}
-
-var (
-	_ Value   = (*Struct)(nil)
-	_ MapLike = (*Struct)(nil)
-)
-
-// NewStruct creates a new *Struct value.
-func NewStruct(descriptor *StructDescriptor, fields []Value) *Struct {
-	return &Struct{descriptor, fields}
-}
-
-func (*Struct) Kind() string {
-	return "map"
-}
-
-// Equal returns true if the rhs is MapLike and all pairs are equal.
-func (s *Struct) Equal(rhs interface{}) bool {
-	return s == rhs || EqMapLike(s, rhs)
-}
-
-func (s *Struct) Hash() uint32 {
-	return HashMapLike(s)
-}
-
-func (s *Struct) Repr(indent int) string {
-	var builder MapReprBuilder
-	builder.Indent = indent
-	for i, name := range s.descriptor.fieldNames {
-		builder.WritePair(parse.Quote(name), indent+2, s.fields[i].Repr(indent+2))
-	}
-	return builder.String()
-}
-
-func (s *Struct) Len() int {
-	return len(s.descriptor.fieldNames)
-}
-
-func (s *Struct) IndexOne(idx Value) Value {
-	return s.fields[s.index(idx)]
-}
-
-func (s *Struct) Assoc(k, v Value) Value {
-	i := s.index(k)
-	fields := make([]Value, len(s.fields))
-	copy(fields, s.fields)
-	fields[i] = v
-	return &Struct{s.descriptor, fields}
-}
-
-func (s *Struct) IterateKey(f func(Value) bool) {
-	for _, field := range s.descriptor.fieldNames {
-		if !f(String(field)) {
-			break
-		}
-	}
-}
-
-func (s *Struct) IteratePair(f func(Value, Value) bool) {
-	for i, field := range s.descriptor.fieldNames {
-		if !f(String(field), s.fields[i]) {
-			break
-		}
-	}
-}
-
-func (s *Struct) HasKey(k Value) bool {
-	index, ok := k.(String)
-	if !ok {
-		return false
-	}
-	_, ok = s.descriptor.fieldIndex[string(index)]
-	return ok
-}
-
-func (s *Struct) index(idx Value) int {
-	index, ok := idx.(String)
-	if !ok {
-		throw(ErrIndexMustBeString)
-	}
-	i, ok := s.descriptor.fieldIndex[string(index)]
-	if !ok {
-		throw(fmt.Errorf("no such field: %s", index.Repr(NoPretty)))
-	}
-	return i
-}
-
-// MarshalJSON encodes the Struct to a JSON Object.
-func (s *Struct) MarshalJSON() ([]byte, error) {
-	var buf bytes.Buffer
-	buf.WriteByte('{')
-	for i, fieldName := range s.descriptor.fieldNames {
-		if i > 0 {
-			buf.WriteByte(',')
-		}
-		buf.Write(s.descriptor.jsonFieldNames[i])
-		buf.WriteByte(':')
-		fieldJSON, err := json.Marshal(s.fields[i])
-		if err != nil {
-			return nil, fmt.Errorf("cannot encode field %q: %v", fieldName, err)
-		}
-		buf.Write(fieldJSON)
-	}
-	buf.WriteByte('}')
-	return buf.Bytes(), nil
-}
-
-// StructDescriptor contains information about the fields in a Struct.
-type StructDescriptor struct {
-	fieldNames     []string
-	jsonFieldNames [][]byte
-	fieldIndex     map[string]int
-}
-
-// NewStructDescriptor creates a new struct descriptor from a list of field
-// names.
-func NewStructDescriptor(fields ...string) *StructDescriptor {
-	fieldNames := append([]string(nil), fields...)
-	jsonFieldNames := make([][]byte, len(fields))
-	fieldIndex := make(map[string]int)
-	for i, name := range fieldNames {
-		fieldIndex[name] = i
-		jsonFieldName, err := json.Marshal(name)
-		// json.Marshal should never fail on string.
-		if err != nil {
-			panic(err)
-		}
-		jsonFieldNames[i] = jsonFieldName
-	}
-	return &StructDescriptor{fieldNames, jsonFieldNames, fieldIndex}
-}
diff -pruN 0.11+ds1-5/eval/types/struct_test.go 0.12+ds1-1/eval/types/struct_test.go
--- 0.11+ds1-5/eval/types/struct_test.go	2018-01-18 02:03:11.000000000 +0000
+++ 0.12+ds1-1/eval/types/struct_test.go	1970-01-01 00:00:00.000000000 +0000
@@ -1,39 +0,0 @@
-package types
-
-import (
-	"testing"
-)
-
-var (
-	testStructDescriptor = NewStructDescriptor("foo", "bar")
-	testStruct           = NewStruct(testStructDescriptor, []Value{String("lorem"), String("ipsum")})
-	testStruct2          = NewStruct(testStructDescriptor, []Value{String("lorem"), String("dolor")})
-)
-
-func TestStructMethods(t *testing.T) {
-	if l := testStruct.Len(); l != 2 {
-		t.Errorf("testStruct.Len() = %d, want 2", l)
-	}
-	if foo := testStruct.IndexOne(String("foo")); foo != String("lorem") {
-		t.Errorf(`testStruct.IndexOne("foo") = %q, want "lorem"`, foo)
-	}
-	if testStruct.Equal(testStruct2) {
-		t.Errorf(`testStruct.Equal(testStruct2) => true, want false`)
-	}
-	if s2 := testStruct.Assoc(String("bar"), String("dolor")); !s2.Equal(testStruct2) {
-		t.Errorf(`testStruct.Assoc(...) => %v, want %v`, s2, testStruct2)
-	}
-	wantRepr := "[&foo=lorem &bar=ipsum]"
-	if gotRepr := testStruct.Repr(NoPretty); gotRepr != wantRepr {
-		t.Errorf(`testStruct.Repr() => %q, want %q`, gotRepr, wantRepr)
-	}
-	wantJSON := `{"foo":"lorem","bar":"ipsum"}`
-	gotJSONBytes, err := testStruct.MarshalJSON()
-	gotJSON := string(gotJSONBytes)
-	if err != nil {
-		t.Errorf(`testStruct.MarshalJSON() => error %v`, err)
-	}
-	if wantJSON != gotJSON {
-		t.Errorf(`testStruct.MarshalJSON() => %q, want %q`, gotJSON, wantJSON)
-	}
-}
diff -pruN 0.11+ds1-5/eval/types/throw.go 0.12+ds1-1/eval/types/throw.go
--- 0.11+ds1-5/eval/types/throw.go	2018-01-18 02:03:11.000000000 +0000
+++ 0.12+ds1-1/eval/types/throw.go	1970-01-01 00:00:00.000000000 +0000
@@ -1,21 +0,0 @@
-package types
-
-import (
-	"fmt"
-
-	"github.com/elves/elvish/util"
-)
-
-func throw(e error) {
-	util.Throw(e)
-}
-
-func throwf(format string, args ...interface{}) {
-	util.Throw(fmt.Errorf(format, args...))
-}
-
-func maybeThrow(err error) {
-	if err != nil {
-		util.Throw(err)
-	}
-}
diff -pruN 0.11+ds1-5/eval/types/types.go 0.12+ds1-1/eval/types/types.go
--- 0.11+ds1-5/eval/types/types.go	2018-01-18 02:03:11.000000000 +0000
+++ 0.12+ds1-1/eval/types/types.go	1970-01-01 00:00:00.000000000 +0000
@@ -1,170 +0,0 @@
-// Package types contains basic types for the Elvish runtime.
-package types
-
-import (
-	"github.com/elves/elvish/util"
-)
-
-// Definitions for Value interfaces, some simple Value types and some common
-// Value helpers.
-
-// Value is an Elvish value.
-type Value interface {
-	Kinder
-	Equaler
-	Hasher
-	Reprer
-}
-
-// Kinder wraps the Kind method.
-type Kinder interface {
-	Kind() string
-}
-
-// Reprer wraps the Repr method.
-type Reprer interface {
-	// Repr returns a string that represents a Value. The string either be a
-	// literal of that Value that is preferably deep-equal to it (like `[a b c]`
-	// for a list), or a string enclosed in "<>" containing the kind and
-	// identity of the Value(like `<fn 0xdeadcafe>`).
-	//
-	// If indent is at least 0, it should be pretty-printed with the current
-	// indentation level of indent; the indent of the first line has already
-	// been written and shall not be written in Repr. The returned string
-	// should never contain a trailing newline.
-	Repr(indent int) string
-}
-
-// NoPretty can be passed to Repr to suppress pretty-printing.
-const NoPretty = util.MinInt
-
-// Equaler wraps the Equal method.
-type Equaler interface {
-	// Equal compares the receiver to another value. Two equal values must have
-	// the same hash code.
-	Equal(other interface{}) bool
-}
-
-// Hasher wraps the Hash method.
-type Hasher interface {
-	// Hash computes the hash code of the receiver.
-	Hash() uint32
-}
-
-// Booler wraps the Bool method.
-type Booler interface {
-	// Bool computes the truth value of the receiver.
-	Bool() bool
-}
-
-// Stringer wraps the String method.
-type Stringer interface {
-	// Stringer converts the receiver to a string.
-	String() string
-}
-
-// ToString converts a Value to string. When the Value type implements
-// String(), it is used. Otherwise Repr(NoPretty) is used.
-func ToString(v Value) string {
-	if s, ok := v.(Stringer); ok {
-		return s.String()
-	}
-	return v.Repr(NoPretty)
-}
-
-// Lener wraps the Len method.
-type Lener interface {
-	// Len computes the length of the receiver.
-	Len() int
-}
-
-// Iterator wraps the Iterate method.
-type Iterator interface {
-	// Iterate calls the passed function with each value within the receiver.
-	// The iteration is aborted if the function returns false.
-	Iterate(func(v Value) bool)
-}
-
-// IteratorValue is an iterable Value.
-type IteratorValue interface {
-	Iterator
-	Value
-}
-
-func CollectFromIterator(it Iterator) []Value {
-	var vs []Value
-	if lener, ok := it.(Lener); ok {
-		vs = make([]Value, 0, lener.Len())
-	}
-	it.Iterate(func(v Value) bool {
-		vs = append(vs, v)
-		return true
-	})
-	return vs
-}
-
-// IterateKeyer wraps the IterateKey method.
-type IterateKeyer interface {
-	// IterateKey calls the passed function with each value within the receiver.
-	// The iteration is aborted if the function returns false.
-	IterateKey(func(k Value) bool)
-}
-
-// IteratePairer wraps the IteratePair method.
-type IteratePairer interface {
-	// IteratePair calls the passed function with each key and value within the
-	// receiver. The iteration is aborted if the function returns false.
-	IteratePair(func(k, v Value) bool)
-}
-
-// Indexer wraps the Index method.
-type Indexer interface {
-	// Index retrieves the values within the receiver at the specified indicies.
-	Index(idx []Value) []Value
-}
-
-// IndexOneer wraps the IndexOne method.
-type IndexOneer interface {
-	// Index retrieves one value from the receiver at the specified index.
-	IndexOne(idx Value) Value
-}
-
-// GetIndexer adapts a Value to an Indexer if there is an adapter.
-func GetIndexer(v Value) (Indexer, bool) {
-	if indexer, ok := v.(Indexer); ok {
-		return indexer, true
-	}
-	if indexOneer, ok := v.(IndexOneer); ok {
-		return IndexOneerIndexer{indexOneer}, true
-	}
-	return nil, false
-}
-
-// IndexOneerIndexer adapts an IndexOneer to an Indexer by calling all the
-// indicies on the IndexOner and collect the results.
-type IndexOneerIndexer struct {
-	IndexOneer
-}
-
-func (ioi IndexOneerIndexer) Index(vs []Value) []Value {
-	results := make([]Value, len(vs))
-	for i, v := range vs {
-		results[i] = ioi.IndexOneer.IndexOne(v)
-	}
-	return results
-}
-
-// Assocer wraps the Assoc method.
-type Assocer interface {
-	// Assoc returns a slightly modified version of the receiver with key k
-	// associated with value v.
-	Assoc(k, v Value) Value
-}
-
-// Dissocer is anything tha can return a slightly modified version of itself with
-// the specified key removed, as a new value.
-type Dissocer interface {
-	// Dissoc returns a slightly modified version of the receiver with key k
-	// dissociated with any value.
-	Dissoc(k Value) Value
-}
diff -pruN 0.11+ds1-5/eval/types/types_test.go 0.12+ds1-1/eval/types/types_test.go
--- 0.11+ds1-5/eval/types/types_test.go	2018-01-18 02:03:11.000000000 +0000
+++ 0.12+ds1-1/eval/types/types_test.go	1970-01-01 00:00:00.000000000 +0000
@@ -1,28 +0,0 @@
-package types
-
-import (
-	"os"
-	"testing"
-
-	"github.com/elves/elvish/tt"
-	"github.com/xiaq/persistent/hashmap"
-	"github.com/xiaq/persistent/vector"
-)
-
-var Args = tt.Args
-
-func kind(k Kinder) string {
-	return k.Kind()
-}
-
-func TestKind(t *testing.T) {
-	tt.Test(t, tt.Fn("kind", kind), tt.Table{
-		Args(Bool(true)).Rets("bool"),
-		Args(String("")).Rets("string"),
-		Args(NewList(vector.Empty)).Rets("list"),
-		Args(NewMap(hashmap.Empty)).Rets("map"),
-		Args(NewStruct(NewStructDescriptor(), nil)).Rets("map"),
-		Args(NewFile(os.Stdin)).Rets("file"),
-		Args(NewPipe(os.Stdin, os.Stdout)).Rets("pipe"),
-	})
-}
diff -pruN 0.11+ds1-5/eval/unwrap.go 0.12+ds1-1/eval/unwrap.go
--- 0.11+ds1-5/eval/unwrap.go	2018-01-18 02:03:11.000000000 +0000
+++ 0.12+ds1-1/eval/unwrap.go	2018-07-11 02:32:25.000000000 +0000
@@ -4,7 +4,7 @@ import (
 	"fmt"
 	"strconv"
 
-	"github.com/elves/elvish/eval/types"
+	"github.com/elves/elvish/eval/vals"
 )
 
 // Unwrappers are helper types for "unwrapping" values, the process for
@@ -21,7 +21,7 @@ type unwrapperInner struct {
 	// error occurs.
 	begin, end int
 	// values contain the Value's to unwrap.
-	values []types.Value
+	values []interface{}
 }
 
 func (u *unwrapperInner) error(want, gotfmt string, gotargs ...interface{}) {
@@ -34,14 +34,16 @@ func (u *unwrapperInner) error(want, got
 type ValuesUnwrapper struct{ *unwrapperInner }
 
 // Unwrap creates an Unwrapper.
-func (ctx *Frame) Unwrap(desc string, begin, end int, vs []types.Value) ValuesUnwrapper {
+func (ctx *Frame) Unwrap(desc string, begin, end int, vs []interface{}) ValuesUnwrapper {
 	return ValuesUnwrapper{&unwrapperInner{ctx, desc, begin, end, vs}}
 }
 
 // ExecAndUnwrap executes a ValuesOp and creates an Unwrapper for the obtained
 // values.
 func (ctx *Frame) ExecAndUnwrap(desc string, op ValuesOp) ValuesUnwrapper {
-	return ctx.Unwrap(desc, op.Begin, op.End, op.Exec(ctx))
+	values, err := op.Exec(ctx)
+	maybeThrow(err)
+	return ctx.Unwrap(desc, op.Begin, op.End, values)
 }
 
 // One unwraps the value to be exactly one value.
@@ -55,21 +57,21 @@ func (u ValuesUnwrapper) One() ValueUnwr
 // ValueUnwrapper unwraps one Value.
 type ValueUnwrapper struct{ *unwrapperInner }
 
-func (u ValueUnwrapper) Any() types.Value {
+func (u ValueUnwrapper) Any() interface{} {
 	return u.values[0]
 }
 
-func (u ValueUnwrapper) String() types.String {
-	s, ok := u.values[0].(types.String)
+func (u ValueUnwrapper) String() string {
+	s, ok := u.values[0].(string)
 	if !ok {
-		u.error("string", "%s", u.values[0].Kind())
+		u.error("string", "%s", vals.Kind(u.values[0]))
 	}
 	return s
 }
 
 func (u ValueUnwrapper) Int() int {
 	s := u.String()
-	i, err := strconv.Atoi(string(s))
+	i, err := strconv.Atoi(s)
 	if err != nil {
 		u.error("integer", "%s", s)
 	}
@@ -85,7 +87,7 @@ func (u ValueUnwrapper) NonNegativeInt()
 }
 
 func (u ValueUnwrapper) FdOrClose() int {
-	s := string(u.String())
+	s := u.String()
 	if s == "-" {
 		return -1
 	}
@@ -95,15 +97,7 @@ func (u ValueUnwrapper) FdOrClose() int
 func (u ValueUnwrapper) Callable() Callable {
 	c, ok := u.values[0].(Callable)
 	if !ok {
-		u.error("callable", "%s", u.values[0].Kind())
+		u.error("callable", "%s", vals.Kind(u.values[0]))
 	}
 	return c
 }
-
-func (u ValueUnwrapper) Iterable() types.Iterator {
-	it, ok := u.values[0].(types.Iterator)
-	if !ok {
-		u.error("iterable", "%s", u.values[0].Kind())
-	}
-	return it
-}
diff -pruN 0.11+ds1-5/eval/util.go 0.12+ds1-1/eval/util.go
--- 0.11+ds1-5/eval/util.go	2018-01-18 02:03:11.000000000 +0000
+++ 0.12+ds1-1/eval/util.go	2018-07-11 02:32:25.000000000 +0000
@@ -1,12 +1,10 @@
 package eval
 
 import (
-	"errors"
 	"fmt"
 	"os"
-	"strings"
 
-	"github.com/elves/elvish/eval/types"
+	"github.com/elves/elvish/eval/vals"
 	"github.com/elves/elvish/parse"
 	"github.com/elves/elvish/util"
 )
@@ -25,6 +23,12 @@ func maybeThrow(err error) {
 	}
 }
 
+// mustScanToGo is like vals.ScanToGo, except that errors are turned into
+// exceptions.
+func mustScanToGo(src interface{}, ptr interface{}) {
+	maybeThrow(vals.ScanToGo(src, ptr))
+}
+
 func mustGetHome(uname string) string {
 	dir, err := util.GetHome(uname)
 	if err != nil {
@@ -33,42 +37,6 @@ func mustGetHome(uname string) string {
 	return dir
 }
 
-func ParseVariable(text string) (explode bool, ns string, name string) {
-	explodePart, qname := ParseVariableSplice(text)
-	nsPart, name := ParseVariableQName(qname)
-	ns = nsPart
-	if len(ns) > 0 {
-		ns = ns[:len(ns)-1]
-	}
-	return explodePart != "", ns, name
-}
-
-func ParseVariableSplice(text string) (explode, qname string) {
-	if strings.HasPrefix(text, "@") {
-		return "@", text[1:]
-	}
-	return "", text
-}
-
-func ParseVariableQName(qname string) (ns, name string) {
-	i := strings.LastIndexByte(qname, ':')
-	if i == -1 {
-		return "", qname
-	}
-	return qname[:i+1], qname[i+1:]
-}
-
-func MakeVariableName(explode bool, ns string, name string) string {
-	prefix := ""
-	if explode {
-		prefix = "@"
-	}
-	if ns != "" {
-		prefix += ns + ":"
-	}
-	return prefix + name
-}
-
 func makeFlag(m parse.RedirMode) int {
 	switch m {
 	case parse.Read:
@@ -83,20 +51,3 @@ func makeFlag(m parse.RedirMode) int {
 		return -1
 	}
 }
-
-var (
-	ErrNoArgAccepted = errors.New("no argument accepted")
-	ErrNoOptAccepted = errors.New("no option accepted")
-)
-
-func TakeNoArg(args []types.Value) {
-	if len(args) > 0 {
-		throw(ErrNoArgAccepted)
-	}
-}
-
-func TakeNoOpt(opts map[string]types.Value) {
-	if len(opts) > 0 {
-		throw(ErrNoOptAccepted)
-	}
-}
diff -pruN 0.11+ds1-5/eval/validator.go 0.12+ds1-1/eval/validator.go
--- 0.11+ds1-5/eval/validator.go	2018-01-18 02:03:11.000000000 +0000
+++ 0.12+ds1-1/eval/validator.go	1970-01-01 00:00:00.000000000 +0000
@@ -1,26 +0,0 @@
-package eval
-
-import (
-	"errors"
-
-	"github.com/elves/elvish/eval/types"
-)
-
-var (
-	errShouldBeFn = errors.New("should be function")
-	errShouldBeNs = errors.New("should be ns")
-)
-
-func ShouldBeFn(v types.Value) error {
-	if _, ok := v.(Callable); !ok {
-		return errShouldBeFn
-	}
-	return nil
-}
-
-func ShouldBeNs(v types.Value) error {
-	if _, ok := v.(Ns); !ok {
-		return errShouldBeNs
-	}
-	return nil
-}
diff -pruN 0.11+ds1-5/eval/vals/assoc.go 0.12+ds1-1/eval/vals/assoc.go
--- 0.11+ds1-5/eval/vals/assoc.go	1970-01-01 00:00:00.000000000 +0000
+++ 0.12+ds1-1/eval/vals/assoc.go	2018-07-11 02:32:25.000000000 +0000
@@ -0,0 +1,72 @@
+package vals
+
+import (
+	"errors"
+
+	"github.com/xiaq/persistent/vector"
+)
+
+// Assocer wraps the Assoc method.
+type Assocer interface {
+	// Assoc returns a slightly modified version of the receiver with key k
+	// associated with value v.
+	Assoc(k, v interface{}) (interface{}, error)
+}
+
+var (
+	errAssocUnsupported        = errors.New("assoc is not supported")
+	errReplacementMustBeString = errors.New("replacement must be string")
+	errAssocWithSlice          = errors.New("assoc with slice not yet supported")
+)
+
+// Assoc takes a container, a key and value, and returns a modified version of
+// the container, in which the key associated with the value. It is implemented
+// for the builtin type string, and types satisfying the listAssocable,
+// mapAssocable or Assocer interface. For other types, it returns an error.
+func Assoc(a, k, v interface{}) (interface{}, error) {
+	switch a := a.(type) {
+	case Assocer:
+		return a.Assoc(k, v)
+	case string:
+		return assocString(a, k, v)
+	case listAssocable:
+		return assocList(a, k, v)
+	case mapAssocable:
+		return a.Assoc(k, v), nil
+	}
+	return nil, errAssocUnsupported
+}
+
+func assocString(s string, k, v interface{}) (interface{}, error) {
+	i, j, err := convertStringIndex(k, s)
+	if err != nil {
+		return nil, err
+	}
+	repl, ok := v.(string)
+	if !ok {
+		return nil, errReplacementMustBeString
+	}
+	return s[:i] + repl + s[j:], nil
+}
+
+type listAssocable interface {
+	Lener
+	Assoc(int, interface{}) vector.Vector
+}
+
+var _ listAssocable = vector.Vector(nil)
+
+func assocList(l listAssocable, k, v interface{}) (interface{}, error) {
+	kstring, ok := k.(string)
+	if !ok {
+		return nil, errIndexMustBeString
+	}
+	index, err := ConvertListIndex(kstring, l.Len())
+	if err != nil {
+		return nil, err
+	}
+	if index.Slice {
+		return nil, errAssocWithSlice
+	}
+	return l.Assoc(index.Lower, v), nil
+}
diff -pruN 0.11+ds1-5/eval/vals/assoc_test.go 0.12+ds1-1/eval/vals/assoc_test.go
--- 0.11+ds1-5/eval/vals/assoc_test.go	1970-01-01 00:00:00.000000000 +0000
+++ 0.12+ds1-1/eval/vals/assoc_test.go	2018-07-11 02:32:25.000000000 +0000
@@ -0,0 +1,44 @@
+package vals
+
+import (
+	"errors"
+	"testing"
+
+	"github.com/elves/elvish/tt"
+)
+
+type customAssocer struct{}
+
+var customAssocError = errors.New("custom assoc error")
+
+func (a customAssocer) Assoc(k, v interface{}) (interface{}, error) {
+	return "custom result", customAssocError
+}
+
+var assocTests = tt.Table{
+	Args("0123", "0", "foo").Rets("foo123", nil),
+	Args("0123", "1:3", "bar").Rets("0bar3", nil),
+	Args("0123", "1:3", 12).Rets(nil, errReplacementMustBeString),
+	Args("0123", "x", "y").Rets(nil, errIndexMustBeNumber),
+
+	Args(MakeList("0", "1", "2", "3"), "0", "foo").Rets(
+		eq(MakeList("foo", "1", "2", "3")), nil),
+	Args(MakeList("0"), MakeList("0"), "1").Rets(nil, errIndexMustBeString),
+	Args(MakeList("0"), "1", "x").Rets(nil, errIndexOutOfRange),
+	// TODO: Support list assoc with slice
+	Args(MakeList("0", "1", "2", "3"), "1:3", MakeList("foo")).Rets(
+		nil, errAssocWithSlice),
+
+	Args(MakeMapFromKV("k", "v", "k2", "v2"), "k", "newv").Rets(
+		eq(MakeMapFromKV("k", "newv", "k2", "v2")), nil),
+	Args(MakeMapFromKV("k", "v"), "k2", "v2").Rets(
+		eq(MakeMapFromKV("k", "v", "k2", "v2")), nil),
+
+	Args(customAssocer{}, "x", "y").Rets("custom result", customAssocError),
+
+	Args(struct{}{}, "x", "y").Rets(nil, errAssocUnsupported),
+}
+
+func TestAssoc(t *testing.T) {
+	tt.Test(t, tt.Fn("Assoc", Assoc), assocTests)
+}
diff -pruN 0.11+ds1-5/eval/vals/bool.go 0.12+ds1-1/eval/vals/bool.go
--- 0.11+ds1-5/eval/vals/bool.go	1970-01-01 00:00:00.000000000 +0000
+++ 0.12+ds1-1/eval/vals/bool.go	2018-07-11 02:32:25.000000000 +0000
@@ -0,0 +1,20 @@
+package vals
+
+// Booler wraps the Bool method.
+type Booler interface {
+	// Bool computes the truth value of the receiver.
+	Bool() bool
+}
+
+// Bool converts a value to bool. It is implemented for the builtin bool type
+// and types implementing the Booler interface. For all other values, it returns
+// true.
+func Bool(v interface{}) bool {
+	switch v := v.(type) {
+	case Booler:
+		return v.Bool()
+	case bool:
+		return v
+	}
+	return true
+}
diff -pruN 0.11+ds1-5/eval/vals/bool_test.go 0.12+ds1-1/eval/vals/bool_test.go
--- 0.11+ds1-5/eval/vals/bool_test.go	1970-01-01 00:00:00.000000000 +0000
+++ 0.12+ds1-1/eval/vals/bool_test.go	2018-07-11 02:32:25.000000000 +0000
@@ -0,0 +1,27 @@
+package vals
+
+import (
+	"testing"
+
+	"github.com/elves/elvish/tt"
+)
+
+type customBooler struct{ b bool }
+
+func (b customBooler) Bool() bool { return b.b }
+
+type customNonBooler struct{}
+
+var boolTests = tt.Table{
+	Args(true).Rets(true),
+	Args(false).Rets(false),
+
+	Args(customBooler{true}).Rets(true),
+	Args(customBooler{false}).Rets(false),
+
+	Args(customNonBooler{}).Rets(true),
+}
+
+func TestBool(t *testing.T) {
+	tt.Test(t, tt.Fn("Bool", Bool), boolTests)
+}
diff -pruN 0.11+ds1-5/eval/vals/concat.go 0.12+ds1-1/eval/vals/concat.go
--- 0.11+ds1-5/eval/vals/concat.go	1970-01-01 00:00:00.000000000 +0000
+++ 0.12+ds1-1/eval/vals/concat.go	2018-07-11 02:32:25.000000000 +0000
@@ -0,0 +1,61 @@
+package vals
+
+import (
+	"errors"
+	"fmt"
+)
+
+// Concatter wraps the Concat method. See Concat for how it is used.
+type Concatter interface {
+	// Concat concatenates the receiver with another value, the receiver being
+	// the left operand. If concatenation is not supported for the given value,
+	// the method can return the special error type ErrCatNotImplemented.
+	Concat(v interface{}) (interface{}, error)
+}
+
+// RConcatter wraps the RConcat method. See Concat for how it is used.
+type RConcatter interface {
+	RConcat(v interface{}) (interface{}, error)
+}
+
+// ErrConcatNotImplemented is a special error value used to signal that
+// concatenation is not implemented. See Concat for how it is used.
+var ErrConcatNotImplemented = errors.New("cat not implemented")
+
+// Concat concatenates two values. If both operands are strings, it returns lhs
+// + rhs, nil. If the left operand implements Concatter, it calls
+// lhs.Concat(rhs). If lhs doesn't implement the interface or returned
+// ErrConcatNotImplemented, it then calls rhs.RConcat(lhs). If all attempts
+// fail, it returns nil and an error.
+func Concat(lhs, rhs interface{}) (interface{}, error) {
+	if v, ok := tryConcatBuiltins(lhs, rhs); ok {
+		return v, nil
+	}
+
+	if lhs, ok := lhs.(Concatter); ok {
+		v, err := lhs.Concat(rhs)
+		if err != ErrConcatNotImplemented {
+			return v, err
+		}
+	}
+
+	if rhs, ok := rhs.(RConcatter); ok {
+		v, err := rhs.RConcat(lhs)
+		if err != ErrConcatNotImplemented {
+			return v, err
+		}
+	}
+
+	return nil, fmt.Errorf("unsupported concat: %s and %s", Kind(lhs), Kind(rhs))
+}
+
+func tryConcatBuiltins(lhs, rhs interface{}) (interface{}, bool) {
+	switch lhs := lhs.(type) {
+	case string:
+		if rhs, ok := rhs.(string); ok {
+			return lhs + rhs, true
+		}
+	}
+
+	return nil, false
+}
diff -pruN 0.11+ds1-5/eval/vals/conversion.go 0.12+ds1-1/eval/vals/conversion.go
--- 0.11+ds1-5/eval/vals/conversion.go	1970-01-01 00:00:00.000000000 +0000
+++ 0.12+ds1-1/eval/vals/conversion.go	2018-07-11 02:32:25.000000000 +0000
@@ -0,0 +1,133 @@
+package vals
+
+import (
+	"fmt"
+	"reflect"
+	"strconv"
+	"unicode/utf8"
+)
+
+// Conversion between native and Elvish values.
+//
+// Elvish uses native Go types most of the time - string, bool, hashmap.Map,
+// vector.Vector, etc., and there is no need for any conversions. There are some
+// exceptions, for instance the numerical types int and float64: in native Go
+// code they are distinct types, but Elvish uses string for all numbers.
+// Similarly, Elvish uses string to represent runes. In all cases, there is a
+// many-to-one relationship between Go types and Elvish types.
+//
+// Conversion from Go value to Elvish value can happen without knowing the
+// destination type: int, float64 and rune values are converted to strings, and
+// values of other types remain unchanged. The opposite is not true: the native
+// Go value corresponding to the Elvish value "1" can be string("1"), int(1),
+// float64(1.0) or rune('1'). Conversion in this direction depends on the
+// destination type.
+
+// ScanToGo converts an Elvish value to a Go value. the pointer points to. It
+// uses the type of the pointer to determine the destination type, and puts the
+// converted value in the location the pointer points to. Conversion only
+// happens when the destination type is int, float64 or rune; in other cases,
+// this function just checks that the source value is already assignable to the
+// destination.
+func ScanToGo(src interface{}, ptr interface{}) error {
+	switch ptr := ptr.(type) {
+	case *int:
+		i, err := elvToInt(src)
+		if err == nil {
+			*ptr = i
+		}
+		return err
+	case *float64:
+		f, err := elvToFloat(src)
+		if err == nil {
+			*ptr = f
+		}
+		return err
+	case *rune:
+		r, err := elvToRune(src)
+		if err == nil {
+			*ptr = r
+		}
+		return err
+	case Scanner:
+		return ptr.ScanElvish(src)
+	default:
+		// Do a generic `*ptr = src` via reflection
+		ptrType := reflect.TypeOf(ptr)
+		if ptrType.Kind() != reflect.Ptr {
+			return fmt.Errorf("need pointer to scan into, got %T", ptr)
+		}
+		dstType := ptrType.Elem()
+		if !reflect.TypeOf(src).AssignableTo(dstType) {
+			return fmt.Errorf("need %s, got %s",
+				Kind(reflect.Zero(dstType).Interface()), Kind(src))
+		}
+		reflect.ValueOf(ptr).Elem().Set(reflect.ValueOf(src))
+		return nil
+	}
+}
+
+// Scanner is implemented by types that can scan an Elvish value into itself.
+type Scanner interface {
+	ScanElvish(interface{}) error
+}
+
+// FromGo converts a Go value to an Elvish value. Conversion happens when the
+// argument is int, float64 or rune (this is consistent with ScanToGo). In other
+// cases, this function just returns the argument.
+func FromGo(a interface{}) interface{} {
+	switch a := a.(type) {
+	case int:
+		return strconv.Itoa(a)
+	case float64:
+		return strconv.FormatFloat(a, 'g', -1, 64)
+	case rune:
+		return string(a)
+	default:
+		return a
+	}
+}
+
+func elvToFloat(arg interface{}) (float64, error) {
+	if _, ok := arg.(string); !ok {
+		return 0, fmt.Errorf("must be string")
+	}
+	s := arg.(string)
+	num, err := strconv.ParseFloat(s, 64)
+	if err != nil {
+		num, err2 := strconv.ParseInt(s, 0, 64)
+		if err2 != nil {
+			return 0, err
+		}
+		return float64(num), nil
+	}
+	return num, nil
+}
+
+func elvToInt(arg interface{}) (int, error) {
+	arg, ok := arg.(string)
+	if !ok {
+		return 0, fmt.Errorf("must be string")
+	}
+	num, err := strconv.ParseInt(arg.(string), 0, 0)
+	if err != nil {
+		return 0, err
+	}
+	return int(num), nil
+}
+
+func elvToRune(arg interface{}) (rune, error) {
+	ss, ok := arg.(string)
+	if !ok {
+		return -1, fmt.Errorf("must be string")
+	}
+	s := ss
+	r, size := utf8.DecodeRuneInString(s)
+	if r == utf8.RuneError {
+		return -1, fmt.Errorf("string is not valid UTF-8")
+	}
+	if size != len(s) {
+		return -1, fmt.Errorf("string has multiple runes")
+	}
+	return r, nil
+}
diff -pruN 0.11+ds1-5/eval/vals/conversion_test.go 0.12+ds1-1/eval/vals/conversion_test.go
--- 0.11+ds1-5/eval/vals/conversion_test.go	1970-01-01 00:00:00.000000000 +0000
+++ 0.12+ds1-1/eval/vals/conversion_test.go	2018-07-11 02:32:25.000000000 +0000
@@ -0,0 +1,55 @@
+package vals
+
+import (
+	"reflect"
+	"testing"
+
+	"github.com/elves/elvish/tt"
+)
+
+type someType struct {
+	foo string
+}
+
+// A wrapper around ScanToGo, to make it easier to test. Instead of supplying a
+// pointer to the destination, an initial value to the destination is supplied
+// and the result is returned.
+func scanToGo2(src interface{}, dstInit interface{}) (interface{}, error) {
+	ptr := reflect.New(reflect.TypeOf(dstInit))
+	err := ScanToGo(src, ptr.Interface())
+	return ptr.Elem().Interface(), err
+}
+
+var scanToGoTests = tt.Table{
+	Args("12", 0).Rets(12),
+	Args("0x12", 0).Rets(0x12),
+	Args("23", 0.0).Rets(23.0),
+	Args("0x23", 0.0).Rets(float64(0x23)),
+	Args("x", ' ').Rets('x'),
+	Args("foo", "").Rets("foo"),
+	Args(someType{"foo"}, someType{}).Rets(someType{"foo"}),
+
+	Args("x", someType{}).Rets(any, anyError),
+	Args(someType{}, 0).Rets(any, anyError),
+	Args("x", 0).Rets(any, anyError),
+	Args(someType{}, 0.0).Rets(any, anyError),
+	Args("x", 0.0).Rets(any, anyError),
+	Args(someType{}, ' ').Rets(any, anyError),
+	Args("\xc3\x28", ' ').Rets(any, anyError), // Invalid UTF8
+	Args("ab", ' ').Rets(any, anyError),
+}
+
+func TestScanToGo(t *testing.T) {
+	tt.Test(t, tt.Fn("ScanToGo", scanToGo2), scanToGoTests)
+}
+
+var fromGoTests = tt.Table{
+	tt.Args(12).Rets("12"),
+	tt.Args(1.5).Rets("1.5"),
+	tt.Args('x').Rets("x"),
+	tt.Args(someType{"foo"}).Rets(someType{"foo"}),
+}
+
+func TestFromGo(t *testing.T) {
+	tt.Test(t, tt.Fn("FromGo", FromGo), fromGoTests)
+}
diff -pruN 0.11+ds1-5/eval/vals/dissoc.go 0.12+ds1-1/eval/vals/dissoc.go
--- 0.11+ds1-5/eval/vals/dissoc.go	1970-01-01 00:00:00.000000000 +0000
+++ 0.12+ds1-1/eval/vals/dissoc.go	2018-07-11 02:32:25.000000000 +0000
@@ -0,0 +1,23 @@
+package vals
+
+// Dissocer wraps the Dissoc method.
+type Dissocer interface {
+	// Dissoc returns a slightly modified version of the receiver with key k
+	// dissociated with any value.
+	Dissoc(k interface{}) interface{}
+}
+
+// Dissoc takes a container and a key, and returns a modified version of the
+// container, with the given key dissociated with any value. It is implemented
+// for types satisfying the mapDissocable or Dissocer interface. For other
+// types, it returns nil.
+func Dissoc(a, k interface{}) interface{} {
+	switch a := a.(type) {
+	case Dissocer:
+		return a.Dissoc(k)
+	case mapDissocable:
+		return a.Dissoc(k)
+	default:
+		return nil
+	}
+}
diff -pruN 0.11+ds1-5/eval/vals/dissoc_test.go 0.12+ds1-1/eval/vals/dissoc_test.go
--- 0.11+ds1-5/eval/vals/dissoc_test.go	1970-01-01 00:00:00.000000000 +0000
+++ 0.12+ds1-1/eval/vals/dissoc_test.go	2018-07-11 02:32:25.000000000 +0000
@@ -0,0 +1,21 @@
+package vals
+
+import (
+	"testing"
+
+	"github.com/elves/elvish/tt"
+)
+
+type customDissocer struct{}
+
+func (customDissocer) Dissoc(interface{}) interface{} { return "custom ret" }
+
+var dissocTests = tt.Table{
+	Args(MakeMapFromKV("k1", "v1", "k2", "v2"), "k1").Rets(
+		eq(MakeMapFromKV("k2", "v2"))),
+	Args(customDissocer{}, "x").Rets("custom ret"),
+}
+
+func TestDissoc(t *testing.T) {
+	tt.Test(t, tt.Fn("Dissoc", Dissoc), dissocTests)
+}
diff -pruN 0.11+ds1-5/eval/vals/equal.go 0.12+ds1-1/eval/vals/equal.go
--- 0.11+ds1-5/eval/vals/equal.go	1970-01-01 00:00:00.000000000 +0000
+++ 0.12+ds1-1/eval/vals/equal.go	2018-07-11 02:32:25.000000000 +0000
@@ -0,0 +1,83 @@
+package vals
+
+import (
+	"reflect"
+
+	"github.com/xiaq/persistent/hashmap"
+)
+
+// Equaler wraps the Equal method.
+type Equaler interface {
+	// Equal compares the receiver to another value. Two equal values must have
+	// the same hash code.
+	Equal(other interface{}) bool
+}
+
+// Equal returns whether two values are equal. It is implemented for the builtin
+// types bool and string, and types satisfying the listEqualable, mapEqualable
+// or Equaler interface. For other types, it uses reflect.DeepEqual to compare
+// the two values.
+func Equal(x, y interface{}) bool {
+	switch x := x.(type) {
+	case Equaler:
+		return x.Equal(y)
+	case bool:
+		return x == y
+	case string:
+		return x == y
+	case listEqualable:
+		if yy, ok := y.(listEqualable); ok {
+			return equalList(x, yy)
+		}
+		return false
+	case mapEqualable:
+		if yy, ok := y.(mapEqualable); ok {
+			return equalMap(x, yy)
+		}
+		return false
+	}
+	return reflect.DeepEqual(x, y)
+}
+
+type listEqualable interface {
+	Lener
+	listIterable
+}
+
+func equalList(x, y listEqualable) bool {
+	if x.Len() != y.Len() {
+		return false
+	}
+	ix := x.Iterator()
+	iy := y.Iterator()
+	for ix.HasElem() && iy.HasElem() {
+		if !Equal(ix.Elem(), iy.Elem()) {
+			return false
+		}
+		ix.Next()
+		iy.Next()
+	}
+	return true
+}
+
+type mapEqualable interface {
+	Lener
+	Index(interface{}) (interface{}, bool)
+	Iterator() hashmap.Iterator
+}
+
+var _ mapEqualable = hashmap.Map(nil)
+
+func equalMap(x, y mapEqualable) bool {
+	if x.Len() != y.Len() {
+		return false
+	}
+	for it := x.Iterator(); it.HasElem(); it.Next() {
+		k, vx := it.Elem()
+		vy, ok := y.Index(k)
+		if !ok || !Equal(vx, vy) {
+			return false
+		}
+	}
+	return true
+}
diff -pruN 0.11+ds1-5/eval/vals/equal_test.go 0.12+ds1-1/eval/vals/equal_test.go
--- 0.11+ds1-5/eval/vals/equal_test.go	1970-01-01 00:00:00.000000000 +0000
+++ 0.12+ds1-1/eval/vals/equal_test.go	2018-07-11 02:32:25.000000000 +0000
@@ -0,0 +1,33 @@
+package vals
+
+import (
+	"testing"
+
+	"github.com/elves/elvish/tt"
+)
+
+type customEqualer struct{ ret bool }
+
+func (c customEqualer) Equal(interface{}) bool { return c.ret }
+
+type customStruct struct{ a, b string }
+
+var equalTests = tt.Table{
+	Args(customEqualer{true}, 2).Rets(true),
+	Args(customEqualer{false}, 2).Rets(false),
+	Args(true, true).Rets(true),
+	Args(true, false).Rets(false),
+	Args("lorem", "lorem").Rets(true),
+	Args("lorem", "ipsum").Rets(false),
+	Args(MakeList("a", "b"), MakeList("a", "b")).Rets(true),
+	Args(MakeList("a", "b"), MakeList("a")).Rets(false),
+	Args(MakeList("a", "b"), MakeList("a", "c")).Rets(false),
+	Args(MakeMapFromKV("k", "v"), MakeMapFromKV("k", "v")).Rets(true),
+	Args(MakeMapFromKV("k", "v"), MakeMapFromKV("k2", "v")).Rets(false),
+	Args(MakeMapFromKV("k", "v", "k2", "v2"), MakeMapFromKV("k", "v")).Rets(false),
+	Args(&customStruct{"a", "b"}, &customStruct{"a", "b"}).Rets(true),
+}
+
+func TestEqual(t *testing.T) {
+	tt.Test(t, tt.Fn("Equal", Equal), equalTests)
+}
diff -pruN 0.11+ds1-5/eval/vals/file.go 0.12+ds1-1/eval/vals/file.go
--- 0.11+ds1-5/eval/vals/file.go	1970-01-01 00:00:00.000000000 +0000
+++ 0.12+ds1-1/eval/vals/file.go	2018-07-11 02:32:25.000000000 +0000
@@ -0,0 +1,37 @@
+package vals
+
+import (
+	"fmt"
+	"os"
+
+	"github.com/elves/elvish/parse"
+	"github.com/xiaq/persistent/hash"
+)
+
+// File wraps a pointer to os.File.
+type File struct {
+	Inner *os.File
+}
+
+var _ interface{} = File{}
+
+// NewFile creates a new File value.
+func NewFile(inner *os.File) File {
+	return File{inner}
+}
+
+func (File) Kind() string {
+	return "file"
+}
+
+func (f File) Equal(rhs interface{}) bool {
+	return f == rhs
+}
+
+func (f File) Hash() uint32 {
+	return hash.UIntPtr(f.Inner.Fd())
+}
+
+func (f File) Repr(int) string {
+	return fmt.Sprintf("<file{%s %p}>", parse.Quote(f.Inner.Name()), f.Inner)
+}
diff -pruN 0.11+ds1-5/eval/vals/hash.go 0.12+ds1-1/eval/vals/hash.go
--- 0.11+ds1-5/eval/vals/hash.go	1970-01-01 00:00:00.000000000 +0000
+++ 0.12+ds1-1/eval/vals/hash.go	2018-07-11 02:32:25.000000000 +0000
@@ -0,0 +1,47 @@
+package vals
+
+import (
+	"github.com/xiaq/persistent/hash"
+)
+
+// Hasher wraps the Hash method.
+type Hasher interface {
+	// Hash computes the hash code of the receiver.
+	Hash() uint32
+}
+
+// Hash returns the 32-bit hash of a value. It is implemented for the builtin
+// types bool and string, and types satisfying the listHashable, mapHashable or
+// Hasher interface. For other values, it returns 0 (which is OK in terms of
+// correctness).
+func Hash(v interface{}) uint32 {
+	switch v := v.(type) {
+	case bool:
+		if v {
+			return 1
+		}
+		return 0
+	case listHashable:
+		h := hash.DJBInit
+		for it := v.Iterator(); it.HasElem(); it.Next() {
+			h = hash.DJBCombine(h, Hash(it.Elem()))
+		}
+		return h
+	case mapIterable:
+		h := hash.DJBInit
+		for it := v.Iterator(); it.HasElem(); it.Next() {
+			k, v := it.Elem()
+			h = hash.DJBCombine(h, Hash(k))
+			h = hash.DJBCombine(h, Hash(v))
+		}
+		return h
+	case string:
+		return hash.String(v)
+	case Hasher:
+		return v.Hash()
+	}
+	return 0
+}
+
+type listHashable listIterable
+type mapHashable mapIterable
diff -pruN 0.11+ds1-5/eval/vals/index.go 0.12+ds1-1/eval/vals/index.go
--- 0.11+ds1-5/eval/vals/index.go	1970-01-01 00:00:00.000000000 +0000
+++ 0.12+ds1-1/eval/vals/index.go	2018-07-11 02:32:25.000000000 +0000
@@ -0,0 +1,57 @@
+package vals
+
+import (
+	"errors"
+)
+
+// Indexer wraps the Index method.
+type Indexer interface {
+	// Index retrieves the value corresponding to the specified key in the
+	// container. It returns the value (if any), and whether it actually exists.
+	Index(k interface{}) (v interface{}, ok bool)
+}
+
+// ErrIndexer wraps the Index method.
+type ErrIndexer interface {
+	// Index retrieves one value from the receiver at the specified index.
+	Index(k interface{}) (interface{}, error)
+}
+
+var errNotIndexable = errors.New("not indexable")
+
+type noSuchKeyError struct {
+	key interface{}
+}
+
+// NoSuchKey returns an error indicating that a key is not found in a map-like
+// value.
+func NoSuchKey(k interface{}) error {
+	return noSuchKeyError{k}
+}
+
+func (err noSuchKeyError) Error() string {
+	return "no such key: " + Repr(err.key, NoPretty)
+}
+
+// Index indexes a value with the given key. It is implemented for types
+// satisfying the ErrIndexer or Indexer interface, types satisfying the
+// listIndexable interface (which covers Vector), and the builtin string type.
+// For other types, it returns a nil value and a non-nil error.
+func Index(a, k interface{}) (interface{}, error) {
+	switch a := a.(type) {
+	case ErrIndexer:
+		return a.Index(k)
+	case Indexer:
+		v, ok := a.Index(k)
+		if !ok {
+			return nil, NoSuchKey(k)
+		}
+		return v, nil
+	case listIndexable:
+		return indexList(a, k)
+	case string:
+		return indexString(a, k)
+	default:
+		return nil, errNotIndexable
+	}
+}
diff -pruN 0.11+ds1-5/eval/vals/index_list.go 0.12+ds1-1/eval/vals/index_list.go
--- 0.11+ds1-5/eval/vals/index_list.go	1970-01-01 00:00:00.000000000 +0000
+++ 0.12+ds1-1/eval/vals/index_list.go	2018-07-11 02:32:25.000000000 +0000
@@ -0,0 +1,117 @@
+package vals
+
+import (
+	"errors"
+	"strconv"
+	"strings"
+
+	"github.com/xiaq/persistent/vector"
+)
+
+var (
+	errIndexMustBeString = errors.New("index must be string")
+	errIndexMustBeNumber = errors.New("index or slice component must be number")
+	errIndexOutOfRange   = errors.New("index out of range")
+)
+
+type listIndexable interface {
+	Lener
+	Index(int) (interface{}, bool)
+	SubVector(int, int) vector.Vector
+}
+
+var _ listIndexable = vector.Vector(nil)
+
+func indexList(l listIndexable, rawIndex interface{}) (interface{}, error) {
+	index, err := ConvertListIndex(rawIndex, l.Len())
+	if err != nil {
+		return nil, err
+	}
+	if index.Slice {
+		return l.SubVector(index.Lower, index.Upper), nil
+	}
+	// Bounds are already checked.
+	value, _ := l.Index(index.Lower)
+	return value, nil
+}
+
+// ListIndex represents a (converted) list index.
+type ListIndex struct {
+	Slice bool
+	Lower int
+	Upper int
+}
+
+// ConvertListIndex parses a list index, check whether it is valid, and returns
+// the converted structure.
+func ConvertListIndex(rawIndex interface{}, n int) (*ListIndex, error) {
+	s, ok := rawIndex.(string)
+	if !ok {
+		return nil, errIndexMustBeString
+	}
+	slice, i, j, err := parseListIndex(s, n)
+	if err != nil {
+		return nil, err
+	}
+	if i < 0 {
+		i += n
+	}
+	if j < 0 {
+		j += n
+	}
+	if slice {
+		if !(0 <= i && i <= j && j <= n) {
+			return nil, errIndexOutOfRange
+		}
+	} else {
+		if !(0 <= i && i < n) {
+			return nil, errIndexOutOfRange
+		}
+	}
+	return &ListIndex{slice, i, j}, nil
+}
+
+// ListIndex = Number |
+//             Number ':' Number
+func parseListIndex(s string, n int) (slice bool, i int, j int, err error) {
+	colon := strings.IndexRune(s, ':')
+	if colon == -1 {
+		// A single number
+		i, err := atoi(s)
+		if err != nil {
+			return false, 0, 0, err
+		}
+		return false, i, 0, nil
+	}
+	if s[:colon] == "" {
+		i = 0
+	} else {
+		i, err = atoi(s[:colon])
+		if err != nil {
+			return false, 0, 0, err
+		}
+	}
+	if s[colon+1:] == "" {
+		j = n
+	} else {
+		j, err = atoi(s[colon+1:])
+		if err != nil {
+			return false, 0, 0, err
+		}
+	}
+	// Two numbers
+	return true, i, j, nil
+}
+
+// atoi is a wrapper around strconv.Atoi, converting strconv.ErrRange to
+// errIndexOutOfRange.
+func atoi(a string) (int, error) {
+	i, err := strconv.Atoi(a)
+	if err != nil {
+		if err.(*strconv.NumError).Err == strconv.ErrRange {
+			return 0, errIndexOutOfRange
+		}
+		return 0, errIndexMustBeNumber
+	}
+	return i, nil
+}
diff -pruN 0.11+ds1-5/eval/vals/index_string.go 0.12+ds1-1/eval/vals/index_string.go
--- 0.11+ds1-5/eval/vals/index_string.go	1970-01-01 00:00:00.000000000 +0000
+++ 0.12+ds1-1/eval/vals/index_string.go	2018-07-11 02:32:25.000000000 +0000
@@ -0,0 +1,34 @@
+package vals
+
+import (
+	"errors"
+	"unicode/utf8"
+)
+
+var errIndexNotAtRuneBoundary = errors.New("index not at rune boundary")
+
+func indexString(s string, index interface{}) (string, error) {
+	i, j, err := convertStringIndex(index, s)
+	if err != nil {
+		return "", err
+	}
+	return s[i:j], nil
+}
+
+func convertStringIndex(rawIndex interface{}, s string) (int, int, error) {
+	index, err := ConvertListIndex(rawIndex, len(s))
+	if err != nil {
+		return 0, 0, err
+	}
+	r, size := utf8.DecodeRuneInString(s[index.Lower:])
+	if r == utf8.RuneError {
+		return 0, 0, errIndexNotAtRuneBoundary
+	}
+	if index.Slice {
+		if r, _ := utf8.DecodeLastRuneInString(s[:index.Upper]); r == utf8.RuneError {
+			return 0, 0, errIndexNotAtRuneBoundary
+		}
+		return index.Lower, index.Upper, nil
+	}
+	return index.Lower, index.Lower + size, nil
+}
diff -pruN 0.11+ds1-5/eval/vals/index_test.go 0.12+ds1-1/eval/vals/index_test.go
--- 0.11+ds1-5/eval/vals/index_test.go	1970-01-01 00:00:00.000000000 +0000
+++ 0.12+ds1-1/eval/vals/index_test.go	2018-07-11 02:32:25.000000000 +0000
@@ -0,0 +1,57 @@
+package vals
+
+import (
+	"testing"
+
+	"github.com/elves/elvish/tt"
+)
+
+var (
+	li0 = EmptyList
+	li4 = MakeList("foo", "bar", "lorem", "ipsum")
+	m   = MakeMapFromKV("foo", "bar", "lorem", "ipsum")
+)
+
+var indexTests = tt.Table{
+	// List indices
+	// ============
+
+	// Simple indicies: 0 <= i < n.
+	Args(li4, "0").Rets("foo", nil),
+	Args(li4, "3").Rets("ipsum", nil),
+	Args(li0, "0").Rets(any, anyError),
+	Args(li4, "4").Rets(any, anyError),
+	Args(li4, "5").Rets(any, anyError),
+	// Negative indices: -n <= i < 0.
+	Args(li4, "-1").Rets("ipsum", nil),
+	Args(li4, "-4").Rets("foo", nil),
+	Args(li4, "-5").Rets(any, anyError), // Too negative.
+	// Decimal indicies are not allowed even if the value is an integer.
+	Args(li4, "0.0").Rets(any, anyError),
+
+	// Slice indicies: 0 <= i <= j <= n.
+	Args(li4, "1:3").Rets(eq(MakeList("bar", "lorem")), nil),
+	Args(li4, "3:4").Rets(eq(MakeList("ipsum")), nil),
+	Args(li4, "4:4").Rets(eq(EmptyList), nil), // i == j == n is allowed.
+	// i defaults to 0
+	Args(li4, ":2").Rets(eq(MakeList("foo", "bar")), nil),
+	Args(li4, ":-1").Rets(eq(MakeList("foo", "bar", "lorem")), nil),
+	// j defaults to n
+	Args(li4, "3:").Rets(eq(MakeList("ipsum")), nil),
+	Args(li4, "-2:").Rets(eq(MakeList("lorem", "ipsum")), nil),
+	// Both indices can be omitted.
+	Args(li0, ":").Rets(eq(li0), nil),
+	Args(li4, ":").Rets(eq(li4), nil),
+
+	// Malformed list indices.
+	Args(li4, "a").Rets(any, anyError),
+	Args(li4, "1:3:2").Rets(any, anyError),
+
+	// Map indicies
+	Args(m, "foo").Rets("bar", nil),
+	Args(m, "bad").Rets(any, anyError),
+}
+
+func TestIndex(t *testing.T) {
+	tt.Test(t, tt.Fn("Index", Index), indexTests)
+}
diff -pruN 0.11+ds1-5/eval/vals/iterate.go 0.12+ds1-1/eval/vals/iterate.go
--- 0.11+ds1-5/eval/vals/iterate.go	1970-01-01 00:00:00.000000000 +0000
+++ 0.12+ds1-1/eval/vals/iterate.go	2018-07-11 02:32:25.000000000 +0000
@@ -0,0 +1,61 @@
+package vals
+
+import (
+	"errors"
+
+	"github.com/xiaq/persistent/vector"
+)
+
+// Iterator wraps the Iterate method.
+type Iterator interface {
+	// Iterate calls the passed function with each value within the receiver.
+	// The iteration is aborted if the function returns false.
+	Iterate(func(v interface{}) bool)
+}
+
+// Iterate iterates the supplied value, and calls the supplied function in each
+// of its elements. The function can return false to break the iteration. It is
+// implemented for the builtin type string, and types satisfying the
+// listIterable or Iterator interface. For these types, it always returns a nil
+// error. For other types, it doesn't do anything and returns an error.
+func Iterate(v interface{}, f func(interface{}) bool) error {
+	switch v := v.(type) {
+	case Iterator:
+		v.Iterate(f)
+	case string:
+		for _, r := range v {
+			b := f(string(r))
+			if !b {
+				break
+			}
+		}
+	case listIterable:
+		for it := v.Iterator(); it.HasElem(); it.Next() {
+			if !f(it.Elem()) {
+				break
+			}
+		}
+	default:
+		return errors.New(Kind(v) + " cannot be iterated")
+	}
+	return nil
+}
+
+type listIterable interface {
+	Iterator() vector.Iterator
+}
+
+var _ listIterable = vector.Vector(nil)
+
+// Collect collects all elements of an iterable value into a slice.
+func Collect(it interface{}) ([]interface{}, error) {
+	var vs []interface{}
+	if len := Len(it); len >= 0 {
+		vs = make([]interface{}, 0, len)
+	}
+	err := Iterate(it, func(v interface{}) bool {
+		vs = append(vs, v)
+		return true
+	})
+	return vs, err
+}
diff -pruN 0.11+ds1-5/eval/vals/iterate_keys.go 0.12+ds1-1/eval/vals/iterate_keys.go
--- 0.11+ds1-5/eval/vals/iterate_keys.go	1970-01-01 00:00:00.000000000 +0000
+++ 0.12+ds1-1/eval/vals/iterate_keys.go	2018-07-11 02:32:25.000000000 +0000
@@ -0,0 +1,50 @@
+package vals
+
+import (
+	"errors"
+
+	"github.com/xiaq/persistent/hashmap"
+)
+
+// KeysIterator wraps the IterateKeys method.
+type KeysIterator interface {
+	// IterateKeys calls the passed function with each key within the receiver.
+	// The iteration is aborted if the function returns false.
+	IterateKeys(func(v interface{}) bool)
+}
+
+// IterateKeys iterates the keys of the supplied value, calling the supplied
+// function for each key. The function can return false to break the iteration.
+// It is implemented for the mapKeysIterable type and types satisfying the
+// IterateKeyser interface. For these types, it always returns a nil error. For
+// other types, it doesn't do anything and returns an error.
+func IterateKeys(v interface{}, f func(interface{}) bool) error {
+	switch v := v.(type) {
+	case KeysIterator:
+		v.IterateKeys(f)
+	case mapKeysIterable:
+		for it := v.Iterator(); it.HasElem(); it.Next() {
+			k, _ := it.Elem()
+			if !f(k) {
+				break
+			}
+		}
+	default:
+		return errors.New(Kind(v) + " cannot have its keys iterated")
+	}
+	return nil
+}
+
+type mapKeysIterable interface {
+	Iterator() hashmap.Iterator
+}
+
+// Feed calls the function with given values, breaking earlier if the function
+// returns false.
+func Feed(f func(interface{}) bool, values ...interface{}) {
+	for _, value := range values {
+		if !f(value) {
+			break
+		}
+	}
+}
diff -pruN 0.11+ds1-5/eval/vals/kind.go 0.12+ds1-1/eval/vals/kind.go
--- 0.11+ds1-5/eval/vals/kind.go	1970-01-01 00:00:00.000000000 +0000
+++ 0.12+ds1-1/eval/vals/kind.go	2018-07-11 02:32:25.000000000 +0000
@@ -0,0 +1,34 @@
+package vals
+
+import (
+	"fmt"
+
+	"github.com/xiaq/persistent/hashmap"
+	"github.com/xiaq/persistent/vector"
+)
+
+// Kinder wraps the Kind method.
+type Kinder interface {
+	Kind() string
+}
+
+// Kind returns the "kind" of the value, a concept similar to type but not yet
+// very well defined. It is implemented for the builtin types bool and string,
+// the Vector and Map types, and types implementing the Kinder interface. For
+// other types, it returns the Go type name of the argument preceeded by "!!".
+func Kind(v interface{}) string {
+	switch v := v.(type) {
+	case Kinder:
+		return v.Kind()
+	case bool:
+		return "bool"
+	case string:
+		return "string"
+	case vector.Vector:
+		return "list"
+	case hashmap.Map:
+		return "map"
+	default:
+		return fmt.Sprintf("!!%T", v)
+	}
+}
diff -pruN 0.11+ds1-5/eval/vals/kind_test.go 0.12+ds1-1/eval/vals/kind_test.go
--- 0.11+ds1-5/eval/vals/kind_test.go	1970-01-01 00:00:00.000000000 +0000
+++ 0.12+ds1-1/eval/vals/kind_test.go	2018-07-11 02:32:25.000000000 +0000
@@ -0,0 +1,24 @@
+package vals
+
+import (
+	"os"
+	"testing"
+
+	"github.com/elves/elvish/tt"
+)
+
+type xtype int
+
+func TestKind(t *testing.T) {
+	tt.Test(t, tt.Fn("Kind", Kind), tt.Table{
+		Args(true).Rets("bool"),
+		Args("").Rets("string"),
+		Args(EmptyList).Rets("list"),
+		Args(EmptyMap).Rets("map"),
+		Args(xtype(0)).Rets("!!vals.xtype"),
+
+		Args(NewStruct(NewStructDescriptor(), nil)).Rets("map"),
+		Args(NewFile(os.Stdin)).Rets("file"),
+		Args(NewPipe(os.Stdin, os.Stdout)).Rets("pipe"),
+	})
+}
diff -pruN 0.11+ds1-5/eval/vals/len.go 0.12+ds1-1/eval/vals/len.go
--- 0.11+ds1-5/eval/vals/len.go	1970-01-01 00:00:00.000000000 +0000
+++ 0.12+ds1-1/eval/vals/len.go	2018-07-11 02:32:25.000000000 +0000
@@ -0,0 +1,24 @@
+package vals
+
+import "github.com/xiaq/persistent/vector"
+
+// Lener wraps the Len method.
+type Lener interface {
+	// Len computes the length of the receiver.
+	Len() int
+}
+
+var _ Lener = vector.Vector(nil)
+
+// Len returns the length of the value, or -1 if the value does not have a
+// well-defined length. It is implemented for the builtin string type and types
+// satisfying the Lener interface. For other types, it returns -1.
+func Len(v interface{}) int {
+	switch v := v.(type) {
+	case Lener:
+		return v.Len()
+	case string:
+		return len(v)
+	}
+	return -1
+}
diff -pruN 0.11+ds1-5/eval/vals/list.go 0.12+ds1-1/eval/vals/list.go
--- 0.11+ds1-5/eval/vals/list.go	1970-01-01 00:00:00.000000000 +0000
+++ 0.12+ds1-1/eval/vals/list.go	2018-07-11 02:32:25.000000000 +0000
@@ -0,0 +1,62 @@
+package vals
+
+import (
+	"bytes"
+	"strings"
+
+	"github.com/xiaq/persistent/vector"
+)
+
+// EmptyList is an empty list.
+var EmptyList = vector.Empty
+
+// MakeList creates a new List from values.
+func MakeList(vs ...interface{}) vector.Vector {
+	vec := vector.Empty
+	for _, v := range vs {
+		vec = vec.Cons(v)
+	}
+	return vec
+}
+
+// MakeList creates a new List from strings.
+func MakeStringList(vs ...string) vector.Vector {
+	vec := vector.Empty
+	for _, v := range vs {
+		vec = vec.Cons(v)
+	}
+	return vec
+}
+
+// ListReprBuilder helps to build Repr of list-like Values.
+type ListReprBuilder struct {
+	Indent int
+	buf    bytes.Buffer
+}
+
+func (b *ListReprBuilder) WriteElem(v string) {
+	if b.buf.Len() == 0 {
+		b.buf.WriteByte('[')
+	}
+	if b.Indent >= 0 {
+		// Pretty printing.
+		//
+		// Add a newline and indent+1 spaces, so that the
+		// starting & lines up with the first pair.
+		b.buf.WriteString("\n" + strings.Repeat(" ", b.Indent+1))
+	} else if b.buf.Len() > 1 {
+		b.buf.WriteByte(' ')
+	}
+	b.buf.WriteString(v)
+}
+
+func (b *ListReprBuilder) String() string {
+	if b.buf.Len() == 0 {
+		return "[]"
+	}
+	if b.Indent >= 0 {
+		b.buf.WriteString("\n" + strings.Repeat(" ", b.Indent))
+	}
+	b.buf.WriteByte(']')
+	return b.buf.String()
+}
diff -pruN 0.11+ds1-5/eval/vals/map.go 0.12+ds1-1/eval/vals/map.go
--- 0.11+ds1-5/eval/vals/map.go	1970-01-01 00:00:00.000000000 +0000
+++ 0.12+ds1-1/eval/vals/map.go	2018-07-11 02:32:25.000000000 +0000
@@ -0,0 +1,70 @@
+package vals
+
+import (
+	"github.com/xiaq/persistent/hashmap"
+)
+
+type mapIterable interface {
+	Iterator() hashmap.Iterator
+}
+type mapAssocable interface {
+	Assoc(k, v interface{}) hashmap.Map
+}
+type mapDissocable interface {
+	Dissoc(interface{}) hashmap.Map
+}
+
+var (
+	_ mapIterable   = hashmap.Map(nil)
+	_ Indexer       = hashmap.Map(nil)
+	_ mapAssocable  = hashmap.Map(nil)
+	_ mapDissocable = hashmap.Map(nil)
+)
+
+// EmptyMap is an empty map.
+var EmptyMap = hashmap.New(Equal, Hash)
+
+// MakeMap converts a native Go map to Map.
+func MakeMap(raw map[interface{}]interface{}) hashmap.Map {
+	m := EmptyMap
+	for k, v := range raw {
+		m = m.Assoc(k, v)
+	}
+	return m
+}
+
+// MakeMapFromKV creates a map from arguments that are alternately keys and
+// values. It panics if the number of arguments is odd.
+func MakeMapFromKV(a ...interface{}) hashmap.Map {
+	if len(a)%2 == 1 {
+		panic("Odd number of arguments to MakeMapFromKV")
+	}
+	m := EmptyMap
+	for i := 0; i < len(a); i += 2 {
+		m = m.Assoc(a[i], a[i+1])
+	}
+	return m
+}
+
+// MapReprBuilder helps building the Repr of a Map. It is also useful for
+// implementing other Map-like values. The zero value of a MapReprBuilder is
+// ready to use.
+type MapReprBuilder struct {
+	ListReprBuilder
+}
+
+func (b *MapReprBuilder) WritePair(k string, indent int, v string) {
+	if indent > 0 {
+		b.WriteElem("&" + k + "=\t" + v)
+	} else {
+		b.WriteElem("&" + k + "=" + v)
+	}
+}
+
+func (b *MapReprBuilder) String() string {
+	s := b.ListReprBuilder.String()
+	if s == "[]" {
+		s = "[&]"
+	}
+	return s
+}
diff -pruN 0.11+ds1-5/eval/vals/pipe.go 0.12+ds1-1/eval/vals/pipe.go
--- 0.11+ds1-5/eval/vals/pipe.go	1970-01-01 00:00:00.000000000 +0000
+++ 0.12+ds1-1/eval/vals/pipe.go	2018-07-11 02:32:25.000000000 +0000
@@ -0,0 +1,40 @@
+package vals
+
+import (
+	"fmt"
+	"os"
+
+	"github.com/xiaq/persistent/hash"
+)
+
+// Pipe wraps a pair of pointers to os.File that are the two ends of the same
+// pipe.
+type Pipe struct {
+	ReadEnd, WriteEnd *os.File
+}
+
+var _ interface{} = Pipe{}
+
+// NewPipe creates a new Pipe value.
+func NewPipe(r, w *os.File) Pipe {
+	return Pipe{r, w}
+}
+
+func (Pipe) Kind() string {
+	return "pipe"
+}
+
+func (p Pipe) Equal(rhs interface{}) bool {
+	return p == rhs
+}
+
+func (p Pipe) Hash() uint32 {
+	h := hash.DJBInit
+	h = hash.DJBCombine(h, hash.UIntPtr(p.ReadEnd.Fd()))
+	h = hash.DJBCombine(h, hash.UIntPtr(p.WriteEnd.Fd()))
+	return h
+}
+
+func (p Pipe) Repr(int) string {
+	return fmt.Sprintf("<pipe{%v %v}>", p.ReadEnd.Fd(), p.WriteEnd.Fd())
+}
diff -pruN 0.11+ds1-5/eval/vals/rat.go 0.12+ds1-1/eval/vals/rat.go
--- 0.11+ds1-5/eval/vals/rat.go	1970-01-01 00:00:00.000000000 +0000
+++ 0.12+ds1-1/eval/vals/rat.go	2018-07-11 02:32:25.000000000 +0000
@@ -0,0 +1,67 @@
+package vals
+
+import (
+	"errors"
+	"fmt"
+	"math/big"
+
+	"github.com/xiaq/persistent/hash"
+)
+
+var ErrOnlyStrOrRat = errors.New("only str or rat may be converted to rat")
+
+// Rat is a rational number.
+type Rat struct {
+	b *big.Rat
+}
+
+var _ interface{} = Rat{}
+
+func (Rat) Kind() string {
+	return "string"
+}
+
+func (r Rat) Equal(a interface{}) bool {
+	if r == a {
+		return true
+	}
+	r2, ok := a.(Rat)
+	if !ok {
+		return false
+	}
+	return r.b.Cmp(r2.b) == 0
+}
+
+func (r Rat) Hash() uint32 {
+	// TODO(xiaq): Use a more efficient implementation.
+	return hash.String(r.String())
+}
+
+func (r Rat) Repr(int) string {
+	return "(rat " + r.String() + ")"
+}
+
+func (r Rat) String() string {
+	if r.b.IsInt() {
+		return r.b.Num().String()
+	}
+	return r.b.String()
+}
+
+// ToRat converts a Value to rat. A str can be converted to a rat if it can be
+// parsed. A rat is returned as-is. Other types of values cannot be converted.
+func ToRat(v interface{}) (Rat, error) {
+	switch v := v.(type) {
+	case Rat:
+		return v, nil
+	case string:
+		r := big.Rat{}
+		_, err := fmt.Sscanln(string(v), &r)
+		if err != nil {
+			return Rat{}, fmt.Errorf("%s cannot be parsed as rat", Repr(v, NoPretty))
+		}
+		return Rat{&r}, nil
+	default:
+		return Rat{}, ErrOnlyStrOrRat
+	}
+}
diff -pruN 0.11+ds1-5/eval/vals/repr.go 0.12+ds1-1/eval/vals/repr.go
--- 0.11+ds1-5/eval/vals/repr.go	1970-01-01 00:00:00.000000000 +0000
+++ 0.12+ds1-1/eval/vals/repr.go	2018-07-11 02:32:25.000000000 +0000
@@ -0,0 +1,64 @@
+package vals
+
+import (
+	"fmt"
+
+	"github.com/elves/elvish/parse"
+	"github.com/elves/elvish/util"
+)
+
+// NoPretty can be passed to Repr to suppress pretty-printing.
+const NoPretty = util.MinInt
+
+// Reprer wraps the Repr method.
+type Reprer interface {
+	// Repr returns a string that represents a Value. The string either be a
+	// literal of that Value that is preferably deep-equal to it (like `[a b c]`
+	// for a list), or a string enclosed in "<>" containing the kind and
+	// identity of the Value(like `<fn 0xdeadcafe>`).
+	//
+	// If indent is at least 0, it should be pretty-printed with the current
+	// indentation level of indent; the indent of the first line has already
+	// been written and shall not be written in Repr. The returned string
+	// should never contain a trailing newline.
+	Repr(indent int) string
+}
+
+// Repr returns the representation for a value, a string that is preferably (but
+// not necessarily) an Elvish expression that evaluates to the argument. If
+// indent >= 0, the representation is pretty-printed. It is implemented for the
+// builtin types bool and string, and types satisfying the listReprable,
+// mapReprable or Reprer interface. For other types, it uses fmt.Sprint with the
+// format "<unknown %v>".
+func Repr(v interface{}, indent int) string {
+	switch v := v.(type) {
+	case Reprer:
+		return v.Repr(indent)
+	case bool:
+		if v {
+			return "$true"
+		}
+		return "$false"
+	case string:
+		return parse.Quote(v)
+	case listReprable:
+		b := ListReprBuilder{Indent: indent}
+		for it := v.Iterator(); it.HasElem(); it.Next() {
+			b.WriteElem(Repr(it.Elem(), indent+1))
+		}
+		return b.String()
+	case mapReprable:
+		builder := MapReprBuilder{}
+		builder.Indent = indent
+		for it := v.Iterator(); it.HasElem(); it.Next() {
+			k, v := it.Elem()
+			builder.WritePair(Repr(k, indent+1), indent+2, Repr(v, indent+2))
+		}
+		return builder.String()
+	default:
+		return fmt.Sprintf("<unknown %v>", v)
+	}
+}
+
+type listReprable listIterable
+type mapReprable mapIterable
diff -pruN 0.11+ds1-5/eval/vals/string.go 0.12+ds1-1/eval/vals/string.go
--- 0.11+ds1-5/eval/vals/string.go	1970-01-01 00:00:00.000000000 +0000
+++ 0.12+ds1-1/eval/vals/string.go	2018-07-11 02:32:25.000000000 +0000
@@ -0,0 +1,20 @@
+package vals
+
+// Stringer wraps the String method.
+type Stringer interface {
+	// Stringer converts the receiver to a string.
+	String() string
+}
+
+// ToString converts a Value to string. When the Value type implements
+// String(), it is used. Otherwise Repr(NoPretty) is used.
+func ToString(v interface{}) string {
+	switch v := v.(type) {
+	case Stringer:
+		return v.String()
+	case string:
+		return v
+	default:
+		return Repr(v, NoPretty)
+	}
+}
diff -pruN 0.11+ds1-5/eval/vals/struct.go 0.12+ds1-1/eval/vals/struct.go
--- 0.11+ds1-5/eval/vals/struct.go	1970-01-01 00:00:00.000000000 +0000
+++ 0.12+ds1-1/eval/vals/struct.go	2018-07-11 02:32:25.000000000 +0000
@@ -0,0 +1,167 @@
+package vals
+
+import (
+	"bytes"
+	"encoding/json"
+	"errors"
+	"fmt"
+
+	"github.com/elves/elvish/parse"
+	"github.com/xiaq/persistent/hash"
+)
+
+var (
+	ErrIndexMustBeString = errors.New("index must be string")
+)
+
+// Struct is like a Map with fixed keys.
+type Struct struct {
+	descriptor *StructDescriptor
+	fields     []interface{}
+}
+
+var _ Indexer = (*Struct)(nil)
+
+// NewStruct creates a new *Struct value.
+func NewStruct(descriptor *StructDescriptor, fields []interface{}) *Struct {
+	return &Struct{descriptor, fields}
+}
+
+func (*Struct) Kind() string {
+	return "map"
+}
+
+// Equal returns true if the rhs is MapLike and all pairs are equal.
+func (s *Struct) Equal(rhs interface{}) bool {
+	if s == rhs {
+		return true
+	}
+	s2, ok := rhs.(*Struct)
+	if !ok {
+		return false
+	}
+	if s.descriptor != s2.descriptor {
+		return false
+	}
+	for i, field := range s.fields {
+		if !Equal(field, s2.fields[i]) {
+			return false
+		}
+	}
+	return true
+}
+
+func (s *Struct) Hash() uint32 {
+	h := hash.DJBInit
+	for _, field := range s.fields {
+		h = hash.DJBCombine(h, Hash(field))
+	}
+	return h
+}
+
+func (s *Struct) Repr(indent int) string {
+	var builder MapReprBuilder
+	builder.Indent = indent
+	for i, name := range s.descriptor.fieldNames {
+		builder.WritePair(parse.Quote(name), indent+2, Repr(s.fields[i], indent+2))
+	}
+	return builder.String()
+}
+
+func (s *Struct) Len() int {
+	return len(s.descriptor.fieldNames)
+}
+
+func (s *Struct) Index(k interface{}) (interface{}, bool) {
+	i, ok := s.index(k)
+	if !ok {
+		return nil, false
+	}
+	return s.fields[i], true
+}
+
+func (s *Struct) Assoc(k, v interface{}) (interface{}, error) {
+	i, ok := s.index(k)
+	if !ok {
+		return nil, NoSuchKey(k)
+	}
+	fields := make([]interface{}, len(s.fields))
+	copy(fields, s.fields)
+	fields[i] = v
+	return &Struct{s.descriptor, fields}, nil
+}
+
+func (s *Struct) IterateKey(f func(interface{}) bool) {
+	for _, field := range s.descriptor.fieldNames {
+		if !f(string(field)) {
+			break
+		}
+	}
+}
+
+func (s *Struct) IteratePair(f func(interface{}, interface{}) bool) {
+	for i, field := range s.descriptor.fieldNames {
+		if !f(string(field), s.fields[i]) {
+			break
+		}
+	}
+}
+
+func (s *Struct) HasKey(k interface{}) bool {
+	_, ok := s.index(k)
+	return ok
+}
+
+func (s *Struct) index(k interface{}) (int, bool) {
+	kstring, ok := k.(string)
+	if !ok {
+		return 0, false
+	}
+	index, ok := s.descriptor.fieldIndex[kstring]
+	return index, ok
+}
+
+// MarshalJSON encodes the Struct to a JSON Object.
+func (s *Struct) MarshalJSON() ([]byte, error) {
+	var buf bytes.Buffer
+	buf.WriteByte('{')
+	for i, fieldName := range s.descriptor.fieldNames {
+		if i > 0 {
+			buf.WriteByte(',')
+		}
+		buf.Write(s.descriptor.jsonFieldNames[i])
+		buf.WriteByte(':')
+		fieldJSON, err := json.Marshal(s.fields[i])
+		if err != nil {
+			return nil, fmt.Errorf("cannot encode field %q: %v", fieldName, err)
+		}
+		buf.Write(fieldJSON)
+	}
+	buf.WriteByte('}')
+	return buf.Bytes(), nil
+}
+
+// StructDescriptor contains information about the fields in a Struct.
+type StructDescriptor struct {
+	fieldNames     []string
+	jsonFieldNames [][]byte
+	fieldIndex     map[string]int
+}
+
+// NewStructDescriptor creates a new struct descriptor from a list of field
+// names.
+func NewStructDescriptor(fields ...string) *StructDescriptor {
+	fieldNames := append([]string(nil), fields...)
+	jsonFieldNames := make([][]byte, len(fields))
+	fieldIndex := make(map[string]int)
+	for i, name := range fieldNames {
+		fieldIndex[name] = i
+		jsonFieldName, err := json.Marshal(name)
+		// json.Marshal should never fail on string.
+		if err != nil {
+			panic(err)
+		}
+		jsonFieldNames[i] = jsonFieldName
+	}
+	return &StructDescriptor{fieldNames, jsonFieldNames, fieldIndex}
+}
diff -pruN 0.11+ds1-5/eval/vals/struct_test.go 0.12+ds1-1/eval/vals/struct_test.go
--- 0.11+ds1-5/eval/vals/struct_test.go	1970-01-01 00:00:00.000000000 +0000
+++ 0.12+ds1-1/eval/vals/struct_test.go	2018-07-11 02:32:25.000000000 +0000
@@ -0,0 +1,43 @@
+package vals
+
+import (
+	"testing"
+)
+
+var (
+	testStructDescriptor = NewStructDescriptor("foo", "bar")
+	testStruct           = NewStruct(testStructDescriptor, []interface{}{"lorem", "ipsum"})
+	testStruct2          = NewStruct(testStructDescriptor, []interface{}{"lorem", "dolor"})
+)
+
+func TestStructMethods(t *testing.T) {
+	if l := testStruct.Len(); l != 2 {
+		t.Errorf("testStruct.Len() = %d, want 2", l)
+	}
+	if foo, ok := testStruct.Index("foo"); foo != "lorem" {
+		t.Errorf(`testStruct.Index("foo") = %q, want "lorem"`, foo)
+	} else if !ok {
+		t.Errorf(`testStruct.Index("foo") => false, want true`)
+	}
+	if testStruct.Equal(testStruct2) {
+		t.Errorf(`testStruct.Equal(testStruct2) => true, want false`)
+	}
+	if s2, err := testStruct.Assoc("bar", "dolor"); !Equal(s2, testStruct2) {
+		t.Errorf(`testStruct.Assoc(...) => %v, want %v`, s2, testStruct2)
+	} else if err != nil {
+		t.Errorf(`testStruct.Assoc(...) => error %s, want no error`, err)
+	}
+	wantRepr := "[&foo=lorem &bar=ipsum]"
+	if gotRepr := testStruct.Repr(NoPretty); gotRepr != wantRepr {
+		t.Errorf(`testStruct.Repr() => %q, want %q`, gotRepr, wantRepr)
+	}
+	wantJSON := `{"foo":"lorem","bar":"ipsum"}`
+	gotJSONBytes, err := testStruct.MarshalJSON()
+	gotJSON := string(gotJSONBytes)
+	if err != nil {
+		t.Errorf(`testStruct.MarshalJSON() => error %v`, err)
+	}
+	if wantJSON != gotJSON {
+		t.Errorf(`testStruct.MarshalJSON() => %q, want %q`, gotJSON, wantJSON)
+	}
+}
diff -pruN 0.11+ds1-5/eval/vals/testutils_test.go 0.12+ds1-1/eval/vals/testutils_test.go
--- 0.11+ds1-5/eval/vals/testutils_test.go	1970-01-01 00:00:00.000000000 +0000
+++ 0.12+ds1-1/eval/vals/testutils_test.go	2018-07-11 02:32:25.000000000 +0000
@@ -0,0 +1,31 @@
+package vals
+
+import "github.com/elves/elvish/tt"
+
+var Args = tt.Args
+
+// anyType matches anything.
+type anyType struct{}
+
+var any = anyType{}
+
+func (anyType) Match(tt.RetValue) bool { return true }
+
+// anyErrorType matches any value that satisfies the error interface.
+type anyErrorType struct{}
+
+var anyError = anyErrorType{}
+
+func (e anyErrorType) Match(a tt.RetValue) bool {
+	_, ok := a.(error)
+	return ok
+}
+
+// equalMatcher matches the return value using Equal.
+type equalMatcher struct {
+	r interface{}
+}
+
+func eq(r interface{}) tt.Matcher { return equalMatcher{r} }
+
+func (em equalMatcher) Match(a tt.RetValue) bool { return Equal(em.r, a) }
diff -pruN 0.11+ds1-5/eval/vals/vals.go 0.12+ds1-1/eval/vals/vals.go
--- 0.11+ds1-5/eval/vals/vals.go	1970-01-01 00:00:00.000000000 +0000
+++ 0.12+ds1-1/eval/vals/vals.go	2018-07-11 02:32:25.000000000 +0000
@@ -0,0 +1,3 @@
+// Package vals contains basic facilities for manipulating values used in the
+// Elvish runtime.
+package vals
diff -pruN 0.11+ds1-5/eval/value.go 0.12+ds1-1/eval/value.go
--- 0.11+ds1-5/eval/value.go	2018-01-18 02:03:11.000000000 +0000
+++ 0.12+ds1-1/eval/value.go	2018-07-11 02:32:25.000000000 +0000
@@ -3,64 +3,47 @@ package eval
 import (
 	"fmt"
 
-	"github.com/elves/elvish/eval/types"
-	"github.com/xiaq/persistent/hashmap"
+	"github.com/elves/elvish/eval/vals"
 )
 
-func mustIndexer(v types.Value, ec *Frame) types.Indexer {
-	indexer, ok := types.GetIndexer(v)
-	if !ok {
-		throw(fmt.Errorf("a %s is not indexable", v.Kind()))
-	}
-	return indexer
-}
-
 // Callable wraps the Call method.
 type Callable interface {
 	// Call calls the receiver in a Frame with arguments and options.
-	Call(ec *Frame, args []types.Value, opts map[string]types.Value)
+	Call(fm *Frame, args []interface{}, opts map[string]interface{}) error
 }
 
 var (
 	// NoArgs is an empty argument list. It can be used as an argument to Call.
-	NoArgs = []types.Value{}
+	NoArgs = []interface{}{}
 	// NoOpts is an empty option map. It can be used as an argument to Call.
-	NoOpts = map[string]types.Value{}
+	NoOpts = map[string]interface{}{}
 )
 
-// Fn is a callable value.
-type Fn interface {
-	types.Value
-	Callable
-}
-
 // FromJSONInterface converts a interface{} that results from json.Unmarshal to
 // a Value.
-func FromJSONInterface(v interface{}) types.Value {
+func FromJSONInterface(v interface{}) interface{} {
 	if v == nil {
 		// TODO Use a more appropriate type
-		return types.String("")
+		return ""
 	}
-	switch v.(type) {
-	case bool:
-		return types.Bool(v.(bool))
-	case float64, string:
+	switch v := v.(type) {
+	case bool, string:
+		return v
+	case float64:
 		// TODO Use a numeric type for float64
-		return types.String(fmt.Sprint(v))
+		return fmt.Sprint(v)
 	case []interface{}:
-		a := v.([]interface{})
-		vs := make([]types.Value, len(a))
-		for i, v := range a {
+		vs := make([]interface{}, len(v))
+		for i, v := range v {
 			vs[i] = FromJSONInterface(v)
 		}
-		return types.MakeList(vs...)
+		return vals.MakeList(vs...)
 	case map[string]interface{}:
-		m := v.(map[string]interface{})
-		mv := hashmap.Empty
-		for k, v := range m {
-			mv = mv.Assoc(types.String(k), FromJSONInterface(v))
+		m := vals.EmptyMap
+		for key, val := range v {
+			m = m.Assoc(key, FromJSONInterface(val))
 		}
-		return types.NewMap(mv)
+		return m
 	default:
 		throw(fmt.Errorf("unexpected json type: %T", v))
 		return nil // not reached
diff -pruN 0.11+ds1-5/eval/value_test.go 0.12+ds1-1/eval/value_test.go
--- 0.11+ds1-5/eval/value_test.go	2018-01-18 02:03:11.000000000 +0000
+++ 0.12+ds1-1/eval/value_test.go	2018-07-11 02:32:25.000000000 +0000
@@ -5,36 +5,36 @@ import (
 	"reflect"
 	"testing"
 
-	"github.com/elves/elvish/eval/types"
+	"github.com/elves/elvish/eval/vals"
 	"github.com/elves/elvish/glob"
 )
 
 var reprTests = []struct {
-	v    types.Value
+	v    interface{}
 	want string
 }{
-	{types.String("233"), "233"},
-	{types.String("a\nb"), `"a\nb"`},
-	{types.String("foo bar"), "'foo bar'"},
-	{types.String("a\x00b"), `"a\x00b"`},
-	{types.Bool(true), "$true"},
-	{types.Bool(false), "$false"},
+	{"233", "233"},
+	{"a\nb", `"a\nb"`},
+	{"foo bar", "'foo bar'"},
+	{"a\x00b", `"a\x00b"`},
+	{true, "$true"},
+	{false, "$false"},
 	{&Exception{nil, nil}, "$ok"},
 	{&Exception{errors.New("foo bar"), nil}, "?(fail 'foo bar')"},
 	{&Exception{
 		PipelineError{[]*Exception{{nil, nil}, {errors.New("lorem"), nil}}}, nil},
 		"?(multi-error $ok ?(fail lorem))"},
 	{&Exception{Return, nil}, "?(return)"},
-	{types.EmptyList, "[]"},
-	{types.MakeList(types.String("bash"), types.Bool(false)), "[bash $false]"},
-	{types.MakeMap(map[types.Value]types.Value{}), "[&]"},
-	{types.MakeMap(map[types.Value]types.Value{&Exception{nil, nil}: types.String("elvish")}), "[&$ok=elvish]"},
+	{vals.EmptyList, "[]"},
+	{vals.MakeList("bash", false), "[bash $false]"},
+	{vals.MakeMap(map[interface{}]interface{}{}), "[&]"},
+	{vals.MakeMap(map[interface{}]interface{}{&Exception{nil, nil}: "elvish"}), "[&$ok=elvish]"},
 	// TODO: test maps of more elements
 }
 
 func TestRepr(t *testing.T) {
 	for _, test := range reprTests {
-		repr := test.v.Repr(types.NoPretty)
+		repr := vals.Repr(test.v, vals.NoPretty)
 		if repr != test.want {
 			t.Errorf("Repr = %s, want %s", repr, test.want)
 		}
diff -pruN 0.11+ds1-5/eval/variable_ref.go 0.12+ds1-1/eval/variable_ref.go
--- 0.11+ds1-5/eval/variable_ref.go	1970-01-01 00:00:00.000000000 +0000
+++ 0.12+ds1-1/eval/variable_ref.go	2018-07-11 02:32:25.000000000 +0000
@@ -0,0 +1,66 @@
+package eval
+
+import "strings"
+
+// ParseVariableRef parses a variable reference.
+func ParseVariableRef(text string) (explode bool, ns string, name string) {
+	return parseVariableRef(text, true)
+}
+
+// ParseIncompleteVariableRef parses an incomplete variable reference.
+func ParseIncompleteVariableRef(text string) (explode bool, ns string, name string) {
+	return parseVariableRef(text, false)
+}
+
+func parseVariableRef(text string, complete bool) (explode bool, ns string, name string) {
+	explodePart, nsPart, name := splitVariableRef(text, complete)
+	ns = nsPart
+	if len(ns) > 0 {
+		ns = ns[:len(ns)-1]
+	}
+	return explodePart != "", ns, name
+}
+
+// SplitVariableRef splits a variable reference into three parts: an optional
+// explode operator (either "" or "@"), a namespace part, and a name part.
+func SplitVariableRef(text string) (explodePart, nsPart, name string) {
+	return splitVariableRef(text, true)
+}
+
+// SplitIncompleteVariableRef splits an incomplete variable reference into three
+// parts: an optional explode operator (either "" or "@"), a namespace part, and
+// a name part.
+func SplitIncompleteVariableRef(text string) (explodePart, nsPart, name string) {
+	return splitVariableRef(text, false)
+}
+
+func splitVariableRef(text string, complete bool) (explodePart, nsPart, name string) {
+	if text == "" {
+		return "", "", ""
+	}
+	e, qname := "", text
+	if text[0] == '@' {
+		e = "@"
+		qname = text[1:]
+	}
+	if qname == "" {
+		return e, "", ""
+	}
+	i := strings.LastIndexByte(qname, ':')
+	if complete && i == len(qname)-1 {
+		i = strings.LastIndexByte(qname[:len(qname)-1], ':')
+	}
+	return e, qname[:i+1], qname[i+1:]
+}
+
+// MakeVariableRef builds a variable reference.
+func MakeVariableRef(explode bool, ns string, name string) string {
+	prefix := ""
+	if explode {
+		prefix = "@"
+	}
+	if ns != "" {
+		prefix += ns + ":"
+	}
+	return prefix + name
+}
diff -pruN 0.11+ds1-5/eval/vars/any.go 0.12+ds1-1/eval/vars/any.go
--- 0.11+ds1-5/eval/vars/any.go	1970-01-01 00:00:00.000000000 +0000
+++ 0.12+ds1-1/eval/vars/any.go	2018-07-11 02:32:25.000000000 +0000
@@ -0,0 +1,20 @@
+package vars
+
+type any struct {
+	ptr *interface{}
+}
+
+func (v any) Set(val interface{}) error {
+	*v.ptr = val
+	return nil
+}
+
+func (v any) Get() interface{} {
+	return *v.ptr
+}
+
+// NewAnyWithInit creates a variable with an initial value. The variable created
+// can be assigned values of any type.
+func NewAnyWithInit(v interface{}) Var {
+	return any{&v}
+}
diff -pruN 0.11+ds1-5/eval/vars/blackhole.go 0.12+ds1-1/eval/vars/blackhole.go
--- 0.11+ds1-5/eval/vars/blackhole.go	1970-01-01 00:00:00.000000000 +0000
+++ 0.12+ds1-1/eval/vars/blackhole.go	2018-07-11 02:32:25.000000000 +0000
@@ -0,0 +1,25 @@
+package vars
+
+type blackhole struct{}
+
+func (blackhole) Set(interface{}) error {
+	return nil
+}
+
+func (blackhole) Get() interface{} {
+	// TODO: Return a special placeholder value.
+	return ""
+}
+
+// NewBlackhole returns a blackhole variable. Assignments to a blackhole
+// variable will be discarded, and getting a blackhole variable always returns
+// an empty string.
+func NewBlackhole() Var {
+	return blackhole{}
+}
+
+// IsBlackhole returns whether the variable is a blackhole variable.
+func IsBlackhole(v Var) bool {
+	_, ok := v.(blackhole)
+	return ok
+}
diff -pruN 0.11+ds1-5/eval/vars/callback.go 0.12+ds1-1/eval/vars/callback.go
--- 0.11+ds1-5/eval/vars/callback.go	1970-01-01 00:00:00.000000000 +0000
+++ 0.12+ds1-1/eval/vars/callback.go	2018-07-11 02:32:25.000000000 +0000
@@ -0,0 +1,34 @@
+package vars
+
+type callback struct {
+	set func(interface{}) error
+	get func() interface{}
+}
+
+// FromSetGet makes a variable from a set callback and a get callback.
+func FromSetGet(set func(interface{}) error, get func() interface{}) Var {
+	return &callback{set, get}
+}
+
+func (cv *callback) Set(val interface{}) error {
+	return cv.set(val)
+}
+
+func (cv *callback) Get() interface{} {
+	return cv.get()
+}
+
+type roCallback func() interface{}
+
+// FromGet makes a variable from a get callback. The variable is read-only.
+func FromGet(get func() interface{}) Var {
+	return roCallback(get)
+}
+
+func (cv roCallback) Set(interface{}) error {
+	return errRoCannotBeSet
+}
+
+func (cv roCallback) Get() interface{} {
+	return cv()
+}
diff -pruN 0.11+ds1-5/eval/vars/del_element.go 0.12+ds1-1/eval/vars/del_element.go
--- 0.11+ds1-5/eval/vars/del_element.go	1970-01-01 00:00:00.000000000 +0000
+++ 0.12+ds1-1/eval/vars/del_element.go	2018-07-11 02:32:25.000000000 +0000
@@ -0,0 +1,39 @@
+package vars
+
+import "github.com/elves/elvish/eval/vals"
+
+// DelElement deletes an element. It uses a similar process to MakeElement,
+// except that the last level of container needs to be Dissoc-able instead of
+// Assoc-able.
+func DelElement(variable Var, indicies []interface{}) error {
+	var err error
+	// In "del a[0][1][2]",
+	//
+	// indicies:  0  1     2
+	// assocers: $a $a[0]
+	// dissocer:          $a[0][1]
+	assocers := make([]interface{}, len(indicies)-1)
+	container := variable.Get()
+	for i, index := range indicies[:len(indicies)-1] {
+		assocers[i] = container
+
+		var err error
+		container, err = vals.Index(container, index)
+		if err != nil {
+			return err
+		}
+	}
+
+	v := vals.Dissoc(container, indicies[len(indicies)-1])
+	if v == nil {
+		return elemErr{len(indicies), "value does not support element removal"}
+	}
+
+	for i := len(assocers) - 1; i >= 0; i-- {
+		v, err = vals.Assoc(assocers[i], indicies[i], v)
+		if err != nil {
+			return err
+		}
+	}
+	return variable.Set(v)
+}
diff -pruN 0.11+ds1-5/eval/vars/element.go 0.12+ds1-1/eval/vars/element.go
--- 0.11+ds1-5/eval/vars/element.go	1970-01-01 00:00:00.000000000 +0000
+++ 0.12+ds1-1/eval/vars/element.go	2018-07-11 02:32:25.000000000 +0000
@@ -0,0 +1,102 @@
+package vars
+
+import (
+	"github.com/elves/elvish/eval/vals"
+)
+
+type elem struct {
+	variable Var
+	assocers []interface{}
+	indices  []interface{}
+	setValue interface{}
+}
+
+func (ev *elem) Set(v0 interface{}) error {
+	var err error
+	v := v0
+	// Evaluate the actual new value from inside out. See comments in
+	// MakeElement for how element assignment works.
+	for i := len(ev.assocers) - 1; i >= 0; i-- {
+		v, err = vals.Assoc(ev.assocers[i], ev.indices[i], v)
+		if err != nil {
+			return err
+		}
+	}
+	err = ev.variable.Set(v)
+	// XXX(xiaq): Remember the set value for use in Get.
+	ev.setValue = v0
+	return err
+}
+
+func (ev *elem) Get() interface{} {
+	// XXX(xiaq): This is only called from fixNilVariables. We don't want to
+	// waste time accessing the variable, so we simply return the value that was
+	// set.
+	return ev.setValue
+}
+
+// NewElement returns an ephemeral variable used for assigning variable element.
+func NewElement(v Var, a []interface{}, i []interface{}) Var {
+	return &elem{v, a, i, ""}
+}
+
+// MakeElement returns a variable, that when set, simulates the mutation of an
+// element.
+func MakeElement(v Var, indicies []interface{}) (Var, error) {
+	// Assignment of indexed variables actually assignes the variable, with
+	// the right hand being a nested series of Assocs. As the simplest
+	// example, `a[0] = x` is equivalent to `a = (assoc $a 0 x)`. A more
+	// complex example is that `a[0][1][2] = x` is equivalent to
+	//	`a = (assoc $a 0 (assoc $a[0] 1 (assoc $a[0][1] 2 x)))`.
+	// Note that in each assoc form, the first two arguments can be
+	// determined now, while the last argument is only known when the
+	// right-hand-side is known. So here we evaluate the first two arguments
+	// of each assoc form and put them in two slices, assocers and indicies.
+	// In the previous example, the two slices will contain:
+	//
+	// assocers: $a $a[0] $a[0][1]
+	// indicies:  0     1        2
+	//
+	// When the right-hand side of the assignment becomes available, the new
+	// value for $a is evaluated by doing Assoc from inside out.
+	assocers := make([]interface{}, len(indicies))
+	varValue := v.Get()
+	assocers[0] = varValue
+	for i, index := range indicies[:len(indicies)-1] {
+		lastAssocer := assocers[i]
+		v, err := vals.Index(lastAssocer, index)
+		if err != nil {
+			return nil, err
+		}
+		assocers[i+1] = v
+	}
+	return NewElement(v, assocers, indicies), nil
+}
+
+type elemErr struct {
+	level int
+	msg   string
+}
+
+func (err elemErr) Error() string {
+	return err.msg
+}
+
+// HeadOfElement gets the underlying head variable of an element variable, or
+// nil if the argument is not an element variable.
+func HeadOfElement(v Var) Var {
+	if ev, ok := v.(*elem); ok {
+		return ev.variable
+	}
+	return nil
+}
+
+// ElementErrorLevel returns the level of an error returned by MakeElement or
+// DelElement. Level 0 represents that the error is about the variable itself.
+// If the argument was not returned from MakeVariable, -1 is returned.
+func ElementErrorLevel(err error) int {
+	if err, ok := err.(elemErr); ok {
+		return err.level
+	}
+	return -1
+}
diff -pruN 0.11+ds1-5/eval/vars/env.go 0.12+ds1-1/eval/vars/env.go
--- 0.11+ds1-5/eval/vars/env.go	1970-01-01 00:00:00.000000000 +0000
+++ 0.12+ds1-1/eval/vars/env.go	2018-07-11 02:32:25.000000000 +0000
@@ -0,0 +1,30 @@
+package vars
+
+import (
+	"errors"
+	"os"
+)
+
+var errEnvMustBeString = errors.New("environment variable can only be set string values")
+
+// envVariable represents an environment variable.
+type envVariable struct {
+	name string
+}
+
+func (ev envVariable) Set(val interface{}) error {
+	if s, ok := val.(string); ok {
+		os.Setenv(ev.name, string(s))
+		return nil
+	}
+	return errEnvMustBeString
+}
+
+func (ev envVariable) Get() interface{} {
+	return string(os.Getenv(ev.name))
+}
+
+// NewEnv returns an environment variable.
+func NewEnv(name string) Var {
+	return envVariable{name}
+}
diff -pruN 0.11+ds1-5/eval/vars/ptr.go 0.12+ds1-1/eval/vars/ptr.go
--- 0.11+ds1-5/eval/vars/ptr.go	1970-01-01 00:00:00.000000000 +0000
+++ 0.12+ds1-1/eval/vars/ptr.go	2018-07-11 02:32:25.000000000 +0000
@@ -0,0 +1,28 @@
+package vars
+
+import (
+	"reflect"
+
+	"github.com/elves/elvish/eval/vals"
+)
+
+type ptr struct {
+	ptr interface{}
+}
+
+// FromPtr creates a variable from a pointer. The variable is kept in sync
+// with the value the pointer points to, using elvToGo and goToElv conversions
+// when Get and Set.
+func FromPtr(p interface{}) Var {
+	return ptr{p}
+}
+
+// Get returns the value pointed by the pointer, after conversion using FromGo.
+func (v ptr) Get() interface{} {
+	return vals.FromGo(reflect.Indirect(reflect.ValueOf(v.ptr)).Interface())
+}
+
+// Get sets the value pointed by the pointer, after conversion using ScanToGo.
+func (v ptr) Set(val interface{}) error {
+	return vals.ScanToGo(val, v.ptr)
+}
diff -pruN 0.11+ds1-5/eval/vars/ptr_test.go 0.12+ds1-1/eval/vars/ptr_test.go
--- 0.11+ds1-5/eval/vars/ptr_test.go	1970-01-01 00:00:00.000000000 +0000
+++ 0.12+ds1-1/eval/vars/ptr_test.go	2018-07-11 02:32:25.000000000 +0000
@@ -0,0 +1,22 @@
+package vars
+
+import "testing"
+
+func TestPtrVariable(t *testing.T) {
+	i := 10
+	variable := FromPtr(&i)
+	if g := variable.Get(); g != "10" {
+		t.Errorf(`Getting ptrVariable returns %v, want "10"`, g)
+	}
+	err := variable.Set("20")
+	if err != nil {
+		t.Errorf(`Setting ptrVariable with "20" returns error %v`, err)
+	}
+	if i != 20 {
+		t.Errorf(`Setting ptrVariable didn't change underlying value`)
+	}
+	err = variable.Set("x")
+	if err == nil {
+		t.Errorf("Setting ptrVariable with incompatible value returns no error")
+	}
+}
diff -pruN 0.11+ds1-5/eval/vars/ro.go 0.12+ds1-1/eval/vars/ro.go
--- 0.11+ds1-5/eval/vars/ro.go	1970-01-01 00:00:00.000000000 +0000
+++ 0.12+ds1-1/eval/vars/ro.go	2018-07-11 02:32:25.000000000 +0000
@@ -0,0 +1,23 @@
+package vars
+
+import (
+	"errors"
+)
+
+var errRoCannotBeSet = errors.New("read-only variable; cannot be set")
+
+type ro struct {
+	value interface{}
+}
+
+func NewRo(v interface{}) Var {
+	return ro{v}
+}
+
+func (rv ro) Set(val interface{}) error {
+	return errRoCannotBeSet
+}
+
+func (rv ro) Get() interface{} {
+	return rv.value
+}
diff -pruN 0.11+ds1-5/eval/vars/vars.go 0.12+ds1-1/eval/vars/vars.go
--- 0.11+ds1-5/eval/vars/vars.go	1970-01-01 00:00:00.000000000 +0000
+++ 0.12+ds1-1/eval/vars/vars.go	2018-07-11 02:32:25.000000000 +0000
@@ -0,0 +1,8 @@
+// Package vars contains basic types for manipulating Elvish variables.
+package vars
+
+// Var represents an Elvish variable.
+type Var interface {
+	Set(v interface{}) error
+	Get() interface{}
+}
diff -pruN 0.11+ds1-5/eval/vars/vars_test.go 0.12+ds1-1/eval/vars/vars_test.go
--- 0.11+ds1-5/eval/vars/vars_test.go	1970-01-01 00:00:00.000000000 +0000
+++ 0.12+ds1-1/eval/vars/vars_test.go	2018-07-11 02:32:25.000000000 +0000
@@ -0,0 +1,85 @@
+package vars
+
+import (
+	"os"
+	"testing"
+)
+
+func TestAnyVariable(t *testing.T) {
+	v := NewAnyWithInit(true)
+	if v.Get() != true {
+		t.Errorf("PtrVariable.Get doesn't return initial value")
+	}
+	if v.Set("233") != nil {
+		t.Errorf("PtrVariable.Set errors")
+	}
+	if v.Get() != "233" {
+		t.Errorf("PtrVariable.Get doesn't return altered value")
+	}
+}
+
+func TestRoVariable(t *testing.T) {
+	v := NewRo("haha")
+	if v.Get() != "haha" {
+		t.Errorf("RoVariable.Get doesn't return initial value")
+	}
+	if v.Set("lala") == nil {
+		t.Errorf("RoVariable.Set doesn't error")
+	}
+}
+
+func TestCbVariable(t *testing.T) {
+	getCalled := false
+	get := func() interface{} {
+		getCalled = true
+		return "cb"
+	}
+	var setCalledWith interface{}
+	set := func(v interface{}) error {
+		setCalledWith = v
+		return nil
+	}
+
+	v := FromSetGet(set, get)
+	if v.Get() != "cb" {
+		t.Errorf("cbVariable doesn't return value from callback")
+	}
+	if !getCalled {
+		t.Errorf("cbVariable doesn't call callback")
+	}
+	v.Set("setting")
+	if setCalledWith != "setting" {
+		t.Errorf("cbVariable.Set doesn't call setter with value")
+	}
+}
+
+func TestRoCbVariable(t *testing.T) {
+	getCalled := false
+	get := func() interface{} {
+		getCalled = true
+		return "cb"
+	}
+	v := FromGet(get)
+	if v.Get() != "cb" {
+		t.Errorf("roCbVariable doesn't return value from callback")
+	}
+	if !getCalled {
+		t.Errorf("roCbVariable doesn't call callback")
+	}
+	if v.Set("lala") == nil {
+		t.Errorf("roCbVariable.Set doesn't error")
+	}
+}
+
+func TestEnvVariable(t *testing.T) {
+	name := "elvish_test"
+	v := envVariable{name}
+	os.Setenv(name, "foo")
+	if v.Get() != "foo" {
+		t.Errorf("envVariable.Get doesn't return env value")
+	}
+	v.Set("bar")
+	if os.Getenv(name) != "bar" {
+		t.Errorf("envVariable.Set doesn't alter env value")
+	}
+}
diff -pruN 0.11+ds1-5/eval/vartypes/blackhole.go 0.12+ds1-1/eval/vartypes/blackhole.go
--- 0.11+ds1-5/eval/vartypes/blackhole.go	2018-01-18 02:03:11.000000000 +0000
+++ 0.12+ds1-1/eval/vartypes/blackhole.go	1970-01-01 00:00:00.000000000 +0000
@@ -1,27 +0,0 @@
-package vartypes
-
-import "github.com/elves/elvish/eval/types"
-
-type blackhole struct{}
-
-func (blackhole) Set(types.Value) error {
-	return nil
-}
-
-func (blackhole) Get() types.Value {
-	// TODO: Return a special placeholder value.
-	return types.String("")
-}
-
-// NewBlackhole returns a blackhole variable. Assignments to a blackhole
-// variable will be discarded, and getting a blackhole variable always returns
-// an empty string.
-func NewBlackhole() Variable {
-	return blackhole{}
-}
-
-// IsBlackhole returns whether the variable is a blackhole variable.
-func IsBlackhole(v Variable) bool {
-	_, ok := v.(blackhole)
-	return ok
-}
diff -pruN 0.11+ds1-5/eval/vartypes/bool.go 0.12+ds1-1/eval/vartypes/bool.go
--- 0.11+ds1-5/eval/vartypes/bool.go	2018-01-18 02:03:11.000000000 +0000
+++ 0.12+ds1-1/eval/vartypes/bool.go	1970-01-01 00:00:00.000000000 +0000
@@ -1,29 +0,0 @@
-package vartypes
-
-import (
-	"errors"
-
-	"github.com/elves/elvish/eval/types"
-)
-
-var errMustBeBool = errors.New("must be bool")
-
-type boolVar struct {
-	ptr *bool
-}
-
-func NewBool(ptr *bool) Variable {
-	return boolVar{ptr}
-}
-
-func (bv boolVar) Get() types.Value {
-	return types.Bool(*bv.ptr)
-}
-
-func (bv boolVar) Set(v types.Value) error {
-	if b, ok := v.(types.Bool); ok {
-		*bv.ptr = bool(b)
-		return nil
-	}
-	return errMustBeBool
-}
diff -pruN 0.11+ds1-5/eval/vartypes/callback.go 0.12+ds1-1/eval/vartypes/callback.go
--- 0.11+ds1-5/eval/vartypes/callback.go	2018-01-18 02:03:11.000000000 +0000
+++ 0.12+ds1-1/eval/vartypes/callback.go	1970-01-01 00:00:00.000000000 +0000
@@ -1,36 +0,0 @@
-package vartypes
-
-import "github.com/elves/elvish/eval/types"
-
-type callback struct {
-	set func(types.Value) error
-	get func() types.Value
-}
-
-// NewCallback makes a variable from a set callback and a get callback.
-func NewCallback(set func(types.Value) error, get func() types.Value) Variable {
-	return &callback{set, get}
-}
-
-func (cv *callback) Set(val types.Value) error {
-	return cv.set(val)
-}
-
-func (cv *callback) Get() types.Value {
-	return cv.get()
-}
-
-type roCallback func() types.Value
-
-// NewRoCallback makes a read-only variable from a get callback.
-func NewRoCallback(get func() types.Value) Variable {
-	return roCallback(get)
-}
-
-func (cv roCallback) Set(types.Value) error {
-	return errRoCannotBeSet
-}
-
-func (cv roCallback) Get() types.Value {
-	return cv()
-}
diff -pruN 0.11+ds1-5/eval/vartypes/del_element.go 0.12+ds1-1/eval/vartypes/del_element.go
--- 0.11+ds1-5/eval/vartypes/del_element.go	2018-01-18 02:03:11.000000000 +0000
+++ 0.12+ds1-1/eval/vartypes/del_element.go	1970-01-01 00:00:00.000000000 +0000
@@ -1,38 +0,0 @@
-package vartypes
-
-import "github.com/elves/elvish/eval/types"
-
-// DelElement deletes an element. It uses a similar process to MakeElement,
-// except that the last level of container needs to be a Dissocer instead of an
-// Assocer.
-func DelElement(variable Variable, indicies []types.Value) error {
-	// In "del a[0][1][2]",
-	//
-	// indicies:  0  1     2
-	// assocers: $a $a[0]
-	// dissocer:          $a[0][1]
-	assocers := make([]types.Assocer, len(indicies)-1)
-	container := variable.Get()
-	for i, index := range indicies[:len(indicies)-1] {
-		indexer, ok := container.(types.IndexOneer)
-		if !ok {
-			return elemErr{i, "value does not support indexing"}
-		}
-		assocer, ok := container.(types.Assocer)
-		if !ok {
-			return elemErr{i, "value does not support indexing for setting"}
-		}
-		assocers[i] = assocer
-		container = indexer.IndexOne(index)
-	}
-	dissocer, ok := container.(types.Dissocer)
-	if !ok {
-		return elemErr{len(indicies), "value does not support element removal"}
-	}
-
-	v := dissocer.Dissoc(indicies[len(indicies)-1])
-	for i := len(assocers) - 1; i >= 0; i-- {
-		v = assocers[i].Assoc(indicies[i], v)
-	}
-	return variable.Set(v)
-}
diff -pruN 0.11+ds1-5/eval/vartypes/element.go 0.12+ds1-1/eval/vartypes/element.go
--- 0.11+ds1-5/eval/vartypes/element.go	2018-01-18 02:03:11.000000000 +0000
+++ 0.12+ds1-1/eval/vartypes/element.go	1970-01-01 00:00:00.000000000 +0000
@@ -1,112 +0,0 @@
-package vartypes
-
-import (
-	"github.com/elves/elvish/eval/types"
-)
-
-type elem struct {
-	variable Variable
-	assocers []types.Assocer
-	indices  []types.Value
-	setValue types.Value
-}
-
-func (ev *elem) Set(v0 types.Value) error {
-	v := v0
-	// Evaluate the actual new value from inside out. See comments in
-	// MakeElement for how element assignment works.
-	for i := len(ev.assocers) - 1; i >= 0; i-- {
-		v = ev.assocers[i].Assoc(ev.indices[i], v)
-	}
-	err := ev.variable.Set(v)
-	// XXX(xiaq): Remember the set value for use in Get.
-	ev.setValue = v0
-	return err
-}
-
-func (ev *elem) Get() types.Value {
-	// XXX(xiaq): This is only called from fixNilVariables. We don't want to
-	// waste time accessing the variable, so we simply return the value that was
-	// set.
-	return ev.setValue
-}
-
-// NewElement returns an ephemeral variable used for assigning variable element.
-func NewElement(v Variable, a []types.Assocer, i []types.Value) Variable {
-	return &elem{v, a, i, types.String("")}
-}
-
-// MakeElement returns a variable, that when set, simulates the mutation of an
-// element.
-func MakeElement(v Variable, indicies []types.Value) (Variable, error) {
-	// Assignment of indexed variables actually assignes the variable, with
-	// the right hand being a nested series of Assocs. As the simplest
-	// example, `a[0] = x` is equivalent to `a = (assoc $a 0 x)`. A more
-	// complex example is that `a[0][1][2] = x` is equivalent to
-	//	`a = (assoc $a 0 (assoc $a[0] 1 (assoc $a[0][1] 2 x)))`.
-	// Note that in each assoc form, the first two arguments can be
-	// determined now, while the last argument is only known when the
-	// right-hand-side is known. So here we evaluate the first two arguments
-	// of each assoc form and put them in two slices, assocers and indicies.
-	// In the previous example, the two slices will contain:
-	//
-	// assocers: $a $a[0] $a[0][1]
-	// indicies:  0     1        2
-	//
-	// When the right-hand side of the assignment becomes available, the new
-	// value for $a is evaluated by doing Assoc from inside out.
-	assocers := make([]types.Assocer, len(indicies))
-	varValue, ok := v.Get().(indexOneAssocer)
-	if !ok {
-		return nil, elemErr{0, "cannot be indexed for setting"}
-	}
-	assocers[0] = varValue
-	for i, index := range indicies[:len(indicies)-1] {
-		lastAssocer, ok := assocers[i].(types.IndexOneer)
-		if !ok {
-			// This cannot occur when i==0, since varValue as already
-			// asserted to be an IndexOnner.
-			return nil, elemErr{i, "cannot be indexed"}
-		}
-		assocer, ok := lastAssocer.IndexOne(index).(types.Assocer)
-		if !ok {
-			return nil, elemErr{i + 1, "cannot be indexed for setting"}
-		}
-		assocers[i+1] = assocer
-	}
-	return NewElement(v, assocers, indicies), nil
-}
-
-// indexOneAssocer combines IndexOneer and Assocer.
-type indexOneAssocer interface {
-	types.IndexOneer
-	types.Assocer
-}
-
-type elemErr struct {
-	level int
-	msg   string
-}
-
-func (err elemErr) Error() string {
-	return err.msg
-}
-
-// GetHeadOfElement gets the underlying head variable of an element variable, or
-// nil if the argument is not an element variable.
-func GetHeadOfElement(v Variable) Variable {
-	if ev, ok := v.(*elem); ok {
-		return ev.variable
-	}
-	return nil
-}
-
-// GetElementErrorLevel returns the level of an error returned by MakeElement or
-// DelElement. Level 0 represents that the error is about the variable itself.
-// If the argument was not returned from MakeVariable, -1 is returned.
-func GetElementErrorLevel(err error) int {
-	if err, ok := err.(elemErr); ok {
-		return err.level
-	}
-	return -1
-}
diff -pruN 0.11+ds1-5/eval/vartypes/env.go 0.12+ds1-1/eval/vartypes/env.go
--- 0.11+ds1-5/eval/vartypes/env.go	2018-01-18 02:03:11.000000000 +0000
+++ 0.12+ds1-1/eval/vartypes/env.go	1970-01-01 00:00:00.000000000 +0000
@@ -1,32 +0,0 @@
-package vartypes
-
-import (
-	"errors"
-	"os"
-
-	"github.com/elves/elvish/eval/types"
-)
-
-var errEnvMustBeString = errors.New("environment variable can only be set string values")
-
-// envVariable represents an environment variable.
-type envVariable struct {
-	name string
-}
-
-func (ev envVariable) Set(val types.Value) error {
-	if s, ok := val.(types.String); ok {
-		os.Setenv(ev.name, string(s))
-		return nil
-	}
-	return errEnvMustBeString
-}
-
-func (ev envVariable) Get() types.Value {
-	return types.String(os.Getenv(ev.name))
-}
-
-// NewEnv returns an environment variable.
-func NewEnv(name string) Variable {
-	return envVariable{name}
-}
diff -pruN 0.11+ds1-5/eval/vartypes/number.go 0.12+ds1-1/eval/vartypes/number.go
--- 0.11+ds1-5/eval/vartypes/number.go	2018-01-18 02:03:11.000000000 +0000
+++ 0.12+ds1-1/eval/vartypes/number.go	1970-01-01 00:00:00.000000000 +0000
@@ -1,32 +0,0 @@
-package vartypes
-
-import (
-	"errors"
-	"strconv"
-
-	"github.com/elves/elvish/eval/types"
-)
-
-var errMustBeNumber = errors.New("must be number")
-
-type number struct {
-	ptr *float64
-}
-
-func NewNumber(ptr *float64) Variable {
-	return number{ptr}
-}
-
-func (nv number) Get() types.Value {
-	return types.String(strconv.FormatFloat(*nv.ptr, 'E', -1, 64))
-}
-
-func (nv number) Set(v types.Value) error {
-	if s, ok := v.(types.String); ok {
-		if num, err := strconv.ParseFloat(string(s), 64); err == nil {
-			*nv.ptr = num
-			return nil
-		}
-	}
-	return errMustBeNumber
-}
diff -pruN 0.11+ds1-5/eval/vartypes/ptr.go 0.12+ds1-1/eval/vartypes/ptr.go
--- 0.11+ds1-5/eval/vartypes/ptr.go	2018-01-18 02:03:11.000000000 +0000
+++ 0.12+ds1-1/eval/vartypes/ptr.go	1970-01-01 00:00:00.000000000 +0000
@@ -1,49 +0,0 @@
-package vartypes
-
-import "github.com/elves/elvish/eval/types"
-
-type ptr struct {
-	valuePtr *types.Value
-}
-
-func (pv ptr) Set(val types.Value) error {
-	*pv.valuePtr = val
-	return nil
-}
-
-func (pv ptr) Get() types.Value {
-	return *pv.valuePtr
-}
-
-func NewPtr(v types.Value) Variable {
-	return ptr{&v}
-}
-
-type validatedPtr struct {
-	valuePtr  *types.Value
-	validator func(types.Value) error
-}
-
-type invalidValueError struct {
-	inner error
-}
-
-func (err invalidValueError) Error() string {
-	return "invalid value: " + err.inner.Error()
-}
-
-func NewValidatedPtr(v types.Value, vld func(types.Value) error) Variable {
-	return validatedPtr{&v, vld}
-}
-
-func (iv validatedPtr) Set(val types.Value) error {
-	if err := iv.validator(val); err != nil {
-		return invalidValueError{err}
-	}
-	*iv.valuePtr = val
-	return nil
-}
-
-func (iv validatedPtr) Get() types.Value {
-	return *iv.valuePtr
-}
diff -pruN 0.11+ds1-5/eval/vartypes/ro.go 0.12+ds1-1/eval/vartypes/ro.go
--- 0.11+ds1-5/eval/vartypes/ro.go	2018-01-18 02:03:11.000000000 +0000
+++ 0.12+ds1-1/eval/vartypes/ro.go	1970-01-01 00:00:00.000000000 +0000
@@ -1,25 +0,0 @@
-package vartypes
-
-import (
-	"errors"
-
-	"github.com/elves/elvish/eval/types"
-)
-
-var errRoCannotBeSet = errors.New("read-only variable; cannot be set")
-
-type ro struct {
-	value types.Value
-}
-
-func NewRo(v types.Value) Variable {
-	return ro{v}
-}
-
-func (rv ro) Set(val types.Value) error {
-	return errRoCannotBeSet
-}
-
-func (rv ro) Get() types.Value {
-	return rv.value
-}
diff -pruN 0.11+ds1-5/eval/vartypes/string.go 0.12+ds1-1/eval/vartypes/string.go
--- 0.11+ds1-5/eval/vartypes/string.go	2018-01-18 02:03:11.000000000 +0000
+++ 0.12+ds1-1/eval/vartypes/string.go	1970-01-01 00:00:00.000000000 +0000
@@ -1,31 +0,0 @@
-package vartypes
-
-import (
-	"errors"
-
-	"github.com/elves/elvish/eval/types"
-)
-
-var errMustBeString = errors.New("must be string")
-
-type stringVar struct {
-	ptr *string
-}
-
-// NewString creates a variable from a string pointer. The Variable can only be
-// set to a String value, and modifications are reflected in the passed string.
-func NewString(ps *string) Variable {
-	return stringVar{ps}
-}
-
-func (sv stringVar) Get() types.Value {
-	return types.String(*sv.ptr)
-}
-
-func (sv stringVar) Set(v types.Value) error {
-	if s, ok := v.(types.String); ok {
-		*sv.ptr = string(s)
-		return nil
-	}
-	return errMustBeString
-}
diff -pruN 0.11+ds1-5/eval/vartypes/validator.go 0.12+ds1-1/eval/vartypes/validator.go
--- 0.11+ds1-5/eval/vartypes/validator.go	2018-01-18 02:03:11.000000000 +0000
+++ 0.12+ds1-1/eval/vartypes/validator.go	1970-01-01 00:00:00.000000000 +0000
@@ -1,44 +0,0 @@
-package vartypes
-
-import (
-	"errors"
-	"strconv"
-
-	"github.com/elves/elvish/eval/types"
-)
-
-var (
-	errShouldBeList   = errors.New("should be list")
-	errShouldBeMap    = errors.New("should be map")
-	errShouldBeBool   = errors.New("should be bool")
-	errShouldBeNumber = errors.New("should be number")
-)
-
-func ShouldBeList(v types.Value) error {
-	if _, ok := v.(types.List); !ok {
-		return errShouldBeList
-	}
-	return nil
-}
-
-func ShouldBeMap(v types.Value) error {
-	if _, ok := v.(types.Map); !ok {
-		return errShouldBeMap
-	}
-	return nil
-}
-
-func ShouldBeBool(v types.Value) error {
-	if _, ok := v.(types.Bool); !ok {
-		return errShouldBeBool
-	}
-	return nil
-}
-
-func ShouldBeNumber(v types.Value) error {
-	if _, ok := v.(types.String); !ok {
-		return errShouldBeNumber
-	}
-	_, err := strconv.ParseFloat(string(v.(types.String)), 64)
-	return err
-}
diff -pruN 0.11+ds1-5/eval/vartypes/variable_test.go 0.12+ds1-1/eval/vartypes/variable_test.go
--- 0.11+ds1-5/eval/vartypes/variable_test.go	2018-01-18 02:03:11.000000000 +0000
+++ 0.12+ds1-1/eval/vartypes/variable_test.go	1970-01-01 00:00:00.000000000 +0000
@@ -1,94 +0,0 @@
-package vartypes
-
-import (
-	"os"
-	"testing"
-
-	"github.com/elves/elvish/eval/types"
-)
-
-func TestPtrVariable(t *testing.T) {
-	v := NewPtr(types.Bool(true))
-	if v.Get() != types.Bool(true) {
-		t.Errorf("PtrVariable.Get doesn't return initial value")
-	}
-	if v.Set(types.String("233")) != nil {
-		t.Errorf("PtrVariable.Set errors")
-	}
-	if v.Get() != types.String("233") {
-		t.Errorf("PtrVariable.Get doesn't return altered value")
-	}
-}
-
-func TestValidatedPtrVariable(t *testing.T) {
-	v := NewValidatedPtr(types.Bool(true), ShouldBeBool)
-	if v.Set(types.String("233")) == nil {
-		t.Errorf("ValidatedPtrVariable.Set doesn't error when setting incompatible value")
-	}
-}
-
-func TestRoVariable(t *testing.T) {
-	v := NewRo(types.String("haha"))
-	if v.Get() != types.String("haha") {
-		t.Errorf("RoVariable.Get doesn't return initial value")
-	}
-	if v.Set(types.String("lala")) == nil {
-		t.Errorf("RoVariable.Set doesn't error")
-	}
-}
-
-func TestCbVariable(t *testing.T) {
-	getCalled := false
-	get := func() types.Value {
-		getCalled = true
-		return types.String("cb")
-	}
-	var setCalledWith types.Value
-	set := func(v types.Value) error {
-		setCalledWith = v
-		return nil
-	}
-
-	v := NewCallback(set, get)
-	if v.Get() != types.String("cb") {
-		t.Errorf("cbVariable doesn't return value from callback")
-	}
-	if !getCalled {
-		t.Errorf("cbVariable doesn't call callback")
-	}
-	v.Set(types.String("setting"))
-	if setCalledWith != types.String("setting") {
-		t.Errorf("cbVariable.Set doesn't call setter with value")
-	}
-}
-
-func TestRoCbVariable(t *testing.T) {
-	getCalled := false
-	get := func() types.Value {
-		getCalled = true
-		return types.String("cb")
-	}
-	v := NewRoCallback(get)
-	if v.Get() != types.String("cb") {
-		t.Errorf("roCbVariable doesn't return value from callback")
-	}
-	if !getCalled {
-		t.Errorf("roCbVariable doesn't call callback")
-	}
-	if v.Set(types.String("lala")) == nil {
-		t.Errorf("roCbVariable.Set doesn't error")
-	}
-}
-
-func TestEnvVariable(t *testing.T) {
-	name := "elvish_test"
-	v := envVariable{name}
-	os.Setenv(name, "foo")
-	if v.Get() != types.String("foo") {
-		t.Errorf("envVariable.Get doesn't return env value")
-	}
-	v.Set(types.String("bar"))
-	if os.Getenv(name) != "bar" {
-		t.Errorf("envVariable.Set doesn't alter env value")
-	}
-}
diff -pruN 0.11+ds1-5/eval/vartypes/vartypes.go 0.12+ds1-1/eval/vartypes/vartypes.go
--- 0.11+ds1-5/eval/vartypes/vartypes.go	2018-01-18 02:03:11.000000000 +0000
+++ 0.12+ds1-1/eval/vartypes/vartypes.go	1970-01-01 00:00:00.000000000 +0000
@@ -1,12 +0,0 @@
-// Package vartypes contains basic types for manipulating Elvish variables.
-package vartypes
-
-import (
-	"github.com/elves/elvish/eval/types"
-)
-
-// Variable represents an Elvish variable.
-type Variable interface {
-	Set(v types.Value) error
-	Get() types.Value
-}
diff -pruN 0.11+ds1-5/Gopkg.lock 0.12+ds1-1/Gopkg.lock
--- 0.11+ds1-5/Gopkg.lock	2018-01-18 02:03:11.000000000 +0000
+++ 0.12+ds1-1/Gopkg.lock	2018-07-11 02:32:25.000000000 +0000
@@ -10,8 +10,8 @@
 [[projects]]
   name = "github.com/kr/pty"
   packages = ["."]
-  revision = "95d05c1eef33a45bd58676b6ce28d105839b8d0b"
-  version = "v1.0.1"
+  revision = "282ce0e5322c82529687d609ee670fac7c7d917c"
+  version = "v1.1.1"
 
 [[projects]]
   name = "github.com/mattn/go-isatty"
@@ -22,14 +22,14 @@
 [[projects]]
   branch = "master"
   name = "github.com/xiaq/persistent"
-  packages = ["hash","hashmap","types","vector"]
-  revision = "221501f424668419263ca5ee4e26637fdb05e229"
+  packages = ["hash","hashmap","vector"]
+  revision = "cd415c64068256386eb46bbbb1e56b461872feed"
 
 [[projects]]
   branch = "master"
   name = "golang.org/x/sys"
   packages = ["unix","windows"]
-  revision = "a3f2cbd54cf5dfe3fbaccf76375fdb12f67654c8"
+  revision = "ff2a66f350cefa5c93a634eadb5d25bb60c85a9c"
 
 [solve-meta]
   analyzer-name = "dep"
diff -pruN 0.11+ds1-5/Makefile 0.12+ds1-1/Makefile
--- 0.11+ds1-5/Makefile	2018-01-18 02:03:11.000000000 +0000
+++ 0.12+ds1-1/Makefile	2018-07-11 02:32:25.000000000 +0000
@@ -1,15 +1,20 @@
 PKG_BASE := github.com/elves/elvish
-PKGS := $(shell go list ./... | sed 's|^$(PKG_BASE)|.|' | grep -v '^./vendor')
-PKG_COVERS := $(shell go list ./... | sed 's|^$(PKG_BASE)|.|' | grep -v '^\./vendor' | grep -v '^\.$$' | sed 's/^\./cover/' | sed 's/$$/.cover/')
+PKGS := $(shell go list ./... | sed 's|^$(PKG_BASE)|.|' | grep -v '^./\(vendor\|website\)')
+PKG_COVERS := $(shell go list ./... | sed 's|^$(PKG_BASE)|.|' | grep -v '^\./\(vendor\|website\)' | grep -v '^\.$$' | sed 's/^\./cover/' | sed 's/$$/.cover/')
 COVER_MODE := set
-VERSION := $(shell git describe --tags --always)
+VERSION := $(shell git describe --tags --always --dirty=-dirty)
 
 GOVERALLS := github.com/mattn/goveralls
 
 default: test get
 
 get:
-	go get -ldflags "-X github.com/elves/elvish/build.Version=$(VERSION) -X github.com/elves/elvish/build.Builder=$(shell id -un)@$(shell hostname)" .
+	go get -ldflags "-X github.com/elves/elvish/buildinfo.Version=$(VERSION) \
+		-X github.com/elves/elvish/buildinfo.GoRoot=$(shell go env GOROOT) \
+		-X github.com/elves/elvish/buildinfo.GoPath=$(shell go env GOPATH)" .
+
+buildall:
+	./buildall.sh
 
 generate:
 	go generate ./...
@@ -33,14 +38,14 @@ cover/all: $(PKG_COVERS)
 # code coverage.
 upload-codecov-travis: cover/all
 	test "$(TRAVIS_PULL_REQUEST)" = false \
-		&& echo "$(TRAVIS_GO_VERSION)" | grep -q '^1.9' \
+		&& echo "$(TRAVIS_GO_VERSION)" | grep -q '^1.10' \
 		&& curl -s https://codecov.io/bash -o codecov.bash \
 		&& bash codecov.bash -f $< \
 		|| echo "not sending to codecov.io"
 
 upload-coveralls-travis: cover/all
 	test "$(TRAVIS_PULL_REQUEST)" = false \
-		&& echo "$(TRAVIS_GO_VERSION)" | grep -q '^1.9' \
+		&& echo "$(TRAVIS_GO_VERSION)" | grep -q '^1.10' \
 		&& go get -d $(GOVERALLS) \
 		&& go build -o goveralls $(GOVERALLS) \
 		&& ./goveralls -coverprofile $< -service=travis-ci \
@@ -56,16 +61,7 @@ upload-coveralls-appveyor: cover/all
 		&& goveralls -coverprofile $< -service=appveyor-ci \
 		|| echo "not sending to coveralls.io"
 
-upload-bin:
-	test "$(TRAVIS_OS_NAME)" = linux \
-		&& echo "$(TRAVIS_GO_VERSION)" | grep -q '^1.9' \
-		&& test "$(TRAVIS_PULL_REQUEST)" = false \
-		&& test -n "$(TRAVIS_TAG)" -o "$(TRAVIS_BRANCH)" = master \
-		&& go build -o ./elvish \
-		&& ./elvish build-and-upload.elv \
-		|| echo "not build-and-uploading"
-
-travis: testmain upload-codecov-travis upload-coveralls-travis upload-bin
+travis: testmain upload-codecov-travis upload-coveralls-travis
 appveyor: testmain upload-codecov-appveyor
 
-.PHONY: default get generate test testmain upload-codecov-travis upload-coveralls-travis upload-codecov-appveyor upload-coveralls-appveyor upload-bin travis
+.PHONY: default get buildall generate test testmain upload-codecov-travis upload-coveralls-travis upload-codecov-appveyor upload-coveralls-appveyor travis
diff -pruN 0.11+ds1-5/parse/check_ast.go 0.12+ds1-1/parse/check_ast.go
--- 0.11+ds1-5/parse/check_ast.go	2018-01-18 02:03:11.000000000 +0000
+++ 0.12+ds1-1/parse/check_ast.go	1970-01-01 00:00:00.000000000 +0000
@@ -1,154 +0,0 @@
-package parse
-
-import (
-	"fmt"
-	"reflect"
-	"strings"
-	"unicode"
-	"unicode/utf8"
-)
-
-// AST checking utilities. Used in test cases.
-
-// ast is an AST specification. The name part identifies the type of the Node;
-// for instance, "Chunk" specifies a Chunk. The fields part is specifies children
-// to check; see document of fs.
-//
-// When a Node contains exactly one child, It can be coalesced with its child
-// by adding "/ChildName" in the name part. For instance, "Chunk/Pipeline"
-// specifies a Chunk that contains exactly one Pipeline. In this case, the
-// fields part specified the children of the Pipeline instead of the Chunk
-// (which has no additional interesting fields anyway). Multi-level coalescence
-// like "Chunk/Pipeline/Form" is also allowed.
-//
-// The dynamic type of the Node being checked is assumed to be a pointer to a
-// struct that embeds the "node" struct.
-type ast struct {
-	name   string
-	fields fs
-}
-
-// fs specifies fields of a Node to check. For the value of field $f in the
-// Node ("found value"), fs[$f] ("wanted value") is used to check against it.
-//
-// If the key is "text", the SourceText of the Node is checked. It doesn't
-// involve a found value.
-//
-// If the wanted value is nil, the found value is checked against nil.
-//
-// If the found value implements Node, then the wanted value must be either an
-// ast, where the checking algorithm of ast applies, or a string, where the
-// source text of the found value is checked.
-//
-// If the found value is a slice whose elements implement Node, then the wanted
-// value must be a slice where checking is then done recursively.
-//
-// If the found value satisfied none of the above conditions, it is checked
-// against the wanted value using reflect.DeepEqual.
-type fs map[string]interface{}
-
-// checkAST checks an AST against a specification.
-func checkAST(n Node, want ast) error {
-	wantnames := strings.Split(want.name, "/")
-	// Check coalesced levels
-	for i, wantname := range wantnames {
-		name := reflect.TypeOf(n).Elem().Name()
-		if wantname != name {
-			return fmt.Errorf("want %s, got %s (%s)", wantname, name, summary(n))
-		}
-		if i == len(wantnames)-1 {
-			break
-		}
-		fields := n.Children()
-		if len(fields) != 1 {
-			return fmt.Errorf("want exactly 1 child, got %d (%s)", len(fields), summary(n))
-		}
-		n = fields[0]
-	}
-
-	ntype := reflect.TypeOf(n).Elem()
-	nvalue := reflect.ValueOf(n).Elem()
-
-	for i := 0; i < ntype.NumField(); i++ {
-		fieldname := ntype.Field(i).Name
-		if !exported(fieldname) {
-			// Unexported field
-			continue
-		}
-		got := nvalue.Field(i).Interface()
-		want, ok := want.fields[fieldname]
-		if ok {
-			err := checkField(got, want, "field "+fieldname+" of: "+summary(n))
-			if err != nil {
-				return err
-			}
-		} else {
-			// Not specified. Check if got is a zero value of its type.
-			if !reflect.DeepEqual(got, reflect.Zero(reflect.TypeOf(got)).Interface()) {
-				return fmt.Errorf("want zero, got %v (field %s of: %s)", got, fieldname, summary(n))
-			}
-		}
-	}
-
-	return nil
-}
-
-var nodeType = reflect.TypeOf((*Node)(nil)).Elem()
-
-// checkField checks a field against a field specification.
-func checkField(got interface{}, want interface{}, ctx string) error {
-	// Want nil.
-	if want == nil {
-		if !reflect.ValueOf(got).IsNil() {
-			return fmt.Errorf("want nil, got %v (%s)", got, ctx)
-		}
-		return nil
-	}
-
-	if got, ok := got.(Node); ok {
-		// Got a Node.
-		return checkNodeInField(got.(Node), want)
-	}
-	tgot := reflect.TypeOf(got)
-	if tgot.Kind() == reflect.Slice && tgot.Elem().Implements(nodeType) {
-		// Got a slice of Nodes.
-		vgot := reflect.ValueOf(got)
-		vwant := reflect.ValueOf(want)
-		if vgot.Len() != vwant.Len() {
-			return fmt.Errorf("want %d, got %d (%s)", vwant.Len(), vgot.Len(), ctx)
-		}
-		for i := 0; i < vgot.Len(); i++ {
-			err := checkNodeInField(vgot.Index(i).Interface().(Node),
-				vwant.Index(i).Interface())
-			if err != nil {
-				return err
-			}
-		}
-		return nil
-	}
-
-	if !reflect.DeepEqual(want, got) {
-		return fmt.Errorf("want %v, got %v (%s)", want, got, ctx)
-	}
-	return nil
-}
-
-func checkNodeInField(got Node, want interface{}) error {
-	switch want := want.(type) {
-	case string:
-		text := got.SourceText()
-		if want != text {
-			return fmt.Errorf("want %q, got %q (%s)", want, text, summary(got))
-		}
-		return nil
-	case ast:
-		return checkAST(got, want)
-	default:
-		panic(fmt.Sprintf("bad want type %T (%s)", want, summary(got)))
-	}
-}
-
-func exported(name string) bool {
-	r, _ := utf8.DecodeRuneInString(name)
-	return unicode.IsUpper(r)
-}
diff -pruN 0.11+ds1-5/parse/check_ast_test.go 0.12+ds1-1/parse/check_ast_test.go
--- 0.11+ds1-5/parse/check_ast_test.go	1970-01-01 00:00:00.000000000 +0000
+++ 0.12+ds1-1/parse/check_ast_test.go	2018-07-11 02:32:25.000000000 +0000
@@ -0,0 +1,154 @@
+package parse
+
+import (
+	"fmt"
+	"reflect"
+	"strings"
+	"unicode"
+	"unicode/utf8"
+)
+
+// AST checking utilities. Used in test cases.
+
+// ast is an AST specification. The name part identifies the type of the Node;
+// for instance, "Chunk" specifies a Chunk. The fields part is specifies children
+// to check; see document of fs.
+//
+// When a Node contains exactly one child, It can be coalesced with its child
+// by adding "/ChildName" in the name part. For instance, "Chunk/Pipeline"
+// specifies a Chunk that contains exactly one Pipeline. In this case, the
+// fields part specified the children of the Pipeline instead of the Chunk
+// (which has no additional interesting fields anyway). Multi-level coalescence
+// like "Chunk/Pipeline/Form" is also allowed.
+//
+// The dynamic type of the Node being checked is assumed to be a pointer to a
+// struct that embeds the "node" struct.
+type ast struct {
+	name   string
+	fields fs
+}
+
+// fs specifies fields of a Node to check. For the value of field $f in the
+// Node ("found value"), fs[$f] ("wanted value") is used to check against it.
+//
+// If the key is "text", the SourceText of the Node is checked. It doesn't
+// involve a found value.
+//
+// If the wanted value is nil, the found value is checked against nil.
+//
+// If the found value implements Node, then the wanted value must be either an
+// ast, where the checking algorithm of ast applies, or a string, where the
+// source text of the found value is checked.
+//
+// If the found value is a slice whose elements implement Node, then the wanted
+// value must be a slice where checking is then done recursively.
+//
+// If the found value satisfied none of the above conditions, it is checked
+// against the wanted value using reflect.DeepEqual.
+type fs map[string]interface{}
+
+// checkAST checks an AST against a specification.
+func checkAST(n Node, want ast) error {
+	wantnames := strings.Split(want.name, "/")
+	// Check coalesced levels
+	for i, wantname := range wantnames {
+		name := reflect.TypeOf(n).Elem().Name()
+		if wantname != name {
+			return fmt.Errorf("want %s, got %s (%s)", wantname, name, summary(n))
+		}
+		if i == len(wantnames)-1 {
+			break
+		}
+		fields := n.Children()
+		if len(fields) != 1 {
+			return fmt.Errorf("want exactly 1 child, got %d (%s)", len(fields), summary(n))
+		}
+		n = fields[0]
+	}
+
+	ntype := reflect.TypeOf(n).Elem()
+	nvalue := reflect.ValueOf(n).Elem()
+
+	for i := 0; i < ntype.NumField(); i++ {
+		fieldname := ntype.Field(i).Name
+		if !exported(fieldname) {
+			// Unexported field
+			continue
+		}
+		got := nvalue.Field(i).Interface()
+		want, ok := want.fields[fieldname]
+		if ok {
+			err := checkField(got, want, "field "+fieldname+" of: "+summary(n))
+			if err != nil {
+				return err
+			}
+		} else {
+			// Not specified. Check if got is a zero value of its type.
+			if !reflect.DeepEqual(got, reflect.Zero(reflect.TypeOf(got)).Interface()) {
+				return fmt.Errorf("want zero, got %v (field %s of: %s)", got, fieldname, summary(n))
+			}
+		}
+	}
+
+	return nil
+}
+
+var nodeType = reflect.TypeOf((*Node)(nil)).Elem()
+
+// checkField checks a field against a field specification.
+func checkField(got interface{}, want interface{}, ctx string) error {
+	// Want nil.
+	if want == nil {
+		if !reflect.ValueOf(got).IsNil() {
+			return fmt.Errorf("want nil, got %v (%s)", got, ctx)
+		}
+		return nil
+	}
+
+	if got, ok := got.(Node); ok {
+		// Got a Node.
+		return checkNodeInField(got.(Node), want)
+	}
+	tgot := reflect.TypeOf(got)
+	if tgot.Kind() == reflect.Slice && tgot.Elem().Implements(nodeType) {
+		// Got a slice of Nodes.
+		vgot := reflect.ValueOf(got)
+		vwant := reflect.ValueOf(want)
+		if vgot.Len() != vwant.Len() {
+			return fmt.Errorf("want %d, got %d (%s)", vwant.Len(), vgot.Len(), ctx)
+		}
+		for i := 0; i < vgot.Len(); i++ {
+			err := checkNodeInField(vgot.Index(i).Interface().(Node),
+				vwant.Index(i).Interface())
+			if err != nil {
+				return err
+			}
+		}
+		return nil
+	}
+
+	if !reflect.DeepEqual(want, got) {
+		return fmt.Errorf("want %v, got %v (%s)", want, got, ctx)
+	}
+	return nil
+}
+
+func checkNodeInField(got Node, want interface{}) error {
+	switch want := want.(type) {
+	case string:
+		text := got.SourceText()
+		if want != text {
+			return fmt.Errorf("want %q, got %q (%s)", want, text, summary(got))
+		}
+		return nil
+	case ast:
+		return checkAST(got, want)
+	default:
+		panic(fmt.Sprintf("bad want type %T (%s)", want, summary(got)))
+	}
+}
+
+func exported(name string) bool {
+	r, _ := utf8.DecodeRuneInString(name)
+	return unicode.IsUpper(r)
+}
diff -pruN 0.11+ds1-5/parse/check_parse_tree_test.go 0.12+ds1-1/parse/check_parse_tree_test.go
--- 0.11+ds1-5/parse/check_parse_tree_test.go	1970-01-01 00:00:00.000000000 +0000
+++ 0.12+ds1-1/parse/check_parse_tree_test.go	2018-07-11 02:32:25.000000000 +0000
@@ -0,0 +1,43 @@
+package parse
+
+import "fmt"
+
+// checkParseTree checks whether the parse tree part of a Node is well-formed.
+func checkParseTree(n Node) error {
+	children := n.Children()
+	if len(children) == 0 {
+		return nil
+	}
+
+	// Parent pointers of all children should point to me.
+	for i, ch := range children {
+		if ch.Parent() != n {
+			return fmt.Errorf("parent of child %d (%s) is wrong: %s", i, summary(ch), summary(n))
+		}
+	}
+
+	// The Begin of the first child should be equal to mine.
+	if children[0].Begin() != n.Begin() {
+		return fmt.Errorf("gap between node and first child: %s", summary(n))
+	}
+	// The End of the last child should be equal to mine.
+	nch := len(children)
+	if children[nch-1].End() != n.End() {
+		return fmt.Errorf("gap between node and last child: %s", summary(n))
+	}
+	// Consecutive children have consecutive position ranges.
+	for i := 0; i < nch-1; i++ {
+		if children[i].End() != children[i+1].Begin() {
+			return fmt.Errorf("gap between child %d and %d of: %s", i, i+1, summary(n))
+		}
+	}
+
+	// Check children recursively.
+	for _, ch := range n.Children() {
+		err := checkParseTree(ch)
+		if err != nil {
+			return err
+		}
+	}
+	return nil
+}
diff -pruN 0.11+ds1-5/parse/parse.go 0.12+ds1-1/parse/parse.go
--- 0.11+ds1-5/parse/parse.go	2018-01-18 02:03:11.000000000 +0000
+++ 0.12+ds1-1/parse/parse.go	2018-07-11 02:32:25.000000000 +0000
@@ -45,7 +45,7 @@ var (
 	errShouldBeCompound           = newError("", "compound")
 	errShouldBeEqual              = newError("", "'='")
 	errBothElementsAndPairs       = newError("cannot contain both list elements and map pairs")
-	errShouldBeNewline            = newError("", "newline")
+	errShouldBeEscapeSequence     = newError("", "escape sequence")
 )
 
 // Chunk = { PipelineSep | Space } { Pipeline { PipelineSep | Space } }
@@ -854,7 +854,7 @@ func (pn *Primary) bareword(ps *Parser,
 // The following are allowed in barewords:
 //
 // * Anything allowed in variable names
-// * The symbols "./\@%+!"
+// * The symbols "./@%+!"
 // * The symbol "=", if ctx != lhsExpr && ctx != strictExpr
 // * The symbol ",", if ctx != bracedExpr && ctx != strictExpr
 // * The symbols "<>*^", if ctx = commandExpr
@@ -862,7 +862,7 @@ func (pn *Primary) bareword(ps *Parser,
 // The seemingly weird inclusion of \ is for easier path manipulation in
 // Windows.
 func allowedInBareword(r rune, ctx ExprCtx) bool {
-	return allowedInVariableName(r) || r == '.' || r == '/' || r == '\\' ||
+	return allowedInVariableName(r) || r == '.' || r == '/' ||
 		r == '@' || r == '%' || r == '+' || r == '!' ||
 		(ctx != LHSExpr && ctx != strictExpr && r == '=') ||
 		(ctx != BracedElemExpr && ctx != strictExpr && r == ',') ||
@@ -945,13 +945,13 @@ spaces:
 		switch {
 		case isSpace(r):
 			ps.next()
-		case r == '`': // line continuation
+		case r == '\\': // line continuation
 			ps.next()
 			switch ps.peek() {
 			case '\n':
 				ps.next()
 			case eof:
-				ps.error(errShouldBeNewline)
+				ps.error(errShouldBeEscapeSequence)
 			default:
 				ps.backup()
 				break spaces
diff -pruN 0.11+ds1-5/parse/parser.go 0.12+ds1-1/parse/parser.go
--- 0.11+ds1-5/parse/parser.go	2018-01-18 02:03:11.000000000 +0000
+++ 0.12+ds1-1/parse/parser.go	2018-07-11 02:32:25.000000000 +0000
@@ -83,7 +83,7 @@ func (ps *Parser) backup() {
 }
 
 func (ps *Parser) errorp(begin, end int, e error) {
-	ps.errors.Add(e.Error(), util.NewSourceRange(ps.srcName, ps.src, begin, end, nil))
+	ps.errors.Add(e.Error(), util.NewSourceRange(ps.srcName, ps.src, begin, end))
 }
 
 func (ps *Parser) error(e error) {
diff -pruN 0.11+ds1-5/parse/parse_test.go 0.12+ds1-1/parse/parse_test.go
--- 0.11+ds1-5/parse/parse_test.go	2018-01-18 02:03:11.000000000 +0000
+++ 0.12+ds1-1/parse/parse_test.go	2018-07-11 02:32:25.000000000 +0000
@@ -230,8 +230,8 @@ var goodCases = []struct {
 		}},
 	)},
 
-	// Line continuation: "`\n" is considered whitespace
-	{"a b`\nc", ast{
+	// Line continuation: "\\\n" is considered whitespace
+	{"a b\\\nc", ast{
 		"Chunk/Pipeline/Form", fs{"Head": "a", "Args": []string{"b", "c"}}}},
 }
 
@@ -245,57 +245,17 @@ func TestParse(t *testing.T) {
 		if err != nil {
 			t.Errorf("Parse(%q) returns bad parse tree: %v", tc.src, err)
 			fmt.Fprintf(os.Stderr, "Parse tree of %q:\n", tc.src)
-			PprintParseTree(bn, os.Stderr)
+			PPrintParseTreeTo(bn, os.Stderr)
 		}
 		err = checkAST(bn, tc.ast)
 		if err != nil {
 			t.Errorf("Parse(%q) returns bad AST: %v", tc.src, err)
 			fmt.Fprintf(os.Stderr, "AST of %q:\n", tc.src)
-			PprintAST(bn, os.Stderr)
+			PPrintASTTo(bn, os.Stderr)
 		}
 	}
 }
 
-// checkParseTree checks whether the parse tree part of a Node is well-formed.
-func checkParseTree(n Node) error {
-	children := n.Children()
-	if len(children) == 0 {
-		return nil
-	}
-
-	// Parent pointers of all children should point to me.
-	for i, ch := range children {
-		if ch.Parent() != n {
-			return fmt.Errorf("parent of child %d (%s) is wrong: %s", i, summary(ch), summary(n))
-		}
-	}
-
-	// The Begin of the first child should be equal to mine.
-	if children[0].Begin() != n.Begin() {
-		return fmt.Errorf("gap between node and first child: %s", summary(n))
-	}
-	// The End of the last child should be equal to mine.
-	nch := len(children)
-	if children[nch-1].End() != n.End() {
-		return fmt.Errorf("gap between node and last child: %s", summary(n))
-	}
-	// Consecutive children have consecutive position ranges.
-	for i := 0; i < nch-1; i++ {
-		if children[i].End() != children[i+1].Begin() {
-			return fmt.Errorf("gap between child %d and %d of: %s", i, i+1, summary(n))
-		}
-	}
-
-	// Check children recursively.
-	for _, ch := range n.Children() {
-		err := checkParseTree(ch)
-		if err != nil {
-			return err
-		}
-	}
-	return nil
-}
-
 var badCases = []struct {
 	src string
 	pos int // expected Begin position of first error
diff -pruN 0.11+ds1-5/parse/parseutil/parseutil.go 0.12+ds1-1/parse/parseutil/parseutil.go
--- 0.11+ds1-5/parse/parseutil/parseutil.go	1970-01-01 00:00:00.000000000 +0000
+++ 0.12+ds1-1/parse/parseutil/parseutil.go	2018-07-11 02:32:25.000000000 +0000
@@ -0,0 +1,44 @@
+// Package parseutil contains utilities built on top of the parse package.
+package parseutil
+
+import (
+	"strings"
+
+	"github.com/elves/elvish/parse"
+)
+
+// LeafNodeAtDot finds the leaf node at a specific position. It returns nil if
+// position is out of bound.
+func FindLeafNode(n parse.Node, p int) parse.Node {
+descend:
+	for len(n.Children()) > 0 {
+		for _, ch := range n.Children() {
+			if ch.Begin() <= p && p <= ch.End() {
+				n = ch
+				continue descend
+			}
+		}
+		return nil
+	}
+	return n
+}
+
+// Wordify turns a piece of source code into words.
+func Wordify(src string) []string {
+	n, _ := parse.Parse("[wordify]", src)
+	return wordifyInner(n, nil)
+}
+
+func wordifyInner(n parse.Node, words []string) []string {
+	if len(n.Children()) == 0 || parse.IsCompound(n) {
+		text := n.SourceText()
+		if strings.TrimFunc(text, parse.IsSpaceOrNewline) != "" {
+			return append(words, text)
+		}
+		return words
+	}
+	for _, ch := range n.Children() {
+		words = wordifyInner(ch, words)
+	}
+	return words
+}
diff -pruN 0.11+ds1-5/parse/parseutil/parseutil_test.go 0.12+ds1-1/parse/parseutil/parseutil_test.go
--- 0.11+ds1-5/parse/parseutil/parseutil_test.go	1970-01-01 00:00:00.000000000 +0000
+++ 0.12+ds1-1/parse/parseutil/parseutil_test.go	2018-07-11 02:32:25.000000000 +0000
@@ -0,0 +1,7 @@
+package parseutil
+
+import "testing"
+
+func Test(t *testing.T) {
+	// Required to get accurate test coverage report.
+}
diff -pruN 0.11+ds1-5/parse/pprint.go 0.12+ds1-1/parse/pprint.go
--- 0.11+ds1-5/parse/pprint.go	2018-01-18 02:03:11.000000000 +0000
+++ 0.12+ds1-1/parse/pprint.go	2018-07-11 02:32:25.000000000 +0000
@@ -1,6 +1,7 @@
 package parse
 
 import (
+	"bytes"
 	"fmt"
 	"io"
 	"reflect"
@@ -13,8 +14,15 @@ const (
 	indentInc     = 2
 )
 
-// PprintAST pretty prints the AST part of a Node.
-func PprintAST(n Node, wr io.Writer) {
+// PPrintAST pretty-prints the AST part of a Node.
+func PPrintAST(n Node) string {
+	var b bytes.Buffer
+	PPrintASTTo(n, &b)
+	return b.String()
+}
+
+// PPrintASTTo pretty-prints the AST part of a Node to a Writer.
+func PPrintASTTo(n Node, wr io.Writer) {
 	pprintAST(n, wr, 0, "")
 }
 
@@ -106,8 +114,15 @@ func pprintAST(n Node, wr io.Writer, ind
 	}
 }
 
-// PprintParseTree pretty prints the parse tree part of a Node.
-func PprintParseTree(n Node, wr io.Writer) {
+// PPrintParseTree pretty-prints the parse tree part of a Node.
+func PPrintParseTree(n Node) string {
+	var b bytes.Buffer
+	PPrintParseTreeTo(n, &b)
+	return b.String()
+}
+
+// PPrintParseTreeTo pretty-prints the parse tree part of a Node to a Writer.
+func PPrintParseTreeTo(n Node, wr io.Writer) {
 	pprintParseTree(n, wr, 0)
 }
 
diff -pruN 0.11+ds1-5/parse/pprint_test.go 0.12+ds1-1/parse/pprint_test.go
--- 0.11+ds1-5/parse/pprint_test.go	2018-01-18 02:03:11.000000000 +0000
+++ 0.12+ds1-1/parse/pprint_test.go	2018-07-11 02:32:25.000000000 +0000
@@ -1,16 +1,15 @@
 package parse
 
 import (
-	"bytes"
 	"testing"
+
+	"github.com/elves/elvish/tt"
 )
 
-var pprintCases = []struct {
-	src           string
-	wantAST       string
-	wantParseTree string
-}{
-	{"ls $x[0]$y[1];echo",
+var n = mustParse("ls $x[0]$y[1];echo")
+
+var pprintASTTests = tt.Table{
+	tt.Args(n).Rets(
 		`Chunk
   Pipeline/Form
     Compound/Indexing/Primary Type=Bareword Value="ls"
@@ -22,7 +21,15 @@ var pprintCases = []struct {
         Primary Type=Variable Value="y"
         Array/Compound/Indexing/Primary Type=Bareword Value="1"
   Pipeline/Form/Compound/Indexing/Primary Type=Bareword Value="echo"
-`,
+`),
+}
+
+func TestPPrintAST(t *testing.T) {
+	tt.Test(t, tt.Fn("PPrintAST", PPrintAST), pprintASTTests)
+}
+
+var pprintParseTreeTests = tt.Table{
+	tt.Args(n).Rets(
 		`Chunk "ls $x[0]$y[1];echo" 0-18
   Pipeline/Form "ls $x[0]$y[1]" 0-13
     Compound/Indexing/Primary "ls" 0-2
@@ -40,26 +47,17 @@ var pprintCases = []struct {
         Sep "]" 12-13
   Sep ";" 13-14
   Pipeline/Form/Compound/Indexing/Primary "echo" 14-18
-`},
+`),
+}
+
+func TestPprintParseTree(t *testing.T) {
+	tt.Test(t, tt.Fn("PPrintParseTree", PPrintParseTree), pprintParseTreeTests)
 }
 
-func TestPprint(t *testing.T) {
-	for _, tc := range pprintCases {
-		n, err := Parse("[test]", tc.src)
-		if err != nil {
-			t.Error(err)
-		}
-		var b bytes.Buffer
-		PprintAST(n, &b)
-		ast := b.String()
-		if b.String() != tc.wantAST {
-			t.Errorf("PprintAST(%q):\n%s\nwant:\n%s", tc.src, ast, tc.wantAST)
-		}
-		b = bytes.Buffer{}
-		PprintParseTree(n, &b)
-		pt := b.String()
-		if pt != tc.wantParseTree {
-			t.Errorf("PprintParseTree(%q):\n%s\nwant:\n%s", tc.src, pt, tc.wantParseTree)
-		}
+func mustParse(src string) Node {
+	n, err := Parse("[test]", src)
+	if err != nil {
+		panic(err)
 	}
+	return n
 }
diff -pruN 0.11+ds1-5/program/program.go 0.12+ds1-1/program/program.go
--- 0.11+ds1-5/program/program.go	2018-01-18 02:03:11.000000000 +0000
+++ 0.12+ds1-1/program/program.go	2018-07-11 02:32:25.000000000 +0000
@@ -34,7 +34,7 @@ type flagSet struct {
 
 	Help, Version, BuildInfo, JSON bool
 
-	CodeInArg, CompileOnly bool
+	CodeInArg, CompileOnly, NoRc bool
 
 	Web  bool
 	Port int
@@ -63,6 +63,7 @@ func newFlagSet() *flagSet {
 
 	f.BoolVar(&f.CodeInArg, "c", false, "take first argument as code to execute")
 	f.BoolVar(&f.CompileOnly, "compileonly", false, "Parse/Compile but do not execute")
+	f.BoolVar(&f.NoRc, "norc", false, "run elvish without invoking rc.elv")
 
 	f.BoolVar(&f.Web, "web", false, "run backend of web interface")
 	f.IntVar(&f.Port, "port", defaultWebPort, "the port of the web backend")
@@ -154,6 +155,6 @@ func FindProgram(flag *flagSet) Program
 		}
 		return web.New(flag.Bin, flag.Sock, flag.DB, flag.Port)
 	default:
-		return shell.New(flag.Bin, flag.Sock, flag.DB, flag.CodeInArg, flag.CompileOnly)
+		return shell.New(flag.Bin, flag.Sock, flag.DB, flag.CodeInArg, flag.CompileOnly, flag.NoRc)
 	}
 }
diff -pruN 0.11+ds1-5/program/shell/interact.go 0.12+ds1-1/program/shell/interact.go
--- 0.11+ds1-5/program/shell/interact.go	2018-01-18 02:03:11.000000000 +0000
+++ 0.12+ds1-1/program/shell/interact.go	2018-07-11 02:32:25.000000000 +0000
@@ -16,7 +16,7 @@ import (
 	"github.com/elves/elvish/util"
 )
 
-func interact(ev *eval.Evaler, dataDir string) {
+func interact(ev *eval.Evaler, dataDir string, norc bool) {
 	// Build Editor.
 	var ed editor
 	if sys.IsATTY(os.Stdin) {
@@ -29,7 +29,7 @@ func interact(ev *eval.Evaler, dataDir s
 	defer ed.Close()
 
 	// Source rc.elv.
-	if dataDir != "" {
+	if !norc && dataDir != "" {
 		err := sourceRC(ev, dataDir)
 		if err != nil {
 			util.PprintError(err)
@@ -72,7 +72,7 @@ func interact(ev *eval.Evaler, dataDir s
 		// No error; reset cooldown.
 		cooldown = time.Second
 
-		err = ev.SourceText(eval.NewInteractiveSource(line))
+		err = ev.EvalSource(eval.NewInteractiveSource(line))
 		if err != nil {
 			util.PprintError(err)
 		}
@@ -89,7 +89,7 @@ func sourceRC(ev *eval.Evaler, dataDir s
 	}
 	code, err := readFileUTF8(absPath)
 
-	return ev.SourceText(eval.NewScriptSource("rc.elv", absPath, code))
+	return ev.SourceRC(eval.NewScriptSource("rc.elv", absPath, code))
 }
 
 func basicReadLine() (string, error) {
diff -pruN 0.11+ds1-5/program/shell/script.go 0.12+ds1-1/program/shell/script.go
--- 0.11+ds1-5/program/shell/script.go	2018-01-18 02:03:11.000000000 +0000
+++ 0.12+ds1-1/program/shell/script.go	2018-07-11 02:32:25.000000000 +0000
@@ -49,7 +49,7 @@ func script(ev *eval.Evaler, args []stri
 		return nil
 	}
 
-	return ev.Eval(op, src)
+	return ev.EvalWithStdPorts(op, src)
 }
 
 var errSourceNotUTF8 = errors.New("source is not UTF-8")
diff -pruN 0.11+ds1-5/program/shell/shell.go 0.12+ds1-1/program/shell/shell.go
--- 0.11+ds1-5/program/shell/shell.go	2018-01-18 02:03:11.000000000 +0000
+++ 0.12+ds1-1/program/shell/shell.go	2018-07-11 02:32:25.000000000 +0000
@@ -21,14 +21,15 @@ type Shell struct {
 	DbPath      string
 	Cmd         bool
 	CompileOnly bool
+	NoRc        bool
 }
 
-func New(binpath, sockpath, dbpath string, cmd, compileonly bool) *Shell {
-	return &Shell{binpath, sockpath, dbpath, cmd, compileonly}
+func New(binpath, sockpath, dbpath string, cmd, compileonly, norc bool) *Shell {
+	return &Shell{binpath, sockpath, dbpath, cmd, compileonly, norc}
 }
 
 // Main runs Elvish using the default terminal interface. It blocks until Elvish
-// quites, and returns the exit code.
+// quits, and returns the exit code.
 func (sh *Shell) Main(args []string) int {
 	defer rescue()
 
@@ -44,7 +45,7 @@ func (sh *Shell) Main(args []string) int
 			return 2
 		}
 	} else {
-		interact(ev, dataDir)
+		interact(ev, dataDir, sh.NoRc)
 	}
 
 	return 0
diff -pruN 0.11+ds1-5/program/subprograms.go 0.12+ds1-1/program/subprograms.go
--- 0.11+ds1-5/program/subprograms.go	2018-01-18 02:03:11.000000000 +0000
+++ 0.12+ds1-1/program/subprograms.go	2018-07-11 02:32:25.000000000 +0000
@@ -3,8 +3,9 @@ package program
 import (
 	"fmt"
 	"os"
+	"runtime"
 
-	"github.com/elves/elvish/build"
+	"github.com/elves/elvish/buildinfo"
 	daemonsvc "github.com/elves/elvish/daemon"
 	"github.com/elves/elvish/program/daemon"
 )
@@ -33,8 +34,7 @@ func (s ShowCorrectUsage) Main([]string)
 type ShowVersion struct{}
 
 func (ShowVersion) Main([]string) int {
-	fmt.Println(build.Version)
-	fmt.Fprintln(os.Stderr, "-version is deprecated and will be removed in 0.12. Use -buildinfo instead.")
+	fmt.Println(buildinfo.Version)
 	return 0
 }
 
@@ -45,11 +45,16 @@ type ShowBuildInfo struct {
 
 func (info ShowBuildInfo) Main([]string) int {
 	if info.JSON {
-		fmt.Printf("{version: %s, builder: %s}\n",
-			quoteJSON(build.Version), quoteJSON(build.Builder))
+		fmt.Printf(`{"version": %s,`, quoteJSON(buildinfo.Version))
+		fmt.Printf(` "goversion": %s,`, quoteJSON(runtime.Version()))
+		fmt.Printf(` "goroot": %s,`, quoteJSON(buildinfo.GoRoot))
+		fmt.Printf(` "gopath": %s}`, quoteJSON(buildinfo.GoPath))
+		fmt.Println()
 	} else {
-		fmt.Println("version:", build.Version)
-		fmt.Println("builder:", build.Builder)
+		fmt.Println("Version:", buildinfo.Version)
+		fmt.Println("Go version:", runtime.Version())
+		fmt.Println("GOROOT at build time:", buildinfo.GoRoot)
+		fmt.Println("GOPATH at build time:", buildinfo.GoPath)
 	}
 	return 0
 }
diff -pruN 0.11+ds1-5/program/web/embedded_html.go 0.12+ds1-1/program/web/embedded_html.go
--- 0.11+ds1-5/program/web/embedded_html.go	2018-01-18 02:03:11.000000000 +0000
+++ 0.12+ds1-1/program/web/embedded_html.go	1970-01-01 00:00:00.000000000 +0000
@@ -1,91 +0,0 @@
-package web
-
-const mainPageHTML = `<html>
-
-  <div id="scrollback">
-    <div id="progress"></div>
-  </div>
-  <textarea id="code" rows="4"></textarea>
-
-  <style>
-    #code {
-      width: 100%;
-    }
-    #scrollback, #code, #progress {
-      font-family: monospace;
-      font-size: 11pt;
-    }
-    .code {
-      color: white;
-      background-color: black;
-    }
-    .error, .exception {
-      color: red;
-    }
-    .exception {
-      font-weight: bold;
-    }
-    .server-error {
-      color: white;
-      background-color: red;
-    }
-  </style>
-
-  <script>
-    // TODO(xiaq): Stream results.
-    var $scrollback = document.getElementById('scrollback'),
-        $code = document.getElementById('code'),
-        $progress = document.getElementById('progress');
-
-    $code.addEventListener('keypress', function(e) {
-      if (e.keyCode == 13 &&
-          !e.shiftKey && !e.ctrlKey && !e.altKey && !e.metaKey) {
-        e.preventDefault();
-        execute();
-      }
-    });
-
-    function execute() {
-      var code = $code.value;
-      $code.value = '';
-      addToScrollback('code', code);
-      $progress.innerText = 'executing...';
-
-      var req = new XMLHttpRequest();
-      req.onloadend = function() {
-        $progress.innerText = '';
-      };
-      req.onload = function() {
-        var res = JSON.parse(req.responseText);
-        addToScrollback('output', res.OutBytes);
-        if (res.OutValues) {
-          for (var v of res.OutValues) {
-            addToScrollback('output-value', v);
-          }
-        }
-        addToScrollback('error', res.ErrBytes);
-        addToScrollback('exception', res.Err);
-      };
-      req.onerror = function() {
-        addToScrollback('server-error', req.responseText
-          || req.statusText
-          || (req.status == req.UNSENT && "lost connection")
-          || "unknown error");
-      };
-      req.open('POST', '/execute');
-      req.send(code);
-    }
-
-    function addToScrollback(className, innerText) {
-      var $div = document.createElement('div')
-      $div.className = className;
-      $div.innerText = innerText;
-      $scrollback.insertBefore($div, $progress);
-
-      window.scrollTo(0, document.body.scrollHeight);
-    }
-
-  </script>
-
-</html>
-`
diff -pruN 0.11+ds1-5/program/web/embed-html 0.12+ds1-1/program/web/embed-html
--- 0.11+ds1-5/program/web/embed-html	2018-01-18 02:03:11.000000000 +0000
+++ 0.12+ds1-1/program/web/embed-html	1970-01-01 00:00:00.000000000 +0000
@@ -1,12 +0,0 @@
-#!/usr/bin/env elvish
-
-out = ./embedded_html.go
-
-{
-	echo "package web"
-	print "const mainPageHTML = `"
-	cat main.html | sed 's/`/`+"`"+`/g'
-	echo "`"
-} > $out
-
-gofmt -w $out
diff -pruN 0.11+ds1-5/program/web/main.html 0.12+ds1-1/program/web/main.html
--- 0.11+ds1-5/program/web/main.html	2018-01-18 02:03:11.000000000 +0000
+++ 0.12+ds1-1/program/web/main.html	1970-01-01 00:00:00.000000000 +0000
@@ -1,88 +0,0 @@
-<html>
-
-  <div id="scrollback">
-    <div id="progress"></div>
-  </div>
-  <textarea id="code" rows="4"></textarea>
-
-  <style>
-    #code {
-      width: 100%;
-    }
-    #scrollback, #code, #progress {
-      font-family: monospace;
-      font-size: 11pt;
-    }
-    .code {
-      color: white;
-      background-color: black;
-    }
-    .error, .exception {
-      color: red;
-    }
-    .exception {
-      font-weight: bold;
-    }
-    .server-error {
-      color: white;
-      background-color: red;
-    }
-  </style>
-
-  <script>
-    // TODO(xiaq): Stream results.
-    var $scrollback = document.getElementById('scrollback'),
-        $code = document.getElementById('code'),
-        $progress = document.getElementById('progress');
-
-    $code.addEventListener('keypress', function(e) {
-      if (e.keyCode == 13 &&
-          !e.shiftKey && !e.ctrlKey && !e.altKey && !e.metaKey) {
-        e.preventDefault();
-        execute();
-      }
-    });
-
-    function execute() {
-      var code = $code.value;
-      $code.value = '';
-      addToScrollback('code', code);
-      $progress.innerText = 'executing...';
-
-      var req = new XMLHttpRequest();
-      req.onloadend = function() {
-        $progress.innerText = '';
-      };
-      req.onload = function() {
-        var res = JSON.parse(req.responseText);
-        addToScrollback('output', res.OutBytes);
-        if (res.OutValues) {
-          for (var v of res.OutValues) {
-            addToScrollback('output-value', v);
-          }
-        }
-        addToScrollback('error', res.ErrBytes);
-        addToScrollback('exception', res.Err);
-      };
-      req.onerror = function() {
-        addToScrollback('server-error', req.responseText
-          || req.statusText
-          || (req.status == req.UNSENT && "lost connection")
-          || "unknown error");
-      };
-      req.open('POST', '/execute');
-      req.send(code);
-    }
-
-    function addToScrollback(className, innerText) {
-      var $div = document.createElement('div')
-      $div.className = className;
-      $div.innerText = innerText;
-      $scrollback.insertBefore($div, $progress);
-
-      window.scrollTo(0, document.body.scrollHeight);
-    }
-
-  </script>
-
-</html>
diff -pruN 0.11+ds1-5/program/web/main.html.go 0.12+ds1-1/program/web/main.html.go
--- 0.11+ds1-5/program/web/main.html.go	1970-01-01 00:00:00.000000000 +0000
+++ 0.12+ds1-1/program/web/main.html.go	2018-07-11 02:32:25.000000000 +0000
@@ -0,0 +1,233 @@
+package web
+
+const mainPageHTML = `<html>
+
+  <body class="light">
+    <div id="content">
+      <div id="scrollback">
+        <div id="progress"></div>
+      </div>
+
+      <div id="cmd" class="cmd">
+        <span id="prompt" class="prompt">&gt;&gt;</span>
+        <textarea id="buffer" class="buffer" rows="1"></textarea>
+      </div>
+
+      <div id="theme-switchers" class="flex">
+        <span class="theme-switcher" id="dark-theme">dark</span>
+        <span class="theme-switcher" id="light-theme">light</span>
+      </div>
+
+    </div>
+  </body>
+
+  <style>
+    /* Colors are taken from Material palette. */
+
+    /* Global styles */
+
+    * {
+      margin: 0;
+      padding: 0;
+      font: 12pt monospace;
+    }
+
+    #content {
+      margin: 20px;
+      padding: 20px;
+    }
+
+    /* Scrollback */
+
+    .exception {
+      font-weight: bold;
+    }
+
+    .server-error {
+      font-style: italic;
+    }
+
+    /* Command line */
+
+    .cmd {
+      display: flex;
+      width: 100%;
+    }
+
+    .cmd > .prompt {
+      display: inline-block;
+      margin-right: 1em;
+    }
+
+    .cmd > .buffer {
+      flex: 1;
+    }
+
+    /* Theme switcher */
+
+    #theme-switchers {
+      display: flex;
+      margin-top: 0.4em;
+    }
+
+    .theme-switcher {
+      cursor: pointer;
+      padding: 0 1em;
+      border: 1px solid;
+    }
+
+    #light-theme {
+      color: black;
+      background-color: white;
+      border-color: black;
+    }
+
+    #dark-theme {
+      color: white;
+      background-color: black;
+      border-color: white;
+    }
+
+    /* Color schemes. Color values from Material palette. */
+
+    body.light {
+      background: #EEEEEE; /* grey 200 */
+    }
+
+    .light * {
+      color: black;
+      background: white;
+    }
+
+    .light #content {
+      background: white;
+    }
+
+    .light .cmd > .prompt {
+      color: #1565C0; /* blue 800 */
+    }
+
+    .light .cmd > #prompt {
+      color: #2E7D32; /* green 800 */
+    }
+
+    .light .error {
+      color: #C62828; /* red 800 */
+    }
+
+
+    .dark * {
+      color: white;
+      background: black;
+    }
+
+    body.dark {
+      background: #212121; /* grey 900 */
+    }
+
+    .dark #content {
+      background: black;
+    }
+
+    .dark .cmd > .prompt {
+      color: #90CAF9; /* blue 200 */
+    }
+
+    .dark .cmd > #prompt {
+      color: #A5D6A7; /* green 200 */
+    }
+
+    .dark .error {
+      color: #EF9A9A; /* red 200 */
+    }
+
+  </style>
+
+  <script>
+    // TODO(xiaq): Stream results.
+    var $prompt = document.getElementById('prompt'),
+        $buffer = document.getElementById('buffer'),
+        $scrollback = document.getElementById('scrollback'),
+        $progress = document.getElementById('progress');
+
+    /* Theme switchers. */
+    document.getElementById('dark-theme').onclick = function() {
+      document.body.className = 'dark';
+    };
+
+    document.getElementById('light-theme').onclick = function() {
+      document.body.className = 'light';
+    };
+
+    $buffer.addEventListener('keypress', function(e) {
+      if (e.keyCode == 13 &&
+          !e.shiftKey && !e.ctrlKey && !e.altKey && !e.metaKey) {
+        e.preventDefault();
+        execute();
+      }
+    });
+
+    function execute() {
+      var code = $buffer.value;
+      addToScrollbackInner(freezeCmd());
+      $buffer.value = '';
+      $progress.innerText = 'executing...';
+
+      var req = new XMLHttpRequest();
+      req.onloadend = function() {
+        $progress.innerText = '';
+      };
+      req.onload = function() {
+        var res = JSON.parse(req.responseText);
+        addToScrollback('output', res.OutBytes);
+        if (res.OutValues) {
+          for (var v of res.OutValues) {
+            addToScrollback('output-value', v);
+          }
+        }
+        addToScrollback('error', res.ErrBytes);
+        addToScrollback('error exception', res.Err);
+      };
+      req.onerror = function() {
+        addToScrollback('error server-error', req.responseText
+          || req.statusText
+          || (req.status == req.UNSENT && "lost connection")
+          || "unknown error");
+      };
+      req.open('POST', '/execute');
+      req.send(code);
+    }
+
+    function addToScrollback(className, innerText) {
+      var div = document.createElement('div');
+      div.className = className;
+      div.innerText = innerText;
+
+      addToScrollbackInner(div);
+    }
+
+    function addToScrollbackInner(element) {
+      $scrollback.insertBefore(element, $progress);
+      window.scrollTo(0, document.body.scrollHeight);
+    }
+
+    function freezeCmd() {
+      var cmd = document.createElement('div'),
+          prompt = document.createElement('span'),
+          buffer = document.createElement('span');
+      cmd.className = 'cmd';
+      prompt.className = 'prompt';
+      prompt.innerText = $prompt.innerText;
+      buffer.className = 'buffer';
+      buffer.innerText = $buffer.value;
+      cmd.appendChild(prompt);
+      cmd.appendChild(buffer);
+      return cmd;
+    }
+
+  </script>
+
+</html>
+`
+
+// vim: se ft=html si et sw=2 ts=2 sts=2:
diff -pruN 0.11+ds1-5/program/web/web.go 0.12+ds1-1/program/web/web.go
--- 0.11+ds1-5/program/web/web.go	2018-01-18 02:03:11.000000000 +0000
+++ 0.12+ds1-1/program/web/web.go	2018-07-11 02:32:25.000000000 +0000
@@ -2,8 +2,6 @@
 // Elvish.
 package web
 
-//go:generate ./embed-html
-
 import (
 	"encoding/json"
 	"fmt"
@@ -14,7 +12,6 @@ import (
 	"os"
 
 	"github.com/elves/elvish/eval"
-	"github.com/elves/elvish/eval/types"
 	"github.com/elves/elvish/parse"
 	"github.com/elves/elvish/runtime"
 )
@@ -32,7 +29,7 @@ type httpHandler struct {
 
 type ExecuteResponse struct {
 	OutBytes  string
-	OutValues []types.Value
+	OutValues []interface{}
 	ErrBytes  string
 	Err       string
 }
@@ -99,7 +96,7 @@ const (
 // return the results collected on stdout and stderr, and the possible error
 // that occurred.
 func evalAndCollect(ev *eval.Evaler, code string) (
-	outBytes []byte, outValues []types.Value, errBytes []byte, err error) {
+	outBytes []byte, outValues []interface{}, errBytes []byte, err error) {
 
 	node, err := parse.Parse("[web]", code)
 	if err != nil {
@@ -120,7 +117,7 @@ func evalAndCollect(ev *eval.Evaler, cod
 		{File: outFile, Chan: outChan},
 		{File: errFile, Chan: eval.BlackholeChan},
 	}
-	err = ev.EvalWithPorts(ports, op, src)
+	err = ev.Eval(op, ports, src)
 
 	outFile.Close()
 	close(outChan)
@@ -164,12 +161,12 @@ func makeBytesWriterAndCollect() (*os.Fi
 // makeValuesWriterAndCollect makes a Value channel for writing, and the written
 // values will be collected in a Value slice that will be put on a channel as
 // soon as the writer is closed.
-func makeValuesWriterAndCollect() (chan types.Value, <-chan []types.Value) {
-	chanValues := make(chan types.Value, outChanBufferSize)
-	chanCollected := make(chan []types.Value)
+func makeValuesWriterAndCollect() (chan interface{}, <-chan []interface{}) {
+	chanValues := make(chan interface{}, outChanBufferSize)
+	chanCollected := make(chan []interface{})
 
 	go func() {
-		var collected []types.Value
+		var collected []interface{}
 		for {
 			for v := range chanValues {
 				collected = append(collected, v)
diff -pruN 0.11+ds1-5/README.md 0.12+ds1-1/README.md
--- 0.11+ds1-5/README.md	2018-01-18 02:03:11.000000000 +0000
+++ 0.12+ds1-1/README.md	2018-07-11 02:32:25.000000000 +0000
@@ -1,21 +1,21 @@
 # Elvish: Friendly and Expressive Shell
 
-[![logo](https://elvish.io/assets/logo.svg)](https://elvish.io/)
+[![logo](https://elv.sh/assets/logo.svg)](https://elv.sh/)
 
 Elvish is a cross-platform shell, supporting Linux, BSDs and Windows. It features an expressive programming language, with features like namespacing and anonymous functions, and a fully programmable user interface with friendly defaults. It is suitable for both interactive use and scripting.
 
 ... which is not 100% true yet. Elvish is already suitable for most daily interactive use, but it is neither complete nor stablized. Contributions are more than welcome!
 
-This README documents the development aspect of Elvish. Other information is to be found on the [website](https://elvish.io).
+This README documents the development aspect of Elvish. Other information is to be found on the [website](https://elv.sh).
 
-[![Build Status on Travis](https://img.shields.io/travis/elves/elvish.svg?label=linux%20%26%20macOS)](https://travis-ci.org/elves/elvish)
+[![Build Status on Travis](https://img.shields.io/travis/elves/elvish.svg?logo=travis&label=linux%20%26%20macOS)](https://travis-ci.org/elves/elvish)
 [![Build status on AppVeyor](https://img.shields.io/appveyor/ci/xiaq/elvish.svg?logo=appveyor&label=windows)](https://ci.appveyor.com/project/xiaq/elvish)
-[![Build Status on VSTS](https://img.shields.io/vso/build/xiaq/13c48a6c-b2dc-472e-af6c-169bf448f8e6/1.svg?label=macOS)](https://xiaq.visualstudio.com/elvish/_build)
+[![Build Status on VSTS](https://img.shields.io/vso/build/xiaq/13c48a6c-b2dc-472e-af6c-169bf448f8e6/1.svg?logo=tfs&label=macOS)](https://xiaq.visualstudio.com/elvish/_build)
 [![Code Coverage on codecov.io](https://img.shields.io/codecov/c/github/elves/elvish.svg?label=codecov)](https://codecov.io/gh/elves/elvish)
 [![Code Coverage on coveralls.io](https://img.shields.io/coveralls/github/elves/elvish.svg?label=coveralls)](https://coveralls.io/github/elves/elvish)
 [![Go Report Card](https://goreportcard.com/badge/github.com/elves/elvish)](https://goreportcard.com/report/github.com/elves/elvish)
 [![GoDoc](https://img.shields.io/badge/godoc-api-blue.svg)](http://godoc.org/github.com/elves/elvish)
-[![License](https://img.shields.io/badge/license-BSD%202--clause-blue.svg)](https://github.com/elves/elvish/blob/master/LICENSE)
+[![License](https://img.shields.io/badge/BSD-2--clause-blue.svg)](https://github.com/elves/elvish/blob/master/LICENSE)
 
 [![Gitter](https://img.shields.io/badge/gitter-elvish--public-blue.svg?logo=gitter-white)](https://gitter.im/elves/elvish-public)
 [![Telegram Group](https://img.shields.io/badge/telegram-@elvish-blue.svg)](https://telegram.me/elvish)
@@ -30,47 +30,31 @@ This README documents the development as
 
 To build Elvish, you need
 
-*   Go >= 1.8.
-
 *   Linux, {Free,Net,Open}BSD, macOS, or Windows (Windows support is experimental).
 
+*   Go >= 1.9.
 
-If you would like to contribute code to Elvish, please read
-[CONTRIBUTING.md](CONTRIBUTING.md).
-
-### The Correct Way
-
-Elvish is a go-gettable package. To build Elvish, first set up your Go workspace according to [How To Write Go Code](http://golang.org/doc/code.html), and then run
+Once you have a suitable environment, simply build Elvish with `go get`:
 
 ```sh
 go get github.com/elves/elvish
 ```
 
-### The Lazy Way
+The binary will be placed in `$GOPATH/bin`. If you haven't configured a
+`GOPATH`, it defaults to `~/go`. Refer to [How To Write Go
+Code](http://golang.org/doc/code.html) on how to set up workspace for Go.
 
-Here is something you can copy-paste into your terminal:
 
-```sh
-export GOPATH=$HOME/go
-export PATH=$PATH:$GOPATH/bin
-mkdir -p $GOPATH
-
-go get github.com/elves/elvish
+Users of macOS can also build Elvish using [Homebrew](http://brew.sh):
 
-for f in ~/.bashrc ~/.zshrc; do
-    printf 'export %s=%s\n' GOPATH '$HOME/go' PATH '$PATH:$GOPATH/bin' >> $f
-done
+```sh
+brew install --HEAD elvish
 ```
 
-The scripts sets up the Go workspace and runs `go get` for you. It assumes that you have a working Go installation and currently use `bash` or `zsh`.
-
-### The Homebrew Way
 
-Users of macOS can build Elvish using [Homebrew](http://brew.sh):
+## Contributing
 
-```sh
-brew install --HEAD elvish
-```
+See [CONTRIBUTING.md](CONTRIBUTING.md).
 
 
 ## Name
diff -pruN 0.11+ds1-5/runtime/runtime.go 0.12+ds1-1/runtime/runtime.go
--- 0.11+ds1-5/runtime/runtime.go	2018-01-18 02:03:11.000000000 +0000
+++ 0.12+ds1-1/runtime/runtime.go	2018-07-11 02:32:25.000000000 +0000
@@ -14,6 +14,8 @@ import (
 	"github.com/elves/elvish/eval"
 	daemonmod "github.com/elves/elvish/eval/daemon"
 	"github.com/elves/elvish/eval/re"
+	storemod "github.com/elves/elvish/eval/store"
+	"github.com/elves/elvish/eval/str"
 	daemonp "github.com/elves/elvish/program/daemon"
 	"github.com/elves/elvish/store/storedefs"
 	"github.com/elves/elvish/util"
@@ -73,7 +75,8 @@ func InitRuntime(binpath, sockpath, dbpa
 
 	ev := eval.NewEvaler()
 	ev.SetLibDir(filepath.Join(dataDir, "lib"))
-	ev.InstallModule("re", re.Ns())
+	ev.InstallModule("re", re.Ns)
+	ev.InstallModule("str", str.Ns)
 	if sockpath != "" && dbpath != "" {
 		spawner := &daemonp.Daemon{
 			BinPath:       binpath,
@@ -91,6 +94,7 @@ func InitRuntime(binpath, sockpath, dbpa
 		// Even if error is not nil, we install daemon-related functionalities
 		// anyway. Daemon may eventually come online and become functional.
 		ev.InstallDaemonClient(client)
+		ev.InstallModule("store", storemod.Ns(client))
 		ev.InstallModule("daemon", daemonmod.Ns(client, spawner))
 	}
 	return ev, dataDir
@@ -202,7 +206,7 @@ func killDaemon(cl *daemon.Client) error
 	if err != nil {
 		return fmt.Errorf("cannot find daemon process (pid=%d): %v", pid, err)
 	}
-	return process.Kill()
+	return process.Signal(os.Interrupt)
 }
 
 // CleanupRuntime cleans up the runtime.
diff -pruN 0.11+ds1-5/store/cmd.go 0.12+ds1-1/store/cmd.go
--- 0.11+ds1-5/store/cmd.go	2018-01-18 02:03:11.000000000 +0000
+++ 0.12+ds1-1/store/cmd.go	2018-07-11 02:32:25.000000000 +0000
@@ -47,9 +47,8 @@ func (s *Store) AddCmd(cmd string) (int,
 	return int(seq), err
 }
 
-// RemoveCmd removes a command from command history referenced by
-// sequence.
-func (s *Store) RemoveCmd(seq int) error {
+// DelCmd deletes a command history item with the given sequence number.
+func (s *Store) DelCmd(seq int) error {
 	return s.db.Update(func(tx *bolt.Tx) error {
 		b := tx.Bucket([]byte(BucketCmd))
 		return b.Delete(marshalSeq(uint64(seq)))
diff -pruN 0.11+ds1-5/store/cmd_test.go 0.12+ds1-1/store/cmd_test.go
--- 0.11+ds1-5/store/cmd_test.go	2018-01-18 02:03:11.000000000 +0000
+++ 0.12+ds1-1/store/cmd_test.go	2018-07-11 02:32:25.000000000 +0000
@@ -73,7 +73,7 @@ func TestCmd(t *testing.T) {
 		}
 	}
 
-	if err := tStore.RemoveCmd(1); err != nil {
+	if err := tStore.DelCmd(1); err != nil {
 		t.Error("Failed to remove cmd")
 	}
 	if seq, err := tStore.Cmd(1); err != storedefs.ErrNoMatchingCmd {
diff -pruN 0.11+ds1-5/store/dir.go 0.12+ds1-1/store/dir.go
--- 0.11+ds1-5/store/dir.go	2018-01-18 02:03:11.000000000 +0000
+++ 0.12+ds1-1/store/dir.go	2018-07-11 02:32:25.000000000 +0000
@@ -62,8 +62,8 @@ func (s *Store) AddDirRaw(d string, scor
 	})
 }
 
-// RemoveDir removes a directory record from history.
-func (s *Store) RemoveDir(d string) error {
+// DelDir deletes a directory record from history.
+func (s *Store) DelDir(d string) error {
 	return s.db.Update(func(tx *bolt.Tx) error {
 		b := tx.Bucket([]byte(BucketDir))
 		return b.Delete([]byte(d))
diff -pruN 0.11+ds1-5/store/dir_test.go 0.12+ds1-1/store/dir_test.go
--- 0.11+ds1-5/store/dir_test.go	2018-01-18 02:03:11.000000000 +0000
+++ 0.12+ds1-1/store/dir_test.go	2018-07-11 02:32:25.000000000 +0000
@@ -13,6 +13,9 @@ var (
 	wantedDirs = []storedefs.Dir{
 		{"/usr", scoreIncrement*scoreDecay*scoreDecay + scoreIncrement},
 		{"/usr/bin", scoreIncrement * scoreDecay}}
+	dirToDel           = "/usr"
+	wantedDirsAfterDel = []storedefs.Dir{
+		{"/usr/bin", scoreIncrement * scoreDecay}}
 )
 
 func TestDir(t *testing.T) {
@@ -28,4 +31,11 @@ func TestDir(t *testing.T) {
 		t.Errorf(`tStore.ListDirs() => (%v, %v), want (%v, <nil>)`,
 			dirs, err, wantedDirs)
 	}
+
+	tStore.DelDir("/usr")
+	dirs, err = tStore.Dirs(black)
+	if err != nil || !reflect.DeepEqual(dirs, wantedDirsAfterDel) {
+		t.Errorf(`After DelDir("/usr"), tStore.ListDirs() => (%v, %v), want (%v, <nil>)`,
+			dirs, err, wantedDirsAfterDel)
+	}
 }
diff -pruN 0.11+ds1-5/store/storedefs/interface.go 0.12+ds1-1/store/storedefs/interface.go
--- 0.11+ds1-5/store/storedefs/interface.go	2018-01-18 02:03:11.000000000 +0000
+++ 0.12+ds1-1/store/storedefs/interface.go	2018-07-11 02:32:25.000000000 +0000
@@ -4,12 +4,14 @@ package storedefs
 type Store interface {
 	NextCmdSeq() (int, error)
 	AddCmd(text string) (int, error)
+	DelCmd(seq int) error
 	Cmd(seq int) (string, error)
 	Cmds(from, upto int) ([]string, error)
 	NextCmd(from int, prefix string) (int, string, error)
 	PrevCmd(upto int, prefix string) (int, string, error)
 
 	AddDir(dir string, incFactor float64) error
+	DelDir(dir string) error
 	Dirs(blacklist map[string]struct{}) ([]Dir, error)
 
 	SharedVar(name string) (string, error)
diff -pruN 0.11+ds1-5/styled/segment.go 0.12+ds1-1/styled/segment.go
--- 0.11+ds1-5/styled/segment.go	1970-01-01 00:00:00.000000000 +0000
+++ 0.12+ds1-1/styled/segment.go	2018-07-11 02:32:25.000000000 +0000
@@ -0,0 +1,108 @@
+package styled
+
+import (
+	"bytes"
+	"fmt"
+	"strings"
+
+	"github.com/elves/elvish/eval/vals"
+)
+
+// Segment is a string that has some style applied to it.
+type Segment struct {
+	Style
+	Text string
+}
+
+func (Segment) Kind() string { return "styled-segment" }
+
+// Repr returns the representation of this Segment. The string can be used to
+// construct an identical Segment. Unset or default attributes are skipped. If
+// the Segment represents an unstyled string only this string is returned.
+func (s Segment) Repr(indent int) string {
+	buf := new(bytes.Buffer)
+	addIfNotEqual := func(key string, val, cmp interface{}) {
+		if val != cmp {
+			fmt.Fprintf(buf, "&%s=%s ", key, vals.Repr(val, 0))
+		}
+	}
+
+	addIfNotEqual("fg-color", s.Foreground, "")
+	addIfNotEqual("bg-color", s.Background, "")
+	addIfNotEqual("bold", s.Bold, false)
+	addIfNotEqual("dim", s.Dim, false)
+	addIfNotEqual("italic", s.Italic, false)
+	addIfNotEqual("underlined", s.Underlined, false)
+	addIfNotEqual("blink", s.Blink, false)
+	addIfNotEqual("inverse", s.Inverse, false)
+
+	if buf.Len() == 0 {
+		return s.Text
+	}
+
+	return fmt.Sprintf("(styled-segment %s %s)", s.Text, strings.TrimSpace(buf.String()))
+}
+
+func (s Segment) IterateKeys(fn func(v interface{}) bool) {
+	vals.Feed(fn, "text", "fg-color", "bg-color", "bold", "dim", "italic", "underlined", "blink", "inverse")
+}
+
+// Index provides access to the attributes of the Segment.
+func (s Segment) Index(k interface{}) (v interface{}, ok bool) {
+	switch k {
+	case "text":
+		v = s.Text
+	case "fg-color":
+		v = s.Foreground
+	case "bg-color":
+		v = s.Background
+	case "bold":
+		v = s.Bold
+	case "dim":
+		v = s.Dim
+	case "italic":
+		v = s.Italic
+	case "underlined":
+		v = s.Underlined
+	case "blink":
+		v = s.Blink
+	case "inverse":
+		v = s.Inverse
+	}
+
+	if v == "" {
+		v = "default"
+	}
+
+	return v, v != nil
+}
+
+// Concat implements Segment+string, Segment+Segment and Segment+Text.
+func (s Segment) Concat(v interface{}) (interface{}, error) {
+	switch rhs := v.(type) {
+	case string:
+		return Text{
+			s,
+			Segment{Text: rhs},
+		}, nil
+	case *Segment:
+		return Text{s, *rhs}, nil
+	case *Text:
+		return Text(append([]Segment{s}, *rhs...)), nil
+	}
+
+	return nil, vals.ErrConcatNotImplemented
+}
+
+// RConcat implements string+Segment.
+func (s Segment) RConcat(v interface{}) (interface{}, error) {
+	switch lhs := v.(type) {
+	case string:
+		return Text{
+			Segment{Text: lhs},
+			s,
+		}, nil
+	}
+
+	return nil, vals.ErrConcatNotImplemented
+}
diff -pruN 0.11+ds1-5/styled/styled.go 0.12+ds1-1/styled/styled.go
--- 0.11+ds1-5/styled/styled.go	1970-01-01 00:00:00.000000000 +0000
+++ 0.12+ds1-1/styled/styled.go	2018-07-11 02:32:25.000000000 +0000
@@ -0,0 +1,98 @@
+package styled
+
+import (
+	"fmt"
+)
+
+// Style specifies how something (mostly a string) shall be displayed.
+type Style struct {
+	Foreground string
+	Background string
+	Bold       bool
+	Dim        bool
+	Italic     bool
+	Underlined bool
+	Blink      bool
+	Inverse    bool
+}
+
+// ImportFromOptions assigns all recognized values from a map to the current
+// Style.
+func (s *Style) ImportFromOptions(options map[string]interface{}) error {
+	assignColor := func(val interface{}, colorField *string) string {
+		if val == "default" {
+			*colorField = ""
+		} else if c, ok := val.(string); ok && isValidColorName(c) {
+			*colorField = c
+		} else {
+			return "valid color string"
+		}
+		return ""
+	}
+	assignBool := func(val interface{}, attrField *bool) string {
+		if b, ok := val.(bool); ok {
+			*attrField = b
+		} else {
+			return "bool value"
+		}
+		return ""
+	}
+
+	for k, v := range options {
+		var need string
+
+		switch k {
+		case "fg-color":
+			need = assignColor(v, &s.Foreground)
+		case "bg-color":
+			need = assignColor(v, &s.Background)
+		case "bold":
+			need = assignBool(v, &s.Bold)
+		case "dim":
+			need = assignBool(v, &s.Dim)
+		case "italic":
+			need = assignBool(v, &s.Italic)
+		case "underlined":
+			need = assignBool(v, &s.Underlined)
+		case "blink":
+			need = assignBool(v, &s.Blink)
+		case "inverse":
+			need = assignBool(v, &s.Inverse)
+
+		default:
+			return fmt.Errorf("unrecognized option '%s'", k)
+		}
+
+		if need != "" {
+			return fmt.Errorf("value to option '%s' must be a %s", k, need)
+		}
+	}
+
+	return nil
+}
+
+func isValidColorName(col string) bool {
+	switch col {
+	case
+		"black",
+		"red",
+		"green",
+		"yellow",
+		"blue",
+		"magenta",
+		"cyan",
+		"lightgray",
+		"gray",
+		"lightred",
+		"lightgreen",
+		"lightyellow",
+		"lightblue",
+		"lightmagenta",
+		"lightcyan",
+		"white":
+		return true
+
+	default:
+		return false
+	}
+}
diff -pruN 0.11+ds1-5/styled/text.go 0.12+ds1-1/styled/text.go
--- 0.11+ds1-5/styled/text.go	1970-01-01 00:00:00.000000000 +0000
+++ 0.12+ds1-1/styled/text.go	2018-07-11 02:32:25.000000000 +0000
@@ -0,0 +1,101 @@
+package styled
+
+import (
+	"bytes"
+	"fmt"
+	"strconv"
+
+	"github.com/elves/elvish/eval/vals"
+)
+
+// Text contains of a list of styled Segments.
+type Text []Segment
+
+func (t Text) Kind() string { return "styled-text" }
+
+// Repr returns the representation of the current Text. It is just a wrapper
+// around the containing Segments.
+func (t Text) Repr(indent int) string {
+	buf := new(bytes.Buffer)
+	for _, s := range t {
+		buf.WriteString(s.Repr(indent + 1))
+	}
+	return fmt.Sprintf("(styled %s)", buf.String())
+}
+
+func (t Text) IterateKeys(fn func(interface{}) bool) {
+	for i := 0; i < len(t); i++ {
+		if !fn(strconv.Itoa(i)) {
+			break
+		}
+	}
+}
+
+// Index provides access to the underlying Segments.
+func (t Text) Index(k interface{}) (interface{}, error) {
+	index, err := vals.ConvertListIndex(k, len(t))
+	if err != nil {
+		return nil, err
+	} else if index.Slice {
+		return t[index.Lower:index.Upper], nil
+	} else {
+		return t[index.Lower], nil
+	}
+}
+
+// Concat implements Text+string, Text+Segment and Text+Text.
+func (t Text) Concat(v interface{}) (interface{}, error) {
+	switch rhs := v.(type) {
+	case string:
+		return Text(append(t, Segment{Text: rhs})), nil
+	case *Segment:
+		return Text(append(t, *rhs)), nil
+	case *Text:
+		return Text(append(t, *rhs...)), nil
+	}
+
+	return nil, vals.ErrConcatNotImplemented
+}
+
+// RConcat implements string+Text.
+func (t Text) RConcat(v interface{}) (interface{}, error) {
+	switch lhs := v.(type) {
+	case string:
+		return Text(append([]Segment{{Text: lhs}}, t...)), nil
+	}
+
+	return nil, vals.ErrConcatNotImplemented
+}
+
+// Partition partitions the Text at n indicies into n+1 Text values.
+func (t Text) Partition(indicies ...int) []Text {
+	out := make([]Text, len(indicies)+1)
+	segs := t
+	consumedSegsLen := 0
+	seg0Consumed := 0
+	for i, idx := range indicies {
+		text := make(Text, 0)
+		for len(segs) > 0 && idx >= consumedSegsLen+len(segs[0].Text) {
+			text = append(text, Segment{
+				segs[0].Style, segs[0].Text[seg0Consumed:]})
+			consumedSegsLen += len(segs[0].Text)
+			seg0Consumed = 0
+			segs = segs[1:]
+		}
+		if len(segs) > 0 && idx > consumedSegsLen {
+			text = append(text, Segment{
+				segs[0].Style, segs[0].Text[:idx-consumedSegsLen]})
+			seg0Consumed = idx - consumedSegsLen
+		}
+		out[i] = text
+	}
+	trailing := make(Text, 0)
+	for len(segs) > 0 {
+		trailing = append(trailing, Segment{
+			segs[0].Style, segs[0].Text[seg0Consumed:]})
+		seg0Consumed = 0
+		segs = segs[1:]
+	}
+	out[len(indicies)] = trailing
+	return out
+}
diff -pruN 0.11+ds1-5/styled/text_test.go 0.12+ds1-1/styled/text_test.go
--- 0.11+ds1-5/styled/text_test.go	1970-01-01 00:00:00.000000000 +0000
+++ 0.12+ds1-1/styled/text_test.go	2018-07-11 02:32:25.000000000 +0000
@@ -0,0 +1,42 @@
+package styled
+
+import (
+	"testing"
+
+	"github.com/elves/elvish/tt"
+)
+
+// TODO: Test other methods of Text.
+
+var Args = tt.Args
+
+var (
+	text0 = Text{}
+	text1 = Text{red("lorem")}
+	text2 = Text{red("lorem"), blue("foobar")}
+)
+
+func red(s string) Segment  { return Segment{Style{Foreground: "red"}, s} }
+func blue(s string) Segment { return Segment{Style{Foreground: "blue"}, s} }
+
+var partitionTests = tt.Table{
+	Args(text0).Rets([]Text{text0}),
+	Args(text1).Rets([]Text{text1}),
+	Args(text1, 0).Rets([]Text{text0, text1}),
+	Args(text1, 1).Rets([]Text{Text{red("l")}, Text{red("orem")}}),
+	Args(text1, 5).Rets([]Text{text1, text0}),
+	Args(text2).Rets([]Text{text2}),
+	Args(text2, 0).Rets([]Text{text0, text2}),
+	Args(text2, 1).Rets([]Text{
+		Text{red("l")}, Text{red("orem"), blue("foobar")}}),
+	Args(text2, 2).Rets([]Text{
+		Text{red("lo")}, Text{red("rem"), blue("foobar")}}),
+	Args(text2, 5).Rets([]Text{Text{red("lorem")}, Text{blue("foobar")}}),
+	Args(text2, 6).Rets([]Text{
+		Text{red("lorem"), blue("f")}, Text{blue("oobar")}}),
+	Args(text2, 11).Rets([]Text{text2, text0}),
+}
+
+func TestPartition(t *testing.T) {
+	tt.Test(t, tt.Fn("Text.Parition", Text.Partition), partitionTests)
+}
diff -pruN 0.11+ds1-5/styled/to_legacy_type.go 0.12+ds1-1/styled/to_legacy_type.go
--- 0.11+ds1-5/styled/to_legacy_type.go	1970-01-01 00:00:00.000000000 +0000
+++ 0.12+ds1-1/styled/to_legacy_type.go	2018-07-11 02:32:25.000000000 +0000
@@ -0,0 +1,43 @@
+package styled
+
+import "github.com/elves/elvish/edit/ui"
+
+func (t Text) ToLegacyType() []*ui.Styled {
+	out := make([]*ui.Styled, len(t))
+	for i, seg := range t {
+		out[i] = seg.ToLegacyType()
+	}
+	return out
+}
+
+func (s Segment) ToLegacyType() *ui.Styled {
+	legacy := &ui.Styled{Text: s.Text}
+	addLegacyStyle := func(style string) {
+		legacy.Styles = append(legacy.Styles, style)
+	}
+	if s.Style.Foreground != "" {
+		addLegacyStyle(s.Style.Foreground)
+	}
+	if s.Style.Background != "" {
+		addLegacyStyle("bg-" + s.Style.Background)
+	}
+	if s.Style.Bold {
+		addLegacyStyle("bold")
+	}
+	if s.Style.Dim {
+		addLegacyStyle("dim")
+	}
+	if s.Style.Italic {
+		addLegacyStyle("italic")
+	}
+	if s.Style.Underlined {
+		addLegacyStyle("underlined")
+	}
+	if s.Style.Blink {
+		addLegacyStyle("blink")
+	}
+	if s.Style.Inverse {
+		addLegacyStyle("inverse")
+	}
+	return legacy
+}
diff -pruN 0.11+ds1-5/styled/to_string.go 0.12+ds1-1/styled/to_string.go
--- 0.11+ds1-5/styled/to_string.go	1970-01-01 00:00:00.000000000 +0000
+++ 0.12+ds1-1/styled/to_string.go	2018-07-11 02:32:25.000000000 +0000
@@ -0,0 +1,91 @@
+package styled
+
+import (
+	"strings"
+)
+
+// todo: Make string conversion variable to environment.
+// E.g. the shell displays colors different than HTML.
+func (t Text) String() string {
+	buf := make([]byte, 0, 64)
+	for _, segment := range t {
+		styleBuf := make([]string, 0, 8)
+
+		if segment.Bold {
+			styleBuf = append(styleBuf, "1")
+		}
+		if segment.Dim {
+			styleBuf = append(styleBuf, "2")
+		}
+		if segment.Italic {
+			styleBuf = append(styleBuf, "3")
+		}
+		if segment.Underlined {
+			styleBuf = append(styleBuf, "4")
+		}
+		if segment.Blink {
+			styleBuf = append(styleBuf, "5")
+		}
+		if segment.Inverse {
+			styleBuf = append(styleBuf, "7")
+		}
+
+		if segment.Foreground != "" {
+			if col, ok := colorTranslationTable[segment.Foreground]; ok {
+				styleBuf = append(styleBuf, col)
+			}
+		}
+		if segment.Background != "" {
+			if col, ok := colorTranslationTable["bg-"+segment.Background]; ok {
+				styleBuf = append(styleBuf, col)
+			}
+		}
+
+		if len(styleBuf) > 0 {
+			buf = append(buf, "\033["...)
+			buf = append(buf, strings.Join(styleBuf, ";")...)
+			buf = append(buf, 'm')
+			buf = append(buf, segment.Text...)
+			buf = append(buf, "\033[m"...)
+		} else {
+			buf = append(buf, segment.Text...)
+		}
+	}
+	return string(buf)
+}
+
+var colorTranslationTable = map[string]string{
+	"black":        "30",
+	"red":          "31",
+	"green":        "32",
+	"yellow":       "33",
+	"blue":         "34",
+	"magenta":      "35",
+	"cyan":         "36",
+	"lightgray":    "37",
+	"gray":         "90",
+	"lightred":     "91",
+	"lightgreen":   "92",
+	"lightyellow":  "93",
+	"lightblue":    "94",
+	"lightmagenta": "95",
+	"lightcyan":    "96",
+	"white":        "97",
+
+	"bg-black":        "40",
+	"bg-red":          "41",
+	"bg-green":        "42",
+	"bg-yellow":       "43",
+	"bg-blue":         "44",
+	"bg-magenta":      "45",
+	"bg-cyan":         "46",
+	"bg-lightgray":    "47",
+	"bg-gray":         "100",
+	"bg-lightred":     "101",
+	"bg-lightgreen":   "102",
+	"bg-lightyellow":  "103",
+	"bg-lightblue":    "104",
+	"bg-lightmagenta": "105",
+	"bg-lightcyan":    "106",
+	"bg-white":        "107",
+}
diff -pruN 0.11+ds1-5/styled/transform.go 0.12+ds1-1/styled/transform.go
--- 0.11+ds1-5/styled/transform.go	1970-01-01 00:00:00.000000000 +0000
+++ 0.12+ds1-1/styled/transform.go	2018-07-11 02:32:25.000000000 +0000
@@ -0,0 +1,91 @@
+package styled
+
+import (
+	"strings"
+)
+
+// FindTransformer looks up a transformer name and if successful returns a
+// function that can be used to transform a styled Segment.
+func FindTransformer(transformerName string) func(Segment) Segment {
+	var innerTransformer func(*Segment)
+
+	switch {
+	// Catch special colors early
+	case transformerName == "default":
+		innerTransformer = func(s *Segment) { s.Foreground = "" }
+	case transformerName == "bg-default":
+		innerTransformer = func(s *Segment) { s.Background = "" }
+
+	case strings.HasPrefix(transformerName, "bg-"):
+		innerTransformer = buildColorTransformer(strings.TrimPrefix(transformerName, "bg-"), false)
+	case strings.HasPrefix(transformerName, "no-"):
+		innerTransformer = buildBoolTransformer(strings.TrimPrefix(transformerName, "no-"), false, false)
+	case strings.HasPrefix(transformerName, "toggle-"):
+		innerTransformer = buildBoolTransformer(strings.TrimPrefix(transformerName, "toggle-"), false, true)
+
+	default:
+		innerTransformer = buildColorTransformer(transformerName, true)
+
+		if innerTransformer == nil {
+			innerTransformer = buildBoolTransformer(transformerName, true, false)
+		}
+	}
+
+	if innerTransformer == nil {
+		return nil
+	}
+
+	return func(segment Segment) Segment {
+		innerTransformer(&segment)
+		return segment
+	}
+}
+
+func buildColorTransformer(transformerName string, setForeground bool) func(*Segment) {
+	if isValidColorName(transformerName) {
+		if setForeground {
+			return func(s *Segment) { s.Foreground = transformerName }
+		} else {
+			return func(s *Segment) { s.Background = transformerName }
+		}
+	}
+
+	return nil
+}
+
+func buildBoolTransformer(transformerName string, val, toggle bool) func(*Segment) {
+	switch transformerName {
+	case "bold":
+		if toggle {
+			return func(s *Segment) { s.Bold = !s.Bold }
+		}
+		return func(s *Segment) { s.Bold = val }
+	case "dim":
+		if toggle {
+			return func(s *Segment) { s.Dim = !s.Dim }
+		}
+		return func(s *Segment) { s.Dim = val }
+	case "italic":
+		if toggle {
+			return func(s *Segment) { s.Italic = !s.Italic }
+		}
+		return func(s *Segment) { s.Italic = val }
+	case "underlined":
+		if toggle {
+			return func(s *Segment) { s.Underlined = !s.Underlined }
+		}
+		return func(s *Segment) { s.Underlined = val }
+	case "blink":
+		if toggle {
+			return func(s *Segment) { s.Blink = !s.Blink }
+		}
+		return func(s *Segment) { s.Blink = val }
+	case "inverse":
+		if toggle {
+			return func(s *Segment) { s.Inverse = !s.Inverse }
+		}
+		return func(s *Segment) { s.Inverse = val }
+	}
+
+	return nil
+}
diff -pruN 0.11+ds1-5/sys/fdset_freebsd.go 0.12+ds1-1/sys/fdset_freebsd.go
--- 0.11+ds1-5/sys/fdset_freebsd.go	2018-01-18 02:03:11.000000000 +0000
+++ 0.12+ds1-1/sys/fdset_freebsd.go	2018-07-11 02:32:25.000000000 +0000
@@ -8,15 +8,16 @@ package sys
 
 import (
 	"reflect"
-	"syscall"
+
+	"golang.org/x/sys/unix"
 )
 
-var nFdBits = (uint)(reflect.TypeOf(syscall.FdSet{}.X__fds_bits[0]).Size() * 8)
+var nFdBits = (uint)(reflect.TypeOf(unix.FdSet{}.X__fds_bits[0]).Size() * 8)
 
-type FdSet syscall.FdSet
+type FdSet unix.FdSet
 
-func (fs *FdSet) s() *syscall.FdSet {
-	return (*syscall.FdSet)(fs)
+func (fs *FdSet) s() *unix.FdSet {
+	return (*unix.FdSet)(fs)
 }
 
 func NewFdSet(fds ...int) *FdSet {
diff -pruN 0.11+ds1-5/sys/fdset_notfreebsd.go 0.12+ds1-1/sys/fdset_notfreebsd.go
--- 0.11+ds1-5/sys/fdset_notfreebsd.go	2018-01-18 02:03:11.000000000 +0000
+++ 0.12+ds1-1/sys/fdset_notfreebsd.go	2018-07-11 02:32:25.000000000 +0000
@@ -3,16 +3,17 @@
 package sys
 
 import (
-	"syscall"
 	"unsafe"
+
+	"golang.org/x/sys/unix"
 )
 
-var nFdBits = uint(8 * unsafe.Sizeof(syscall.FdSet{}.Bits[0]))
+var nFdBits = uint(8 * unsafe.Sizeof(unix.FdSet{}.Bits[0]))
 
-type FdSet syscall.FdSet
+type FdSet unix.FdSet
 
-func (fs *FdSet) s() *syscall.FdSet {
-	return (*syscall.FdSet)(fs)
+func (fs *FdSet) s() *unix.FdSet {
+	return (*unix.FdSet)(fs)
 }
 
 func NewFdSet(fds ...int) *FdSet {
diff -pruN 0.11+ds1-5/sys/select_bsd.go 0.12+ds1-1/sys/select_bsd.go
--- 0.11+ds1-5/sys/select_bsd.go	2018-01-18 02:03:11.000000000 +0000
+++ 0.12+ds1-1/sys/select_bsd.go	2018-07-11 02:32:25.000000000 +0000
@@ -2,8 +2,8 @@
 
 package sys
 
-import "syscall"
+import "golang.org/x/sys/unix"
 
-func Select(nfd int, r *FdSet, w *FdSet, e *FdSet, timeout *syscall.Timeval) (err error) {
-	return syscall.Select(nfd, r.s(), w.s(), e.s(), timeout)
+func Select(nfd int, r *FdSet, w *FdSet, e *FdSet) error {
+	return unix.Select(nfd, r.s(), w.s(), e.s(), nil)
 }
diff -pruN 0.11+ds1-5/sys/select_linux.go 0.12+ds1-1/sys/select_linux.go
--- 0.11+ds1-5/sys/select_linux.go	2018-01-18 02:03:11.000000000 +0000
+++ 0.12+ds1-1/sys/select_linux.go	2018-07-11 02:32:25.000000000 +0000
@@ -2,9 +2,12 @@
 
 package sys
 
-import "syscall"
+import "golang.org/x/sys/unix"
 
-func Select(nfd int, r *FdSet, w *FdSet, e *FdSet, timeout *syscall.Timeval) error {
-	_, err := syscall.Select(nfd, r.s(), w.s(), e.s(), timeout)
+func Select(nfd int, r *FdSet, w *FdSet, e *FdSet) error {
+	// On ARM64, MIPS64 and MIPS64LE, unix.Select is emulated in userland and
+	// will dereference timeout. In that case, we use Pselect to work around the
+	// problem. Bug: https://github.com/golang/go/issues/24189
+	_, err := unix.Pselect(nfd, r.s(), w.s(), e.s(), nil, nil)
 	return err
 }
diff -pruN 0.11+ds1-5/sys/select_test.go 0.12+ds1-1/sys/select_test.go
--- 0.11+ds1-5/sys/select_test.go	2018-01-18 02:03:11.000000000 +0000
+++ 0.12+ds1-1/sys/select_test.go	2018-07-11 02:32:25.000000000 +0000
@@ -3,8 +3,9 @@
 package sys
 
 import (
-	"syscall"
 	"testing"
+
+	"golang.org/x/sys/unix"
 )
 
 func TestFdSet(t *testing.T) {
@@ -30,8 +31,8 @@ func TestFdSet(t *testing.T) {
 
 func TestSelect(t *testing.T) {
 	var p1, p2 [2]int
-	mustNil(syscall.Pipe(p1[:]))
-	mustNil(syscall.Pipe(p2[:]))
+	mustNil(unix.Pipe(p1[:]))
+	mustNil(unix.Pipe(p2[:]))
 	fs := NewFdSet(p1[0], p2[0])
 	var maxfd int
 	if p1[0] > p2[0] {
@@ -40,16 +41,16 @@ func TestSelect(t *testing.T) {
 		maxfd = p2[0] + 1
 	}
 	go func() {
-		syscall.Write(p1[1], []byte("to p1"))
-		syscall.Write(p2[1], []byte("to p2"))
-		syscall.Close(p1[1])
-		syscall.Close(p2[1])
+		unix.Write(p1[1], []byte("to p1"))
+		unix.Write(p2[1], []byte("to p2"))
+		unix.Close(p1[1])
+		unix.Close(p2[1])
 	}()
-	e := Select(maxfd+1, fs, nil, nil, nil)
+	e := Select(maxfd+1, fs, nil, nil)
 	if e != nil {
-		t.Errorf("Select(%v, %v, nil, nil, nil) => %v, want <nil>",
+		t.Errorf("Select(%v, %v, nil, nil) => %v, want <nil>",
 			maxfd+1, fs, e)
 	}
-	syscall.Close(p1[0])
-	syscall.Close(p2[0])
+	unix.Close(p1[0])
+	unix.Close(p2[0])
 }
diff -pruN 0.11+ds1-5/sys/waitforread_unix.go 0.12+ds1-1/sys/waitforread_unix.go
--- 0.11+ds1-5/sys/waitforread_unix.go	2018-01-18 02:03:11.000000000 +0000
+++ 0.12+ds1-1/sys/waitforread_unix.go	2018-07-11 02:32:25.000000000 +0000
@@ -20,7 +20,7 @@ func WaitForRead(files ...*os.File) (rea
 		}
 		fdset.Set(fd)
 	}
-	err = Select(maxfd+1, fdset, nil, nil, nil)
+	err = Select(maxfd+1, fdset, nil, nil)
 	ready = make([]bool, len(files))
 	for i, file := range files {
 		ready[i] = fdset.IsSet(int(file.Fd()))
diff -pruN 0.11+ds1-5/sys/winsize_unix.go 0.12+ds1-1/sys/winsize_unix.go
--- 0.11+ds1-5/sys/winsize_unix.go	2018-01-18 02:03:11.000000000 +0000
+++ 0.12+ds1-1/sys/winsize_unix.go	2018-07-11 02:32:25.000000000 +0000
@@ -9,14 +9,13 @@ package sys
 import (
 	"fmt"
 	"os"
-	"syscall"
 	"unsafe"
 
 	"golang.org/x/sys/unix"
 )
 
 // SIGWINCH is the Window size change signal.
-const SIGWINCH = syscall.SIGWINCH
+const SIGWINCH = unix.SIGWINCH
 
 // winSize mirrors struct winsize in the C header.
 // The following declaration matches struct winsize in the headers of
diff -pruN 0.11+ds1-5/sys/winsize_windows.go 0.12+ds1-1/sys/winsize_windows.go
--- 0.11+ds1-5/sys/winsize_windows.go	2018-01-18 02:03:11.000000000 +0000
+++ 0.12+ds1-1/sys/winsize_windows.go	2018-07-11 02:32:25.000000000 +0000
@@ -10,6 +10,8 @@ import (
 
 // SIGWINCH is the Window size change signal. On Windows this signal does not
 // exist, so we use -1, an impossible value for signals.
+// NOTE: This has to use the syscall package before the x/sys/* packages also
+// use syscall.Signal to define signals.
 const SIGWINCH = syscall.Signal(-1)
 
 // GetWinsize queries the size of the terminal referenced by the given file.
diff -pruN 0.11+ds1-5/.travis.yml 0.12+ds1-1/.travis.yml
--- 0.11+ds1-5/.travis.yml	2018-01-18 02:03:11.000000000 +0000
+++ 0.12+ds1-1/.travis.yml	2018-07-11 02:32:25.000000000 +0000
@@ -1,5 +1,5 @@
 language: go
-go: 1.9
+go: "1.10"
 env: CGO_ENABLED=0
 script: make travis
 
@@ -9,4 +9,4 @@ os:
     - osx
 matrix:
   include:
-    - go: 1.8
+    - go: "1.9"
diff -pruN 0.11+ds1-5/tt/tt.go 0.12+ds1-1/tt/tt.go
--- 0.11+ds1-5/tt/tt.go	2018-01-18 02:03:11.000000000 +0000
+++ 0.12+ds1-1/tt/tt.go	2018-07-11 02:32:25.000000000 +0000
@@ -26,22 +26,14 @@ func Args(args ...interface{}) *Case {
 }
 
 // Rets modifies the test case so that it requires the return values to match
-// the given values. It returns the receiver.
+// the given values. It returns the receiver. The arguments may implement the
+// Matcher interface, in which case its Match method is called with the actual
+// return value. Otherwise, reflect.DeepEqual is used to determine matches.
 func (c *Case) Rets(matchers ...interface{}) *Case {
 	c.retsMatchers = append(c.retsMatchers, matchers)
 	return c
 }
 
-func match(matchers, actual []interface{}) bool {
-	for i, matcher := range matchers {
-		// TODO: Support custom matching strategy
-		if !reflect.DeepEqual(matcher, actual[i]) {
-			return false
-		}
-	}
-	return true
-}
-
 // FnToTest describes a function to test.
 type FnToTest struct {
 	name    string
@@ -99,6 +91,32 @@ func Test(t T, fn *FnToTest, tests Table
 	}
 }
 
+// RetValue is an empty interface used in the Matcher interface.
+type RetValue interface{}
+
+// Matcher wraps the Match method.
+type Matcher interface {
+	// Match reports whether a return value is considered a match. The argument
+	// is of type RetValue so that it cannot be implemented accidentally.
+	Match(RetValue) bool
+}
+
+func match(matchers, actual []interface{}) bool {
+	for i, matcher := range matchers {
+		if !matchOne(matcher, actual[i]) {
+			return false
+		}
+	}
+	return true
+}
+
+func matchOne(m, a interface{}) bool {
+	if m, ok := m.(Matcher); ok {
+		return m.Match(a)
+	}
+	return reflect.DeepEqual(m, a)
+}
+
 func sprintArgs(args ...interface{}) string {
 	return sprintCommaDelimited(args...)
 }
diff -pruN 0.11+ds1-5/util/multierror.go 0.12+ds1-1/util/multierror.go
--- 0.11+ds1-5/util/multierror.go	2018-01-18 02:03:11.000000000 +0000
+++ 0.12+ds1-1/util/multierror.go	2018-07-11 02:32:25.000000000 +0000
@@ -27,17 +27,26 @@ func (es MultiError) Error() string {
 }
 
 // Errors concatenate multiple errors into one. If all errors are nil, it
-// returns nil, otherwise the return value is a MultiError containing all the
-// non-nil arguments.
+// returns nil. If there is one non-nil error, it is returned. Otherwise the
+// return value is a MultiError containing all the non-nil arguments. Arguments
+// of the type MultiError are flattened.
 func Errors(errs ...error) error {
 	var nonNil []error
 	for _, err := range errs {
 		if err != nil {
-			nonNil = append(nonNil, err)
+			if multi, ok := err.(MultiError); ok {
+				nonNil = append(nonNil, multi.Errors...)
+			} else {
+				nonNil = append(nonNil, err)
+			}
 		}
 	}
-	if len(nonNil) == 0 {
+	switch len(nonNil) {
+	case 0:
 		return nil
+	case 1:
+		return nonNil[0]
+	default:
+		return MultiError{nonNil}
 	}
-	return MultiError{nonNil}
 }
diff -pruN 0.11+ds1-5/util/source_range.go 0.12+ds1-1/util/source_range.go
--- 0.11+ds1-5/util/source_range.go	2018-01-18 02:03:11.000000000 +0000
+++ 0.12+ds1-1/util/source_range.go	2018-07-11 02:32:25.000000000 +0000
@@ -13,14 +13,13 @@ type SourceRange struct {
 	Source string
 	Begin  int
 	End    int
-	Next   *SourceRange
 
 	savedPprintInfo *rangePprintInfo
 }
 
 // NewSourceRange creates a new SourceRange.
-func NewSourceRange(name, source string, begin, end int, next *SourceRange) *SourceRange {
-	return &SourceRange{name, source, begin, end, next, nil}
+func NewSourceRange(name, source string, begin, end int) *SourceRange {
+	return &SourceRange{name, source, begin, end, nil}
 }
 
 // rangePprintInfo is information about the source range that are needed for
diff -pruN 0.11+ds1-5/util/source_range_test.go 0.12+ds1-1/util/source_range_test.go
--- 0.11+ds1-5/util/source_range_test.go	2018-01-18 02:03:11.000000000 +0000
+++ 0.12+ds1-1/util/source_range_test.go	2018-07-11 02:32:25.000000000 +0000
@@ -62,5 +62,5 @@ func parseSourceRange(s, starter, ender
 	if endAfter {
 		end += len(ender)
 	}
-	return NewSourceRange("[test]", s, strings.Index(s, starter), end, nil)
+	return NewSourceRange("[test]", s, strings.Index(s, starter), end)
 }
diff -pruN 0.11+ds1-5/.vsts.sh 0.12+ds1-1/.vsts.sh
--- 0.11+ds1-5/.vsts.sh	2018-01-18 02:03:11.000000000 +0000
+++ 0.12+ds1-1/.vsts.sh	2018-07-11 02:32:25.000000000 +0000
@@ -1,10 +1,10 @@
-mkdir ~/bin ~/go
-export PATH=$HOME/bin:$PATH
-
-curl -sL -o ~/bin/gimme https://raw.githubusercontent.com/travis-ci/gimme/master/gimme
-chmod +x ~/bin/gimme
-eval "$(gimme 1.9)"
+curl -sL -o gimme https://raw.githubusercontent.com/travis-ci/gimme/master/gimme
+chmod +x gimme
+eval "$(./gimme 1.9)"
 
 export GOPATH=$HOME/go
-go get github.com/elves/elvish
-go test github.com/elves/elvish/...
+SRCDIR=$GOPATH/src/github.com/elves/elvish
+mkdir -p $(dirname $SRCDIR)
+cp -r $PWD $SRCDIR
+cd $SRCDIR
+make test
diff -pruN 0.11+ds1-5/website/assets/elvish.css 0.12+ds1-1/website/assets/elvish.css
--- 0.11+ds1-5/website/assets/elvish.css	1970-01-01 00:00:00.000000000 +0000
+++ 0.12+ds1-1/website/assets/elvish.css	2018-07-11 02:32:25.000000000 +0000
@@ -0,0 +1,143 @@
+/* elv.sh-specific stylings. */
+
+a {
+  color: #0645ad;
+}
+
+a:visited {
+  color: #0b0080;
+}
+
+/* Used to several places. */
+.no-display {
+  display: none !important;
+}
+
+#blog-title {
+  cursor: default;
+}
+
+#blog-title a {
+  color: white;
+  text-decoration: none;
+  cursor: default;
+}
+
+#blog-title a.command {
+  color: #5b5;
+}
+
+.toc {
+  background-color: #f0f0f0;
+  padding: 1em;
+  margin: 0 16px 16px 0;
+  border-radius: 6px;
+}
+
+/* The first <h1> clears the TOC */
+.article-content h1 {
+  clear: both;
+}
+
+#toc-list {
+  margin-left: -0.6em;
+}
+
+@media (min-width: 600px) and (max-width: 899px) {
+  #toc-list {
+    column-count: 2;
+  }
+}
+
+@media (min-width: 900px) {
+  #toc-list {
+    column-count: 3;
+  }
+}
+
+
+#toc-list li {
+  list-style: none;
+  /* Keep first-level ToC within one column */
+  break-inside: avoid;
+}
+
+pre.ttyshot {
+  font-size: 12pt;
+  line-height: 1 !important;
+  border: 1px solid black;
+  display: inline-block;
+  margin-bottom: 0 !important;
+}
+
+pre.ttyshot, pre.ttyshot code {
+  background-color: white;
+}
+
+@media screen and (max-width: 600px) {
+  pre.ttyshot {
+    font-size: 2.6vw;
+  }
+}
+
+/* Keys */
+.key {
+  display: inline-block;
+  border: 1px solid black;
+  border-radius: 3px;
+  padding: 0 2px;
+  margin: 1px;
+  font-size: 85%;
+}
+
+/* Code highlights. */
+.sgr-1 {
+  font-weight: bold;
+}
+.sgr-7 {
+  color: white;
+  background-color: black;
+}
+.sgr-31 {
+  color: darkred; /* red in tty */
+}
+.sgr-41 {
+  background-color: darkred; /* red in tty */
+}
+.sgr-32 {
+  color: green; /* green in tty */
+}
+.sgr-42, .sgr-7.sgr-32 {
+  background-color: green; /* green in tty */
+}
+.sgr-33 {
+  color: goldenrod; /* yellow in tty */
+}
+.sgr-43, .sgr-7.sgr-33 {
+  background-color: goldenrod; /* yellow in tty */
+}
+.sgr-34 {
+  color: blue;
+}
+.sgr-44, .sgr-7.sgr-34 {
+  color: white; /* Hacky hacky, just to make the nav ttyshot work */
+  background-color: blue;
+}
+.sgr-35 {
+  color: darkorchid; /* magenta in tty */
+}
+.sgr-45, .sgr-7.sgr-35 {
+  background-color: darkorchid; /* magenta in tty */
+}
+.sgr-36 {
+  color: darkcyan; /* cyan in tty */
+}
+.sgr-46, .sgr-7.sgr-36 {
+  background-color: darkcyan; /* cyan in tty */
+}
+.sgr-37 {
+  color: lightgray;
+}
+.sgr-47, .sgr-7.sgr-37 {
+  background-color: gray;
+}
diff -pruN 0.11+ds1-5/website/assets/genblog.css 0.12+ds1-1/website/assets/genblog.css
--- 0.11+ds1-5/website/assets/genblog.css	1970-01-01 00:00:00.000000000 +0000
+++ 0.12+ds1-1/website/assets/genblog.css	2018-07-11 02:32:25.000000000 +0000
@@ -0,0 +1,289 @@
+/*
+Copied from:  http://code.google.com/p/html5resetcss/
+
+html5doctor.com Reset Stylesheet
+v1.6.1
+Last Updated: 2010-09-17
+Author: Richard Clark - http: //richclarkdesign.com
+Twitter: @rich_clark
+*/
+html, body, div, span, object, iframe, h1, h2, h3, h4, h5, h6, p,
+blockquote, pre, abbr, address, cite, code, del, dfn, em, img, ins,
+kbd, q, samp, small, strong, sub, sup, var, b, i, dl, dt, dd, ol, ul, li,
+fieldset, form, label, legend, table, caption, tbody, tfoot, thead, tr, th, td,
+article, aside, canvas, details, figcaption, figure, footer, header, hgroup,
+menu, nav, section, summary, time, mark, audio,
+video { margin: 0; padding: 0; border: 0; outline: 0;
+            font-size: 100%; vertical-align: baseline; background: transparent; }
+body { line-height: 1; }
+article, aside, details, figcaption, figure, footer, header, hgroup, menu, nav,
+section { display: block; }
+nav ul { list-style: none; }
+blockquote, q { quotes: none; }
+blockquote:before, blockquote:after, q:before, q:after { content: ''; content: none; }
+a { margin: 0; padding: 0; font-size: 100%; vertical-align: baseline; background: transparent; }
+ins { background-color: #ff9; color: #000; text-decoration: none; }
+mark { background-color: #ff9; color: #000; font-style: italic; font-weight: bold; }
+del { text-decoration: line-through; }
+abbr[title], dfn[title] { border-bottom: 1px dotted; cursor: help; }
+table { border-collapse: collapse; border-spacing: 0; }
+hr { display: block; height: 1px; border: 0; border-top: 1px solid #cccccc; margin: 1em 0; padding: 0; }
+input, select { vertical-align: middle; }
+/* end HTML5 reset */
+/* Global styling. */
+
+html {
+  /* Prevent font scaling in landscape while allowing user zoom */
+  -webkit-text-size-adjust: 100%;
+}
+
+body {
+    font-family: "Helvetica Neue", Helvetica, "Segoe UI", Arial, freesans, sans-serif, "Apple Color Emoji", "Segoe UI Emoji", "Segoe UI Symbol", "Microsoft Yahei", FZHei-B01, "WenQuanYi Micro Hei", SimHei;
+    font-size: 16px;
+    background-color: white;
+}
+
+/*
+#blog-title, h1, h2, h3 {
+    font-family: Georgia, Times, Serif, FZShuSong-Z01, SimSun;
+}
+*/
+
+/* Layout primitives. */
+
+#navbar-container {
+    top: 0;
+    width: 1280px;
+    max-width: 100%;
+    margin: auto;
+    color: white;
+    background-color: #1a1a1a;
+}
+
+.card {
+    background-color: white;
+}
+
+.card, #navbar {
+    margin: 0 auto;
+    max-width: 1024px;
+}
+
+.nav-link, .nav-link:visited {
+    color: black;
+    display: inline;
+    text-decoration: none;
+    line-height: 1.4em;
+}
+
+.clear {
+    clear: both;
+}
+
+img {
+    max-width: 100%;
+}
+
+a {
+    color: #6666ff;
+}
+
+/* Global header card. */
+
+#navbar {
+    padding-bottom: 16px;
+}
+
+#blog-title {
+    font-size: 2em;
+    color: white;
+    display: inline-block;
+    margin: 16px 2% 0 4%;
+}
+
+.card-splitter {
+    height: 1px;
+    width: 100%;
+    background-color: #667;
+}
+
+#category-list {
+    display: inline-block;
+    /* The top margin is used when the category list is not on the same line
+     * as the title. Otherwise it is not important. */
+    margin-top: 16px;
+    /* Used (together with margin-right of #blog-title) to separate category
+     * list and blog title when they are on the same line. Otherwise, used as
+     * spacing from the page edge. This needs to be the same as that of
+     * #blog-title to keep them left-aligned. */
+    margin-left: 2%;
+}
+
+#category-list > li {
+    list-style: none;
+    display: inline-block;
+    /* This margin is important for the correct spacing when the category list
+     * runs multiple lines. */
+    margin-top: 10px;
+}
+
+#category-list > li > .nav-link {
+    color: white;
+    position: relative;
+    top: -4px; /* Cancel padding-bottom to align with baseline. */
+    padding: 4px 10px 4px;
+}
+
+#category-list > li > .nav-link:hover {
+    background-color: #333;
+}
+
+#category-list > li.current > .nav-link {
+    color: black;
+    background-color: white;
+}
+
+/* Article content. */
+
+.timestamp {
+    margin-bottom: 0.6em;
+}
+
+.article-header {
+    padding: 2.4% 4% 1.6%;
+}
+
+.article-content {
+    padding: 4%;
+}
+
+.article p, .article ul, .article pre {
+    margin-bottom: 16px;
+}
+
+.article p {
+    line-height: 1.5;
+}
+
+.article li {
+    margin-top: 0.5em;
+    margin-bottom: 0.5em;
+}
+
+.article li > p {
+    margin-top: 1em;
+    margin-bottom: 1em;
+}
+
+.article pre {
+    padding: 1em;
+    line-height: 1.45;
+    overflow: auto;
+}
+
+.article p code {
+    padding: 0.2em 0;
+}
+
+.article p code::before, .article p code::after {
+    letter-spacing: -0.2em;
+    content: "\00a0";
+}
+
+code, pre {
+    font-family: Consolas, Menlo, "Bitstream Vera Sans Mono", "DejaVu Sans Mono", monospace;
+}
+
+.article code, .article pre {
+    background-color: #f0f0f0;
+    border-radius: 3px;
+}
+
+.article code {
+    font-size: 85%;
+}
+
+/* Use headers up to h2 should be used in .article-content */
+.article h1, .article h2 {
+    line-height: 1.25;
+}
+
+.article .article-header h1 {
+    font-size: 2em;
+    margin: 0;
+    padding: 0;
+    border: none;
+}
+
+.article h1, .article h2 {
+    margin-top: 24px;
+    margin-bottom: 16px;
+    font-weight: 600;
+}
+
+.article h1 {
+    font-size: 1.5em;
+    padding-bottom: 0.3em;
+    border-bottom: 1px solid #eee;
+}
+
+.article h2 {
+    font-size: 1.25em;
+}
+
+.article ul, .article ol {
+    margin-left: 1em;
+}
+
+/* Category content. */
+
+.category-prelude {
+    padding: 4% 5.8% 0;
+    margin-bottom: -20px;
+}
+
+.article-list {
+    padding: 4%;
+}
+
+.article-list > li {
+    list-style: square inside;
+    padding: 3px;
+    line-height: 1.4;
+}
+
+.article-list > li:hover {
+    background-color: #c0c0c0;
+}
+
+.article-list > li > .nav-link {
+    border-bottom: 1px solid black;
+}
+
+.article-timestamp {
+    float: right;
+    display: inline-block;
+    margin-left: 1em;
+}
+
+/* Utilities usable in articles. */
+
+hr {
+    clear: both;
+    border-color: #aaa;
+    text-align: center;
+}
+
+hr:after {
+    content: "❧";
+    text-shadow: 0px 0px 2px #667;
+    display: inline-block;
+    position: relative;
+    top: -0.5em;
+    padding: 0 0.25em;
+    font-size: 1.1em;
+    color: black;
+    background-color: white;
+}
+
+/* vi: se ts=4 sts=4 sw=4: */
diff -pruN 0.11+ds1-5/website/assets/home-demos.css 0.12+ds1-1/website/assets/home-demos.css
--- 0.11+ds1-5/website/assets/home-demos.css	1970-01-01 00:00:00.000000000 +0000
+++ 0.12+ds1-1/website/assets/home-demos.css	2018-07-11 02:32:25.000000000 +0000
@@ -0,0 +1,111 @@
+#demo-window {
+  background-color: #f0ede2;
+  margin-bottom: 16px;
+  border-radius: 5px;
+  /* Graceful degrading: the demo window starts with a scrollbar, which gets
+   * hiden only when the JavaScript code finishes adding demo switchers. */
+  overflow: scroll;
+}
+
+.overflow-hidden {
+  overflow: hidden !important;
+}
+
+.animated-transition {
+  transition: all 300ms cubic-bezier(0.165, 0.84, 0.44, 1);
+}
+
+#demo-container {
+  transform: translateX(0);
+  width: 500%;
+}
+
+.demo-wrapper {
+  width: 20%;
+  float: left;
+}
+
+/* Expanded view. */
+#demo-container.expanded {
+  width: 100%;
+}
+#demo-container.expanded .demo-wrapper {
+  width: 100%;
+}
+
+.demo:after {
+  content: "";
+  display: table;
+  clear: both;
+}
+
+.demo {
+  padding: 16px;
+}
+
+.demo-col {
+  width: 100%;
+  height: 100%;
+  float: left;
+}
+
+.demo-col .demo-ttyshot {
+  text-align: center;
+}
+
+.demo-col .demo-ttyshot .ttyshot {
+  text-align: left;
+}
+
+.demo-description {
+  margin: 16px 5% 0;
+}
+
+@media screen and (min-width: 1024px) {
+  .demo-col.left {
+    width: 60%;
+  }
+  .demo-col.right {
+    width: 40%;
+  }
+  .demo-description {
+    margin: 0 5%;
+  }
+  .animated-transition {
+    transition-duration: 400ms;
+  }
+}
+
+.demo-title {
+  font-size: 1.3em;
+}
+
+.demo-col p {
+  margin-top: 16px;
+  margin-bottom: 0;
+}
+
+ul#demo-switcher {
+  display: inline;
+  margin: 0;
+}
+
+ul#demo-switcher > li {
+  list-style: none;
+  display: inline-block;
+}
+
+ul#demo-switcher > li > a {
+  color: black;
+  padding: 4px 14px;
+}
+
+ul#demo-switcher > li > a.current , ul#demo-switcher > li > a.current:hover {
+  color: white;
+  background-color: black;
+}
+
+ul#demo-switcher > li > a:hover {
+  background-color: #ccc;
+  cursor: pointer;
+}
diff -pruN 0.11+ds1-5/website/assets/home-demos.js 0.12+ds1-1/website/assets/home-demos.js
--- 0.11+ds1-5/website/assets/home-demos.js	1970-01-01 00:00:00.000000000 +0000
+++ 0.12+ds1-1/website/assets/home-demos.js	2018-07-11 02:32:25.000000000 +0000
@@ -0,0 +1,184 @@
+(function() {
+  var current = 0,
+      expanded = true;
+      expander = document.getElementById('demo-expander'),
+      demoWindow = document.getElementById("demo-window"),
+      demoContainer = document.getElementById("demo-container"),
+      demoSwitcher = document.getElementById("demo-switcher"),
+      demoWrappers = document.getElementsByClassName("demo-wrapper"),
+      nDemos = demoWrappers.length,
+      switcherLinks = [];
+
+  var scrollTo = function(to, instant) {
+    if (expanded) {
+      return;
+    }
+    if (current != null) {
+      switcherLinks[current].className = "";
+    }
+    var translate = -demoWrappers[0].offsetWidth * to; 
+    demoContainer.className = instant ? "" : "animated-transition";
+    demoContainer.style.transform = "translateX(" + translate + "px)";
+    switcherLinks[to].className = "current";
+    current = to;
+  };
+  var scrollToNext = function() {
+    scrollTo(current < nDemos - 1 ? current + 1 : current);
+  };
+  var scrollToPrev = function() {
+    scrollTo(current > 0 ? current - 1 : current);
+  };
+  function scrollToCurrent() {
+    scrollTo(current);
+  }
+
+  function expand() {
+    expanded = true;
+    expander.className = "current";
+    switcherLinks[current].className = "";
+    demoContainer.className = "expanded";
+    demoContainer.style.transform = "";
+  }
+  function collapse() {
+    switcherLinks[current].className = "current";
+    expander.className = "";
+    demoContainer.className = "";
+  }
+  expander.onclick = expand;
+
+  /* Build demo switcher. */
+  for (var i = 0; i < nDemos; i++) {
+    var li = document.createElement("li"),
+        link = document.createElement("a");
+    link.textContent = i + 1;
+    link.onclick = (function(to) { return function() {
+      if (expanded) {
+        expanded = false;
+        collapse();
+      }
+      scrollTo(to);
+    }; })(i);
+    if (i == 0) {
+      link.className = "current";
+    }
+    switcherLinks.push(link);
+    li.appendChild(link);
+    demoSwitcher.appendChild(li);
+  }
+
+  /* Switcher built. Hide the warning text, show the expand button, unexpand
+   * and hide scrollbar. */
+  document.getElementById('demo-js-warning').className = "no-display";
+  document.getElementById('demo-expander-li').className = "";
+  demoContainer.className = "";
+  demoWindow.className = "overflow-hidden";
+  expanded = false;
+
+  /* Resizing breaks sliding, fix it. */
+  window.addEventListener('resize', function() { scrollTo(current, true); });
+
+  /* Scrolling primitives. */
+  var scrollXTrigger = 5, scrollYTrigger = 5;
+  var scrollX = false, scrollY = false;
+  var offsetX = 0, offsetY = 0, baseOffset = 0;
+  function handleScroll(ev) {
+    if (!scrollX && !scrollY) {
+      if (Math.abs(offsetX) > scrollXTrigger) {
+        baseOffset = offsetX;
+        scrollX = true;
+      } else if (Math.abs(offsetY) > scrollYTrigger) {
+        baseOffset = offsetY;
+        scrollY = true;
+      }
+    }
+    if (!scrollX) {
+      return;
+    }
+    // No overscrolling.
+    var calculatedOffset = offsetX - baseOffset;
+    if ((current == 0 && calculatedOffset > 0) ||
+        (current == nDemos-1 && calculatedOffset < 0)) {
+      calculatedOffset = 0;
+    }
+    var translate = calculatedOffset - demoWrappers[0].offsetWidth * current;
+    demoContainer.style.transform = "translateX(" + translate + "px)";
+    ev.preventDefault();
+  }
+  function settleScroll() {
+    if (scrollX) {
+      var threshold = Math.min(60, demoWindow.offsetWidth / 4);
+      if (offsetX < -threshold) {
+        scrollToNext();
+      } else if (offsetX > threshold) {
+        scrollToPrev();
+      } else {
+        scrollToCurrent();
+      }
+    }
+    offsetX = offsetY = baseOffset = 0;
+    scrollX = scrollY = false;
+  }
+
+  /* Support scrolling by touch. */
+  var initX, initY;
+  demoWindow.addEventListener('touchstart', function(ev) {
+    if (expanded) {
+      return;
+    }
+    initX = ev.touches[0].clientX;
+    initY = ev.touches[0].clientY;
+    demoContainer.className = "";
+  });
+  demoWindow.addEventListener('touchmove', function(ev) {
+    if (expanded) {
+      return;
+    }
+    if (ev.touches.length == 1) {
+      var lastX = ev.touches[0].clientX;
+      var lastY = ev.touches[0].clientY;
+      offsetX = lastX - initX;
+      offsetY = lastY - initY;
+      handleScroll(ev);
+      // document.getElementById('demo-debug').innerText = '(' + offsetX + ', ' + offsetY + '), ' + scrollX + ', ' + scrollY;
+    }
+  });
+  demoWindow.addEventListener('touchcancel', function() {
+    if (expanded) {
+      return;
+    }
+    scrollTo(current);
+  });
+  demoWindow.addEventListener('touchend', function() {
+    if (expanded) {
+      return;
+    }
+    settleScroll();
+  });
+
+  // Keyboard bindings.
+  window.addEventListener('keypress', function(ev) {
+    var char = String.fromCodePoint(ev.keyCode || ev.charCode);
+    if (char == 'h') {
+      scrollToPrev();
+    } else if (char == 'l') {
+      scrollToNext();
+    } else if (char == 'o') {
+      expanded = !expanded;
+      if (expanded) {
+        expand();
+      } else {
+        collapse();
+        scrollToCurrent();
+      }
+    } else {
+      var i = parseInt(char);
+      if (1 <= i && i <= nDemos) {
+        if (expanded) {
+          expanded = false;
+          collapse();
+        }
+        scrollTo(i-1);
+      }
+    }
+  });
+})();
diff -pruN 0.11+ds1-5/website/assets/logo.svg 0.12+ds1-1/website/assets/logo.svg
--- 0.11+ds1-5/website/assets/logo.svg	1970-01-01 00:00:00.000000000 +0000
+++ 0.12+ds1-1/website/assets/logo.svg	2018-07-11 02:32:25.000000000 +0000
@@ -0,0 +1,107 @@
+<?xml version="1.0" encoding="UTF-8" standalone="no"?>
+<!-- Created with Inkscape (http://www.inkscape.org/) -->
+
+<svg
+   xmlns:dc="http://purl.org/dc/elements/1.1/"
+   xmlns:cc="http://creativecommons.org/ns#"
+   xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#"
+   xmlns:svg="http://www.w3.org/2000/svg"
+   xmlns="http://www.w3.org/2000/svg"
+   xmlns:sodipodi="http://sodipodi.sourceforge.net/DTD/sodipodi-0.dtd"
+   xmlns:inkscape="http://www.inkscape.org/namespaces/inkscape"
+   width="256"
+   height="64"
+   viewBox="0 0 67.733336 16.933333"
+   version="1.1"
+   id="svg8"
+   inkscape:version="0.92.1 r"
+   sodipodi:docname="logo.svg">
+  <defs
+     id="defs2" />
+  <sodipodi:namedview
+     id="base"
+     pagecolor="#ffffff"
+     bordercolor="#666666"
+     borderopacity="1.0"
+     inkscape:pageopacity="0"
+     inkscape:pageshadow="2"
+     inkscape:zoom="2"
+     inkscape:cx="117.4042"
+     inkscape:cy="-7.1611659"
+     inkscape:document-units="mm"
+     inkscape:current-layer="layer1"
+     showgrid="false"
+     fit-margin-top="20"
+     fit-margin-left="20"
+     fit-margin-right="20"
+     fit-margin-bottom="20"
+     units="px"
+     inkscape:window-width="1149"
+     inkscape:window-height="703"
+     inkscape:window-x="0"
+     inkscape:window-y="23"
+     inkscape:window-maximized="0" />
+  <metadata
+     id="metadata5">
+    <rdf:RDF>
+      <cc:Work
+         rdf:about="">
+        <dc:format>image/svg+xml</dc:format>
+        <dc:type
+           rdf:resource="http://purl.org/dc/dcmitype/StillImage" />
+        <dc:title></dc:title>
+      </cc:Work>
+    </rdf:RDF>
+  </metadata>
+  <g
+     inkscape:label="Layer 1"
+     inkscape:groupmode="layer"
+     id="layer1"
+     transform="translate(-24.796419,-91.142756)">
+    <rect
+       style="fill:#000000;fill-opacity:1;stroke-width:0.26458332"
+       id="rect5161"
+       width="67.733414"
+       height="16.93334"
+       x="24.796419"
+       y="91.142754" />
+    <g
+       aria-label="~&gt; elvish"
+       transform="matrix(0.26458333,0,0,0.26458333,0.09146774,0.94877938)"
+       style="font-style:normal;font-weight:normal;font-size:40px;line-height:1.25;font-family:sans-serif;letter-spacing:0px;word-spacing:0px;fill:#000000;fill-opacity:1;stroke:none"
+       id="flowRoot3680">
+      <path
+         d="m 134.36328,372.67578 v 3.39844 q -1.46484,1.15234 -2.87109,1.69922 -1.38672,0.52734 -2.92969,0.52734 -1.75781,0 -3.96484,-0.99609 -0.42969,-0.19532 -0.64453,-0.27344 -1.50391,-0.64453 -2.51954,-0.87891 -0.99609,-0.23437 -1.99218,-0.23437 -1.54297,0 -2.92969,0.58594 -1.36719,0.5664 -2.79297,1.79687 v -3.39844 q 1.52344,-1.17187 2.94922,-1.69922 1.42578,-0.52734 3.04687,-0.52734 1.03516,0 2.01172,0.21484 0.97657,0.21485 2.42188,0.82032 0.23437,0.0976 0.64453,0.29297 2.26562,1.07421 4.08203,1.07421 1.36719,0 2.69531,-0.58593 1.34766,-0.60547 2.79297,-1.81641 z"
+         style="font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;font-family:'DejaVu Sans Mono';-inkscape-font-specification:'DejaVu Sans Mono';fill:#ffffff;fill-opacity:1"
+         id="path6427" />
+      <path
+         d="m 137.78125,369.14062 v -3.55468 l 20.64453,8.16406 v 3.24219 l -20.64453,8.16406 v -3.57422 l 16.60156,-6.19141 z"
+         style="font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;font-family:'DejaVu Sans Mono';-inkscape-font-specification:'DejaVu Sans Mono';fill:#ffffff;fill-opacity:1"
+         id="path6429" />
+      <path
+         d="m 205.90625,376.07422 v 1.75781 h -15.56641 v 0.11719 q 0,3.57422 1.85547,5.52734 1.875,1.95313 5.27344,1.95313 1.71875,0 3.59375,-0.54688 1.875,-0.54687 4.00391,-1.66015 v 3.57421 q -2.05079,0.83985 -3.96485,1.25 -1.89453,0.42969 -3.67187,0.42969 -5.09766,0 -7.96875,-3.04687 -2.8711,-3.06641 -2.8711,-8.4375 0,-5.23438 2.8125,-8.35938 2.8125,-3.125 7.5,-3.125 4.17969,0 6.58203,2.83203 2.42188,2.83203 2.42188,7.73438 z m -3.59375,-1.05469 q -0.0781,-3.16406 -1.50391,-4.80469 -1.40625,-1.66015 -4.0625,-1.66015 -2.59765,0 -4.27734,1.71875 -1.67969,1.71875 -1.99219,4.76562 z"
+         style="font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;font-family:'DejaVu Sans Mono';-inkscape-font-specification:'DejaVu Sans Mono';fill:#55bb55;fill-opacity:1"
+         id="path6431" />
+      <path
+         d="m 220.73047,379.98047 q 0,2.42187 0.8789,3.65234 0.89844,1.23047 2.63672,1.23047 h 4.19922 v 3.04688 h -4.55078 q -3.22266,0 -5,-2.07032 -1.75781,-2.07031 -1.75781,-5.85937 V 360.11719 H 211.375 v -2.8125 h 9.35547 z"
+         style="font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;font-family:'DejaVu Sans Mono';-inkscape-font-specification:'DejaVu Sans Mono';fill:#55bb55;fill-opacity:1"
+         id="path6433" />
+      <path
+         d="m 234.26562,366.03516 h 3.73047 l 6.34766,18.35937 6.36719,-18.35937 h 3.73047 l -7.77344,21.875 h -4.62891 z"
+         style="font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;font-family:'DejaVu Sans Mono';-inkscape-font-specification:'DejaVu Sans Mono';fill:#55bb55;fill-opacity:1"
+         id="path6435" />
+      <path
+         d="m 261.375,366.03516 h 9.19922 v 19.08203 h 7.1289 v 2.79297 h -17.85156 v -2.79297 h 7.12891 V 368.82812 H 261.375 Z m 5.60547,-8.51563 h 3.59375 v 4.55078 h -3.59375 z"
+         style="font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;font-family:'DejaVu Sans Mono';-inkscape-font-specification:'DejaVu Sans Mono';fill:#55bb55;fill-opacity:1"
+         id="path6437" />
+      <path
+         d="m 299.44141,366.79687 v 3.51563 q -1.54297,-0.89844 -3.10547,-1.34766 -1.5625,-0.44922 -3.1836,-0.44922 -2.4414,0 -3.65234,0.80079 -1.19141,0.78125 -1.19141,2.40234 0,1.46484 0.89844,2.1875 0.89844,0.72266 4.47266,1.40625 l 1.44531,0.27344 q 2.67578,0.50781 4.04297,2.03125 1.38672,1.52343 1.38672,3.96484 0,3.24219 -2.30469,5.07813 -2.30469,1.8164 -6.40625,1.8164 -1.62109,0 -3.39844,-0.35156 -1.77734,-0.33203 -3.84765,-1.01563 v -3.71093 q 2.01171,1.03515 3.84765,1.5625 1.83594,0.50781 3.47656,0.50781 2.38282,0 3.69141,-0.95703 1.30859,-0.97656 1.30859,-2.71485 0,-2.5 -4.78515,-3.45703 l -0.15625,-0.0391 -1.34766,-0.27344 q -3.10547,-0.60547 -4.53125,-2.03125 -1.42578,-1.44531 -1.42578,-3.92578 0,-3.14453 2.12891,-4.84375 2.1289,-1.71875 6.07422,-1.71875 1.75781,0 3.3789,0.33203 1.6211,0.3125 3.1836,0.95703 z"
+         style="font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;font-family:'DejaVu Sans Mono';-inkscape-font-specification:'DejaVu Sans Mono';fill:#55bb55;fill-opacity:1"
+         id="path6439" />
+      <path
+         d="m 325.02734,374.35547 v 13.55469 h -3.61328 v -13.55469 q 0,-2.94922 -1.03515,-4.33594 -1.03516,-1.38672 -3.24219,-1.38672 -2.51953,0 -3.88672,1.79688 -1.34766,1.77734 -1.34766,5.11718 v 12.36329 h -3.59375 v -30.39063 h 3.59375 v 11.79688 q 0.95703,-1.875 2.59766,-2.83204 1.64062,-0.97656 3.88672,-0.97656 3.33984,0 4.98047,2.20703 1.66015,2.1875 1.66015,6.64063 z"
+         style="font-style:normal;font-variant:normal;font-weight:normal;font-stretch:normal;font-family:'DejaVu Sans Mono';-inkscape-font-specification:'DejaVu Sans Mono';fill:#55bb55;fill-opacity:1"
+         id="path6441" />
+    </g>
+  </g>
+</svg>
Binary files 0.11+ds1-5/website/favicons/android-chrome-192x192.png and 0.12+ds1-1/website/favicons/android-chrome-192x192.png differ
Binary files 0.11+ds1-5/website/favicons/android-chrome-384x384.png and 0.12+ds1-1/website/favicons/android-chrome-384x384.png differ
Binary files 0.11+ds1-5/website/favicons/apple-touch-icon.png and 0.12+ds1-1/website/favicons/apple-touch-icon.png differ
diff -pruN 0.11+ds1-5/website/favicons/browserconfig.xml 0.12+ds1-1/website/favicons/browserconfig.xml
--- 0.11+ds1-5/website/favicons/browserconfig.xml	1970-01-01 00:00:00.000000000 +0000
+++ 0.12+ds1-1/website/favicons/browserconfig.xml	2018-07-11 02:32:25.000000000 +0000
@@ -0,0 +1,9 @@
+<?xml version="1.0" encoding="utf-8"?>
+<browserconfig>
+    <msapplication>
+        <tile>
+            <square150x150logo src="/mstile-150x150.png"/>
+            <TileColor>#da532c</TileColor>
+        </tile>
+    </msapplication>
+</browserconfig>
Binary files 0.11+ds1-5/website/favicons/favicon-16x16.png and 0.12+ds1-1/website/favicons/favicon-16x16.png differ
Binary files 0.11+ds1-5/website/favicons/favicon-32x32.png and 0.12+ds1-1/website/favicons/favicon-32x32.png differ
Binary files 0.11+ds1-5/website/favicons/favicon.ico and 0.12+ds1-1/website/favicons/favicon.ico differ
diff -pruN 0.11+ds1-5/website/favicons/manifest.json 0.12+ds1-1/website/favicons/manifest.json
--- 0.11+ds1-5/website/favicons/manifest.json	1970-01-01 00:00:00.000000000 +0000
+++ 0.12+ds1-1/website/favicons/manifest.json	2018-07-11 02:32:25.000000000 +0000
@@ -0,0 +1,18 @@
+{
+    "name": "",
+    "icons": [
+        {
+            "src": "/android-chrome-192x192.png",
+            "sizes": "192x192",
+            "type": "image/png"
+        },
+        {
+            "src": "/android-chrome-384x384.png",
+            "sizes": "384x384",
+            "type": "image/png"
+        }
+    ],
+    "theme_color": "#ffffff",
+    "background_color": "#ffffff",
+    "display": "standalone"
+}
\ No newline at end of file
Binary files 0.11+ds1-5/website/favicons/mstile-150x150.png and 0.12+ds1-1/website/favicons/mstile-150x150.png differ
diff -pruN 0.11+ds1-5/website/favicons/safari-pinned-tab.svg 0.12+ds1-1/website/favicons/safari-pinned-tab.svg
--- 0.11+ds1-5/website/favicons/safari-pinned-tab.svg	1970-01-01 00:00:00.000000000 +0000
+++ 0.12+ds1-1/website/favicons/safari-pinned-tab.svg	2018-07-11 02:32:25.000000000 +0000
@@ -0,0 +1,25 @@
+<?xml version="1.0" standalone="no"?>
+<!DOCTYPE svg PUBLIC "-//W3C//DTD SVG 20010904//EN"
+ "http://www.w3.org/TR/2001/REC-SVG-20010904/DTD/svg10.dtd">
+<svg version="1.0" xmlns="http://www.w3.org/2000/svg"
+ width="480.000000pt" height="480.000000pt" viewBox="0 0 480.000000 480.000000"
+ preserveAspectRatio="xMidYMid meet">
+<metadata>
+Created by potrace 1.11, written by Peter Selinger 2001-2013
+</metadata>
+<g transform="translate(0.000000,480.000000) scale(0.100000,-0.100000)"
+fill="#000000" stroke="none">
+<path d="M2165 4789 c-254 -24 -579 -117 -819 -234 -232 -113 -504 -310 -672
+-487 -626 -660 -832 -1573 -548 -2427 239 -718 797 -1276 1514 -1515 1022
+-340 2128 31 2738 917 560 814 560 1900 0 2714 -497 722 -1337 1114 -2213
+1032z m589 -950 c201 -41 367 -129 506 -269 182 -181 291 -402 345 -700 12
+-65 19 -169 22 -332 l5 -238 -993 -2 -994 -3 2 -97 c6 -364 167 -646 448 -783
+149 -73 235 -90 465 -89 164 0 212 4 300 23 185 40 409 120 592 212 l78 39 0
+-225 0 -226 -33 -14 c-85 -35 -279 -97 -405 -129 -326 -82 -632 -94 -907 -37
+-577 121 -938 551 -1006 1200 -15 142 -6 464 15 580 74 398 249 694 532 897
+166 119 378 192 639 218 82 8 293 -5 389 -25z"/>
+<path d="M2285 3466 c-300 -70 -511 -293 -594 -627 -11 -46 -23 -108 -27 -136
+l-7 -53 763 0 762 0 -6 78 c-22 252 -89 430 -209 559 -71 76 -145 122 -252
+159 -70 24 -98 28 -225 30 -99 3 -164 -1 -205 -10z"/>
+</g>
+</svg>
diff -pruN 0.11+ds1-5/website/.gitignore 0.12+ds1-1/website/.gitignore
--- 0.11+ds1-5/website/.gitignore	1970-01-01 00:00:00.000000000 +0000
+++ 0.12+ds1-1/website/.gitignore	2018-07-11 02:32:25.000000000 +0000
@@ -0,0 +1,5 @@
+/src/*.html
+!/src/template.html
+/src/*/*.html
+/dst
+/*.tool
diff -pruN 0.11+ds1-5/website/highlighter.go 0.12+ds1-1/website/highlighter.go
--- 0.11+ds1-5/website/highlighter.go	1970-01-01 00:00:00.000000000 +0000
+++ 0.12+ds1-1/website/highlighter.go	2018-07-11 02:32:25.000000000 +0000
@@ -0,0 +1,150 @@
+package main
+
+import (
+	"bufio"
+	"bytes"
+	"fmt"
+	"html"
+	"log"
+	"os"
+	"strings"
+
+	"github.com/elves/elvish/edit/highlight"
+	"github.com/elves/elvish/parse"
+)
+
+var (
+	scanner = bufio.NewScanner(os.Stdin)
+	lineno  = 0
+)
+
+func scan() bool {
+	lineno++
+	return scanner.Scan()
+}
+
+func main() {
+	for scan() {
+		line := scanner.Text()
+		trimmed := strings.TrimLeft(line, " ")
+		indent := line[:len(line)-len(trimmed)]
+		if trimmed == "```elvish" || trimmed == "```elvish-bad" {
+			bad := trimmed == "```elvish-bad"
+			highlighted := convert(collectFenced(indent), bad)
+			fmt.Printf("%s<pre><code>%s</code></pre>\n", indent, highlighted)
+		} else if trimmed == "```elvish-transcript" {
+			highlighted := convertTranscript(collectFenced(indent))
+			fmt.Printf("%s<pre><code>%s</code></pre>\n", indent, highlighted)
+		} else {
+			fmt.Println(line)
+		}
+	}
+}
+
+func collectFenced(indent string) string {
+	var buf bytes.Buffer
+	for scan() {
+		line := scanner.Text()
+		if !strings.HasPrefix(line, indent) {
+			log.Fatalf("bad indent of line %d: %q", lineno, line)
+		}
+		unindented := line[len(indent):]
+		if unindented == "```" {
+			break
+		}
+		if buf.Len() > 0 {
+			buf.WriteRune('\n')
+		}
+		buf.WriteString(unindented)
+	}
+	return buf.String()
+}
+
+const (
+	ps1 = "~> "
+	ps2 = "   "
+)
+
+func convertTranscript(transcript string) string {
+	scanner := bufio.NewScanner(bytes.NewBufferString(transcript))
+	var buf bytes.Buffer
+
+	overread := false
+	var line string
+	for overread || scanner.Scan() {
+		if overread {
+			overread = false
+		} else {
+			line = scanner.Text()
+		}
+		if strings.HasPrefix(line, ps1) {
+			elvishBuf := bytes.NewBufferString(line[len(ps1):] + "\n")
+			for scanner.Scan() {
+				line = scanner.Text()
+				if strings.HasPrefix(line, ps2) {
+					elvishBuf.WriteString(line + "\n")
+				} else {
+					overread = true
+					break
+				}
+			}
+			buf.WriteString(html.EscapeString(ps1))
+			buf.WriteString(convert(elvishBuf.String(), false))
+		} else {
+			buf.WriteString(html.EscapeString(line))
+			buf.WriteString("<br>")
+		}
+	}
+	return buf.String()
+}
+
+func convert(text string, bad bool) string {
+	n, err := parse.Parse("highlight", text)
+	if err != nil && !bad {
+		log.Printf("parsing %q: %v", text, err)
+	}
+
+	styling := highlight.Styling{}
+
+	e := highlight.Emitter{
+		func(s string) bool {
+			return true
+		},
+		styling.Add,
+	}
+	e.EmitAll(n)
+
+	var buf bytes.Buffer
+
+	openedSpan := false
+	currentStyle := ""
+
+	applier := styling.Apply()
+	for i, r := range text {
+		applier.At(i)
+		style := applier.Get()
+		if style != currentStyle {
+			if openedSpan {
+				buf.WriteString("</span>")
+			}
+			var classes []string
+			for _, c := range strings.Split(style, ";") {
+				if c != "" {
+					classes = append(classes, "sgr-"+c)
+				}
+			}
+			fmt.Fprintf(&buf, `<span class="%s">`, strings.Join(classes, " "))
+			openedSpan = true
+			currentStyle = style
+		}
+		if r == '\n' {
+			buf.WriteString("<br>")
+		} else {
+			buf.WriteString(html.EscapeString(string(r)))
+		}
+	}
+	if openedSpan {
+		buf.WriteString("</span>")
+	}
+	return buf.String()
+}
diff -pruN 0.11+ds1-5/website/macros.go 0.12+ds1-1/website/macros.go
--- 0.11+ds1-5/website/macros.go	1970-01-01 00:00:00.000000000 +0000
+++ 0.12+ds1-1/website/macros.go	2018-07-11 02:32:25.000000000 +0000
@@ -0,0 +1,86 @@
+package main
+
+import (
+	"bufio"
+	"bytes"
+	"fmt"
+	"io/ioutil"
+	"log"
+	"os"
+	"path"
+	"strings"
+)
+
+const (
+	ttyshot = "$ttyshot "
+	cf      = "$cf "
+	dl      = "$dl "
+)
+
+func main() {
+	scanner := bufio.NewScanner(os.Stdin)
+	for scanner.Scan() {
+		line := scanner.Text()
+		line = expandTtyshot(line)
+		line = expandCf(line)
+		line = expandDl(line)
+		fmt.Println(line)
+	}
+}
+
+func expandTtyshot(line string) string {
+	i := strings.Index(line, ttyshot)
+	if i < 0 {
+		return line
+	}
+	name := line[i+len(ttyshot):]
+	content, err := ioutil.ReadFile(path.Join("tty", name+".html"))
+	if err != nil {
+		log.Fatal(err)
+	}
+	var buf bytes.Buffer
+	buf.WriteString(line[:i])
+	buf.WriteString(`<pre class="ttyshot"><code>`)
+	buf.Write(bytes.Replace(
+		content, []byte("\n"), []byte("<br>"), -1))
+	buf.WriteString("</code></pre>")
+	return buf.String()
+}
+
+func expandCf(line string) string {
+	i := strings.Index(line, cf)
+	if i < 0 {
+		return line
+	}
+	targets := strings.Split(line[i+len(cf):], " ")
+	var buf bytes.Buffer
+	buf.WriteString("See also")
+	for i, target := range targets {
+		var sep string
+		if i == 0 {
+			sep = " "
+		} else if i == len(targets)-1 {
+			sep = " and "
+		} else {
+			sep = ", "
+		}
+		fmt.Fprintf(&buf, "%s[`%s`](#%s)", sep, target, target)
+	}
+	buf.WriteString(".")
+	return buf.String()
+}
+
+func expandDl(line string) string {
+	i := strings.Index(line, dl)
+	if i < 0 {
+		return line
+	}
+	fields := strings.SplitN(line[i+len(dl):], " ", 2)
+	name := fields[0]
+	url := name
+	if len(fields) == 2 {
+		url = fields[1]
+	}
+	return line[:i] + fmt.Sprintf(
+		`<a href="https://dl.elv.sh/%s">%s</a>`, url, name)
+}
diff -pruN 0.11+ds1-5/website/Makefile 0.12+ds1-1/website/Makefile
--- 0.11+ds1-5/website/Makefile	1970-01-01 00:00:00.000000000 +0000
+++ 0.12+ds1-1/website/Makefile	2018-07-11 02:32:25.000000000 +0000
@@ -0,0 +1,40 @@
+PUBLISH_DIR := ./publish
+
+MDS := $(wildcard src/*.md src/*/*.md)
+HTMLS := $(MDS:.md=.html)
+
+TOOL_SOURCES := $(wildcard *.go)
+TOOL_BINARIES := $(TOOL_SOURCES:.go=.tool)
+
+default: gen
+
+%.html: %.md $(TOOL_BINARIES) md-to-html
+	./md-to-html $< $@
+
+%.tool: %.go
+	go build -o $@ $<
+
+# Generates the website into the dst directory.
+gen: $(HTMLS) maybe-genblog
+	genblog src dst
+	ln -sf `pwd`/assets dst/
+	ln -sf `pwd`/favicons/* dst/
+
+# Installs genblog only if not found.
+maybe-genblog:
+	which genblog || make genblog
+
+# Installs latest genblog.
+genblog:
+	go get -u github.com/xiaq/genblog
+	genblog -print-default-css > assets/genblog.css
+
+# Copies the generated website into $(PUBLISH_DIR).
+publish: gen
+	rsync -aLv --delete ./dst/ $(PUBLISH_DIR)/
+
+clean:
+	rm -f $(HTMLS) $(TOOL_BINARIES)
+
+.PHONY: default gen maybe-genblog genblog publish clean
+.SECONDARY: $(TOOL_BINARIES)
diff -pruN 0.11+ds1-5/website/md-to-html 0.12+ds1-1/website/md-to-html
--- 0.11+ds1-5/website/md-to-html	1970-01-01 00:00:00.000000000 +0000
+++ 0.12+ds1-1/website/md-to-html	2018-07-11 02:32:25.000000000 +0000
@@ -0,0 +1,21 @@
+#!/bin/sh
+in=$1
+out=$2
+opts=
+
+has() {
+    head -n1 $in | grep "$@" >/dev/null
+}
+
+has toc && {
+    opts="$opts --toc --template=toc-and-body"
+}
+has number-sections && {
+    opts="$opts --number-sections"
+}
+
+mydir=$(dirname "$0")
+
+$mydir/highlighter.tool < $1 |
+  $mydir/macros.tool |
+  pandoc -f gfm+smart --data-dir=$mydir/pandoc-data -o $2 $opts
diff -pruN 0.11+ds1-5/website/pandoc-data/templates/toc-and-body.html 0.12+ds1-1/website/pandoc-data/templates/toc-and-body.html
--- 0.11+ds1-5/website/pandoc-data/templates/toc-and-body.html	1970-01-01 00:00:00.000000000 +0000
+++ 0.12+ds1-1/website/pandoc-data/templates/toc-and-body.html	2018-07-11 02:32:25.000000000 +0000
@@ -0,0 +1,25 @@
+<div class="toc">
+  <p>Table of Content: <a id="toc-toggle" href="">[Hide]</a></p>
+  <div id="toc-list">
+    $toc$
+  </div>
+  <script>
+  (function() {
+    var shown = true,
+        tocToggle = document.getElementById('toc-toggle'),
+        tocList = document.getElementById('toc-list');
+    tocToggle.onclick = function(ev) {
+      shown = !shown;
+      if (shown) {
+        tocToggle.innerText = "[Hide]";
+        tocList.className = "";
+      } else {
+        tocToggle.innerText = "[Show]";
+        tocList.className = "no-display";
+      }
+      ev.preventDefault();
+    };
+  })();
+  </script>
+</div>
+$body$
diff -pruN 0.11+ds1-5/website/README.md 0.12+ds1-1/website/README.md
--- 0.11+ds1-5/website/README.md	1970-01-01 00:00:00.000000000 +0000
+++ 0.12+ds1-1/website/README.md	2018-07-11 02:32:25.000000000 +0000
@@ -0,0 +1,20 @@
+This directory contains source for Elvish's official website. They are
+converted to HTML with a custom toolchain with the following dependencies:
+
+*   GNU Make (any "reasonably modern" version should do).
+
+*   pandoc 2.2.1 (other versions in the 2.x series might also work).
+
+*   A Go toolchain, for building [genblog](https://github.com/xiaq/genblog)
+    and some custom preprocessors in the `tools` directory.
+
+The documents are written in GitHub-flavored markdown sprinkled with some HTML
+and custom macros. Most of them can be viewed directly in GitHub.
+
+# History
+
+These files used to live in a [separate
+repository](https://github.com/elves/elvish.io). However, because @xiaq did
+not merge the repositories in the correct way (he simply copied all the
+files), the commit history is lost. Please see that repository for a full list
+of contributors.
diff -pruN 0.11+ds1-5/website/src/blog/0.10-release-notes.md 0.12+ds1-1/website/src/blog/0.10-release-notes.md
--- 0.11+ds1-5/website/src/blog/0.10-release-notes.md	1970-01-01 00:00:00.000000000 +0000
+++ 0.12+ds1-1/website/src/blog/0.10-release-notes.md	2018-07-11 02:32:25.000000000 +0000
@@ -0,0 +1,133 @@
+Version 0.10 has been released two and a half months after 0.9, bringing many new features and enhancements. The [second issue](newsletter-sep-2017.html) of Elvish Newsletter accompanies this release.
+
+# Breaking changes
+
+*   If you are upgrading from an earlier version, Elvish will complain that
+    your database is not valid. This is because Elvish now uses BoltDB for
+    storage. A [database migration
+    tool](https://github.com/elves/upgrade-db-for-0.10) is available.
+
+*   Breaking changes to the editor API:
+
+    *   The `$edit:completer` map is now known as `$edit:arg-completer`.
+
+    *   The keybinding API has been changed (again). Keybindings for different now
+        live in their own subnamespaces. For instance, keybindings for insert mode
+        used to be `edit:binding[insert]` but is now `edit:insert:binding`.
+
+    *   Module names of some editor modes have also been changed for consistency.
+        The completion mode now uses the `edit:completion` module (used to be
+        `edit:compl`). Location mode: `edit:location`; navigation mode:
+        `edit:navigation`.
+
+    *   Byte output from prompts now preserve newlines. For instance, if you
+        have `edit:prompt = { echo haha }`, you will now have a trailing
+        newline in the prompt, making your command appear on the next line. To
+        fix this, simple replace `echo` with `print`, which does not print a
+        trailing newline. ([#354](https://github.com/elves/elvish/issues/354))
+
+*   Breaking changes to the language core:
+
+    *   Due to the switch to persistent data structures, assignments of maps
+        now behave as if they copy the entire container. See the section in
+        [some unique semantics](/learn/unique-semantics.html) for an
+        explanation.
+
+    *   The implicit `$args` variable is gone, as well as its friends:
+        positional variables `$0`, `$1`, ..., and the special `$@` shorthand
+        for `$@args`. Lambdas defined without argument list (`{ some code }`)
+        now behave as if they have an empty argument list.
+        ([#397](https://github.com/elves/elvish/issues/397))
+
+        Old lambdas that rely on `$args` or its friends now must declare their
+        arguments explicitly. For instance, `fn ls { e:ls --color=auto $@ }`
+        needs to be rewritten to `fn ls [@a]{ e:ls --color=auto $@a }`.
+
+    * Support for using backquotes for output capture (e.g. ``echo uname is
+      `uname` ``) has been removed. Use parentheses instead (e.g. `echo uname
+      is (uname)`).
+      
+    * Backquotes are repurposed for line continuation. A backquote followed by
+      a newline is equivalent to a space.
+      ([#417](https://github.com/elves/elvish/issues/417))
+
+*   The signature of the `splits` builtin has been changed. The separator used
+    to be an option `sep` but is now the first argument. For instance, `splits
+    &sep=: a:b:c` should now be written as `splits : a:b:c`.
+
+
+# Notable fixes and enhancements
+
+*   Thanks to the BoltDB migration, Elvish is now a pure Go project! This
+    allows for fully statically linked executables and easy cross compilation.
+    ([#377](https://github.com/elves/elvish/issues/377))
+
+*   Enhancements to the language core:
+
+    *   It is now possible to define options when declaring functions
+        ([#82](https://github.com/elves/elvish/issues/82)).
+
+    * Interrupting Elvish code with Ctrl-C now works more reliably ([#388](https://github.com/elves/elvish/issues/388)).
+
+    * Piping value outputs into a command that does not read the value input (e.g. `range 1000 | echo haha`) no longer hangs ([#389](https://github.com/elves/elvish/issues/389)).
+
+*   New builtins functions and variables (documented in the [builtin module reference](/ref/builtin.html)):
+
+    * New `assoc` and `dissoc` builtin that outputs modified versions of container types.
+
+    * New `keys`, `has-keys` and `has-values` builtins ([#432](https://github.com/elves/elvish/issues/432), [#398](https://github.com/elves/elvish/issues/398), [#407](https://github.com/elves/elvish/issues/407)).
+
+    * A new blackhole variable `$_` has been added. ([#401](https://github.com/elves/elvish/issues/401))
+
+    * New `replaces` builtin for replacing strings. ([#463](https://github.com/elves/elvish/issues/463))
+
+    * New `not-eq` builtin for inequality.
+
+    * New `drop` builtin, mirroring `take`.
+
+*   Enhancements to the editor:
+
+    *   Matching algorithm used in completion is now programmable with
+        `$edit:-matcher` ([#430](https://github.com/elves/elvish/issues/430));
+        see [documentation](/ref/edit.html).
+
+    *   Elvish can now able to complete arguments with variables. For instance,
+        if you have a directory with `a.mp4` and `a.txt`, and variable `$foo`
+        containing `a`, `echo $foo.<Tab>` now works
+        ([#446](https://github.com/elves/elvish/issues/446)). However, the
+        completion will expand `$foo` into `a`, which is not intended
+        ([#474](https://github.com/elves/elvish/issues/474)).
+
+    *   It is now possible to manipulate the cursor position using the
+        experimental `$edit:-dot` variable
+        ([415](https://github.com/elves/elvish/issues/415)).
+
+    *   The default prompt now replaces `>` with a red `#` when uid = 0.
+
+*   An experimental custom listing mode (known as "narrow mode" for now) has
+    been introduced and can be started with `edit:-narrow-read`. This means
+    that it is now to implement listing modes entirely in Elvish script.
+
+    Experimental re-implementations of the several standard listing modes
+    (location mode, history lising mode and lastcmd mode) are provided as the
+    bundled `narrow` module. Read [its source in
+    eval/narrow.elv](https://github.com/elves/elvish/blob/e7a8b96d7d4fccb7bafe01f27db9c0fe06c568b4/eval/narrow.elv)
+    for more details.
+
+*   Improvements to the daemon:
+
+    *   The daemon now quits automatically when all Elvish sessions are closed.
+        ([#419](https://github.com/elves/elvish/issues/419))
+
+    *   The daemon can now spawned be correctly when Elvish is not installed in
+        `PATH`.
+
+*   Elvish no longer quits on SIGQUIT (usually triggered by `Ctrl-\`), matching
+    the behavior of other shells. It still prints a stack trace though, which
+    can be useful for debugging.
+    ([#411](https://github.com/elves/elvish/issues/411))
+
+*   A `-compileonly` flag for the Elvish binary is added. It makes Elvish
+    compiles a script (in memory) but does not execute it. It can be used for
+    checking the well-formedness of programs and is useful in editor plugins.
+    ([#458](https://github.com/elves/elvish/issues/458))
diff -pruN 0.11+ds1-5/website/src/blog/0.11-release-notes.md 0.12+ds1-1/website/src/blog/0.11-release-notes.md
--- 0.11+ds1-5/website/src/blog/0.11-release-notes.md	1970-01-01 00:00:00.000000000 +0000
+++ 0.12+ds1-1/website/src/blog/0.11-release-notes.md	2018-07-11 02:32:25.000000000 +0000
@@ -0,0 +1,158 @@
+Version 0.11 has been released four months after 0.10, bringing many new
+features and bugfixes. There is no newsletter accompanying this release
+(instead, there is a
+[tweet](https://twitter.com/RealElvishShell/status/953781788706557952)).
+
+As usual, prebuilt binaries can be found in [download](../download).
+
+# Breaking Changes
+
+*   The syntax for importing modules in nested directories with `use` has
+    changed.
+
+    Previously, `use` accepts colon-delimited components, and replace the
+    colons with slashes to derive the path: for instance, `use a:b:c` imports
+    the module `a/b/c.elv` under `~/.elvish/lib`, under the namespace `a:b:c`.
+    Now, to import this module, you should use `use a/b/c` instead, and it
+    will import the same file under the namespace `c`.
+
+    It is not yet possible to rename the module when importing it; this makes
+    it hard to import modules with the same filename living under different
+    directories and will be addressed in the next version.
+
+    The current implementation of `use` still supports the use of colons to
+    affect the name under which the module is imported: for instance, `use
+    a/b:c` imports the file `a/b/c.elv` under the name `b:c`. However, this
+    feature should be considered undocumented and will be removed in the next
+    version.
+
+*   Module imports are now scoped lexically, akin to how variables are scoped.
+    For instance, `use re` in one module does not affect other files; neither
+    does `{ use re }` affect the outer scope.
+
+*   The variable a function `func` maps to is now `$func~` instead of
+    `$&func`. The ampersand `&` is now no longer allowed in variable names,
+    while the tilde `~` is. A
+    [tool](https://github.com/elves/upgrade-scripts-for-0.11) has been
+    provided to rewrite old code.
+
+*   Strings are no longer callable
+    ([#552](https://github.com/elves/elvish/issues/552)). A new
+    [external](/ref/builtin.html#external) has been added to support calling
+    external programs dynamically.
+
+*   It is now forbidden to assign non-strings to environment variables. For
+    instance, `E:X = []` used to assign the environment variable `X` the
+    string value `"[]"`; now it causes an exception.
+
+
+# Notable Fixes and Enhancements
+
+## Supported Platforms
+
+*   Support for Go 1.7 has been dropped, and support for Go 1.9 has been
+    added.
+
+*   Elvish now has experimental support for Windows 10. Terminal and
+    filesystem features may be buggy. [Prebuilt binaries](/download) for
+    Windows are also available.
+
+*   Prebuilt binaries for AMD64 and ARM64 architectures on Linux are provided.
+
+## Language
+
+*   It is now possible to `use` relative paths. For instance, in module
+    `a/b/c/foo.elv` (under `~/.elvish/lib`), `use ./bar` is the same as `use
+    a/b/c/bar`, and `use ../bar` is the same as `use a/b/bar`. The resolved
+    path must not escape the `lib` directory; for instance, `use ../bar` from
+    `~/.elvish/lib/foo.elv` will throw cause a compilation error.
+
+*   A new builtin variable,
+    [`$value-out-indicator`](/ref/builtin.html#value-out-indicator), can now be
+    used to customize the marker for value outputs
+    ([#473](https://github.com/elves/elvish/issues/473)).
+
+*   A new builtin command [`to-string`](/ref/builtin.html#to-string) has been
+    added.
+
+*   Special forms like `if` now works correctly with redirections and
+    temporary assignments ([#486](https://github.com/elves/elvish/issues/486)).
+
+*   A primitive for running functions in parallel,
+    [`run-parellel`](/ref/builtin.html#run-parallel), has been added
+    ([#485](https://github.com/elves/elvish/issues/485)).
+
+*   The [`splits`](/ref/builtin.html#splits) builtin now supports a `&max`
+    option.
+
+*   A new [`src`](/ref/builtin.html#src) builtin can now be used to get
+    information about the current source.
+
+*   A new builtin variable [`$args`](/ref/builtin.html#args) can now be used
+    to access command-line arguments.
+
+*   The [`del`](/ref/language.html#deleting-variable-or-element-del) special
+    command can now delete map elements
+    ([#79](https://github.com/elves/elvish/issues/79)).
+
+## Editor
+
+*   A maximum wait time can be specified with
+    [$edit:-prompts-max-wait](/ref/edit.html#edit-prompts-max-wait) to prevent
+    slow prompt functions from blocking UI updates
+    ([#482](https://github.com/elves/elvish/issues/482)).
+
+*   Execution of hook functions are now correctly isolated
+    ([#515](https://github.com/elves/elvish/issues/515)).
+
+*   A new [matcher](/ref/edit.html#matcher) `edit:match-substr` has been added.
+
+*   The editor is now able to handle Alt-modified function keys in more
+    terminals ([#181](https://github.com/elves/elvish/issues/181)).
+
+*   Location mode now always hides the current directory
+    ([#531](https://github.com/elves/elvish/issues/531)).
+
+*   Ctrl-H is now treated the same as Backspace
+    ([#539](https://github.com/elves/elvish/issues/539)).
+
+*   It is now possible to scroll file previews with
+    <span class="key">Alt-Up</span> and <span class="key">Alt-Down</span>.
+
+*   The height the editor can take up can now be restricted with
+    [`$edit:max-height`](/ref/edit.html#editmax-height).
+
+## Misc
+
+*   The Elvish command supports a new `-buildinfo` flag, which causes Elvish
+    to print out the version and builder of the binary and exit. Another
+    `-json` flag has also been introduced; when present, it causes
+    `-buildinfo` to print out a JSON object.
+
+*   Elvish now handles SIGHUP by relaying it to the entire process group
+    ([#494](https://github.com/elves/elvish/issues/494)).
+
+*   The daemon now detects the path of the Elvish executable more reliabily,
+    notably when Elvish is used as login shell
+    ([#496](https://github.com/elves/elvish/issues/496)).
+
+*   When an exception is thrown and the traceback contains only one entry,
+    the traceback is now shown more compactly.
+
+    Before:
+
+    ```elvish-transcript
+    ~> fail x
+    Exception: x
+    Traceback:
+      [interactive], line 1:
+        fail x
+    ```
+
+    Now:
+
+    ```elvish-transcript
+    ~> fail x
+    Exception: x
+    [tty], line 1: fail x
+    ```
diff -pruN 0.11+ds1-5/website/src/blog/0.9-release-notes.md 0.12+ds1-1/website/src/blog/0.9-release-notes.md
--- 0.11+ds1-5/website/src/blog/0.9-release-notes.md	1970-01-01 00:00:00.000000000 +0000
+++ 0.12+ds1-1/website/src/blog/0.9-release-notes.md	2018-07-11 02:32:25.000000000 +0000
@@ -0,0 +1,92 @@
+Version 0.9 has been released to coincide with the official publication of the
+Elvish website, which will be hosting all release notes in the future.
+
+This version is released slightly more than a month after 0.8. Despite the
+short interval, there are some interesting additions and changes.
+
+# Breaking Changes
+
+*   Lists have become immutable.
+
+    Support for assigning individual list elements has been temporarily
+    removed. For instance, the following is no longer possible:
+
+    ```elvish
+    li = [lorem ipsum foo bar]
+    li[1] = not-ipsum
+    ```
+
+    You need to use this for now:
+
+    ```elvish
+    li = [(explode $li[:1]) not-ipsum (explode $li[2:])]
+    ```
+
+    Element assignment will be reintroduced as a syntax sugar, after the
+    conversion to persistent data structure is finished.
+
+    Assignments to map elements are not affected.
+
+*   The `true` and `false` builtin commands have been removed. They have been
+    equivalent to `put $true` and `put $false` for a while.
+
+*   The default keybinding for last command mode has been changed from <span
+    class="key">Alt-,</span> to <span class="key">Alt-1</span>.
+
+*   The "bang mode" is now known as "last command mode".
+
+*   The `le:` module (for accessing the Elvish editor) has been renamed to
+    `edit:`. You can do a simple substitution `s/le:/edit:/g` to fix your
+    `rc.elv`.
+
+*   The 3 listing modes -- location mode, history listing mode and last
+    command mode -- are being merged into one generic listing mode.
+
+    Most of their builtins have been merged: for instance, they use to come
+    with their own builtins for changing focused candidate, `le:loc:up`,
+    `le:histlist:up` and `le:bang:up`. These have been merged into simply
+    `edit:listing:up`, that operates on whichever listing mode is active.
+
+    A new binding table, `$edit:binding[listing]` has also been introduced.
+    Bindings put there will be available in all 3 listing modes, with bindings
+    in their own tables (`$edit:binding[loc]`, `$edit:binding[histlist]` and
+    `$edit:binding[lastcmd]`) having higher precedence.
+
+*   The readline-style binding module has been renamed from
+    `embedded:readline-binding` to just `readline-binding`. Future embedded
+    modules will no longer have an `embedded:` prefix either.
+
+
+# Notable Fixes and Enhancements
+
+*   This release has seen more progress towards breaking up the huge, untested
+    [edit](https://github.com/elves/elvish/tree/master/edit) package. For
+    instance, the syntax highlighter and command history helpers now live in
+    their own packages, and have better test coverages.
+
+*   An experimental web interface has been added. It can be used by supplying
+    the `-web` flag when running Elvish, i.e. `elvish -web`. The default port
+    is 3171, which is [a way](https://en.wikipedia.org/wiki/Leet) to write
+    "ELVI". An alternative port can be specified using `-port`, e.g. `elvish
+    -web -port 2333`.
+
+*   Per-session command history has been reintroduced
+    ([#355](https://github.com/elves/elvish/issues/355)).
+
+*   Elvish now forks a daemon for mediating access to the database. This is to
+    prepare for the switch to a pure Go database and removing the current C
+    dependency on SQLite. A new `daemon:` module has been introduced.
+
+*   A new `edit:complex-candidate` builtin has been introduced to construct
+    complex candidates from completers.
+
+*   A new `re:` module, containing regular expression utilities, has been
+    introduced.
+
+
+# Known Issues
+
+The daemon implementation has a known issue of some intermediate process not
+being reaped correctly and there is an outstanding [pull
+request](https://github.com/elves/elvish/pull/373/) for it. In the worst case,
+this will leave 2 processes hanging in the system.
diff -pruN 0.11+ds1-5/website/src/blog/index.toml 0.12+ds1-1/website/src/blog/index.toml
--- 0.11+ds1-5/website/src/blog/index.toml	1970-01-01 00:00:00.000000000 +0000
+++ 0.12+ds1-1/website/src/blog/index.toml	2018-07-11 02:32:25.000000000 +0000
@@ -0,0 +1,29 @@
+[[articles]]
+name = "0.11-release-notes"
+title = "Elvish 0.11 Release Notes"
+timestamp = "2018-01-17"
+
+[[articles]]
+name = "newsletter-sep-2017"
+title = "Elvish Newsletter, Sep 2017 Issue"
+timestamp = "2017-09-17"
+
+[[articles]]
+name = "0.10-release-notes"
+title = "Elvish 0.10 Release Notes"
+timestamp = "2017-09-17"
+
+[[articles]]
+name = "newsletter-july-2017"
+title = "Elvish Newsletter, July 2017 Issue"
+timestamp = "2017-07-10"
+
+[[articles]]
+name = "0.9-release-notes"
+title = "Elvish 0.9 Release Notes"
+timestamp = "2017-07-03"
+
+[[articles]]
+name = "live"
+title = "The Elvish Website is Officially Live!"
+timestamp = "2017-07-03"
diff -pruN 0.11+ds1-5/website/src/blog/live.md 0.12+ds1-1/website/src/blog/live.md
--- 0.11+ds1-5/website/src/blog/live.md	1970-01-01 00:00:00.000000000 +0000
+++ 0.12+ds1-1/website/src/blog/live.md	2018-07-11 02:32:25.000000000 +0000
@@ -0,0 +1,31 @@
+After being "in construction" for ages, the Elvish website is now officially
+live. It is (still) not complete yet, but in the spirit of "release early",
+here it is :)
+
+This website will be the entry point of all Elvish information. It will host
+documents (including both [learning materials](/learn) and
+[references](/ref)), as well as release notes or other announcements.  The
+[GitHub repository](https://github.com/elves/elvish) continues to host code
+and issues, of course.
+
+The website has no comment facilities: commenting is outsourced to [Hacker
+News](https://news.ycombinator.com) and [r/elv](https://www.reddit.com/r/elv/).
+
+The [homepage](/) introduces each section, and highlights some key features of
+Elvish with "ttyshots". Please start there to explore this website!
+
+
+# Technical Stack
+
+The website is built with a [static website
+generator](https://github.com/xiaq/genblog) from a bunch of [MarkDown,
+configuration files and messy scripts](https://github.com/elves/elvish.io),
+and hosted on a [DigitalOcean](https://www.digitalocean.com) server with
+[nginx](http://nginx.org).
+
+Thanks to [Cloudflare](https://www.cloudflare.com), this website has a good
+chance of withstanding kisses of death from Hacker News, should there be any.
+It uses Cloudflare's "strict" SSL configuration, meaning that both traffic from
+you to Cloudflare and Cloudflare to the origin server are fully encrypted with
+verified certificates. The origin server obtains and renews its certificate
+from the wonderful [Let's Encrypt](https://letsencrypt.org/) service.
diff -pruN 0.11+ds1-5/website/src/blog/newsletter-july-2017.md 0.12+ds1-1/website/src/blog/newsletter-july-2017.md
--- 0.11+ds1-5/website/src/blog/newsletter-july-2017.md	1970-01-01 00:00:00.000000000 +0000
+++ 0.12+ds1-1/website/src/blog/newsletter-july-2017.md	2018-07-11 02:32:25.000000000 +0000
@@ -0,0 +1,117 @@
+Welcome to the first issue of Elvish Newsletter!
+
+Elvish is a shell that seeks to combine a full-fledged programming language
+with a friendly user interface. This newsletter is a summary of its progress
+and future plans.
+
+
+# Status Updates
+
+*   18 pull requests to the [main repo](https://github.com/elves/elvish) have
+    been merged in the past four weeks. Among them 13 were made by @xofyargs,
+    and the rest by @myfreeweb, @jiujieti, @HeavyHorst, @silvasur and @ALSchwalm.
+    The [website repo](https://github.com/elves/elvish.io) has also
+    merged 3 pull reqeusts from @bengesoff, @zhsj and @silvasur. Many kudos!
+
+*   The [website](https://elvish.io) was [officially
+    live](/blog/live.html) on 3 July. Although the initial
+    [submission](https://news.ycombinator.com/item?id=14691639) to HN was a
+    failure, Elvish gained
+    [quite](https://www.reddit.com/r/programming/comments/6l38nd/elvish_friendly_and_expressive_shell/)
+    [some](https://www.reddit.com/r/golang/comments/6l3aev/elvish_friendly_and_expressive_shell_written_in_go/)
+    [popularity](https://www.reddit.com/r/linux/comments/6l6wcs/elvish_friendly_and_expressive_shell_now_ready/)
+    on Reddit, and [another](https://news.ycombinator.com/item?id=14698187) HN
+    submission made to the homepage. These, among others, have brought 40k
+    unique visitors to the website, totalling 340k HTTP requests. Thank you
+    Internet :)
+
+*   A lot of discussions have happened over the IM channels and the issue
+    tracker, and it has become necessary to better document the current status
+    of Elvish and organize the development effort, and this newsletter is
+    part of the response.
+
+    There is no fixed schedule yet, but the current plan is to publish
+    newsletters roughly every month. Preview releases of Elvish, which used to
+    happen quite arbitrarily, will also be done to coincide with the publication
+    of newsletters.
+
+*   There are now IM channels for developers, see below for details.
+
+
+# Short-Term and Mid-Term Plans
+
+The next preview release will be 0.10, and there is now a
+[milestone](https://github.com/elves/elvish/milestone/2) for it, a list of
+issues considered vital for the release. If you would like to contribute, you
+are more than welcome to pick an issue from that list, although you are also
+more than welcome to pick just any issue.
+
+Aside from the short-term goal of releasing 0.10, here are the current mid-term
+focus areas of Elvish development:
+
+*   Stabilizing the language core.
+
+    The core of Elvish is still pretty immature, and it is definitely not as
+    usable as any other dynamic language, say Python or Clojure. Among others,
+    the 0.10 milestone now plans changes to the implementation of maps
+    ([#414](https://github.com/elves/elvish/issues/414)), a new semantics of
+    element assignment ([#422](https://github.com/elves/elvish/issues/422))
+    and enhanced syntax for function definition
+    ([#82](https://github.com/elves/elvish/issues/82) and
+    [#397](https://github.com/elves/elvish/issues/397)). You probably wouldn't
+    expect such fundamental changes in a mature language :)
+
+    A stable language core is a prerequisite for a 1.0 release. Elvish 1.x
+    will maintain backwards compatibility with code written for earlier 1.x
+    versions.
+
+*   Enhance usability of the user interface, and provide basic
+    programmability.
+
+    The goal is to build a fully programmable user interface, and there are a
+    lot to be done. Among others, the 0.10 milestone plans to support
+    manipulating the cursor
+    ([#415](https://github.com/elves/elvish/issues/415)) programmatically,
+    scrolling of previews in navigation mode previews
+    ([#381](https://github.com/elves/elvish/issues/381)), and invoking
+    external editors for editing code
+    ([#393](https://github.com/elves/elvish/issues/393)).
+
+    The user interface is important for two reasons. Enhancements to the UI
+    can improve the power of Elvish directly and significantly; its API is also
+    a very good place for testing the language. By developing the language and
+    the user interface in parallel, we can make sure that they work well
+    together.
+
+Like many other open source projects, you are welcome to discuss and challenge
+the current plan, or come up with your ideas regarding the design and
+implementation.
+
+(So what's the long-term goal of Elvish? The long-term goal is to remove the
+"seeks to" part from the introduction of Elvish at the beginning of the post.)
+
+
+# Development IM Channels
+
+To better coordinate development, there are now IM channels for Elvish
+development:
+[#elvish-dev](http://webchat.freenode.net/?channels=elvish-dev) on freenode,
+[elves/elvish-dev](https://gitter.im/elves/elvish-dev) on Gitter and
+[@elvish_dev](https://telegram.me/elvish_dev) on Telegram. These channels are
+all connected together thanks to [fishroom](https://github.com/tuna/fishroom).
+
+For general questions, you are welcome in
+[#elvish](https://webchat.freenode.net/?channels=elvish) on Freenode,
+[elves/elvish-public](https://gitter.im/elves/elvish-public) on Gitter, or
+[@elvish](https://telegram.me/elvish) on Telegram.
+
+
+# Conclusion
+
+This concludes this first issue of the newsletter. Hopefully future issues of
+this newsletter will also feature blog posts from Elvish users like *Elvish for
+Python Users* and popular Elvish modules like *Tetris in Your Shell* :)
+
+Have Fun with Elvish!
+
+\- xiaq
diff -pruN 0.11+ds1-5/website/src/blog/newsletter-sep-2017.md 0.12+ds1-1/website/src/blog/newsletter-sep-2017.md
--- 0.11+ds1-5/website/src/blog/newsletter-sep-2017.md	1970-01-01 00:00:00.000000000 +0000
+++ 0.12+ds1-1/website/src/blog/newsletter-sep-2017.md	2018-07-11 02:32:25.000000000 +0000
@@ -0,0 +1,77 @@
+Welcome to the second issue of Elvish Newsletter!
+
+Elvish is a shell that seeks to combine a full-fledged programming language
+with a friendly user interface. This newsletter is a summary of its progress
+and future plans.
+
+
+# Release of 0.10 Version
+
+This newsletter accompanies the release of the 0.10 version. This release
+contains 125 commits, with contributions from @xofyarg, @tw4452852, @ALSchwalm,
+@zhsj, @HeavyHorst, @silvasur, @zzamboni, @Chilledheart, @myfreeweb, @xchenan
+and @jiujieti.
+
+Elvish used to depend on SQLite for storage. As a result, compiling Elvish
+relied on cgo and required a C compiler. This release sees the switch to
+BoltDB, making Elvish a pure-Go project. Elvish can now be compiled much
+faster, and into a fully statically linked binary. Cross-compilation is also
+much easier, as the Go compiler has fantastic cross-compiling support.
+
+Maps (`[&k=v &k2=v2]`) are now implemented using persistent hash maps. This
+concludes the transition to persistent data structures for all primary data
+types (strings, lists, maps). Persistent data structures are immutable, and
+thus have a simpler semantics and are automatically concurrency-safe. This
+does have an interesting impact on the semantics of assignments, which is now
+documented in a new section on the [unique
+semantics](/learn/unique-semantics.html) page.
+
+For a complete list of changes, see the [release notes](0.10-release-notes.html).
+
+
+# Community
+
+*   We now have an official list of awesome unofficial Elvish libraries:
+    [elves/awesome-elvish](https://github.com/elves/awesome-elvish). Among
+    others, we now have at least two very advanced prompt themes, chain.elv
+    from @zzamboni and powerline.elv from @muesli :)
+
+*   Diego Zamboni (@zzamboni), the author of chain.elv, has written very
+    passionately on Elvish: [Elvish, an awesome Unix
+    shell](http://zzamboni.org/post/elvish-an-awesome-unix-shell/).
+
+*   Patrick Callahan has given an awesome talk on [Delightful Command-Line
+    Experiences](https://dl.elvish.io/resources/callahan-delightful-commandline-experiences.pdf),
+    featuring Elvish as a "very lively, ambitious shell".
+
+*   Elvish is now [packaged](https://packages.debian.org/elvish) in Debian.
+
+*   The number of followers to
+    [@RealElvishShell](https://twitter.com/RealElvishShell/) has grown to 23.
+
+
+# Plans
+
+The mid-term remains the same as in the [previous
+issue](newsletter-july-2017.html): stabilizing the language core and enhancing
+usability of the user interface.
+
+The short-term plan is captured in the
+[milestone](https://github.com/elves/elvish/milestone/3) for the 0.11 version.
+Among other things, 0.11 is expected to ship with `epm`, [the standard package
+manager](https://github.com/elves/elvish/issues/239) for Elvish, and a more
+responsive interface by running
+[prompts](https://github.com/elves/elvish/issues/482) and
+[completions](https://github.com/elves/elvish/issues/483) asynchronously. Stay
+very tuned.
+
+
+# Conclusions
+
+In the last newsletter, I predicted that we will be featuring *Elvish for
+Python Users* and *Tetris in Your Shell* in a future newsletter. It seems we
+are getting close to that pretty steadily.
+
+Have fun with Elvish!
+
+\- xiaq
diff -pruN 0.11+ds1-5/website/src/download/index.toml 0.12+ds1-1/website/src/download/index.toml
--- 0.11+ds1-5/website/src/download/index.toml	1970-01-01 00:00:00.000000000 +0000
+++ 0.12+ds1-1/website/src/download/index.toml	2018-07-11 02:32:25.000000000 +0000
@@ -0,0 +1 @@
+prelude = "prelude"
diff -pruN 0.11+ds1-5/website/src/download/prelude.md 0.12+ds1-1/website/src/download/prelude.md
--- 0.11+ds1-5/website/src/download/prelude.md	1970-01-01 00:00:00.000000000 +0000
+++ 0.12+ds1-1/website/src/download/prelude.md	2018-07-11 02:32:25.000000000 +0000
@@ -0,0 +1,239 @@
+Prebuilt, statically linked binaries for some common platforms are provided
+below.
+If your environment is not listed above, you may still be able to build Elvish
+from [source](https://github.com/elves/elvish).
+For users in China, the
+[mirror](https://mirrors.tuna.tsinghua.edu.cn/elvish) hosted by TUNA may be
+faster.
+
+Note that Elvish is **pre-release software**. It can be unstable, and does not
+maintain backward compatibility from version to version.
+
+<style>
+  table {
+    border-collapse: collpase;
+    width: 100%;
+    margin-bottom: 16px;
+  }
+  td, th {
+    border: 1px solid #aaa;
+    text-align: left;
+    padding: 0.4em;
+  }
+  tr:nth-child(even) {
+    background-color: #ddd;
+  }
+</style>
+
+<table>
+  <colgroup>
+    <col style="width:34%">
+    <col style="width:22%">
+    <col style="width:22%">
+    <col style="width:22%">
+  </colgroup>
+  <tr>
+    <th>Version</th>
+    <th>Linux</th>
+    <th>macOS</th>
+    <th>Windows</th>
+  </tr>
+  <tr>
+    <td>HEAD</td>
+    <td>
+      $dl amd64 elvish-linux-amd64-HEAD.tar.gz
+      $dl 386 elvish-linux-386-HEAD.tar.gz
+      $dl arm64 elvish-linux-arm64-HEAD.tar.gz
+    </td>
+    <td>
+      $dl amd64 elvish-darwin-amd64-HEAD.tar.gz
+    </td>
+    <td>
+      $dl amd64 elvish-windows-amd64-HEAD.zip
+      $dl 386 elvish-windows-386-HEAD.zip
+    </td>
+  </tr>
+  <tr>
+    <td>
+      0.11 (<a href="/blog/0.11-release-notes.html">Release Note</a>)
+    </td>
+    <td>
+      $dl amd64 elvish-linux-amd64-0.11.tar.gz
+      $dl 386 elvish-linux-386-0.11.tar.gz
+      $dl arm64 elvish-linux-arm64-0.11.tar.gz
+    </td>
+    <td>
+      $dl amd64 elvish-darwin-amd64-0.11.tar.gz
+    </td>
+    <td>
+      $dl amd64 elvish-windows-amd64-0.11.zip
+      $dl 386 elvish-windows-386-0.11.zip
+    </td>
+  </tr>
+</table>
+
+
+Old versions:
+
+<table>
+  <colgroup>
+    <col style="width:34%">
+    <col style="width:22%">
+    <col style="width:22%">
+    <col style="width:22%">
+  </colgroup>
+  <tr>
+    <th>Version</th>
+    <th>Linux</th>
+    <th>macOS</th>
+    <th>Windows</th>
+  </tr>
+  <tr>
+    <td>0.10.1 (<a href="/blog/0.10-release-notes.html">Release Note</a>)</td>
+    <td>
+      $dl amd64 elvish-0.10.1-linux.tar.gz
+    </td>
+    <td>
+      $dl amd64 elvish-0.10.1-osx.tar.gz
+    </td>
+    <td>N/A</td>
+  </tr>
+  <tr>
+    <td>0.9 (<a href="/blog/0.9-release-notes.html">Release Note</a>)</td>
+    <td>
+      $dl amd64 elvish-0.9-linux.tar.gz
+    </td>
+    <td>
+      $dl amd64 elvish-0.9-osx.tar.gz
+    </td>
+    <td>N/A</td>
+  </tr>
+  <tr>
+    <td>0.8 (<a href="https://github.com/elves/elvish/releases/tag/0.8">Release Note</a>)</td>
+    <td>
+      $dl amd64 elvish-0.8-linux.tar.gz
+    </td>
+    <td>
+      $dl amd64 elvish-0.8-osx.tar.gz
+    </td>
+    <td>N/A</td>
+  </tr>
+  <tr>
+    <td>0.7 (<a href="https://github.com/elves/elvish/releases/tag/0.7">Release Note</a>)</td>
+    <td>
+      $dl amd64 elvish-0.7-linux.tar.gz
+    </td>
+    <td>
+      $dl amd64 elvish-0.7-osx.tar.gz
+    </td>
+    <td>N/A</td>
+  </tr>
+  <tr>
+    <td>0.6 (<a href="https://github.com/elves/elvish/releases/tag/0.6">Release Note</a>)</td>
+    <td>
+      $dl amd64 elvish-0.6-linux.tar.gz
+    </td>
+    <td>
+      $dl amd64 elvish-0.6-osx.tar.gz
+    </td>
+    <td>N/A</td>
+  </tr>
+  <tr>
+    <td>0.5 (<a href="https://github.com/elves/elvish/releases/tag/0.5">Release Note</a>)</td>
+    <td>
+    $dl amd64 elvish-0.5-linux.tar.gz
+    </td>
+    <td>
+      $dl amd64 elvish-0.5-osx.tar.gz
+    </td>
+    <td>N/A</td>
+  </tr>
+  <tr>
+    <td>0.4</td>
+    <td>
+      $dl amd64 elvish-0.4-linux.tar.gz
+    </td>
+    <td>
+      $dl amd64 elvish-0.4-osx.tar.gz
+    </td>
+    <td>N/A</td>
+  </tr>
+  <tr>
+    <td>0.3</td>
+    <td>
+      $dl amd64 elvish-0.3-linux.tar.gz
+    </td>
+    <td>
+      $dl amd64 elvish-0.3-osx.tar.gz
+    </td>
+    <td>N/A</td>
+  </tr>
+  <tr>
+    <td>0.2</td>
+    <td>
+      $dl amd64 elvish-0.2-linux.tar.gz
+    </td>
+    <td>
+      $dl amd64 elvish-0.2-osx.tar.gz
+    </td>
+    <td>N/A</td>
+  </tr>
+  <tr>
+    <td>0.1</td>
+    <td>
+      $dl amd64 elvish-0.1-linux.tar.gz
+    </td>
+    <td>
+      $dl amd64 elvish-0.1-osx.tar.gz
+    </td>
+    <td>N/A</td>
+  </tr>
+</table>
+
+
+# OS-Specific Packages
+
+## Fedora
+
+RPM packages are available from [the FZUG Repo](https://github.com/FZUG/repo/wiki/Add-FZUG-Repository):
+
+```elvish
+# Add FZUG repo
+dnf config-manager --add-repo=http://repo.fdzh.org/FZUG/FZUG.repo
+# Install Elvish
+dnf install elvish
+```
+
+## Debian / Ubuntu
+
+Elvish is packaged by [Debian](https://packages.debian.org/elvish) since
+buster and by [Ubuntu](http://packages.ubuntu.com/elvish) since 17.10.
+
+However, packages in official repositories are likely outdated. You can
+install the latest release from
+[PPA](https://launchpad.net/~zhsj/+archive/ubuntu/elvish):
+
+```elvish
+# Add Elvish PPA repo
+sudo wget -O /etc/apt/trusted.gpg.d/elvish \
+  'https://sks-keyservers.net/pks/lookup?search=0xE9EA75D542E35A20&options=mr&op=get'
+sudo gpg --dearmor /etc/apt/trusted.gpg.d/elvish
+sudo rm /etc/apt/trusted.gpg.d/elvish
+echo 'deb http://ppa.launchpad.net/zhsj/elvish/ubuntu xenial main' |
+  sudo tee /etc/apt/sources.list.d/elvish.list
+sudo apt-get update
+
+# Install Elvish
+sudo apt-get install elvish
+```
+
+## macOS (Homebrew)
+
+Elvish is packaged in Homebrew:
+
+```elvish
+# Install latest release
+brew install elvish
+# Or install HEAD:
+brew install --HEAD elvish
+```
diff -pruN 0.11+ds1-5/website/src/home.md 0.12+ds1-1/website/src/home.md
--- 0.11+ds1-5/website/src/home.md	1970-01-01 00:00:00.000000000 +0000
+++ 0.12+ds1-1/website/src/home.md	2018-07-11 02:32:25.000000000 +0000
@@ -0,0 +1,171 @@
+**Elvish** is a friendly and expressive shell for Linux, BSDs, macOS and
+Windows.
+
+<!--
+<pre id="demo-debug">
+</pre>
+-->
+
+<ul id="demo-switcher">
+  <li id="demo-js-warning">
+    Enable JavaScript for the option of showing demos as slides.
+  </li>
+  <li id="demo-expander-li" class="no-display"><a id="demo-expander">↧</a></li>
+</ul>
+
+<div id="demo-window"> <div id="demo-container" class="expanded">
+  <div class="demo-wrapper"> <div class="demo">
+    <div class="demo-col left"><div class="demo-ttyshot">
+      $ttyshot pipelines
+    </div></div>
+    <div class="demo-col right"> <div class="demo-description">
+      <div class="demo-title">Powerful Pipelines</div>
+      <p>
+        Text pipelines are intuitive and powerful. However, if your data have
+        inherently complex structures, passing them through the pipeline
+        often requires a lot of ad-hoc, hard-to-maintain text processing code.
+      </p>
+      <p>
+        Pipelines in Elvish can carry structured data, not just text. You can
+        stream lists, maps and even functions through the pipeline.
+      </p>
+    </div> </div>
+  </div> </div>
+
+  <div class="demo-wrapper"> <div class="demo">
+    <div class="demo-col left"><div class="demo-ttyshot">
+      $ttyshot control-structures
+    </div></div>
+    <div class="demo-col right"> <div class="demo-description">
+      <div class="demo-title">Intuitive Control Structures</div>
+      <p>
+        If you know programming, you probably already know how
+        <code>if</code> looks in C. So why learn another syntax?
+      </p>
+      <p>
+        Control structures in Elvish have an intuitive C-like syntax.
+      </p>
+    </div> </div>
+  </div> </div>
+
+  <div class="demo-wrapper"> <div class="demo">
+    <div class="demo-col left"><div class="demo-ttyshot">
+      $ttyshot location-mode
+    </div></div>
+    <div class="demo-col right"> <div class="demo-description">
+      <div class="demo-title">Directory History</div>
+      <p>
+        Is <code>cd /a/long/nested/directory</code> the first thing you
+        do every day? Struggling to remember where your logs and
+        configurations?
+      </p>
+      <p>
+        Elvish remembers where you have been. Press Ctrl-L and search, like in a
+        browser.
+      </p>
+    </div> </div>
+  </div> </div>
+
+  <div class="demo-wrapper"> <div class="demo">
+    <div class="demo-col left"><div class="demo-ttyshot">
+      $ttyshot histlist-mode
+    </div></div>
+    <div class="demo-col right"> <div class="demo-description">
+      <div class="demo-title">Command History</div>
+      <p>
+        Want to find the magical <code>ffmpeg</code> command that you used to
+        transcode a video file two months ago?
+      </p>
+      <p>
+        Just dig through your command history with Ctrl-R. Same key, more
+        useful.
+      </p>
+      <p>
+        (To be fair, you can do this in bash with <code>history | grep
+        ffmpeg</code>, but it's far fewer keystrokes in Elvish :)
+      </p>
+    </div> </div>
+  </div> </div>
+
+  <div class="demo-wrapper"> <div class="demo">
+    <div class="demo-col left"><div class="demo-ttyshot">
+      $ttyshot navigation-mode
+    </div></div>
+    <div class="demo-col right"> <div class="demo-description">
+      <div class="demo-title">Built-in File Manager</div>
+      <p>
+        Power of the shell or convenience of a file manager?
+      </p>
+      <p>
+        Choose both. Press Ctrl-N to quickly navigate directories and preview
+        files, with full shell power.
+      </p>
+    </div> </div>
+  </div> </div>
+</div> </div>
+
+<link href="/assets/home-demos.css" rel="stylesheet">
+<script src="/assets/home-demos.js"></script>
+
+# Getting Elvish
+
+Elvish is still in development, but has enough features and stability for
+daily use.
+
+*   [Download](/download/) prebuilt binaries if you are running Linux or macOS on
+    an x86-64 CPU.
+
+*   Source code is available on the [GitHub repository](https://github.com/elves/elvish).
+
+# Speaking Elvish
+
+*   [Learn](/learn/) to speak Elvish by following tutorials.
+
+    If you are not experienced with any shell, start with the
+    [fundamentals](/learn/fundamentals.html). (This tutorial is still a work in
+    progress, though.)
+
+    If you come from other shells, read the [cookbook](/learn/cookbook.html)
+    to get started quickly, and learn about Elvish's [unique
+    semantics](learn/unique-semantics.html).
+
+*   [Reference](/ref/) documents describe Elvish in a more formal and complete
+    way.
+
+    Read about the [philosophy](/ref/philosophy.html), the
+    [language](/ref/language.html), the [builtin module](/ref/builtin.html), and
+    more.
+
+*   [The blog](/blog/) contains news on Elvish.
+
+    It is the place for release notes, notes on the internals of Elvish, and
+    other announcements or musings from the developers.
+
+*   [The feed](/feed.atom) contains updates to all sections of the website (not
+    just the blog).
+
+*   [Follow](https://twitter.com/RealElvishShell/) Elvish on Twitter.
+
+# Meeting Other Elves
+
+*   Join [#elvish](https://webchat.freenode.net/?channels=elvish) on Freenode,
+    [elves/elvish-public](https://gitter.im/elves/elvish-public) on Gitter, or
+    [@elvish](https://telegram.me/elvish) on Telegram.
+
+    The wonderful [fishroom](https://github.com/tuna/fishroom) service
+    connects all of them together. So just join whichever channel suits you
+    best, and you won't miss discussions happening in other channels.
+
+*   If you are interested in contributing to Elvish, you can also discuss at
+    [#elvish-dev](http://webchat.freenode.net/?channels=elvish-dev) on
+    freenode, [elves/elvish-dev](https://gitter.im/elves/elvish-dev) on Gitter
+    or [@elvish_dev](https://telegram.me/elvish_dev) on Telegram.
+
+*   Chinese speakers are also welcome in
+    [#elvish-zh](https://webchat.freenode.net/?channels=elvish-zh) (Freenode)
+    and [@elvish_zh](https://telegram.me/elvish_zh) (Telegram). There are
+    also [#elvish-dev-zh](https://webchat.freenode.net/?channels=elvish-dev-zh)
+    (Freenode) and [@elvish_dev_zh](https://telegram.me/elvish_dev_zh) (Telegram).
+
+*   The [issue tracker](https://github.com/elves/elvish/issues) is the place
+    for bug reports and feature requests.
diff -pruN 0.11+ds1-5/website/src/index.toml 0.12+ds1-1/website/src/index.toml
--- 0.11+ds1-5/website/src/index.toml	1970-01-01 00:00:00.000000000 +0000
+++ 0.12+ds1-1/website/src/index.toml	2018-07-11 02:32:25.000000000 +0000
@@ -0,0 +1,25 @@
+title = "Elvish: Friendly and Expressive Shell"
+author = "Qi Xiao"
+feedPosts = 10
+template = "template.html"
+rootURL = "https://elv.sh"
+
+[index]
+name = "home"
+title = "Home"
+
+[[categories]]
+name = "download"
+title = "Download"
+
+[[categories]]
+name = "blog"
+title = "Blog"
+
+[[categories]]
+name = "learn"
+title = "Learn"
+
+[[categories]]
+name = "ref"
+title = "Reference"
diff -pruN 0.11+ds1-5/website/src/learn/cookbook.md 0.12+ds1-1/website/src/learn/cookbook.md
--- 0.11+ds1-5/website/src/learn/cookbook.md	1970-01-01 00:00:00.000000000 +0000
+++ 0.12+ds1-1/website/src/learn/cookbook.md	2018-07-11 02:32:25.000000000 +0000
@@ -0,0 +1,267 @@
+<!-- toc -->
+
+If you come from other shells, hopefully the following recipes will get you started quickly:
+
+# UI Recipes
+
+*   Put your startup script in `~/.elvish/rc.elv`. There is no `alias` yet,
+    but you can achieve the goal by defining a function:
+
+    ```elvish
+    fn ls [@a]{ e:ls --color $@a }
+    ```
+
+    The `e:` prefix (for "external") ensures that the external command named
+    `ls` will be called. Otherwise this definition will result in infinite
+    recursion.
+
+*   The left and right prompts can be customized by assigning functions to
+    `edit:prompt` and `edit:rprompt`. Their outputs are concatenated (with no
+    spaces in between) before being used as the respective prompts. The
+    following simulates the default prompts but uses fancy Unicode:
+
+    ```elvish
+    # "tilde-abbr" abbreviates home directory to a tilde.
+    edit:prompt = { tilde-abbr $pwd; put '❱ ' }
+    # "constantly" returns a function that always writes the same value(s) to
+    # output; "edit:styled" writes styled output.
+    edit:rprompt = (constantly (edit:styled (whoami)✸(hostname) inverse))
+    ```
+
+    Here is a terminalshot of the alternative prompts:
+
+    $ttyshot unicode-prompts
+
+*   Press <span class="key">▲&#xfe0e;</span> to search through history. It
+    uses what you have typed to do prefix match. To cancel, press <span
+    class="key">Escape</span>.
+
+    $ttyshot history-mode
+
+*   Press <span class="key">Tab</span> to start completion. Use arrow keys
+    <span class="key">▲&#xfe0e;</span>
+    <span class="key">▼&#xfe0e;</span>
+    <span class="key">◀&#xfe0e;</span>
+    <span class="key">▶&#xfe0e;</span>
+    or <span class="key">Tab</span> and <span class="key">Shift-Tab</span>
+    to select the candidate. Press <span class="key">Enter</span>, or just
+    continue typing to accept. To cancel, press <span
+    class="key">Escape.</span> It even comes with a scrollbar! :) In fact,
+    all interactive modes show a scrollbar when there is more output to see.
+
+    $ttyshot completion-mode
+
+*   You can make completion case-insensitive with the following code:
+
+    ```elvish
+    edit:-matcher[''] = [p]{ edit:match-prefix &ignore-case $p }
+    ```
+
+    You can also make the completion use "smart case" by changing
+    `&ignore-case` to `&smart-case`. This means that if your pattern
+    is entirely lower case it ignores case, otherwise it's case
+    sensitive.
+
+*   <a name="navigation-mode"></a>Press <span class="key">Ctrl-N</span> to start the builtin filesystem
+    navigator, aptly named "navigation mode." Use arrow keys to navigate.
+    <span class="key">Enter</span> inserts the selected filename to your
+    command line. If you want to insert the filename and stay in the mode
+    (e.g. when you want to insert several filenames), use <span
+    class="key">Alt-Enter</span>.
+
+    You can continue typing your command when you are in navigation mode.
+    Press <span class="key">Ctrl-H</span> to toggle hidden files; and like in
+    other modes, <span class="key">Escape</span> gets you back to the default
+    (insert) mode.
+
+    $ttyshot navigation-mode
+
+*   Try typing `echo [` and press <span class="key">Enter</span>. Elvish knows that the command is unfinished due to the unclosed `[` and inserts a newline instead of accepting the command. Moreover, common errors like syntax errors and missing variables are highlighted in real time.
+
+*   Elvish remembers which directories you have visited. Press <span
+    class="key">Ctrl-L</span> to list visited directories. Like in completion,
+    use arrow keys
+    <span class="key">▲&#xfe0e;</span>
+    <span class="key">▼&#xfe0e;</span>
+    or <span class="key">Tab</span> and <span class="key">Shift-Tab</span> to
+    select a directory and use Enter to `cd` into it. Press <span
+    class="key">Escape</span> to cancel.
+
+    $ttyshot location-mode
+
+    Type to filter:
+
+    $ttyshot location-mode-filter
+
+    The filtering algorithm is tailored for matching paths; you need only type
+    a prefix of each component. In the screenshot, x/p/v matches
+    **x**iaq/**p**ersistent/**v**ector.
+
+*   Elvish doesn't support history expansion like `!!`. Instead, it has a
+    "last command mode" offering the same functionality, triggered by <span
+    class="key">Alt-1</span> by default (resembling how you type `!` using
+    <span class="key">Shift-1</span>). In this mode, you can pick individual
+    arguments from the last command using numbers, or the entire command by
+    typing <span class="key">Alt-1</span> again.
+
+    This is showing me trying to fix a forgotten `sudo`:
+
+    $ttyshot lastcmd
+
+# Language Recipes
+
+*   Lists look like `[a b c]`, and maps look like `[&key1=value1
+    &key2=value2]`. Unlike other shells, a list never expands to multiple
+    words, unless you explicitly explode it by prefixing the variable name
+    with `@`:
+
+    ```elvish-transcript
+    ~> li = [1 2 3]
+    ~> put $li
+    ▶ [1 2 3]
+    ~> put $@li
+    ▶ 1
+    ▶ 2
+    ▶ 3
+    ~> map = [&k1=v1 &k2=v2]
+    ~> echo $map[k1]
+    v1
+    ```
+
+*   Environment variables live in a separate `E:` (for "environment") namespace and must be explicitly qualified:
+
+    ```elvish-transcript
+    ~> put $E:HOME
+    ▶ /home/xiaq
+    ~> E:PATH = $E:PATH":/bin"
+    ```
+
+*   You can manipulate search paths through the special list `$paths`, which is synced with `$E:PATH`:
+
+    ```elvish-transcript
+    ~> echo $paths
+    [/bin /sbin]
+    ~> paths = [/opt/bin $@paths /usr/bin]
+    ~> echo $paths
+    [/opt/bin /bin /sbin /usr/bin]
+    ~> echo $E:PATH
+    /opt/bin:/bin:/sbin:/usr/bin
+    ```
+
+*   You can manipulate the keybinding in the default insert mode through the
+    map `$edit:insert:binding`. For example, this binds
+    <span class="key">Ctrl-L</span> to clearing the terminal:
+
+    ```elvish
+    edit:insert:binding[Ctrl-L] = { clear > /dev/tty }
+    ```
+
+    Use `pprint $edit:insert:binding` to get a nice (albeit long) view of the
+    current keybinding.
+
+    **NOTE**: Bindings for letters modified by Alt are case-sensitive. For
+    instance, `Alt-a` means pressing `Alt` and `A`, while `Alt-A` means
+    pressing `Alt`, `Shift` and `A`. This will probably change in the future.
+
+*   There is no interpolation inside double quotes (yet). For example, the
+    output of `echo "$user"` is simply the string `$user`. Use implicit string
+    concatenation to build strings:
+
+    ```elvish-transcript
+    ~> name = xiaq
+    ~> echo "My name is "$name"."
+    My name is xiaq.
+    ```
+
+    Sometimes string concatenation will force you to use string literals
+    instead of barewords:
+
+    ```elvish-transcript
+    ~> noun = language
+    ~> echo $noun's'
+    languages
+    ```
+
+    You cannot write `s` as a bareword because Elvish would think you are
+    trying to write the variable `$nouns`. It's hard to make such mistakes
+    when working interactively, as Elvish highlights variables and complains
+    about nonexistent variables as you type.
+
+*   Double quotes support C-like escape sequences (``\n`` for newline,
+    etc.):
+
+    ```elvish-transcript
+    ~> echo "a\nb"
+    a
+    b
+    ```
+
+    **NOTE**: If you run `echo "a\nb"` in bash or zsh, you might get the same
+    result (depending on the value of some options), and this might lead you
+    to believe that they support C-like escape sequences in double quotes as
+    well. This is not the case; bash and zsh preserve the backslash in double
+    quotes, and it is the `echo` builtin command that interpret the escape
+    sequences. This difference becomes apparent if you change `echo` to
+    `touch`: In Elvish, `touch "a\nb"` creates a file whose name has a
+    newline; while in bash or zsh, it creates a file whose name contains a
+    backslash followed by `n`.
+
+*   Elementary floating-point arithmetics as well as comparisons are builtin,
+    with a prefix syntax:
+
+    ```elvish-transcript
+    ~> + 1 2
+    ▶ 3
+    ~> / (* 2 3) 4
+    ▶ 1.5
+    ~> > 1 2
+    ▶ $false
+    ~> < 1 2
+    ▶ $true
+    ```
+
+    **NOTE**: Elvish has special parsing rules to recognize `<` and `>` as
+    command names. That means that you cannot put redirections in the
+    beginning; bash and zsh allows `< input cat`, which is equivalent to
+    `cat < input`; in Elvish, you can only use the latter syntax.
+
+*   Functions are defined with `fn`. You can name arguments:
+
+    ```elvish-transcript
+    ~> fn square [x]{
+         * $x $x
+         }
+    ~> square 4
+    ▶ 16
+    ```
+
+*   Output of some builtin commands start with a funny `▶`. It is not part of
+    the output itself, but shows that such commands output a stream of values
+    instead of bytes. As such, their internal structures as well as boundaries
+    between values are preserved. This allows us to manipulate structured data
+    in the shell.
+
+    Read [semantics highlights](/learn/semantics-highlights.html) for details.
+
+*   When calling Elvish commands, options use the special `&option=value`
+    syntax. For instance, the `echo` builtin command supports a `sep` option
+    for specifying an alternative separator:
+
+    ```elvish-transcript
+    ~> echo &sep="," a b c
+    a,b,c
+    ```
+
+    The mixture of options and arguments is a classical problem in traditional
+    shell programming. For instance, if you want to print a file whose name is
+    in `$x`, `cat $x` is the obvious thing to do, but it does not do this
+    reliably -- if `$x` starts with `-` (e.g. `-v`), `cat` thinks that it is
+    an option. The correct way is `cat -- $x`.
+
+    Elvish commands are free from this problem. However, the option facility
+    is only available to builtin commands and user-defined functions, not
+    external commands, meaning that you still need to do `cat -- $x`.
+
+    In principle, it is possible to write safe wrapper
+    for external commands and there is a
+    [plan](https://github.com/elves/elvish/issues/371) for this.
diff -pruN 0.11+ds1-5/website/src/learn/effective-elvish.md 0.12+ds1-1/website/src/learn/effective-elvish.md
--- 0.11+ds1-5/website/src/learn/effective-elvish.md	1970-01-01 00:00:00.000000000 +0000
+++ 0.12+ds1-1/website/src/learn/effective-elvish.md	2018-07-11 02:32:25.000000000 +0000
@@ -0,0 +1,455 @@
+<!-- toc -->
+
+Elvish is not an entirely new language. Its programming techniques have two
+primary sources: traditional Unix shells and functional programming languages,
+both dating back to many decades ago. However, the way Elvish combines those
+two paradigms is unique in many ways, which enables new ways to write code.
+
+This document is an advanced tutorial focusing on how to write idiomatic
+Elvish code, code that is concise and clear, and takes full advantage of
+Elvish's features.
+
+An appropriate adjective for idiomatic Elvish code, like *Pythonic* for Python
+or *Rubyesque* for Ruby, is **Elven**. In [Roguelike
+games](https://en.wikipedia.org/wiki/Roguelike), Elven items are known to be
+high-quality, artful and resilient. So is Elven code.
+
+
+# Style
+
+## Naming
+
+Use `dash-delimited-words` for names of variables and functions. Underscores
+are allowed in variable and function names, but their use should be limited to
+environment variables (e.g. `$E:LC_ALL`) and external commands (e.g. `pkg_add`).
+
+When building a module, use a leading dash to communicate that a variable or
+function is subject to change in future and cannot be relied upon, either
+because it is an experimental feature or implementation detail.
+
+Elvish's core libraries follow the naming convention above.
+
+## Indentation
+
+Indent by two spaces.
+
+## Code Blocks
+
+In Elvish, code blocks in control structures are delimited by curly braces.
+This is perhaps the most visible difference of Elvish from most other shells
+like bash, zsh or fish. The following bash code:
+
+```bash
+if true; then
+  echo true
+fi
+```
+
+Is written like this in Elvish:
+
+```elvish
+if $true {
+  echo true
+}
+```
+
+If you have used lambdas in Elvish, you will notice that code blocks are
+syntactically just parameter-list-less lambdas.
+
+In Elvish, you cannot put opening braces of code blocks on the next line. This
+won't work:
+
+```elvish
+if $true
+{ # wrong!
+  echo true
+}
+```
+
+Instead, you must write:
+
+```elvish
+if $true {
+  echo true
+}
+```
+
+This is because in Elvish, control structures like `if` follow the same syntax
+as normal commands, hence newlines terminate them. To make the code block part
+of the `if` command, it must appear on the same line.
+
+
+# Using the Pipeline
+
+Elvish is equipped with a powerful tool for passing data: the pipeline. Like
+in traditional shells, it is an intuitive notation for data processing: data
+flows from left to right, undergoing one transformation after another. Unlike
+in traditional shells, it is not restricted to unstructured bytes: all Elvish
+values, including lists, maps and even closures, can flow in the pipeline.
+This section documents how to make the most use of pipelines.
+
+## Returning Values with Structured Output
+
+Unlike functions in most other programming languages, Elvish commands do not
+have return values. Instead, they can write to *structured output*, which is
+similar to the traditional byte-based stdout, but preserves all internal
+structures of aribitrary Elvish values. The most fundamental command that does
+this is `put`:
+
+```elvish-transcript
+~> put foo
+▶ foo
+~> x = (put foo)
+~> put $x
+▶ foo
+```
+
+This is hardly impressive - you can output and recover simple strings using
+good old byte-based output as well. But let's try this:
+
+```elvish-transcript
+~> put "a\nb" [foo bar]
+▶ "a\nb"
+▶ [foo bar]
+~> s li = (put "a\nb" [foo bar])
+~> put $s
+▶ "a\nb"
+~> put $li[0]
+▶ foo
+```
+
+Here, two things are worth mentioning: the first value we `put` contains a
+newline, and the second value is a list. When we capture the output, we get
+those exact values back. Passing structured data is difficult with byte-based
+output, but trivial with value output.
+
+Besides `put`, many other builtin commands also write to structured output,
+like `splits`:
+
+```elvish-transcript
+~> splits , foo,bar
+▶ foo
+▶ bar
+~> words = [(splits , foo,bar)]
+~> put $words
+▶ [foo bar]
+```
+
+User-defined functions behave in the same way: they "return" values by writing
+to structured stdout. Without realizing that "return values" are just outputs
+in Elvish, it is easy to think of `put` as **the** command to "return" values
+and write code like this:
+
+```elvish-transcript
+~> fn split-by-comma [s]{ put (splits , $s) }
+~> split-by-comma foo,bar
+▶ foo
+▶ bar
+```
+
+The `split-by-comma` function works, but it can be written more concisely
+as:
+
+```elvish-transcript
+~> fn split-by-comma [s]{ splits , $s }
+~> split-by-comma foo,bar
+▶ foo
+▶ bar
+```
+
+In fact, the pattern `put (some-cmd)` is almost always redundant and
+equivalent to just `some-command`.
+
+Similarly, it is seldom necessary to write `echo (some-cmd)`: it is almost
+always equivalent to just `some-cmd`. As an exercise, try simplifying the
+following function:
+
+```elvish
+fn git-describe { echo (git describe --tags --always) }
+```
+
+## Mixing Bytes and Values
+
+Each pipe in Elvish comprises two components: one traditional byte pipe that
+carries unstructured bytes, and one value pipe that carries Elvish values. You
+can write to both, and output capture will capture both:
+
+```elvish-transcript
+~> fn f { echo bytes; put value }
+~> f
+bytes
+▶ value
+~> outs = [(f)]
+~> put $outs
+▶ [bytes value]
+```
+
+This also illustrates that the output capture operator `(...)` works with both
+byte and value outputs, and it can recover the output sent to `echo`. When
+byte output contains multiple lines, each line becomes one value:
+
+```elvish-transcript
+~> x = [(echo "lorem\nipsum")]
+~> put $x
+▶ [lorem ipsum]
+```
+
+Most Elvish builtin functions also work with with both byte and value inputs.
+Similarly to output capture, they split their byte intput by newlines. For
+example:
+
+```elvish-transcript
+~> use str
+~> put lorem ipsum | each $str:to-upper~
+▶ LOREM
+▶ IPSUM
+~> echo "lorem\nipsum" | each $str:to-upper~
+▶ LOREM
+▶ IPSUM
+```
+
+This line-oriented processing of byte input is consistent with traditional
+Unix tools like `grep`, `sed` and `awk`. In fact, it is easy to write your own
+`grep` in Elvish:
+
+```elvish-transcript
+~> use re
+~> fn mygrep [p]{ each [line]{ if (re:match $p $line) { echo $line } } }
+~> cat in.txt
+abc
+123
+lorem
+456
+~> cat in.txt | mygrep '[0-9]'
+123
+456
+```
+
+(Note that it is more concise to write `mygrep ... < in.txt`, but due to [a
+bug](https://github.com/elves/elvish/issues/600) this does not work.)
+
+However, this line-oriented behavior is not always desirable: not all Unix
+commands output newline-separated data. When you want to get the output as is,
+as a single string, you can use the `slurp` command:
+
+```elvish-transcript
+~> echo "a\nb\nc" | slurp
+▶ "a\nb\nc\n"
+```
+
+One immediate use of `slurp` is to read a whole file into a string:
+
+```elvish-transcript
+~> cat hello.go
+package main
+
+import "fmt"
+
+func main() {
+            fmt.Println("vim-go")
+}
+~> hello-go = (slurp < hello.go)
+~> put $hello-go
+▶ "package main\n\nimport \"fmt\"\n\nfunc main()
+{\n\tfmt.Println(\"vim-go\")\n}\n"
+```
+
+It is also useful, for example, when working with NUL-separated output:
+
+```elvish-transcript
+~> touch "a\nb.go"
+~> mkdir d
+~> touch d/f.go
+~> find . -name '*.go' -print0 | splits "\000" (slurp)
+▶ "./a\nb.go"
+▶ ./d/f.go
+▶ ''
+```
+
+In the above command, `slurp` turns the input into one string, which is then
+used as an argument to `splits`. The `splits` command then splits the whole
+input by NUL bytes.
+
+Note that in Elvish, strings can contain NUL bytes; in fact, they can contain
+any byte; this makes Elvish suitable for working with binary data. (Also, note
+that the `find` command terminates its output with a NUL byte, hence we see a
+trailing empty string in the output.)
+
+One side note: In the first example, we saw that `bytes` appeared before `value`.
+This is not guaranteed: byte output and value output are separate, it is
+possible to get `value` before `bytes` in more complex cases. Writes to one
+component, however, always have their orders preserved, so in `put x; put y`,
+`x` will always appear before `y`.
+
+## Prefer Pipes Over Parentheses
+
+If you have experience with Lisp, you will discover that you can write Elvish
+code very similar to Lisp. For instance, to split a string containing
+comma-separated value, reduplicate each value (using commas as separators),
+and rejoin them with semicolons, you can write:
+
+```elvish-transcript
+~> csv = a,b,foo,bar
+~> joins ';' [(each [x]{ put $x,$x } [(splits , $csv)])]
+▶ 'a,a;b,b;foo,foo;bar,bar'
+```
+
+This code works, but it is a bit unreadable. In particular, since `splits`
+outputs multiple values but `each` wants a list argument, you have to wrap the
+output of `splits` in a list with `[(splits ...)]`. Then you have to do this
+again in order to pass the output of `each` to `joins`. You might wonder why
+commands like `splits` and `each` do not simply output a list to make this
+easier.
+
+The answer to that particular question is in the next subsection, but for the
+program at hand, there is a much better way to write it:
+
+```elvish-transcript
+~> csv = a,b,foo,bar
+~> splits , $csv | each [x]{ put $x,$x } | joins ';'
+▶ 'a,a;b,b;foo,foo;bar,bar'
+```
+
+Besides having fewer pairs of parentheses (and brackets), this program is also
+more readable, because the data flows from left to right, and there is no
+nesting. You can see that `$csv` is first split by commas, then each value
+gets reduplicated, and then finally everything is joined by semicolons. It
+matches exactly how you would describe the algorithm in spoken English -- or
+for that matter, any spoken language!
+
+Both versions work, because commands like `each` and `joins` that work with
+multiple inputs can take their inputs in two ways: they can take the inputs as
+one list argument, like in the first version; or from the pipeline, like the
+second version. Whenever possible, you should prefer the input-from-pipeline
+form: it makes for programs that have little nesting, read naturally.
+
+One exception to the recommendation is when the input is a small set of things
+known beforehand. For example:
+
+```elvish-transcript
+~> each $str:to-upper~ [lorem ipsum]
+▶ LOREM
+▶ IPSUM
+```
+
+Here, using the input-from-argument is completely fine: if you want to use the
+input-from-input form, you have to supply the input using `put`, which is also
+OK but a bit more wordy:
+
+```elvish-transcript
+~> put lorem ipsum | each $str:to-upper~
+▶ LOREM
+▶ IPSUM
+```
+
+However, not all commands support taking input from the pipeline. For example,
+if we want to first join some values with space and then split at commas, this
+won't work:
+
+```elvish-transcript
+~> joins ' ' [a,b c,d] | splits ,
+Exception: want 2 arguments, got 1
+[tty], line 1: joins ' ' [a,b c,d] | splits ,
+```
+
+This is because the `splits` command only ever works with one input (one
+string to split), and was not implemented to support taking input from
+pipeline; hence it always takes 2 arguments and we got an exception.
+
+It is easy to remedy this situation however. The `all` command passes its
+input to its output, and by capturing its output, we can turn the input into
+an argument:
+
+```elvish-transcript
+~> joins ' ' [a,b c,d] | splits , (all)
+▶ a
+▶ 'b c'
+▶ d
+```
+
+## Streaming Multiple Outputs
+
+In the previous subsection, we remarked that commands like `splits` and `each`
+write multiple output values instead of one list. Why?
+
+This has to do with another advantage of passing data through the pipeline: in
+a pipeline, all commands are executed in parallel. A command in a pipeline
+does not need to wait for its previous command to finish running before it can
+start processing data. Try this in your terminal:
+
+```elvish-transcript
+~> each $str:to-upper~ | each [x]{ put $x$x }
+(Start typing)
+abc
+▶ ABCABC
+xyz
+▶ XYZXYZ
+(Press ^D)
+```
+
+You will notice that as soon as you press Enter after typing `abc`, the output
+`ABCABC` is shown. As soon as one input is available, it goes through the
+entire pipeline, each command doing its work. This gives you immediate
+feedback, and makes good use of multi-core CPUs on modern computers. Pipelines
+are like assembly lines in the manufacturing industry.
+
+If instead of passing multiple values, we pass a list through the pipeline:
+that means that each command will now be waiting for its previous command to
+do all the processing and pack the results in a list before it can start doing
+anything. Now, although the commands themselves are run in parallel, they all
+need to be waiting for their previous commands to finish before they can start
+doing real work.
+
+This is why commands like `each` and `splits` produce multiple values instead
+of one list. When writing your functions, try to make them produce multiple
+values as well: they will cooperate better with builtin commands, and they can
+benefit from the efficiency of parallel computations.
+
+
+# Working with Multiple Values
+
+In Elvish, many constructs can evaluate to multiple values. This can be
+surprising if you are not familiar with it.
+
+To start with, output captures evaluate to all the captured values, instead of
+a list:
+
+```elvish-transcript
+~> splits , a,b,c
+▶ a
+▶ b
+▶ c
+~> li = (splits , a,b,c)
+Exception: arity mismatch
+[tty], line 1: li = (splits , a,b,c)
+```
+
+The assignment fails with "arity mismatch" because the right hand side
+evaluates to 3 values, but you are attempting to assign them to just one
+variable. If you want to capture the results into a list, you have to
+explicitly do so, either by constructing a list or using rest variables:
+
+```elvish-transcript
+~> li = [(splits , a,b,c)]
+~> put $li
+▶ [a b c]
+~> @li = (splits , a,b,c) # equivalent and slightly shorter
+```
+
+## Assigning Multiple Variables
+
+
+# To Be Continued...
+
+As of writing, Elvish is neither stable nor complete. The builtin libraries
+still have missing pieces, the package manager is in its early days, and
+things like a type system and macros have been proposed and considered, but
+not yet worked on. Deciding best practices for using feature *x* can be a bit
+tricky when that feature *x* doesn't yet exist!
+
+The current version of the document is what the lead developer of Elvish
+(@xiaq) has collected as best practices for writing Elvish code in early 2018,
+between the release of Elvish 0.11 and 0.12. They apply to aspects of the
+Elvish language that are relatively complete and stable; but as Elvish
+evolves, the document will co-evolve. You are invited to revisit this document
+once in a while!
diff -pruN 0.11+ds1-5/website/src/learn/fundamentals.md 0.12+ds1-1/website/src/learn/fundamentals.md
--- 0.11+ds1-5/website/src/learn/fundamentals.md	1970-01-01 00:00:00.000000000 +0000
+++ 0.12+ds1-1/website/src/learn/fundamentals.md	2018-07-11 02:32:25.000000000 +0000
@@ -0,0 +1,326 @@
+<!-- toc -->
+
+**This tutorial is quite incomplete, and it is being constantly expanded.**
+
+This tutorial introduces the fundamentals of shell programming with Elvish.
+It does not assume familiarity with other shells, but some understanding of
+basic programming concepts is required.
+
+This tutorial evolves around a "hello" program. But if your primary interest
+in shell programming is not saying hello (which, by the way, is a shame), the
+tutorial also contains some hints and examples on how to apply to knowledge to
+non-hello applications.
+
+**Note**: Elvish is very similar to other shell languages in many aspects, but
+also very different in others. When transferring your knowledge of Elvish to
+another shell, it is worthwhile to first check how things work there.
+
+
+# Hello, world!
+
+Let's begin with the most traditional "hello world" program. In Elvish,
+you invoke the `echo` **command** to print something on the terminal:
+
+```elvish-transcript
+~> echo "Hello, world!"
+Hello, world!
+```
+
+In Elvish, as in other shells, command invocations follow a simple structure: you
+write the command name, followed by arguments, all separated by spaces (or
+tabs). No parentheses or commas are needed.
+
+We enclose our text here in double quotes, making it a **string literal**.
+Compared to other languages, shell languages are a bit sloppy in that they
+allow you to write strings *without* quotes. The following also works:
+
+```elvish-transcript
+~> echo Hello, world!
+Hello, world!
+```
+
+However, the way it works has a subtle difference: here `Hello,` and `world!`
+are two arguments (remember that spaces separate arguments), and `echo` joins
+them together with a space. This is apparent if you put multiple spaces
+between them:
+
+```elvish-transcript
+~> echo Hello,      world!
+Hello, world!
+~> echo "Hello,     world!"
+Hello,     world!
+```
+
+When you write your message without quotes, no matter how many spaces there
+are, it is always the same two arguments `Hello,` and `world!`. If you quote
+your message, the spaces are part of the string and thus preserved.
+
+It is a good idea to always quote your string when it contains spaces or any
+special symbols other than period (`.`), dash (`-`) or underscore (`_`).
+
+
+## It Doesn't Have to be Hello
+
+All command invocations in shell have the same basic structure: name of
+command, followed by arguments. Elvish provides a lot of useful [builtin
+commands](/ref/builtin.html), and `echo` is just one of them. As another
+example, there is one for generating random numbers, which you can use as a
+digital dice:
+
+```elvish-transcript
+~> # randint a b generates an integer from range a...b-1
+   randint 1 7
+▶ 3
+```
+
+Arithmetic operations are also commands. Since they also follow the same order
+of command name first, the syntax deviates a bit from usual mathematical
+notations:
+
+```elvish-transcript
+~> * 17 28 # multiplication
+▶ 476
+~> ^ 2 10 # exponention
+▶ 1024
+```
+
+The commands introduced above -- `echo`, `randint`, `*` and `^` -- are all
+**builtin** commands, commands that Elvish provides for you.
+
+As a shell language, however, Elvish also makes it trivial to use **external**
+commands, commands implemented as separate programs. Chances are you have
+already used some of them like `ls` or `cat`. Here we show you how to obtain
+Elvish entirely from the command line: you can use `wget` to download files,
+`shasum` to verify its checksum, and `tar` to uncompress them, all of which
+are external commands:
+
+```elvish-transcript
+~> wget https://dl.elv.sh/elvish-linux-amd64-HEAD.tar.gz
+... omit ...
+elvish-linux.tar.gz  100%[======================>]   4.91M  10.9MB/s    in 0.4s
+~> shasum -a 256 elvish-linux.tar.gz
+0fc3c145a81345a1c49576b86ef12156d4eba1829e1bb20e9c39d115991a9c7b elvish-linux.tar.gz
+~> tar xvf elvish-linux.tar.gz
+x elvish
+```
+
+With the most basic knowledge of how to invoke commands, there is already a
+myriad of functionalities at your fingertip.
+
+
+# Hello, {insert user name}!
+
+The "hello world" program is a classic, but the fact that it always prints the
+same simple message does make it a little bit boring.
+
+One way to make programs more useful is to teach them to do different things
+depending on the context. In the case of our "hello world" program, why not
+teach it to greet whoever is running the program?
+
+```elvish-transcript
+~> echo "Hello, world! Hello, "$E:USER"!"
+Hello, world! Hello, xiaq!
+```
+
+There are several things happening here. First, `$E:USER` represents the `USER`
+**environment variable** ("E" being mnemonic for "environment"). In UNIX
+environments, it is usually set to the name of the current user. Second, we
+are running several strings and a variable all together: in this case, Elvish
+will concatenate them for you. Hence the result we see.
+
+Depending on your taste, you might feel that it's nicer to greet the world and
+the user on separate lines. To do this, we can insert `\n`, an **escape
+sequence** representing a newline in our string:
+
+```elvish-transcript
+~> echo "Hello, world!\nHello, "$E:USER"!"
+Hello, world!
+Hello, xiaq!
+```
+
+There are many such sequences starting with a backslash, including `\\` which
+represents the backslash itself. Beware that such escape sequences only work
+within double quotes.
+
+Environment variables are not the only way to learn about a computer system;
+we can also gain more information by invoking commands. For instance, the
+`uname` command tells you which operation system the computer is running:
+
+```elvish-transcript
+~> uname
+Darwin
+```
+
+([Darwin](https://en.wikipedia.org/wiki/Darwin_(operating_system)), by the
+way, is the open-source core of macOS and iOS.)
+
+To incorporate the output of `uname` into our hello message, we can first
+**capture** its output using parentheses and keep it in a variable using the
+assignment form `variable = value`:
+
+```elvish-transcript
+~> os = (uname)
+~> # Variable "os" now contains "Darwin"
+```
+
+We can then use this variable, in a similar way to how we used the environment
+variable, just without the `E:` namespace:
+
+```elvish-transcript
+~> echo "Hello, "$os" user!"
+Hello, Darwin user!
+```
+
+I used a variable for demonstration, but it's also possible to forego the
+variable and use the captured output directly:
+
+```elvish-transcript
+~> echo "Hello, "(uname)" user!"
+Hello, Darwin user!
+```
+
+## It Doesn't Have to be Hello
+
+Output captures can get you quite far in combining commands. For instance, you
+can use output captures to construct do complex arithmetic involving more than
+one operation:
+
+```elvish-transcript
+~> # compute the answer to life, universe and everything
+   * (+ 3 4) (- 100 94)
+▶ 42
+```
+
+
+<!--
+# Hello, everyone!
+
+
+Now let's say you want to say hello to several people, and typing `Hello` repeatedly is tiresome. You can save some work by using a **for-loop**:
+
+```elvish
+for name [Julius Pompey Marcus] {
+    echo 'Hello, '$name'!'
+}
+```
+
+In elvish you can put newlines between the elements to loop over, as long as they are terminated by `; do`.
+
+For easier reuse, you can also create a **list** to store the names:
+
+```elvish
+triumvirate = [Julius Pompey Marcus]
+```
+
+Lists are surrounded by square brackets, like in several other languages. Elements are separated by whitespaces.
+
+As you may have noticed, dashes are allowed in variable names. You are encouraged to use them instead of underscores; they are easier to type and more readable (after a little getting-used-to).
+
+Now it's time to use our list of the first triumvirate:
+
+```elvish
+for name in $first-triumvirate; do
+    echo 'Hello, '$name'!'
+done
+```
+
+This will, however, results in an error, saying that a string and a list cannot be concatenated. Why? Remember that `$x` is always one value. This is even true for lists, so the `for` loop only sees one value to loop over, namely the list itself.
+
+To make multiple words out of a list, you must explicitly **splice** the list with an `@` before the variable name:
+
+```elvish
+for name in $@first-triumvirate; do
+    echo 'Hello, '$name'!'
+done
+```
+
+# Each person gets $hello~'ed
+
+The for-loop we just show can also be written in a functional style:
+
+```elvish
+each [name]{
+    echo 'Hello, '$name'!'
+} $first-triumvirate
+```
+
+This looks similar to the for-loop version, but it makes use of a remarkable construct -- an **anonymous function**, also known as a **lambda**. In elvish, a lambda is syntactically formed by an argument list followed immediately (without space) by a function body enclosed in braces. Here, `[name]{ echo 'Hello, '$name'!' }` is a lambda that takes exactly one argument and calls `echo` to do the helloing. We pass it along a list to the `each` builtin, which runs the function on each element of the list.
+
+Functions, like strings and lists, can be stored in variables:
+
+```elvish
+hello=[name]{ echo 'Hello, '$name'!' }
+each $hello $first-triumvirate
+```
+
+To call a function, simply use it as a command:
+
+```elvish
+$hello 'Mark Antony' # Hello, Mark Anthony!
+```
+
+You must have noticed that you have to use `$hello` instead of `hello` to call the function. This is because the *hello-the-variable* and *hello-the-command* are different enitites. To define new commands, use the `fn` special form:
+
+```elvish
+fn hello [name]{
+    echo 'Hello, '$name'!'
+}
+hello Cicero # Hello, Cicero!
+```
+
+Users of traditional shells and Common Lisp will find this separation of the variable namespace and command namespace familiar.
+
+However, in elvish this separation is only superficial; what `fn hello` really does is just defining a variable called `hello~`. You can prove this:
+
+```elvish
+echo $hello~ # <closure ...>
+$hello~ Brutus # Hello, Brutus!
+each $hello~ $first-triumvirate # (Hello to the first triumvirate)
+```
+
+Conversely, defining a variable `hello~` will also create a command named `hello`:
+
+```elvish
+hello~ = [name]{ echo "Hello, hello, "$name"!" }
+hello Augustus # Hello, Augustus!
+```
+
+<!--
+```
+What I want to get into this document:
+
+[ ] Command substitution
+
+[ ] Rich pipeline
+
+[X] Lists
+
+[ ] Maps
+
+[X] Lambdas
+
+[X] fn
+
+[X] $&
+
+[X] One variable, one argument
+
+[X] String syntax
+
+[X] Lack of interpolation
+
+[X] Several builtins -- each println
+
+[ ] Editor API
+
+[ ] Exception and verdict
+
+[X] E: namespace for environment variables
+
+[ ] e: namespace for external commands
+
+[ ] Modules
+
+Write for readers with a moderate knowledge of a POSIXy shell (bash, zsh, ...)
+-->
+
diff -pruN 0.11+ds1-5/website/src/learn/index.toml 0.12+ds1-1/website/src/learn/index.toml
--- 0.11+ds1-5/website/src/learn/index.toml	1970-01-01 00:00:00.000000000 +0000
+++ 0.12+ds1-1/website/src/learn/index.toml	2018-07-11 02:32:25.000000000 +0000
@@ -0,0 +1,15 @@
+[[articles]]
+name = "fundamentals"
+title = "Fundamentals"
+
+[[articles]]
+name = "cookbook"
+title = "Cookbook"
+
+[[articles]]
+name = "unique-semantics"
+title = "Some Unique Semantics"
+
+[[articles]]
+name = "effective-elvish"
+title = "Effective Elvish"
diff -pruN 0.11+ds1-5/website/src/learn/unique-semantics.md 0.12+ds1-1/website/src/learn/unique-semantics.md
--- 0.11+ds1-5/website/src/learn/unique-semantics.md	1970-01-01 00:00:00.000000000 +0000
+++ 0.12+ds1-1/website/src/learn/unique-semantics.md	2018-07-11 02:32:25.000000000 +0000
@@ -0,0 +1,438 @@
+<!-- toc -->
+
+<div class="clear"></div>
+
+The semantics of Elvish is unique in many aspects when compared to other
+shells. This can be surprising if you are used to other shells, and it is a
+result of the [design choice](/ref/philosophy.html) of making Elvish a
+full-fledged programming language.
+
+# Structureful IO
+
+Elvish offers the ability to build elaborate data structures, "return" them
+from functions, and pass them through pipelines.
+
+## Motivation
+
+Traditional shells use strings for all kinds of data. They can be stored in
+variables, used as function arguments, written to output and read from input.
+Strings are very simple to use, but they fall short if your data has an
+inherent structure. A common solution is using pseudo-structures of "each line
+representing a record, and each (whitespace-separated) field represents a
+property", which is fine as long as your data do not contain whitespaces. If
+they do, you will quickly run into problems with escaping and quotation and
+find yourself doing black magics with strings.
+
+Some shells provide data structures like lists and maps, but they are usually
+not first-class values. You can store them in variables, but you might not to
+be able to nest them, pass them to functions, or returned them from functions.
+
+## Data Structures and "Returning" Them
+
+Elvish offers first-class support for data structures such as lists and maps.
+Here is an example that uses a list:
+
+```elvish-transcript
+~> li = [foo bar 'lorem ipsum']
+~> kind-of $li # "kind" is like type
+▶ list
+~> count $li # count the number of elements in a list
+▶ 3
+```
+
+(See the [language reference](/ref/language.html) for a more complete description
+of the builtin data structures.)
+
+As you can see, you can store lists in variables and use them as command
+arguments. But they would be much less useful if you cannot **return** them
+from a function. A naive way to do this is by `echo`ing the list and use output
+capture to recover it:
+
+```elvish-transcript
+~> fn f {
+     echo [foo bar 'lorem ipsum']
+   }
+~> li = (f) # (...) is output capture, like $(...) in other shells
+~> kind-of $li
+▶ string
+~> count $li # count the number of bytes, since $li is now a string
+▶ 23
+```
+
+As we have seen, our attempt to output a list has turned it into a string.
+This is because the `echo` command in Elvish, like in other shells, is
+string-oriented. To echo a list, it has to be first converted to a string.
+
+Elvish provides a `put` command to output structured values as they are:
+
+```elvish-transcript
+~> fn f {
+     put [foo bar 'lorem ipsum']
+   }
+~> li = (f)
+~> kind-of $li
+▶ list
+~> count $li
+▶ 3
+```
+
+So how does `put` work differently from `echo` under the hood?
+
+In Elvish, the standard output is made up of two parts: one traditional
+byte-oriented **file**, and one internal **value-oriented channel**. The
+`echo` command writes to the file, so it has to serialize its arguments into
+strings; the `put` command writes to the value-oriented channel, preserving
+all the internal structures of the values.
+
+If you invoke `put` directly from the command prompt, the values it output have
+a leading `▶`:
+
+```elvish-transcript
+~> put [foo bar]
+▶ [foo bar]
+```
+
+The leading arrow is a way to visualize that a command has written something
+onto the value channel, and not part of the value itself.
+
+In retrospect, you may discover that the `kind-of` and `count` builtin commands
+also write their output to the value channel.
+
+
+## Passing Data Structures Through the Pipeline
+
+When I said that standard output in Elvish comprises two parts, I was not
+telling the full story: pipelines in Elvish also have these two parts, in a
+very similar way. Data structures can flow in the value-oriented part of the
+pipeline as well. For instance, the `each` command takes **input** from the
+value-oriented channel, and apply a function to each value:
+
+```elvish-transcript
+~> put lorem ipsum | each [x]{ echo "Got "$x }
+Got lorem
+Got ipsum
+```
+
+There are many builtin commands that inputs or outputs values. As another
+example, the `take` commands retains a fixed number of items:
+
+```elvish-transcript
+~> put [lorem ipsum] "foo\nbar" [&key=value] | take 2
+▶ [lorem ipsum]
+▶ "foo\nbar"
+```
+
+## Interoperability with External Commands
+
+Unfortunately, the ability of passing structured values is not available to
+external commands. However, Elvish comes with a pair of commands for JSON
+serialization/deserialization. The following snippet illustrates how
+to interoperate with a Python script:
+
+```elvish-transcript
+~> cat sort-list.py
+import json, sys
+li = json.load(sys.stdin)
+li.sort()
+json.dump(li, sys.stdout)
+~> put [lorem ipsum foo bar] | to-json | python sort-list.py | from-json
+▶ [bar foo ipsum lorem]
+```
+
+It is easy to write a wrapper for such external commands:
+
+```elvish-transcript
+~> fn sort-list { to-json | python sort-list.py | from-json }
+~> put [lorem ipsum foo bar] | sort-list
+▶ [bar foo ipsum lorem]
+```
+
+More serialization/deserialization commands may be added to the language in
+the future.
+
+
+# Exit Status and Exceptions
+
+Unix commands exit with a non-zero value to signal errors. This is available
+traditionally as a `$?` variable in other shells:
+
+```bash
+true
+echo $? # prints "0"
+false
+echo $? # prints "1"
+```
+
+Builtin commands and user-defined functions also do this to signal errors,
+although they are not Unix commands:
+
+```bash
+bad() {
+  return 2
+}
+bad
+echo $? # prints "2"
+```
+
+This model is fine, only if most errors are non-fatal (so that errors from a
+previous command normally do not affect the execution of subsequence ones) and
+the script author remembers to check `$?` for the rare fatal errors.
+
+Elvish has no concept of exit status. Instead, it has exceptions that, when
+thrown, interrupt the flow of execution. The equivalency of the `bad` function
+in elvish is as follows:
+
+```elvish
+fn bad {
+  fail "bad things have happened" # throw an exception
+}
+bad # will print a stack trace and stop execution
+echo "after bad" # not executed
+```
+
+(If you run this interactively, you need to enter a literal newline after
+`bad` by pressing <span class="key">Alt-Enter</span> to make sure that it is
+executed in the same chunk as `echo "after bad"`.)
+
+And, non-zero exit status from external commands are turned into exceptions:
+
+```elvish
+false # will print a stack trace and stop execution
+echo "after false"
+```
+
+(If you are running a version of Elvish older than 0.9, you might need to use
+`e:false`, as Elvish used to have a builtin version of `false` that does
+something different.)
+
+An alternative way to describe this is that Elvish **does** have exit
+statuses, but non-zero exit statuses terminates execution by default.
+
+If you come from POSIX shell, this is aknin to `set -e` or having a `&&`
+between each pair of pipelines. In a sense, `&&` is written as `;` (or more
+commonly, a newline) in Elvish.
+
+Defaulting on stopping execution when bad things happen makes Elvish safer and
+code behavior more predictable.
+
+
+## Predicates and `if`
+
+The use of exit status is not limited to errors, however. In the Unix toolbox,
+quite a few commands exit with 0 to signal "true" and 1 to signal "false".
+Notably ones are:
+
+*   `test` aka `[`: testing file types, comparing numbers and strings;
+
+*   `grep`: exits with 0 when there are matches, with 1 otherwise;
+
+*   `diff`: exits with 0 when files are the same, with 1 otherwise;
+
+*   `true` and `false`, always exit with 0 and 1 respectively.
+
+The `if` control structure in POSIX shell is designed to work with such
+predicate commands: it takes a pipeline, and executes the body if the last
+command in the pipeline exits with 0. Examples:
+
+```sh
+# command 1
+if true; then
+  echo 'always executes'
+fi
+
+# command 2
+n=10
+if test $n -gt 2; then
+  echo 'executes when $n > 2'
+fi
+
+# command 3
+if diff a.txt b.txt; then
+  echo 'a.txt and b.txt are the same'
+fi
+```
+
+Since Elvish treats non-zero exit status as a kind of exception, the way that
+predicate commands and `if` work in POSIX shell does not work well for Elvish.
+Instead, Elvish's `if` is like most non-shell programming languages: it takes
+a value, and executes the body if the value is booleanly true. The first
+command above is written in Elvish as:
+
+```elvish
+if $true {
+  echo 'always executes'
+}
+```
+
+The way to write the second command in Elvish warrants an explanation of how
+predicates work in Elvish first. Predicates in Elvish simply write a boolean
+output, either `$true` or `$false`:
+
+```elvish-transcript
+~> > 10 2
+▶ $true
+~> > 1 2
+▶ $false
+```
+
+To use predicates in `if`, you simply capture its output with `()`. So the
+second command is written in Elvish as:
+
+```elvish
+n = 10
+if (> $n 2) {
+  echo 'executes when $n > 2'
+}
+```
+
+The parentheses after `if` are different to those in C: In C it is a
+syntactical requirement to put them around the condition; in Elvish, it
+functions as output capture operator.
+
+Sometimes it can be useful to have a condition on whether an external commands
+exits with 0. In this case, you can use the exception capture operator `?()`:
+
+```elvish
+if ?(diff a.txt b.txt) {
+  echo 'a.txt and b.txt are the same'
+}
+```
+
+In Elvish, all exceptions are booleanly false, while the special `$ok` value
+is booleanly true. If the `diff` exits with 0, the `?(...)` construct
+evaluates to `$ok`, which is booleanly true. Otherwise, it evaluates to an
+exception, which is booleanly false. Overall, this leads to a similar
+semantics with the POSIX `if` command.
+
+Note that the following code does have a severe downside: `?()` will prevent
+any kind of exceptions from throwing. In this case, we only want to turn one
+sort of exception into a boolean: `diff` exits with 1. If `diff` exits with 2,
+it usually means that there was a genuine error (e.g. `a.txt` does not exist).
+Swallowing this error defeats Elvish's philosophy of erring on the side of
+caution; a more sophisticated system of handling exit status is still being
+considered.
+
+
+# Phases of Code Execution
+
+A piece of code that gets evaluated as a whole is called a **chunk** (a
+loanword from Lua). If you run `elvish some-script.elv` from the command line,
+the entire script is one chunk; in interactive mode, each time you hit Enter,
+the code you have written is one chunk.
+
+Elvish interprets a code chunk in 3 phases: it first **parse**s the code into
+a syntax tree, then **compile**s the syntax tree code to an internal
+representation, and finally **evaluate**s the just-generated internal
+representation.
+
+If any error happens during the first two phases, Elvish rejects the chunk
+without executing any of it. For instance, in Elvish unclosed parenthesis is a
+syntax error (error in the first parsing phase). The following code, when
+executed as a chunk, does nothing other than printing the parse error:
+
+```elvish-bad
+echo before
+echo (
+```
+
+The same code, interpreted as bash, also contains a syntax error. However, if
+you save this file to `bad.bash` and run `bash bad.bash`, bash will execute the
+first line before complaining about the syntax error on the second line.
+
+Likewise, in Elvish using an unassigned variable is a compilation error, so the
+following code does nothing either:
+
+```elvish
+# assuming $nonexistent was not assigned
+echo before
+echo $nonexistent
+```
+
+There seems to be no equivalency of compilation errors in other shells, but
+this extra compilation phase makes the language safer. In future, optional
+type checking may be introduced, which will fit into the compilation phase.
+
+
+# Assignment Semantics
+
+In Python, JavaScript and many other languages, if you assign a container
+(e.g. a map) to multiple variables, modifications via those variables mutate
+the same container. This is best illustrated with an example:
+
+```python
+m = {'foo': 'bar', 'lorem': 'ipsum'}
+m2 = m
+m2['foo'] = 'quux'
+print(m['foo']) # prints "quux"
+```
+
+This is because in such languages, variables do not hold the "actual" map, but
+a reference to it. After the assignment `m2 = m`, both variables refer to the
+same map. The subsequent element assignment `m2['foo'] = 'quux'` mutates the
+underlying map, so `m['foo']`  is also changed.
+
+This is not the case for Elvish:
+
+```elvish-transcript
+~> m = [&foo=bar &lorem=ipsum]
+~> m2 = $m
+~> m2[foo] = quux
+~> put $m[foo]
+▶ bar
+```
+
+It seems that when you assign `m2 = $m`, the entire map is copied from `$m`
+into `$m2`, so any subsequent changes to `$m2` does not affect the original
+map in `$m`. You can entirely think of it this way: thinking
+**assignment as copying** correctly models the behavior of Elvish.
+
+But wouldn't it be expensive to copy an entire list or map every time
+assignment happens? No, the "copying" is actually very cheap. Is it
+implemented as [copy-on-write](https://en.wikipedia.org/wiki/Copy-on-write) --
+i.e. the copying is delayed until `$m2` gets modified? No, subsequent
+modifications to the new `$m2` is also very cheap. Read on if you are
+interested in how it is possible.
+
+## Implementation Detail: Persistent Data Structures
+
+Like in Python and JavaScript, Elvish variables like `$m` and `$m2` also only
+hold a reference to the underlying map. However, that map is **immutable**,
+meaning that they never change after creation. That explains why `$m` did not
+change: because the map `$m` refers to never changes. But how is it possible
+to do `m2[foo] = quux` if the map is immutable?
+
+The map implementation of Elvish has another property: although the map is
+immutable, it is easy to create a slight variation of one map. Given a map, it
+is easy to create another map that is almost the same, either 1) with one more
+key/value pair, or 2) with the value for one key changed, or 3) with one fewer
+key/value pair. This operation is fast, even if the original map is very
+large.
+
+This low-level functionality is exposed by the `assoc` (associate) and
+`dissoc` (dissociate) builtins:
+
+```elvish-transcript
+~> assoc [&] foo quux # "add" one pair
+▶ [&foo=quux]
+~> assoc [&foo=bar &lorem=ipsum] foo quux # "modify" one pair
+▶ [&lorem=ipsum &foo=quux]
+~> dissoc [&foo=bar &lorem=ipsum] foo # "remove" one pair
+▶ [&lorem=ipsum]
+```
+
+Now, although maps are immutable, variables are mutable. So when you try to
+assign an element of `$m2`, Elvish turns that into an assignment of `$m2`
+itself:
+
+```elvish
+m2[foo] = quux
+# is just syntax sugar for:
+m2 = (assoc $m2 foo quux)
+```
+
+The sort of immutable data structures that support cheap creation of "slight
+variations" are called [persistent data
+structures](https://en.wikipedia.org/wiki/Persistent_data_structure) and is
+used in functional programming languages. However, the way Elvish turns
+assignment to `$m2[foo]` into an assignment to `$m2` itself seems to be a new
+approach.
diff -pruN 0.11+ds1-5/website/src/ref/builtin.md 0.12+ds1-1/website/src/ref/builtin.md
--- 0.11+ds1-5/website/src/ref/builtin.md	1970-01-01 00:00:00.000000000 +0000
+++ 0.12+ds1-1/website/src/ref/builtin.md	2018-07-11 02:32:25.000000000 +0000
@@ -0,0 +1,1840 @@
+<!-- toc -->
+
+# Introduction
+
+The builtin module contains facilities that are potentially useful to all
+users. It occupies the `builtin:` namespace. You rarely have to explicitly
+specify the namespace though, since it is one of the namespaces consulted when
+resolving unqualified names.
+
+## Usage Notation
+
+The usage of a builtin command is described by giving an example usage, using
+variables as arguments. For instance, The `repeat` command takes two arguments
+and are described as:
+
+```elvish
+repeat $n $v
+```
+
+Optional arguments are represented with a trailing `?`, while variadic
+arguments with a trailing `...`. For instance, the `count` command takes an optional list:
+
+```elvish
+count $input-list?
+```
+
+While the `put` command takes an arbitrary number of arguments:
+
+```elvish
+put $values...
+```
+
+Options are given along with their default values. For instance, the `echo`
+command takes an `sep` option and arbitrary arguments:
+
+```elvish
+echo &sep=' ' $value...
+```
+
+(When you calling functions, options are always optional.)
+
+
+## Supplying Input
+
+Some builtin functions, e.g. `count` and `each`, can take their input in one
+of two ways:
+
+1. From pipe:
+
+    ```elvish-transcript
+    ~> put lorem ipsum | count # count number of inputs
+    2
+    ~> put 10 100 | each [x]{ + 1 $x } # apply function to each input
+    ▶ 11
+    ▶ 101
+    ```
+
+    Byte pipes are also possible; one line becomes one input:
+
+    ```elvish-transcript
+    ~> echo "a\nb\nc" | count # count number of lines
+    ▶ 3
+    ```
+
+1. From an argument -- an iterable value:
+
+    ```elvish-transcript
+    ~> count [lorem ipsum] # count number of elements in argument
+    2
+    ~> each [x]{ + 1 $x } [10 100] # apply to each element in argument
+    ▶ 11
+    ▶ 101
+    ```
+
+    Strings, and in future, other sequence types are also possible:
+
+    ```elvish-transcript
+    ~> count lorem
+    ▶ 5
+    ```
+
+When documenting such commands, the optional argument is always written as
+`$input-list?`. On the other hand, a trailing `$input-list?` always indicates
+that a command can take its input in one of two ways above: this fact is not
+repeated below.
+
+**Note**: You should prefer the first form, unless using it requires explicit
+`put` commands. Avoid `count [(some-command)]` or `each $some-func
+[(some-command)]`; they are, most of the time, equivalent to `some-command |
+count` or `some-command | each $some-func`.
+
+**Rationale**: An alternative way to design this is to make (say) `count` take
+an arbitrary number of arguments, and count its arguments; when there is 0
+argument, count inputs. However, this leads to problems in code like `count
+*`; the intention is clearly to count the number of files in the current
+directory, but when the current directory is empty, `count` will wait for
+inputs. Hence it is required to put the input in a list: `count [*]`
+unambiguously supplies input in the argument, even if there is no file.
+
+## Numerical Commands
+
+Commands that operate on numbers are quite flexible about the format of the
+numbers. Integers can be specified as decimals (e.g. `233`) or hexadecimals
+(e.g. `0xE9`) and floating-point numbers can be specified using the scientific
+notation (e.g. `2.33e2`). These are different strings, but equal when
+considered as commands.
+
+Elvish has no special syntax or data type for numbers. Instead, they are just
+strings. For this reason, builtin commands for strings and numbers are
+completely separate. For instance, the numerical equality command is `==`,
+while the string equality command is `==s`. Another example is the `+`
+builtin, which only operates on numbers and does not function as a string
+concatenation commands.
+
+
+## Predicates
+
+Predicates are functions that write exactly one output that is either `$true`
+or `$false`. They are described like "Determine ..." or "Test ...". See
+[`is`](#is) for one example.
+
+
+## "Do Not Use" Functions and Variables
+
+The name of some variables and functions have a leading `-`. This is a
+convention to say that it is subject to change and should not be depended
+upon. They are either only useful for debug purposes, or have known issues in
+the interface or implementation, and in the worst case will make Elvish crash.
+(Before 1.0, all features are subject to change, but those ones are sure to be
+changed.)
+
+Those functions and variables are documented near the end of the respective
+sections. Their known problem is also discussed.
+
+
+# Builtin Functions
+
+## + - * /
+
+```elvish
++ $summand...
+- $minuend $subtrahend...
+* $factor...
+/ $dividend $divisor...
+```
+
+Basic arithmetic operations of adding, substraction, multiplication and
+division respectively.
+
+All of them can take multiple arguments:
+
+```elvish-transcript
+~> + 2 5 7 # 2 + 5 + 7
+▶ 14
+~> - 2 5 7 # 2 - 5 - 7
+▶ -10
+~> * 2 5 7 # 2 * 5 * 7
+▶ 70
+~> / 2 5 7 # 2 / 5 / 7
+▶ 0.05714285714285715
+```
+
+When given one element, they all output their sole argument (given that it
+is a valid number). When given no argument,
+
+*   `+` outputs 0, and `*` outputs 1. You can think that they both have
+    a "hidden" argument of 0 or 1, which does not alter their behaviors (in
+    mathematical terms, 0 and 1 are [identity
+    elements](https://en.wikipedia.org/wiki/Identity_element) of addition and
+    multiplication, respectively).
+
+*   `-` throws an exception.
+
+*   `/` becomes a synonym for `cd /`, due to the implicit cd feature. (The
+    implicit cd feature will probably change to avoid this oddity).
+
+## %
+
+```elvish
+% $dividend $divisor
+```
+
+Output the remainder after dividing `$dividend` by `$divisor`. Both must be
+integers. Example:
+
+```elvish-transcript
+~> % 23 7
+▶ 2
+```
+
+## ^
+
+```elvish
+^ $base $exponent
+```
+
+Output the result of raising `$base` to the power of `$exponent`. Examples:
+
+```elvish-transcript
+~> ^ 2 10
+▶ 1024
+~> ^ 2 0.5
+▶ 1.4142135623730951
+```
+
+## &lt; &lt;= == != &gt; &gt;=
+
+```elvish
+<  $number... # less
+<= $number... # less or equal
+== $number... # equal
+!= $number... # not equal
+>  $number... # greater
+>= $number... # greater or equal
+```
+
+Number comparisons. All of them accept an arbitrary number of arguments:
+
+1.  When given fewer than two arguments, all output `$true`.
+
+2.  When given two arguments, output whether the two arguments satisfy the
+    named relationship.
+
+3.  When given more than two arguments, output whether every adjacent pair of
+    numbers satisfy the named relationship.
+
+Examples:
+
+```elvish-transcript
+~> == 3 3.0
+▶ $true
+~> < 3 4
+▶ $true
+~> < 3 4 10
+▶ $true
+~> < 6 9 1
+▶ $false
+```
+
+As a consequence of rule 3, the `!=` command outputs `$true` as long as any
+*adjacent* pair of numbers are not equal, even if some numbers that are not
+adjacent are equal:
+
+```elvish-transcript
+~> != 5 5 4
+▶ $false
+~> != 5 6 5
+▶ $true
+```
+
+
+## &lt;s &lt;=s ==s !=s &gt;s &gt;=s
+
+```elvish
+<s  $string... # less
+<=s $string... # less or equal
+==s $string... # equal
+!=s $string... # not equal
+>s  $string... # greater
+>=s $string... # greater or equal
+```
+
+String comparisons. They behave similarly to their number counterparts when
+given multiple arguments. Examples:
+
+```elvish-transcript
+~> >s lorem ipsum
+▶ $true
+~> ==s 1 1.0
+▶ $false
+~> >s 8 12
+▶ $true
+```
+
+
+## all
+
+```elvish
+all
+```
+
+Pass inputs, both bytes and values, to the output.
+
+This is an identity function in pipelines: `a | all | b` is equivalent to `a |
+b`. It is mainly useful for turning inputs into arguments, like:
+
+```elvish-transcript
+~> put 'lorem,ipsum' | splits , (all)
+▶ lorem
+▶ ipsum
+```
+
+Or capturing all inputs in a variable:
+
+```elvish-transcript
+~> x = [(all)]
+foo
+bar
+(Press ^D)
+~> put $x
+▶ [foo bar]
+```
+
+
+## assoc
+
+```elvish
+assoc $container $k $v
+```
+
+Output a slighly modified version of `$container`, such that its value at `$k`
+is `$v`. Applies to both lists and to maps.
+
+When `$container` is a list, `$k` may be a negative index. However, slice is
+not yet supported.
+
+
+```elvish-transcript
+~> assoc [foo bar quux] 0 lorem
+▶ [lorem bar quux]
+~> assoc [foo bar quux] -1 ipsum
+▶ [foo bar ipsum]
+~> assoc [&k=v] k v2
+▶ [&k=v2]
+~> assoc [&k=v] k2 v2
+▶ [&k2=v2 &k=v]
+```
+
+Etymology: [Clojure](https://clojuredocs.org/clojure.core/assoc).
+
+$cf dissoc
+
+
+## bool
+
+```elvish
+bool $value
+```
+
+Convert a value to boolean. In Elvish, only `$false` and errors are booleanly
+false. Everything else, including 0, empty strings and empty lists, is
+booleanly true:
+
+```elvish-transcript
+~> bool $true
+▶ $true
+~> bool $false
+▶ $false
+~> bool $ok
+▶ $true
+~> bool ?(fail haha)
+▶ $false
+~> bool ''
+▶ $true
+~> bool []
+▶ $true
+~> bool abc
+▶ $true
+```
+
+$cf not
+
+## cd
+
+```elvish
+cd $dirname
+```
+
+Change directory.
+
+Note that Elvish's `cd` does not support `cd -`.
+
+## constantly
+
+```elvish
+constantly $value...
+```
+
+Output a function that takes no arguments and outputs `$value`s when called.
+Examples:
+
+```elvish-transcript
+~> f=(constantly lorem ipsum)
+~> $f
+▶ lorem
+▶ ipsum
+```
+
+The above example is actually equivalent to simply `f = []{ put lorem ipsum }`;
+it is most useful when the argument is **not** a literal value, e.g.
+
+```elvish-transcript
+~> f = (constantly (uname))
+~> $f
+▶ Darwin
+~> $f
+▶ Darwin
+```
+
+The above code only calls `uname` once, while if you do `f = []{ put (uname) }`,
+every time you invoke `$f`, `uname` will be called.
+
+Etymology: [Clojure](https://clojuredocs.org/clojure.core/constantly).
+
+
+## count
+
+```elvish
+count $input-list?
+```
+
+Count the number of inputs.
+
+Examples:
+
+```elvish-transcript
+~> count lorem # count bytes in a string
+▶ 5
+~> count [lorem ipsum]
+▶ 2
+~> range 100 | count
+▶ 100
+~> seq 100 | count
+▶ 100
+```
+
+
+## dir-history
+
+```elvish
+dir-history
+```
+
+Return a list containing the directory history. Each element is a map
+with two keys: `path` and `score`. The list is sorted by descending
+score.
+
+Example:
+
+```elvish-transcript
+~> dir-history | take 1
+▶ [&path=/Users/foo/.elvish &score=96.79928]
+```
+
+## dissoc
+
+```elvish
+dissoc $map $k
+```
+
+Output a slightly modified version of `$map`, with the key `$k` removed. If
+`$map` does not contain `$k` as a key, the same map is returned.
+
+```elvish-transcript
+~> dissoc [&foo=bar &lorem=ipsum] foo
+▶ [&lorem=ipsum]
+~> dissoc [&foo=bar &lorem=ipsum] k
+▶ [&lorem=ipsum &foo=bar]
+```
+
+$cf assoc
+
+## drop
+
+```elvish
+drop $n $input-list?
+```
+
+Drop the first `$n` elements of the input. If `$n` is larger than the number of
+input elements, the entire input is dropped.
+
+Example:
+
+```elvish-transcript
+~> drop 2 [a b c d e]
+▶ c
+▶ d
+▶ e
+~> splits ' ' 'how are you?' | drop 1
+▶ are
+▶ 'you?'
+~> range 2 | drop 10
+```
+
+Etymology: Haskell.
+
+$cf take
+
+
+## each
+
+```elvish
+each $f $input-list?
+```
+
+Call `$f` on all inputs. Examples:
+
+```elvish-transcript
+~> range 5 8 | each [x]{ ^ $x 2 }
+▶ 25
+▶ 36
+▶ 49
+~> each [x]{ put $x[:3] } [lorem ipsum]
+▶ lor
+▶ ips
+```
+
+$cf peach
+
+Etymology: Various languages, as `for each`. Happens to have the same name as
+the iteration construct of
+[Factor](http://docs.factorcode.org/content/word-each,sequences.html).
+
+
+## eawk
+
+```elvish
+eawk $f $input-list?
+```
+
+For each input, call `$f` with the input followed by all its fields.
+
+It should behave the same as the following functions:
+
+```elvish
+fn eawk [f @rest]{
+  each [line]{
+    @fields = (re:split '[ \t]+'
+                        (re:replace '^[ \t]+|[ \t]+$' '' $line))
+    $f $line $@fields
+  } $@rest
+}
+```
+
+This command allows you to write code very similar to `awk` scripts using
+anonymous functions. Example:
+
+```elvish-transcript
+~> echo ' lorem ipsum
+   1 2' | awk '{ print $1 }'
+lorem
+1
+~> echo ' lorem ipsum
+   1 2' | eawk [line a b]{ put $a }
+▶ lorem
+▶ 1
+```
+
+
+## echo
+
+```elvish
+echo &sep=' ' $value...
+```
+
+Print all arguments, joined by the `sep` option, and followed by a newline.
+
+Examples:
+
+```elvish-transcript
+~> echo Hello   elvish
+Hello elvish
+~> echo "Hello   elvish"
+Hello   elvish
+~> echo &sep=, lorem ipsum
+lorem,ipsum
+```
+
+Notes: The `echo` builtin does not treat `-e` or `-n` specially. For instance,
+`echo -n` just prints `-n`. Use double-quoted strings to print special
+characters, and `print` to suppress the trailing newline.
+
+$cf print
+
+Etymology: Bourne sh.
+
+
+## eq
+
+```elvish
+eq $values...
+```
+
+Determine whether all `$value`s are structurally equivalent. Writes `$true`
+when given no or one argument.
+
+```elvish-transcript
+~> eq a a
+▶ $true
+~> eq [a] [a]
+▶ $true
+~> eq [&k=v] [&k=v]
+▶ $true
+~> eq a [b]
+▶ $false
+```
+
+$cf is not-eq
+
+Etymology: [Perl](https://perldoc.perl.org/perlop.html#Equality-Operators).
+
+
+## exec
+
+```elvish
+exec $command?
+```
+
+Replace the Elvish process with an external `$command`, defaulting to
+`elvish`.
+
+
+## exit
+
+```elvish
+exit $status?
+```
+
+Exit the Elvish process with `$status` (defaulting to 0).
+
+## explode
+
+```elvish
+explode $iterable
+```
+
+Put all elements of `$iterable` on the structured stdout. Like `flatten` in
+functional languages. Equivalent to `[li]{ put $@li }`.
+
+Example:
+
+```elvish-transcript
+~> explode [a b [x]]
+▶ a
+▶ b
+▶ [x]
+```
+
+Etymology: [PHP](http://php.net/manual/en/function.explode.php). PHP's
+`explode` is actually equivalent to Elvish's `splits`, but the author liked
+the name too much to not use it.
+
+
+## external
+
+```elvish
+external $program
+```
+
+Construct a callable value for the external program `$program`. Example:
+
+```elvish-transcript
+~> x = (external man)
+~> $x ls # opens the manpage for ls
+```
+
+$cf has-external search-external
+
+## fail
+
+```elvish
+fail $message
+```
+
+Throw an exception.
+
+```elvish-transcript
+~> fail bad
+Exception: bad
+Traceback:
+  [interactive], line 1:
+    fail bad
+~> put ?(fail bad)
+▶ ?(fail bad)
+```
+
+**Note**: Exceptions are now only allowed to carry string messages. You cannot
+do `fail [&cause=xxx]` (this will, ironically, throw a different exception
+complaining that you cannot throw a map). This is subject to change. Builtins
+will likely also throw structured exceptions in future.
+
+
+## fclose
+
+```elvish
+fclose $file
+```
+
+Close a file opened with `fopen`.
+
+$cf fopen
+
+
+## fopen
+
+```elvish
+fopen $filename
+```
+
+Open a file. Currently, `fopen` only supports opening a file for reading. File
+must be closed with `fclose` explicitly. Example:
+
+```elvish-transcript
+~> cat a.txt
+This is
+a file.
+~> f = (fopen a.txt)
+~> cat < $f
+This is
+a file.
+~> fclose $f
+```
+
+$cf fclose
+
+
+## from-json
+
+```elvish
+from-json
+```
+
+Takes bytes stdin, parses it as JSON and puts the result on structured stdout.
+The input can contain multiple JSONs, which can, but do not have to, be
+separated with whitespaces.
+
+Examples:
+
+```elvish-transcript
+~> echo '"a"' | from-json
+▶ a
+~> echo '["lorem", "ipsum"]' | from-json
+▶ [lorem ipsum]
+~> echo '{"lorem": "ipsum"}' | from-json
+▶ [&lorem=ipsum]
+~> # multiple JSONs running together
+   echo '"a""b"["x"]' | from-json
+▶ a
+▶ b
+▶ [x]
+~> # multiple JSONs separated by newlines
+   echo '"a"
+   {"k": "v"}' | from-json
+▶ a
+▶ [&k=v]
+```
+
+$cf to-json
+
+
+## has-external
+
+```elvish
+has-external $command
+```
+
+Test whether `$command` names a valid external command. Examples (your output
+might differ):
+
+```elvish-transcript
+~> has-external cat
+▶ $true
+~> has-external lalala
+▶ $false
+```
+
+$cf external search-external
+
+
+## has-prefix
+
+```elvish
+has-prefix $string $prefix
+```
+
+Determine whether `$prefix` is a prefix of `$string`. Examples:
+
+```elvish-transcript
+~> has-prefix lorem,ipsum lor
+▶ $true
+~> has-prefix lorem,ipsum foo
+▶ $false
+```
+
+## has-suffix
+
+```elvish
+has-suffix $string $suffix
+```
+
+Determine whether `$suffix` is a suffix of `$string`. Examples:
+
+```elvish-transcript
+~> has-suffix a.html .txt
+▶ $false
+~> has-suffix a.html .html
+▶ $true
+```
+
+
+## is
+
+```elvish
+is $values...
+```
+
+Determine whether all `$value`s have the same identity. Writes `$true` when
+given no or one argument.
+
+The definition of identity is subject to change. Do not rely on its behavior.
+
+```elvish-transcript
+~> is a a
+▶ $true
+~> is a b
+▶ $false
+~> is [] []
+▶ $true
+~> is [a] [a]
+▶ $false
+```
+
+$cf eq
+
+Etymology: [Python](https://docs.python.org/3/reference/expressions.html#is).
+
+
+## joins
+
+```elvish
+joins $sep $input-list?
+```
+
+Join inputs with `$sep`. Examples:
+
+```elvish-transcript
+~> put lorem ipsum | joins ,
+▶ lorem,ipsum
+~> joins , [lorem ipsum]
+▶ lorem,ipsum
+```
+
+The suffix "s" means "string" and also serves to avoid colliding with the
+well-known [join](https://en.wikipedia.org/wiki/join_(Unix)) utility.
+
+Etymology: Various languages as `join`, in particular
+[Python](https://docs.python.org/3.6/library/stdtypes.html#str.join).
+
+$cf splits
+
+
+## keys
+
+```elvish
+keys $map
+```
+
+Put all keys of `$map` on the structured stdout.
+
+Example:
+
+```elvish-transcript
+~> keys [&a=foo &b=bar &c=baz]
+▶ a
+▶ c
+▶ b
+```
+
+Note that there is no guaranteed order for the keys of a map.
+
+
+## kind-of
+
+```elvish
+kind-of $value...
+```
+
+Output the kinds of `$value`s. Example:
+
+```elvish-transcript
+~> kind-of lorem [] [&]
+▶ string
+▶ list
+▶ map
+```
+
+The terminology and definition of "kind" is subject to change.
+
+## nop
+
+```elvish
+nop &any-opt= $value...
+```
+
+Accepts arbitrary arguments and options and does exactly nothing.
+
+Examples:
+
+```elvish-transcript
+~> nop
+~> nop a b c
+~> nop &k=v
+```
+
+Etymology: Various languages, in particular NOP in [assembly
+languages](https://en.wikipedia.org/wiki/NOP).
+
+
+## not
+
+```elvish
+not $value
+```
+
+Boolean negation. Examples:
+
+```elvish-transcript
+~> not $true
+▶ $false
+~> not $false
+▶ $true
+~> not $ok
+▶ $false
+~> not ?(fail error)
+▶ $true
+```
+
+**NOTE**: `and` and `or` are implemented as special commands.
+
+$cf bool
+
+
+## not-eq
+
+```elvish
+not-eq $values...
+```
+
+Determines whether every adjacent pair of `$value`s are not equal. Note that
+this does not imply that `$value`s are all distinct. Examples:
+
+```elvish-transcript
+~> not-eq 1 2 3
+▶ $true
+~> not-eq 1 2 1
+▶ $true
+~> not-eq 1 1 2
+▶ $false
+```
+
+$cf eq
+
+
+## ord
+
+```elvish
+ord $string
+```
+
+Output value of each codepoint in `$string`, in hexadecimal. Examples:
+
+```elvish-transcript
+~> ord aA
+▶ 0x61
+▶ 0x41
+~> ord 你好
+▶ 0x4f60
+▶ 0x597d
+```
+
+The output format is subject to change.
+
+Etymology: [Python](https://docs.python.org/3/library/functions.html#ord).
+
+## path-*
+
+```elvish
+path-abs $path
+path-base $path
+path-clean $path
+path-dir $path
+path-ext $path
+```
+
+See [godoc of path/filepath](https://godoc.org/path/filepath). Go errors are
+turned into exceptions.
+
+## peach
+
+```elvish
+peach $f $input-list?
+```
+
+Call `$f` on all inputs, possibly in parallel.
+
+Example (your output will differ):
+
+```elvish-transcript
+~> range 1 7 | peach [x]{ + $x 10 }
+▶ 12
+▶ 11
+▶ 13
+▶ 16
+▶ 15
+▶ 14
+```
+
+This command is intended for homogenous processing of possibly unbound data.
+If you need to do a fixed number of heterogenous things in parallel, use
+`run-parallel`.
+
+$cf each run-parallel
+
+
+## pipe
+
+```elvish
+pipe
+```
+
+Create a new Unix pipe that can be used in redirections.
+
+A pipe contains both the read FD and the write FD. When redirecting command
+input to a pipe with `<`, the read FD is used. When redirecting command output
+to a pipe with `>`, the write FD is used. It is not supported to redirect both
+input and output with `<>` to a pipe.
+
+Pipes have an OS-dependent buffer, so writing to a pipe without an active
+reader does not necessarily block. Pipes **must** be explicitly closed with
+`prclose` and `pwclose`.
+
+Putting values into pipes will cause those values to be discarded.
+
+Examples (assuming the pipe has a large enough buffer):
+
+```elvish-transcript
+~> p = (pipe)
+~> echo 'lorem ipsum' > $p
+~> head -n1 < $p
+lorem ipsum
+~> put 'lorem ipsum' > $p
+~> head -n1 < $p
+# blocks
+# $p should be closed with prclose and pwclose afterwards
+```
+
+$cf prclose pwclose
+
+
+## prclose
+
+```elvish
+prclose $pipe
+```
+
+Close the read end of a pipe.
+
+$cf pwclose pipe
+
+
+## put
+
+```elvish
+put $value...
+```
+
+Takes arbitrary arguments and write them to the structured stdout.
+
+Examples:
+
+```elvish-transcript
+~> put a
+▶ a
+~> put lorem ipsum [a b] { ls }
+▶ lorem
+▶ ipsum
+▶ [a b]
+▶ <closure 0xc4202607e0>
+```
+
+Etymology: Various languages, in particular
+[C](https://manpages.debian.org/stretch/manpages-dev/puts.3.en.html) and [Ruby](https://ruby-doc.org/core-2.2.2/IO.html#method-i-puts) as `puts`.
+
+
+## pprint
+
+```elvish
+pprint $value...
+```
+
+Pretty-print representations of Elvish values. Examples:
+
+```elvish-transcript
+~> pprint [foo bar]
+[
+ foo
+ bar
+]
+~> pprint [&k1=v1 &k2=v2]
+[
+ &k2=
+  v2
+ &k1=
+  v1
+]
+```
+
+The output format is subject to change.
+
+$cf repr
+
+## print
+
+```elvish
+print &sep=' ' $value...
+```
+
+Like `echo`, just without the newline.
+
+$cf echo
+
+Etymology: Various languages, in particular
+[Perl](https://perldoc.perl.org/functions/print.html) and
+[zsh](http://zsh.sourceforge.net/Doc/Release/Shell-Builtin-Commands.html),
+whose `print`s do not print a trailing newline.
+
+
+## pwclose
+
+```elvish
+pwclose $pipe
+```
+
+Close the write end of a pipe.
+
+$cf prclose pipe
+
+
+## range
+
+```elvish
+range &step=1 $low? $high
+```
+
+Output `$low`, `$low` + `$step`, ..., proceeding as long as smaller than
+`$high`. If not given, `$low` defaults to 0.
+
+Examples:
+
+```elvish-transcript
+~> range 4
+▶ 0
+▶ 1
+▶ 2
+▶ 3
+~> range 1 6 &step=2
+▶ 1
+▶ 3
+▶ 5
+```
+
+Beware floating point oddities:
+
+```elvish-transcript
+~> range 0 0.8 &step=.1
+▶ 0
+▶ 0.1
+▶ 0.2
+▶ 0.30000000000000004
+▶ 0.4
+▶ 0.5
+▶ 0.6
+▶ 0.7
+▶ 0.7999999999999999
+```
+
+Etymology:
+[Python](https://docs.python.org/3/library/functions.html#func-range).
+
+
+## rand
+
+```elvish
+rand
+```
+
+Output a pseudo-random number in the interval [0, 1). Example:
+
+```elvish-transcript
+~> rand
+▶ 0.17843564133528436
+```
+
+## randint
+
+```elvish
+randint $low $high
+```
+
+Output a pseudo-random integer in the interval [$low, $high). Example:
+
+```elvish-transcript
+~> # Emulate dice
+   randint 1 7
+▶ 6
+```
+
+
+## repeat
+
+```elvish
+repeat $n $value
+```
+
+Output `$value` for `$n` times. Example:
+
+```elvish-transcript
+~> repeat 0 lorem
+~> repeat 4 NAN
+▶ NAN
+▶ NAN
+▶ NAN
+▶ NAN
+```
+
+Etymology: [Clojure](https://clojuredocs.org/clojure.core/repeat).
+
+## replaces
+
+```elvish
+replaces &max=-1 $old $repl $source
+```
+
+Replace all occurrences of `$old` with `$repl` in `$source`. If `$max` is
+non-negative, it determines the max number of substitutions.
+
+**Note**: `replaces` does not support searching by regular
+expressions, `$old` is always interpreted as a plain string. Use
+[re:replace](re.html#replace) if you need to search by regex.
+
+## repr
+
+```elvish
+repr $value...
+```
+
+Writes representation of `$value`s, separated by space and followed by a
+newline. Example:
+
+```elvish-transcript
+~> repr [foo 'lorem ipsum'] "aha\n"
+[foo 'lorem ipsum'] "aha\n"
+```
+
+$cf pprint
+
+Etymology: [Python](https://docs.python.org/3/library/functions.html#repr).
+
+
+## resolve
+
+```elvish
+resolve $command
+```
+
+Resolve `$command`. Command resolution is described in the [language
+reference](/ref/language.html). (TODO: actually describe it there.)
+
+Example:
+
+```elvish-transcript
+~> resolve echo
+▶ <builtin echo>
+~> fn f { }
+~> resolve f
+▶ <closure 0xc4201c24d0>
+~> resolve cat
+▶ <external cat>
+```
+
+
+## run-parallel
+
+```elvish
+run-parallel $callable ...
+```
+
+Run several callables in parallel, and wait for all of them to finish.
+
+If one or more callables throw exceptions, the other callables continue
+running, and a composite exception is thrown when all callables finish
+execution.
+
+The behavior of `run-parallel` is consistent with the behavior of pipelines,
+except that it does not perform any redirections.
+
+Here is an example that lets you pipe the stdout and stderr of a command to
+two different commands:
+
+```elvish
+pout = (pipe)
+perr = (pipe)
+run-parallel {
+  foo > $pout 2> $perr
+  pwclose $pout
+  pwclose $perr
+} {
+  bar < $pout
+  prclose $pout
+} {
+  bar2 < $perr
+  prclose $perr
+}
+```
+
+This command is intended for doing a fixed number of heterogenous things in
+parallel. If you need homogenous parallel processing of possibly unbound data,
+use `peach` instead.
+
+$cf peach
+
+
+## search-external
+
+```elvish
+search-external $command
+```
+
+Output the full path of the external `$command`. Throws an exception when not
+found. Example (your output might vary):
+
+```elvish-transcript
+~> search-external cat
+▶ /bin/cat
+```
+
+$cf external has-external
+
+## slurp
+
+```elvish
+slurp
+```
+
+Reads bytes input into a single string, and put this string on structured
+stdout.
+
+Example:
+
+```elvish-transcript
+~> echo "a\nb" | slurp
+▶ "a\nb\n"
+```
+
+Etymology: Perl, as
+[`File::Slurp`](http://search.cpan.org/~uri/File-Slurp-9999.19/lib/File/Slurp.pm).
+
+
+## splits
+
+```elvish
+splits $sep $string
+```
+
+Split `$string` by `$sep`. If `$sep` is an empty string, split it into
+codepoints.
+
+```elvish-transcript
+~> splits , lorem,ipsum
+▶ lorem
+▶ ipsum
+~> splits '' 你好
+▶ 你
+▶ 好
+```
+
+**Note**: `splits` does not support splitting by regular expressions,
+`$sep` is always interpreted as a plain string. Use
+[re:split](re.html#split) if you need to split by regex.
+
+Etymology: Various languages as `split`, in particular
+[Python](https://docs.python.org/3.6/library/stdtypes.html#str.split).
+
+$cf joins
+
+
+## src
+
+```elvish
+src
+```
+
+Output a map-like value describing the current source being evaluated. The
+value contains the following fields:
+
+*   `type`, which can be one of `interactive`, `script` or `module`;
+
+*   `name`, which is set to the name under which a script is executed or a
+    module is imported. It is an empty string when `type` = `interactive`;
+
+*   `path`, which is the path to the current source. It is an empty string
+    when `type` = `interactive`;
+
+*   `code`, which is the full body of the current source.
+
+Examples:
+
+```elvish-transcript
+~> put (src)[type name path code]
+▶ interactive
+▶ ''
+▶ ''
+▶ 'put (src)[type name path code]'
+~> echo 'put (src)[type name path code]' > foo.elv
+~> elvish foo.elv
+▶ script
+▶ foo.elv
+▶ /home/xiaq/foo.elv
+▶ "put (src)[type name path code]\n"
+~> echo 'put (src)[type name path code]' > ~/.elvish/lib/m.elv
+~> use m
+▶ module
+▶ m
+▶ /home/xiaq/.elvish/lib/m.elv
+▶ "put (src)[type name path code]\n"
+```
+
+Note: this builtin always returns information of the source of the **calling
+function**. Example:
+
+```elvish-transcript
+~> echo 'fn f { put (src)[type name path code] }' > ~/.elvish/lib/n.elv
+~> use n
+~> n:f
+▶ module
+▶ n
+▶ /home/xiaq/.elvish/lib/n.elv
+▶ "fn f { put (src)[type name path code] }\n"
+```
+
+
+## styled
+
+```elvish
+styled $object $style-transformer...
+```
+
+Construct a styled text by applying the supplied transformers to the supplied
+object. `$object` can be either a string, a styled segment (see below), a styled
+text or an arbitrary concatenation of them. A `$style-transformer` is either:
+
+* The name of a builtin style transformer
+  * On of the attribute names `bold`, `dim`, `italic`, `underlined`, `blink` or
+    `inverse` for setting the corresponding attribute
+  * An attribute name prefixed by `no-` for unsetting the attribute
+  * An attribute name prefixed by `toggle-` for toggling the attribute between set
+    and unset
+  * One of the color names `black`, `red`, `green`, `yellow`, `blue`, `magenta`,
+    `cyan`, `lightgray`, `gray`, `lightred`, `lightgreen`, `lightyellow`,
+    `lightblue`, `lightmagenta`, `lightcyan` or `white` for setting the text color
+  * A color name prefixed by `bg-` to set the background color
+* A lambda that receives a styled segment as the only argument and returns a
+  single styled segment
+* A function with the same properties as the lambda (provided via the
+  `$transformer~` syntax)
+
+When a styled text is converted to a string the corresponding [ANSI SGR code](https://en.wikipedia.org/wiki/ANSI_escape_code#SGR_.28Select_Graphic_Rendition.29_parameters)
+is built to render the style.
+
+A styled text is nothing more than a wrapper around a list of styled segments. They
+can be accessed by indexing into it.
+
+```elvish
+s = (styled abc red)(styled def green)
+put $s[0] $s[1]
+```
+
+
+## styled-segment
+
+```elvish
+styled-segment $object &fg-color=default &bg-color=default &bold=$false &dim=$false &italic=$false &underlined=$false &blink=$false &inverse=$false
+```
+
+Constructs a styled segment and is a helper function for styled transformers.
+`$object` can be a plain string, a styled segment or a concatenation thereof.
+Probably the only reason to use it is to build custom style transformers:
+
+```elvish
+fn my-awesome-style-transformer [seg]{ styled-segment $seg &bold=(not $seg[dim]) &dim=(not $seg[italic]) &italic=$seg[bold] }
+styled abc $my-awesome-style-transformer~
+```
+
+As just seen the properties of styled segments can be inspected by indexing into it.
+Valid indices are the same as the options to `styled-segment` plus `text`.
+
+```elvish
+s = (styled-segment abc &bold)
+put $s[text]
+put $s[fg-color]
+put $s[bold]
+```
+
+
+## take
+
+```elvish
+take $n $input-list?
+```
+
+Retain the first `$n` input elements. If `$n` is larger than the number of
+input elements, the entire input is retained. Examples:
+
+```elvish-transcript
+~> take 3 [a b c d e]
+▶ a
+▶ b
+▶ c
+~> splits ' ' 'how are you?' | take 1
+▶ how
+~> range 2 | take 10
+▶ 0
+▶ 1
+```
+
+
+Etymology: Haskell.
+
+
+## tilde-abbr
+
+```elvish
+tilde-abbr $path
+```
+
+If `$path` represents a path under the home directory, replace the
+home directory with `~`.  Examples:
+
+```elvish-transcript
+~> echo $E:HOME
+/Users/foo
+~> tilde-abbr /Users/foo
+▶ '~'
+~> tilde-abbr /Users/foobar
+▶ /Users/foobar
+~> tilde-abbr /Users/foo/a/b
+▶ '~/a/b'
+```
+
+
+## to-json
+
+```elvish
+to-json
+```
+
+Takes structured stdin, convert it to JSON and puts the result on bytes
+stdout.
+
+```elvish-transcript
+~> put a | to-json
+"a"
+~> put [lorem ipsum] | to-json
+["lorem","ipsum"]
+~> put [&lorem=ipsum] | to-json
+{"lorem":"ipsum"}
+```
+
+$cf from-json
+
+## to-string
+
+```elvish
+to-string $value...
+```
+
+Convert arguments to string values.
+
+```elvish-transcript
+~> to-string foo [a] [&k=v]
+▶ foo
+▶ '[a]'
+▶ '[&k=v]'
+```
+
+## wcswidth
+
+```elvish
+wcswidth $string
+```
+
+Output the width of `$string` when displayed on the terminal. Examples:
+
+```elvish-transcript
+~> wcswidth a
+▶ 1
+~> wcswidth lorem
+▶ 5
+~> wcswidth 你好，世界
+▶ 10
+```
+
+
+## -gc
+
+```elvish
+-gc
+```
+
+Force the Go garbage collector to run.
+
+This is only useful for debug purposes.
+
+
+## -ifaddrs
+
+```elvish
+-ifaddrs
+```
+
+Output all IP addresses of the current host.
+
+This should be part of a networking module instead of the builtin module.
+
+## -log
+
+```elvish
+-log $filename
+```
+
+Direct internal debug logs to the named file.
+
+This is only useful for debug purposes.
+
+
+## -stack
+
+```elvish
+-stack
+```
+
+
+Print a stack trace.
+
+This is only useful for debug purposes.
+
+
+## -source
+
+```elvish
+-source $filename
+```
+
+Read the named file, and evaluate it in the current scope.
+
+Examples:
+
+```elvish-transcript
+~> cat x.elv
+echo 'executing x.elv'
+foo = bar
+~> -source x.elv
+executing x.elv
+~> echo $foo
+bar
+```
+
+Note that while in the example, you can reference `$foo` after sourcing
+`x.elv`, putting the `-source` command and reference to `$foo` in the **same
+code chunk** (e.g. by using <span class="key">Alt-Enter</span> to insert a
+literal Enter, or using `;`) is invalid:
+
+```elvish-transcript
+~> # A new Elvish session
+~> cat x.elv
+echo 'executing x.elv'
+foo = bar
+~> -source x.elv; echo $foo
+Compilation error: variable $foo not found
+  [interactive], line 1:
+    -source x.elv; echo $foo
+```
+
+This is because the reading of the file is done in the evaluation phase, while
+the check for variables happens at the compilation phase (before evaluation).
+So the compiler has no evidence showing that `$foo` is actually valid, and
+will complain. (See [here](/learn/unique-semantics.html#execution-phases) for
+a more detailed description of execution phases.)
+
+To work around this, you can add a forward declaration for `$foo`:
+
+```elvish-transcript
+~> # Another new session
+~> cat x.elv
+echo 'executing x.elv'
+foo = bar
+~> foo = ''; -source x.elv; echo $foo
+executing x.elv
+bar
+```
+
+
+## -time
+
+```elvish
+-time $callable
+```
+
+Run the callable, and write the time used to run it. Example:
+
+```elvish-transcript
+~> -time { sleep 1 }
+1.006060647s
+```
+
+When the callable also produces outputs, they are a bit tricky to separate
+from the output of `-time`. The easiest workaround is to redirect the output
+into a temporary file:
+
+```elvish-transcript
+~> f = (mktemp)
+~> -time { { echo output; sleep 1 } > $f }
+1.005589823s
+~> cat $f
+output
+~> rm $f
+```
+
+
+# Builtin Variables
+
+## $_
+
+A blackhole variable.
+
+Values assigned to it will be discarded. Trying to use its value (like `put
+$_`) causes an exception.
+
+## $after-chdir
+
+A list of functions to run after changing directory. These functions are
+always called with directory to change it, which might be a relative path. The
+following example also shows `$before-chdir`:
+
+```elvish-transcript
+~> before-chdir = [[dir]{ echo "Going to change to "$dir", pwd is "$pwd }]
+~> after-chdir = [[dir]{ echo "Changed to "$dir", pwd is "$pwd }]
+~> cd /usr
+Going to change to /usr, pwd is /Users/xiaq
+Changed to /usr, pwd is /usr
+/usr> cd local
+Going to change to local, pwd is /usr
+Changed to local, pwd is /usr/local
+/usr/local>
+```
+
+$cf before-chdir
+
+## $args
+
+A list containing command-line arguments. Analogous to `argv` in some other
+languages. Examples:
+
+```elvish-transcript
+~> echo 'put $args' > args.elv
+~> elvish args.elv foo -bar
+▶ [foo -bar]
+~> elvish -c 'put $args' foo -bar
+▶ [foo -bar]
+```
+
+As demonstrated above, this variable does not contain the name of the script
+used to invoke it. For that information, use the `src` command.
+
+$cf src
+
+## $before-chdir
+
+A list of functions to run before changing directory. These functions are
+always called with the new working directory.
+
+$cf after-chdir
+
+## $false
+
+The boolean false value.
+
+## $ok
+
+The special value used by `?()` to signal absence of exceptions.
+
+## $value-out-indicator
+
+A string put before value outputs (such as those of of `put`). Defaults to
+`'▶ '`. Example:
+
+```elvish-transcript
+~> put lorem ipsum
+▶ lorem
+▶ ipsum
+~> value-out-indicator = 'val> '
+~> put lorem ipsum
+val> lorem
+val> ipsum
+```
+
+Note that you almost always want some trailing whitespace for readability.
+
+## $paths
+
+A list of search paths, kept in sync with `$E:PATH`. It is easier to use than
+`$E:PATH`.
+
+## $pid
+
+The process ID of the current Elvish process.
+
+## $pwd
+
+The present working directory. Setting this variable has the same effect as
+`cd`. This variable is most useful in temporary assignment.
+
+Example:
+
+```elvish
+## Updates all git repositories
+for x [*/] {
+    pwd=$x {
+        if ?(test -d .git) {
+            git pull
+        }
+    }
+}
+```
+
+Etymology: the `pwd` command.
+
+## $true
+
+The boolean true value.
diff -pruN 0.11+ds1-5/website/src/ref/bundled.md 0.12+ds1-1/website/src/ref/bundled.md
--- 0.11+ds1-5/website/src/ref/bundled.md	1970-01-01 00:00:00.000000000 +0000
+++ 0.12+ds1-1/website/src/ref/bundled.md	2018-07-11 02:32:25.000000000 +0000
@@ -0,0 +1,11 @@
+<!-- toc -->
+
+The Elvish binary is bundled with some Elvishscript modules.
+
+# readline-binding
+
+Contains readline-like keybindings. To use, put the following in `~/.elvish/rc.elv`:
+
+```elvish
+use readline-binding
+```
diff -pruN 0.11+ds1-5/website/src/ref/edit.md 0.12+ds1-1/website/src/ref/edit.md
--- 0.11+ds1-5/website/src/ref/edit.md	1970-01-01 00:00:00.000000000 +0000
+++ 0.12+ds1-1/website/src/ref/edit.md	2018-07-11 02:32:25.000000000 +0000
@@ -0,0 +1,610 @@
+<!-- toc -->
+
+The `edit:` module is the interface to the Elvish editor.
+
+Function usages are given in the same format as in the [reference for the
+builtin module](/ref/builtin.html).
+
+*This document is incomplete.*
+
+# Overview
+
+## Modes and Submodules
+
+The Elvish editor has different **modes**, and exactly one mode is active at
+the same time.  Each mode has its own UI and keybindings. For instance, the
+default **insert mode** lets you modify the current command. The **completion
+mode** (triggered by <span class="key">Tab</span> by default) shows you all
+candidates for completion, and you can use arrow keys to navigate those
+candidates.
+
+$ttyshot completion-mode
+
+Each mode has its own submodule under `edit:`. For instance, builtin functions
+and configuration variables for the completion mode can be found in the
+`edit:completion:` module.
+
+The primary modes supported now are `insert`, `completion`, `navigation`,
+`history`, `histlist`, `location`, and `lastcmd`. The last 4 are "listing
+modes", and their particularity is documented below.
+
+
+# Prompts
+
+Elvish has two prompts: the (normal) left-hand prompt and the right-side
+prompt (rprompt). Most of this section only documents the left-hand prompt,
+but API for rprompt is the same other than the variable name: just replace
+`prompt` with `rprompt`.
+
+To customize the prompt, assign a function to `edit:prompt`. The function may
+write value outputs or byte outputs. Value outputs may be either strings or
+`edit:styled` values; they are joiend with no spaces in between. Byte outputs
+are output as-is, including any newlines, but control characters will be
+escaped: you should use `edit:styled` to output styled text. If you mix value
+and byte outputs, the order in which they appear is non-deterministic.
+
+The default prompt and rprompt are equivalent to:
+
+```elvish
+edit:prompt = { tilde-abbr $pwd; put '> ' }
+edit:rprompt = (constantly (edit:styled (whoami)@(hostname) inverse))
+```
+
+More prompt functions:
+
+```elvish-transcript
+~> edit:prompt = { tilde-abbr $pwd; edit:styled '> ' green }
+~> # ">" is now green
+~> edit:prompt = { echo '$' }
+$
+# Cursor will be on the next line as `echo` outputs a trailing newline
+```
+
+## Stale Prompt
+
+Elvish never waits for the prompt function to finish. Instead, the prompt
+function is always executed on a separate thread, and Elvish updates the
+screen when the function finishes.
+
+However, this can be misleading when the function is slow: this means that the
+prompt on the screen may not contain the latest information. To deal with
+this, if the prompt function does not finish within a certain threshold - by
+default 0.2 seconds, Elvish marks the prompt as **stale**: it still shows the
+old stale prompt content, but transforms it using a **stale transformer**. The
+default stale transformer applies reverse-video to the whole prompt.
+
+The threshold is customizable with `$edit:prompt-stale-threshold`; it
+specifies the threshold in seconds.
+
+The transformer is customizable with `$edit:prompt-stale-transform`. It is a
+function; the function is called with no arguments, and `styled` values as
+inputs, and the output is interpreted in the same way as prompt functions.
+Since `styled` values can be used as outputs in prompt functions, a function
+that simply passes all the input values through as outputs is a valid stale
+transformer.
+
+As an example, try running following code:
+
+```elvish
+n = 0
+edit:prompt = { sleep 2; put $n; n = (+ $n 1); put ': ' }
+edit:-prompt-eagerness = 10 # update prompt on each keystroke
+edit:prompt-stale-threshold = 0.5
+```
+
+And then start typing. Type one character; the prompt becomes inverse after
+0.5 second: this is when Elvish starts to consider the prompt as stale. The
+prompt will return normal after 2 seconds, and the counter in the prompt is
+updated: this is when the prompt function finishes.
+
+Another thing you will notice is that, if you type a few characters quickly
+(in less than 2 seconds, to be precise), the prompt is only updated twice.
+This is because Elvish never does two prompt updates in parallel: prompt
+updates are serialized. If a prompt update is required when the prompt
+function is still running, Elvish simply queues another update. If an update
+is already queued, Elvish does not queue another update. The reason why
+exactly two updates happen in this case, and how this algorithm ensures
+freshness of the prompt is left as an exercise to the reader.
+
+
+## Prompt Eagerness
+
+The occassions when the prompt should get updated can be controlled with
+`$edit:-prompt-eagerness`:
+
+*   The prompt is always updated when the editor becomes active -- when Elvish
+    starts, or a command finishes execution, or when the user presses Enter.
+
+*   If `$edit-prompt-eagerness` >= 5, it is updated when the working directory
+    changes.
+
+*   If `$edit-prompt-eagerness` >= 10, it is updated on each keystroke.
+
+The default value is 5.
+
+
+## RPrompt Persistency
+
+By default, the rprompt is only shown while the editor is active: as soon as
+you press Enter, it is erased. If you want to keep it, simply set
+`$edit:rprompt-persistent` to `$true`:
+
+```elvish
+edit:rprompt-persistent = $true
+```
+
+
+# Keybindings
+
+Each mode has its own keybinding, accessible as the `binding` variable in its
+module. For instance, the binding table for insert mode is
+`$edit:insert:binding`. To see current bindings, simply print the binding
+table: `pprint $edit:insert:binding` (replace `insert` with any other mode).
+
+A binding tables is simply a map that maps keys to functions. For instance, to
+bind `Alt-x` in insert mode to exit Elvish, simply do:
+
+```elvish
+edit:insert:binding[Alt-x] = { exit }
+```
+
+Outputs from a bound function always appear above the Elvish prompt. You can see this by doing the following:
+
+```elvish
+edit:insert:binding[Alt-x] = { echo 'output from a bound function!' }
+```
+
+and press <span class="key">Alt-x</span> in insert mode. It allows you to put
+debugging outputs in bound functions without messing up the terminal.
+
+Internally, this is implemented by connecting their output to a pipe.
+This does the correct thing in most cases, but if you are sure you want to do
+something to the terminal, redirect the output to `/dev/tty`. For instance, the
+following binds <span class="key">Ctrl-L</span> to clearing the terminal:
+
+```elvish
+edit:insert:binding[Ctrl-L] = { clear > /dev/tty }
+```
+
+Bound functions have their inputs redirected to /dev/null.
+
+
+## Format of Keys
+
+TBD
+
+## Listing Modes
+
+The modes `histlist`, `loc` and `lastcmd` are all **listing modes**: They all
+show a list, and you can filter items and accept items.
+
+Because they are very similar, you may want to change their bindings at the
+same time. This is made possible by the `$edit:listing:binding` binding table
+(`listing` is not a "real" mode but an "abstract" mode). These modes still have
+their own binding tables like `$edit:histlist:binding`, and bindings there have
+highter precedence over those in the shared `$edit:listing:binding` table.
+
+Moreover, there are a lot of builtin functions in the `edit:listing` module
+like `edit:listing:down` (for moving down selection). They always apply to
+whichever listing mode is active.
+
+
+## Caveat: Bindings to Start Modes
+
+Note that keybindings to **start** modes live in the binding table of the
+insert mode, not the target mode. For instance, if you want to be able to use
+<span class="key">Alt-l</span> to start location mode, you should modify
+`$edit:insert:binding[Alt-l]`:
+
+```elvish
+edit:insert:binding[Alt-l] = { edit:location:start }
+```
+
+One tricky case is the history mode. You can press
+<span class="key">▲&#xfe0e;</span> to start searching for history, and continue
+pressing it to search further. However, when the first press happens, the
+editor is in insert mode, while with subsequent presses, the editor is in
+history mode. Hence this binding actually relies on two entries,
+`$edit:insert:binding[Up]` and `$edit:history:binding[Up]`.
+
+So for instance if you want to be able to use
+<span class="key">Ctrl-P</span> for this, you need to modify both bindings:
+
+```elvish
+edit:insert:binding[Up] = { edit:history:start }
+edit:history:binding[Up] = { edit:history:up }
+```
+
+
+# Completion API
+
+## Argument Completer
+
+There are two types of completions in Elvish: completion for internal data and
+completion for command arguments. The former includes completion for variable
+names (e.g. `echo $`<span class="key">Tab</span>) and indicies (e.g. `echo
+$edit:insert:binding[`<span class="key">Tab</span>). These are the completions
+that Elvish can provide itself because they only depend on the internal state
+of Elvish.
+
+The latter, in turn, is what happens when you type e.g. `cat `<span
+class="key">Tab</span>. Elvish cannot provide completions for them without full
+knowledge of the command.
+
+Command argument completions are programmable via the `$edit:completion:arg-completer`
+variable. When Elvish is completing an argument of command `$x`, it will call
+the value stored in `$edit:completion:arg-completer[$x]`, with all the existing arguments,
+plus the command name in the front.
+
+For example, if the user types `man 1`<span class="key">Tab</span>, Elvish will call:
+
+```elvish
+$edit:completion:arg-completer[man] man 1
+```
+
+If the user is starting a new argument when hitting <span
+class="key">Tab</span>, Elvish will call the completer with a trailing empty
+string. For instance, if you do `man 1`<span class="key">Space</span><span
+class="key">Tab</span>, Elvish will call:
+
+```elvish
+$edit:completion:arg-completer[man] man 1 ""
+```
+
+The output of this call becomes candidates. There are several ways of
+outputting candidates:
+
+*   Writing byte output, e.g. "echo cand1; echo cand2". Each line becomes a
+    candidate. This has the drawback that you cannot put newlines in
+    candidates. Only use this if you are sure that you candidates will not
+    contain newlines -- e.g. package names, usernames, but **not** file names,
+    etc..
+
+*   Write strings to value output, e.g. "put cand1 cand2". Each string output
+    becomes a candidate.
+
+*   Use the `edit:complex-candidate` command:
+
+    ```elvish
+    edit:complex-candidate &code-suffix='' &display-suffix='' &style='' $stem
+    ```
+
+    **TODO**: Document this.
+
+After receiving your candidates, Elvish will match your candidates against what
+the user has typed. Hence, normally you don't need to (and shouldn't) do any
+matching yourself.
+
+That means that in many cases you can (and should) simpy ignore the last
+argument to your completer. However, they can be useful for deciding what
+**kind** of things to complete. For instance, if you are to write a completer
+for `ls`, you want to see whether the last argument starts with `-` or not: if
+it does, complete an option; and if not, complete a filename.
+
+Here is a very basic example of configuring a completer for the `apt` command.
+It only supports completing the `install` and `remove` command and package
+names after that:
+
+```elvish
+all-packages = [(apt-cache search '' | eawk [0 1 @rest]{ put $1 })]
+
+edit:completion:arg-completer[apt] = [@args]{
+    n = (count $args)
+    if (== $n 2) {
+        # apt x<Tab> -- complete a subcommand name
+        put install uninstall
+    } elif (== $n 3) {
+        put $@all-packages
+    }
+}
+```
+
+Here is another slightly more complex example for the `git` command. It
+supports completing some common subcommands and then branch names after that:
+
+```elvish
+fn all-git-branches {
+    # Note: this assumes a recent version of git that supports the format
+    # string used.
+    git branch -a --format="%(refname:strip=2)" | eawk [0 1 @rest]{ put $1 }
+}
+
+common-git-commands = [
+  add branch checkout clone commit diff init log merge
+  pull push rebase reset revert show stash status
+]
+
+edit:arg-completer[git] = [@args]{
+    n = (count $args)
+    if (== $n 2) {
+        put $@common-git-commands
+    } elif (>= $n 3) {
+        all-git-branches
+    }
+}
+```
+
+
+## Matcher
+
+As stated above, after the completer outputs candidates, Elvish matches them
+with them with what the user has typed. For clarity, the part of the user input
+that is relevant to tab completion is called for the **seed** of the
+completion. For instance, in `echo x`<span class="key">Tab</span>, the seed is
+`x`.
+
+Elvish first indexes the matcher table -- `$edit:completion:matcher` -- with the
+completion type to find a **matcher**. The **completion type** is currently one
+of `variable`, `index`, `command`, `redir` or `argument`. If the
+`$edit:completion:matcher` lacks the suitable key, `$edit:completion:matcher['']` is used.
+
+
+Elvish then calls the matcher with one argument -- the seed, and feeds
+the *text* of all candidates to the input. The mather must output an identical
+number of booleans, indicating whether the candidate should be kept.
+
+As an example, the following code configures a prefix matcher for all
+completion types:
+
+```elvish
+edit:completion:matcher[''] = [seed]{ each [cand]{ has-prefix $cand $seed } }
+```
+
+Elvish provides three builtin matchers, `edit:match-prefix`,
+`edit:match-substr` and `edit:match-subseq`. In addition to conforming to the
+matcher protocol, they accept two options `&ignore-case` and `&smart-case`.
+For example, if you want completion of arguments to use prefix matching and
+ignore case, use:
+
+```elvish
+edit:completion:matcher[argument] = [seed]{ edit:match-prefix $seed &ignore-case=$true }
+```
+
+The default value of `$edit:completion:matcher` is `[&''=$edit:match-prefix~]`, hence
+that candidates for all completion types are matched by prefix.
+
+
+# Hooks
+
+Hooks are functions that are executed at certain points in time. In Elvish,
+this functionality is provided by lists of functions.
+
+There are current two hooks:
+
+*   `$edit:before-readline`, whose elements are called before the editor reads
+    code, with no arguments.
+
+*   `$edit:after-readline`, whose elements are called, after the editor reads
+    code, with a sole element -- the line just read.
+
+Example usage:
+
+```elvish
+edit:before-readline = [{ echo 'going to read' }]
+edit:after-readline = [[line]{ echo 'just read '$line }]
+```
+
+Then every time you accept a chunk of code (and thus leaving the editor),
+`just read ` followed by the code is printed; and at the very beginning of an
+Elvish session, or after a chunk of code is executed, `going to read` is
+printed.
+
+
+# Functions and Variables
+
+Functions and variables who name start with `-` are experimental, will have
+their names or behaviors changed in the near future. Others are more stable
+(unless noted explicitly) but are also subject to change before the 1.0
+release.
+
+## edit:-dump-buf
+
+Dump the content of onscreen buffer as HTML. This command is used to generate
+"ttyshots" on [the homepage](/).
+
+Example:
+
+```elvish
+ttyshot = ~/a.html
+edit:insert:binding[Ctrl-X] = { edit:-dump-buf > $tty }
+```
+
+
+## edit:complete-filename
+
+```elvish
+edit:complete-filename @args
+```
+
+Produces a list of filenames found in the directory of the last argument. All other arguments are ignored. If the last argument does not contain a path (either absolute or relative to the current directory), then the current directory is used. Relevant files are output as `edit:complex-candidate` objects.
+
+This function is the default handler for any commands without explicit handlers in `$edit:completion:arg-completer`. See [Argument Completer](#argument-completer).
+
+Example:
+
+```elvish-transcript
+~> edit:complete-filename ''
+▶ (edit:complex-candidate Applications &code-suffix=/ &display-suffix='' &style='01;34')
+▶ (edit:complex-candidate Books &code-suffix=/ &display-suffix='' &style='01;34')
+▶ (edit:complex-candidate Desktop &code-suffix=/ &display-suffix='' &style='01;34')
+▶ (edit:complex-candidate Docsafe &code-suffix=/ &display-suffix='' &style='01;34')
+▶ (edit:complex-candidate Documents &code-suffix=/ &display-suffix='' &style='01;34')
+...
+~> edit:complete-filename .elvish/
+▶ (edit:complex-candidate .elvish/aliases &code-suffix=/ &display-suffix='' &style='01;34')
+▶ (edit:complex-candidate .elvish/db &code-suffix=' ' &display-suffix='' &style='')
+▶ (edit:complex-candidate .elvish/epm-installed &code-suffix=' ' &display-suffix='' &style='')
+▶ (edit:complex-candidate .elvish/lib &code-suffix=/ &display-suffix='' &style='01;34')
+▶ (edit:complex-candidate .elvish/rc.elv &code-suffix=' ' &display-suffix='' &style='')
+```
+
+## edit:complete-getopt
+
+```elvish
+edit:complete-getopt $args $opts $handlers
+```
+
+Produces completions according to a specification of accepted command-line options (both short and long options are handled), positional handler functions for each command position, and the current arguments in the command line. The arguments are as follows:
+
+- `$args` is an array containing the current arguments in the command line (without the command itself). These are the arguments as passed to the  [Argument Completer](#argument-completer) function.
+- `$opts` is an array of maps, each one containing the definition of one possible command-line option. Matching options will be provided as completions when the last element of `$args` starts with a dash, but not otherwise. Each map can contain the following keys (at least one of `short` or `long` needs to be specified):
+  - `short` contains the one-letter short option, if any, without the dash.
+  - `long` contains the long option name, if any, without the initial two dashes.
+  - `arg-optional`, if set to `$true`, specifies that the option receives an optional argument.
+  - `arg-mandatory`, if set to `$true`, specifies that the option receives a mandatory argument. Only one of `arg-optional` or `arg-mandatory` can be set to `$true`.
+  - `desc` can be set to a human-readable description of the option which will be displayed in the completion menu.
+- `$handlers` is an array of functions, each one returning the possible completions for that position in the arguments. Each function receives as argument the last element of `$args`, and should return zero or more possible values for the completions at that point. The returned values can be plain strings or the output of `edit:complex-candidate`. If the last element of the list is the string `...`, then the last handler is reused for all following arguments.
+
+Example:
+
+```elvish-transcript
+~> for args [ [''] ['-'] ['-a' ''] [ arg1 '' ] [ arg1 arg2 '']] {
+     echo "args = "(to-string $args)
+     edit:complete-getopt $args [ [&short=a &long=all &desc="Show all"] ] [ [_]{ put first1 first2 } [_]{ put second1 second2 } ... ]
+   }
+args = ['']
+▶ first1
+▶ first2
+args = [-]
+▶ (edit:complex-candidate -a &code-suffix='' &display-suffix=' (Show all)' &style='')
+▶ (edit:complex-candidate --all &code-suffix='' &display-suffix=' (Show all)' &style='')
+args = [-a '']
+▶ first1
+▶ first2
+args = [arg1 '']
+▶ second1
+▶ second2
+args = [arg1 arg2 '']
+▶ second1
+▶ second2
+```
+
+## edit:styled
+
+```elvish
+edit:styled $text $styles
+```
+
+Construct an abstract styled text. The `$text` argument can be an arbitrary
+string, while the `$styles` argument is a list or semicolon-delimited string
+of the following styles:
+
+* Text styles: `bold`, `dim`, `italic`, `underlined`, `blink`, `inverse`.
+
+* Text colors: `black`, `red`, `green`, `yellow`, `blue`, `magenta`, `cyan`,
+  `lightgray`, and their corresponding light colors: `gray`, `lightred`,
+  `lightgreen`, `lightyellow`, `lightblue`, `lightmagenta`, `lightcyan` and
+  `white`.
+
+* Background colors: any of the text colors with a `bg-` prefix (e.g. `bg-red`
+  for red background), plus `bg-default` for default background color.
+
+* An [ANSI SGR
+  parameter](https://en.wikipedia.org/wiki/ANSI_escape_code#SGR_.28Select_Graphic_Rendition.29_parameters)
+  code (e.g. `1` for bold), subject to terminal support.
+
+Note that the result of `edit:styled` is an abstract data structure, not an
+ANSI sequence. However, it stringifies to an ANSI sequence, so you rarely have
+to convert it. To force a conversion, use `to-string`:
+
+```elvish-transcript
+~> edit:styled haha green
+▶ (edit:styled haha [green])
+~> echo (edit:styled haha green) # output is green
+haha
+~> to-string (edit:styled haha green)
+▶ "\e[32mhaha\e[m"
+```
+
+The forced conversion is useful when e.g. assigning to `$value-out-indicator`:
+
+```elvish-transcript
+~> value-out-indicator = (to-string (edit:styled '> ' green))
+~> put lorem ipsum # leading '> ' is green
+> lorem
+> ipsum
+```
+
+The styled text can be inspected by indexing:
+
+```elvish-transcript
+~> s = (edit:styled haha green)
+~> put $s[text] $s[styles]
+▶ haha
+▶ [green]
+```
+
+
+## $edit:current-command
+
+Contains the content of the current input. Setting the variable will cause the
+cursor to move to the very end, as if
+`edit-dot = (count $edit:current-command)` has been invoked.
+
+This API is subject to change.
+
+
+## $edit:-dot
+
+Contains the current position of the curosr, as a byte position within
+`$edit:current-command`.
+
+
+## $edit:max-height
+
+Maximum height the editor is allowed to use, defaults to `+Inf`.
+
+By default, the height of the editor is only restricted by the terminal
+height. Some modes like location mode can use a lot of lines; as a result, it
+can often occupy the entire terminal, and push up your scrollback buffer.
+Change this variable to a finite number to restrict the height of the editor.
+
+
+## $edit:completion:matcher
+
+See [the Matcher section](#matcher).
+
+
+## $edit:prompt
+
+See [Prompts](#prompts).
+
+
+## $edit:-prompt-eagerness
+
+See [Prompt Eagerness](#prompt-eagerness).
+
+
+## $edit:prompt-stale-threshold
+
+See [Stale Prompt](#stale-prompt).
+
+
+## $edit:prompt-stale-transformer
+
+See [Stale Prompt](#stale-prompt).
+
+
+## $edit:rprompt
+
+See [Prompts](#prompts).
+
+
+## $edit:-rprompt-eagerness
+
+See [Prompt Eagerness](#prompt-eagerness).
+
+
+## $edit:rprompt-persistent
+
+See [RPrompt Persistency](#rprompt-persistency).
+
+
+## $edit:rprompt-stale-threshold
+
+See [Stale Prompt](#stale-prompt).
+
+
+## $edit:rprompt-stale-transformer
+
+See [Stale Prompt](#stale-prompt).
diff -pruN 0.11+ds1-5/website/src/ref/epm.md 0.12+ds1-1/website/src/ref/epm.md
--- 0.11+ds1-5/website/src/ref/epm.md	1970-01-01 00:00:00.000000000 +0000
+++ 0.12+ds1-1/website/src/ref/epm.md	2018-07-11 02:32:25.000000000 +0000
@@ -0,0 +1,194 @@
+<!-- toc -->
+
+# Introduction
+
+The Elvish Package Manager (`epm`) is a module bundled with Elvish for
+managing third-party packages.
+
+In Elvish terminology, a **module** is a `.elv` file that can be imported with
+the `use` command, while a **package** is a collection of modules that are
+usually kept in the same repository as one coherent project and may have
+interdependencies. The Elvish language itself only deals with modules; the
+concept of package is a matter of how to organize modules.
+
+Like the `go` command, Elvish does **not** have a central registry of
+packages. A package is simply identified by the URL of its code repository,
+e.g. [github.com/elves/sample-pkg](https://github.com/elves/sample-pkg). To
+install the package, one simply uses the following:
+
+```elvish
+use epm
+epm:install github.com/elves/sample-pkg
+```
+
+`epm` knows out-of-the-box how to manage packages hosted in GitHub,
+BitBucket and GitLab, and requires the `git` command to be
+available. It can also copy files via `git` or `rsync` from arbitrary
+locations (see [Custom package domains](#custom-package-domains) for
+details).
+
+Once installed, modules in this package can be imported with `use
+github.com/elves/sample-pkg/...`. This package has a module named
+`sample-mod` containing a function `sample-fn`, and can be used like
+this:
+
+```elvish-transcript
+~> use github.com/elves/sample-pkg/sample-mod
+~> sample-mod:sample-fn
+This is a sample function in a sample module in a sample package
+```
+
+The next section describes functions in the `epm` module, using the
+same notation as the [doc for the builtin
+module](builtin.html#usage-notation).
+
+# Functions
+
+## install
+
+```elvish
+epm:install &silent-if-installed=$false $pkg...
+```
+
+Install the named packages. By default, if a package is already
+installed, a message will be shown. This can be disabled by passing
+`&silent-if-installed=$true`, so that already-installed packages are
+silently ignored.
+
+## installed
+
+```elvish
+epm:installed
+```
+
+Return an array with all installed packages. `epm:list` can be used as
+an alias for `epm:installed`.
+
+## is-installed
+
+```elvish
+epm:is-installed $pkg
+```
+
+Returns a boolean value indicating whether the given package is
+installed.
+
+## metadata
+
+```elvish
+epm:metadata $pkg
+```
+
+Returns a hash containing the metadata for the given package. Metadata
+for a package includes the following base attributes:
+
+- `name`: name of the package
+- `installed`: a boolean indicating whether the package is currently installed
+- `method`: method by which it was installed (`git` or `rsync`)
+- `src`: source URL of the package
+- `dst`: where the package is (or would be) installed. Note that this attribute is returned even if `installed` is `$false`.
+
+Additionally, packages can define arbitrary metadata attributes in a
+file called `metadata.json` in their top directory. The following attributes are recommended:
+
+- `description`: a human-readable description of the package
+- `maintainers`: an array containing the package maintainers, in `Name <email>` format.
+- `homepage`: URL of the homepage for the package, if it has one.
+
+## query
+
+```elvish
+epm:query $pkg
+```
+
+Pretty print the available metadata of the given package.
+
+## uninstall
+
+```elvish
+epm:uninstall $pkg...
+```
+
+Uninstall named packages.
+
+## upgrade
+
+```elvish
+epm:upgrade $pkg...
+```
+
+Upgrade named packages. If no package name is given, upgrade all installed
+packages.
+
+# Custom package domains
+
+Package names in `epm` have the following structure:
+`domain/path`. The `domain` is usually the hostname from where the
+package is to be fetched, such as `github.com`. The `path` can have
+one or more components separated by slashes. Usually, the full name of
+the package corresponds with the URL from where it can be fetched. For
+example, the package hosted at https://github.com/elves/sample-pkg is
+identified as `github.com/elves/sample-pkg`.
+
+Packages are stored under `~/.elvish/lib/` in a path identical to
+their name. For example, the package mentioned above is stored at
+`~/.elvish/lib/github.com/elves/sample-pkg`.
+
+Each domain must be configured with the following information:
+
+* The method to use to fetch packages from the domain. The two
+  supported methods are `git` and `rsync`.
+
+* The number of directory levels under the domain directory in which
+  the packages are found. For example, for `github.com` the number of
+  levels is 2, since package paths have two levels
+  (e.g. `elves/sample-pkg`). All packages from a given domain have the
+  same number of levels.
+
+* Depending on the method, other attributes are needed:
+
+    - `git` needs a `protocol` attribute, which can be `https` or
+      `http`, and determines how the URL is constructed.
+
+    - `rsync` needs a `location` attribute, which must be a valid source
+      directory recognized by the `rsync` command.
+
+`epm` includes default domain configurations for `github.com`,
+`gitlab.com` and `bitbucket.org`. These three domains share the same
+configuration:
+
+```json
+{
+   "method" : "git",
+   "protocol" : "https",
+   "levels" : "2"
+}
+```
+
+You can define your own domain by creating a file named
+`epm-domain.cfg` in the appropriate directory under
+`~/.elvish/lib/`. For example, if you want to define an `elvish-dev`
+domain which installs packages from your local `~/dev/elvish/`
+directory, you must create the file
+`~/.elvish/lib/elvish-dev/epm-domain.cfg` with the following JSON
+content:
+
+```json
+{
+   "method" : "rsync",
+   "location" : "~/dev/elvish",
+   "levels" : "1"
+}
+```
+
+You can then install any directory under `~/dev/elvish/` as a
+package. For example, if you have a directory
+`~/dev/elvish/utilities/`, the following command will install it under
+`~/.elvish/lib/elvish-dev/utilities`:
+
+```elvish
+epm:install elvish-dev/utilities
+```
+
+When you make any changes to your source directory, `epm:upgrade` will
+synchronize those changes to `~/.elvish/lib`.
diff -pruN 0.11+ds1-5/website/src/ref/index.toml 0.12+ds1-1/website/src/ref/index.toml
--- 0.11+ds1-5/website/src/ref/index.toml	1970-01-01 00:00:00.000000000 +0000
+++ 0.12+ds1-1/website/src/ref/index.toml	2018-07-11 02:32:25.000000000 +0000
@@ -0,0 +1,27 @@
+[[articles]]
+name = "philosophy"
+title = "The Philosophy"
+
+[[articles]]
+name = "language"
+title = "The Language"
+
+[[articles]]
+name = "builtin"
+title = "The Builtin Module"
+
+[[articles]]
+name = "edit"
+title = "The Editor Module (edit:)"
+
+[[articles]]
+name = "re"
+title = "The Regular Expression Module (re:)"
+
+[[articles]]
+name = "epm"
+title = "The Elvish Package Manager (epm:)"
+
+[[articles]]
+name = "bundled"
+title = "Bundled Modules"
diff -pruN 0.11+ds1-5/website/src/ref/language.md 0.12+ds1-1/website/src/ref/language.md
--- 0.11+ds1-5/website/src/ref/language.md	1970-01-01 00:00:00.000000000 +0000
+++ 0.12+ds1-1/website/src/ref/language.md	2018-07-11 02:32:25.000000000 +0000
@@ -0,0 +1,1699 @@
+<!-- toc number-sections -->
+
+# Introduction
+
+This document describes the Elvish programming language. It tries to be both a
+specification and an advanced tutorial; if it turns out to be impossible to
+do these two things at the same time, this document will evolve to a formal
+specification, and more readable tutorials will be created.
+
+Examples for one construct might use constructs that have not yet been
+introduced, so some familiarity with the language is assumed. If you are new
+to Elvish, start with the [learning materials](/learn).
+
+**Note to the reader**. Like Elvish itself, this document is a work in
+progress. Some materials are missing, and some are documented sparingly. If
+you have found something that should be improved -- even if there is already a
+"TODO" for it -- please feel free to ask on any of the chat channels
+advertised on the [homepage](/). Some developer will explain to you, and then
+update the document. Question-driven documentation :)
+
+
+# Syntax Convention
+
+Elvish source code must be UTF-8-encoded. In this document, **character** is a
+synonym of [Unicode codepoint](https://en.wikipedia.org/wiki/Code_point) or
+its UTF-8 encoding.
+
+Also like most shells, Elvish uses whitespaces -- instead of commas, periods
+or semicolons -- to separate constructs. In this document, an **inline
+whitespace** is any of:
+
+*   A space (ASCII 0x20) or tab (ASCII 0x9, `"\t"`);
+
+*   A comment: starting with `#` and ending before the next newline or end of
+    file;
+
+*   Line continuation: a backslash followed by a newline.
+
+A **whitespace** is either an **inline whitespace** or a newline (`"\n"`).
+
+Like most shells, Elvish has a syntax structure that can be divided into two
+levels: a **statement** level and an **expression** level. For instance, on
+the expression level, `"echo"` is a quoted string that evaluates to `echo`;
+but on the statement level, it is a command that outputs an empty line. This
+distinction is often clear from the context and sometime the verb used. A
+statements **executes** to produce side effects; an expression **evaluates**
+to some values. (The traditional terms for the two levels are "commands" and
+"words", but those terms are quite ambiguous.)
+
+
+# String
+
+The most common data structure in shells is the string. String literals can be
+quoted or unquoted (barewords).
+
+## Quoted
+
+There are two types of quoted strings in Elvish, single-quoted strings and
+double-quoted strings.
+
+In single-quoted strings, all characters represent themselves, except single
+quotes, which need to be doubled. For instance, ``'*\'`` evaluates to ``*\``,
+and ``'it''s'`` evaluates to ``it's``.
+
+In double-quoted strings, the backslash ``\`` introduces a **escape
+sequence**. For instance, ``"\n"`` evaluates to a newline; ``"\\"`` evaluates
+to a backslash; invalid escape sequences like ``"\*"`` result in a syntax
+error.
+
+**TODO**: Document the full list of supported escape sequences.
+
+Unlike most other shells, double-quoted strings do not support interpolation.
+For instance, `"$USER"` simply evaluates to the string `$USER`. To get a
+similar effect, simply concatenate strings: instead of `"my name is $name"`,
+write `"my name is "$name`. Under the hood this is a [compound
+expression](#compound-expression-and-braced-lists).
+
+## Barewords
+
+If a string only consists of bareword characters, it can be written without
+any quote; this is called a **bareword**. Examples are `a.txt`,
+`long-bareword`, and `/usr/local/bin`. The set of bareword characters include:
+
+*   ASCII letters (a-z and A-Z) and numbers (0-9);
+
+*   The symbols `-_:%+,./@!`;
+
+*   Non-ASCII codepoints that are printable, as defined by
+    [unicode.IsPrint](https://godoc.org/unicode#IsPrint) in Go's standard
+    library.
+
+The following are bareword characters depending on their position:
+
+*   The tilde `~`, unless it appears at the beginning of a compound
+    expression, in which case it is subject to [tilde
+    expansion](#tilde-expansion);
+
+*   The equal sign `=`, unless it is used for terminating [map keys](#map) or
+    [option keys](#arguments-and-options), or denoting
+    [assignments](#assignment) or [temporary
+    assignments](#temporary-assignment).
+
+Unlike traditional shells, an unquoted backslash ``\`` does not escape
+metacharacters; use quoted strings instead. For instance, to echo a star, write
+`echo "*"` or `echo '*'`, not ``echo \*``. Unquote backslashes are now only
+used in line continuations; their use elsewhere is reserved will cause a
+syntax error.
+
+## Notes
+
+The three syntaxes above all evaluate to strings, and they are
+interchangeable. For instance, `xyz`, `'xyz'` and `"xyz"` are different
+syntaxes for the same string, and they are always equivalent.
+
+Elvish does **not** have a separate number type. For instance, in the command
+`+ 1 2`, both `1` and `2` are strings, and it is the command `+` that knows to
+treat its arguments as numbers. This design is driven by syntax -- because
+barewords are always treated as strings, and digits are barewords, we cannot
+treat words like `1` as number literals. At some point the language may get a
+dedicated number type (or several number types), but they will likely need to
+be constructed explicitly, e.g. `(number 1)`.
+
+
+# List and Map
+
+Lists and maps are the basic container types in Elvish.
+
+## List
+
+Lists are surround by square brackets `[ ]`, with elements separated by
+whitespaces. Examples:
+
+```elvish-transcript
+~> put [lorem ipsum]
+▶ [lorem ipsum]
+~> put [lorem
+        ipsum
+        foo
+        bar]
+▶ [lorem ipsum foo bar]
+```
+
+Note that commas have no special meanings and are valid bareword characters,
+so don't use them to separate elements:
+
+```elvish-transcript
+~> li = [a, b]
+~> put $li
+▶ [a, b]
+~> put $li[0]
+▶ a,
+```
+
+## Map
+
+Maps are also surrounded by square brackets; a key/value pair is written
+`&key=value` (reminiscent to HTTP query parameters), and pairs are separated
+by whitespaces. Whitespaces are allowed after `=`, but not before `=`.
+Examples:
+
+```elvish-transcript
+~> put [&foo=bar &lorem=ipsum]
+▶ [&foo=bar &lorem=ipsum]
+~> put [&a=   10
+        &b=   23
+        &sum= (+ 10 23)]
+▶ [&a=10 &b=23 &sum=33]
+```
+
+An empty map is written as `[&]`.
+
+
+# Variable
+
+Variables are named holders of values. The following characters can be used in
+variable names (a subset of bareword characters):
+
+*   ASCII letters (a-z and A-Z) and numbers (0-9);
+
+*   The symbols `-_:~`. The colon `:` is special; it is normally used for
+    separating namespaces or denoting namespace variables;
+
+*   Non-ASCII codepoints that are printable, as defined by
+    [unicode.IsPrint](https://godoc.org/unicode#IsPrint) in Go's standard
+    library.
+
+In most other shells, variables can map directly to environmental variables:
+`$PATH` is the same as the `PATH` environment variable. This is not the case
+in Elvish. Instead, environment variables are put in a dedicated `E:`
+namespace; the environment variable `PATH` is known as `$E:PATH`. The `$PATH`
+variable, on the other hand, does not exist initially, and if you have defined
+it, only lives in a certain lexical scope within the Elvish interpreter.
+
+You will notice that variables sometimes have a leading dollar `$`, and
+sometimes not. The tradition is that they do when they are used for their
+values, and do not otherwise (e.g. in assignment). This is consistent with
+most other shells.
+
+## Assignment
+
+A variable can be assigned by writing its name, `=`, and the value to assign.
+There must be inline whitespaces both before and after `=`. Example:
+
+```elvish-transcript
+~> foo = bar
+```
+
+You can assign multiple values to multiple variables simultaneously, simply by writing several variable names (separated by inline whitespaces) on the left-hand side, and several values on the right-hand side:
+
+```elvish-transcript
+~> x y = 3 4
+```
+
+## Referencing
+
+Use a variable by adding `$` before the name:
+
+```elvish-transcript
+~> foo = bar
+~> x y = 3 4
+~> put $foo
+▶ bar
+~> put $x
+▶ 3
+```
+
+Variables must be assigned before use. Attempting to use an unassigned
+variable causes a compilation error:
+
+```elvish-transcript
+~> echo $x
+Compilation error: variable $x not found
+[tty], line 1: echo $x
+~> { echo $x }
+Compilation error: variable $x not found
+[tty], line 1: { echo $x }
+```
+
+## Explosion and Rest Variable
+
+If a variable contains a list value, you can add `@` before the variable name
+to get all its element values. This is called **exploding** the variable:
+
+```elvish-transcript
+~> li = [lorem ipsum foo bar]
+~> put $li
+▶ [lorem ipsum foo bar]
+~> put $@li
+▶ lorem
+▶ ipsum
+▶ foo
+▶ bar
+```
+
+(This notation is restricted to exploding variables. To explode arbitrary values, use the builtin [explode](/ref/builtin.html#explode) command.)
+
+When assigning variables, if you prefix the name of the last variable with
+`@`, it gets assigned a list containing all remaining values. That variable is
+called a **rest variable**. Example:
+
+```elvish-transcript
+~> a b @rest = 1 2 3 4 5 6 7
+~> put $a $b $rest
+▶ 1
+▶ 2
+▶ [3 4 5 6 7]
+```
+
+Schematically this is a reverse operation to variable explosion, which is why
+they share the `@` sign.
+
+
+## Temporary Assignment
+
+You can prepend a command with **temporary assignments**, which gives
+variables temporarily values during the execution of that command.
+
+In the following
+example, `$x` and `$y` are temporarily assigned 100 and 200:
+
+```elvish-transcript
+~> x y = 1 2
+~> x=100 y=200 + $x $y
+▶ 300
+~> echo $x $y
+1 2
+```
+
+In contrary to normal assignments, there should be no whitespaces around the
+equal sign `=`. To have multiple variables in the left-hand side, use braces:
+
+```elvish-transcript
+~> x y = 1 2
+~> fn f { put 100 200 }
+~> {x,y}=(f) + $x $y
+▶ 300
+```
+
+If you use a previously undefined variable in a temporary assignment, its
+value will become the empty string after the command finishes. This behavior
+will likely change; don't rely on it.
+
+Since ordinary assignments are also a kind of command, they can also be
+prepended with temporary assignments:
+
+```elvish-transcript
+~> x=1
+~> x=100 y = (+ 133 $x)
+~> put $x $y
+▶ 1
+▶ 233
+```
+
+Temporary assignments must all appear before the command. As soon as something
+that is not a temporary assignments is parsed, Elvish no longer parses
+temporary assignments. For instance, in `x=1 echo x=1`, the second `x=1` is not
+a temporary assignment, but a bareword.
+
+**Note**: Elvish's behavior differs from bash (or zsh) in one important place.
+In bash, temporary assignments to variables do not affect their direct
+appearance in the command:
+
+```sh-transcript
+bash-4.4$ x=1
+bash-4.4$ x=100 echo $x
+1
+```
+
+
+## Scoping rule
+
+Elvish has lexical scoping. Scopes are introduced by [lambdas](#lambda) or
+[user-defined modules](#user-defined-modules).
+
+When you use a variable, Elvish looks for it in the current lexical scope,
+then its parent lexical scope and so forth, until the outermost scope:
+
+```elvish-transcript
+~> x = 12
+~> { echo $x } # $x is in the global scope
+12
+~> { y = bar; { echo $y } } # $y is in the outer scope
+bar
+```
+
+If a variable is not in any of the lexical scopes, Elvish tries to resolve it
+in the `builtin:` namespace, and if that also fails, cause an error:
+
+```elvish-transcript
+~> echo $pid # builtin
+36613
+~> echo $nonexistent
+Compilation error: variable $nonexistent not found
+  [interactive], line 1:
+    echo $nonexistent
+```
+
+Note that Elvish resolves all variables in a code chunk before starting to
+execute any of it; that is why the error message above says *compilation
+error*. This can be more clearly observed in the following example:
+
+```elvish-transcript
+~> echo pre-error; echo $nonexistent
+Compilation error: variable $nonexistent not found
+[tty], line 1: echo pre-error; echo $nonexistent
+```
+
+When you assign a variable, Elvish does a similar searching. If the variable
+cannot be found, it will be created in the current scope:
+
+```elvish-transcript
+~> x = 12
+~> { x = 13 } # assigns to x in the global scope
+~> echo $x
+13
+~> { z = foo } # creates z in the inner scope
+~> echo $z
+Compilation error: variable $z not found
+[tty], line 1: echo $z
+```
+
+One implication of this behavior is that Elvish will not shadow your variable
+in outer scopes.
+
+There is a `local:` namespace that always refers to the current scope, and by
+using it it is possible to force Elvish to shadow variables:
+
+```elvish-transcript
+~> x = 12
+~> { local:x = 13; echo $x } # force shadowing
+13
+~> echo $x
+12
+```
+
+After force shadowing, you can still access the variable in the outer scope
+using the `up:` namespace, which always **skips** the innermost scope:
+
+```elvish-transcript
+~> x = 12
+~> { local:x = 14; echo $x $up:x }
+14 12
+```
+
+The `local:` and `up:` namespaces can also be used on unshadowed variables,
+although they are not useful in those cases:
+
+```elvish-transcript
+~> foo = a
+~> { echo $up:foo } # $up:foo is the same as $foo
+a
+~> { bar = b; echo $local:bar } # $local:bar is the same as $bar
+b
+```
+
+It is not possible to refer to a specific outer scope.
+
+You cannot create new variables in the `builtin:` namespace, although existing
+variables in it can be assigned new values.
+
+
+# Lambda
+
+A function literal, or lambda, is a [code chunk](#code-chunk) surrounded by
+curly braces:
+
+```elvish-transcript
+~> f = { echo "Inside a lambda" }
+~> put $f
+▶ <closure 0x18a1a340>
+```
+
+One or more whitespace characters after `{` is required: Elvish relies on the
+presence of whitespace to disambiguate lambda literals and [braced
+lists](#braced-lists). It is good style to put some whitespace before the
+closing `}` as well, but this is not required by the syntax.
+
+Functions are first-class values in Elvish. They can be kept in variables, used
+as arguments, output on the value channel, and embedded in other data
+structures. They can also be used as commands:
+
+```elvish-transcript
+~> $f
+Inside a lambda
+~> { echo "Inside a literal lambda" }
+Inside a literal lambda
+```
+
+The last command resembles a code block in C-like languages in syntax. But
+under the hood, it defines a function on the fly and calls it immediately.
+
+Functions defined using the basic syntax above do not accept any arguments or
+options. To do so, you need to write a signature.
+
+## Signature
+
+A **signature** specifies the arguments a function can accept:
+
+```elvish-transcript
+~> f = [a b]{ put $b $a }
+~> $f lorem ipsum
+▶ ipsum
+▶ lorem
+```
+
+There should be no space between `]` and `{`; otherwise Elvish will parse the
+signature as a list, followed by a lambda without signature:
+
+```elvish-transcript
+~> put [a]{ nop }
+▶ <closure 0xc420153d80>
+~> put [a] { nop }
+▶ [a]
+▶ <closure 0xc42004a480>
+```
+
+Like in the left hand of assignments, if you prefix the last argument with
+`@`, it becomes a **rest argument**, and its value is a list containing all
+the remaining arguments:
+
+```elvish-transcript
+~> f = [a @rest]{ put $a $rest }
+~> $f lorem
+▶ lorem
+▶ []
+~> $f lorem ipsum dolar sit
+▶ lorem
+▶ [ipsum dolar sit]
+```
+
+You can also declare options in the signature. The syntax is `&name=default`
+(like a map pair), where `default` is the default value for the option:
+
+```elvish-transcript
+~> f = [&opt=default]{ echo "Value of $opt is "$opt }
+~> $f
+Value of $opt is default
+~> $f &opt=foobar
+Value of $opt is foobar
+```
+
+Options must have default values: Options should be **option**al.
+
+If you call a function with too few arguments, too many arguments or unknown
+options, an exception is thrown:
+
+```elvish-transcript
+~> [a]{ echo $a } foo bar
+Exception: need 1 arguments, got 2
+[tty], line 1: [a]{ echo $a } foo bar
+~> [a b]{ echo $a $b } foo
+Exception: need 2 arguments, got 1
+[tty], line 1: [a b]{ echo $a $b } foo
+~> [a b @rest]{ echo $a $b $rest } foo
+Exception: need 2 or more arguments, got 1
+[tty], line 1: [a b @rest]{ echo $a $b $rest } foo
+~> [&k=v]{ echo $k } &k2=v2
+Exception: unknown option k2
+[tty], line 1: [&k=v]{ echo $k } &k2=v2
+```
+
+
+## Closure Semantics
+
+User-defined functions are also known as "closures", because they have
+[closure
+semantics](https://en.wikipedia.org/wiki/Closure_(computer_programming)).
+
+In the following example, the `make-adder` function outputs two functions,
+both referring to a local variable `$n`. Closure semantics means that:
+
+1.  Both functions can continue to refer to the `$n` variable after
+    `make-adder` has returned.
+
+2.  Multiple calls to the `make-adder` function generates distinct instances
+    of the `$n` variables.
+
+```elvish-transcript
+~> fn make-adder {
+     n = 0
+     put { put $n } { n = (+ $n 1) }
+   }
+~> getter adder = (make-adder)
+~> $getter # $getter outputs $n
+▶ 0
+~> $adder # $adder increments $n
+~> $getter # $getter and $setter refer to the same $n
+▶ 1
+~> getter2 adder2 = (make-adder)
+~> $getter2 # $getter2 and $getter refer to different $n
+▶ 0
+~> $getter
+▶ 1
+```
+
+Variables that get "captured" in closures are called **upvalues**; this is why
+the pseudo-namespace for variables in outer scopes is called `up:`. When
+capturing upvalues, Elvish only captures the variables that are used. In the
+following example, `$m` is not an upvalue of `$g` because it is not used:
+
+```elvish-transcript
+~> fn f { m = 2; n = 3; put { put $n } }
+~> g = (f)
+```
+
+This effect is not currently observable, but will become so when namespaces
+[become introspectable](https://github.com/elves/elvish/issues/492).
+
+
+# Indexing
+
+Indexing is done by putting one or more **index expressions** in brackets `[]`
+after a value.
+
+## List Indexing
+
+Lists can be indexed with any of the following:
+
+*   A non-negative integer, an offset counting from the beginning of the list.
+    For example, `$li[0]` is the first element of `$li`.
+
+*   A negative integer, an offset counting from the back of the list. For
+    instance, `$li[-1]` is the last element `$li`.
+
+*   A slice `$a:$b`, where both `$a` and `$b` are integers. The result is
+    sublist of `$li[$a]` up to, but not including, `$li[$b]`. For instance,
+    `$li[4:7]` equals `[$li[4] $li[5] $li[6]]`, while `$li[1:-1]` contains all
+    elements from `$li` except the first and last one.
+
+    Both integers may be omitted; `$a` defaults to 0 while `$b` defaults to the
+    length of the list. For instance, `$li[:2]` is equivalent to `$li[0:2]`,
+    `$li[2:]` is equivalent to `$li[2:(count $li)]`, and `$li[:]` makes a copy
+    of `$li`. The last form is rarely useful, as lists are immutable.
+
+    Note that the slice needs to be a **single** string, so there cannot be
+    any spaces within the slice. For instance, `$li[2:10]` cannot be written
+    as `$li[2: 10]`; the latter contains two indicies and is equivalent to
+    `$li[2:] $li[10]` (see [Multiple Indicies](#multiple-indicies)).
+
+*   Not yet implemented: The string `@`. The result is all the values in the
+    list. Note that this is not the same as `:`: if `$li` has 10 elements,
+    `$li[@]` evaluates to 10 values (all the elements in the list), while
+    `$li[:]` evaluates to just one value (a copy of the list).
+
+    When used on a variable like `$li`, it is equivalent to the explosion
+    construct `$li[@]`. It is useful, however, when used on other constructs,
+    like output capture or other
+
+Examples:
+
+```elvish-transcript
+~> li = [lorem ipsum foo bar]
+~> put $li[0]
+▶ lorem
+~> put $li[-1]
+▶ bar
+~> put $li[0:2]
+▶ [lorem ipsum]
+```
+
+(Negative indicies and slicing are borrowed from Python.)
+
+## String indexing
+
+**NOTE**: String indexing will likely change.
+
+Strings should always be UTF-8, and they can indexed by **byte indicies at
+which codepoints start**, and indexing results in **the codepoint that starts
+there**.  This is best explained with examples:
+
+*   In the string `elv`, every codepoint is encoded with only one byte, so 0,
+    1, 2 are all valid indices:
+
+    ```elvish-transcript
+    ~> put elv[0]
+    ▶ e
+    ~> put elv[1]
+    ▶ l
+    ~> put elv[2]
+    ▶ v
+    ```
+
+*   In the string `世界`, each codepoint is encoded with three bytes. The first
+    codepoint occupies byte 0 through 2, and the second occupies byte 3 through
+    5\. Hence valid indicies are 0 and 3:
+
+    ```elvish-transcript
+    ~> put 世界[0]
+    ▶ 世
+    ~> put 世界[3]
+    ▶ 界
+    ```
+
+Strings can also be indexed by slices.
+
+(This idea of indexing codepoints by their byte positions is borrowed from
+Julia.)
+
+## Map indexing
+
+Maps are simply indexed by their keys. There is no slice indexing, and `:` does
+not have a special meaning. Examples:
+
+```elvish-transcript
+~> map = [&a=lorem &b=ipsum &a:b=haha]
+~> echo $map[a]
+lorem
+~> echo $map[a:b]
+haha
+```
+
+## Multiple Indices
+
+If you put multiple values in the index, you get multiple values: `$li[x y z]`
+is equivalent to `$li[x] $li[y] $li[z]`. This applies to all indexable values.
+Examples:
+
+```elvish-transcript
+~> put elv[0 2 0:2]
+▶ e
+▶ v
+▶ el
+~> put [lorem ipsum foo bar][0 2 0:2]
+▶ lorem
+▶ foo
+▶ [lorem ipsum]
+~> put [&a=lorem &b=ipsum &a:b=haha][a a:b]
+▶ lorem
+▶ haha
+```
+
+
+# Output Capture
+
+Output capture is formed by putting parentheses `()` around a [code
+chunk](#code-chunk). It redirects the output of the chunk into an internal
+pipe, and evaluates to all the values that have been output.
+
+```elvish-transcript
+~> + 1 10 100
+▶ 111
+~> x = (+ 1 10 100)
+~> put $x
+▶ 111
+~> put lorem ipsum
+▶ lorem
+▶ ipsum
+~> x y = (put lorem ipsum)
+~> put $x
+▶ lorem
+~> put $y
+▶ ipsum
+```
+
+If the chunk outputs bytes, Elvish strips the last newline (if any), and split them by newlines, and consider each line to be one string value:
+
+```elvish-transcript
+~> put (echo "a\nb")
+▶ a
+▶ b
+```
+
+**Note 1**. Only the last newline is ever removed, so empty lines are
+preserved; `(echo "a\n")` evaluates to two values, `"a"` and `""`.
+
+**Note 2**. One consequence of this mechanism is that you can not distinguish
+outputs that lack a trailing newline from outputs that have one; `(echo what)`
+evaluates to the same value as `(print what)`. If such a distinction is
+needed, use [`slurp`](/ref/builtin.html#slurp) to preserve the original
+bytes output.
+
+If the chunk outputs both values and bytes, the values of output capture will
+contain both value outputs and lines. However, the ordering between value
+output and byte output might not agree with the order in which they happened:
+
+```elvish-transcript
+~> put (put a; echo b) # value order need not be the same as output order
+▶ b
+▶ a
+```
+
+
+# Exception Capture
+
+Exception capture is formed by putting `?()` around a code chunk. It runs the chunk and evaluates to the exception it throws.
+
+```elvish-transcript
+~> fail bad
+Exception: bad
+Traceback:
+  [interactive], line 1:
+    fail bad
+~> put ?(fail bad)
+▶ ?(fail bad)
+```
+
+If there was no error, it evaluates to the special value `$ok`:
+
+```elvish-transcript
+~> nop
+~> put ?(nop)
+▶ $ok
+```
+
+Exceptions are booleanly false and `$ok` is booleanly true. This is useful in
+`if` (introduced later):
+
+```elvish-transcript
+if ?(test -d ./a) {
+  # ./a is a directory
+}
+```
+
+Exception captures do not affect the output of the code chunk. You can combine
+output capture and exception capture:
+
+```elvish
+output = (error = ?(commands-that-may-fail))
+```
+
+
+# Tilde Expansion
+
+Tildes are special when they appear at the beginning of an expression (the
+exact meaning of "expression" will be explained later). The string after it, up
+to the first `/` or the end of the word, is taken as a user name; and they
+together evaluate to the home directory of that user. If the user name is
+empty, the current user is assumed.
+
+In the following example, the home directory of the current user is `/home/xiaq`, while that of the root user is `/root`:
+
+```elvish-transcript
+~> put ~
+▶ /home/xiaq
+~> put ~root
+▶ /root
+~> put ~/xxx
+▶ /home/xiaq/xxx
+~> put ~root/xxx
+▶ /root/xxx
+```
+
+Note that tildes are not special when they appear elsewhere in a word:
+
+```elvish-transcript
+~> put a~root
+▶ a~root
+```
+
+If you need them to be, surround them with braces (the reason this works will be explained later):
+
+```elvish-transcript
+~> put a{~root}
+▶ a/root
+```
+
+
+# Wildcard Patterns
+
+**Wildcard patterns** are patterns containing **wildcards**, and they evaluate to all filenames they match.
+
+We will use this directory tree in examples:
+
+```
+.x.conf
+a.cc
+ax.conf
+foo.cc
+d/
+|__ .x.conf
+|__ ax.conf
+|__ y.cc
+.d2/
+|__ .x.conf
+|__ ax.conf
+```
+
+Elvish supports the following wildcards:
+
+*   `?` matches one arbitrary character except `/`. For example, `?.cc`
+    matches `a.cc`;
+
+*   `*` matches any number of arbitrary characters except `/`. For example,
+    `*.cc` matches `a.cc` and `foo.cc`;
+
+*   `**` matches any number of arbitrary characters including `/`. For example,
+    `**.cc` matches `a.cc`, `foo.cc` and `b/y.cc`.
+
+The following behaviors are default, although they can be altered by modifiers:
+
+*   When the entire wildcard pattern has no match, an error is thrown.
+
+*   None of the wildcards matches `.` at the beginning of filenames. For example:
+
+    *   `?x.conf` does not match `.x.conf`;
+
+    *   `d/*.conf` does not match `d/.x.conf`;
+
+    *   `**.conf` does not match `d/.x.conf`.
+
+
+## Modifiers
+
+Wildcards can be **modified** using the same syntax as indexing. For instance,
+in `*[match-hidden]` the `*` wildcard is modified with the `match-hidden`
+modifier. Multiple matchers can be chained like `*[set:abc][range:0-9]`. There
+are two kinds of modifiers:
+
+**Global modifiers** apply to the whole pattern and can be placed after any
+wildcard:
+
+*   `nomatch-ok` tells Elvish not to throw an error when there is no match for
+    the pattern. For instance, in the example directory `put bad*` will be an
+    error, but `put bad*[nomatch-ok]` does exactly nothing.
+
+*   `but:xxx` (where `xxx` is any filename) excludes the filename from the final
+    result.
+
+Although global modifiers affect the entire wildcard pattern, you can add it
+after any wildcard, and the effect is the same. For example, `put
+*/*[nomatch-ok].cpp` and `put *[nomatch-ok]/*.cpp` do the same thing.
+
+On the other hand, you must add it after a wildcard, instead of after the
+entire pattern: `put */*.cpp[nomatch-ok]` unfortunately does not do the correct
+thing. (This will probably be fixed.)
+
+**Local modifiers** only apply to the wildcard it immediately follows:
+
+*   `match-hidden` tells the wildcard to match `.` at the beginning of
+    filenames, e.g. `*[match-hidden].conf` matches `.x.conf` and `ax.conf`.
+
+    Being a local modifier, it only applies to the wildcard it immediately
+    follows. For instance, `*[match-hidden]/*.conf` matches `d/ax.conf` and
+    `.d2/ax.conf`, but not `d/.x.conf` or `.d2/.x.conf`.
+
+*   Character matchers restrict the characters to match:
+
+    *   Character sets, like `set:aeoiu`;
+
+    *   Character ranges like `range:a-z` (including `z`) or `range:a~z` (excluding `z`);
+
+    *   Character classes: `control`, `digit`, `graphic`, `letter`, `lower`,
+        `mark`, `number`, `print`, `punct`, `space`, `symbol`, `title`, and
+        `upper`. See the Is* functions [here](https://godoc.org/unicode) for
+        their definitions.
+
+    Note the following caveats:
+
+    *   Multiple matchers, they are OR'ed. For instance, ?[set:aeoiu][digit]
+        matches `aeoiu` and digits.
+
+    *   Dots at the beginning of filenames always require an explicit
+        `match-hidden`, even if the matcher includes `.`. For example,
+        `?[set:.a]x.conf` does **not** match `.x.conf`; you have to `?[set:.a
+        match-hidden]x.conf`.
+
+    *   Likewise, you always need to use `**` to match slashes, even if the
+        matcher includes `/`. For example `*[set:abc/]` is the same as
+        `*[set:abc]`.
+
+
+# Compound Expression and Braced Lists
+
+Writing several expressions together with no space in between will concatenate
+them. This creates a **compound expression**, because it mimics the formation
+of compound words in natural languages. Examples:
+
+```elvish-transcript
+~> put 'a'b"c" # compounding three string literals
+▶ abc
+~> v = value
+~> put '$v is '$v # compounding one string literal with one string variable
+▶ '$v is value'
+```
+
+Many constructs in Elvish can generate multiple values, like indexing with
+multiple indices and output captures. Compounding multiple values with other
+values generates all possible combinations:
+
+```elvish-transcript
+~> put (put a b)-(put 1 2)
+▶ a-1
+▶ a-2
+▶ b-1
+▶ b-2
+```
+
+Note the order of the generated values. The value that comes later changes faster.
+
+**NOTE**: There is a perhaps a better way to explain the ordering, but you can think of the previous code as equivalent to this:
+
+```elvish-transcript
+for x [a b] {
+  for y [1 2] {
+    put $x-$y
+  }
+}
+```
+
+## Braced Lists
+
+In practice, you never have to write `(put a b)`: you can use a braced list `{a,b}`:
+
+```elvish-transcript
+~> put {a,b}-{1,2}
+▶ a-1
+▶ a-2
+▶ b-1
+▶ b-2
+```
+
+Elements in braced lists can also be separated with whitespaces, or a combination of comma and whitespaces (the latter not recommended):
+
+```elvish-transcript
+~> put {a b , c,d}
+▶ a
+▶ b
+▶ c
+▶ d
+```
+
+(In future, the syntax might be made more strict.)
+
+Braced list is merely a syntax for grouping multiple values. It is not a data
+structure.
+
+
+# Expression Structure and Precedence
+
+Braced lists are evaluated before being compounded with other values. You can
+use this to affect the order of evaluation. For instance, `put *.txt` gives you
+all filenames that end with `.txt` in the current directory; while `put
+{*}.txt` gives you all filenames in the current directory, appended with
+`.txt`.
+
+**TODO**: document evaluation order regarding tilde and wildcards.
+
+
+# Ordinary Command
+
+The **command** is probably the most important syntax construct in shell
+languages, and Elvish is no exception. The word **command** itself, is
+overloaded with meanings. In the terminology of this document, the term
+**command** include the following:
+
+*   An ordinary assignment, introduced above;
+
+*   An ordinary command, introduced in this section;
+
+*   A special command, introduced in the next section.
+
+An **ordinary command** consists of a compulsory head, and any number of
+arguments, options and redirections.
+
+## Head
+
+The **head** must appear first. It is an arbitrary word that determines what will be run. Examples:
+
+```elvish-transcript
+~> ls -l # the string ls is the head
+(output omitted)
+~> (put [@a]{ ls $@a }) -l
+(same output)
+```
+
+The head must evaluate to one value. For instance, the following does not work:
+
+```elvish-transcript
+~> (put [@a]{ ls $@a } -l)
+Exception: head of command must be a single value; got 2 values
+[tty], line 1: (put [@a]{ ls $@a } -l)
+```
+
+The definition of barewords is relaxed for the head to include `<`, `>`, `*`
+and `^`. These are all names of numeric builtins:
+
+```elvish-transcript
+~> < 3 5 # less-than
+▶ $true
+~> > 3 5 # greater-than
+▶ $false
+~> * 3 5 # multiplication
+▶ 15
+~> ^ 3 5 # power
+▶ 243
+```
+
+## Arguments and Options
+
+**Arguments** (args for short) and **options** (opts for short) can be supplied
+to commands. Arguments are arbitrary words, while options have the same
+syntax as map pairs. They are separated by inline whitespaces:
+
+```elvish-transcript
+~> echo &sep=, a b c # seq=, is an option; a b c are arguments
+a,b,c
+```
+
+
+## Redirections
+
+Redirections are used for modifying file descriptors (FD).
+
+The most common form of redirections opens a file and associates it with an FD.
+The form consists of an optional destination FD (like `2`), a redirection
+operator (like `>`) and a filename (like `error.log`):
+
+*   The **destination fd** determines which FD to modify. If absent, it is
+    determined from the redirection operator. If present, there must be no
+    space between the fd and the redirection operator. (Otherwise Elvish parses
+    it as an argument.)
+
+*   The **redirection operator** determines the mode to open the file, and the
+    destination fd if it is not explicitly specified.
+
+*   The **filename** names the file to open.
+
+Possible redirection operators and their default FDs are:
+
+*   `<` for reading. The default FD is 0 (stdin).
+
+*   `>` for writing. The default FD is 1 (stdout).
+
+*   `>>` for appending. The default FD is 1 (stdout).
+
+*   `<>` for reading and writing. The default FD is 1 (stdout).
+
+Examples:
+
+```elvish-transcript
+~> echo haha > log
+~> cat log
+haha
+~> cat < log
+haha
+~> ls --bad-arg 2> error
+Exception: ls exited with 2
+Traceback:
+  [interactive], line 1:
+    ls --bad-arg 2> error
+~> cat error
+/bin/ls: unrecognized option '--bad-arg'
+Try '/bin/ls --help' for more information.
+```
+
+Redirections can also be used for closing or duplicating FDs. Instead of
+writing a filename, use `&n` (where `n` is a number) for duplicating, or `&-`
+for closing. In this case, the redirection operator only determines the default
+destination FD (and is totally irrevelant if a destination FD is specified).
+Examples:
+
+```elvish-transcript
+~> ls >&- # close stdout
+/bin/ls: write error: Bad file descriptor
+Exception: ls exited with 2
+Traceback:
+  [interactive], line 1:
+    ls >&-
+```
+
+If you have multiple related redirections, they are applied in the order they appear. For instance:
+
+```elvish-transcript
+~> fn f { echo out; echo err >&2 } # echoes "out" on stdout, "err" on stderr
+~> f >log 2>&1 # use file "log" for stdout, then use (changed) stdout for stderr
+~> cat log
+out
+err
+```
+
+
+## Ordering
+
+Elvish does not impose any ordering of arguments, options and redirections:
+they can intermix each other. The only requirement is that the head must come
+first. This is different from POSIX shells, where redirections may appear
+before the head. For instance, the following two both work in POSIX shell, but
+only the former works in Elvish:
+
+
+```sh
+echo something > file
+> file echo something # mistaken for the comparison builtin ">" in Elvish
+```
+
+
+# Special Commands
+
+**Special commands** obey the same syntax rules as normal commands (i.e.
+syntactically special commands can be treated the same as ordinary commands),
+but have evaluation rules that are custom to each command. To explain this, we
+use the following example:
+
+```elvish-transcript
+~> or ?(echo x) ?(echo y) ?(echo z)
+x
+▶ $ok
+```
+
+In the example, the `or` command first evaluates its first argument, which has the value `$ok` (a truish value) and the side effect of outputting `x`. Due to the custom evaluation rule of `or`, the rest of the arguments are not evaluated.
+
+If `or` were a normal command, the code above is still syntactically correct. However, Elvish would then evaluate all its arguments, with the side effect of outputting `x`, `y` and `z`, before calling `or`.
+
+
+## Deleting variable or element: `del`
+
+The `del` special command can be used to either delete variables or elements
+of lists and maps. Operands should be specified without a leading dollar sign,
+like the left-hand side of assignments.
+
+Example of deleting variable:
+
+```elvish-transcript
+~> x = 2
+~> echo $x
+2
+~> del x
+~> echo $x
+Compilation error: variable $x not found
+[tty], line 1: echo $x
+```
+
+Example of deleting element:
+
+```elvish-transcript
+~> m = [&k=v &k2=v2]
+~> del m[k2]
+~> put $m
+▶ [&k=v]
+~> l = [[&k=v &k2=v2]]
+~> del l[0][k2]
+~> put $l
+▶ [[&k=v]]
+```
+
+
+## Logics: `and` and `or`
+
+The `and` special command evaluates its arguments from left to right; as soon
+as a booleanly false value is obtained, it outputs the value and stops. When
+given no arguments, it outputs `$true`.
+
+The `or` special command is the same except that it stops when a booleanly true
+value is obtained. When given no arguments, it outpus `$false`.
+
+
+## Condition: `if`
+
+**TODO**: Document the syntax notation, and add more examples.
+
+Syntax:
+
+```elvish-transcript
+if <condition> {
+    <body>
+} elif <condition> {
+    <body>
+} else {
+    <else-body>
+}
+```
+
+The `if` special command goes through the conditions one by one: as soon as one
+evaluates to a booleanly true value, its corresponding body is executed. If
+none of conditions are booleanly true and an else body is supplied, it is
+executed.
+
+The condition part is an expression, not a command like in other shells.
+
+Tip: a combination of `if` and `?()` gives you a semantics close to other shells:
+
+```elvish-transcript
+if ?(test -d .git) {
+    # do something
+}
+```
+
+However, for Elvish's builtin predicates that output values instead of throw exceptions, the output capture construct `()` should be used.
+
+**TODO**: add more examples.
+
+## Conditional Loop: `while`
+
+Syntax:
+
+```elvish-transcript
+while <condition> {
+    <body>
+} else {
+    <else-body>
+}
+```
+
+Execute the body as long as the condition evaluates to a booleanly true value.
+
+The else body, if present, is executed if the body has never been executed
+(i.e. the condition evaluates to a booleanly false value in the very
+beginning).
+
+
+## Iterative Loop: `for`
+
+Syntax:
+
+```elvish-transcript
+for <var> <container> {
+    <body>
+} else {
+    <body>
+}
+```
+
+Iterate the container (e.g. a list). In each iteration, assign the variable to
+an element of the container and execute the body.
+
+The else body, if present, is executed if the body has never been executed
+(i.e. the iteration value has no elements).
+
+
+## Exception Control: `try`
+
+(If you just want to capture the exception, you can use the more concise
+exception capture construct `?()` instead.)
+
+Syntax:
+
+```elvish-transcript
+try {
+    <try-block>
+} except except-varname {
+    <except-block>
+} else {
+    <else-block>
+} finally {
+    <finally-block>
+}
+```
+
+Only `try` and `try-block` are required. This control structure behaves as follows:
+
+1.  The `try-block` is always executed first.
+
+2.  If `except` is present and an exception occurs in `try-block`, it is
+    caught and stored in `except-varname`, and `except-block` is then
+    executed. Example:
+
+    ```elvish-transcript
+    ~> try { fail bad } except e { put $e }
+    ▶ ?(fail bad)
+    ```
+
+    Note that if `except` is not present, exceptions thrown from `try` are not
+    caught: for instance, `try { fail bad }` throws `bad`; it is equivalent to
+    a plain `fail bad`.
+
+    Note that the word after `except` names a variable, not a matching
+    condition. Exception matching is not supported yet. For instance, you may
+    want to only match exceptions that were created with `fail bad` with
+    `except bad`, but in fact this creates a variable `$bad` that contains
+    whatever exception was thrown.
+
+3.  If no exception occurs and `else` is present, `else-block` is executed.
+    Example:
+
+    ```elvish-transcript
+    ~> try { nop } else { echo well }
+    well
+    ```
+
+4.  If `finally-block` is present, it is executed. Examples:
+
+    ```elvish-transcript
+    ~> try { fail bad } finally { echo final }
+    final
+    Exception: bad
+    Traceback:
+      [tty], line 1:
+        try { fail bad } finally { echo final }
+    ~> try { echo good } finally { echo final }
+    good
+    final
+    ```
+
+5.  If the exception was not caught (i.e. `except` is not present), it is
+    rethrown.
+
+Exceptions thrown in blocks other than `try-block` are not caught. If an
+exception was thrown and either `except-block` or `finally-block` throws
+another exception, the original exception is lost. Examples:
+
+```elvish-transcript
+~> try { fail bad } except e { fail worse }
+Exception: worse
+Traceback:
+  [tty], line 1:
+    try { fail bad } except e { fail worse }
+~> try { fail bad } except e { fail worse } finally { fail worst }
+Exception: worst
+Traceback:
+  [tty], line 1:
+    try { fail bad } except e { fail worse } finally { fail worst }
+```
+
+
+## Function Definition: `fn`
+
+Syntax:
+
+```elvish-transcript
+fn <name> <lambda>
+```
+
+Define a function with a given name. The function behaves in the same way to
+the lambda used to define it, except that it "captures" `return`. In other
+words, `return` will fall through lambdas not defined with `fn`, and continues
+until it exits a function defined with `fn`:
+
+```elvish-transcript
+~> fn f {
+     { echo a; return }
+     echo b # will not execute
+   }
+~> f
+a
+~> {
+     f
+     echo c # executed, because f "captures" the return
+   }
+a
+c
+```
+
+**TODO**: Find a better way to describe this. Hopefully the example is
+illustrative enough, though.
+
+
+# Command Resolution
+
+When using a literal string as the head of a command, it is first **resolved**
+during the compilation phase, using the following order:
+
+1.  Special commands.
+
+2.  Functions defined on any of the containing lexical scopes, with inner
+    scopes looked up first. (This is exactly the same process as variable
+    resolution).
+
+3.  The `builtin:` namespace.
+
+4.  Should all these steps fail, it is resolved to be an external command.
+    Determination of the path of the external command does not happen in the
+    resolution process; that happens during evaluation.
+
+You can use the [resolve](/ref/builtin.html#resolve) command to see which
+command Elvish resolves a string to.
+
+During the evaluation phase, external commands are then subject to
+**searching**. This can be observed with the
+[search-external](/ref/builtin.html#search-builtin) command.
+
+
+# Pipeline
+
+A **pipeline** is formed by joining one or more commands together with the pipe
+sign (`|`).
+
+## IO Semantics
+
+For each pair of adjacent commands `a | b`, the output of `a` is connected to
+the input of `b`. Both the byte pipe and the value channel are connected, even
+if one of them is not used.
+
+Command redirections are applied before the connection happens. For instance,
+the following writes `foo` to `a.txt` instead of the output:
+
+```elvish-transcript
+~> echo foo > a.txt | cat
+~> cat a.txt
+foo
+```
+
+## Execution Flow
+
+All of the commands in a pipeline are executed in parallel, and the execution
+of the pipeline finishes when all of its commands finish execution.
+
+If one or more command in a pipeline throws an exception, the other commands
+will continue to execute as normal. After all commands finish execution, an
+exception is thrown, the value of which depends on the number of commands that
+have thrown an exception:
+
+*   If only one command has thrown an exception, that exception is rethrown.
+
+*   If more than one commands have thrown exceptions, a "composite exception",
+    containing information all exceptions involved, is thrown.
+
+## Background Pipeline
+
+Adding an ampersand `&` to the end of a pipeline will cause it to be executed
+in the background. In this case, the rest of the code chunk will continue to
+execute without waiting for the pipeline to finish. Exceptions thrown from the
+background pipeline do not affect the code chunk that contains it.
+
+When a background pipeline finishes, a message is printed to the terminal if
+the shell is interactive.
+
+
+# Code Chunk
+
+A **code chunk** is formed by joining zero or more pipelines together,
+separating them with either newlines or semicolons.
+
+Pipelines in a code chunk are executed in sequence. If any pipeline throws an
+exception, the execution of the whole code chunk stops, propagating that
+exception.
+
+
+# Exception and Flow Commands
+
+Exceptions have similar semantics to those in Python or Java. They can be
+thrown with the [fail](/ref/builtin.html#fail) command and caught with either
+exception capture `?()` or the `try` special command.
+
+If an external command exits with a non-zero status, Elvish treats that as an
+exception.
+
+Flow commands -- `break`, `continue` and `return` -- are ordinary builtin
+commands that raise special "flow control" exceptions. The `for` and `while`
+commands capture `break` and `continue`, while `fn` modifies its closure to
+capture `return`.
+
+One interesting implication is that since flow commands are just ordinary
+commands you can build functions on top of them. For instance, this function
+`break`s randomly:
+
+```elvish
+fn random-break {
+  if eq (randint 2) 0 {
+    break
+  }
+}
+```
+
+The function `random-break` can then be used in for-loops and while-loops.
+
+Note that the `return` flow control exception is only captured by functions
+defined with `fn`. It falls through ordinary lambdas:
+
+```elvish
+fn f {
+  {
+    # returns f, falling through the innermost lambda
+    return
+  }
+}
+```
+
+
+# Namespaces and Modules
+
+Namespace in Elvish helps prevent name collisions and is important for
+building modules.
+
+## Syntax
+
+Prepend `namespace:` to command names and variable names to specify the
+namespace. The following code
+
+```elvish
+e:echo $E:PATH
+```
+
+uses the `echo` command from the `e` namespace and the `PATH` variable from
+the `E` namespace.
+
+Names of namespaces can contain colons themselves. For instance, in `$x:y:z`
+the namespace is `x:y`. More precisely, when parsing command names and
+variable names, everything up to the last colon is considered to be the
+namespace.
+
+A convention used in articles is when referring to a namespace is to add a
+trailing colon: for instance, `edit:` means "the namespace `edit`". This is
+remniscent to the syntax but is not syntactically valid.
+
+## Special Namespaces
+
+The following namespaces have special meanings to the language:
+
+*   `local:` and `up:` refer to lexical scopes, and have been documented above.
+
+*   `e:` refers to externals. For instance, `e:ls` refers to the external
+    command `ls`.
+
+    Most of the time you can rely on the rules of [command
+    resolution](#command-resolution) and do not need to use this explicitly,
+    unless a function defined by you (or an Elvish builtin) shadows an
+    external command.
+
+*   `E:` refers to environment variables. For instance, `$E:USER` is the
+    environment variable `USER`.
+
+    This **is** always needed, because unlike command resolution, variable
+    resolution does not fall back onto environment variables.
+
+*   `builtin:` refers to builtin functions and variables.
+
+    You don't need to use this explicitly unless you have defined names that
+    shadows builtin counterparts.
+
+## Pre-Imported Modules
+
+Namespaces that are not special (i,e. one of the above) are also called
+**modules**. Aside from these special namespaces, Elvish also comes with the
+following modules:
+
+*   `edit:` for accessing the Elvish editor. This module is available in
+    interactive mode.
+
+    See [reference](/ref/edit.html).
+
+*   `re:` for regular expression facilities.
+    This module is always available. See [reference](/ref/re.html).
+
+*   `daemon:` for manipulating the daemon. This module is always available.
+
+    This is not yet documented.
+
+## User-Defined Modules
+
+You can define your own modules with Elvishscript but putting them under
+`~/.elvish/lib` and giving them a `.elv` extension. For instance, to define a
+module named `a`, store it in `~/.elvish/lib/a.elv`:
+
+```elvish-transcript
+~> cat ~/.elvish/lib/a.elv
+echo "mod a loading"
+fn f {
+  echo "f from mod a"
+}
+```
+
+To import the module, use `use`:
+
+```elvish-transcript
+~> use a
+mod a loading
+~> a:f
+f from mod a
+```
+
+The argument to `use` is called the **usespec** and will be explained in more
+details below. In the simplest case, it is simply the module name.
+
+Modules are evaluated in a seprate scope. That means that functions and
+variables defined in the module does not pollute the default namespace, and
+vice versa. For instance, if you define `ls` as a wrapper function in
+`rc.elv`:
+
+```elvish
+fn ls [@a]{
+    e:ls --color=auto $@a
+}
+```
+
+That definition is not visible in module files: `ls` will still refer to the
+external command `ls`, unless you shadow it in the very same module.
+
+## Modules in Nested Directories
+
+It is often useful to put modules under directories. When importing such
+modules, you can control which (trailing) parts becomes the module name.
+
+For instance, if you have the following content in `x/y/z.elv` (relative to `~/.elvish/lib`, of course):
+
+```elvish
+fn f {
+    echo 'In a deeply nested module'
+}
+```
+
+It is possible to import the module as `z`, `y:z`, or `x:y:z`:
+
+```elvish-transcript
+~> use x/y/z # imports as "z"
+~> z:f
+In a deeply nested module
+~> use x/y:z # imports as "y:z"
+~> y:z:f
+In a deeply nested module
+~> use x:y:z # imports as "x:y:z"
+~> x:y:z:f
+In a deeply nested module
+```
+
+To be precise:
+
+*   The path of the module to import is derived by replacing all `:` with `/`
+    and adding `.elv`. In the above example, all of `x/y/z`, `x/y:z` and
+    `x:y:z` have the same path `x/y/z.elv` and refer to the same module
+
+*   The part after the last `/` becomes the module name.
+
+## Lexical Scoping of Imports
+
+Namespace imports are also lexically scoped. For instance, if you `use` a
+module within an inner scope, it is not available outside that scope:
+
+```elvish
+{
+    use some-mod
+    some-mod:some-func
+}
+some-mod:some-func # not valid
+```
+
+## Re-Importing
+
+Modules are cached after one import. Subsequent imports do not re-execute the
+module; they only serve the bring it into the current scope. Moreover, the
+cache is keyed by the path of the module, not the name under which it is
+imported. For instance, if you have the following in `~/.elvish/lib/a/b.elv`:
+
+```elvish
+echo importing
+```
+
+The following code only prints one `importing`:
+
+```elvish
+{ use a:b }
+use a:b # only brings mod into the lexical scope
+```
+
+As does the following:
+
+```elvish
+use a/b
+use a:b
+```
diff -pruN 0.11+ds1-5/website/src/ref/philosophy.md 0.12+ds1-1/website/src/ref/philosophy.md
--- 0.11+ds1-5/website/src/ref/philosophy.md	1970-01-01 00:00:00.000000000 +0000
+++ 0.12+ds1-1/website/src/ref/philosophy.md	2018-07-11 02:32:25.000000000 +0000
@@ -0,0 +1,54 @@
+<!-- toc -->
+
+The development of Elvish is driven by a set of ideas, a **design philosophy**.
+
+# The language
+
+*   Elvish should be a real, expressive programming language.
+
+    Shells are often considered domain-specific languages (DSL), but Elvish
+    does not restrict itself to this notion. It embraces such concepts as
+    namespaces, first-class functions and exceptions. Whatever you may find in
+    a modern general-purpose programming language is likely to be found in
+    Elvish.
+
+    Elvish is not alone in this respect. There are multiple ongoing efforts;
+    [this page](https://github.com/oilshell/oil/wiki/ExternalResources) on
+    the wiki of oilshell (which is one of the efforts) is a good reference.
+
+*   Elvish should try to preserve and extend traditional shell programming
+    techniques, as long as they don't conflict with the previous tenet. Some
+    examples are:
+
+    *   Barewords are simply strings.
+
+    *   Prefix notation dominates, like Lisp. For example, arithmetics is done
+        like `+ 10 (/ 105 5)`.
+
+    *   Pipeline is the main tool for function composition. To make pipelines
+        suitable for complex data manipulation, Elvish extends them to be able
+        to carry structured data (as opposed to just bytes).
+
+    *   Output capture is the auxiliary tool for function composition. Elvish
+        functions may write structured data directly to the output, and
+        capturing the output yields the same structured data.
+
+# The user interface
+
+*   The user interface should be usable without any customizations. It should
+    be simple and consistent by default:
+
+    *   Prefer to extend well-known functionalities in other shell to inventing
+        brand new ones. For instance, in Elvish Ctrl-R summons the "history
+        listing" for searching history, akin to how Ctrl-R works in bash, but
+        more powerful.
+
+    *   When a useful feature has no prior art in other shells, borrow from
+        other programs. For instance, the [navigation
+        mode](/learn/cookbook.html#navigation-mode), summoned by Ctrl-N,
+        mimics [Ranger](http://ranger.nongnu.org); while the "location mode"
+        used for quickly changing location, mimics location bars in GUI
+        browsers (and is summoned by the same key combination Ctrl-L).
+
+*   Customizability should be achieved via progammability, not an enormous
+    inventory of options that interact with each other in obscure ways.
diff -pruN 0.11+ds1-5/website/src/ref/re.md 0.12+ds1-1/website/src/ref/re.md
--- 0.11+ds1-5/website/src/ref/re.md	1970-01-01 00:00:00.000000000 +0000
+++ 0.12+ds1-1/website/src/ref/re.md	2018-07-11 02:32:25.000000000 +0000
@@ -0,0 +1,138 @@
+<!-- toc -->
+
+# Introduction
+
+The `re:` module wraps Go's `regexp` package. See the
+[doc](http://godoc.org/regexp) for the Go package for syntax of regular
+expressions and replacement patterns.
+
+# Functions
+
+Function usages notations follow the same convention as the [builtin module
+doc](/ref/builtin.html).
+
+The following options are supported by multiple functions in this module:
+
+*   `&posix` (defaults to `$false`): Use POSIX ERE syntax. See also
+    [doc](http://godoc.org/regexp#CompilePOSIX) in Go package.
+
+*   `&longest` (defaults to `$false`): Prefer leftmost-longest match. See also
+    [doc](http://godoc.org/regexp#Regexp.Longest) in Go package.
+
+*   `&max` (defaults to -1): If non-negative, maximum number of results.
+
+## find
+
+```elvish
+re:find &posix=$false &longest=$false &max=-1 $pattern $source
+```
+
+Find all matches of `$pattern` in `$source`.
+
+Each match is represented by a map-like value `$m`; `$m[text]`, `$m[start]`
+and `$m[end]` are the text, start and end positions (as byte indicies into
+`$source`) of the match; `$m[groups]` is a list of submatches for capture
+groups in the pattern. A submatch has a similar structure to a match, except
+that it does not have a `group` key. The entire pattern is an implicit capture
+group, and it always appears first.
+
+Examples:
+
+```elvish-transcript
+~> re:find . ab
+▶ [&text=a &start=0 &end=1 &groups=[[&text=a &start=0 &end=1]]]
+▶ [&text=b &start=1 &end=2 &groups=[[&text=b &start=1 &end=2]]]
+~> re:find '[A-Z]([0-9])' 'A1 B2'
+▶ [&text=A1 &start=0 &end=2 &groups=[[&text=A1 &start=0 &end=2] [&text=1 &start=1 &end=2]]]
+▶ [&text=B2 &start=3 &end=5 &groups=[[&text=B2 &start=3 &end=5] [&text=2 &start=4 &end=5]]]
+```
+
+## match
+
+```elvish
+re:match &posix=$false $pattern $source
+```
+
+Determine whether `$pattern` matches `$source`. The pattern is not anchored.
+Examples:
+
+```elvish-transcript
+~> re:match . xyz
+▶ $true
+~> re:match . ''
+▶ $false
+~> re:match '[a-z]' A
+▶ $false
+```
+
+## replace
+
+```elvish
+re:replace &posix=$false &longest=$false &literal=$false $pattern $repl $source
+```
+
+Replace all occurrences of `$pattern` in `$source` with `$repl`.
+
+The replacement `$repl` can be either
+
+1.  A string-typed replacement template. The template can use `$name` or
+    `${name}` patterns to refer to capture groups, where `name` consists of
+    letters, digits and underscores. Numbered patterns like `$1` refer to
+    capture groups by their order, while named patterns like `$stem` refer to
+    capture groups by their names (specified using the syntax
+    `(?P<name>...)`). Use `$$` for a literal dollar sign. The name is taken as
+    long as possible; for instance, `$1a` is the same as `${1a}`.
+
+    See also doc of Go's regexp package on the [template
+    syntax](https://godoc.org/regexp#Regexp.Expand).
+
+2.  A function that takes a string argument and outputs a string. For each
+    match, the function is called with the content of the match, and its
+    output is used as the replacement.
+
+If `$literal` is true, `$repl` must be a string and is treated literally
+instead of as a pattern.
+
+Example:
+
+```elvish-transcript
+~> re:replace '(ba|z)sh' '${1}SH' 'bash and zsh'
+▶ 'baSH and zSH'
+~> re:replace '(ba|z)sh' elvish 'bash and zsh rock'
+▶ 'elvish and elvish rock'
+~> re:replace '(ba|z)sh' [x]{ put [&bash=BaSh &zsh=ZsH][$x] } 'bash and zsh'
+▶ 'BaSh and ZsH'
+```
+
+## split
+
+```elvish
+re:split &posix=$false &longest=$false &max=-1 $pattern $source
+```
+
+Split `$source`, using `$pattern` as separators. Examples:
+
+```elvish-transcript
+~> re:split : /usr/sbin:/usr/bin:/bin
+▶ /usr/sbin
+▶ /usr/bin
+▶ /bin
+~> re:split &max=2 : /usr/sbin:/usr/bin:/bin
+▶ /usr/sbin
+▶ /usr/bin:/bin
+```
+
+## quote
+
+```elvish
+re:quote $string
+```
+
+Quote `$string`. Examples:
+
+```elvish-transcript
+~> re:quote a.txt
+▶ a\.txt
+~> re:quote '(*)'
+▶ '\(\*\)'
+```
diff -pruN 0.11+ds1-5/website/src/template.html 0.12+ds1-1/website/src/template.html
--- 0.11+ds1-5/website/src/template.html	1970-01-01 00:00:00.000000000 +0000
+++ 0.12+ds1-1/website/src/template.html	2018-07-11 02:32:25.000000000 +0000
@@ -0,0 +1,117 @@
+<!doctype html>
+<html>
+<head>
+  <meta charset="utf-8">
+  <meta name="viewport" content="width=device-width, initial-scale=1.0" />
+  {{ if is "homepage" }}
+    <link rel="alternate" type="application/atom+xml"
+          href="{{ rootURL }}/feed.atom">
+  {{ end }}
+
+  <title>
+    {{ if is "homepage" }}
+      {{ .BlogTitle }}
+    {{ else if is "category" }}
+      {{ index .CategoryMap .Category }}
+    {{ else }}
+      {{ .Title }}
+    {{ end }}
+  </title>
+
+  <!-- Generated by realfavicongenerator.net -->
+  <link rel="apple-touch-icon" sizes="180x180" href="/apple-touch-icon.png">
+  <link rel="icon" type="image/png" sizes="32x32" href="/favicon-32x32.png">
+  <link rel="icon" type="image/png" sizes="16x16" href="/favicon-16x16.png">
+  <link rel="manifest" href="/manifest.json">
+  <link rel="mask-icon" href="/safari-pinned-tab.svg" color="#55bb55">
+  <meta name="theme-color" content="#ffffff">
+
+  <link href="/assets/genblog.css" rel="stylesheet">
+  <link href="/assets/elvish.css" rel="stylesheet">
+</head>
+
+<body>
+  <div id="navbar-container"> <div id="navbar">
+    <div id="blog-title">
+      <code>
+        <a href="/">~&gt;</a>
+        <a href="/download" class="command">elvish</a>
+      </code>
+    </div>
+    <ul id="category-list">
+      {{ $homepageTitle := .HomepageTitle }}
+      {{ $curcat := .Category }}
+      <li class="{{ if eq $curcat "homepage" }}current{{ end }}">
+        <a href="{{ rootURL }}" class="nav-link">
+          {{ $homepageTitle }}
+        </a>
+      </li>
+      {{ range $info := .Categories }}
+        <li class="{{ if eq $curcat $info.Name}}current{{ end }}">
+          <a href="{{ rootURL }}/{{ $info.Name }}/" class="nav-link">
+            {{ $info.Title }}
+          </a>
+        </li>
+      {{ end }}
+    </ul>
+    <div class="clear"></div>
+  </div> </div>
+
+  {{/*
+    The reference to "content" is a free one and has to be fixed elsewhere.
+    The *-content templates defined below are intended to be used for this.
+
+    For instance, by adding the following code, this whole template file will
+    function as the template for articles:
+
+        {{ define "content" }} {{ template "article-content" . }} {{ end }}
+
+    This snippet can be generated by contentIs("article").
+  */}}
+  {{ template "content" . }}
+</body>
+</html>
+
+{{ define "article-content" }}
+  <div class="card">
+    <article class="article">
+      {{ if not .IsHomepage }}
+        <div class="article-header">
+          <div class="timestamp"> {{ .Timestamp }} </div>
+          <h1> {{ .Title }} </h1>
+          <div class="clear"></div>
+        </div>
+        <div class="card-splitter"></div>
+      {{ end }}
+      <div class="article-content">
+        {{ .Content }}
+      </div>
+      <div class="clear"></div>
+    </article>
+  </div>
+{{ end }}
+
+{{ define "category-content" }}
+  {{ $category := .Category }}
+  <div class="card">
+    {{ if ne .Prelude "" }}
+      <div class="category-prelude article">
+        <article class="article">
+          {{ .Prelude }}
+        </article>
+      </div>
+    {{ end }}
+    <ul class="article-list">
+      {{ range $article := .Articles }}
+        <li>
+          <a href="{{ rootURL }}/{{ $category }}/{{ $article.Name }}.html"
+             class="nav-link">{{ $article.Title }}</a>
+          <span class="article-timestamp">
+            {{ $article.Timestamp }}
+          </span>
+          <div class="clear"></div>
+        </li>
+      {{ end }}
+    </ul>
+  </div>
+{{ end }}
diff -pruN 0.11+ds1-5/website/tty/completion-mode.html 0.12+ds1-1/website/tty/completion-mode.html
--- 0.11+ds1-5/website/tty/completion-mode.html	1970-01-01 00:00:00.000000000 +0000
+++ 0.12+ds1-1/website/tty/completion-mode.html	2018-07-11 02:32:25.000000000 +0000
@@ -0,0 +1,5 @@
+~/go/src/github.com/elves/elvish&gt; <span class="sgr-32">vim</span><span class="sgr-"> </span><span class="sgr-4">CONTRIBUTING.md </span>
+<span class="sgr-1 sgr-37 sgr-45"> COMPLETING argument </span><span class="sgr-">  </span><span class="sgr-35 sgr-7">                      </span><span class="sgr-35">━━━━━━━━━━━━</span>
+ <span class="sgr- sgr-7">CONTRIBUTING.md</span><span class="sgr-">  Gopkg.toml  README.md  </span><span class="sgr-01 sgr-34">edit  </span><span class="sgr-">  </span><span class="sgr-01 sgr-34">getopt </span><span class="sgr-"> </span>
+ Dockerfile       LICENSE     <span class="sgr-01 sgr-34">cover    </span><span class="sgr-">  </span><span class="sgr-01 sgr-34">errors</span><span class="sgr-">  </span><span class="sgr-01 sgr-34">glob   </span><span class="sgr-"> </span>
+ Gopkg.lock       Makefile    <span class="sgr-01 sgr-34">daemon   </span><span class="sgr-">  </span><span class="sgr-01 sgr-34">eval  </span><span class="sgr-">  main.go </span>
diff -pruN 0.11+ds1-5/website/tty/control-structures.html 0.12+ds1-1/website/tty/control-structures.html
--- 0.11+ds1-5/website/tty/control-structures.html	1970-01-01 00:00:00.000000000 +0000
+++ 0.12+ds1-1/website/tty/control-structures.html	2018-07-11 02:32:25.000000000 +0000
@@ -0,0 +1,15 @@
+~&gt; <span class="sgr-32">if</span><span class="sgr-"> </span><span class="sgr-35">$true</span><span class="sgr-"> </span><span class="sgr-1">{</span><span class="sgr-"> </span><span class="sgr-32">echo</span><span class="sgr-"> good </span><span class="sgr-1">}</span><span class="sgr-"> </span><span class="sgr-33">else</span><span class="sgr-"> </span><span class="sgr-1">{</span><span class="sgr-"> </span><span class="sgr-32">echo</span><span class="sgr-"> bad </span><span class="sgr-1">}</span>               
+good
+~&gt; <span class="sgr-32">for</span><span class="sgr-"> </span><span class="sgr-35">x</span><span class="sgr-"> </span><span class="sgr-1">[</span><span class="sgr-">lorem ipsum</span><span class="sgr-1">]</span><span class="sgr-"> </span><span class="sgr-1">{</span>
+     <span class="sgr-32">echo</span><span class="sgr-"> </span><span class="sgr-35">$x</span><span class="sgr-">.pdf</span>
+   <span class="sgr-1">}</span>
+lorem.pdf
+ipsum.pdf
+~&gt; <span class="sgr-32">try</span><span class="sgr-"> </span><span class="sgr-1">{</span>
+     <span class="sgr-32">fail</span><span class="sgr-"> </span><span class="sgr-33">&#39;bad error&#39;</span>
+   <span class="sgr-1">}</span><span class="sgr-"> </span><span class="sgr-33">except</span><span class="sgr-"> </span><span class="sgr-35">e</span><span class="sgr-"> </span><span class="sgr-1">{</span>
+     <span class="sgr-32">echo</span><span class="sgr-"> error </span><span class="sgr-35">$e</span>
+   <span class="sgr-1">}</span><span class="sgr-"> </span><span class="sgr-33">else</span><span class="sgr-"> </span><span class="sgr-1">{</span>
+     <span class="sgr-32">echo</span><span class="sgr-"> ok</span>
+   <span class="sgr-1">}</span>
+error ?(fail 'bad error')
diff -pruN 0.11+ds1-5/website/tty/histlist-mode.html 0.12+ds1-1/website/tty/histlist-mode.html
--- 0.11+ds1-5/website/tty/histlist-mode.html	1970-01-01 00:00:00.000000000 +0000
+++ 0.12+ds1-1/website/tty/histlist-mode.html	2018-07-11 02:32:25.000000000 +0000
@@ -0,0 +1,15 @@
+~&gt;                                           <span class="sgr-7">xiaq@xiaqsmbp</span>
+<span class="sgr-1 sgr-37 sgr-45"> HISTORY </span><span class="sgr-"> </span>
+13345 make tools/ttyshot                                 <span class="sgr-35">│</span>
+13346 make                                               <span class="sgr-35">│</span>
+13347 ./assets/                                          <span class="sgr-35">│</span>
+13348 ls                                                 <span class="sgr-35">│</span>
+13349 ls                                                 <span class="sgr-35">│</span>
+13350 rm *.png                                           <span class="sgr-35">│</span>
+<span class="sgr-7">13351 git st                                             </span><span class="sgr-35">│</span>
+13352 ..                                                 <span class="sgr-35">│</span>
+13353 git st                                             <span class="sgr-35">│</span>
+13354 git add .                                          <span class="sgr-35">│</span>
+13355 git st                                             <span class="sgr-35">│</span>
+13356 git commit                                         <span class="sgr-35">│</span>
+13357 git push                                           <span class="sgr-35">│</span>
diff -pruN 0.11+ds1-5/website/tty/history-mode.html 0.12+ds1-1/website/tty/history-mode.html
--- 0.11+ds1-5/website/tty/history-mode.html	1970-01-01 00:00:00.000000000 +0000
+++ 0.12+ds1-1/website/tty/history-mode.html	2018-07-11 02:32:25.000000000 +0000
@@ -0,0 +1,2 @@
+~&gt; <span class="sgr-32">echo</span><span class="sgr-"> </span><span class="sgr-4">$daemon:pid</span><span class="sgr-">                          </span><span class="sgr-7">xiaq@xiaqsmbp</span>
+<span class="sgr-1 sgr-37 sgr-45"> HISTORY #11610 </span><span class="sgr-"> </span>
\ No newline at end of file
diff -pruN 0.11+ds1-5/website/tty/lastcmd.html 0.12+ds1-1/website/tty/lastcmd.html
--- 0.11+ds1-5/website/tty/lastcmd.html	1970-01-01 00:00:00.000000000 +0000
+++ 0.12+ds1-1/website/tty/lastcmd.html	2018-07-11 02:32:25.000000000 +0000
@@ -0,0 +1,12 @@
+~&gt; <span class="sgr-32">rm</span><span class="sgr-"> -rf /var</span>
+rm: /var: Operation not permitted
+Exception: <span class="sgr-1 sgr-31">rm exited with 1</span>
+Traceback:
+  [interactive], line 1:
+    rm -rf /var
+~&gt;                                           <span class="sgr-7">xiaq@xiaqsmbp</span>
+<span class="sgr-1 sgr-37 sgr-45"> LASTCMD </span><span class="sgr-"> </span>
+<span class="sgr-7">M-1 rm -rf /var                                           </span>
+  0 rm                                                    
+  1 -rf                                                   
+  2 /var                                                  
\ No newline at end of file
diff -pruN 0.11+ds1-5/website/tty/location-mode-filter.html 0.12+ds1-1/website/tty/location-mode-filter.html
--- 0.11+ds1-5/website/tty/location-mode-filter.html	1970-01-01 00:00:00.000000000 +0000
+++ 0.12+ds1-1/website/tty/location-mode-filter.html	2018-07-11 02:32:25.000000000 +0000
@@ -0,0 +1,3 @@
+~&gt;                                           <span class="sgr-7">xiaq@xiaqsmbp</span>
+<span class="sgr-1 sgr-37 sgr-45"> LOCATION </span><span class="sgr-"> </span><span class="sgr-4">x/p/v</span>
+<span class="sgr-7"> 1 ~/go/src/github.com/xiaq/persistent/vector             </span>
\ No newline at end of file
diff -pruN 0.11+ds1-5/website/tty/location-mode.html 0.12+ds1-1/website/tty/location-mode.html
--- 0.11+ds1-5/website/tty/location-mode.html	1970-01-01 00:00:00.000000000 +0000
+++ 0.12+ds1-1/website/tty/location-mode.html	2018-07-11 02:32:25.000000000 +0000
@@ -0,0 +1,15 @@
+~&gt;                                           <span class="sgr-7">xiaq@xiaqsmbp</span>
+<span class="sgr-1 sgr-37 sgr-45"> LOCATION </span><span class="sgr-"> </span>
+<span class="sgr-7">  * ~                                                    </span><span class="sgr-35 sgr-7"> </span>
+  * ~/go/src/github.com/elves/elvish                     <span class="sgr-35">│</span>
+110 ~/on/elvish-site/code                                <span class="sgr-35">│</span>
+ 62 ~/on/elvish-site/code/src                            <span class="sgr-35">│</span>
+ 52 ~/go/src/github.com/elves/elvish/edit                <span class="sgr-35">│</span>
+ 34 ~/on/elvish-site/code/tty                            <span class="sgr-35">│</span>
+ 33 ~/on/elvish-site/code/assets                         <span class="sgr-35">│</span>
+ 32 ~/go/src/github.com/elves/elvish/eval                <span class="sgr-35">│</span>
+ 26 ~/on/chat-app/code                                   <span class="sgr-35">│</span>
+ 24 ~/on/elvish-site/code/dst                            <span class="sgr-35">│</span>
+ 20 ~/go/src/github.com/elves/md-highlighter             <span class="sgr-35">│</span>
+ 14 ~/on/chat-app/code/public                            <span class="sgr-35">│</span>
+ 13 ~/.elvish                                            <span class="sgr-35">│</span>
diff -pruN 0.11+ds1-5/website/tty/navigation-mode.html 0.12+ds1-1/website/tty/navigation-mode.html
--- 0.11+ds1-5/website/tty/navigation-mode.html	1970-01-01 00:00:00.000000000 +0000
+++ 0.12+ds1-1/website/tty/navigation-mode.html	2018-07-11 02:32:25.000000000 +0000
@@ -0,0 +1,15 @@
+~/go/src/github.com/elves/elvish&gt;            <span class="sgr-7">xiaq@xiaqsmbp</span>
+<span class="sgr-1 sgr-37 sgr-45"> NAVIGATING </span><span class="sgr-"> </span>
+<span class="sgr-01 sgr-34 sgr-7"> elvish         </span><span class="sgr-">  CONTRIBUTING.md </span><span class="sgr-35 sgr-7"> </span><span class="sgr-">  FROM golang:onbuild </span>
+<span class="sgr-01 sgr-34"> fix-for-0.7    </span><span class="sgr-"> </span><span class="sgr- sgr-7"> Dockerfile      </span><span class="sgr-35 sgr-7"> </span><span class="sgr-">                      </span>
+<span class="sgr-01 sgr-34"> images         </span><span class="sgr-">  Gopkg.lock      </span><span class="sgr-35 sgr-7"> </span>
+<span class="sgr-01 sgr-34"> md-highlighter </span><span class="sgr-">  Gopkg.toml      </span><span class="sgr-35 sgr-7"> </span>
+                  LICENSE         <span class="sgr-35 sgr-7"> </span>
+                  Makefile        <span class="sgr-35 sgr-7"> </span>
+                  README.md       <span class="sgr-35 sgr-7"> </span>
+                 <span class="sgr-01 sgr-34"> cover           </span><span class="sgr-35">│</span>
+                 <span class="sgr-01 sgr-34"> daemon          </span><span class="sgr-35">│</span>
+                 <span class="sgr-01 sgr-34"> edit            </span><span class="sgr-35">│</span>
+                 <span class="sgr-01 sgr-34"> errors          </span><span class="sgr-35">│</span>
+                 <span class="sgr-01 sgr-34"> eval            </span><span class="sgr-35">│</span>
+                 <span class="sgr-01 sgr-34"> getopt          </span><span class="sgr-35">│</span>
diff -pruN 0.11+ds1-5/website/tty/pipelines.html 0.12+ds1-1/website/tty/pipelines.html
--- 0.11+ds1-5/website/tty/pipelines.html	1970-01-01 00:00:00.000000000 +0000
+++ 0.12+ds1-1/website/tty/pipelines.html	2018-07-11 02:32:25.000000000 +0000
@@ -0,0 +1,15 @@
+~&gt; <span class="sgr-32">curl</span><span class="sgr-"> https://api.github.com/repos/elves/elvish/issues </span><span class="sgr-32">|</span>
+     <span class="sgr-32">from-json</span><span class="sgr-"> </span><span class="sgr-32">|</span><span class="sgr-"> </span><span class="sgr-32">each</span><span class="sgr-"> <span class="sgr-35">$explode~</span> </span><span class="sgr-32">|</span>
+     <span class="sgr-32">each</span><span class="sgr-"> </span><span class="sgr-1">[</span><span class="sgr-">issue</span><span class="sgr-1">]{</span><span class="sgr-"> </span><span class="sgr-32">echo</span><span class="sgr-"> </span><span class="sgr-35">$issue</span><span class="sgr-1">[</span><span class="sgr-">number</span><span class="sgr-1">]</span><span class="sgr-">: </span><span class="sgr-35">$issue</span><span class="sgr-1">[</span><span class="sgr-">title</span><span class="sgr-1">]</span><span class="sgr-"> </span><span class="sgr-1">}</span><span class="sgr-"> </span><span class="sgr-32">|</span>
+     <span class="sgr-32">head</span><span class="sgr-"> -n 11</span>
+366: Support searching file from elvish directly
+364: Ctrl-C in elvish kills Atom in background
+357: Asynchronous syntax highlighting
+356: In web backend, run commands with pty IO, not pipe
+354: Support multi-line prompts from byte output
+353: Completers should detect context in a top-down manner
+352: Quoted command names are highlighted randomly
+351: keep navigation mode open after command
+350: Raw mode requires two presses of ^V
+344: Elvish won't compile
+343: Possible to suppress job control messages?
diff -pruN 0.11+ds1-5/website/tty/unicode-prompts.html 0.12+ds1-1/website/tty/unicode-prompts.html
--- 0.11+ds1-5/website/tty/unicode-prompts.html	1970-01-01 00:00:00.000000000 +0000
+++ 0.12+ds1-1/website/tty/unicode-prompts.html	2018-07-11 02:32:25.000000000 +0000
@@ -0,0 +1 @@
+~❱ <span class="sgr-36"># Fancy unicode prompts!</span><span class="sgr-">                  </span><span class="sgr-7">xiaq✸xiaqsmbp</span>
\ No newline at end of file
